# Copyright 2012, 2013 Canonical Ltd.
# This software is licensed under the GNU Affero General Public License
# version 3 (see the file LICENSE).

from oath import hotp

from acceptance import base, pages

import sst.actions


class TestAddDevice(base.SSTTestCaseWithLogIn):

    def setUp(self):
        super(TestAddDevice, self).setUp()
        self.page = pages.AddNewAuthenticationDevice(open_page=True)

    def test_cancel_device_addition(self):
        your_authentication_devices = self.page.cancel()
        self.assertEqual(your_authentication_devices.get_devices(), [])

    def test_default_device(self):
        self.assertEqual(
            self.page.get_selected_device(),
            'type_google')

    def test_paper_device_enabled(self):
        self.assertTrue(
            self.page.is_paper_device_displayed())


class TestAddGenericDevice(base.SSTTestCaseWithLogIn):

    def setUp(self):
        super(TestAddGenericDevice, self).setUp()
        add_device_page = pages.AddNewAuthenticationDevice(open_page=True)
        self.page = add_device_page.add_generic_device()

    def test_open_add_generic_device_page(self):
        name, one_time_password = self.page.get_form_values()
        self.assertEqual(name, 'Authentication device')
        self.assertEqual(one_time_password, '')
        name_error, one_time_password_error = self.page.get_form_errors()
        self.assertEqual(name_error, None)
        self.assertEqual(one_time_password_error, None)

    def test_cancel_generic_device_addition(self):
        your_authentication_devices = self.page.cancel()
        self.assertEqual(your_authentication_devices.get_devices(), [])

    def test_add_generic_device_with_errors(self):
        # Test with default name and wrong password.
        add_device = self.page.add_device_with_errors(
            name=None, one_time_password='some invalid key')
        name_error, one_time_password_error = add_device.get_form_errors()
        self.assertEqual(name_error, None)
        self.assertEqual(
            one_time_password_error,
            'Please enter a 6-digit or 8-digit one-time password')
        # Test with missing name and wrong password.
        add_device = add_device.add_device_with_errors(
            name='', one_time_password=None)
        name_error, one_time_password_error = add_device.get_form_errors()
        self.assertEqual(name_error, 'This field is required.')
        self.assertEqual(
            one_time_password_error,
            'Please enter a 6-digit or 8-digit one-time password')
        # Test with missing name, correct password.
        aes_key = add_device.get_key()
        valid_otp = hotp(aes_key, 0)
        add_device = add_device.add_device_with_errors(
            name=None, one_time_password=valid_otp)
        name_error, one_time_password_error = add_device.get_form_errors()
        self.assertEqual(name_error, 'This field is required.')
        self.assertEqual(one_time_password_error, None)

    def test_add_generic_device(self):
        aes_key = self.page.get_key()
        one_time_password = hotp(aes_key, 0)
        your_authentication_devices = self.page.add_device(
            'Test generic device', one_time_password)
        self.assertEqual(
            your_authentication_devices.get_devices(), ['Test generic device'])

    def test_adding_second_device_requires_two_factor(self):
        aes_key = self.page.get_key()
        one_time_password = hotp(aes_key, 0)
        devices_page = self.page.add_device(
            'Test generic device', one_time_password)
        logout_landing_page = devices_page.log_out()
        login_page = logout_landing_page.click_log_in_or_create_account()
        account_page = login_page.log_in_to_site_recognized(self.user)
        devices_page = account_page.menu.go_to_authentication_devices()
        add_link = sst.actions.get_element_by_css(
            '*[data-qa-id="add_new_device"]')
        sst.actions.click_link(add_link)
        # On instantiation, the page is asserted to be open.
        pages.TwoFactorVerify(open_page=False)


class TestAddGoogleDevice(base.SSTTestCaseWithLogIn):

    def setUp(self):
        super(TestAddGoogleDevice, self).setUp()
        add_device_page = pages.AddNewAuthenticationDevice(open_page=True)
        self.page = add_device_page.add_google_device()

    def test_add_google_device(self):
        aes_key = self.page.get_key(self.user.email)
        one_time_password = hotp(aes_key, 0)
        your_authentication_devices = self.page.add_device(
            'Test Google device', one_time_password)
        self.assertEqual(
            your_authentication_devices.get_devices(), ['Test Google device'])


class TestAddYubikeyDevice(base.SSTTestCaseWithLogIn):

    def setUp(self):
        super(TestAddYubikeyDevice, self).setUp()
        add_device_page = pages.AddNewAuthenticationDevice(open_page=True)
        self.page = add_device_page.add_yubikey_device()

    def test_add_yubikey_device(self):
        self.page.assert_warning()
        aes_key = self.page.get_key()
        one_time_password = hotp(aes_key, 0)
        your_authentication_devices = self.page.add_device(
            'Test Yubikey device', one_time_password)
        self.assertEqual(
            your_authentication_devices.get_devices(), ['Test Yubikey device'])


class TestAddPaperDevice(base.SSTTestCaseWithLogIn):

    def setUp(self):
        super(TestAddPaperDevice, self).setUp()
        add_device_page = pages.AddNewAuthenticationDevice(open_page=True)
        self.page = add_device_page.add_paper_device()

    def test_add_paper_device(self):
        self.page.assert_codes_present()
        self.page.assert_print_button_visible()
        your_auth_devices = self.page.go_back_to_device_list()
        self.assertEqual(
            your_auth_devices.get_devices(), ['Printable Backup Codes'])
