# Copyright 2013 Canonical Ltd.
# This software is licensed under the GNU Affero General Public License
# version 3 (see the file LICENSE).

import uuid

import mock
import testscenarios

from ssoclient.v2 import errors
from u1testutils.sso import api, data

from acceptance import base, fixture_setup, pages
from acceptance.tests import devices


# The qa anchor of all pages is asserted when they are instantiated.


class UserRegisteredFixtureTestCase(base.IdentityProviderSSTTestCase):

    def test_user_registered(self):
        user = data.User.make_from_configuration(new_user=True)
        fixture = fixture_setup.UserRegistered(self.base_url, user)
        self.useFixture(fixture)
        # It will raise an error because the user is already registered.
        self.assertRaises(errors.AlreadyRegistered, self.useFixture, fixture)

    def test_create_new_account_api_error(self):
        with mock.patch.object(
                api.APIClient, 'create_new_account') as mock_create:
            mock_create.return_value = False
            fixture = fixture_setup.UserRegistered(
                self.base_url, user='not important')
            error = self.assertRaises(
                AssertionError, self.useFixture, fixture)
        self.assertEqual(
            str(error),
            'Failed to register the user using the Ubuntu SSO API.')


class UserLoggedInFixtureTestCase(
        testscenarios.TestWithScenarios, base.IdentityProviderSSTTestCase):

    def setUp(self):
        super(UserLoggedInFixtureTestCase, self).setUp()
        self.user = self._register_test_user()

    def _register_test_user(self):
        unique_id = str(uuid.uuid1())
        user = data.User.make_from_configuration(
            new_user=True, unique_id=unique_id)
        self.useFixture(fixture_setup.UserRegistered(self.base_url, user))
        return user

    def test_user_logged_in(self):
        fixture = fixture_setup.UserLoggedIn(self.user)
        self.useFixture(fixture)
        your_account_page = pages.YourAccount(open_page=False)
        self.assertEqual(
            your_account_page.get_user_name(), self.user.full_name)


class DevicesFixturesTestCase(base.SSTTestCaseWithLogIn):

    def test_add_one_device_then_open_device_list(self):
        device_name = 'test device'
        self.useFixture(
            devices.AuthenticationDevicesPageOpenedWithDevicesAdded(
                [device_name]))
        your_auth_devices_page = pages.YourAuthenticationDevices(
            open_page=False)
        self.assertEqual(your_auth_devices_page.get_devices(), [device_name])

    def test_add_many_devices_then_open_device_list(self):
        device_names = ['test device1', 'test device2', 'test device3']
        self.useFixture(
            devices.AuthenticationDevicesPageOpenedWithDevicesAdded(
                device_names))
        your_auth_devices_page = pages.YourAuthenticationDevices(
            open_page=False)
        self.assertEqual(your_auth_devices_page.get_devices(), device_names)

    def test_open_delete_device_page(self):
        self.useFixture(devices.DeleteDevicesPageOpened())
        pages.DeleteAuthenticationDevice(open_page=False)

    def test_open_rename_device_page(self):
        self.useFixture(devices.RenameDevicePageOpened())
        pages.RenameAuthenticationDevice(open_page=False)

    def test_open_two_factor_verification_page(self):
        device_name = 'test device1'
        self.useFixture(devices.TwoFactorPageOpenedWithTwoFactorRequired(
            [device_name], self.user))
        pages.TwoFactorVerify(open_page=False)
