#!/bin/bash

set -ue

function usage {
  cat  << EOF
Usage: $0 [-t TESTCASE] [-h|-f] [-p PORT]
Run SSO acceptance tests on a random port.
  -t [TESTCASE]   run using TESTCASE regex. Will not reset db
  -h              run headless (default)
  -f              run headful
  -p [PORT]       use a specific port (e.g. -p 8001)
  -c              force a clean db (if using -t option)
  -1              stop on the first failure (AKA "fail-fast")
EOF
}

# random port by default
PORT=$(./scripts/random-port.sh)
TESTCASE=
HEADLESS="-x"
CLEAN="false"
FAILFAST=

while getopts ":t:hfc1p:" opt; do
  case $opt in
    t) TESTCASE=$OPTARG ;;
    h) HEADLESS="-x" ;;
    f) HEADLESS="" ;;
    p) PORT=$OPTARG ;;
    c) CLEAN="true" ;;
    1) FAILFAST="--failfast" ;;
    \?)
      echo "Invalid option: -$OPTARG" >&2
      usage
      exit 1
      ;;
    :)
      echo "Option -$OPTARG requires an argument." >&2
      usage
      exit 1
      ;;
  esac
done

export PORT=$PORT
export SSO_HOSTNAME="0.0.0.0:$PORT"
export SSO_ROOT_URL="http://$SSO_HOSTNAME"
export SST_BASE_URL="http://$SSO_HOSTNAME"

function clean_up {
    # terminate sso server
    screen -X -S $SSO_SCREEN_NAME quit
    # terminate gunicorn
    kill `cat logs/gunicorn.pid`
    # terminate emailserver server
    screen -X -S emailserver quit
    exit
}

# only reset db if we're doing a full run or explicitly told to
if [ -z $TESTCASE ] || $CLEAN
then
  echo "Resetting db"
  # start clean
  make reset-db
  make manage ARGS="loaddata test"
  make manage ARGS="create_test_team"
  # add openid RP config
  # we need this to explicitly allow unverified logins
  make manage ARGS="add_openid_rp_config $SST_BASE_URL/consumer --allow-unverified --allowed-user-attribs=\"fullname,nickname,email,language,account_verified\""
fi

# handle exit cleanly (mainly Ctrl-C)
trap clean_up SIGHUP SIGINT SIGTERM

echo -n Starting SSO in background on port $PORT...
SSO_SCREEN_NAME="sso-"$PORT
screen -dmS $SSO_SCREEN_NAME make run ARGS="$SSO_HOSTNAME"
echo done

echo -n Starting localmail test server in background...
screen -dmS emailserver env/bin/python -c 'import localmail; localmail.run()'
echo done

echo -n "Waiting for SSO to start"
while [ $(netstat -lnt | grep ":$PORT" | grep -v grep | wc -l) != 1 ]
do
    echo -n .
    sleep 0.5
done
echo "done"

SST_ARGS="-s -r xml -q -c 1 $FAILFAST $HEADLESS"

if [ -n "$TESTCASE" ]; then
  SST_ARGS="$SST_ARGS \"$TESTCASE\""
fi
echo $SST_ARGS

make run-acceptance ARGS="$SST_ARGS"

clean_up $SSO_SCREEN_NAME
