# Copyright 2010-2012 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

from __future__ import unicode_literals

import logging
import json

from datetime import timedelta

from django.conf import settings
from django.core.exceptions import ValidationError
from django.core.serializers.json import DateTimeAwareJSONEncoder
from django.http import (
    HttpResponseBadRequest,
    HttpResponseForbidden,
)
from django.shortcuts import render
from django.utils.timezone import now
from django.utils.translation import ugettext_lazy as _
from oauth_backend.models import Token
from piston.emitters import Emitter
from piston.handler import BaseHandler

from api.v10.decorators import (
    api_user_required,
    check_leak,
    plain_user_required,
    named_operation,
)
from api.v10.forms import WebserviceCreateAccountForm
from identityprovider import emailutils
from identityprovider.models import (
    Account,
    AuthToken,
    EmailAddress,
    get_team_memberships_for_user,
)
from identityprovider.models.captcha import (
    Captcha,
    NewCaptchaError,
    VerifyCaptchaError,
)
from identityprovider.models.const import AuthTokenType, EmailStatus
from identityprovider.signals import (
    account_created,
    account_email_validated,
    application_token_created,
    application_token_invalidated,
)
from identityprovider.timeline_helpers import get_request_timing_function
from identityprovider.validators import PASSWORD_LEAKED


logger = logging.getLogger(__name__)


def api_error(cls, msg):
    """Ensure 4xx response content in plain text to avoid XSS"""
    response = cls(msg, content_type='text/plain')
    # Add back bit expected by django-piston but removed from django 1.5
    response._is_string = True
    return response


class LazrRestfulEmitter(Emitter):
    """JSON emitter, lazr.restful flavoured.

    Reports content-type 'application/json', without specifying a charset
    as that seems to confuse lazr.restfulclient.
    """
    def render(self, request):
        # ensure_ascii=False was causing bug #820216
        # it looks like this is because simplejson's implementation is
        # broken
        # http://code.google.com/p/simplejson/issues/detail?id=48
        # https://code.djangoproject.com/ticket/6234
        seria = json.dumps(self.construct(),
                           cls=DateTimeAwareJSONEncoder, indent=4)

        return seria
Emitter.register('lazr.restful', LazrRestfulEmitter, 'application/json')


class LazrRestfulHandler(BaseHandler):
    allowed_methods = ('GET', 'POST')

    response = {}

    base_url = settings.SSO_ROOT_URL.strip('/')

    def __init__(self):
        self.response = self.response.copy()
        for key, value in self.response.items():
            if key.endswith('_link'):
                self.response[key] = value % self.base_url

    def read(self, request):
        if 'ws.op' not in request.GET:
            return self.response
        return self.named_operation(request, request.GET)

    def create(self, request):
        if 'ws.op' not in request.POST:
            return api_error(HttpResponseBadRequest,
                             'No operation name given.')
        return self.named_operation(request, request.POST)

    def named_operation(self, request, serialized):
        method_name = serialized['ws.op']
        method = getattr(self, method_name, None)
        is_named_operation = getattr(method, 'is_named_operation', False)
        if not is_named_operation:
            return api_error(HttpResponseBadRequest,
                             'No such operation: %s' % method_name)
        request.data = self.lazr_restful_deserialize(serialized)
        return method(request)

    def lazr_restful_deserialize(self, serialized):
        data = {}
        for key in serialized:
            if key == 'ws.op':
                continue
            try:
                value = json.loads(serialized[key])
            except (ValueError, TypeError):
                value = serialized[key]
            data[key] = value
        return data


class RootHandler(LazrRestfulHandler):
    allowed_methods = ('GET',)

    response = {
        "registrations_collection_link": "%s/api/1.0/registration",
        "captchas_collection_link": "%s/api/1.0/captchas",
        "validations_collection_link": "%s/api/1.0/validation",
        "authentications_collection_link": "%s/api/1.0/authentications",
        "resource_type_link": "%s/api/1.0/#service-root",
        "accounts_collection_link": "%s/api/1.0/accounts",
    }

    wadl_template = 'api/wadl1.0.xml'

    def read(self, request):
        if ('application/vd.sun.wadl+xml' in request.META['HTTP_ACCEPT'] or
                'application/vnd.sun.wadl+xml' in request.META['HTTP_ACCEPT']):
            context = {'baseurl': self.base_url}
            response = render(request, self.wadl_template, context)
            response._is_string = True
            return response
        else:
            return self.response


class CaptchaHandler(LazrRestfulHandler):
    response = {
        "total_size": 0,
        "start": None,
        "resource_type_link": "%s/api/1.0/#captchas",
        "entries": []
    }

    @named_operation
    def new(self, request):
        timer = get_request_timing_function(request)
        try:
            return Captcha.new(timer=timer).serialize()
        except NewCaptchaError as e:
            logger.exception("Failed to connect to reCaptcha server")
            # TODO: Some better return here?
            return e.dummy.serialize()


class RegistrationHandler(LazrRestfulHandler):
    response = {
        "total_size": 0,
        "start": None,
        "resource_type_link": "%s/api/1.0/#registrations",
        "entries": []
    }

    @named_operation
    def register(self, request):
        data = request.data
        data['remote_ip'] = request.environ['REMOTE_ADDR']
        form = WebserviceCreateAccountForm(data)
        try:
            if not form.is_valid():
                errors = dict((k, map(unicode, v))
                              for (k, v) in form.errors.items())
                # XXX: cope with client trimming error messages
                if unicode(PASSWORD_LEAKED) in errors.get('password', ''):
                    errors['password'] = unicode(_(
                        'unsafe: leaked by security breach '
                        'on another website'))
                result = {'status': 'error', 'errors': errors}
                return result
        except VerifyCaptchaError:
            logger.exception("reCaptcha connection error")
            msg = unicode(
                _('Unable to verify captcha. Please try again shortly.'))
            return {
                'status': 'error',
                'errors': {'captcha_solution': [msg]}
            }

        cleaned_data = form.cleaned_data
        requested_email = cleaned_data['email']
        emails = EmailAddress.objects.filter(email__iexact=requested_email)
        if len(emails) > 0:
            return {'status': 'error',
                    'errors': {'email': ['Email already registered']}}

        platform = cleaned_data['platform']
        redirection_url = cleaned_data['validate_redirect_to']
        password = cleaned_data['password']
        email = cleaned_data['email']
        displayname = cleaned_data['displayname']
        if not displayname:
            displayname = email.split('@', 1)[0]

        account = Account.objects.create_account(
            displayname, email, password, email_validated=False)
        root_url = request.build_absolute_uri('/')
        emailutils.send_new_user_email(
            root_url, account=account, email=email,
            redirection_url=redirection_url, platform=platform)
        account_created.send(
            sender=self, openid_identifier=account.openid_identifier)

        return {
            'status': 'ok',
            'message': "Email verification required."
        }

    @named_operation
    def request_password_reset_token(self, request):
        data = request.data
        email = data['email']

        account = Account.objects.get_by_email(email)
        if account is None or not account.can_reset_password:
            return api_error(HttpResponseForbidden,
                             "CanNotResetPasswordError: Can't reset password "
                             "for this account.")

        emailutils.send_password_reset_email(account, email)

        return {
            'status': 'ok',
            'message': "Password reset token sent."
        }

    @named_operation
    def set_new_password(self, request):
        data = request.data
        new_password = data['new_password']
        try:
            token = AuthToken.get_token(
                email=data['email'], raw_token=data['token'],
                token_type=AuthTokenType.PASSWORDRECOVERY, date_consumed=None)
        except AuthToken.DoesNotExist:
            return api_error(HttpResponseForbidden,
                             "CanNotSetPasswordError: Invalid token.")
        if not token.requester:
            token.delete()
            return {
                'status': 'error',
                'message': "Wrong token, request new one."
            }

        try:
            token.requester.set_password(new_password)
        except ValidationError as e:
            return {
                'status': 'error',
                'errors': e.messages,
            }
        token.consume()

        return {
            'status': 'ok',
            'message': "Password changed"
        }


def _serialize_account(user):
    emails = EmailAddress.objects.filter(account=user,
                                         status=EmailStatus.VALIDATED)
    preferred_email = user.preferredemail
    if preferred_email is not None:
        preferred_email = preferred_email.email

    if user.person:
        username = user.person_name
    else:
        username = user.openid_identifier

    return {
        'username': username,
        'displayname': user.displayname,
        'openid_identifier': user.openid_identifier,
        'preferred_email': preferred_email,
        'verified_emails': [e.email for e in emails],
        'unverified_emails': [e.email for e in user.unverified_emails()],
    }


class AuthenticationHandler(LazrRestfulHandler):
    """All these methods assume that they're run behind Basic Auth."""
    response = {
        "total_size": 0,
        "start": None,
        "resource_type_link": "%s/api/1.0/#authentications",
        "entries": []
    }

    @plain_user_required
    @named_operation
    def authenticate(self, request):
        data = request.data
        account = request.user
        token = account.create_v1_oauth_token(data['token_name'])
        application_token_created.send(
            sender=self, openid_identifier=account.openid_identifier)
        return token.serialize()

    @api_user_required
    @named_operation
    def list_tokens(self, request):
        data = request.data
        tokens = Token.objects.filter(
            consumer__user__username=data['consumer_key'])
        result = [{'token': t.token, 'name': t.name} for t in tokens]
        return result

    @api_user_required
    @named_operation
    def validate_token(self, request):
        data = request.data
        token = data.get('token')
        openid = data.get('consumer_key')
        max_age = data.get('max_age')

        try:
            token = Token.objects.get(
                consumer__user__username=openid, token=token)
        except Token.DoesNotExist:
            return False

        if max_age is not None:
            age = now() - token.updated_at
            if age > timedelta(seconds=max_age):
                # token is too old
                return False
        return token.serialize()

    @api_user_required
    @named_operation
    def invalidate_token(self, request):
        data = request.data
        tokens = Token.objects.filter(
            token=data['token'], consumer__user__username=data['consumer_key'])
        tokens.delete()
        application_token_invalidated.send(
            sender=self, openid_identifier=data['consumer_key'])

    @api_user_required
    @named_operation
    def team_memberships(self, request):
        data = request.data
        accounts = Account.objects.filter(
            openid_identifier=data['openid_identifier'])
        accounts = list(accounts)

        if len(accounts) == 1:
            account = accounts[0]
            memberships = (
                get_team_memberships_for_user(
                    data['team_names'], account, include_private=True))
            return memberships
        else:
            return []

    @api_user_required
    @named_operation
    def account_by_email(self, request):
        data = request.data
        account = Account.objects.get_by_email(data['email'])
        if account:
            return _serialize_account(account)
        else:
            return None

    @api_user_required
    @named_operation
    def account_by_openid(self, request):
        data = request.data
        try:
            account = Account.objects.get(openid_identifier=data['openid'])
        except Account.DoesNotExist:
            return None
        else:
            return _serialize_account(account)


class AccountsHandler(LazrRestfulHandler):
    @check_leak
    @named_operation
    def me(self, request):
        account = request.user
        return _serialize_account(account)

    @check_leak
    @named_operation
    def team_memberships(self, request):
        team_names = request.data['team_names']

        memberships = (
            get_team_memberships_for_user(
                team_names, request.user, include_private=True))
        return memberships

    @check_leak
    @named_operation
    def validate_email(self, request):
        email_token = request.data['email_token']
        try:
            # Additional filtering is done to check that the token has no
            # consumption date and that the email that's being validated
            # belongs to the user who requested the token.
            user_mails_qs = request.user.emailaddress_set.values('email')
            token = AuthToken.get_token(
                requester=request.user,
                token_type=AuthTokenType.VALIDATEEMAIL,
                raw_token=email_token, date_consumed=None,
                email__in=user_mails_qs)

            email = EmailAddress.objects.get(email__iexact=token.email)
            email.validate()

            token.consume()

            account_email_validated.send(
                openid_identifier=request.user.openid_identifier,
                sender=self)
            return {'email': email.email}
        except AuthToken.DoesNotExist:
            return {'errors': {'email_token': ["Bad email token!"]}}
