# encoding: utf-8

from __future__ import unicode_literals

import json

from datetime import datetime, timedelta

from django.core import mail
from django.core.urlresolvers import reverse
from django.test.utils import override_settings
from django.utils.http import urlencode
from django.utils.timezone import now
from mock import Mock, patch
from lazr.restfulclient.errors import HTTPError
from timeline import Timeline

from api.v10.handlers import (
    AuthenticationHandler,
    CaptchaHandler,
    LazrRestfulHandler,
)
from api.v10.tests.utils import (
    AnonAPITestCase,
    http_authorization_extra,
)
from identityprovider.models import (
    Account,
    AuthToken,
    APIUser,
    EmailAddress,
)
from identityprovider.models.captcha import Captcha
from identityprovider.models.const import (
    AccountStatus,
    AuthTokenType,
    EmailStatus,
)
from identityprovider.tests import (
    DEFAULT_API_PASSWORD,
    DEFAULT_USER_PASSWORD,
)
from identityprovider.tests.utils import SSOBaseTestCase
from identityprovider.validators import PASSWORD_LEAKED, PASSWORD_POLICY_ERROR


class TestLazrRestfulHandler(SSOBaseTestCase):
    handler = LazrRestfulHandler()

    def test_named_operation_error_is_plaintext(self):
        script = '<script>alert("boo");</script>'
        response = self.handler.named_operation(None, {'ws.op': script})
        self.assertEqual(response['content-type'], 'text/plain')

    def test_create_error_is_plaintext(self):
        request = self.factory.make_request(method='POST')
        response = self.handler.create(request)
        self.assertEqual(response['content-type'], 'text/plain')


class RegistrationHandlerTestCase(AnonAPITestCase):

    def test_authtoken_without_requester_is_handled_properly(self):
        authtoken = self.factory.make_authtoken(
            email="test@example.com",
            token_type=AuthTokenType.PASSWORDRECOVERY)
        assert authtoken.requester is None

        response = self.api.registrations.set_new_password(
            email='test@example.com', token=authtoken.raw_token,
            new_password='pass')
        self.assertEqual(response['status'], "error")

    def test_set_new_password_ok(self):
        email_address = self.factory.make_email_address()
        account = self.factory.make_account(email=email_address)
        account.create_v1_oauth_token('some-token')
        authtoken = self.factory.make_authtoken(
            requester=account, requester_email=email_address,
            email=email_address, token_type=AuthTokenType.PASSWORDRECOVERY)

        response = self.api.registrations.set_new_password(
            email=email_address, token=authtoken.raw_token,
            new_password='testing-password')
        self.assertEqual(response['status'], "ok")
        self.assertEqual(account.v1_oauth_tokens().count(), 0)

    def test_set_new_password_invalid(self):
        email_address = self.factory.make_email_address()
        account = self.factory.make_account(email=email_address)
        authtoken = self.factory.make_authtoken(
            requester=account, requester_email=email_address,
            email=email_address, token_type=AuthTokenType.PASSWORDRECOVERY)

        response = self.api.registrations.set_new_password(
            email=email_address, token=authtoken.raw_token,
            new_password='pass')
        self.assertEqual(response['status'], "error")
        self.assertEqual(
            response['errors'], [PASSWORD_POLICY_ERROR.format(min_length=8)])

    def test_set_new_password_with_leaked_creds(self):
        email_address = self.factory.make_email_address()
        account = self.factory.make_account(email=email_address)
        authtoken = self.factory.make_authtoken(
            requester=account, requester_email=email_address,
            email=email_address, token_type=AuthTokenType.PASSWORDRECOVERY)
        self.factory.make_leaked_credential(email=email_address,
                                            password=DEFAULT_USER_PASSWORD)

        response = self.api.registrations.set_new_password(
            email=email_address, token=authtoken.raw_token,
            new_password=DEFAULT_USER_PASSWORD)
        self.assertEqual(response['status'], "error")
        self.assertEqual(response['errors'], [unicode(PASSWORD_LEAKED)])

    def test_reset_token_when_email_is_invalid(self):
        with self.assertRaises(HTTPError) as ctx:
            self.api.registrations.request_password_reset_token(
                email='non-existing@example.com')

        self.assertEqual(ctx.exception.response.status, 403)
        self.assertEqual(ctx.exception.response['content-type'], "text/plain")
        self.assertTrue(
            ctx.exception.content.startswith("CanNotResetPasswordError:"))

    def test_reset_token_when_account_is_disabled(self):
        email_address = self.factory.make_email_address()
        self.factory.make_account(email=email_address,
                                  status=AccountStatus.SUSPENDED)
        with self.assertRaises(HTTPError) as ctx:
            self.api.registrations.request_password_reset_token(
                email=email_address)

        self.assertEqual(ctx.exception.response.status, 403)
        self.assertEqual(ctx.exception.response['content-type'], "text/plain")
        self.assertTrue(
            ctx.exception.content.startswith("CanNotResetPasswordError:"))

    def test_reset_password_when_not_existing_token_is_passed(self):
        email_address = self.factory.make_email_address()
        with self.assertRaises(HTTPError) as ctx:
            self.api.registrations.set_new_password(
                email=email_address, token='not-existing-token',
                new_password='password')

        self.assertEqual(ctx.exception.response.status, 403)
        self.assertEqual(ctx.exception.response['content-type'], "text/plain")
        self.assertEqual(
            ctx.exception.content, "CanNotSetPasswordError: Invalid token.")

    def test_register_email_error_is_in_list(self):
        self.factory.make_account(email='register@example.com',
                                  email_validated=False)

        response = self.api.registrations.register(
            email='register@example.com', password='blogdf3Daa',
            captcha_solution='solution', captcha_id='id')

        self.assertEqual(response['errors'],
                         {'email': ["Email already registered"]})
        self.assertEqual(len(mail.outbox), 0)

    def test_register_without_displayname(self):
        email = 'pepe@foo.com'

        response = self.api.registrations.register(
            email=email, password='MySecretPassword1',
            captcha_solution='foobar', captcha_id='id')
        self.assertEqual(response['status'], 'ok')

        account = Account.objects.get_by_email(email)
        self.assertEqual(account.displayname, 'pepe')
        self.assert_new_user_email_sent(email)

    def assert_new_user_email_sent(self, email_address, platform='desktop'):
        self.assertEqual(len(mail.outbox), 1)
        email = mail.outbox[0]
        self.assertIn(email_address, email.to)
        self.assertIn('Finish your registration', email.subject)
        if platform == 'desktop':
            self.assertIn('desktop', email.body)
        else:
            self.assertIn('Welcome', email.body)

    def test_register_with_leaked_creds(self):
        email = 'pepe@foo.com'
        password = 'MySecretPassword1'
        self.factory.make_leaked_credential(email=email, password=password)

        response = self.api.registrations.register(
            email=email, password=password,
            captcha_solution='foobar', captcha_id='id')
        self.assertEqual(response['status'], 'error')
        self.assertEqual(
            response['errors'],
            {'password': 'unsafe: leaked by security breach on '
                         'another website'})

        self.assertIsNone(Account.objects.get_by_email(email))

    def test_register_with_displayname(self):
        email = self.factory.make_email_address()

        response = self.api.registrations.register(
            email=email, password='MySecretPassword1',
            captcha_solution='foobar', captcha_id='id',
            displayname='Test User')
        self.assertEqual(response['status'], 'ok')

        account = Account.objects.get_by_email(email)
        self.assertEqual(account.displayname, 'Test User')
        self.assert_new_user_email_sent(email)

    def test_register_with_lowered_password_requirements(self):
        email = self.factory.make_email_address()

        # password: 8 characters long (no uppercase or numbers, only lowercase)
        response = self.api.registrations.register(
            email=email, password='abcdefgh',
            captcha_solution='foobar', captcha_id='id',
            displayname='Test User')
        self.assertEqual(response['status'], 'ok')

        account = Account.objects.get_by_email(email)
        self.assertEqual(account.displayname, 'Test User')
        self.assert_new_user_email_sent(email)

    def test_register_without_platform(self):
        email = self.factory.make_email_address()

        response = self.api.registrations.register(
            email=email, password='MySecretPassword1',
            captcha_solution='foobar', captcha_id='id')
        self.assertEqual(response['status'], 'ok')
        self.assert_new_user_email_sent(email)

    def test_register_with_platform(self):
        email = self.factory.make_email_address()

        response = self.api.registrations.register(
            email=email, password='MySecretPassword1',
            captcha_solution='foobar', captcha_id='id',
            platform='mobile')
        self.assertEqual(response['status'], 'ok')
        self.assert_new_user_email_sent(email, 'mobile')

    def test_register_with_invalid_platform(self):
        email = self.factory.make_email_address()

        response = self.api.registrations.register(
            email=email, password='MySecretPassword1',
            captcha_solution='foobar', captcha_id='id',
            platform='foo')

        self.assertEqual(response['status'], 'error')
        self.assertTrue('platform' in response['errors'])
        self.assertEqual(len(mail.outbox), 0)

    def test_register_without_redirection_url(self):
        email_address = self.factory.make_email_address()

        response = self.api.registrations.register(
            email=email_address, password='MySecretPassword1',
            captcha_solution='foobar', captcha_id='id',
            platform='desktop')

        self.assertEqual(response['status'], 'ok')
        self.assert_new_user_email_sent(email_address)

    def test_register_from_desktop_with_redirection_url(self):
        email_address = self.factory.make_email_address()

        response = self.api.registrations.register(
            email=email_address, password='MySecretPassword1',
            captcha_solution='foobar', captcha_id='id',
            platform='desktop', validate_redirect_to='http://foo')

        self.assertEqual(response['status'], 'ok')
        self.assert_new_user_email_sent(email_address)
        token = AuthToken.objects.get(token_type=AuthTokenType.VALIDATEEMAIL)
        self.assertEqual(token.redirection_url, 'http://foo')

    def test_register_from_mobile_with_redirection_url(self):
        email_address = self.factory.make_email_address()

        response = self.api.registrations.register(
            email=email_address, password='MySecretPassword1',
            captcha_solution='foobar', captcha_id='id',
            platform='mobile', validate_redirect_to='http://foo')

        self.assertEqual(response['status'], 'ok')
        account = Account.objects.get_by_email(email_address)
        unverified = account.unverified_emails()
        self.assertEqual([email_address],
                         [emailaddress.email for emailaddress in unverified])

    def test_register_from_mobile_without_redirection_url(self):
        email_address = self.factory.make_email_address()

        response = self.api.registrations.register(
            email=email_address, password='MySecretPassword1',
            captcha_solution='foobar', captcha_id='id',
            platform='mobile')

        self.assertEqual(response['status'], 'ok')
        account = Account.objects.get_by_email(email_address)

        unverified = account.unverified_emails()
        self.assertEqual([email_address],
                         [emailaddress.email for emailaddress in unverified])

    def test_set_new_password_consumed_token_fails(self):
        email_address = self.factory.make_email_address()
        account = self.factory.make_account(email=email_address)
        account.create_v1_oauth_token('some-token')
        authtoken = self.factory.make_authtoken(
            requester=account, requester_email=email_address,
            email=email_address, token_type=AuthTokenType.PASSWORDRECOVERY)
        # Mark the token as consumed...
        authtoken.date_consumed = now()
        authtoken.save()

        with self.assertRaises(HTTPError) as ctx:
            self.api.registrations.set_new_password(
                email=email_address, token=authtoken.raw_token,
                new_password='testing-password')
        # Oauth token not reset
        self.assertEqual(account.v1_oauth_tokens().count(), 1)
        # 403 response with "Invalid token" expected
        self.assertEqual(ctx.exception.response.status, 403)
        self.assertEqual(ctx.exception.response['content-type'], "text/plain")
        self.assertTrue(ctx.exception.content.startswith(
            "CanNotSetPasswordError: Invalid token"))

    def test_error_raises(self):
        self.patch(
            'api.v10.handlers.Account.objects.create_account',
            side_effect=Exception())
        email_address = self.factory.make_email_address()

        with self.assertRaises(Exception):
            self.client.post(
                '/api/1.0/registration',
                {'ws.op': 'register',
                 'email': email_address,
                 'password': 'MySecretPassword1',
                 'captcha_solution': 'foobar',
                 'captcha_id': 'id'})


class AuthenticationTestCase(SSOBaseTestCase):

    def setUp(self):
        super(AuthenticationTestCase, self).setUp()
        self.authentication = AuthenticationHandler()
        self.email_address = self.factory.make_email_address()
        self.account = self.factory.make_account(email=self.email_address)
        self.apiuser = self.factory.make_apiuser()

    def put(self, url, data, **extra):
        return self.client.put(
            url,
            urlencode(data),
            content_type='application/x-www-form-urlencoded',
            **extra
        )

    def authenticate(self, token_name='', token=None, method='GET',
                     api_version='1.0', parse_content=True):
        data = {
            'ws.op': 'authenticate',
            'token_name': token_name,
        }
        if token is not None:
            data['token'] = token

        extra = http_authorization_extra(self.email_address,
                                         password=DEFAULT_USER_PASSWORD)
        if method == 'PUT':
            data = urlencode(data)
            extra['content_type'] = 'application/x-www-form-urlencoded'
        dispatcher = getattr(self.client, method.lower())
        response = dispatcher("/api/%s/authentications" % api_version, data,
                              **extra)
        if parse_content:
            response = json.loads(response.content)
        return response

    def validate_token(self, token, openid, max_age=None):
        data = {
            'ws.op': 'validate_token',
            'token': token,
            'consumer_key': openid,
        }
        if max_age is not None:
            data['max_age'] = max_age
        extra = http_authorization_extra(self.apiuser.username,
                                         password=DEFAULT_API_PASSWORD)
        response = self.client.get('/api/1.0/authentications', data, **extra)
        content = json.loads(response.content)
        return content

    def test_account_by_openid_with_valid_openid(self):
        account = self.factory.make_account()
        openid = account.openid_identifier

        request = self.factory.make_request(user=APIUser())
        request.data = {'openid': openid}

        account = self.authentication.account_by_openid(request)

        self.assertEqual(account['openid_identifier'], openid)

    def test_account_by_openid_with_invalid_openid(self):
        request = self.factory.make_request(user=APIUser())
        request.data = {'openid': "bad-openid"}

        account = self.authentication.account_by_openid(request)

        self.assertTrue(account is None)

    def test_authenticate_unquoted_token(self):
        expected_keys = ['consumer_key', 'consumer_secret', 'name', 'token',
                         'token_secret', 'created', 'updated']

        content = self.authenticate(token_name='some-token')
        self.assertEqual(set(content.keys()), set(expected_keys))
        self.assertEqual(content['name'], 'some-token')

    def test_authenticate_quoted_token(self):
        expected_keys = ['consumer_key', 'consumer_secret', 'name', 'token',
                         'token_secret', 'created', 'updated']

        content = self.authenticate(token_name='"some-token"')
        self.assertEqual(set(content.keys()), set(expected_keys))
        self.assertEqual(content['name'], 'some-token')

    def test_authenticate_latin1_username(self):
        username = 'Franço'
        data = {
            'ws.op': 'authenticate',
            'token_name': 'some-token',
        }

        extra = http_authorization_extra(username.encode('latin1'),
                                         password=DEFAULT_USER_PASSWORD)
        response = self.client.get('/api/1.0/authentications', data, **extra)
        self.assertEqual(response.status_code, 401)

    def test_authenticate_with_leaked_creds(self):
        self.account.need_password_reset = True
        self.account.save()

        response = self.authenticate(token_name='foo', parse_content=False)
        self.assertEqual(response.status_code, 403)

    def test_team_memberships_public_teams(self):
        team = self.factory.make_team()
        otherteam = self.factory.make_team()
        self.factory.make_apiuser(username='foobar')
        account = self.factory.make_account()
        self.factory.add_account_to_team(account, team)

        # Request membership via API
        extra = http_authorization_extra('foobar',
                                         password=DEFAULT_API_PASSWORD)
        data = {
            'ws.op': 'team_memberships',
            'team_names': '["%s", "%s"]' % (team.name, otherteam.name),
            'openid_identifier': account.openid_identifier,
        }
        response = self.client.get('/api/1.0/authentications', data, **extra)
        self.assertContains(response, team.name)
        self.assertNotContains(response, otherteam.name)

    def test_team_membership_private_teams(self):
        team = self.factory.make_team(private=True)
        self.factory.make_apiuser(username='foobar')
        account = self.factory.make_account()
        self.factory.add_account_to_team(account, team)

        # Request membership via API
        extra = http_authorization_extra('foobar',
                                         password=DEFAULT_API_PASSWORD)
        data = {
            'ws.op': 'team_memberships',
            'team_names': '["%s"]' % team.name,
            'openid_identifier': account.openid_identifier,
        }
        response = self.client.get('/api/1.0/authentications', data, **extra)
        self.assertContains(response, team.name)

    def test_team_membership_basic_auth_protected(self):
        data = {
            'ws.op': 'team_memberships',
            'team_names': '["something"]',
            'openid_identifier': 'test',
        }
        response = self.client.get('/api/1.0/authentications', data)
        self.assertEqual(401, response.status_code)

    def test_team_membership_fails_for_regular_user(self):
        email_address = self.factory.make_email_address()
        self.factory.make_account(email=email_address)
        extra = http_authorization_extra(email_address,
                                         password=DEFAULT_USER_PASSWORD)
        data = {
            'ws.op': 'team_memberships',
            'team_names': '["something"]',
            'openid_identifier': 'test',
        }
        response = self.client.get('/api/1.0/authentications', data, **extra)
        self.assertEqual(403, response.status_code)

    def test_validate_unknown_token(self):
        openid = self.account.openid_identifier

        response = self.validate_token('foo', openid)
        self.assertFalse(response)

    def test_validate_token(self):
        openid = self.account.openid_identifier
        token = self.authenticate(token_name='foo')

        response = self.validate_token(token['token'], openid)
        self.assertEqual(response, token)

    def test_validate_fresh_token(self):
        openid = self.account.openid_identifier
        token = self.authenticate(token_name='foo')

        response = self.validate_token(token['token'], openid, max_age=900)
        self.assertEqual(response, token)

    def test_validate_stale_token(self):
        openid = self.account.openid_identifier
        token = self.authenticate(token_name='foo')

        updated = datetime.strptime(token['updated'], '%Y-%m-%d %H:%M:%S.%f')
        # 5 seconds have passed since we created the token
        time_at_call = updated + timedelta(seconds=5)
        with patch('api.v10.handlers.now') as mock_now:
            mock_now.return_value = time_at_call
            response = self.validate_token(token['token'], openid, max_age=4)

        self.assertFalse(response)


class AccountsTestCase(SSOBaseTestCase):

    url = reverse('api-10-accounts')

    def auth_header(self, token, url=None, parameters=None):
        if url is None:
            url = self.url
        return self.factory.make_authorization_header_from_token(
            url, token, http_method='GET', parameters=parameters)

    def test_me_is_oauth_protected(self):
        email_address = self.factory.make_email_address()
        extra = http_authorization_extra(email_address,
                                         password=DEFAULT_USER_PASSWORD)
        data = {'ws.op': 'me'}
        response = self.client.get(self.url, data, **extra)
        self.assertEqual(401, response.status_code)

    def test_me_success(self):
        name = 'Ñânḑù'
        account = self.factory.make_account(displayname=name)
        token = account.create_v1_oauth_token('token-name')

        data = {'ws.op': 'me'}
        response = self.client.get(
            self.url, data, **self.auth_header(token, parameters=data))

        data = json.loads(unicode(response.content))
        expected = ['username', 'preferred_email', 'displayname',
                    'unverified_emails', 'verified_emails',
                    'openid_identifier']
        self.assertEqual(set(expected), set(data.keys()))
        self.assertEqual(data['displayname'], name)

    def test_me_with_leaked_creds(self):
        account = self.factory.make_account()
        account.need_password_reset = True
        account.save()
        token = account.create_v1_oauth_token('token-name')

        data = {'ws.op': 'me'}
        response = self.client.get(
            self.url, data, **self.auth_header(token, parameters=data))

        self.assertEqual(response.status_code, 403)

    def test_team_membership_public_teams(self):
        account = self.factory.make_account()
        team = self.factory.make_team()
        otherteam = self.factory.make_team()
        self.factory.add_account_to_team(account, team)
        token = account.create_v1_oauth_token('token-name')

        data = {'ws.op': 'team_memberships',
                'team_names': '["%s", "%s"]' % (team.name, otherteam.name)}
        response = self.client.get(
            self.url, data, **self.auth_header(token, parameters=data))

        self.assertContains(response, team.name)
        self.assertNotContains(response, otherteam.name)

    def test_team_membership_private_teams(self):
        account = self.factory.make_account()
        team = self.factory.make_team(private=True)
        self.factory.add_account_to_team(account, team)
        token = account.create_v1_oauth_token('token-name')

        data = {'ws.op': 'team_memberships',
                'team_names': '["%s"]' % team.name}
        response = self.client.get(
            self.url, data, **self.auth_header(token, parameters=data))

        self.assertContains(response, team.name)

    def test_team_memberships_with_leak_creds(self):
        account = self.factory.make_account()
        account.need_password_reset = True
        account.save()
        team = self.factory.make_team(private=True)
        self.factory.add_account_to_team(account, team)
        token = account.create_v1_oauth_token('token-name')

        data = {'ws.op': 'team_memberships',
                'team_names': '["%s"]' % team.name}
        response = self.client.get(
            self.url, data, **self.auth_header(token, parameters=data))

        self.assertEqual(response.status_code, 403)

    def test_validate_email_bad_token(self):
        account = self.factory.make_account()
        token = account.create_v1_oauth_token('token-name')

        data = {'ws.op': 'validate_email', 'email_token': 'ABCD'}
        response = self.client.get(
            self.url, data, **self.auth_header(token, parameters=data))

        self.assertContains(response, 'Bad email token!')

    def test_validate_email_success(self):
        account = self.factory.make_account()
        email = self.factory.make_email_for_account(
            account, status=EmailStatus.NEW)
        authtoken = self.factory.make_authtoken(
            requester=account, email=email.email,
            token_type=AuthTokenType.VALIDATEEMAIL)

        token = account.create_v1_oauth_token('token-name')

        data = {'ws.op': 'validate_email', 'email_token': authtoken.raw_token}
        response = self.client.get(
            self.url, data, **self.auth_header(token, parameters=data))

        self.assertContains(response, email.email)
        updatedemail = EmailAddress.objects.get(email=email.email)
        self.assertEqual(EmailStatus.VALIDATED, updatedemail.status)

    def test_validate_email_consumed_token_fail(self):
        account = self.factory.make_account()
        email = self.factory.make_email_for_account(
            account, status=EmailStatus.NEW)
        authtoken = self.factory.make_authtoken(
            requester=account, email=email.email,
            token_type=AuthTokenType.VALIDATEEMAIL)
        # Mark the token as consumed
        authtoken.date_consumed = now()
        authtoken.save()

        token = account.create_v1_oauth_token('token-name')

        data = {'ws.op': 'validate_email', 'email_token': authtoken.raw_token}
        response = self.client.get(
            self.url, data, **self.auth_header(token, parameters=data))
        # Check error message is in response and EmailAddress status wasn't
        # changed.
        self.assertContains(response, "Bad email token!")
        updatedemail = EmailAddress.objects.get(email=email.email)
        self.assertEqual(EmailStatus.NEW, updatedemail.status)

    def test_validate_email_address_mismatch_fail(self):
        """Trying to validate e-mail not linked to this account should fail."""
        account = self.factory.make_account()
        email = self.factory.make_email_for_account(
            account, status=EmailStatus.NEW)
        bogus_account = self.factory.make_account()
        bogus_email = self.factory.make_email_for_account(
            bogus_account, status=EmailStatus.NEW)
        # This weird token is for account but wants to validate an e-mail
        # address not belonging to it.
        authtoken = self.factory.make_authtoken(
            requester=account, email=bogus_email.email,
            token_type=AuthTokenType.VALIDATEEMAIL)

        token = account.create_v1_oauth_token('token-name')

        data = {'ws.op': 'validate_email', 'email_token': authtoken.raw_token}
        response = self.client.get(
            self.url, data, **self.auth_header(token, parameters=data))
        # Check error message is in response and EmailAddress status wasn't
        # changed.
        self.assertContains(response, "Bad email token!")
        updatedemail = EmailAddress.objects.get(email=email.email)
        self.assertEqual(EmailStatus.NEW, updatedemail.status)

    def test_validate_email_with_leaked_creds(self):
        account = self.factory.make_account()
        account.need_password_reset = True
        account.save()
        email = self.factory.make_email_for_account(
            account, status=EmailStatus.NEW)
        authtoken = self.factory.make_authtoken(
            requester=account, email=email.email,
            token_type=AuthTokenType.VALIDATEEMAIL)

        token = account.create_v1_oauth_token('token-name')

        data = {'ws.op': 'validate_email', 'email_token': authtoken.raw_token}
        response = self.client.get(
            self.url, data, **self.auth_header(token, parameters=data))

        self.assertEqual(response.status_code, 403)


@override_settings(CAPTCHA_PUBLIC_KEY='public', CAPTCHA_PRIVATE_KEY='private')
class CaptchasTestCase(SSOBaseTestCase):

    url = reverse('api-10-captchas')

    @patch.object(Captcha, '_open')
    def test_returns_dummy(self, mock_open):
        resp = self.client.post(self.url, {'ws.op': 'new'},
                                HTTP_ACCEPT='application/json')
        self.assertEqual(None, json.loads(resp.content)['image_url'])

    @patch.object(Captcha, '_open')
    def test_puts_entry_in_timeline(self, mock_open):
        handler = CaptchaHandler()
        request = Mock()
        timeline = Timeline()
        request.META = {'timeline.timeline': timeline}
        handler.new(request)
        self.assertEqual(1, len(timeline.actions))
        self.assertEqual('captcha-new', timeline.actions[0].category)
        self.assertEqual(mock_open.call_args[0][0],
                         timeline.actions[0].detail)
