# Copyright 2010-2012 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).
import json
import logging

from piston.resource import Resource

from api.v20.authorization import is_authorized
from api.v20.utils import errors
from identityprovider.auth import SSOOAuthAuthentication


logger = logging.getLogger(__name__)


class ApiResource(Resource):

    # Patch piston to support the http PATCH method
    callmap = dict(Resource.callmap.items() + [('PATCH', 'patch')])

    def __call__(self, request, *args, **kwargs):
        request.args = args
        request.kwargs = kwargs
        response = super(ApiResource, self).__call__(request, *args, **kwargs)
        # cope with piston stripping out all response headers
        if response.status_code == 429:
            retry_after = json.loads(response.content)['extra'].get(
                'Retry-After', '')
            if retry_after:
                response['Retry-After'] = retry_after
        return response

    def form_validation_response(self, e):
        """Method to return form validation error information."""
        return errors.INVALID_DATA(**e.form.errors)


class ApiOAuthAuthentication(SSOOAuthAuthentication):

    def _has_ownership(self, request):
        return True

    def is_authenticated(self, request):
        result = super(ApiOAuthAuthentication, self).is_authenticated(request)
        if result:
            # valid oauth credentials is not enough, we need to guarantee this
            # user also has ownership over the resource
            result = self._has_ownership(request)

            if not result:
                # oauth creds are valid, but user does not own the resource,
                # check if user has special perms over the account.
                #
                # NOTE new API endpoints should handle authorization
                # separately with @authorization_required.
                result = is_authorized(request)

        return result

    def challenge(self):
        """Return a json body 401 response."""
        response = errors.INVALID_CREDENTIALS()
        response['WWW-Authenticate'] = 'OAuth realm="%s"' % self.realm
        return response


class ApiAccountsAuthentication(ApiOAuthAuthentication):

    def _has_ownership(self, request):
        openid = request.kwargs.get('openid')  # given openid
        result = request.user.openid_identifier == openid
        return result


class ApiAccountRegistrationAuthentication(ApiOAuthAuthentication):

    def _has_ownership(self, request):
        # No one is owner of a not-yet-created account
        return False


class ApiEmailsAuthentication(ApiOAuthAuthentication):

    def _has_ownership(self, request):
        email = request.kwargs.get('email')  # given email
        emails = request.user.emailaddress_set.filter(email=email)
        result = emails.count() > 0
        return result


class ApiTokensAuthentication(ApiOAuthAuthentication):

    def _has_ownership(self, request):
        token = request.kwargs.get('token')  # given token key
        result = request.user.oauth_token_by_key(token) is not None
        return result
