# Copyright 2010-2016 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

import json
import logging
import time
import urllib
import urlparse

from django.conf import settings
from django.core.exceptions import (
    MultipleObjectsReturned,
    ObjectDoesNotExist,
    ValidationError,
)
from django.core.validators import validate_email
from django.http import HttpResponse
from django.utils.translation import ugettext_lazy as _
from django_statsd.clients import statsd
from gargoyle import gargoyle
from piston.handler import AnonymousBaseHandler, BaseHandler
from piston.utils import rc, require_mime, validate

from api.v20 import registration
from api.v20.authorization import authorization_required
from api.v20.decorators import add_piston_http_patch_support, throttle
from api.v20.utils import (
    errors,
    get_account_data,
    get_email_data,
    get_minimal_account_data,
    get_token_data,
)
from api.v20.whitelist import URL_WHITELIST
from identityprovider import auth, emailutils
from identityprovider.fields import FIELD_REQUIRED
from identityprovider.forms import AccountStatusForm
from identityprovider.login import (
    AccountDeactivated,
    AccountSuspended,
    AuthenticationError,
    EmailInvalidated,
    PasswordPolicyError,
    authenticate_user,
)
from identityprovider.models import (
    Account,
    AuthToken,
    EmailAddress,
    InvalidatedEmailAddress,
    Token,
    twofactor,
)
from identityprovider.models.captcha import Captcha, VerifyCaptchaError
from identityprovider.models.const import (
    AccountStatus,
    AuthLogType,
    AuthTokenType,
    TokenScope,
)
from identityprovider.signals import login_failed, login_succeeded
from identityprovider.stats import stats
from identityprovider.timeline_helpers import get_request_timing_function
from identityprovider.utils import redirection_url_for_token
from webservices.launchpad import get_lp_ssh_keys


logger = logging.getLogger(__name__)


def statsd_timing(request, suffix, start_timestamp):
    """Helper to track metrics for Piston handlers.

    This helper would format the metrics in the same namespace as default
    Piston metrics are taken, such as:

    statsd.{env}.sso.view.api.v20.handlers.{handler}.{GET,POST}.{suffix}

    - env is one of staging, production (automatically setup)
    - handler is the handler class name (AccountsHandler, RequestsHandler, etc.
      automatically set on our custom PistonRequestTimingMiddleware).

    The given 'suffix' will be appended to the metric name, and a timing call
    will be recorded using current timestamp minus the given 'start_timestamp'.

    """
    ms = int((time.time() - start_timestamp) * 1000)  # measure time
    data = dict(module=request._view_module, name=request._view_name,
                method=request.method, suffix=suffix)
    statsd.timing('view.{module}.{name}.{method}.{suffix}'.format(**data), ms)


def valid_email(email):
    if email is None:
        return errors.INVALID_DATA(email=[FIELD_REQUIRED])
    try:
        validate_email(email)
    except ValidationError as e:
        result = errors.INVALID_DATA(email=e.messages)
    else:
        result = email
    return result


def get_openid_from_args(request, openid):
    return openid


def get_email_from_request_data(request):
    return valid_email(request.data.get('email'))


def get_token_from_args(request, token):
    return token


def get_email_or_openid(request, account):
    """Validate 'account', which might be an email address or an openid_id."""
    if '@' in account:
        return valid_email(account)
    return account


class AnonymousAccountsHandler(AnonymousBaseHandler):
    allowed_methods = ('GET',)

    @throttle(extra_callback=get_openid_from_args)
    def read(self, request, openid):
        account = Account.objects.active_by_openid(openid)
        if account is None:
            return errors.RESOURCE_NOT_FOUND()

        return get_minimal_account_data(account)


class AccountsHandler(BaseHandler):
    allowed_methods = ('GET', 'PATCH')
    anonymous = AnonymousAccountsHandler

    @throttle(extra_callback=get_openid_from_args)
    def read(self, request, openid):
        """Return account information."""
        account = Account.objects.active_by_openid(openid)
        if account is None:
            return errors.RESOURCE_NOT_FOUND()

        expand = (request.GET.get('expand', '').lower() == 'true')
        return get_account_data(account, expand=expand)

    @authorization_required
    @require_mime('json')
    @add_piston_http_patch_support
    @validate(AccountStatusForm, 'data')
    @throttle(extra_callback=get_openid_from_args)
    def patch(self, request, openid):
        account = Account.objects.active_by_openid(openid)
        if account is None:
            return errors.RESOURCE_NOT_FOUND()

        account.status = request.form.cleaned_data['status']
        account.status_comment = request.form.cleaned_data['status_comment']
        account.save()
        return get_account_data(account)


class PasswordResetTokenHandler(BaseHandler):
    allowed_methods = ('POST',)

    @require_mime('json')
    @throttle()
    def create(self, request):
        data = request.data
        email = valid_email(data.get('email'))
        if isinstance(email, HttpResponse):
            return email

        token_string = data.get('token')

        account = Account.objects.get_by_email(email)
        invalidated_email = InvalidatedEmailAddress.objects.filter(
            email__iexact=email)

        if account is None:
            # they've tried to reset with an invalid email, so send them
            # an email on how to create an account
            root_url = request.build_absolute_uri('/')
            emailutils.send_invitation_after_password_reset(root_url, email)

            if invalidated_email.exists():
                # there is no account with this email
                # but there was some with it invalidated
                return errors.EMAIL_INVALIDATED()
            else:
                msg = _("No account associated with {email}")
                # there is no account, neither with this email invalidated
                return errors.INVALID_DATA(email=[msg.format(email=email)])

        if not account.can_reset_password:
            if account.status == AccountStatus.SUSPENDED:
                # log why email was not sent
                condition = ("account '%s' is not active" %
                             account.displayname)
                logger.debug("PasswordResetTokenHandler.create: email was "
                             "not sent out because %s" % condition)
                return errors.ACCOUNT_SUSPENDED()
            elif account.status == AccountStatus.DEACTIVATED:
                # log why email was not sent
                condition = ("account '%s' is not active" %
                             account.displayname)
                logger.debug("PasswordResetTokenHandler.create: email was "
                             "not sent out because %s" % condition)
                return errors.ACCOUNT_DEACTIVATED()

            # user does not have any verified email address
            # he should contact support
            condition = ("account '%s' is not allowed to reset password" %
                         account.displayname)
            logger.debug("PasswordResetTokenHandler.create: email was not "
                         "sent out because %s" % condition)
            return errors.CAN_NOT_RESET_PASSWORD()

        validated = account.emailaddress_set.verified().filter(
            email__iexact=email).exists()
        if not validated and account.preferredemail is not None:
            email = account.preferredemail.email

        tokens = AuthToken.objects.filter(
            token_type=AuthTokenType.PASSWORDRECOVERY, requester_email=email,
            email=email, date_consumed=None)
        if tokens.count() >= settings.MAX_PASSWORD_RESET_TOKENS:
            return errors.TOO_MANY_TOKENS()

        # create new token and send email to user
        redirection_url = redirection_url_for_token(token_string)
        root_url = request.build_absolute_uri('/')
        auth_token, inv_auth_token = emailutils.send_password_reset_email(
            root_url, account, email, redirection_url)

        # return response
        data = dict(email=auth_token.email)
        response = rc.CREATED
        response.content = data
        return response


class AnonymousAccountRegistrationHandler(AnonymousBaseHandler):
    allowed_methods = ('POST',)

    @require_mime('json')
    @throttle()
    def create(self, request):
        """Create/register a new account."""
        data = request.data

        try:
            email = data['email']
            password = data['password']
            displayname = data['displayname']
        except KeyError:
            expected = set(('email', 'password', 'displayname'))
            missing = dict((k, [FIELD_REQUIRED]) for k in expected - set(data))
            return errors.INVALID_DATA(**missing)

        captcha_required = (gargoyle.is_active('CAPTCHA', request) and
                            gargoyle.is_active('CAPTCHA_NEW_ACCOUNT', request))

        captcha_id = data.get('captcha_id')
        captcha_solution = data.get('captcha_solution')
        if not (captcha_solution and captcha_id) and captcha_required:
            extra = {}
            if data.get('create_captcha', True):
                timer = get_request_timing_function(request)
                try:
                    extra = Captcha.new(timer=timer).serialize()
                except Exception:
                    logger.exception('failed to create reCaptcha')
            return errors.CAPTCHA_REQUIRED(**extra)

        elif captcha_id:
            remote_addr = request.environ['REMOTE_ADDR']
            captcha = Captcha(captcha_id)
            verified = False
            try:
                timer = get_request_timing_function(request)
                verified = captcha.verify(
                    captcha_solution, remote_addr, data['email'],
                    timer=timer)
            except VerifyCaptchaError as e:
                # only expose HTTP error codes to client
                code = e.response.code if e.response.code > 200 else None
                return errors.CAPTCHA_ERROR(
                    recaptcha_reason=e.response.reason,
                    recaptcha_status_code=code,
                    recaptcha_body=e.response.body)
            except Exception:
                logger.exception("reCaptcha error")

            if not verified:
                message = getattr(captcha, 'message', '')
                return errors.CAPTCHA_FAILURE(captcha_message=message)

        root_url = request.build_absolute_uri('/')
        try:
            account = registration.register(
                root_url, email, password, displayname,
                creation_source=data.get('creation_source'),
            )
        except ValidationError as e:
            return errors.INVALID_DATA(**e.message_dict)
        except registration.EmailAlreadyRegistered:
            return errors.ALREADY_REGISTERED(email=data['email'])
        except PasswordPolicyError as e:
            return errors.PASSWORD_POLICY_ERROR(
                reason=unicode(e), location=root_url)

        data = get_account_data(account)
        response = rc.CREATED
        response.content = data
        return response


class AccountRegistrationHandler(BaseHandler):
    allowed_methods = ('POST',)
    anonymous = AnonymousAccountRegistrationHandler

    # no throttle limit since this endpoint is used by privileged accounts only
    @require_mime('json')
    def create(self, request):
        """Create/register a new account as a privileged user.

        The newly created account will be verified and passwordless,
        and the owner will not be able to login until a password reset is done.

        Captcha is completely ignored.

        """
        data = request.data

        try:
            email = data['email']
            displayname = data['displayname']
        except KeyError:
            expected = set(('email', 'displayname'))
            missing = dict((k, [FIELD_REQUIRED]) for k in expected - set(data))
            return errors.INVALID_DATA(**missing)

        root_url = request.build_absolute_uri('/')
        try:
            account = registration.register(
                root_url, email, password=None, displayname=displayname,
                email_validated=True, send_email=False,
                creation_source=data.get('creation_source'),
            )
        except ValidationError as e:
            return errors.INVALID_DATA(**e.message_dict)
        except registration.EmailAlreadyRegistered as exc:
            account = exc.account
            response = rc.ALL_OK
        else:
            response = rc.CREATED

        data = get_account_data(account)
        response.content = data
        return response


class AccountLoginHandler(BaseHandler):
    allowed_methods = ('POST',)

    @require_mime('json')
    @throttle(extra_callback=get_email_from_request_data)
    def create(self, request):
        data = request.data
        try:
            email = data['email']
            password = data['password']
            token_name = data['token_name']
        except KeyError:
            expected = set(('email', 'password', 'token_name'))
            missing = dict((k, [FIELD_REQUIRED]) for k in expected - set(data))
            return errors.INVALID_DATA(**missing)

        account = None
        response = None
        try:
            account = authenticate_user(email, password, request=request)
        except AccountSuspended:
            response = errors.ACCOUNT_SUSPENDED()
        except AccountDeactivated:
            response = errors.ACCOUNT_DEACTIVATED()
        except EmailInvalidated:
            response = errors.EMAIL_INVALIDATED()
        except PasswordPolicyError as e:
            root_url = request.build_absolute_uri('/')
            response = errors.PASSWORD_POLICY_ERROR(
                reason=unicode(e), location=root_url)
        except AuthenticationError:
            response = errors.INVALID_CREDENTIALS()

        if account is not None:
            otp = data.get('otp')
            if otp is not None:
                try:
                    twofactor.authenticate(account, otp)
                except AuthenticationError:
                    account = None
                    response = errors.TWOFACTOR_FAILURE()
            elif account.twofactor_required:
                response = errors.TWOFACTOR_REQUIRED()

        if account is None:
            # track failed login attempt
            login_failed.send(sender=self, request=request,
                              credentials=dict(email=email, password=password),
                              authlogtype=AuthLogType.API_FAIL)

        if response is not None:
            return response

        token, created = account.get_or_create_oauth_token(token_name)
        if created:
            response = rc.CREATED
            log_type = AuthLogType.API_NEW
        else:
            response = rc.ALL_OK
            log_type = AuthLogType.API_EXISTING

        data = get_token_data(token)
        data['openid'] = account.openid_identifier
        response.content = data

        login_succeeded.send(
            sender=self, user=account, request=request, authlogtype=log_type,
            email=email)
        return response


class MacaroonDischargeHandler(BaseHandler):
    allowed_methods = ('POST',)

    @require_mime('json')
    @throttle(extra_callback=get_email_from_request_data)
    def create(self, request):
        if settings.CRYPTO_SSO_PRIVKEY is None:
            return errors.RESOURCE_NOT_FOUND()

        data = request.data
        try:
            email = data['email']
            password = data['password']
        except KeyError:
            expected = {'email', 'password'}
            missing = dict((k, [FIELD_REQUIRED]) for k in expected - set(data))
            return errors.INVALID_DATA(**missing)

        if 'macaroon' in data:
            simple_macaroon = True
            root_macaroons_info = [(None, data['macaroon'])]
            stats.increment('macaroon.discharge', key='single_macaroon')
        elif 'macaroons' in data:
            simple_macaroon = False
            root_macaroons_info = data['macaroons']
            stats.increment('macaroon.discharge', key='multi_macaroons')
        elif 'caveat_id' in data:
            simple_macaroon = True
            root_macaroons_info = None
            caveat_id = data['caveat_id']
            stats.increment('macaroon.discharge', key='caveat_id')
        else:
            # let's flag the new one only, not the deprecated ones
            missing = {'caveat_id': [FIELD_REQUIRED]}
            return errors.INVALID_DATA(**missing)

        account = None
        response = None
        try:
            account = authenticate_user(email, password, request=request)
        except AccountSuspended:
            response = errors.ACCOUNT_SUSPENDED()
        except AccountDeactivated:
            response = errors.ACCOUNT_DEACTIVATED()
        except EmailInvalidated:
            response = errors.EMAIL_INVALIDATED()
        except PasswordPolicyError as e:
            root_url = request.build_absolute_uri('/')
            response = errors.PASSWORD_POLICY_ERROR(
                reason=unicode(e), location=root_url)
        except AuthenticationError:
            response = errors.INVALID_CREDENTIALS()

        if account is not None:
            otp = data.get('otp')
            if otp is not None:
                try:
                    twofactor.authenticate(account, otp)
                except AuthenticationError:
                    account = None
                    response = errors.TWOFACTOR_FAILURE()
            elif account.twofactor_required:
                response = errors.TWOFACTOR_REQUIRED()

        if account is None:
            # track failed login attempt
            login_failed.send(sender=self, request=request,
                              credentials=dict(email=email, password=password),
                              authlogtype=AuthLogType.API_FAIL)

        if response is not None:
            return response

        try:
            if root_macaroons_info is None:
                discharge = auth.build_discharge_macaroon(account, caveat_id)
                all_discharges = [(None, discharge.serialize())]
            else:
                # 2016-05-03: this is deprecated, just supporting it until
                # no more clients send root macaroon(s)
                all_discharges = []
                for mac_id, root_macaroon_raw in root_macaroons_info:
                    discharge = auth.build_discharge_macaroon_from_root(
                        account, root_macaroon_raw)
                    all_discharges.append((mac_id, discharge.serialize()))
        except ValidationError:
            return errors.INVALID_DATA()
        except AuthenticationError:
            return errors.INVALID_CREDENTIALS()

        response = rc.ALL_OK
        if simple_macaroon:
            response.content = dict(discharge_macaroon=all_discharges[0][1])
        else:
            response.content = dict(discharge_macaroons=all_discharges)

        log_type = AuthLogType.API_MACAROON_DISCHARGE_NEW
        login_succeeded.send(sender=self, user=account, request=request,
                             authlogtype=log_type, email=email)
        return response


class MacaroonRefreshHandler(BaseHandler):
    allowed_methods = ('POST',)

    @require_mime('json')
    def create(self, request):
        if settings.CRYPTO_SSO_PRIVKEY is None:
            return errors.RESOURCE_NOT_FOUND()

        data = request.data
        try:
            discharge_macaroon_raw = data['discharge_macaroon']
        except KeyError:
            missing = {'discharge_macaroon': [FIELD_REQUIRED]}
            return errors.INVALID_DATA(**missing)

        if 'root_macaroon' in data:
            # (deprecated 2016-05-03) use the root macaroon to old-fashion
            # refresh, that also returns a bound discharge
            method = auth.refresh_macaroons
            params = (data['root_macaroon'], discharge_macaroon_raw)
            stats.increment('macaroon.refresh', key='root_macaroon')
        else:
            method = auth.refresh_discharge
            params = (discharge_macaroon_raw,)
            stats.increment('macaroon.refresh', key='just_discharge')

        try:
            new_discharge = method(*params)
        except ValidationError:
            return errors.INVALID_DATA()
        except AccountDeactivated:
            return errors.ACCOUNT_DEACTIVATED()
        except AuthenticationError:
            return errors.INVALID_CREDENTIALS()

        response = rc.ALL_OK
        response.content = dict(discharge_macaroon=new_discharge.serialize())
        return response


class EmailsHandler(BaseHandler):
    allowed_methods = ('DELETE', 'GET', 'PATCH')
    model = EmailAddress

    @throttle(extra_callback=lambda request, email: valid_email(email))
    def read(self, request, email):
        try:
            result = get_email_data(self.model.objects.get(email=email))
        except ObjectDoesNotExist:
            result = errors.RESOURCE_NOT_FOUND()
        except MultipleObjectsReturned:
            result = errors.INVALID_DATA()
        return result

    @throttle(extra_callback=lambda request, email: valid_email(email))
    def delete(self, request, email):
        result = super(EmailsHandler, self).delete(request, email=email)
        # piston is returning 410 (Gone) when the email does not exist
        # this is not correct, we should return 404
        if result.status_code == rc.NOT_HERE.status_code:
            result = errors.RESOURCE_NOT_FOUND()
        return result

    @authorization_required
    @require_mime('json')
    @add_piston_http_patch_support
    @validate(AccountStatusForm, 'data')
    @throttle(extra_callback=lambda request, email: valid_email(email))
    def patch(self, request, email):
        account = Account.objects.get_by_email(email)
        if account is None:
            return errors.RESOURCE_NOT_FOUND()

        account.status = request.form.cleaned_data['status']
        account.status_comment = request.form.cleaned_data['status_comment']
        account.save()
        return get_account_data(account)


class RequestsHandler(BaseHandler):

    allowed_methods = ('POST',)

    def log_warning(self, msg, *args):
        sanitized_args = [
            '<REDACTED>' if 'PLAINTEXT' in repr(arg) else arg
            for arg in args
        ]
        logger.warning(msg, *sanitized_args)

    def serialize_error(self, exception):
        result = exception.__class__.__name__
        msg = ', '.join(repr(i) for i in exception.args)
        if msg:
            result += ': %s' % msg
        return result

    def token_ok(self, token, http_url):
        result = False
        scope = getattr(token, 'scope', None)
        if scope is not None:
            result = (scope == TokenScope.VERSION_2)
        else:
            # is the url whitelisted?
            for url in URL_WHITELIST:
                if url.match(http_url) is not None:
                    result = True
                    break

        return result

    @require_mime('json')
    def create(self, request):
        """Return whether a given OAuth signature is valid.

        Receive an oauth Authorization header, an url and an http method and
        decide if the signature is valid (was created using valid credentials
        and belongs to an active account).

        """
        data = request.data
        http_method = data.get('http_method', '')
        http_url = data.get('http_url', '')
        authorization = data.get('authorization', None)
        query_string = data.get('query_string', '')

        result = dict(is_valid=False, account_verified=False)
        start_timestamp = time.time()  # metrics

        if not http_url or not http_method:
            statsd_timing(  # metrics
                request, 'missing-' + ('url' if http_method else 'method'),
                start_timestamp)
            return result

        headers = None
        if authorization:
            headers = {'Authorization': authorization}

        if query_string or not authorization:
            # Backwards compatible change, covering 2 scenarios:
            #
            # 1- query_string is no longer needed, but we merge it with the
            #    http_url if not already there, for possible oauth values or
            #    real query string values that were signed but not included
            #    in the http_url value (u1db-server).
            #
            # 2- http_url with every oauth paramater in the query string, but
            #    with buggy urlencoding, and because oauthlib is pretty strict,
            #    we get errors such as: ValueError: Error trying to decode a
            #    non urlencoded string. Found invalid characters: '/' (pay-ui).
            #    Reported as LP: #1504548 on 2015-10-09.

            original_http_url = http_url
            parts = urlparse.urlparse(http_url)

            # the original_http_url has query string to be merged
            parse_qsl = lambda qs: urlparse.parse_qsl(qs.encode('utf-8'))
            parsed_query_string = set(parse_qsl(query_string))
            original_query = set(parse_qsl(parts.query))
            qs = original_query.union(parsed_query_string)
            qs = urllib.urlencode(sorted(qs))

            http_url = urlparse.urlunparse((
                parts.scheme, parts.netloc, parts.path, parts.params,
                qs, parts.fragment))
            self.log_warning(
                'RequestsHandler.create: caller for validating http_url %r '
                'sent query_string %r or did not send authorization (got %r), '
                'will merge it into original url.',
                original_http_url, query_string, authorization)

        def token_allowed(token):
            result = self.token_ok(token, http_url)
            if not result:
                self.log_warning(
                    'RequestsHandler.create: NOT validating (otherwise valid) '
                    'signature %r because the token is not allowed to access '
                    'url %r via %r', authorization, http_url, http_method)
            return result

        try:
            valid, oauth_request = auth.validate_oauth_signature(
                http_url, http_method=http_method, headers=headers,
                token_filter_fn=token_allowed)
        except Exception as e:
            result['error'] = self.serialize_error(e)
            logger.exception(
                'RequestsHandler.create: could not verify request (data %r):',
                data)
            valid = False

        if valid:
            # signature is valid, account exists and is active
            openid_identifier = oauth_request.client_key
            verified = Account.objects.verified().filter(
                openid_identifier=openid_identifier).exists()
            result['is_valid'] = True
            result['account_verified'] = verified
            result['identifier'] = openid_identifier

        statsd_timing(  # metrics
            request, 'oauth-valid' if valid else 'oauth-invalid',
            start_timestamp)

        return result


class TokensHandler(BaseHandler):
    allowed_methods = ('DELETE', 'GET')
    model = Token

    @throttle(extra_callback=get_token_from_args)
    def read(self, request, token):
        """Return token information."""
        result = super(TokensHandler, self).read(request, key=token)
        if isinstance(result, self.model):
            result = get_token_data(result)
            result.pop('token_secret', None)
            result.pop('consumer_secret', None)
        else:
            result = errors.RESOURCE_NOT_FOUND()
        return result

    @throttle(extra_callback=get_token_from_args)
    def delete(self, request, token):
        """Return token information."""
        result = super(TokensHandler, self).delete(request, key=token)
        # piston is returning 410 (Gone) when the token does not exist
        # this is not correct, we should return 404
        if result.status_code == rc.NOT_HERE.status_code:
            result = errors.RESOURCE_NOT_FOUND()
        return result


class SSHKeyHandler(BaseHandler):
    allowed_methods = ('GET',)

    @throttle(extra_callback=get_email_or_openid)
    def read(self, request, account):
        """Return ssh keys for an SSO account.

        Accounts are identified by either an email address or an openid
        identifier. Since openid identifiers can't contain the '@' character,
        we can safely differentiate the two different methods of access.
        """
        account = self.get_account_by_openid_or_email(account)
        if not account or not account.person_name:
            content = {'error_type': 'no_keys',
                       'message': 'There are no keys for this address.'}
            return HttpResponse(content=json.dumps(content), status=404,
                                content_type='application/json; charset=utf-8')

        keys = get_lp_ssh_keys(account.openid_identifier)
        resp = rc.ALL_OK
        resp.content = {
            'ssh_keys': keys,
            'username': account.person_name,
            'openid_identifier': account.openid_identifier,
        }
        return resp

    def get_account_by_openid_or_email(self, openid_or_email):
        if '@' in openid_or_email:
            return Account.objects.get_by_email(openid_or_email)
        else:
            return Account.objects.active_by_openid(openid_or_email)
