# Copyright 2010-2012 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

from django.core.exceptions import ValidationError
from django.template.loader import render_to_string

from identityprovider.models import (
    EmailAddress,
    Account,
)
from identityprovider.models.const import AccountCreationRationale
from identityprovider.emailutils import (
    send_impersonation_email,
    send_new_user_email,
)
from identityprovider.forms import NewAccountForm
from identityprovider.login import PasswordPolicyError
from identityprovider.signals import account_created
from identityprovider.validators import PASSWORD_LEAKED


class EmailAlreadyRegistered(Exception):

    def __init__(self, account, *args, **kwargs):
        self.account = account
        super(EmailAlreadyRegistered, self).__init__(*args, **kwargs)


def register(
        root_url, email, password, displayname, email_validated=False,
        send_email=True, creation_source=None):
    """Register a new account"""
    emails = EmailAddress.objects.filter(email__iexact=email)
    if emails.count() > 0:
        other_email = emails[0]
        # Only send email if the account is active; otherwise, a disabled
        # account can be spammed.
        if other_email.account.is_active and send_email:
            send_impersonation_email(root_url, other_email.email)
        raise EmailAlreadyRegistered(other_email.account)

    if password is not None:
        data = {
            'email': email,
            'password': password,
            'passwordconfirm': password,
            'displayname': displayname,
            'accept_tos': True,
        }
        form = NewAccountForm(data)
        if not form.is_valid():
            leaked = unicode(PASSWORD_LEAKED) in form.errors.get(
                'password', '')
            if leaked:
                message = render_to_string(
                    'common/password_leaked_warning.txt')
                raise PasswordPolicyError(message, None)
            else:
                raise ValidationError(form.errors)

    account = Account.objects.create_account(
        displayname, email, password, email_validated=email_validated,
        creation_rationale=AccountCreationRationale.API_CREATED,
        creation_source=creation_source,
    )

    account_created.send('api', openid_identifier=account.openid_identifier)

    if send_email:
        send_new_user_email(root_url=root_url, account=account, email=email)

    return account
