# -*- coding: utf-8 -*-

from __future__ import unicode_literals

import json
from urllib import urlencode

from django.core.urlresolvers import reverse
from django.test.client import RequestFactory
from piston.handler import AnonymousBaseHandler
from piston.resource import Resource

from api.v20.auth import (
    ApiAccountsAuthentication,
    ApiEmailsAuthentication,
    ApiOAuthAuthentication,
    ApiTokensAuthentication,
    ApiResource,
)
from identityprovider.tests import DEFAULT_USER_PASSWORD
from identityprovider.tests.utils import SSOBaseTestCase


class ApiResourceTestCase(SSOBaseTestCase):

    def setUp(self):
        super(ApiResourceTestCase, self).setUp()
        self.resource = ApiResource(handler=AnonymousBaseHandler)
        self.request = RequestFactory().get('/')  # url is irrelevant

        assert getattr(self.request, 'args', None) is None
        assert getattr(self.request, 'kwargs', None) is None

    def test_is_piston_resource(self):
        self.assertIsInstance(self.resource, Resource)

    def test_args_are_stored_on_call(self):
        args = ('foo', object(), dict(zaraza=42))
        with self.assertRaises(AttributeError):
            self.resource(self.request, *args)

        self.assertEqual(getattr(self.request, 'args', None), args)

    def test_kwargs_are_stored_on_call(self):
        kwargs = dict(bar='foo', something=object(), extra=dict(zaraza=42))
        with self.assertRaises(AttributeError):
            self.resource(self.request, **kwargs)

        self.assertEqual(getattr(self.request, 'kwargs', None), kwargs)


class ApiOAuthAuthenticationTestCase(SSOBaseTestCase):

    auth_class = ApiOAuthAuthentication
    url_name = 'account-index'  # url is not relevant for this test case
    url_kwargs = {}

    def setUp(self):
        super(ApiOAuthAuthenticationTestCase, self).setUp()
        self.auth = self.auth_class()
        self.email = 'test@canonical.com'
        self.account = self.factory.make_account(
            email=self.email, password=DEFAULT_USER_PASSWORD)
        self.token = self.account.create_oauth_token('api-test')

    def get_request(self, method='GET', token=None, qs_params=None):
        url = reverse(self.url_name, kwargs=self.url_kwargs)
        if qs_params is not None:
            url += '?' + urlencode(qs_params)

        headers = {}
        if token is not None:
            headers = self.factory.make_authorization_header_from_token(
                url, token, http_method=method.upper())

        request = getattr(RequestFactory(), method.lower())(url, **headers)
        # the following is ensured by ApiResource
        request.args = ()
        request.kwargs = self.url_kwargs
        return request

    def test_challenge(self):
        response = self.auth.challenge()

        self.assertEqual(response.status_code, 401)
        self.assertEqual(response['WWW-Authenticate'],
                         'OAuth realm="Ubuntu Single Sign On API"')
        self.assertEqual(response['content-type'],
                         'application/json; charset=utf-8')

        content = json.loads(response.content)
        self.assertEqual(content['code'], 'INVALID_CREDENTIALS')

    def test_is_not_authenticated(self):
        request = self.get_request()
        result = self.auth.is_authenticated(request)

        self.assertFalse(result)

    def test_authenticated_with_django_is_not_authenticated(self):
        assert self.client.login(
            username=self.email, password=DEFAULT_USER_PASSWORD)
        request = self.get_request()
        result = self.auth.is_authenticated(request)

        self.assertFalse(result)

    def test_is_authenticated_if_has_ownership(self):
        request = self.get_request(token=self.token)

        result = self.auth.is_authenticated(request)

        self.assertTrue(result)

    def test_not_authenticated_if_v1_token(self):
        v1_token = self.account.create_v1_oauth_token('v1 token test')
        request = self.get_request(token=v1_token)

        result = self.auth.is_authenticated(request)

        self.assertFalse(result)

    def test_add_troubling_query_strings(self):
        params = [
            ('name', 'ñoño ñandú'.encode('utf-8')),
            ('email', 'foo@foo.com'.encode('utf-8')),
            ('name', 'algo más'.encode('utf-8')),
        ]
        request = self.get_request(token=self.token, qs_params=params)

        result = self.auth.is_authenticated(request)

        self.assertTrue(result)


class ApiAccountsAuthenticationTestCase(ApiOAuthAuthenticationTestCase):

    auth_class = ApiAccountsAuthentication
    url_name = 'api-account'

    def setUp(self):
        super(ApiAccountsAuthenticationTestCase, self).setUp()
        self.url_kwargs = dict(openid=self.account.openid_identifier)

        other = self.factory.make_account()
        self.other_openid = other.openid_identifier

        assert self.account.openid_identifier != self.other_openid

    def _is_authenticated_if_no_ownership(self, perm, method):
        if perm is not None:
            self.factory.add_permission_to_account(self.account, perm)

        self.url_kwargs = dict(openid=self.other_openid)
        request = self.get_request(method=method, token=self.token)

        return self.auth.is_authenticated(request)

    def test_is_not_authenticated_if_no_ownership(self):
        result = self._is_authenticated_if_no_ownership(
            perm=None, method='GET')
        self.assertFalse(result)

    def test_is_not_authenticated_if_no_ownership_edit_perm_get(self):
        result = self._is_authenticated_if_no_ownership(
            perm='api_edit_account_details', method='GET')
        self.assertFalse(result)

    def test_is_authenticated_if_no_ownership_edit_perm_post(self):
        result = self._is_authenticated_if_no_ownership(
            perm='api_edit_account_details', method='POST')
        self.assertTrue(result)

    def test_is_authenticated_if_no_ownership_edit_perm_put(self):
        result = self._is_authenticated_if_no_ownership(
            perm='api_edit_account_details', method='PUT')
        self.assertTrue(result)

    def test_is_authenticated_if_no_ownership_edit_perm_delete(self):
        result = self._is_authenticated_if_no_ownership(
            perm='api_edit_account_details', method='DELETE')
        self.assertTrue(result)

    def test_is_authenticated_if_no_ownership_view_perm_get(self):
        result = self._is_authenticated_if_no_ownership(
            perm='api_view_account_details', method='GET')
        self.assertTrue(result)

    def test_is_not_authenticated_if_no_ownership_view_perm_post(self):
        result = self._is_authenticated_if_no_ownership(
            perm='api_view_account_details', method='POST')
        self.assertFalse(result)

    def test_is_not_authenticated_if_no_ownership_view_perm_put(self):
        result = self._is_authenticated_if_no_ownership(
            perm='api_view_account_details', method='PUT')
        self.assertFalse(result)

    def test_is_not_authenticated_if_no_ownership_view_perm_delete(self):
        result = self._is_authenticated_if_no_ownership(
            perm='api_view_account_details', method='DELETE')
        self.assertFalse(result)


class ApiAccountRegistrationAuthenticationTestCase(
        ApiOAuthAuthenticationTestCase):

    auth_class = ApiAccountsAuthentication
    url_name = 'api-registration'

    def test_is_authenticated_if_has_ownership(self):
        request = self.get_request(token=self.token)

        result = self.auth.is_authenticated(request)

        self.assertFalse(result)

    def test_add_troubling_query_strings(self):
        params = [
            ('name', 'ñoño ñandú'.encode('utf-8')),
            ('email', 'foo@foo.com'.encode('utf-8')),
            ('name', 'algo más'.encode('utf-8')),
        ]
        request = self.get_request(token=self.token, qs_params=params)

        result = self.auth.is_authenticated(request)

        self.assertFalse(result)  # never authenticated because false ownership


class ApiEmailsAuthenticationTestCase(ApiAccountsAuthenticationTestCase):

    auth_class = ApiEmailsAuthentication
    url_name = 'api-email'

    def setUp(self):
        super(ApiEmailsAuthenticationTestCase, self).setUp()
        self.url_kwargs = dict(email=self.email)

        other = self.factory.make_account()
        self.other_email = other.preferredemail

        assert self.email != self.other_email

    def _is_authenticated_if_no_ownership(self, perm, method):
        if perm is not None:
            self.factory.add_permission_to_account(self.account, perm)

        self.url_kwargs = dict(email=self.other_email)
        request = self.get_request(method=method, token=self.token)

        return self.auth.is_authenticated(request)


class ApiTokensAuthenticationTestCase(ApiAccountsAuthenticationTestCase):

    auth_class = ApiTokensAuthentication
    url_name = 'api-token'

    def setUp(self):
        super(ApiTokensAuthenticationTestCase, self).setUp()
        self.url_kwargs = dict(token=self.token.pk)

        other = self.factory.make_account()
        self.other_token_key = other.create_oauth_token('other').key

        assert self.token.key != self.other_token_key

    def _is_authenticated_if_no_ownership(self, perm, method):
        if perm is not None:
            self.factory.add_permission_to_account(self.account, perm)

        self.url_kwargs = dict(token=self.other_token_key)
        request = self.get_request(method=method, token=self.token)

        return self.auth.is_authenticated(request)
