import json
import string
import time

from django.conf.urls import patterns, url
from django.core.cache import cache
from django.core.urlresolvers import reverse
from memcache import Client
from piston.handler import BaseHandler
from piston.resource import Resource
from urls import urlpatterns as original_patterns

from api.v20.decorators import throttle
from identityprovider.tests.utils import SSOBaseTestCase


class DummyHandler(BaseHandler):
    allowed_methods = ('GET',)

    @throttle(extra_callback=lambda r, c: c)
    def read(self, request, testing_param):
        return {'success': testing_param}


urlpatterns = original_patterns + patterns(
    '',
    url(r'^dummy-api-test/(?P<testing_param>.+)$',
        Resource(handler=DummyHandler), name='dummy-api'),
)


class ThrottleTestCase(SSOBaseTestCase):

    urls = 'api.v20.tests.test_decorators'

    def setUp(self):
        super(ThrottleTestCase, self).setUp()
        self.now = time.time()

    def assert_handler_response(
            self, param='foo', status_code=200, expected=None, **kwargs):
        url = reverse('dummy-api', args=(param,))
        response = self.client.get(url, **kwargs)

        self.assertEqual(response.status_code, status_code)
        if expected is None:
            expected = {'success': param}
        self.assertEqual(json.loads(response.content), expected)

    def test_letters_and_digits(self):
        black_list_chars = [';']
        for i in string.letters + string.digits + string.punctuation:
            # XXX 2016-01-05 psivaa: django 1.8 does treat `;`
            # as unsafe now and escapes
            if i not in black_list_chars:
                self.assert_handler_response(i)

    def test_control_characters(self):
        error = 'Control characters not allowed, local_extra: %r'
        for i in '\t\x0b\x0c\r ':  # string.whitespace without \n
            with self.assertRaises(Client.MemcachedKeyCharacterError) as ctx:
                self.assert_handler_response(i)

            self.assertEqual(unicode(ctx.exception), error % unicode(i))

    def test_max_requests(self):
        throttled = {
            'code': 'TOO_MANY_REQUESTS',
            'message': 'Too many requests. Please try again later.',
            'extra': {'Retry-After': '3599'},
        }
        with self.settings(THROTTLE_MAX_REQUESTS=2):
            self.assert_handler_response()
            self.assert_handler_response(status_code=429, expected=throttled)

    def test_throttling_key(self):
        assert cache.get('127.0.0.1:foo') is None
        self.assert_handler_response(param='foo')

        value = cache.get('127.0.0.1:foo')
        self.assertIsNotNone(value)
        self.assertEqual(value[0], 2)

    def test_uses_real_user_ip(self):
        assert cache.get('127.0.0.1:foo') is None
        forwarded_ip = '1.2.13.14'
        assert cache.get(forwarded_ip + ':foo') is None

        self.assert_handler_response(HTTP_X_FORWARDED_FOR=forwarded_ip)

        self.assertIsNone(cache.get('127.0.0.1:foo'))
        value = cache.get(forwarded_ip + ':foo')
        self.assertIsNotNone(value)
        self.assertEqual(value[0], 2)
