# -*- coding: utf-8 -*-

from __future__ import unicode_literals

import json
import re
import time
import uuid

from StringIO import StringIO
from urllib import quote, urlencode
from urlparse import parse_qsl, urlparse, urlunparse

from django.conf import settings
from django.contrib.auth.models import AnonymousUser, make_password
from django.core import mail
from django.core.serializers.json import DjangoJSONEncoder
from django.core.urlresolvers import NoReverseMatch, reverse
from django.test.client import RequestFactory
from django.test.utils import override_settings
from django.utils.timezone import now
from gargoyle.testutils import switches
from mock import Mock, patch
from pymacaroons import Macaroon, Verifier
from timeline import Timeline

from api.v20 import handlers, whitelist
from identityprovider.auth import build_discharge_macaroon
from identityprovider.login import PasswordPolicyError
from identityprovider.models import (
    Account,
    AuthToken,
    EmailAddress,
    Token,
    captcha,
)
from identityprovider.models.const import (
    AccountCreationRationale,
    AccountStatus,
    AuthLogType,
    AuthTokenType,
    EmailStatus,
    TokenScope,
)
from identityprovider.signals import login_failed
from identityprovider.tests import DEFAULT_USER_PASSWORD
from identityprovider.tests.test_auth import AuthLogTestCaseMixin
from identityprovider.tests.utils import SSOBaseTestCase, TimelineActionMixin
from identityprovider.utils import redirection_url_for_token


OVERRIDES = dict(
    EMAIL_WHITELIST_REGEXP_LIST=['^canonicaltest(?:\+.+)?@gmail\.com$']
)


class POSTOnlyResourceMixin(object):

    def test_get_not_allowed(self):
        response = self.client.get(self.url)
        self.assertEqual(response.status_code, 405)
        self.assertEqual(response.get('Allow', ''), 'POST')

    def test_requires_post_data_json_encoded(self):
        response = self.client.post(self.url, data={'foo': ['bar', 'baz']})
        # django-piston doesn't set this required attribute
        response._base_content_is_iter = False
        self.assertContains(response, 'Bad Request', status_code=400)


class BaseTestCase(SSOBaseTestCase):

    url = None

    def setUp(self):
        super(BaseTestCase, self).setUp()
        self.account = self.factory.make_account(
            password=DEFAULT_USER_PASSWORD)
        self.token = self.account.create_oauth_token('token-name')

    def get_response(
            self, url=None, token=None, method='GET', status_code=200,
            **kwargs):
        if url is None:
            url = self.url

        if token is not None:
            headers = self.factory.make_authorization_header_from_token(
                url, token, http_method=method.upper())
            kwargs.update(headers)

        response = getattr(self.client, method.lower())(url, **kwargs)
        if status_code is not None:
            self.assertEqual(response.status_code, status_code)
        return response

    def do_request(self, method, data=None, raw_data=None, url=None,
                   status_code=200, token=None,
                   content_type='application/json', **kwargs):
        if raw_data is not None:
            data = raw_data
        elif data:
            data = json.dumps(data)
        response = self.get_response(
            url=url, token=token, method=method, status_code=status_code,
            data=data, content_type=content_type, **kwargs)
        return self.from_json(response)

    def do_get(self, *args, **kwargs):
        return self.do_request('GET', *args, **kwargs)

    def do_patch(self, *args, **kwargs):
        return self.do_request('PATCH', *args, **kwargs)

    def do_post(self, *args, **kwargs):
        return self.do_request('POST', *args, **kwargs)

    def do_put(self, *args, **kwargs):
        return self.do_request('PUT', *args, **kwargs)

    def from_json(self, response):
        self.assertIn(
            response['Content-Type'],
            ('application/json', 'application/json; charset=utf-8'))
        try:
            result = json.loads(response.content)
        except:
            self.fail('Could not JSON-decode response, content is %r.' %
                      response.content)
        return result

    def to_json(self, expected):
        return json.loads(json.dumps(expected, cls=DjangoJSONEncoder))

    def assert_bad_request(self, data, extra=None, **kwargs):
        if extra is None:
            extra = {}

        json_body = self.do_post(data, status_code=400, **kwargs)

        self.assertEqual(json_body['code'], "INVALID_DATA")
        self.assertIn("Invalid request data", json_body['message'])
        self.assertEqual(json_body['extra'], extra)

    def assert_field_required(self, data, field, **kwargs):
        data.pop(field, None)
        self.assert_bad_request(
            data, extra={field: [handlers.FIELD_REQUIRED]}, **kwargs)

    def assert_any_server_error_is_json(self, api_call, **api_call_kwargs):
        api_call_kwargs['status_code'] = 500

        error = RuntimeError('Something very bad happened')

        with patch('django.db.models.manager.Manager.get_queryset',
                   side_effect=error):
            json_body = api_call(**api_call_kwargs)

        self.assertEqual(
            json_body['message'],
            '%s: %s' % (error.__class__.__name__, unicode(error)))
        self.assertEqual(json_body['code'], 'SERVER_ERROR')
        self.assertEqual(json_body['extra'], {})


class AccountsHandlerTestCase(BaseTestCase):
    url_name = 'api-account'

    def setUp(self):
        super(AccountsHandlerTestCase, self).setUp()
        self.url = reverse(self.url_name,
                           args=(self.account.openid_identifier,))

        account = self.factory.make_account(
            email='super@user.com', permissions=['api_edit_account_details'])
        self.super_token = account.create_oauth_token('super-token')
        self.data = {'status': 'SUSPENDED', 'status_comment': 'Test comment'}

    def test_any_server_error_is_json(self):
        self.assert_any_server_error_is_json(self.do_get)
        self.assert_any_server_error_is_json(self.do_patch, token=self.token)

    def test_openid_empty(self):
        self.assertRaises(NoReverseMatch, reverse, self.url_name, args=('',))

    def test_get_unauthenticated_gives_minimal(self):
        json_body = self.do_get()
        expected = handlers.get_minimal_account_data(self.account)
        self.assertEqual(json_body, self.to_json(expected))

    def test_patch_unauthenticated_401(self):
        self.do_patch(status_code=401)
        self.do_patch(self.data, status_code=401)

    def test_get_authenticated_with_django_gives_minimal(self):
        assert self.client.login(username=self.account.preferredemail.email,
                                 password=DEFAULT_USER_PASSWORD)
        json_body = self.do_get()
        expected = handlers.get_minimal_account_data(self.account)
        self.assertEqual(json_body, self.to_json(expected))

    def test_patch_authenticated_with_django_401(self):
        assert self.client.login(username=self.account.preferredemail.email,
                                 password=DEFAULT_USER_PASSWORD)
        self.do_patch(status_code=401)
        self.do_patch(self.data, status_code=401)

    def test_get_authenticated_with_oauth(self):
        json_body = self.do_get(token=self.token)
        expected = handlers.get_account_data(self.account)
        self.assertEqual(json_body, self.to_json(expected))

    def test_patch_authenticated_with_oauth_unprivileged(self):
        json_body = self.do_patch(self.data, token=self.token, status_code=403)
        self.assertEqual(json_body['code'], 'FORBIDDEN')
        self.assertEqual(json_body['extra'], {})

    def test_patch_authenticated_with_oauth_privileged(self):
        account = Account.objects.get(id=self.account.id)
        expected_status = getattr(AccountStatus, self.data['status'])
        assert account.status != expected_status
        self.do_patch(self.data, token=self.super_token)
        account.refresh_from_db()
        self.assertEqual(account.status, expected_status)
        self.assertEqual(account.status_comment, self.data['status_comment'])

    def test_patch_authenticated_with_oauth_privileged_invalid_data(self):
        invalid_data = (
            {},
            {'status': 'SUSPENDED'},
            {'status_comment': 'Valid value'},
            {'status': 'SUSPENDED', 'extra_field': 1},
            {'status': 'SUSPENDED', 'status_comment': ''},
            {'status': '', 'status_comment': 'Valid value'},
            {'status': '__dict__', 'status_comment': 'Valid value'},
            {'extra:field': 'something', 'status_comment': 'Valid value'},
        )
        for data in invalid_data:
            response = self.do_patch(
                data, token=self.super_token, status_code=400)
            self.assertEqual(
                response.get('message', ''), 'Invalid request data')
            self.assertEqual(response.get('code', ''), 'INVALID_DATA')

    def test_patch_authenticated_with_oauth_privileged_invalid_json(self):
        raw_data = "invalid ][ json"
        response = self.do_patch(
            raw_data=raw_data, token=self.super_token, status_code=400)
        self.assertEqual(
            response.get('message', ''), 'Invalid request data')
        self.assertEqual(response.get('code', ''), 'INVALID_DATA')

    def test_authenticated_openid_no_match(self):
        assert self.account.openid_identifier != 'foo'
        url = reverse(self.url_name, args=('foo',))
        self.do_get(url=url, token=self.token, status_code=404)
        self.do_patch(self.data, url=url, token=self.token, status_code=401)

    def test_authenticated_openid_no_match_but_has_permission(self):
        self.factory.add_permission_to_account(
            self.account, 'api_view_account_details')

        other = self.factory.make_account()
        assert self.account.openid_identifier != other.openid_identifier
        url = reverse(self.url_name, args=(other.openid_identifier,))
        json_body = self.do_get(url=url, token=self.token)

        expected = handlers.get_account_data(other)
        self.assertEqual(json_body, self.to_json(expected))

    def test_authenticated_openid_no_match_but_has_permission_404(self):
        self.factory.add_permission_to_account(
            self.account, 'api_view_account_details')

        assert self.account.openid_identifier != 'foo'
        assert Account.objects.active_by_openid('foo') is None
        url = reverse(self.url_name, args=('foo',))
        self.do_get(url=url, token=self.token, status_code=404)
        self.do_patch(self.data, url=url, token=self.super_token,
                      status_code=404)

    def test_get_account_data_no_preferred_email(self):
        self.account.emailaddress_set.all().delete()
        account = Account.objects.get(id=self.account.id)

        assert account.preferredemail is None
        result = handlers.get_account_data(account)

        self.assertIsNone(result['email'])

    def test_get_account_data_verified(self):
        account = self.account
        assert account.emailaddress_set.verified().count() > 0

        result = handlers.get_account_data(account)
        self.assertTrue(result['verified'])

    def test_get_account_data_no_verified(self):
        account = self.account
        account.emailaddress_set.verified().delete()

        result = handlers.get_account_data(account)
        self.assertFalse(result['verified'])

    def assert_account_data_expanded(self, expand_key, expand):
        url = self.url + '?' + urlencode({'expand': expand_key})
        json_body = self.do_get(url=url, token=self.token)

        expected = handlers.get_account_data(self.account, expand=expand)
        self.assertEqual(json_body, self.to_json(expected))

    def test_expanded_data(self):
        self.assert_account_data_expanded('true', True)

    def test_expanded_data_case_insensitive(self):
        self.assert_account_data_expanded('True', True)
        self.assert_account_data_expanded('TRUE', True)
        self.assert_account_data_expanded('trUE', True)

    def test_not_expanded_data(self):
        self.assert_account_data_expanded('false', False)

    def test_not_expanded_data_other_values(self):
        self.assert_account_data_expanded('False', False)
        self.assert_account_data_expanded('0', False)
        self.assert_account_data_expanded('no', False)
        self.assert_account_data_expanded(' ', False)


class AnonymousAccountRegistrationHandlerTestCase(BaseTestCase):

    url = reverse('api-registration')

    def setUp(self):
        super(AnonymousAccountRegistrationHandlerTestCase, self).setUp()
        self.data = {
            'email': self.factory.make_email_address(),
            'password': 'asdfASDF1',
            'displayname': 'Ricardo the Magnificent',
        }

    def assert_field_required(self, field):
        parent = super(AnonymousAccountRegistrationHandlerTestCase, self)
        parent.assert_field_required(data=self.data, field=field)

    def assert_account_count(self, account_count):
        account = Account.objects.get_by_email(self.data['email'])
        self.assertEqual(account is None, account_count == 0)

    def assert_account_created(self):
        account = Account.objects.get_by_email(self.data['email'])
        self.assertIsNotNone(account)
        self.assertEqual(account.displayname, self.data['displayname'])
        self.assertEqual(account.status, AccountStatus.ACTIVE)
        self.assertEqual(account.is_verified, False)
        self.assertEqual(
            account.creation_rationale, AccountCreationRationale.API_CREATED)
        self.assertEqual(
            account.creation_source, self.data.get('creation_source'))
        return account

    def test_any_server_error_is_json(self):
        self.assert_any_server_error_is_json(self.do_post, data=self.data)

    def test_registration_handler_invalid_data(self):
        data = {
            'email': 'x@x.com',
            'password': 'y',
            'displayname': 'foo',
        }
        extra = {'password': ['Password must be at least 8 characters long']}
        self.assert_bad_request(data, extra=extra)

    def test_no_email(self):
        self.assert_field_required('email')

    def test_no_password(self):
        self.assert_field_required('password')

    def test_no_displayname(self):
        self.assert_field_required('displayname')

    def test_registration_already_registered(self):
        self.data['email'] = self.account.preferredemail.email

        json_body = self.do_post(self.data, status_code=409)

        self.assertEqual(json_body['code'], "ALREADY_REGISTERED")
        self.assertIn("already registered", json_body['message'])
        self.assertIn('email', json_body['extra'])
        self.assertIn(self.data['email'], json_body['extra']['email'])

    def test_registration_success(self):
        json_body = self.do_post(self.data, status_code=201)

        self.assertIn('href', json_body)
        account = self.assert_account_created()
        self.assertEqual(json_body['openid'], account.openid_identifier)
        self.assertEqual(json_body['email'], self.data['email'])
        self.assertEqual(json_body['displayname'], self.data['displayname'])
        self.assertEqual(json_body['status'], 'Active')
        self.assertEqual(json_body['verified'], False)
        self.assertEqual(len(json_body['emails']), 1)
        self.assertIn(quote(self.data['email'], safe='@'),
                      json_body['emails'][0]['href'])

        # user can login after registration
        loggedin = self.client.login(
            username=self.data['email'], password=self.data['password'])
        self.assertTrue(loggedin)

    def test_registration_with_leaked_credentials(self):
        data = {
            'email': 'foo@example.com',
            'password': 'compromised',
            'displayname': 'foo',
        }

        with patch('api.v20.registration.register') as mock_register:
            mock_register.side_effect = PasswordPolicyError('password leaked',
                                                            None)
            json_body = self.do_post(data, status_code=403)

        self.assertEqual(json_body['code'], "PASSWORD_POLICY_ERROR")
        self.assertEqual("Your password doesn't meet security constraints.",
                         json_body['message'])
        self.assertEqual('password leaked', json_body['extra']['reason'])

    def test_creation_source_used(self):
        self.data['creation_source'] = 'zaraza created foo'

        self.do_post(self.data, status_code=201)

        self.assert_account_created()

    def test_throttle_only_by_ip(self):
        data = self.data.copy()
        other_email = 'other@email.com'
        assert other_email != data['email']

        with self.settings(THROTTLE_MAX_REQUESTS=2):
            response = self.do_post(data, status_code=201)
            self.assert_account_created()

            data['email'] = other_email
            response = self.do_post(data, status_code=429)
            self.assertEqual(response['code'], 'TOO_MANY_REQUESTS')

    def test_invalid_email(self):
        self.data['email'] = 'with space@example.com'
        self.assert_bad_request(
            data=self.data, extra={'email': ['Invalid email']})


@override_settings(CAPTCHA_PUBLIC_KEY='public', CAPTCHA_PRIVATE_KEY='private')
class AnonymousAccountRegistrationWithCaptchaHandlerTestCase(BaseTestCase):

    url = reverse('api-registration')

    def setUp(self):
        super(AnonymousAccountRegistrationWithCaptchaHandlerTestCase,
              self).setUp()
        p = switches(CAPTCHA=True, CAPTCHA_NEW_ACCOUNT=True)
        p.patch()
        self.addCleanup(p.unpatch)

        self.captcha_response = captcha.CaptchaResponse(
            code=42, response=StringIO("challenge: '999'"))
        self.mock_captcha_open = self.patch(
            'api.v20.handlers.Captcha._open',
            return_value=self.captcha_response)

        self.data = {
            'email': self.factory.make_email_address(),
            'password': 'asdfASDF1',
            'displayname': 'Ricardo the Magnificent',
            'captcha_id': '999',
            'captcha_solution': 'foo bar',
        }

    def test_register_captcha_required(self):
        captcha_data = {
            'captcha_id': '999',
            'image_url': settings.CAPTCHA_IMAGE_URL_PATTERN % '999'}
        del self.data['captcha_id']
        del self.data['captcha_solution']

        json_body = self.do_post(self.data, status_code=401)

        self.assertEqual(json_body['code'], "CAPTCHA_REQUIRED")
        self.assertIn('A captcha challenge is required', json_body['message'])
        self.assertIsNone(Account.objects.get_by_email(self.data['email']))
        self.assertEqual(json_body['extra'], captcha_data)

    def test_register_captcha_success(self):
        self.captcha_response.response = StringIO('true\nok')
        json_body = self.do_post(self.data, status_code=201)

        url_request = self.mock_captcha_open.call_args[0][0]
        self.assertEqual(
            url_request.get_full_url(), settings.CAPTCHA_VERIFY_URL)
        data = dict(parse_qsl(url_request.data))
        expected = {
            'challenge': '999', 'privatekey': settings.CAPTCHA_PRIVATE_KEY,
            'remoteip': '127.0.0.1', 'response': 'foo bar'}
        self.assertEqual(data, expected)

        self.assertIn('openid', json_body)
        self.assertIn('href', json_body)
        self.assertEqual(json_body['email'], self.data['email'])
        self.assertEqual(json_body['displayname'], self.data['displayname'])
        self.assertEqual(json_body['status'], 'Active')
        self.assertEqual(len(json_body['emails']), 1)
        self.assertIn(quote(self.data['email'], safe='@'),
                      json_body['emails'][0]['href'])

    def test_register_captcha_failure(self):
        self.captcha_response.is_error = False

        self.data['captcha_id'] = '999'
        self.data['captcha_solution'] = 'foo bar'

        json_body = self.do_post(self.data, status_code=403)

        url_request = self.mock_captcha_open.call_args[0][0]
        self.assertEqual(
            url_request.get_full_url(), settings.CAPTCHA_VERIFY_URL)
        data = dict(parse_qsl(url_request.data))
        expected = {
            'challenge': '999', 'privatekey': settings.CAPTCHA_PRIVATE_KEY,
            'remoteip': '127.0.0.1', 'response': 'foo bar'}
        self.assertEqual(data, expected)

        self.assertEqual(json_body['code'], "CAPTCHA_FAILURE")
        self.assertIn(
            'Failed response to captcha challenge.', json_body['message'])
        self.assertIsNone(Account.objects.get_by_email(self.data['email']))

    def test_register_captcha_whitelist(self):
        self.data['email'] = 'canonicaltest@gmail.com'
        self.data['captcha_id'] = '999'
        self.data['captcha_solution'] = 'foo bar'

        with self.settings(**OVERRIDES):
            self.do_post(self.data, status_code=201)

        self.assertIsNotNone(Account.objects.get_by_email(self.data['email']))
        self.assertFalse(self.mock_captcha_open.called)

    def test_register_captcha_whitelist_with_uuid(self):
        self.data['email'] = 'canonicaltest+something@gmail.com'
        self.data['captcha_id'] = '999'
        self.data['captcha_solution'] = 'foo bar'

        with self.settings(**OVERRIDES):
            self.do_post(self.data, status_code=201)

        self.assertIsNotNone(Account.objects.get_by_email(self.data['email']))
        self.assertFalse(self.mock_captcha_open.called)

    def test_register_captcha_whitelist_fail(self):
        self.data['captcha_id'] = '999'
        self.data['captcha_solution'] = 'foo bar'
        self.data['email'] = 'notcanonicaltest@gmail.com'

        self.captcha_response.is_error = False
        self.captcha_response.response = StringIO('false\nmessage')

        with self.settings(**OVERRIDES):
            self.do_post(self.data, status_code=403)

        self.assertIsNone(Account.objects.get_by_email(self.data['email']))
        self.assertTrue(self.mock_captcha_open.called)

    def test_register_captcha_http_error(self):
        response = captcha.CaptchaResponse(
            500, None, "", "Server error", "Unexpected recaptcha error")
        error = captcha.VerifyCaptchaError(response)
        self.mock_captcha_open.side_effect = error

        self.data['captcha_id'] = '999'
        self.data['captcha_solution'] = 'foo bar'

        json_body = self.do_post(self.data, status_code=502)

        self.assertEqual(json_body['code'], "CAPTCHA_ERROR")
        self.assertIn('Unable to get a valid response from reCaptcha service',
                      json_body['message'])
        extra = json_body['extra']
        self.assertEqual(extra['recaptcha_reason'], "Server error")
        self.assertEqual(extra['recaptcha_status_code'], 500)
        self.assertEqual(extra['recaptcha_body'], "Unexpected recaptcha error")
        self.assertIsNone(Account.objects.get_by_email(self.data['email']))

    def test_register_captcha_network_error(self):
        response = captcha.CaptchaResponse(111, None, "", "Connection refused")
        error = captcha.VerifyCaptchaError(response)
        self.mock_captcha_open.side_effect = error

        self.data['captcha_id'] = '999'
        self.data['captcha_solution'] = 'foo bar'

        json_body = self.do_post(self.data, status_code=502)

        self.assertEqual(json_body['code'], "CAPTCHA_ERROR")
        self.assertIn('Unable to get a valid response from reCaptcha service',
                      json_body['message'])
        extra = json_body['extra']
        self.assertEqual(extra['recaptcha_reason'], "Connection refused")
        self.assertEqual(extra['recaptcha_status_code'], None)
        self.assertEqual(extra['recaptcha_body'], None)
        self.assertIsNone(Account.objects.get_by_email(self.data['email']))

    def test_timeline_for_captcha_generation(self):
        timeline = Timeline()
        meta = {
            'timeline.timeline': timeline,
        }
        data = self.data.copy()
        data.pop('captcha_id')
        data.pop('captcha_solution')
        data['create_captcha'] = True

        self.do_post(data, status_code=401, **meta)

        self.assertEqual(1, len(timeline.actions))
        self.assertEqual('captcha-new', timeline.actions[0].category)
        self.assertEqual(self.mock_captcha_open.call_args[0][0],
                         timeline.actions[0].detail)

    def test_timeline_for_captcha_verification(self):
        self.mock_captcha_open.return_value = captcha.CaptchaResponse(
            500, StringIO('false\nsomething'))
        timeline = Timeline()
        meta = {
            'timeline.timeline': timeline,
        }

        self.do_post(self.data, status_code=403, **meta)

        self.assertEqual(1, len(timeline.actions))
        self.assertEqual('captcha-verify', timeline.actions[0].category)


class AccountRegistrationHandlerTestCase(BaseTestCase):
    url = reverse('api-registration')

    def setUp(self):
        super(AccountRegistrationHandlerTestCase, self).setUp()
        account = self.factory.make_account(
            email='super@user.com', permissions=['api_edit_account_details'])
        self.super_token = account.create_oauth_token('super-token')
        self.data = {
            'email': self.factory.make_email_address(),
            'displayname': 'Foo Bar Baz',
        }
        self.now = now()

    def assert_field_required(self, field):
        super(AccountRegistrationHandlerTestCase, self).assert_field_required(
            data=self.data, field=field, token=self.super_token)

    def assert_account_data(self, json_body, new_account=True):
        account = Account.objects.get_by_email(self.data['email'])
        self.assertIsNotNone(account)
        if new_account:
            self.assertEqual(
                account.creation_rationale,
                AccountCreationRationale.API_CREATED)
            self.assertGreaterEqual(account.date_created, self.now)
        else:
            self.assertEqual(
                account.creation_rationale,
                AccountCreationRationale.OWNER_CREATED_LAUNCHPAD)
            self.assertLessEqual(account.date_created, self.now)

        self.assertEqual(
            account.creation_source, self.data.get('creation_source'))
        self.assertIn('href', json_body)
        self.assertEqual(json_body['openid'], account.openid_identifier)
        self.assertEqual(json_body['email'], self.data['email'])
        self.assertEqual(json_body['displayname'], self.data['displayname'])
        self.assertEqual(json_body['status'], 'Active')
        self.assertEqual(json_body['verified'], True)
        self.assertEqual(len(json_body['emails']), 1)
        self.assertIn(quote(self.data['email'], safe='@'),
                      json_body['emails'][0]['href'])
        return account

    def test_any_server_error_is_json(self):
        self.assert_any_server_error_is_json(
            self.do_post, data=self.data, token=self.super_token)

    def test_no_email(self):
        self.assert_field_required('email')

    def test_no_displayname(self):
        self.assert_field_required('displayname')

    def test_registration_success(self):
        json_body = self.do_post(
            self.data, status_code=201, token=self.super_token)

        self.assert_account_data(json_body)

    def test_registration_already_registered(self):
        self.data['email'] = self.account.preferredemail.email

        json_body = self.do_post(
            self.data, status_code=200, token=self.super_token)

        # display name does not change
        self.data['displayname'] = self.account.displayname
        self.assert_account_data(json_body, new_account=False)

    def test_password_ignored_if_given(self):
        self.data['password'] = '12345678'

        json_body = self.do_post(
            self.data, status_code=201, token=self.super_token)

        self.assert_account_data(json_body)
        # user can not login
        loggedin = self.client.login(
            username=self.data['email'], password=self.data['password'])
        self.assertFalse(loggedin)

    def test_regular_v2_token_is_considered_anonymous(self):
        not_super_token = self.account.create_oauth_token('test')
        self.assert_bad_request(
            self.data, token=not_super_token,
            extra={'password': [handlers.FIELD_REQUIRED]})
        # no account was created
        self.assertIsNone(Account.objects.get_by_email(self.data['email']))

    def test_creation_source_used(self):
        self.data['creation_source'] = 'zaraza created foo by superuser'

        json_body = self.do_post(
            self.data, status_code=201, token=self.super_token)

        self.assert_account_data(json_body)

    def test_not_throttled(self):
        with self.settings(THROTTLE_MAX_REQUESTS=2):
            for i in range(3):
                self.data['email'] = 'foo-%i@example.com' % i
                json_body = self.do_post(
                    self.data, status_code=201, token=self.super_token)
                self.assert_account_data(json_body)


class EmailsHandlerTestCase(BaseTestCase):
    url_name = 'api-email'

    def setUp(self):
        super(EmailsHandlerTestCase, self).setUp()
        self.email = self.account.preferredemail
        self.url = reverse(self.url_name, kwargs=dict(email=self.email.email))

        account = self.factory.make_account(
            email='super@user.com', permissions=['api_edit_account_details'])
        self.super_token = account.create_oauth_token('super-token')
        self.data = {'status': 'SUSPENDED', 'status_comment': 'Test comment'}

    def assert_email_result(self, result, email):
        expected = dict(
            email=email.email, verified=True,
            date_created=email.date_created,
            href=reverse(self.url_name, kwargs=dict(email=email.email)),
        )
        self.assertEqual(result, self.to_json(expected))

    def test_any_server_error_is_json(self):
        self.assert_any_server_error_is_json(self.do_get, token=self.token)
        self.assert_any_server_error_is_json(self.do_patch, token=self.token)

    def test_email_empty(self):
        self.assertRaises(
            NoReverseMatch, reverse, self.url_name, kwargs=dict(email=''))

    def test_get_unauthenticated_401(self):
        self.do_get(status_code=401)

    def test_patch_unauthenticated_401(self):
        self.do_patch(status_code=401)
        self.do_patch(self.data, status_code=401)

    def test_get_authenticated_with_django_401(self):
        assert self.client.login(username=self.account.preferredemail.email,
                                 password=DEFAULT_USER_PASSWORD)
        self.do_get(status_code=401)

    def test_patch_authenticated_with_django_401(self):
        assert self.client.login(username=self.account.preferredemail.email,
                                 password=DEFAULT_USER_PASSWORD)
        self.do_patch(status_code=401)
        self.do_patch(self.data, status_code=401)

    def test_get_authenticated_with_oauth(self):
        json_body = self.do_get(token=self.token)
        self.assert_email_result(json_body, self.email)

    def test_patch_authenticated_with_oauth_unprivileged(self):
        json_body = self.do_patch(self.data, token=self.token, status_code=403)
        self.assertEqual(json_body['code'], 'FORBIDDEN')
        self.assertEqual(json_body['extra'], {})

    def test_patch_authenticated_with_oauth_privileged(self):
        account = Account.objects.get(id=self.account.id)
        expected_status = getattr(AccountStatus, self.data['status'])
        assert account.status != expected_status
        self.do_patch(self.data, token=self.super_token)
        account.refresh_from_db()
        self.assertEqual(account.status, expected_status)
        self.assertEqual(account.status_comment, self.data['status_comment'])

    def test_patch_authenticated_with_oauth_privileged_invalid_data(self):
        invalid_data = (
            {},
            {'status': 'SUSPENDED'},
            {'status_comment': 'Valid value'},
            {'status': 'SUSPENDED', 'extra_field': 1},
            {'status': 'SUSPENDED', 'status_comment': ''},
            {'status': '', 'status_comment': 'Valid value'},
            {'status': '__dict__', 'status_comment': 'Valid value'},
            {'extra:field': 'something', 'status_comment': 'Valid value'},
        )
        for data in invalid_data:
            response = self.do_patch(
                data, token=self.super_token, status_code=400)
            self.assertEqual(
                response.get('message', ''), 'Invalid request data')
            self.assertEqual(response.get('code', ''), 'INVALID_DATA')

    def test_patch_authenticated_with_oauth_privileged_invalid_json(self):
        raw_data = "invalid ][ json"
        response = self.do_patch(
            raw_data=raw_data, token=self.super_token, status_code=400)
        self.assertEqual(
            response.get('message', ''), 'Invalid request data')
        self.assertEqual(response.get('code', ''), 'INVALID_DATA')

    def test_can_get_others_if_elevated_permission(self):
        other_email = self.factory.make_account().preferredemail
        self.factory.add_permission_to_account(
            self.account, 'api_view_account_details')

        url = reverse(self.url_name, kwargs=dict(email=other_email.email))
        json_body = self.do_get(url=url, token=self.token)
        self.assert_email_result(json_body, other_email)

    def test_forbidden_if_not_owner_not_elevated_permission_on_get(self):
        other_email = self.factory.make_account().preferredemail

        url = reverse(self.url_name, kwargs=dict(email=other_email.email))
        self.do_get(url=url, token=self.token, status_code=401)
        self.do_patch(self.data, url=url, token=self.token, status_code=401)

    def test_email_does_not_exist(self):
        self.factory.add_permission_to_account(
            self.account, 'api_view_account_details')
        url = reverse(self.url_name, kwargs=dict(email='foobarbaz@no.com'))
        self.do_get(url=url, token=self.token, status_code=404)

        self.factory.add_permission_to_account(
            self.account, 'api_edit_account_details')
        self.do_patch(self.data, url=url, token=self.token, status_code=404)

    def test_delete_authenticated_with_oauth(self):
        response = self.get_response(
            token=self.token, method='DELETE', status_code=204)

        self.assertEqual(response.content, '')
        self.assertEqual(
            EmailAddress.objects.filter(email=self.email.email).count(), 0)

    def test_can_delete_others_if_elevated_permission(self):
        other_email = self.factory.make_account().preferredemail
        self.factory.add_permission_to_account(
            self.account, 'api_edit_account_details')

        url = reverse(self.url_name, kwargs=dict(email=other_email.email))
        response = self.get_response(
            url=url, token=self.token, method='DELETE', status_code=204)

        self.assertEqual(response.content, '')
        self.assertEqual(
            EmailAddress.objects.filter(email=other_email.email).count(), 0)

    def test_forbidden_if_not_owner_not_elevated_permission_on_delete(self):
        other_email = self.factory.make_account().preferredemail

        url = reverse(self.url_name, kwargs=dict(email=other_email.email))
        self.get_response(
            url=url, token=self.token, method='DELETE', status_code=401)
        self.assertEqual(
            EmailAddress.objects.filter(email=other_email.email).count(), 1)

    def test_delete_email_does_not_exist(self):
        self.factory.add_permission_to_account(
            self.account, 'api_edit_account_details')

        url = reverse(self.url_name, kwargs=dict(email='foobarbaz@no.com'))
        self.get_response(
            url=url, token=self.token, method='DELETE', status_code=404)


class RequestsTestCase(BaseTestCase, POSTOnlyResourceMixin):

    url_name = 'api-requests'
    default_http_method = 'GET'

    def setUp(self):
        super(RequestsTestCase, self).setUp()
        self.url = reverse(self.url_name)
        self.external_url = 'http://example.com/foo/bar?some=1&nothing=foo'
        self.mock_logger = self.patch('api.v20.handlers.logger')

    def build_data(
            self, external_url=None, token=None, consumer=None,
            as_query_string=False, plaintext=False, **kwargs):
        if external_url is None:
            external_url = self.external_url
        if token is None:
            token = self.token
        uri, header, body = self.factory.make_oauth_authorization_from_token(
            external_url, token, http_method=self.default_http_method,
            consumer=consumer, as_query_string=as_query_string,
            plaintext=plaintext)
        data = dict(
            http_url=external_url, http_method=self.default_http_method)
        if as_query_string:
            data['http_url'] = uri
        else:
            data['authorization'] = header['Authorization']

        # update data with kwargs to allow tests mess up with parameters
        data.update(kwargs)
        return data

    def assert_request_response(
            self, response, is_valid, account_verified=False,
            openid_identifier=None, error=None, data=None):
        if not is_valid:
            assert not account_verified
        expected = dict(is_valid=is_valid, account_verified=account_verified)
        if error:
            expected['error'] = error
        # now is_valid implies 'identifier' is present
        if openid_identifier is None:
            openid_identifier = self.account.openid_identifier
        if is_valid:
            expected['identifier'] = openid_identifier
        self.assertEqual(response, expected)

        if error is None:
            self.assertFalse(self.mock_logger.exception.called)
        else:
            self.mock_logger.exception.assert_called_once_with(
                'RequestsHandler.create: could not verify request (data %r):',
                data)

    def assert_workaround_warning_logged(
            self, http_url, query_string, authorization):
        qs_workaround_warning = (
            'RequestsHandler.create: caller for validating http_url %r sent '
            'query_string %r or did not send authorization (got %r), will '
            'merge it into original url.')
        self.mock_logger.warning.assert_called_once_with(
            qs_workaround_warning, http_url, query_string, authorization)

    def assert_token_filtered_warning_logged(
            self, authorization, http_url, http_method):
        self.mock_logger.warning.assert_called_with(
            'RequestsHandler.create: NOT validating (otherwise valid) '
            'signature %r because the token is not allowed to access url %r '
            'via %r', authorization, http_url, http_method)

    def do_post_and_assert_statsd_metric_recorded(self, data, suffix):
        with patch('api.v20.handlers.time') as mock_time:
            mock_time.time.side_effect = [2000, 2001]
            with patch('api.v20.handlers.statsd') as mock_statsd:
                self.do_post(data)

        mock_statsd.timing.assert_called_once_with(
            'view.api.v20.handlers.RequestsHandler.POST.' + suffix, 1000)

    def test_invalid_if_not_method_given(self):
        data = self.build_data()
        data.pop('http_method')

        response = self.do_post(data)

        self.assert_request_response(response, is_valid=False)

    def test_invalid_if_method_none(self):
        data = self.build_data(http_method=None)

        response = self.do_post(data)

        self.assert_request_response(response, is_valid=False)

    def test_invalid_if_not_url_given(self):
        data = self.build_data()
        data.pop('http_url')

        response = self.do_post(data)

        self.assert_request_response(response, is_valid=False)

    def test_invalid_if_url_none(self):
        data = self.build_data(http_url=None)

        response = self.do_post(data)

        self.assert_request_response(response, is_valid=False)

    def test_url_encoded(self):
        data = self.build_data(external_url='http://foo/ñoño ñandú/')

        response = self.do_post(data)

        self.assert_request_response(
            response, is_valid=True, account_verified=True)

    def test_invalid_if_verify_request_fails(self):
        self.patch(
            'api.v20.handlers.auth.validate_oauth_signature',
            side_effect=KeyError('\xf1'))
        data = self.build_data()

        response = self.do_post(data)

        self.assert_request_response(
            response, is_valid=False, account_verified=False,
            error="KeyError: u'\\xf1'", data=data)

    def test_invalid_if_header_none(self):
        data = self.build_data(as_query_string=False, authorization=None)

        response = self.do_post(data)

        self.assert_request_response(response, is_valid=False)

    def test_invalid_if_not_header_given(self):
        data = self.build_data(as_query_string=False)
        data.pop('authorization')

        response = self.do_post(data)

        self.assert_request_response(response, is_valid=False)

    def test_invalid_if_no_header_or_query_string(self):
        data = self.build_data()
        del data['authorization']

        response = self.do_post(data)

        self.assert_request_response(response, is_valid=False)

    def test_invalid_if_bad_query_string(self):
        data = self.build_data(query_string='oauth=invalid')
        # make sure there is no authorization header
        del data['authorization']

        response = self.do_post(data)

        self.assert_request_response(response, is_valid=False)

    def test_invalid_if_oauth_query_string_invalid_param(self):
        data = self.build_data(as_query_string=True)
        # tamper with query string parameters
        data['http_url'] = data['http_url'].replace(
            'oauth_consumer_key=%s' % self.token.consumer.openid_identifier,
            'oauth_consumer=foo')

        # make sure we don't have an authorization header set
        assert 'authorization' not in data

        # do a post with correct params
        response = self.do_post(data)

        self.assert_request_response(response, is_valid=False)

    def test_url_mismatch(self):
        new_url = self.external_url[:-1]
        assert new_url != self.external_url
        data = self.build_data()
        data['http_url'] = new_url
        response = self.do_post(data)

        self.assert_request_response(response, is_valid=False)

    def test_method_mismatch(self):
        data = self.build_data()
        data['http_method'] = self.default_http_method * 2
        response = self.do_post(data)

        self.assert_request_response(response, is_valid=False)

    def test_valid_request(self):
        # do a post with correct params
        data = self.build_data()
        response = self.do_post(data)

        self.assert_request_response(
            response, is_valid=True, account_verified=True)

    def test_valid_request_unverified_user(self):
        self.account.verified_emails().delete()
        assert not self.account.is_verified

        # do a post with correct params
        data = self.build_data()
        response = self.do_post(data)

        self.assert_request_response(
            response, is_valid=True, account_verified=False)

    def test_valid_request_for_non_existent_account(self):
        data = self.build_data()
        self.account.delete()
        # do a post with correct params for a non existent account
        response = self.do_post(data)

        self.assert_request_response(response, is_valid=False)

    def test_valid_request_for_inactive_account(self):
        for status, name in AccountStatus._get_choices():
            if status == AccountStatus.ACTIVE:
                continue
            self.account.status = status
            self.account.save()

            # restore oauth token since some account changes invalidate those
            token = self.account.create_oauth_token('another')
            data = self.build_data(token=token)

            # do a post with correct params for an inactive account
            response = self.do_post(data)

            self.assert_request_response(response, is_valid=False)

    def test_valid_request_with_only_query_string(self):
        data = self.build_data(as_query_string=True)
        assert 'authorization' not in data

        # do a post with correct params
        response = self.do_post(data)

        self.assert_request_response(
            response, is_valid=True, account_verified=True)

    def test_query_string_merged_with_url(self):
        data = self.build_data(as_query_string=True)

        http_url = data.pop('http_url')
        parts = urlparse(http_url)
        new_url = urlunparse(
            (parts.scheme, parts.netloc, parts.path, parts.params, '',
             parts.fragment))
        data['http_url'] = new_url
        data['query_string'] = parts.query

        # do a post with correct params, but url had oauth qs chopped off
        response = self.do_post(data)

        self.assert_request_response(
            response, is_valid=True, account_verified=True)
        self.assert_workaround_warning_logged(new_url, parts.query, None)

    def test_query_string_not_duplicated_when_merged_with_url(self):
        data = self.build_data(as_query_string=True)

        http_url = data['http_url']
        parts = urlparse(http_url)
        data['query_string'] = parts.query
        assert parts.query in http_url

        # do a post with correct params, but url had oauth qs chopped off
        response = self.do_post(data)

        self.assert_request_response(
            response, is_valid=True, account_verified=True)

        # the final url is a new url with query string params in any order,
        # because it was a set as an intermediary step. Use ANY instead
        # fighting ordering.
        self.assert_workaround_warning_logged(http_url, parts.query, None)

    def test_query_string_merged_if_authorization_header_found(self):
        query_string = 'include_deleted=True'
        assert query_string not in self.external_url
        url = self.external_url + '&' + query_string
        data = self.build_data(external_url=url, query_string=query_string)
        assert data['authorization']

        # do a post with correct params
        response = self.do_post(data)

        self.assert_request_response(
            response, is_valid=True, account_verified=True)
        self.assert_workaround_warning_logged(
            url, query_string, data['authorization'])

    def test_query_string_merged_non_ascii(self):
        params = [
            ('name'.encode('utf-8'), 'ñoño ñandú'.encode('utf-8')),
            ('name'.encode('utf-8'), 'algo más'.encode('utf-8')),
        ]
        query_string = [
            ('heart'.encode('utf-8'), '☙ ♡ ♥ ღ 💌 🏩 '.encode('utf-8')),
            ('música'.encode('utf-8'), '🎵♩🎶'.encode('utf-8')),
        ]
        whole_qs = urlencode(params + query_string)
        url = 'http://foobarbaz/please?' + whole_qs

        data = self.build_data(
            external_url=url, query_string=urlencode(query_string))
        assert data['authorization'] is not None

        response = self.do_post(data)

        self.assert_request_response(
            response, is_valid=True, account_verified=True)
        self.assert_workaround_warning_logged(
            url, urlencode(query_string), data['authorization'])

    def test_http_url_non_ascii(self):
        url = 'http://moño.com/å ä ö/'
        data = self.build_data(external_url=url)

        response = self.do_post(data)

        self.assert_request_response(
            response, is_valid=True, account_verified=True)

    def test_replay_attack(self):
        data = self.build_data()
        response = self.do_post(data)
        self.assert_request_response(
            response, is_valid=True, account_verified=True)

        response = self.do_post(data)
        self.assert_request_response(response, is_valid=False)

    def test_consumer_and_token_v2_for_the_same_account(self):
        """Ensure consumer matches the given V2 token."""
        other_consumer = self.factory.make_account()
        other_token = other_consumer.create_oauth_token('other')

        # build an invalid signature, use one consumer and token for another
        data = self.build_data(token=other_token, consumer=self.account)
        response = self.do_post(data)
        self.assert_request_response(response, is_valid=False)

        # same swapping consumer and token
        data = self.build_data(token=self.token, consumer=other_consumer)
        response = self.do_post(data)
        self.assert_request_response(response, is_valid=False)

    def test_consumer_and_token_v1_for_the_same_account(self):
        """Ensure consumer matches the given V1 token."""
        other_consumer = self.factory.make_account()
        other_token = other_consumer.create_v1_oauth_token('other')

        # choose any whitelisted URL for V1
        url = whitelist.URLS[0]

        # build an invalid signature, use one consumer and token for another
        data = self.build_data(url, token=other_token, consumer=self.account)
        response = self.do_post(data)
        self.assert_request_response(response, is_valid=False)

        # same swapping consumer and token
        data = self.build_data(url, token=self.token, consumer=other_consumer)
        response = self.do_post(data)
        self.assert_request_response(response, is_valid=False)

    def test_v1_token_is_allowed_only_if_whitelisted(self):
        v1_token = self.account.create_v1_oauth_token('v1 token')

        for url in whitelist.URLS:
            data = self.build_data(url, token=v1_token)
            response = self.do_post(data)
            self.assert_request_response(
                response, is_valid=True, account_verified=True)

    def test_v1_token_is_allowed_only_if_whitelisted_matches_prefix(self):
        v1_token = self.account.create_v1_oauth_token('v1 token')

        for url in whitelist.URLS:
            url += '?foo=bar&zaraza=1'
            data = self.build_data(url, token=v1_token)
            response = self.do_post(data)
            self.assert_request_response(
                response, is_valid=True, account_verified=True)
            self.assertEqual(self.mock_logger.mock_calls, [])

    def test_v1_token_not_allowed_if_not_whitelisted_bad_prefix(self):
        v1_token = self.account.create_v1_oauth_token('v1 token')

        for url in whitelist.URLS:
            parts = urlparse(url)
            url = urlunparse(
                (parts.scheme, parts.netloc, '/foo' + parts.path, parts.params,
                 parts.query, parts.fragment))
            data = self.build_data(url, token=v1_token)
            response = self.do_post(data)
            self.assert_request_response(response, is_valid=False)
            self.assert_token_filtered_warning_logged(
                data['authorization'], url, self.default_http_method)

    def test_v1_token_not_allowed_url_is_prefix_of_whitelisted_url(self):
        v1_token = self.account.create_v1_oauth_token('v1 token')

        for url in whitelist.URLS:
            # remove the last /
            url = url.rsplit('/', 1)[0]
            data = self.build_data(url, token=v1_token)
            response = self.do_post(data)
            self.assert_request_response(response, is_valid=False)

            # remove the last path
            url = url.rsplit('/', 1)[0]
            data = self.build_data(url, token=v1_token)
            response = self.do_post(data)
            self.assert_request_response(response, is_valid=False)
            self.assert_token_filtered_warning_logged(
                data['authorization'], url, self.default_http_method)

    def test_v1_token_not_allowed_if_not_whitelisted(self):
        v1_token = self.account.create_v1_oauth_token('v1 token')

        url = 'http://example.com'
        assert url not in whitelist.URLS

        data = self.build_data(url, token=v1_token)
        response = self.do_post(data)
        self.assert_request_response(response, is_valid=False)
        self.assert_token_filtered_warning_logged(
            data['authorization'], url, self.default_http_method)

    def test_token_web_login_is_invalid(self):
        weblogin_token = self.account.create_oauth_token(
            'webonly', scope=TokenScope.WEB_LOGIN)
        data = self.build_data(token=weblogin_token)
        response = self.do_post(data)
        self.assert_request_response(response, is_valid=False)

    def test_token_web_login_is_invalid_even_if_url_whitelisted(self):
        weblogin_token = self.account.create_oauth_token(
            'foo', scope=TokenScope.WEB_LOGIN)
        for url in whitelist.URLS:
            data = self.build_data(url, token=weblogin_token)
            response = self.do_post(data)
            self.assert_request_response(response, is_valid=False)

    def test_many_query_strings_same_key(self):
        url = 'http://example.com/?name=foo&name=bar'
        data = self.build_data(url)

        response = self.do_post(data)

        self.assert_request_response(
            response, is_valid=True, account_verified=True)

    def test_num_queries_signature_valid_v2(self):
        data = self.build_data()

        # Right now, 3 queries for:

        # 1- get client secret from client key (filter active accounts)
        # SELECT * FROM account WHERE openid_identifier = 'key' AND status = 20

        # 2- get token secret from token key
        # SELECT * FROM identityprovider_token WHERE key = 'token key'

        # 3- if signature valid, check if account is verified from emails
        # SELECT DISTINCT (1) AS "a", account.*
        # FROM account INNER JOIN emailaddress
        # ON (account.id = emailaddress.account)
        # WHERE (emailaddress.status IN (2, 4) AND
        #        account.openid_identifier = 'key')
        with self.assertNumQueries(3):
            response = self.do_post(data)

        self.assert_request_response(
            response, is_valid=True, account_verified=True)

    def test_num_queries_signature_invalid_v2(self):
        # build an invalid signature
        data = self.build_data()
        # delete the token from the DB
        self.token.delete()

        # Request not valid, the 3rd query for account verified is not issued,
        # but the fallback V1 token is queried for.
        with self.assertNumQueries(3):
            response = self.do_post(data)

        self.assert_request_response(response, is_valid=False)

    def test_num_queries_signature_valid_v2_bad_scope(self):
        weblogin_token = self.account.create_oauth_token(
            'foo', scope=TokenScope.WEB_LOGIN)
        data = self.build_data(token=weblogin_token)

        # Request not valid, the 3rd query for account verified is not issued
        with self.assertNumQueries(2):
            response = self.do_post(data)

        self.assert_request_response(response, is_valid=False)

    def test_num_queries_signature_valid_v1(self):
        v1_token = self.account.create_v1_oauth_token('v1 token')

        # One extra query for old V1 tokens because we need to first exhaust
        # new tokens before fetchins old ones.
        for url in whitelist.URLS:
            data = self.build_data(external_url=url, token=v1_token)
            with self.assertNumQueries(4):
                response = self.do_post(data)
            self.assert_request_response(
                response, is_valid=True, account_verified=True)

    def test_predefined_signature_non_ascii_params(self):
        params = [
            ('name', 'ñoño ñandú'.encode('utf-8')),
            ('email', 'foo@foo.com'.encode('utf-8')),
            ('name', 'algo más'.encode('utf-8')),
        ]
        # URL params have to be urlencoded
        url = 'http://sso-trusty:8000/error?' + urlencode(params)

        account = self.factory.make_account(
            openid_identifier='1234567', oauth_secret='1234567890')
        account.token_set.create(
            key='12345678901234567890123456789012345678901234567890',
            secret='0987654321', name='foo')
        assert account.secret == '1234567890', account.secret

        # use fix signature from http://nouncer.com/oauth/authentication.html
        # timestamp 1442966400
        # nonce 12345
        header = (
            'OAuth realm="http://sso-trusty:8000/error",'
            'oauth_consumer_key="1234567",'
            'oauth_token="12345678901234567890123456789012345678901234567890",'
            'oauth_nonce="12345",'
            'oauth_timestamp="1442966400",'
            'oauth_signature_method="HMAC-SHA1",'
            'oauth_version="1.0",'
            'oauth_signature="cOCcOe6F5qLDl43G1uvn1tx9Eig%3D"'
        )
        data = {
            'http_url': url, 'http_method': 'GET',
            'authorization': header,
        }
        # ensure the hard coded timestamp is valid
        with self.settings(OAUTH_REQUEST_MAX_AGE_SECONDS=9999999999):
            response = self.do_post(data)

        self.assert_request_response(
            response, is_valid=True, account_verified=True,
            openid_identifier=account.openid_identifier)

    def test_http_url_query_string_urlencoded_if_no_authorization(self):
        # build a predefined signature with a / in it
        account = self.factory.make_account(
            openid_identifier='1234567', oauth_secret='1234567890')
        account.token_set.create(
            key='12345678901234567890123456789012345678901234567890',
            secret='0987654321', name='foo')
        assert account.secret == '1234567890', account.secret

        # use fix signature from http://nouncer.com/oauth/authentication.html
        # timestamp 1444338517
        # nonce 12345
        url = 'http://example.com/?currency=USD&'
        # do not use urlencode because we need to mimic buggy clients
        oauth_qs = '&'.join((
            'oauth_consumer_key=1234567',
            'oauth_nonce=12345',
            'oauth_timestamp=1444338517',
            'oauth_token=12345678901234567890123456789012345678901234567890',
            'oauth_signature_method=HMAC-SHA1',
            'oauth_version=1.0',
            'oauth_signature=SyWDtG5Z9/s49g0HQmQGHE5q0Xs='))

        # no query_string nor authorization given, all OAuth is in the http_url
        buggy_url = url + oauth_qs
        data = {'http_url': buggy_url, 'http_method': 'GET'}

        # ensure the hard coded timestamp is valid
        with self.settings(OAUTH_REQUEST_MAX_AGE_SECONDS=9999999999):
            response = self.do_post(data)

        self.assert_request_response(
            response, is_valid=True, account_verified=True,
            openid_identifier=account.openid_identifier)
        self.assert_workaround_warning_logged(buggy_url, '', None)

    def test_statsd_metric_recorded_oauth_valid(self):
        data = self.build_data()
        self.do_post_and_assert_statsd_metric_recorded(data, 'oauth-valid')

    def test_statsd_metric_recorded_oauth_invalid(self):
        data = {'http_url': 'http://example.com/?oauth_signature=invalid',
                'http_method': self.default_http_method}
        self.do_post_and_assert_statsd_metric_recorded(data, 'oauth-invalid')

    def test_statsd_metric_recorded_missing_http_url(self):
        data = self.build_data()
        data.pop('http_url')
        self.do_post_and_assert_statsd_metric_recorded(data, 'missing-url')

    def test_statsd_metric_recorded_missing_http_method(self):
        data = self.build_data()
        data.pop('http_method')
        self.do_post_and_assert_statsd_metric_recorded(data, 'missing-method')

    def test_plaintext_signatures_not_logged_when_merging_query_string(self):
        data = self.build_data(as_query_string=True, plaintext=True)

        response = self.do_post(data)

        self.assert_request_response(
            response, is_valid=True, account_verified=True)
        self.assert_workaround_warning_logged('<REDACTED>', '', None)

    def test_plaintext_signatures_not_logged_when_token_not_allowed(self):
        v1_token = self.account.create_v1_oauth_token('v1 token')
        data = self.build_data(token=v1_token, plaintext=True)

        response = self.do_post(data)

        self.assert_request_response(response, is_valid=False)
        self.assert_token_filtered_warning_logged(
            '<REDACTED>', data['http_url'], self.default_http_method)


class RequestsWithHEADTestCase(RequestsTestCase):
    default_http_method = 'HEAD'


class RequestsWithDELETETestCase(RequestsTestCase):
    default_http_method = 'DELETE'


class RequestsWithPATCHTestCase(RequestsTestCase):
    default_http_method = 'PATCH'


class RequestsWithPOSTTestCase(RequestsTestCase):
    default_http_method = 'POST'


class RequestsWithPUTTestCase(RequestsTestCase):
    default_http_method = 'PUT'


class PasswordResetTokenHandlerTestCase(BaseTestCase, POSTOnlyResourceMixin):
    url = reverse('api-password-reset')

    def setUp(self):
        super(PasswordResetTokenHandlerTestCase, self).setUp()
        self.email = EmailAddress.objects.get(account=self.account)
        self.data = {'email': self.email.email}
        self.mock_logger = self.patch('api.v20.handlers.logging')

    def test_any_server_error_is_json(self):
        self.assert_any_server_error_is_json(self.do_post, data=self.data)

    def test_email_missing(self):
        self.assert_field_required(data={}, field='email')

    def test_invalid_email(self):
        mock_send_invitation = self.patch(
            'api.v20.handlers.emailutils.send_invitation_after_password_reset')

        extra = {'email': ['No account associated with invalid@foo.com']}
        self.assert_bad_request(data={'email': 'invalid@foo.com'}, extra=extra)
        mock_send_invitation.assert_called_once_with(
            self.sso_root_url, 'invalid@foo.com')

    def test_suspended_account(self):
        self.account.suspend()

        assert not self.account.can_reset_password
        content = self.do_post(data=self.data, status_code=403)

        self.assertEqual(content, {
            'message': 'Your account has been suspended. Please contact '
                       'login support to re-enable it',
            'code': 'ACCOUNT_SUSPENDED',
            'extra': {}})

        condition = "account '%s' is not active" % self.account.displayname
        self.mock_logger.debug("PasswordResetTokenHandler.create: email was "
                               "not sent out because %s" % condition)

    def test_deactivated_account(self):
        self.account.status = AccountStatus.DEACTIVATED
        self.account.save()

        name = 'identityprovider.models.account.Account.can_reset_password'
        with patch(name, False):
            content = self.do_post(data=self.data, status_code=403)

        self.assertEqual(content, {
            'message': 'Your account has been deactivated. To reactivate it, '
                       'please reset your password',
            'code': 'ACCOUNT_DEACTIVATED',
            'extra': {}})

        condition = "account '%s' is not active" % self.account.displayname
        self.mock_logger.debug("PasswordResetTokenHandler.create: email was "
                               "not sent out because %s" % condition)

    def test_can_not_reset_password(self):
        name = 'identityprovider.models.account.Account.can_reset_password'
        with patch(name, False):
            content = self.do_post(data=self.data, status_code=403)

        self.assertEqual(content, {
            'message': 'Can not reset password. Please contact login support',
            'code': 'CAN_NOT_RESET_PASSWORD',
            'extra': {}})

        condition = "account '%s' is not active" % self.account.displayname
        self.mock_logger.debug("PasswordResetTokenHandler.create: email was "
                               "not sent out because %s" % condition)

    def test_reset_password(self):
        email = self.email.email
        body = self.do_post({'email': email}, status_code=201)

        self.assertEqual(body['email'], email)

        self.assertEqual(len(mail.outbox), 1)
        mail_content = unicode(mail.outbox[0].message())
        links = re.search(
            ('http://.+?/token/(.+?)/\+resetpassword'), mail_content)
        self.assertIsNotNone(links, "Found no +resetpassword links with token")
        token = links.group(1)
        self.assertIn(token, mail_content)
        self.assertIn(email, mail_content)

    def test_reset_password_uses_preferred_if_given_not_validated(self):
        email = self.factory.make_email_for_account(
            self.account, status=EmailStatus.NEW)
        self.do_post({'email': email.email}, status_code=201)

        tokens = AuthToken.objects.filter(
            token_type=AuthTokenType.PASSWORDRECOVERY,
            email=email)
        self.assertEqual(tokens.count(), 0)

        tokens = AuthToken.objects.filter(
            token_type=AuthTokenType.PASSWORDRECOVERY,
            email=self.account.preferredemail)
        self.assertEqual(tokens.count(), 1)

        self.assertEqual(len(mail.outbox), 1)
        mail_content = unicode(mail.outbox[0].message())
        self.assertNotIn(email.email, mail_content)
        self.assertIn(self.account.preferredemail.email, mail_content)

    def test_reset_password_uses_given_email_if_validated(self):
        email = self.factory.make_email_for_account(
            self.account, status=EmailStatus.VALIDATED)
        assert self.account.preferredemail.email != email.email
        self.do_post({'email': email.email}, status_code=201)

        tokens = AuthToken.objects.filter(
            token_type=AuthTokenType.PASSWORDRECOVERY,
            email=self.account.preferredemail)
        self.assertEqual(tokens.count(), 0)

        tokens = AuthToken.objects.filter(
            token_type=AuthTokenType.PASSWORDRECOVERY,
            email=email)
        self.assertEqual(tokens.count(), 1)

        self.assertEqual(len(mail.outbox), 1)
        mail_content = unicode(mail.outbox[0].message())
        self.assertNotIn(self.account.preferredemail.email, mail_content)
        self.assertIn(email.email, mail_content)

    def test_too_many_tokens(self):
        with self.settings(MAX_PASSWORD_RESET_TOKENS=0):
            content = self.do_post(data=self.data, status_code=403)

        self.assertEqual(content, {
            'message': 'Too many non-consumed tokens exist. Further token '
                       'creation is not allowed until existing tokens '
                       'are consumed.',
            'code': 'TOO_MANY_TOKENS',
            'extra': {}})

    def test_too_many_active_tokens(self):
        self.do_post(data=self.data, status_code=201)
        tokens = AuthToken.objects.filter(
            token_type=AuthTokenType.PASSWORDRECOVERY, email=self.email,
            date_consumed=None)
        assert tokens.count() == 1

        with self.settings(MAX_PASSWORD_RESET_TOKENS=1):
            content = self.do_post(data=self.data, status_code=403)
            self.assertEqual(content, {
                'message': 'Too many non-consumed tokens exist. Further '
                           'token creation is not allowed until existing '
                           'tokens are consumed.',
                'code': 'TOO_MANY_TOKENS',
                'extra': {}})

    def test_invalidated_email(self):
        invalidated_email = self.email.invalidate()

        name = ('api.v20.handlers.emailutils.'
                'send_invitation_after_password_reset')
        with patch(name) as mock_send_invitation:
            content = self.do_post(
                {'email': invalidated_email.email}, status_code=403)

        mock_send_invitation.assert_called_once_with(self.sso_root_url,
                                                     self.email.email)
        expected = {
            'message': ('This email address has been invalidated. '
                        'Please contact login support.'),
            'code': 'EMAIL_INVALIDATED',
            'extra': {},
        }
        self.assertEqual(content, expected)

    @patch('api.v20.handlers.emailutils.send_password_reset_email')
    def test_redirect_from_token(self, mock_send_password_reset_email):
        token = 'ABCDEFGH' * 2
        redirection_url = redirection_url_for_token(token)
        # We expect this method to return a tuple
        mock_send_password_reset_email.return_value = (Mock(), Mock())
        self.do_post(
            {'email': self.email.email, 'token': token}, status_code=201)

        mock_send_password_reset_email.assert_called_once_with(
            self.sso_root_url, self.account, self.email.email, redirection_url)

    def test_invalid_email_throttle_decorator(self):
        self.assert_bad_request(
            data={'email': 'with space@example.com'},
            extra={'email': ['Enter a valid email address.']})


class TokensHandlerTestCase(BaseTestCase):
    url_name = 'api-token'

    def setUp(self):
        super(TokensHandlerTestCase, self).setUp()
        self.url = reverse(
            self.url_name, kwargs=dict(token=self.token.pk))

    def assert_token_result(self, result, token):
        expected = dict(
            consumer_key=token.consumer.key,
            token_key=token.key,
            token_name=token.name,
            date_created=token.date_created,
            date_updated=token.date_updated,
            href=reverse(self.url_name, kwargs=dict(token=token.pk)),
        )
        self.assertEqual(result, self.to_json(expected))

    def test_any_server_error_is_json(self):
        self.assert_any_server_error_is_json(self.do_get, token=self.token)

    def test_token_empty(self):
        self.assertRaises(
            NoReverseMatch, reverse, self.url_name, kwargs=dict(token=''))

    def test_unauthenticated_401(self):
        self.do_get(status_code=401)

    def test_authenticated_with_django_401(self):
        assert self.client.login(username=self.account.preferredemail.email,
                                 password=DEFAULT_USER_PASSWORD)
        self.do_get(status_code=401)

    def test_get_authenticated_with_oauth(self):
        json_body = self.do_get(token=self.token)
        self.assert_token_result(json_body, self.token)

    def test_secrets_not_returned(self):
        json_body = self.do_get(token=self.token)
        for k in ('token_secret', 'consumer_secret'):
            self.assertNotIn(k, json_body)

    def test_can_get_others_if_elevated_permission(self):
        other_token = self.factory.make_account().create_oauth_token('other')
        self.factory.add_permission_to_account(
            self.account, 'api_view_account_details')

        url = reverse(self.url_name, kwargs=dict(token=other_token.pk))
        json_body = self.do_get(url=url, token=self.token)
        self.assert_token_result(json_body, other_token)

    def test_forbidden_if_not_owner_not_elevated_permission_on_get(self):
        other_token = self.factory.make_account().create_oauth_token('other')

        url = reverse(self.url_name, kwargs=dict(token=other_token.pk))
        self.do_get(url=url, token=self.token, status_code=401)

    def test_get_token_does_not_exist(self):
        self.factory.add_permission_to_account(
            self.account, 'api_view_account_details')

        url = reverse(self.url_name, kwargs=dict(token='foobarbaz'))
        self.do_get(url=url, token=self.token, status_code=404)

    def test_delete_authenticated_with_oauth(self):
        response = self.get_response(
            token=self.token, method='DELETE', status_code=204)

        self.assertEqual(response.content, '')
        self.assertEqual(Token.objects.filter(pk=self.token.pk).count(), 0)

    def test_can_delete_others_if_elevated_permission(self):
        other_token = self.factory.make_account().create_oauth_token('other')
        self.factory.add_permission_to_account(
            self.account, 'api_edit_account_details')

        url = reverse(self.url_name, kwargs=dict(token=other_token.pk))
        response = self.get_response(
            url=url, token=self.token, method='DELETE', status_code=204)

        self.assertEqual(response.content, '')
        self.assertEqual(Token.objects.filter(pk=other_token.pk).count(), 0)

    def test_forbidden_if_not_owner_not_elevated_permission_on_delete(self):
        other_token = self.factory.make_account().create_oauth_token('other')

        url = reverse(self.url_name, kwargs=dict(token=other_token.pk))
        self.get_response(
            url=url, token=self.token, method='DELETE', status_code=401)
        self.assertEqual(Token.objects.filter(pk=other_token.pk).count(), 1)

    def test_delete_token_does_not_exist(self):
        self.factory.add_permission_to_account(
            self.account, 'api_edit_account_details')

        url = reverse(self.url_name, kwargs=dict(token='foobarbaz'))
        self.get_response(
            url=url, token=self.token, method='DELETE', status_code=404)


class SSHKeyHandlerTestCase(BaseTestCase):
    url_name = 'api-keys'
    key_tmpl = '{} {} {}'.format('ssh-rsa ', uuid.uuid4(), uuid.uuid4())

    def _create_test_scenario(self, num_keys=1, make_person=True):
        keys = [self.key_tmpl for key in range(num_keys)]
        mock = self.patch('api.v20.handlers.get_lp_ssh_keys')
        mock.return_value = keys
        account = self.factory.make_account()
        if make_person:
            self.factory.make_person(account=account)
        return account, keys

    def assert_response_contains_account_details(self, response, account):
        self.assertEqual(
            account.openid_identifier, response['openid_identifier'])
        self.assertEqual(account.person_name, response['username'])

    def test_post_not_allowed(self):
        account = self.factory.make_account()
        self.factory.make_person(account=account)
        url = reverse(self.url_name, kwargs=dict(account=account.user.email))
        response = self.client.post(url)
        self.assertEqual(response.status_code, 405)
        self.assertEqual(response.get('Allow', ''), 'GET')

    def test_ssh_key_no_keys_by_email(self):
        account, keys = self._create_test_scenario(num_keys=0)

        url = reverse(self.url_name, kwargs=dict(account=account.user.email))
        response = self.do_get(url=url, status_code=200)

        self.assertEqual(keys, response['ssh_keys'])
        self.assert_response_contains_account_details(response, account)

    def test_ssh_key_no_keys_by_openid(self):
        account, keys = self._create_test_scenario(num_keys=0)

        url = reverse(
            self.url_name, kwargs=dict(account=account.openid_identifier))
        response = self.do_get(url=url, status_code=200)

        self.assertEqual(keys, response['ssh_keys'])
        self.assert_response_contains_account_details(response, account)

    def test_ssh_key_one_key_by_email(self):
        account, keys = self._create_test_scenario(num_keys=1)

        url = reverse(self.url_name, kwargs=dict(account=account.user.email))
        response = self.do_get(url=url, status_code=200)

        self.assertEqual(keys, response['ssh_keys'])
        self.assert_response_contains_account_details(response, account)

    def test_ssh_key_one_key_by_openid(self):
        account, keys = self._create_test_scenario(num_keys=1)

        url = reverse(
            self.url_name, kwargs=dict(account=account.openid_identifier))
        response = self.do_get(url=url, status_code=200)

        self.assertEqual(keys, response['ssh_keys'])
        self.assert_response_contains_account_details(response, account)

    def test_ssh_key_multiple_keys_by_email(self):
        account, keys = self._create_test_scenario(num_keys=5)

        url = reverse(self.url_name, kwargs=dict(account=account.user.email))
        response = self.do_get(url=url, status_code=200)

        self.assertEqual(keys, response['ssh_keys'])
        self.assert_response_contains_account_details(response, account)

    def test_ssh_key_multiple_keys_by_openid(self):
        account, keys = self._create_test_scenario(num_keys=5)

        url = reverse(
            self.url_name, kwargs=dict(account=account.openid_identifier))
        response = self.do_get(url=url, status_code=200)

        self.assertEqual(keys, response['ssh_keys'])
        self.assert_response_contains_account_details(response, account)

    def test_ssh_key_valid_email_no_username(self):
        account, keys = self._create_test_scenario(make_person=False)

        url = reverse(self.url_name, kwargs=dict(account=account.user.email))
        response = self.do_get(url=url, status_code=404)

        self.assertEqual('no_keys', response['error_type'])

    def test_ssh_key_email_without_account(self):
        url = reverse(self.url_name, kwargs=dict(account='bogus@example.com'))
        response = self.do_get(url=url, status_code=404)

        self.assertEqual('no_keys', response['error_type'])

    def test_ssh_key_invalid_email(self):
        url = reverse(self.url_name, kwargs=dict(account='bo gus@example.com'))
        response = self.do_get(url=url, status_code=400)

        self.assertEqual('Invalid request data', response['message'])
        self.assertEqual('INVALID_DATA', response['code'])

    def test_ssh_key_valid_openid_no_username(self):
        account = self.factory.make_account()
        url = reverse(
            self.url_name, kwargs=dict(account=account.openid_identifier))
        response = self.do_get(url=url, status_code=404)

        self.assertEqual('no_keys', response['error_type'])

    def test_ssh_key_invalid_openid(self):
        url = reverse(self.url_name, kwargs=dict(account='0000000'))
        response = self.do_get(url=url, status_code=404)

        self.assertEqual('no_keys', response['error_type'])


# FIXME: These tests are time sensitive (both the test and the code do call
# time.time()) and can (and do) break under load (with AssertionError: '41' !=
# u'42'). They could either control time.time() or inject the right cache
# entries and use them (instead of mocking them) -- vila 2015-11-13
class ThrottledTestCase(BaseTestCase):

    def patch_piston_cache(self):
        return self.patch('piston.utils.cache')

    def throttle_all(self):
        mock_cache = self.patch_piston_cache()
        # hit max number of requests allowed
        mock_cache.get.return_value = (3600, time.time() + 42.1)

    def throttle_by_openid(self, openid):
        mock_cache = self.patch_piston_cache()

        def mock_get(key, default=None, version=None):
            # only look for openid in extra part
            # the first part of the key is the requester username
            _, extra = key.split(':', 1)
            if openid in extra:
                return (3600, time.time() + 42.1)
            else:
                return (42, time.time() + 42.1)
        mock_cache.get = mock_get

    def throttle_by_email(self, email):
        mock_cache = self.patch_piston_cache()

        def mock_get(key, default=None, version=None):
            if email in key:
                return (3600, time.time() + 42.1)
            else:
                return (42, time.time() + 42.1)
        mock_cache.get = mock_get

    def throttle_by_token(self, token):
        mock_cache = self.patch_piston_cache()

        def mock_get(key, default=None, version=None):
            if token in key:
                return (3600, time.time() + 42.1)
            else:
                return (42, time.time() + 42.1)
        mock_cache.get = mock_get

    def assert_handler_throttled(
            self, request, token=None, throttle_func=None, throttled=True,
            **kwargs):
        if throttle_func is None:
            self.throttle_all()
        else:
            throttle_func()

        response = self.get_response(
            request.path, token=token, method=request.method, status_code=None,
            **kwargs)

        if throttled:
            self.assertEqual(response.status_code, 429)
            self.assertEqual(response['Retry-After'], '42')
        else:
            self.assertNotEqual(response.status_code, 429)

    def assert_handler_not_throttled(self, *args, **kwargs):
        self.assert_handler_throttled(*args, throttled=False, **kwargs)

    def test_anonymous_accounts_handler_throttled_by_openid(self):
        openid = self.account.openid_identifier
        url = reverse('api-account', args=[openid])
        request = RequestFactory().get(url)
        request.user = AnonymousUser()

        self.assert_handler_throttled(
            request, throttle_func=lambda: self.throttle_by_openid(openid))

        assert 'foo1234' != openid
        path = reverse('api-account', args=['foo1234'])
        request.path = path
        self.assert_handler_not_throttled(
            request, throttle_func=lambda: self.throttle_by_openid(openid))

    def test_accounts_handler_throttled_by_openid(self):
        openid = self.account.openid_identifier
        url = reverse('api-account', args=[openid])
        request = RequestFactory().get(url)
        request.user = self.account

        self.assert_handler_throttled(
            request, token=self.token,
            throttle_func=lambda: self.throttle_by_openid(openid))

        assert 'foo1234' != openid
        path = reverse('api-account', args=['foo1234'])
        request.path = path
        self.assert_handler_not_throttled(
            request, token=self.token,
            throttle_func=lambda: self.throttle_by_openid(openid))

    def test_accounts_handler_whitelisted(self):
        url = reverse('api-account', args=[self.account.openid_identifier])
        request = RequestFactory().get(url)
        request.user = self.account

        emails = [self.account.preferredemail.email]
        with self.settings(THROTTLE_WHITELIST_EMAIL_REGEXP_LIST=emails):
            self.assert_handler_not_throttled(request, token=self.token)

    def test_password_reset_token_handler_throttled_by_email(self):
        url = reverse('api-password-reset')
        request = RequestFactory().post(url)
        request.user = AnonymousUser()
        email = self.account.preferredemail.email

        self.assert_handler_throttled(
            request, data=json.dumps({'email': email}),
            content_type='application/json')

        self.assert_handler_throttled(
            request, data=json.dumps({'email': 'foo@foo.com'}),
            content_type='application/json')

    def test_account_registration_handler_throttled_by_ip(self):
        url = reverse('api-registration')
        request = RequestFactory().post(url)
        request.user = AnonymousUser()
        email = 'foo@foo.com'

        data = {
            'captcha_id': 'some-id',
            'captcha_solution': 'some solution',
            'email': 'foo@foo.com',
            'password': 'some-password-123',
            'displayname': 'Some User',
        }

        self.assert_handler_throttled(
            request, data=json.dumps(data),
            content_type='application/json')

        data['email'] = 'bar@foo.com'
        self.assert_handler_not_throttled(
            request, data=json.dumps(data),
            content_type='application/json',
            throttle_func=lambda: self.throttle_by_email(email))

    def test_account_login_handler_throttled_by_email(self):
        url = reverse('api-login')
        request = RequestFactory().post(url)
        request.user = AnonymousUser()
        email = self.account.preferredemail.email

        data = {
            'email': email,
            'password': DEFAULT_USER_PASSWORD,
            'token_name': 'foo',
        }

        self.assert_handler_throttled(
            request, data=json.dumps(data),
            content_type='application/json',
            throttle_func=lambda: self.throttle_by_email(email))

        data['email'] = 'bar@foo.com'
        self.assert_handler_not_throttled(
            request, data=json.dumps(data),
            content_type='application/json',
            throttle_func=lambda: self.throttle_by_email(email))

    def test_emails_handler_get_throttled_by_email(self):
        email = self.account.preferredemail.email
        url = reverse('api-email', args=[email])
        request = RequestFactory().get(url)
        request.user = self.account

        self.assert_handler_throttled(
            request, token=self.token,
            throttle_func=lambda: self.throttle_by_email(email))

        assert 'foo@foo.com' != email
        url = reverse('api-email', args=['foo@foo.com'])
        request.path = url
        self.assert_handler_not_throttled(
            request, token=self.token,
            throttle_func=lambda: self.throttle_by_email(email))

    def test_emails_handler_get_whitelisted(self):
        email = self.account.preferredemail.email
        url = reverse('api-email', args=[email])
        request = RequestFactory().get(url)
        request.user = self.account

        emails = [self.account.preferredemail.email]
        with self.settings(THROTTLE_WHITELIST_EMAIL_REGEXP_LIST=emails):
            self.assert_handler_not_throttled(request, token=self.token)

    def test_email_handler_delete_throttled_by_email(self):
        email = self.account.preferredemail.email
        url = reverse('api-email', args=[email])
        request = RequestFactory().delete(url)
        request.user = self.account

        self.assert_handler_throttled(
            request, token=self.token,
            throttle_func=lambda: self.throttle_by_email(email))

        assert 'foo@foo.com' != email
        url = reverse('api-email', args=['foo@foo.com'])
        request.path = url
        self.assert_handler_not_throttled(
            request, token=self.token,
            throttle_func=lambda: self.throttle_by_email(email))

    def test_email_handler_delete_whitelisted(self):
        email = self.account.preferredemail.email
        url = reverse('api-email', args=[email])
        request = RequestFactory().delete(url)
        request.user = self.account

        emails = [self.account.preferredemail.email]
        with self.settings(THROTTLE_WHITELIST_EMAIL_REGEXP_LIST=emails):
            self.assert_handler_not_throttled(request, token=self.token)

    def test_tokens_handler_get_throttled(self):
        token = self.token.pk
        url = reverse('api-token', args=[self.token])
        request = RequestFactory().get(url)
        request.user = self.account

        self.assert_handler_throttled(
            request, token=self.token,
            throttle_func=lambda: self.throttle_by_token(token))

        assert 'some-token' != token
        path = reverse('api-token', args=['some-token'])
        request.path = path
        self.assert_handler_not_throttled(
            request, token=self.token,
            throttle_func=lambda: self.throttle_by_token(token))

    def test_tokens_handler_get_whitelisted(self):
        url = reverse('api-token', args=[self.token])
        request = RequestFactory().get(url)
        request.user = self.account

        emails = [self.account.preferredemail.email]
        with self.settings(THROTTLE_WHITELIST_EMAIL_REGEXP_LIST=emails):
            self.assert_handler_not_throttled(request, token=self.token)

    def test_tokens_handler_delete_throttled(self):
        token = self.token.pk
        url = reverse('api-token', args=[self.token])
        request = RequestFactory().delete(url)
        request.user = self.account

        self.assert_handler_throttled(
            request, token=self.token,
            throttle_func=lambda: self.throttle_by_token(token))

        assert 'some-token' != token
        path = reverse('api-token', args=['some-token'])
        request.path = path
        self.assert_handler_not_throttled(
            request, token=self.token,
            throttle_func=lambda: self.throttle_by_token(token))

    def test_tokens_handler_delete_whitelisted(self):
        url = reverse('api-token', args=[self.token])
        request = RequestFactory().delete(url)
        request.user = self.account

        emails = [self.account.preferredemail.email]
        with self.settings(THROTTLE_WHITELIST_EMAIL_REGEXP_LIST=emails):
            self.assert_handler_not_throttled(request, token=self.token)

    def test_ssh_key_handler_throttled_by_email(self):
        account = self.factory.make_account()
        email = account.user.email
        url = reverse('api-keys', args=[email])
        request = RequestFactory().get(url)
        request.user = self.account

        self.assert_handler_throttled(
            request, token=self.token,
            throttle_func=lambda: self.throttle_by_email(email))

    def test_ssh_key_handler_throttled_by_token(self):
        account = self.factory.make_account()
        token = account.openid_identifier
        url = reverse('api-keys', args=[token])
        request = RequestFactory().get(url)
        request.user = self.account

        self.assert_handler_throttled(
            request, token=self.token,
            throttle_func=lambda: self.throttle_by_token(token))

    def test_reports_metrics_when_throttled(self):
        email = self.account.preferredemail.email
        url = reverse('api-email', args=[email])
        request = RequestFactory().get(url)
        request.user = self.account
        called_metrics = []

        with patch(
                'api.v20.decorators.stats.increment', called_metrics.append):
            self.assert_handler_throttled(request, token=self.token)
        self.assertEqual(
            ['throttling.throttled',
             'throttling.api.v20.handlers.EmailsHandler.read.throttled'],
            called_metrics)

    def test_doesnt_reports_metrics_when_not_throttled(self):
        url = reverse('api-account', args=[self.account.openid_identifier])
        request = RequestFactory().get(url)
        request.user = self.account
        called_metrics = []

        with patch(
                'api.v20.decorators.stats.increment', called_metrics.append):
            with patch('piston.utils.cache') as mock_cache:
                mock_cache.get.return_value = (0, time.time() + 42.1)
                self.get_response(request.path, method=request.method)
        self.assertEqual([], called_metrics)


class MacaroonHandlerBaseTestCase(SSOBaseTestCase):

    def track_failed_logins(self, **kw):
        self.login_failed_calls.append(kw)

    def setUp(self):
        super(MacaroonHandlerBaseTestCase, self).setUp()

        self.login_failed_calls = []
        login_failed.connect(self.track_failed_logins, dispatch_uid=self.id())

        # XXX cjwatson 2016-05-11:
        # MacaroonRefreshHandlerTestCase.test_macaroon_refreshed needs to be
        # fixed before changing the version here.
        self.root_macaroon, self.macaroon_random_key, _ = self.build_macaroon(
            version=0)
        self.data = dict(
            email='foo@bar.com', password='foobar123',
            macaroon=self.root_macaroon.serialize())
        self.account = self.factory.make_account(
            email=self.data['email'], password=self.data['password'])

        self.stats_increment = self.patch('api.v20.handlers.stats.increment')


class MacaroonDischargeHandlerTestCase(MacaroonHandlerBaseTestCase,
                                       AuthLogTestCaseMixin):

    url = reverse('api-macaroon-discharge')

    def do_post(self, data=None, expected_status_code=200,
                content_type='application/json', **kwargs):
        if data is None:
            data = self.data

        response = self.client.post(self.url, data=json.dumps(data),
                                    content_type=content_type, **kwargs)

        self.assertEqual(
            response.status_code, expected_status_code,
            "Bad status code! expected={} got={} response={}".format(
                expected_status_code, response.status_code, response))
        self.assertEqual(response['Content-type'],
                         'application/json; charset=utf-8',
                         response)
        return json.loads(response.content)

    def assert_response_correct(self, json_body, account, bind=False):
        """Check we received a good discharge macaroon for the root sent.

        Proper verification of the received macaroon internals happens in
        the tests of the macaroon builder.
        """
        discharge_macaroon_raw = json_body['discharge_macaroon']
        discharge_macaroon = Macaroon.deserialize(discharge_macaroon_raw)
        if bind:
            discharge_macaroon = self.root_macaroon.prepare_for_request(
                discharge_macaroon)
        v = Verifier()
        v.satisfy_general(lambda c: True)
        v.verify(
            self.root_macaroon, self.macaroon_random_key, [discharge_macaroon])

    def assert_failed_login(self, code, data, extra=None,
                            expected_status_code=403, check_login_failed=True):
        if extra is None:
            extra = {}

        json_body = self.do_post(expected_status_code=expected_status_code,
                                 data=data)
        self.assertEqual(json_body['code'], code)
        self.assertEqual(json_body['extra'], extra)

        if check_login_failed:
            self.assertEqual(len(self.login_failed_calls), 1)
            failed = self.login_failed_calls[0]
            self.assertEqual(
                failed['credentials'],
                {'email': data['email'], 'password': data['password']})

        return json_body

    def test_login_required_parameters(self):
        json_body = self.do_post(expected_status_code=400, data={})

        self.assertEqual(json_body, {
            'code': 'INVALID_DATA',
            'extra': {'email': [handlers.FIELD_REQUIRED]},
            'message': 'Invalid request data'},
        )
        self.assertEqual(self.login_failed_calls, [])

    def test_account_suspended(self):
        self.account.suspend()
        self.assert_failed_login('ACCOUNT_SUSPENDED', self.data)

    def test_account_deactivated(self):
        self.account.deactivate()
        self.assert_failed_login('ACCOUNT_DEACTIVATED', self.data)

    def test_failed_login(self):
        self.account.set_password('other-thing')
        self.account.save()
        self.assert_failed_login('INVALID_CREDENTIALS',
                                 self.data, expected_status_code=401)

    def test_email_invalidated(self):
        self.account.preferredemail.invalidate()
        self.assert_failed_login('EMAIL_INVALIDATED', self.data)

    def test_password_policy(self):
        self.account.accountpassword.password = make_password('short')
        self.account.accountpassword.save()

        self.data['password'] = 'short'
        extra = {
            'reason': 'Password must be at least 8 characters long',
            'location': 'http://testserver/'}
        self.assert_failed_login(
            'PASSWORD_POLICY_ERROR', data=self.data, extra=extra)

    def test_twofactor_required(self):
        self.account.twofactor_required = True
        self.account.save()

        json_body = self.do_post(expected_status_code=401)
        self.assertEqual(json_body['code'], 'TWOFACTOR_REQUIRED')
        self.assertEqual(self.login_failed_calls, [])

    def test_twofactor(self):
        device = self.factory.make_device(account=self.account)
        self.account.twofactor_required = True
        self.account.save()

        self.data['otp'] = self.factory.make_next_otp(device)
        json_body = self.do_post(data=self.data)

        self.assert_response_correct(json_body, self.account)

    def test_twofactor_wrong_otp(self):
        device = self.factory.make_device(account=self.account)
        self.account.twofactor_required = True
        self.account.save()

        self.data['otp'] = str(int(self.factory.make_next_otp(device)) - 1)
        self.assert_failed_login('TWOFACTOR_FAILURE', self.data)

    def test_failed_login_creates_authlog(self):
        self.account.set_password('other-thing')
        self.account.save()
        expected_values = {
            'login_email': self.data['email'],
            'log_type': AuthLogType.API_FAIL,
            'referer': '',
            'user_agent': '',
            'remote_ip': '127.0.0.1',
            'account': None
        }
        with self.assert_authlog_matches(expected_values):
            self.assert_failed_login('INVALID_CREDENTIALS',
                                     self.data, expected_status_code=401)

    def test_login_creates_authlog(self):
        expected_values = {
            'login_email': self.data['email'],
            'log_type': AuthLogType.API_MACAROON_DISCHARGE_NEW,
            'referer': '',
            'user_agent': '',
            'remote_ip': '127.0.0.1',
            'account': self.account.id
        }
        # This context manager magically checks when_created values
        with self.assert_authlog_matches(expected_values):
            self.do_post()

    def test_macaroon_created(self):
        json_body = self.do_post()
        self.assert_response_correct(json_body, self.account)
        self.stats_increment.assert_any_call(
            'macaroon.discharge', key='single_macaroon')

    def test_root_macaroon_corrupt(self):
        data = dict(
            email='foo@bar.com', password='foobar123',
            macaroon="I'm a seriously corrupted macaroon")
        self.assert_failed_login('INVALID_DATA', data,
                                 expected_status_code=400,
                                 check_login_failed=False)

    def test_root_macaroon_not_for_sso(self):
        macaroon, _, _ = self.build_macaroon(
            service_location="other service", version=1)
        data = dict(email='foo@bar.com', password='foobar123',
                    macaroon=macaroon.serialize())
        self.assert_failed_login('INVALID_CREDENTIALS', data,
                                 expected_status_code=401,
                                 check_login_failed=False)

    def test_missing_macaroon_info(self):
        data = dict(email='foo@bar.com', password='foobar123')
        extra = {'caveat_id': [handlers.FIELD_REQUIRED]}
        self.assert_failed_login('INVALID_DATA', data, extra=extra,
                                 expected_status_code=400,
                                 check_login_failed=False)

    def test_multiple_macaroons(self):
        # build *several* macaroons and send them all at once
        roots, rkeys, _ = zip(*[
            self.build_macaroon(version=1) for _ in range(3)])
        ids = range(3)
        roots = dict(zip(ids, roots))
        rkeys = dict(zip(ids, rkeys))
        macaroons = [(mac_id, m.serialize()) for mac_id, m in roots.items()]
        data = dict(email='foo@bar.com', password='foobar123',
                    macaroons=macaroons)
        json_body = self.do_post(data=data)

        # get all the discharges and verify their integrity
        verifying_function = lambda c: True
        for mac_id, discharge_macaroon_raw in json_body['discharge_macaroons']:
            discharge_macaroon = Macaroon.deserialize(discharge_macaroon_raw)
            root_macaroon = roots[mac_id]
            key = rkeys[mac_id]
            v = Verifier()
            v.satisfy_general(verifying_function)
            v.verify(root_macaroon, key, [discharge_macaroon])

        self.stats_increment.assert_any_call(
            'macaroon.discharge', key='multi_macaroons')

    def test_multiple_mixed(self):
        bad_macaroon, _, _ = self.build_macaroon(
            service_location="other service", version=1)
        data = dict(email='foo@bar.com', password='foobar123', macaroons=[
            ('good', self.root_macaroon.serialize()),
            ('bad', bad_macaroon.serialize()),
        ])
        self.assert_failed_login('INVALID_CREDENTIALS', data,
                                 expected_status_code=401,
                                 check_login_failed=False)

    def test_caveat_id_discharge_created(self):
        # discharge the test macaroon
        (caveat,) = [c for c in self.root_macaroon.third_party_caveats()
                     if c.location == settings.MACAROON_SERVICE_LOCATION]
        data = dict(email='foo@bar.com', password='foobar123',
                    caveat_id=caveat.caveat_id)

        json_body = self.do_post(data=data)
        self.assert_response_correct(json_body, self.account, bind=True)
        self.stats_increment.assert_any_call(
            'macaroon.discharge', key='caveat_id')

    def test_caveat_id_corrupt(self):
        data = dict(email='foo@bar.com', password='foobar123',
                    caveat_id="I'm a seriously corrupted caveat")
        self.assert_failed_login('INVALID_DATA', data,
                                 expected_status_code=400,
                                 check_login_failed=False)


class MacaroonHandlerTimelineTestCase(MacaroonHandlerBaseTestCase,
                                      TimelineActionMixin):

    url = reverse('api-macaroon-discharge')

    def test_login_timeline_records_password_checking(self):
        # Prepare and inject a timeline in the client's POST.
        empty_timeline = Timeline()
        response = self.client.post(
            self.url, data=json.dumps(self.data),
            content_type='application/json',
            **{'timeline.timeline': empty_timeline})
        self.assert_timeline_contains(request=response.wsgi_request,
                                      category='django-password-checking',
                                      detail=str(self.account.id))

    def test_login_ignore_weird_timeline(self):
        """A weird object passed as timeline should be ignored/not used."""
        weird_timeline = []
        response = self.client.post(
            self.url, data=json.dumps(self.data),
            content_type='application/json',
            **{'timeline.timeline': weird_timeline})

        # If the timeline is a weird object we shouldn't have touched it
        self.assertEqual([], response.wsgi_request.META['timeline.timeline'])

    def test_login_no_timeline(self):
        """Shouldn't barf if there's no timeline in the request at all."""
        response = self.client.post(
            self.url, data=json.dumps(self.data),
            content_type='application/json')

        self.assertEqual(response.status_code, 200)
        self.assertNotIn('timeline.timeline', response.wsgi_request.META)


class MacaroonRefreshHandlerTestCase(MacaroonHandlerBaseTestCase):

    url = reverse('api-macaroon-refresh')

    def setUp(self):
        super(MacaroonRefreshHandlerTestCase, self).setUp()

        # discharge the test macaroon
        (caveat,) = [c for c in self.root_macaroon.third_party_caveats()
                     if c.location == settings.MACAROON_SERVICE_LOCATION]

        unbound_discharge = build_discharge_macaroon(
            self.account, caveat.caveat_id)
        self.discharge_macaroon = self.root_macaroon.prepare_for_request(
            unbound_discharge)

        self.data = {'discharge_macaroon': unbound_discharge.serialize()}

    def do_post(self, data=None, expected_status_code=200,
                content_type='application/json', **kwargs):
        if data is None:
            data = self.data

        response = self.client.post(self.url, data=json.dumps(data),
                                    content_type=content_type, **kwargs)

        self.assertEqual(
            response.status_code, expected_status_code,
            "Bad status code! expected={} got={} response={}".format(
                expected_status_code, response.status_code, response))
        self.assertEqual(response['Content-type'],
                         'application/json; charset=utf-8',
                         response)
        return json.loads(response.content)

    def assert_failed_login(self, code, data, extra=None,
                            expected_status_code=403, check_login_failed=True):
        if extra is None:
            extra = {}

        json_body = self.do_post(expected_status_code=expected_status_code,
                                 data=data)
        self.assertEqual(json_body['code'], code)
        self.assertEqual(json_body['extra'], extra)
        return json_body

    def test_required_parameters_discharge(self):
        json_body = self.do_post(expected_status_code=400, data={})

        self.assertEqual(json_body, {
            'code': 'INVALID_DATA',
            'extra': {'discharge_macaroon': [handlers.FIELD_REQUIRED]},
            'message': 'Invalid request data'},
        )
        self.assertEqual(self.login_failed_calls, [])

    def test_account_missing(self):
        self.account.deactivate()
        self.assert_failed_login('ACCOUNT_DEACTIVATED', self.data)

    def test_root_macaroon_corrupt(self):
        data = dict(root_macaroon="I'm a seriously corrupted macaroon",
                    discharge_macaroon="Also broken")
        self.assert_failed_login('INVALID_DATA', data,
                                 expected_status_code=400,
                                 check_login_failed=False)

    def test_macaroon_bad_authinfo(self):
        macaroon, _, _ = self.build_macaroon(
            service_location="other service", version=1)
        data = dict(root_macaroon=macaroon.serialize(),
                    discharge_macaroon=self.discharge_macaroon.serialize())
        self.assert_failed_login('INVALID_CREDENTIALS', data,
                                 expected_status_code=401,
                                 check_login_failed=False)

    def test_macaroon_refreshed(self):
        data = {'root_macaroon': self.root_macaroon.serialize(),
                'discharge_macaroon': self.discharge_macaroon.serialize()}
        json_body = self.do_post(data=data)

        # get new discharge macaroon and verify with old root (its internals
        # are verified by the tests of the 'refresh_macaroons' function itself)
        discharge_macaroon_raw = json_body['discharge_macaroon']
        discharge_macaroon = Macaroon.deserialize(discharge_macaroon_raw)
        v = Verifier()
        v.satisfy_general(lambda c: True)
        v.verify(
            self.root_macaroon, self.macaroon_random_key, [discharge_macaroon])
        self.stats_increment.assert_any_call(
            'macaroon.refresh', key='root_macaroon')

    def test_just_discharge_refreshed(self):
        json_body = self.do_post()

        # get new discharge macaroon, apply it and verify with old root (its
        # internals are verified by the tests of the 'refresh_macaroons'
        # function itself)
        unbound_discharge_raw = json_body['discharge_macaroon']
        unbound_discharge = Macaroon.deserialize(unbound_discharge_raw)
        discharge = self.root_macaroon.prepare_for_request(unbound_discharge)
        v = Verifier()
        v.satisfy_general(lambda c: True)
        v.verify(self.root_macaroon, self.macaroon_random_key, [discharge])
        self.stats_increment.assert_any_call(
            'macaroon.refresh', key='just_discharge')
