import json

from django.contrib.auth.models import make_password
from django.core.serializers.json import DateTimeAwareJSONEncoder
from django.core.urlresolvers import reverse
from django.utils.timezone import now

from api.v20 import handlers
from identityprovider.models import Token
from identityprovider.models.const import AuthLogType
from identityprovider.signals import login_failed
from identityprovider.tests.test_auth import AuthLogTestCaseMixin
from identityprovider.tests.utils import SSOBaseTestCase, TimelineActionMixin

from timeline import Timeline

handler = handlers.AccountLoginHandler()


class LoginHandlerBaseTestCase(SSOBaseTestCase):

    url = reverse('api-login')

    def track_calls(self, **kw):
        self.login_failed_calls.append(kw)

    def setUp(self):
        super(LoginHandlerBaseTestCase, self).setUp()
        self.login_failed_calls = []
        login_failed.connect(self.track_calls, dispatch_uid=self.id())

        self.data = dict(
            email='foo@bar.com', password='foobar123', token_name='token_name')
        self.account = self.factory.make_account(
            email=self.data['email'], password=self.data['password'])


class LoginHandlerTestCase(LoginHandlerBaseTestCase, AuthLogTestCaseMixin):

    def do_post(
            self, status_code=200, data=None, content_type='application/json',
            **kwargs):
        if data is None:
            data = self.data

        response = self.client.post(
            self.url, data=json.dumps(data), content_type=content_type,
            **kwargs)

        self.assertEqual(response.status_code, status_code, response)
        assert response['Content-type'] == 'application/json; charset=utf-8', (
            response)
        return json.loads(response.content)

    def assert_response_correct(self, json_body, account):
        # an oauth token was created for this user
        self.assertEqual(Token.objects.all().count(), 1)
        token = Token.objects.get()
        expected_response = {
            'href': '/api/v2/tokens/oauth/%s' % token.pk,
            'token_key': token.key,
            'token_secret': token.secret,
            'token_name': 'token_name',
            'consumer_key': token.consumer.key,
            'consumer_secret': token.consumer.secret,
            'openid': account.openid_identifier,
        }
        for k, v in expected_response.iteritems():
            self.assertEqual(json_body.get(k), v)
        self.assertEqual(self.login_failed_calls, [])

    def assert_failed_login(self, code, data, extra=None):
        status_code = 403
        if code == 'INVALID_CREDENTIALS':
            status_code = 401

        if extra is None:
            extra = {}

        json_body = self.do_post(status_code=status_code, data=data)
        self.assertEqual(json_body['code'], code)
        self.assertEqual(json_body['extra'], extra)

        self.assertEqual(len(self.login_failed_calls), 1)
        failed = self.login_failed_calls[0]
        self.assertEqual(
            failed['credentials'],
            {'email': data['email'], 'password': data['password']})

        return json_body

    def test_login_required_parameters(self):
        json_body = self.do_post(status_code=400, data={})

        self.assertEqual(json_body, {
            'code': 'INVALID_DATA',
            'extra': {'email': [handlers.FIELD_REQUIRED]},
            'message': 'Invalid request data'},
        )
        self.assertEqual(self.login_failed_calls, [])

    def test_account_suspended(self):
        self.account.suspend()
        self.assert_failed_login('ACCOUNT_SUSPENDED', self.data)

    def test_account_deactivated(self):
        self.account.deactivate()
        self.assert_failed_login('ACCOUNT_DEACTIVATED', self.data)

    def test_failed_login(self):
        self.account.set_password('other-thing')
        self.account.save()
        self.assert_failed_login('INVALID_CREDENTIALS', self.data)

    def test_email_invalidated(self):
        self.account.preferredemail.invalidate()
        self.assert_failed_login('EMAIL_INVALIDATED', self.data)

    def test_password_policy(self):
        self.account.accountpassword.password = make_password('short')
        self.account.accountpassword.save()

        self.data['password'] = 'short'
        extra = {
            'reason': 'Password must be at least 8 characters long',
            'location': 'http://testserver/'}
        self.assert_failed_login(
            'PASSWORD_POLICY_ERROR', data=self.data, extra=extra)

    def test_login_token_created(self):
        assert Token.objects.all().count() == 0, 'No tokens in the system'

        before = now()
        json_body = self.do_post(status_code=201)
        after = now()

        self.assert_response_correct(json_body, self.account)
        for timestamp in ('date_created', 'date_updated'):
            timestamp = json_body.get(timestamp)
            self.assertGreaterEqual(after.isoformat(), timestamp)
            self.assertGreaterEqual(timestamp, before.isoformat())

    def test_login_token_updated(self):
        token = self.account.create_oauth_token(self.data['token_name'])
        assert Token.objects.all().count() == 1, 'One tokens in the system'

        before = now()
        json_body = self.do_post(status_code=200)
        after = now()

        self.assert_response_correct(json_body, self.account)
        expected = json.loads(
            json.dumps(token.date_created, cls=DateTimeAwareJSONEncoder))
        self.assertEqual(json_body['date_created'], expected)
        self.assertGreater(json_body['date_updated'], before.isoformat())
        self.assertLess(json_body['date_updated'], after.isoformat())

    def test_twofactor_required(self):
        self.account.twofactor_required = True
        self.account.save()

        json_body = self.do_post(status_code=401)
        self.assertEqual(json_body['code'], 'TWOFACTOR_REQUIRED')
        self.assertEqual(self.login_failed_calls, [])

    def test_twofactor(self):
        device = self.factory.make_device(account=self.account)
        self.account.twofactor_required = True
        self.account.save()

        self.data['otp'] = self.factory.make_next_otp(device)
        json_body = self.do_post(status_code=201, data=self.data)

        self.assert_response_correct(json_body, self.account)

    def test_twofactor_wrong_otp(self):
        device = self.factory.make_device(account=self.account)
        self.account.twofactor_required = True
        self.account.save()

        self.data['otp'] = str(int(self.factory.make_next_otp(device)) - 1)
        self.assert_failed_login('TWOFACTOR_FAILURE', self.data)

    def test_failed_login_creates_authlog(self):
        self.account.set_password('other-thing')
        self.account.save()
        expected_values = {
            'login_email': self.data['email'],
            'log_type': AuthLogType.API_FAIL,
            'referer': '',
            'user_agent': '',
            'remote_ip': '127.0.0.1',
            'account': None
        }
        with self.assert_authlog_matches(expected_values):
            self.assert_failed_login('INVALID_CREDENTIALS', self.data)

    def test_login_creates_authlog(self):
        expected_values = {
            'login_email': self.data['email'],
            'log_type': AuthLogType.API_NEW,
            'referer': '',
            'user_agent': '',
            'remote_ip': '127.0.0.1',
            'account': self.account.id
        }
        # This context manager magically checks when_created values
        with self.assert_authlog_matches(expected_values):
            self.do_post(status_code=201)

    def test_repeat_login_creates_authlog_with_existing(self):
        expected_values = {
            'login_email': self.data['email'],
            'log_type': AuthLogType.API_EXISTING,
            'referer': '',
            'user_agent': '',
            'remote_ip': '127.0.0.1',
            'account': self.account.id
        }
        # Create a token (i.e. the user already requested auth and a token was
        # created)
        self.account.create_oauth_token(self.data['token_name'])
        # This context manager magically checks when_created values
        with self.assert_authlog_matches(expected_values):
            self.do_post()


class LoginHandlerTimelineTestCase(LoginHandlerBaseTestCase,
                                   TimelineActionMixin):

    def test_login_timeline_records_password_checking(self):
        # Prepare and inject a timeline in the client's POST.
        empty_timeline = Timeline()
        response = self.client.post(
            self.url, data=json.dumps(self.data),
            content_type='application/json',
            **{'timeline.timeline': empty_timeline})
        self.assert_timeline_contains(request=response.wsgi_request,
                                      category='django-password-checking',
                                      detail=str(self.account.id))

    def test_login_ignore_weird_timeline(self):
        """A weird object passed as timeline should be ignored/not used."""
        weird_timeline = []
        response = self.client.post(
            self.url, data=json.dumps(self.data),
            content_type='application/json',
            **{'timeline.timeline': weird_timeline})

        # If the timeline is a weird object we shouldn't have touched it
        self.assertEqual([], response.wsgi_request.META['timeline.timeline'])

    def test_login_no_timeline(self):
        """Shouldn't barf if there's no timeline in the request at all."""
        response = self.client.post(
            self.url, data=json.dumps(self.data),
            content_type='application/json')

        self.assertEqual(response.status_code, 201)
        self.assertNotIn('timeline.timeline', response.wsgi_request.META)
