from django.core.exceptions import ValidationError

from identityprovider.fields import FIELD_REQUIRED
from identityprovider.login import PasswordPolicyError
from identityprovider.models import Account
from identityprovider.models.const import (
    AccountCreationRationale,
    AccountStatus,
)
from identityprovider.tests import DEFAULT_USER_PASSWORD
from identityprovider.tests.utils import SSOBaseTestCase

from api.v20 import registration


class RegistrationTestCase(SSOBaseTestCase):

    def setUp(self):
        super(RegistrationTestCase, self).setUp()
        self.email = self.factory.make_email_address()

        self.mock_send_impersonation_email = self.patch(
            'api.v20.registration.send_impersonation_email')
        self.mock_send_new_user_email = self.patch(
            'api.v20.registration.send_new_user_email')

    def test_register_already_registered_and_inactive(self):
        account = self.factory.make_account(
            email=self.email, status=AccountStatus.DEACTIVATED)
        assert not account.is_active

        with self.assertRaises(registration.EmailAlreadyRegistered):
            registration.register(self.sso_root_url, self.email, "", "")

        self.assertFalse(self.mock_send_impersonation_email.called)
        self.assertFalse(self.mock_send_new_user_email.called)

    def test_register_already_registered_and_active(self):
        account = self.factory.make_account(email=self.email)
        assert account.is_active

        with self.assertRaises(registration.EmailAlreadyRegistered):
            registration.register(self.sso_root_url, self.email, "", "")

        self.mock_send_impersonation_email.assert_called_once_with(
            self.sso_root_url, self.email)
        self.assertFalse(self.mock_send_new_user_email.called)

    def test_register_success(self):
        assert Account.objects.all().count() == 0

        registration.register(
            self.sso_root_url, self.email, 'SecretPassword1', 'displayname')

        self.assertEqual(Account.objects.all().count(), 1)
        account = Account.objects.get()
        self.assertEqual(account.displayname, 'displayname')
        self.assertEqual(account.preferredemail.email, self.email)
        self.assertFalse(account.is_verified)

        self.assertTrue(
            self.client.login(username=self.email, password='SecretPassword1'),
            'Password was not properly set on new account.')
        self.mock_send_new_user_email.assert_called_once_with(
            root_url=self.sso_root_url, account=account, email=self.email)
        self.assertFalse(self.mock_send_impersonation_email.called)

    def test_invalid_email_and_pw(self):
        with self.assertRaises(ValidationError) as ctx:
            registration.register(self.sso_root_url, "", '', 'displayname')

        e = ctx.exception
        self.assertEqual(e.message_dict['email'], [FIELD_REQUIRED])
        self.assertEqual(e.message_dict['password'], [FIELD_REQUIRED])

    def test_leaked_credentials(self):
        self.factory.make_leaked_credential(email='foo@example.com',
                                            password=DEFAULT_USER_PASSWORD)
        with self.assertRaises(PasswordPolicyError) as ctx:
            registration.register(self.sso_root_url, 'foo@example.com',
                                  DEFAULT_USER_PASSWORD, 'displayname')

        self.assertIn('there was a security breach', str(ctx.exception))
        self.assertIn('Ubuntu One', str(ctx.exception))

    def test_email_not_validated(self):
        account = registration.register(
            self.sso_root_url, self.email, 'SecretPassword1', 'displayname',
            email_validated=False)
        self.assertFalse(account.is_verified)

    def test_email_validated(self):
        account = registration.register(
            self.sso_root_url, self.email, 'SecretPassword1', 'displayname',
            email_validated=True)
        self.assertTrue(account.is_verified)

    def test_creation_rationale_unset_source(self):
        account = registration.register(
            self.sso_root_url, self.email, 'SecretPassword1', 'displayname')
        self.assertEqual(
            account.creation_rationale, AccountCreationRationale.API_CREATED)
        self.assertIsNone(account.creation_source)

    def test_creation_rationale_set_source(self):
        account = registration.register(
            self.sso_root_url, self.email, 'SecretPassword1', 'displayname',
            creation_source='foo bar baz')
        self.assertEqual(
            account.creation_rationale, AccountCreationRationale.API_CREATED)
        self.assertEqual(account.creation_source, 'foo bar baz')
