# Copyright 2010-2012 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).
import json

from django.core.urlresolvers import reverse
from django.http import HttpResponse
from django.utils.translation import ugettext_lazy as _

from identityprovider.models.const import AccountStatus


class EnsureJSONResponseOnAPIErrorMiddleware(object):

    def process_exception(self, request, exception):
        response = None
        if request.META['PATH_INFO'].startswith('/api/v2/'):
            # response from inside the API are always expected to have a JSON
            # content
            msg = '%s: %s' % (exception.__class__.__name__, unicode(exception))
            response = api_error(status=500, code='SERVER_ERROR', message=msg)

        return response


def api_error(status, code, message, extra=None):
    if extra is None:
        extra = {}

    extra_data = {}
    for k, v in extra.iteritems():
        if isinstance(v, list):
            v = [unicode(i) for i in v]
        elif isinstance(v, basestring):
            v = unicode(v)
        extra_data[k] = v

    data = {
        'code': code,
        'message': unicode(message),
        'extra': extra_data,
    }
    response = HttpResponse(content=json.dumps(data), status=status,
                            content_type='application/json; charset=utf-8')
    response['Vary'] = 'Accept'
    return response


ERRORS = dict(
    INVALID_DATA=(400, _("Invalid request data")),

    CAPTCHA_REQUIRED=(
        401, _('A captcha challenge is required to complete the request.')),
    INVALID_CREDENTIALS=(401, _('Provided email/password is not correct.')),
    TWOFACTOR_REQUIRED=(401, _("2-factor authentication required.")),

    ACCOUNT_SUSPENDED=(
        403, _('Your account has been suspended. Please contact login '
               'support to re-enable it')),
    ACCOUNT_DEACTIVATED=(
        403, _('Your account has been deactivated. To reactivate it, '
               'please reset your password')),
    EMAIL_INVALIDATED=(
        403, _("This email address has been invalidated. Please contact "
               "login support.")),
    CAN_NOT_RESET_PASSWORD=(403, _('Can not reset password. Please contact '
                                   'login support')),
    CAPTCHA_FAILURE=(403, _('Failed response to captcha challenge.')),
    TOO_MANY_TOKENS=(403, _('Too many non-consumed tokens exist. Further '
                            'token creation is not allowed until existing '
                            'tokens are consumed.')),
    TWOFACTOR_FAILURE=(
        403, _("The provided 2-factor key is not recognised.")),
    PASSWORD_POLICY_ERROR = (403, _("Your password doesn't meet "
                                    "security constraints.")),
    FORBIDDEN=(403, _("Insufficient permissions.")),

    RESOURCE_NOT_FOUND=(404, _("The resource requested was not found.")),

    ALREADY_REGISTERED=(409, _("The email address is already registered")),

    TOO_MANY_REQUESTS=(429, _('Too many requests. Please try again later.')),

    CAPTCHA_ERROR=(502, _('Unable to get a valid response from '
                          'reCaptcha service')),
)


class Errors(object):
    def wrap(self, code):
        status, message = ERRORS[code]

        def wrapper(**extra):
            return api_error(status, code, message, extra)
        wrapper.__name__ = code
        return wrapper

    def __getattr__(self, attr):
        if attr in ERRORS:
            return self.wrap(attr)
        else:
            raise AttributeError

errors = Errors()


def _get_account_status_text(status):
    return unicode(AccountStatus._verbose[status])


def get_minimal_account_data(account):
    """Return the minimal non-private data for the account."""
    href = reverse('api-account', args=(account.openid_identifier,))
    data = dict(
        href=href,
        openid=account.openid_identifier,
        verified=account.is_verified,
    )
    return data


def get_account_data(account, limit=10, expand=False):
    """Get the relevant data from an account to be serialized.

    Emails and tokens will be limited to the latest 10 results, where latest
    for email means creation date, and for tokens mean updated date.

    """
    data = get_minimal_account_data(account)
    email = None
    if account.preferredemail is not None:
        email = account.preferredemail.email

    emails = account.emails(limit=limit)
    tokens = account.oauth_tokens(limit=limit)
    data.update(
        email=email,
        displayname=account.displayname,
        status=_get_account_status_text(account.status),
        emails=[get_email_data(e, expand) for e in emails],
        tokens=[get_token_data(t, expand) for t in tokens],
    )
    return data


def get_email_data(email, expand=True):
    result = dict(
        verified=email.is_verified,
        href=reverse('api-email', args=(email.email,)),
    )
    if expand:
        result.update(dict(
            email=email.email,
            date_created=email.date_created,
        ))
    return result


def get_token_data(token, expand=True):
    result = dict(
        token_name=token.name,
        href=reverse('api-token', args=(token.pk,)),
    )
    if expand:
        result.update(dict(
            date_created=token.date_created,
            date_updated=token.date_updated,
            consumer_key=token.consumer.key,
            consumer_secret=token.consumer.secret,
            token_key=token.key,
            token_secret=token.secret,
        ))
    return result
