# Copyright 2010,2013 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

from __future__ import absolute_import

from urllib import urlencode

import gargoyle

from django import forms
from django.contrib import admin
from django.contrib.admin.views.main import ChangeList
from django.contrib.auth.admin import UserAdmin
from django.contrib.auth.models import User, make_password
from django.core import validators
from django.core.urlresolvers import reverse
from django.template.defaultfilters import force_escape
from django.db import connections
from django.db.models.query import QuerySet
from oauth_backend.models import Consumer as V1Consumer, Token as V1Token

from identityprovider.const import (
    AX_DATA_LABELS,
    SREG_LABELS,
)
from identityprovider.fields import CommaSeparatedField
from identityprovider.models import (
    Account,
    AccountPassword,
    APIUser,
    AuthenticationDevice,
    AuthLog,
    AuthToken,
    EmailAddress,
    InvalidatedEmailAddress,
    LeakedCredential,
    OpenIDRPConfig,
    Person,
    TeamParticipation,
    Token,
)
from identityprovider.models.const import (
    AccountStatus,
    AuthLogType,
    EmailStatus,
)
from identityprovider.widgets import (
    CommaSeparatedWidget,
    LPUsernameWidget,
    ReadOnlyDateTimeWidget,
    StatusWidget,
    account_to_lp_link,
)


DEVICE_TYPE_CHOICES = [
    (None, 'Unset'),
    ('paper', 'Printable backup codes'),
    ('yubi', 'YubiKey'),
    ('google', 'Smartphone or tablet'),
    ('generic', 'Generic HOTP device')
]


def build_account_link(account):
    url = reverse(
        'admin:identityprovider_account_change', args=(account.id,))
    snippet = '<a href="%s">%s</a> (OpenID: %s)'
    info = (url, force_escape(account.displayname), account.openid_identifier)
    return snippet % info


def build_tokens_link(account):
    qs = urlencode({'q': account.openid_identifier})
    new_style_url = reverse(
        'admin:identityprovider_token_changelist') + '?' + qs
    old_style_url = reverse(
        'admin:oauth_backend_token_changelist') + '?' + qs
    snippet = '<a href="%s">New-style</a> or <a href="%s">Old-style</a>'
    info = (new_style_url, old_style_url)
    return snippet % info


def build_auth_tokens_link(account):
    qs = urlencode({'q': account.openid_identifier})
    url = reverse('admin:identityprovider_authtoken_changelist') + '?' + qs
    return '<a href="%s">By OpenID</a>' % url


def build_auth_logs_link(label, obj):
    qs = urlencode({'q': obj})
    url = reverse('admin:identityprovider_authlog_changelist') + '?' + qs
    return '<a href="%s">By %s</a>' % (url, label)


class ApproxCountQuerySet(QuerySet):
    """Counting all rows is very expensive on large tables.

    This is a replacement for QuerySet that returns an approximation if count()
    is called with no additional constraints. In all other cases it should
    behave exactly as QuerySet.
    """

    def count(self):
        if self._result_cache is not None and not getattr(self, '_iter', None):
            return len(self._result_cache)

        query = self.query
        if any([query.where,
                query.high_mark is not None,
                query.low_mark != 0,
                query.select,
                query.group_by,
                query.having]):
            # query has constraints, run the real thing
            return self.query.get_count(using=self.db)
        else:
            # If query has no constraints, we would be simply doing
            # "SELECT COUNT(*) FROM foo"; get an approximation instead.
            cursor = connections[self.db].cursor()
            cursor.execute(
                "SELECT reltuples FROM pg_class WHERE relname = %s",
                [query.model._meta.db_table])
            return int(cursor.fetchone()[0])


class LargeTableMixin(object):

    def get_queryset(self, request):
        qs = super(LargeTableMixin, self).get_queryset(request)
        return qs._clone(klass=ApproxCountQuerySet)


class OpenIDRPConfigForm(forms.ModelForm):
    displayname = forms.CharField(label="Display name")
    trust_root = forms.CharField(validators=[validators.URLValidator()])
    logo = forms.CharField(required=False)
    allowed_user_attribs = CommaSeparatedField(
        choices=AX_DATA_LABELS.items(),
        required=False,
        widget=CommaSeparatedWidget,
    )
    allowed_ax = CommaSeparatedField(
        choices=AX_DATA_LABELS.items(),
        required=False,
        widget=CommaSeparatedWidget,
    )
    allowed_sreg = CommaSeparatedField(
        choices=SREG_LABELS.items(),
        required=False,
        widget=CommaSeparatedWidget
    )
    can_query_any_team = forms.BooleanField(label="Can query private teams",
                                            required=False)
    auto_authorize = forms.BooleanField(label="Auto-authorize",
                                        required=False)
    allow_unverified = forms.BooleanField(label="Allow unverified accounts",
                                          required=False)

    class Meta:
        model = OpenIDRPConfig
        # RemovedInDjango18Warning: Creating a ModelForm without either the
        # 'fields' attribute or the 'exclude' attribute is deprecated
        fields = '__all__'


class OpenIDRPConfigAdmin(admin.ModelAdmin):

    form = OpenIDRPConfigForm
    list_display = ('displayname', 'trust_root', 'can_query_any_team',
                    'auto_authorize', 'twofactor_required')
    list_display_links = ('displayname', 'trust_root')
    list_editable = ('twofactor_required',)
    list_filter = ('twofactor_required',)
    ordering = ('displayname',)
    search_fields = ('displayname', 'trust_root')

    class Media:
        css = {'all': ("/assets/identityprovider/admin-fixes.css",)}


class AccountPasswordInlineForm(forms.ModelForm):
    password = forms.CharField(
        widget=forms.PasswordInput, required=False,
        label="Password",
        help_text="Enter the new password in plain text, it will "
                  "be properly encrypted into the database.")

    class Meta:
        model = AccountPassword
        # RemovedInDjango18Warning: Creating a ModelForm without either the
        # 'fields' attribute or the 'exclude' attribute is deprecated
        fields = '__all__'

    def __init__(self, *args, **kwargs):
        initial = kwargs.get('initial', {})
        initial.update({'password': ''})
        kwargs['initial'] = initial
        super(AccountPasswordInlineForm, self).__init__(*args, **kwargs)


class AccountPasswordFormset(forms.models.BaseInlineFormSet):

    def __init__(self, *args, **kwargs):
        super(AccountPasswordFormset, self).__init__(*args, **kwargs)
        self.can_delete = False


class AccountPasswordInline(admin.TabularInline):

    model = AccountPassword
    form = AccountPasswordInlineForm
    formset = AccountPasswordFormset


class AuthenticationDeviceInlineForm(forms.ModelForm):

    key = forms.CharField(widget=forms.TextInput(attrs={'size': 40}),
                          label="OATH/HOTP key")
    name = forms.CharField(widget=forms.TextInput(attrs={'size': 15}),
                           label="Device name")
    device_type = forms.ChoiceField(choices=DEVICE_TYPE_CHOICES,
                                    label="Device type")

    class Meta:
        model = AuthenticationDevice
        exclude = ('account',)


class AuthenticationDeviceInline(admin.TabularInline):

    model = AuthenticationDevice
    form = AuthenticationDeviceInlineForm
    extra = 0


class EmailAddressInlineForm(forms.ModelForm):

    email = forms.CharField(widget=forms.TextInput(attrs={'size': 40}),
                            label="Email address")

    class Meta:
        model = EmailAddress
        # RemovedInDjango18Warning: Creating a ModelForm without either the
        # 'fields' attribute or the 'exclude' attribute is deprecated
        fields = '__all__'


class EmailAddressInlineFormSet(forms.models.BaseInlineFormSet):

    def clean(self):
        super(EmailAddressInlineFormSet, self).clean()
        count = 0
        for form in self.forms:
            if form.is_valid() and form.cleaned_data:
                status = form.cleaned_data['status']
                if status == EmailStatus.PREFERRED:
                    count += 1
        if count > 1:
            raise forms.ValidationError(
                "Only one email address can be preferred.")


class EmailAddressInline(admin.TabularInline):

    model = EmailAddress
    form = EmailAddressInlineForm
    formset = EmailAddressInlineFormSet
    verbose_name_plural = "Email addresses"


class LPUsernameField(forms.Field):

    widget = LPUsernameWidget


class AccountAdminForm(forms.ModelForm):

    date_created = forms.DateTimeField(widget=ReadOnlyDateTimeWidget,
                                       required=False)
    creation_source = forms.CharField(required=False)
    lp_username = LPUsernameField(label="LP username", required=False)
    status = forms.TypedChoiceField(widget=StatusWidget, coerce=int,
                                    choices=AccountStatus._get_choices())
    status_comment = forms.CharField(required=False)

    displayname = forms.CharField(label="Display name",
                                  widget=forms.TextInput)
    openid_identifier = forms.CharField(label="OpenID identifier",
                                        widget=forms.TextInput)

    class Meta:
        model = Account
        exclude = ('date_status_set',)

    def __init__(self, *args, **kwargs):
        instance = kwargs.get('instance')
        if instance:
            widget = self.declared_fields['status'].widget
            widget.date_status_set = instance.date_status_set

            widget = self.declared_fields['date_created'].widget
            widget.value = instance.date_created

            widget = self.declared_fields['lp_username'].widget
            widget.account = instance
        super(AccountAdminForm, self).__init__(*args, **kwargs)


class AccountAdmin(LargeTableMixin, admin.ModelAdmin):
    inlines = [AccountPasswordInline, EmailAddressInline,
               AuthenticationDeviceInline]
    fieldsets = (
        (None, {
            'fields': (
                'date_created', 'lp_username',
                ('creation_rationale', 'creation_source'),
                'status', 'status_comment',
                ('displayname', 'openid_identifier'),
                ('twofactor_required', 'warn_about_backup_device',
                 'need_password_reset'),
            ),
        }),
    )
    form = AccountAdminForm
    # preferred email is not listed becuse is expensive!
    list_display = (
        '__unicode__', 'openid_identifier', 'lp_username', 'status_and_source',
        'tokens_link', 'auth_tokens_link', 'auth_logs_link')
    search_fields = ('displayname', '=openid_identifier')
    list_filter = ('date_created', 'status', 'date_status_set',
                   'need_password_reset')

    def lp_username(self, account):
        return account_to_lp_link(account)

    lp_username.allow_tags = True

    def status_and_source(self, account):
        status = unicode(AccountStatus.label_for(account.status))
        if account.creation_source:
            result = '%s (%s)' % (status, account.creation_source)
        else:
            result = status
        return result

    def tokens_link(self, obj):
        return build_tokens_link(obj)

    tokens_link.allow_tags = True

    def auth_tokens_link(self, obj):
        return build_auth_tokens_link(obj)

    auth_tokens_link.allow_tags = True

    def auth_logs_link(self, obj):
        return build_auth_logs_link("Account", obj)

    auth_logs_link.allow_tags = True

    def save_formset(self, request, form, formset, change):
        if formset.model is AccountPassword:
            password = formset.forms[0].cleaned_data.get('password')
            instances = formset.save(commit=False)
            if len(instances) == 1 and password:
                instance = instances[0]
                encrypted_password = make_password(password)
                instance.password = encrypted_password
                instance.save()
        elif (formset.model is EmailAddress and
                form.instance.status == AccountStatus.DELETED):
            # we're not saving the formset as the email addresses have
            # been deleted. We need to set the new / changed / deleted
            # objects attributes normally set by saving.
            formset.new_objects = []
            formset.changed_objects = []
            formset.deleted_objects = []
            return
        else:
            super(AccountAdmin, self).save_formset(
                request, form, formset, change)


class APIUserAdminForm(forms.ModelForm):

    password = forms.CharField(required=False)

    class Meta:
        model = APIUser
        # RemovedInDjango18Warning: Creating a ModelForm without either the
        # 'fields' attribute or the 'exclude' attribute is deprecated
        fields = '__all__'

    def __init__(self, *args, **kwargs):
        kwargs['initial'] = {'password': ''}
        super(APIUserAdminForm, self).__init__(*args, **kwargs)


class APIUserAdmin(admin.ModelAdmin):

    model = APIUser
    form = APIUserAdminForm
    search_fields = ['username']

    def save_model(self, request, obj, form, change):
        password = form.cleaned_data.get('password')
        if password:
            obj.set_password(password)
        obj.save()


class EmailAddressForm(forms.ModelForm):

    email = forms.CharField(widget=forms.TextInput(attrs={'size': 40}),
                            label="Email address")

    class Meta:
        model = EmailAddress
        # RemovedInDjango18Warning: Creating a ModelForm without either the
        # 'fields' attribute or the 'exclude' attribute is deprecated
        fields = '__all__'


class EmailAddressAdmin(LargeTableMixin, admin.ModelAdmin):
    form = EmailAddressForm
    model = EmailAddress
    raw_id_fields = ('account',)
    list_display = (
        'email', 'status', 'account_link', 'tokens_link', 'auth_tokens_link',
        'auth_logs_link')
    search_fields = ('=email',)

    def account_link(self, obj):
        if obj.account is None:
            return "None"
        return build_account_link(obj.account)

    account_link.allow_tags = True

    def tokens_link(self, obj):
        if obj.account is None:
            return "None"
        return build_tokens_link(obj.account)

    tokens_link.allow_tags = True

    def auth_tokens_link(self, obj):
        if obj.account is None:
            return "None"
        return build_auth_tokens_link(obj.account)

    auth_tokens_link.allow_tags = True

    def auth_logs_link(self, obj):
        return build_auth_logs_link("Email", obj)

    auth_logs_link.allow_tags = True


class InvalidatedEmailAddressForm(forms.ModelForm):

    email = forms.CharField(widget=forms.TextInput(attrs={'size': 40}),
                            label="Email address")

    class Meta:
        model = InvalidatedEmailAddress
        # RemovedInDjango18Warning: Creating a ModelForm without either the
        # 'fields' attribute or the 'exclude' attribute is deprecated
        fields = '__all__'


class InvalidatedEmailAddressAdmin(EmailAddressAdmin):
    form = InvalidatedEmailAddressForm
    model = InvalidatedEmailAddress
    list_display = ('email', 'account_link')


class AuthLogChangeList(ChangeList):

    """ChangeList for AuthLog entries.

    Get a custom title that removes the "Change" legend and return pertinent
    links depending on whether we have an account linked to the record.
    """

    def __init__(self, *args, **kwargs):
        super(AuthLogChangeList, self).__init__(*args, **kwargs)
        self.title = 'Authentication logs'

    def url_for_result(self, result):
        if result.account:
            return reverse(
                'admin:identityprovider_account_change', args=(
                    result.account.id,))
        return ''


class AuthLogAdmin(LargeTableMixin, admin.ModelAdmin):

    """Admin for AuthLog entries.

    AuthLog is "read-only" since editing logs doesn't make much sense. So
    permission to add or delete items is disabled. Also, a custom ChangeList
    class is used for further read-only-oriented customization.

    The record change page is a read-only view of the data.
    """

    model = AuthLog
    date_hierarchy = 'when_created'
    search_fields = ('=login_email',)
    list_display = ('when_created', 'verbose_log_type', 'account_or_empty',
                    'login_email', 'remote_ip', 'user_agent', 'referer')
    list_display_links = ('account_or_empty',)
    fields = readonly_fields = (
        'when_created', 'verbose_log_type', 'account', 'login_email',
        'remote_ip', 'user_agent', 'referer')

    def account_or_empty(self, inst):
        if inst.account:
            return inst.account
        return ''
    account_or_empty.short_description = 'Account'

    def verbose_log_type(self, inst):
        return unicode(AuthLogType._label_for(inst.log_type))
    verbose_log_type.short_description = 'Type of login'

    def has_add_permission(self, request, obj=None):
        return False

    def has_delete_permission(self, request, obj=None):
        return False

    def get_changelist(self, request, **kwargs):
        return AuthLogChangeList


class AuthTokenAdmin(LargeTableMixin, admin.ModelAdmin):
    model = AuthToken
    search_fields = ('email', 'requester_email', 'displayname',
                     '=requester__openid_identifier', 'hashed_token')
    list_filter = ('token_type',)
    list_display = ('short_token', 'token_type', 'requester', 'email',
                    'displayname', 'date_created', 'date_consumed')
    readonly_fields = ('hashed_token',)

    def get_search_results(self, request, queryset, search_term):
        """Search also for the hashed version of the entered raw token."""
        queryset, use_distinct = super(
            AuthTokenAdmin, self).get_search_results(
                request, queryset, search_term)
        # Augment with search for hash
        queryset |= self.model.objects.filter(
            hashed_token=self.model.hash_raw(search_term))
        return queryset, use_distinct


class TokenAdmin(LargeTableMixin, admin.ModelAdmin):
    model = Token
    search_fields = ('name', 'scope', 'consumer__displayname',
                     '=consumer__openid_identifier')
    list_filter = ('scope',)
    list_display = ('name', 'key', 'scope', 'consumer',
                    'date_created', 'date_updated')


class LeakedCredentialAdmin(LargeTableMixin, admin.ModelAdmin):
    form = EmailAddressForm
    model = LeakedCredential
    list_display = ('email', 'source')
    search_fields = ('email',)
    list_filter = ('source',)


class TeamParticipationInline(admin.TabularInline):
    model = TeamParticipation
    fk_name = 'person'
    extra = 0


class PersonAdmin(admin.ModelAdmin):
    inlines = [TeamParticipationInline]
    model = Person

    # this model can not be edited since is sycned from Launchpad

    def get_readonly_fields(self, request, obj=None):
        return [field.name for field in obj._meta.fields]

    def has_add_permission(self, request, obj=None):
        return False

    def has_delete_permission(self, request, obj=None):
        return False


class V1ConsumerAdmin(admin.ModelAdmin):

    list_display = ('user', 'created_at', 'updated_at')
    search_fields = ('=user__username',)


class V1TokenAdmin(admin.ModelAdmin):

    list_display = ('name', 'consumer', 'created_at', 'updated_at', 'token')
    search_fields = ('=consumer__user__username',)


admin.site.register(Account, AccountAdmin)
admin.site.register(APIUser, APIUserAdmin)
admin.site.register(AuthToken, AuthTokenAdmin)
admin.site.register(AuthLog, AuthLogAdmin)
admin.site.register(EmailAddress, EmailAddressAdmin)
admin.site.register(InvalidatedEmailAddress, InvalidatedEmailAddressAdmin)
admin.site.register(LeakedCredential, LeakedCredentialAdmin)
admin.site.register(V1Consumer, V1ConsumerAdmin)
admin.site.register(V1Token, V1TokenAdmin)
admin.site.register(Person, PersonAdmin)
admin.site.register(OpenIDRPConfig, OpenIDRPConfigAdmin)
admin.site.register(Token, TokenAdmin)


# customize User admin for large tables
class LargeTableUserAdmin(LargeTableMixin, UserAdmin):

    # Drop 'groups' from the standard user list_filter, as we
    # don't use them, and they lead to slower queries.
    list_filter = ('is_staff', 'is_superuser', 'is_active')


admin.site.unregister(User)
admin.site.register(User, LargeTableUserAdmin)


# Add custom help text in gargoyle.Switch ModelAdmin
class CustomSwitchForm(forms.ModelForm):
    def __init__(self, *args, **kwargs):
        super(CustomSwitchForm, self).__init__(*args, **kwargs)
        self.fields['value'].help_text = (
            'For GeoIP based location, set status = SELECTIVE and'
            ' value = {"request": {"country_code": [["e", "CN"]]}}'
            ' (eg. Switch disabled if client geolocated in China)')


class SwitchAdmin(gargoyle.admin.SwitchAdmin):
    form = CustomSwitchForm


admin.site.unregister(gargoyle.models.Switch)
admin.site.register(gargoyle.models.Switch, SwitchAdmin)
