from __future__ import unicode_literals

from urlparse import urlparse
from cStringIO import StringIO

from django.conf import settings
from django.core.handlers.base import BaseHandler
from django.core.handlers.wsgi import WSGIRequest
from django.test import RequestFactory
from requests import Response
from requests.cookies import cookiejar_from_dict
from requests.hooks import dispatch_hook
from requests.structures import CaseInsensitiveDict
from requests.utils import get_encoding_from_headers

from ssoclient.v2 import V2ApiClient
from ssoclient.v2.http import ApiSession

from identityprovider.timeline_helpers import (
    maybe_contextmanager,
    get_request_timing_function,
)


class TimelineRecordingApiSession(ApiSession):

    def __init__(self, endpoint, timer=None, http_host=None):
        super(TimelineRecordingApiSession, self).__init__(endpoint)
        self.timer = timer
        self.http_host = http_host

    def request(self, method, url, **kwargs):
        # When accessing the API via the Web UI, we need to set the proper
        # Host header so Django methods such as request.build_absolute_uri
        # return valid URLS (instead of localhost URLs).
        if self.http_host:
            kwargs.setdefault('headers', {}).setdefault('Host', self.http_host)
        category = 'ssoclient-{}'.format(method)
        with maybe_contextmanager(self.timer, category, url,
                                  allow_nested=True):
            return super(TimelineRecordingApiSession, self).request(
                method, url, **kwargs)


class TimelineRecordingApiClient(V2ApiClient):

    def __init__(self, endpoint, **kwargs):
        self.session = TimelineRecordingApiSession(endpoint, **kwargs)


def get_api_client(request):
    # TODO: configure caching, auth
    timer = get_request_timing_function(request)

    http_host = request.META.get('HTTP_HOST')
    if not http_host:
        http_host = urlparse(request.build_absolute_uri()).netloc

    api = TimelineRecordingApiClient(
        settings.API_HOST + settings.API_URL, timer=timer, http_host=http_host)
    if request.META['SERVER_NAME'] == "testserver":
        api.session.mount('http://', DjangoWSGIAdapter(request))
        api.session.mount('https://', DjangoWSGIAdapter(request))
    return api


class SubRequestHandler(BaseHandler):
    """A HTTP Handler for sub-requests.

    Uses the WSGI interface to compose requests, but returns
    the raw HttpResponse object.

    A modified version of django.test.client.ClientHandler.
    """

    def __call__(self, environ):
        # Set up middleware if needed. We couldn't do this earlier, because
        # settings weren't available.
        if self._request_middleware is None:
            self.load_middleware()
        request = WSGIRequest(environ)
        response = self.get_response(request)
        return response


class DjangoSubRequester(RequestFactory):

    handler = SubRequestHandler()

    def request(self, **request):
        return self.handler(self._base_environ(**request))


class DjangoWSGIAdapter(object):
    """A requests transport adapter for django.

    Resolves the request internally via django rather that extenally via HTTP.
    To be used when the target url is also hosted by the same django app.
    """

    _encode_body_methods = set(['PATCH', 'POST', 'PUT', 'TRACE'])

    def __init__(self, request):
        super(DjangoWSGIAdapter, self).__init__()
        self.parent_request = request

    def send(self, request, **kwargs):
        django_response = self.get_django_response(request)
        return self.build_response(request, django_response)

    def close(self):
        pass

    def get_django_response(self, request):
        url = request.url
        data = {}
        method = request.method
        kwargs = {}
        if method in self._encode_body_methods:
            if 'Content-Type' in request.headers:
                content_type = request.headers['Content-Type'].split(';', 1)[0]
                kwargs['content_type'] = content_type
            data = request.body

        env = {
            'REMOTE_ADDR': '127.0.0.1',
            'SERVER_NAME': self.parent_request.META['SERVER_NAME'],
            'SERVER_PORT': self.parent_request.META['SERVER_PORT'],
            'wsgi.url_scheme': self.parent_request.META['wsgi.url_scheme'],
        }
        # convert regular headers into django style headers
        headers = request.headers.items()
        env.update(
            ('HTTP_' + k.replace('-', '_').upper(), v) for k, v in headers
        )
        requester = DjangoSubRequester(**env)
        action = getattr(requester, method.lower())
        kwargs.update(env)
        return action(url, data, **kwargs)

    def build_response(self, req, resp):
        response = Response()

        response.status_code = resp.status_code
        response.headers = CaseInsensitiveDict((k, v) for k, v in resp.items())

        response.encoding = get_encoding_from_headers(response.headers)
        response.raw = StringIO(resp.content)
        response.reason = None

        if isinstance(req.url, bytes):
            response.url = req.url.decode('utf-8')
        else:
            response.url = req.url

        # Convert from django's SimpleCookie to request's CookieJar
        cookiejar_from_dict(resp.cookies, response.cookies)

        # context
        response.request = req
        response.connection = self

        response = dispatch_hook('response', req.hooks, response)
        return response
