# -*- coding: utf-8 -*-
#
# Copyright 2010-2016 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

from __future__ import unicode_literals

import base64
import datetime
import json
import logging

from django.conf import settings
from django.contrib.auth.backends import ModelBackend
from django.contrib.auth.hashers import BCryptPasswordHasher
from django.contrib.auth.models import (
    check_password,
    is_password_usable,
    make_password,
)
from django.core.cache import cache
from django.core.exceptions import ValidationError
from django.http import HttpResponse
from django.utils.translation import ugettext_lazy as _
from django.utils.timezone import now
from oauthlib.oauth1 import RequestValidator, ResourceEndpoint
from pymacaroons import Macaroon, Verifier

from identityprovider.login import AuthenticationError, AccountDeactivated
from identityprovider.models import (
    Account,
    AccountPassword,
    LeakedCredential,
    Token,
)
from identityprovider.models.api import APIUser
from identityprovider.models.const import EmailStatus, TokenScope
from identityprovider.stats import stats
from identityprovider.timeline_helpers import (
    get_request_timing_function,
    maybe_contextmanager,
)
from identityprovider.utils import (
    validate_launchpad_password, get_object_or_none)
from oauth_backend.models import Token as V1Token


logger = logging.getLogger(__name__)

MACAROON_TIMESTAMP_FORMAT = '%Y-%m-%dT%H:%M:%S.%f'


class BCrypt13PasswordHasher(BCryptPasswordHasher):
    rounds = 13
    algorithm = 'bcrypt13'


class LaunchpadBackend(object):
    INVALID = _('Password invalid.')

    valid_email_statuses = (
        EmailStatus.PREFERRED, EmailStatus.VALIDATED, EmailStatus.NEW)

    def __init__(self):
        self.password_dirty = False

    def authenticate(self, username=None, password=None,
                     encrypted_password=None, request=None):
        account = None
        if request:
            authenticated = SSOOAuthAuthentication().is_authenticated(request)
            if authenticated:
                account = request.user
            else:
                timer = get_request_timing_function(request)
                account = self.authenticate_by_username(
                    username, password, encrypted_password, timer=timer)
        else:
            account = self.authenticate_by_username(username, password,
                                                    encrypted_password)
        if account:
            account.last_login = now()
        return account

    def authenticate_by_username(self, username=None, password=None,
                                 encrypted_password=None, timer=None):
        try:
            account = Account.objects.select_related('accountpassword')
            account = account.get(
                emailaddress__email__iexact=username,
                emailaddress__status__in=self.valid_email_statuses)
        except Account.DoesNotExist:
            return None
        if not account.is_active:
            return None

        try:
            ap = account.accountpassword
            if ap is None:
                return None
        except AccountPassword.DoesNotExist:
            # password lost somehow, user should reset it
            return None

        try:
            ap = self._validate_password(
                ap, password, encrypted_password, timer=timer)
        except ValidationError:
            return None

        # Skip leak checks in read-only mode, since it makes no sense to ask
        # the user to change his password if he really can't.
        # (Plus we can't even update the last_leak_check time).
        if not settings.READ_ONLY_MODE:
            # Snapshot the last_leak_check value here, to avoid a race
            # condition where a leak is added after we've verified this
            # password but before we set its last_leak_check to that point in
            # time.
            last_leak_check = now()
            # check if password was leaked for this email.
            if self.password_leaked(username, password, ap, timer=timer):
                account.need_password_reset = True
                account.save()
            ap.last_leak_check = last_leak_check
            ap.save(update_fields=['last_leak_check'])

        return ap.account

    def password_leaked(self, email, password, ap, timer=None):
        leaks = LeakedCredential.objects.filter(email=email)
        # last_leak_check can be null for records which were created before
        # that field was added, and which haven't been updated by a password
        # change, login attempt, or leak check. Handle that here.
        if ap and ap.last_leak_check:
            # Ony check against leaks newer or equal than the last time this
            # password was checked for leaks
            leaks = leaks.filter(when_created__gte=ap.last_leak_check)
        for leak in leaks:
            try:
                self._validate_raw_password(
                    ap, password, hashed_password=leak.password,
                    update=False, timer=timer)
                return True
            except ValidationError:
                # password did not match, keep trying
                pass
        return False

    def _validate_password(self, ap, password, hashed_password, timer=None):
        try:
            if password is not None:
                ap = self._validate_raw_password(ap, password, timer=timer)
            elif hashed_password is not None:
                if hashed_password != ap.password:
                    raise ValidationError(self.INVALID)
            else:
                raise ValidationError(self.INVALID)
        except (UnicodeDecodeError, UnicodeEncodeError):
            # Password not ASCII, we must fail
            raise ValidationError(self.INVALID)

        return ap

    def _validate_raw_password(self, ap, password, hashed_password=None,
                               update=True, timer=None):
        valid = False
        if hashed_password is None:
            hashed_password = ap.password
        identifier = (str(ap.account.id) if ap
                      else "(direct password comparison)")
        with maybe_contextmanager(
                timer, "django-password-checking", identifier):
            usable_by_django = is_password_usable(hashed_password)
            if usable_by_django:
                valid = check_password(password, hashed_password,
                                       setter=self._set_dirty)
            else:
                self._set_dirty(password)

        if not valid:
            with maybe_contextmanager(
                    timer, "validate-launchpad-password", identifier):
                valid = validate_launchpad_password(password, hashed_password)

        if not valid:
            raise ValidationError(self.INVALID)

        if update and self.password_dirty:
            new_password = make_password(password)
            AccountPassword.objects.filter(id=ap.id).update(
                password=new_password)
            # reload from db to update in-mem object
            ap = AccountPassword.objects.get(id=ap.id)

        return ap

    def _set_dirty(self, password):
        self.password_dirty = True

    def get_user(self, user_id):
        """Return the Account object instead of Django's User object."""
        return get_object_or_none(Account.objects.select_related(), pk=user_id)

    def supports_anonymous_user(self):
        return False

    def supports_object_permissions(self):
        return False

    def has_perm(self, account, *args, **kwargs):
        user = getattr(account, 'user', account)
        return ModelBackend().has_perm(user, *args, **kwargs)


def decode(value):
    # try to decode as UTF-8
    try:
        decoded = value.decode('utf-8')
    except UnicodeDecodeError:
        # or fallback to ISO-8859-1 according to RFC 2617
        decoded = value.decode('iso-8859-1')
    return decoded


def basic_authenticate(username, password):
    """Function used by lazr.authentication middleware"""
    # decode username and password before checking against the database
    username = decode(username)
    password = decode(password)
    api_user = APIUser.authenticate(username, password)
    if api_user is not None:
        return api_user
    user = LaunchpadBackend().authenticate(username, password)
    return user


class SSORequestValidator(RequestValidator):
    """Custom OAuth-signed request validator for SSO.

    This class implements all needed methods from RequestValidator so we can
    use it along with the ResourceEndpoint to provide proper OAuth signature
    validations.

    For details about each property and method below, please refer to:
    http://oauthlib.readthedocs.org/en/latest/oauth1/validator.html

    """

    def __init__(self, token_filter_fn=None):
        super(SSORequestValidator, self).__init__()
        self.token_filter_fn = token_filter_fn or (lambda token: True)

    def _nonce_key(self, client_key, token_key, nonce):
        return 'oauth-%s-%s-%s' % (client_key, token_key, nonce)

    @property
    def enforce_ssl(self):
        return settings.OAUTH_ENFORCE_SSL

    @property
    def safe_characters(self):
        result = super(SSORequestValidator, self).safe_characters
        result.update({'-', '_', '.', '~'})
        return result

    @property
    def timestamp_lifetime(self):
        return settings.OAUTH_REQUEST_MAX_AGE_SECONDS

    @property
    def nonce_length(self):
        # The nonce length will vary between OAuth libraries. The old
        # python-oauth will use very small nonces, and oauthlib uses nonce with
        # length 29. Provide min and max values that make sense and cover
        # common usages.
        return 5, 100

    @property
    def client_key_length(self):
        # Client keys are the account openid identifier, which is 7 chars,
        # hard coded in the SQL stored procedure "generate_openid_identifier".
        return 7, 7

    @property
    def access_token_length(self):
        # V1 tokens are 50 chars, V2 tokens are 60 chars (min, max).
        return 50, 60

    @property
    def dummy_client(self):
        """Dummy client used when an invalid client key is supplied.

        The dummy client should be associated with either a client secret, a
        rsa key or both depending on which signature methods are supported.
        Providers should make sure that get_client_secret(dummy_client) or
        get_rsa_key(dummy_client) return a valid secret or key for the dummy
        client.

        """
        return 'dummy_client'

    @property
    def dummy_access_token(self):
        """Dummy access token used when an invalid token was supplied.

        The dummy access token should be associated with an access token secret
        such that get_access_token_secret(.., dummy_access_token) returns a
        valid secret.

        """
        return 'dummy_access_token'

    def validate_client_key(self, client_key, request):
        """Validate that client key is a registered and valid client."""
        # Do nothing, avoid extra access to the DB -- we will fetch the client
        # (Account) on get_client_key
        return True

    def validate_access_token(self, client_key, token, request):
        """Validate that supplied access token is registered and valid."""
        # Do nothing, avoid extra access to the DB -- we will fetch the client
        # (Token) on get get_access_token_secret
        return True

    def validate_realms(self, client_key, token, request, **kwargs):
        """Validate access to the request realm."""
        return True  # We don't check realms values.

    def validate_timestamp_and_nonce(
            self, client_key, timestamp, nonce, request, request_token=None,
            access_token=None):
        """Validate that the nonce has not been used before.

        Parameters:
            client_key - The client/consumer key.
            timestamp - The oauth_timestamp parameter.
            nonce - The oauth_nonce parameter.
            request_token - Request token string, if any.
            access_token - Access token string, if any.
            request - An oauthlib.common.Request object.

        Return True or False.

        """
        # timestamp freshness was already checked by the ResourceEndpoint
        # using the self.timestamp_lifetime value

        key = self._nonce_key(client_key, access_token, nonce)
        nonce_valid = cache.get(key) is None
        if nonce_valid:
            cache.set(key, nonce, timeout=settings.OAUTH_NONCE_TIMEOUT_SECONDS)
        else:
            logger.info(
                'SSORequestValidator.validate_timestamp_and_nonce: client_key '
                '%r nonce INVALID %r.', client_key, nonce)

        return nonce_valid

    def get_client_secret(self, client_key, request):
        """Retrieve the client secret associated with the client key.

        Parameters:
            client_key - The client/consumer key.
            request - An oauthlib.common.Request object.

        Return the client secret as a string.

        This method must allow the use of a dummy client_key value. Fetching
        the secret using the dummy key must take the same amount of time as
        fetching a secret for a valid client.

        """
        logger.debug('SSORequestValidator.get_client_secret: client_key %r.',
                     client_key)
        account = Account.objects.active_by_openid(client_key)
        if account is not None:
            result = account.secret
        else:
            result = self.dummy_client
        return result

    def _get_v2_token(self, client_key, token, request):
        """Retrieve a VERSION_2 scoped Token."""
        try:
            v2_token = Token.objects.get(
                pk=token, consumer__openid_identifier=client_key)
        except Token.DoesNotExist:
            v2_token = None
        return v2_token

    def _get_v1_token(self, client_key, token, request):
        """Retrieve a V1Token, return it only if target url is whitelisted."""
        try:
            v1_token = V1Token.objects.get(
                pk=token, consumer__user__username=client_key)
        except V1Token.DoesNotExist:
            v1_token = None
        return v1_token

    def get_access_token_secret(self, client_key, token, request):
        """Retrieve the shared secret associated with the access token.

        Parameters:
            client_key - The client/consumer key.
            token - The access token string.
            request - An oauthlib.common.Request object.

        Return the token secret as a string.

        This method must allow the use of a dummy values and the running time
        must be roughly equivalent to that of the running time of valid values.

        """
        logger.debug('SSORequestValidator.get_access_token_secret: client_key '
                     '%r token %r.', client_key, token)
        token_obj = self._get_v2_token(client_key, token, request)
        if token_obj is None:
            token_obj = self._get_v1_token(client_key, token, request)

        if token_obj is not None and self.token_filter_fn(token_obj):
            result = token_obj.secret
        else:
            result = self.dummy_access_token

        return result


def validate_oauth_signature(
        http_url, http_method, headers, token_filter_fn=None):
    """Validate an OAuth signature for http_url."""
    validator = SSORequestValidator(token_filter_fn)
    provider = ResourceEndpoint(validator)
    valid, oauth_request = provider.validate_protected_resource_request(
        http_url, http_method=http_method, headers=headers)
    if not valid:
        oauth_request = None
    return valid, oauth_request


class SSOOAuthAuthentication(object):
    """Custom class to perform OAuth authentication.

    Validate the given OAuth signature, either by header or query string, and
    confirm that the consumer belongs to an SSO active Account.
    """

    def __init__(self, *args, **kwargs):
        self.require_v2_tokens = kwargs.pop('require_v2_tokens', True)
        self.realm = kwargs.pop('realm', 'Ubuntu Single Sign On API')
        super(SSOOAuthAuthentication, self).__init__()

    def is_authenticated(self, request):
        """Return if the given request is correctly oauth-signed.

        If credentials are valid, ensure the SSO Account attached to the
        request is active.

        Decouple this method from piston.authentication.OAuthAuthentication so
        we can check token and account properties without duplicating access
        to the DB.

        """

        def token_allowed(token):
            # Require that the signing token was generated via v2 api
            is_v2_token = getattr(token, 'scope', None) == TokenScope.VERSION_2
            return not self.require_v2_tokens or is_v2_token

        headers = {}
        authorization = request.META.get('HTTP_AUTHORIZATION')
        if authorization:
            headers['Authorization'] = authorization
        try:
            valid, oauth_request = validate_oauth_signature(
                http_url=request.build_absolute_uri(),
                http_method=request.method, headers=headers,
                token_filter_fn=token_allowed)
        except ValueError:
            valid = False
        except:
            logger.exception(
                'SSOOAuthAuthentication.is_authenticated: could not validate:')
            raise

        if valid:
            # signature is valid, account exists and is active
            openid_identifier = oauth_request.client_key
            account = Account.objects.get(openid_identifier=openid_identifier)
            request.user = account
            request.consumer = account
            request.throttle_extra = account.id

        return valid

    def challenge(self):
        response = HttpResponse("Authorization Required", status=401)
        response['WWW-Authenticate'] = 'OAuth realm="%s"' % self.realm
        return response


def _get_own_caveat(macaroon):
    """Get a caveat for this service from the Macaroon."""
    try:
        # isolate 3rd-party caveat that concerns to us (hinted by 'location')
        (sso_caveat,) = [c for c in macaroon.third_party_caveats()
                         if c.location == settings.MACAROON_SERVICE_LOCATION]
    except:
        # the macaroon doesn't have a location for this service
        raise AuthenticationError("The received macaroon is not for ours")

    caveat_info = _decrypt_caveat(sso_caveat.caveat_id)
    return sso_caveat.caveat_id, caveat_info


def _open_caveat_v0(info):
    """Open the caveat_id, v0 (unversioned, really) format."""
    caveat_info_raw = settings.CRYPTO_SSO_PRIVKEY.decrypt(
        base64.b64decode(info))
    caveat_info = json.loads(caveat_info_raw)
    return caveat_info


def _open_caveat_v1(info):
    """Open the caveat_id, v1 format."""
    assert info['version'] == 1  # only supported version
    info['3rdparty'] = settings.CRYPTO_SSO_PRIVKEY.decrypt(
        base64.b64decode(info.pop('secret')))
    return info


def _decrypt_caveat(raw_caveat_id):
    """Decrypt and decode the caveat id info."""
    try:
        caveat_info = json.loads(raw_caveat_id)
        function = _open_caveat_v1
        stats_key = '1'
    except:
        # bad data, or maybe the v0 (unversioned, really) format
        caveat_info = raw_caveat_id
        function = _open_caveat_v0
        stats_key = '0'

    try:
        caveat_info = function(caveat_info)
    except:
        raise ValidationError("Bad info in the caveat_id")

    stats.increment('macaroon.auth.caveat_version', key=stats_key)
    return caveat_info


def _build_discharge(macaroon_key, caveat_id, account,
                     last_auth=None, expires=None):
    """Build a typical discharge macaroon."""
    service_location = settings.MACAROON_SERVICE_LOCATION

    # create the new discharge macaroon with same location, key and
    # identifier than it's original 3rd-party caveat (so they can
    # be matched and verified)
    d = Macaroon(
        location=service_location,
        key=macaroon_key,
        identifier=caveat_id,
    )

    # add the account info
    account_info = base64.b64encode(json.dumps({
        'openid': account.openid_identifier,
        'email': account.preferredemail.email,
        'displayname': account.displayname,
        'is_verified': account.is_verified,
    }).encode("utf8"))
    d.add_first_party_caveat(service_location + '|account|' + account_info)

    # add timestamp values
    now_string = now().strftime(MACAROON_TIMESTAMP_FORMAT)
    if last_auth is None:
        last_auth = now_string
    if expires is None:
        expires_ts = now() + datetime.timedelta(seconds=settings.MACAROON_TTL)
        expires = expires_ts.strftime(MACAROON_TIMESTAMP_FORMAT)

    d.add_first_party_caveat(service_location + '|valid_since|' + now_string)
    d.add_first_party_caveat(service_location + '|last_auth|' + last_auth)
    d.add_first_party_caveat(service_location + '|expires|' + expires)
    return d


def build_discharge_macaroon_from_root(account, root_macaroon_raw):
    """Build a discharge macaroon from a root one.

    This function is deprecated; will be removed when clients stop hitting
    the handler passing the root macaroon.
    """
    try:
        root_macaroon = Macaroon.deserialize(root_macaroon_raw)
    except:
        raise ValidationError("The received Macaroon is corrupt")

    # get the raw caveat id and the decrypted deserialized info
    raw_caveat_id, caveat_info = _get_own_caveat(root_macaroon)

    # create a discharge macaroon with same location, key and
    # identifier than it's original 3rd-party caveat (so they can
    # be matched and verified)
    d = _build_discharge(caveat_info['3rdparty'], raw_caveat_id, account)

    # return the properly prepared discharge macaroon
    discharge = root_macaroon.prepare_for_request(d)
    return discharge


def build_discharge_macaroon(account, raw_caveat_id):
    """Build a discharge macaroon from the caveat id."""
    # get the decrypted deserialized info
    caveat_info = _decrypt_caveat(raw_caveat_id)

    # create a discharge macaroon with same location, key and
    # identifier than it's original 3rd-party caveat (so they can
    # be matched and verified)
    discharge = _build_discharge(
        caveat_info['3rdparty'], raw_caveat_id, account)

    # return the unbound discharge macaroon
    return discharge


def _account_from_macaroon(macaroon, key, discharge=None):
    """Get the new discharge macaroon."""
    service_location = settings.MACAROON_SERVICE_LOCATION
    info_holder = {}

    def checker(caveat):
        """Extract the openid."""
        source, key, value = caveat.split("|", 2)
        if source == service_location and key == 'account':
            acc = json.loads(base64.b64decode(value).decode("utf8"))
            info_holder['openid'] = acc['openid']
        elif source == service_location and key == 'last_auth':
            info_holder['last_auth'] = value
        elif source == service_location and key == 'expires':
            info_holder['expires'] = value
        return True

    # assure that the received pair is safe
    v = Verifier()
    v.satisfy_general(checker)
    try:
        if discharge is None:
            v.verify(macaroon, key)
        else:
            v.verify(macaroon, key, [discharge])
    except:
        raise AuthenticationError("Not verifying macaroons")

    # verify that the account is still active and password didn't change
    account = Account.objects.active_by_openid(info_holder['openid'])
    if account is None:
        raise AccountDeactivated()
    if account.accountpassword.date_changed is not None:
        last_auth_ts = datetime.datetime.strptime(info_holder['last_auth'],
                                                  MACAROON_TIMESTAMP_FORMAT)
        if account.accountpassword.date_changed > last_auth_ts:
            raise AuthenticationError("Password changed")

    return account, info_holder


def refresh_macaroons(root_macaroon_raw, discharge_macaroon_raw):
    """Refresh a root/discharge pair with a new discharge macaroon.

    This function is deprecated; will be removed when clients stop hitting
    the handler passing the root macaroon.
    """
    try:
        root_macaroon = Macaroon.deserialize(root_macaroon_raw)
        discharge_macaroon = Macaroon.deserialize(discharge_macaroon_raw)
    except:
        raise ValidationError("The received Macaroons are corrupt")

    # get the raw caveat id and the decrypted deserialized info
    raw_caveat_id, caveat_info = _get_own_caveat(root_macaroon)

    account, macaroon_info = _account_from_macaroon(
        root_macaroon, caveat_info['roothash'], discharge_macaroon)

    # create the new discharge macaroon with same location, key and
    # identifier than it's original 3rd-party caveat (so they can
    # be matched and verified), and keeping the last_auth and expires from
    # the original discharge macaroon
    d = _build_discharge(
        caveat_info['3rdparty'], raw_caveat_id, account,
        last_auth=macaroon_info['last_auth'], expires=macaroon_info['expires'])

    # return the properly prepared discharge macaroon
    discharge = root_macaroon.prepare_for_request(d)
    return discharge


def refresh_discharge(discharge_macaroon_raw):
    """Refresh a discharge with a new discharge macaroon."""
    try:
        discharge_macaroon = Macaroon.deserialize(discharge_macaroon_raw)
    except:
        raise ValidationError("The received Macaroons are corrupt")

    # get the decrypted deserialized info
    caveat_info = _decrypt_caveat(discharge_macaroon.identifier)
    account, macaroon_info = _account_from_macaroon(
        discharge_macaroon, caveat_info['3rdparty'])

    # create the new discharge macaroon with same location, key and
    # identifier than it's original 3rd-party caveat (so they can
    # be matched and verified), and keeping the last_auth and expires from
    # the original discharge macaroon
    discharge = _build_discharge(
        caveat_info['3rdparty'], discharge_macaroon.identifier, account,
        last_auth=macaroon_info['last_auth'], expires=macaroon_info['expires'])

    # return the unbound discharge macaroon
    return discharge
