# Copyright 2016 Canonical
# All Rights Reserved

"""Layer to provide easy cryptographic functions."""

from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.asymmetric import padding
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization


_RSA_PADDING = padding.OAEP(
    mgf=padding.MGF1(algorithm=hashes.SHA1()),
    algorithm=hashes.SHA1(),
    label=None)


class Crypto(object):
    """Lightweight layer with simple interface to encrypt/decrypt."""

    def __init__(self, key):
        self.key = key

    def encrypt(self, data):
        """Encrypt the received data."""
        return self.key.encrypt(data, _RSA_PADDING)

    def decrypt(self, data):
        """Decrypt the received data."""
        return self.key.decrypt(data, _RSA_PADDING)


def parse_public_key(data):
    """Parse a public key from the raw data, in SSH format."""
    key = serialization.load_ssh_public_key(data, backend=default_backend())
    return Crypto(key)


def parse_private_key(data):
    """Parse a private key from the raw data, in PEM format."""
    key = serialization.load_pem_private_key(
        data, password=None, backend=default_backend())
    return Crypto(key)


def _load_key(filepath, parser):
    """Generic loader from file.

    On errors ignore them and return None, as this function is targeted to
    be used from the config files.
    """
    try:
        with open(filepath, 'rb') as fh:
            raw = fh.read()
        key = parser(raw)
    except:
        key = None
    return key


def load_public_key(filepath):
    """Load a public key from file in SSH format."""
    return _load_key(filepath, parse_public_key)


def load_private_key(filepath):
    """Load a private key from file in PEM format."""
    return _load_key(filepath, parse_private_key)
