# Copyright 2010 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

import re
from django import forms
from django.utils.translation import ugettext_lazy as _

from identityprovider.widgets import (
    CommaSeparatedWidget,
    ROAwareTextInput,
    ROAwareSelect,
)

FIELD_REQUIRED = _(u'Field required')
INVALID_EMAIL = _(u'Invalid email')
ONE_TIME_PWD_ERROR = _('Please enter a 6-digit or 8-digit one-time password')

default_errors = {
    'required': FIELD_REQUIRED,
    'invalid': _(u'Invalid value'),
}
email_errors = {
    'required': FIELD_REQUIRED,
    # Translators: This is used on all email-related input forms.
    'invalid': INVALID_EMAIL,
}
default_email_placeholder = _('Your email address')
default_password_placeholder = _('Password')


class CommaSeparatedField(forms.MultipleChoiceField):
    widget = CommaSeparatedWidget

    def clean(self, value):
        return ','.join(super(CommaSeparatedField, self).clean(value))


class OATHPasswordField(forms.CharField):
    """A string of between 6 or 8 digits."""
    widget = forms.widgets.TextInput(attrs={
        'autocomplete': 'off',
        'autofocus': 'autofocus',
        'required': 'required',
        'type': 'tel',
    })
    SIX = re.compile('[0-9]{6}$')
    EIGHT = re.compile('[0-9]{8}$')

    def clean(self, value):
        """Validate otp and detect type"""
        # remove any whitespace from the string
        if value:
            value = value.strip().replace(' ', '')
        value = super(OATHPasswordField, self).clean(value)
        if self.SIX.match(value):
            return value
        elif self.EIGHT.match(value):
            return value
        raise forms.ValidationError(ONE_TIME_PWD_ERROR)


class PreferredEmailField(forms.ModelChoiceField):

    def __init__(self, *args, **kwargs):
        kwargs['widget'] = ROAwareSelect
        kwargs['error_messages'] = email_errors
        kwargs['empty_label'] = None
        super(PreferredEmailField, self).__init__(*args, **kwargs)

    def label_from_instance(self, obj):
        return obj.email


class EmailField(forms.EmailField):

    def __init__(self, *args, **kwargs):
        placeholder = kwargs.pop('placeholder', default_email_placeholder)
        attrs = {'class': 'textType', 'type': 'email'}
        # add placeholder, autofocus and email type
        attrs['placeholder'] = placeholder
        attrs['autofocus'] = 'autofocus'

        # after django 1.5, we can replace the following with an EmailInput
        kwargs['widget'] = forms.TextInput(attrs=attrs)
        kwargs['error_messages'] = email_errors
        super(EmailField, self).__init__(*args, **kwargs)


class PasswordField(forms.CharField):

    def __init__(self, *args, **kwargs):
        placeholder = kwargs.pop('placeholder', default_password_placeholder)
        attrs = {
            'class': 'disableAutoComplete textType',
            'autocomplete': 'off',
            'placeholder': placeholder,
        }

        kwargs['widget'] = forms.PasswordInput(attrs=attrs)
        kwargs['error_messages'] = default_errors
        super(PasswordField, self).__init__(*args, **kwargs)


class DisplayNameField(forms.CharField):

    def __init__(self, *args, **kwargs):
        placeholder = kwargs.pop('placeholder', _('Your name'))
        attrs = {'class': 'textType', 'placeholder': placeholder}

        kwargs['widget'] = ROAwareTextInput(attrs=attrs)
        kwargs['error_messages'] = default_errors
        super(DisplayNameField, self).__init__(*args, **kwargs)


class UsernameField(forms.CharField):

    def __init__(self, *args, **kwargs):
        placeholder = kwargs.pop('placeholder', _('Username'))
        attrs = {
            'class': 'textType',
            'placeholder': placeholder,
            }
        if kwargs.pop('disabled', False):
            attrs['disabled'] = 'disabled'

        kwargs['widget'] = ROAwareTextInput(attrs=attrs)
        kwargs['error_messages'] = default_errors
        super(UsernameField, self).__init__(*args, **kwargs)
