# Copyright 2010-2013 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

import logging

from django.contrib import auth
from django.template.loader import render_to_string
from django.utils.translation import ugettext_lazy as _

from identityprovider.models import Account, InvalidatedEmailAddress
from identityprovider.models.const import (
    AccountStatus,
    EmailStatus,
)
from identityprovider.validators import ValidationError, get_password_validator

LOGIN_INCORRECT = _('Incorrect email/password combination')
logger = logging.getLogger(__name__)


class AuthenticationError(Exception):
    pass


class AccountSuspended(AuthenticationError):
    pass


class AccountDeactivated(AuthenticationError):
    pass


class EmailInvalidated(AuthenticationError):
    pass


class PasswordPolicyError(AuthenticationError):
    def __init__(self, policy_message, account):
        super(PasswordPolicyError, self).__init__(policy_message)
        self.account = account


def authenticate_user(email, password, rpconfig=None, request=None):
    """Attempt to authenticate a user.

    Return the user's account on success, or throw AuthenticationErrors
    on failure.

    """
    account = auth.authenticate(username=email, password=password,
                                request=request)

    if account is None:
        # further specific error messages for the user
        account = Account.objects.get_by_email(email)
        if account is not None:
            email_obj = account.emailaddress_set.get(email__iexact=email)

            if not account.is_active:
                if account.status == AccountStatus.SUSPENDED:
                    raise AccountSuspended(
                        _('Your account has been suspended. Please contact '
                          'login support to re-enable it'))
                else:
                    raise AccountDeactivated(
                        _('Your account has been deactivated. To reactivate '
                          'it, please reset your password'))
        else:
            # check if this is an invalidated email address
            invalidated = InvalidatedEmailAddress.objects.filter(
                email__iexact=email)
            if invalidated.exists():
                raise EmailInvalidated(
                    _("This email address has been invalidated. "
                      "Please contact login support."))

        raise AuthenticationError(LOGIN_INCORRECT)

    if not isinstance(account, Account):
        raise AuthenticationError(LOGIN_INCORRECT)

    validate_policy = get_password_validator(account=account, request=request)
    try:
        validate_policy(password)
    except ValidationError as e:
        message = e.messages[0]
        raise PasswordPolicyError(message, account)

    # check if account needs a password reset
    if account.need_password_reset:
        message = render_to_string('common/password_leaked_warning.txt')
        raise PasswordPolicyError(message, account)

    if rpconfig is not None and not rpconfig.allow_unverified:
        email_obj = account.emailaddress_set.get(email__iexact=email)
        if email_obj.status == EmailStatus.NEW:
            raise AuthenticationError(LOGIN_INCORRECT)

    return account
