# Copyright 2015 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

from optparse import make_option

from django.core.management.base import BaseCommand
from django.db import transaction
from django.utils.timezone import now, timedelta

from identityprovider.models import AuthToken


class Command(BaseCommand):

    option_list = BaseCommand.option_list + (
        make_option('-a', '--older-than', dest='max_age', default=32,
                    action='store',
                    help='Cleanup records older than this number of days.'
                    'Defaults to 32 to delete tokens older than 1 month.'),
        make_option('--dry-run', dest='dry_run', default=False,
                    action='store_true', help='Do not delete for real, just '
                    'print what would have been done.'),
    )
    help = 'Clean Authentication Tokens older than specified number of days.'

    def handle(self, *args, **options):
        result = self.clean_old_authtokens(
            max_age=options['max_age'], dry_run=options['dry_run'])
        if result is not None:
            self.stdout.write(result)

    def clean_old_authtokens(self, max_age, dry_run):
        """Delete authtokens older than max_age (days).

        If dry_run is given, only a summary of total and matching tokens will
        be shown.
        """
        cutoff = now() - timedelta(days=int(max_age))
        old_tokens = AuthToken.objects.filter(date_created__lte=cutoff)
        self.stdout.write("Total Auth Tokens: %s" % (
            AuthToken.objects.all().count(), ))
        if dry_run:
            self.stdout.write(
                "%s tokens older than %s days would be deleted." % (
                    old_tokens.count(), max_age))
        if not dry_run:
            self.stdout.write("Deleting %s tokens older than %s days." % (
                old_tokens.count(), max_age))
            with transaction.atomic():
                # The transaction ensures the delete() call is blocking and
                # running two instances of this is safe.
                old_tokens.delete()
