# Copyright 2010 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

import time
from datetime import date, datetime, timedelta
from optparse import make_option

import openid.store.nonce
from django.conf import settings
from django.core.management.base import BaseCommand
from django.db import connection, transaction

from identityprovider.models import Account, EmailAddress


SESSION_SQL = """DELETE FROM django_session
WHERE session_key = ANY(SELECT session_key FROM django_session
                        WHERE expire_date < CURRENT_TIMESTAMP LIMIT %(limit)s
                        FOR UPDATE %(_nowait)s)"""
NONCES_SQL = """DELETE FROM openidnonce
WHERE timestamp = ANY(SELECT timestamp FROM openidnonce
                      WHERE timestamp < %(nonce_expire_stamp)s LIMIT %(limit)s
                      FOR UPDATE %(_nowait)s)"""
NO_ITEMS = """No items selected to clean up. Please select at least one of:

--sessions
--nonces
--testdata
"""


class Command(BaseCommand):

    option_list = BaseCommand.option_list + (
        make_option('-s', '--sessions', dest='sessions', default=False,
                    action='store_true', help='Cleanup sessions.'),
        make_option('-n', '--nonces', dest='nonces', default=False,
                    action='store_true', help='Cleanup nonces.'),
        make_option('-t', '--testdata', dest='testdata', default=False,
                    action='store_true', help='Cleanup test data.'),
        make_option('-l', '--limit', dest='limit', default=10000,
                    action='store',
                    help='Number of rows to process per batch.'),
        make_option('-d', '--date-created', dest='date_created',
                    default=None, action='store',
                    help='Cleanup records created before this date.'),
    )
    help = 'Clean unnecessary/stalled data from database.'

    def handle(self, *args, **options):
        limit = int(options['limit'])
        # Used for tests.
        # * _nowait specifies whether the select_for_update query should have
        #   NOWAIT added, so it immediately exceptions out if it can't get
        #   a lock (where default behavior is to block until the lock can be
        #   obtained).
        # * _wait_for_release should block (possibly until some mechanism
        #   releases its block), so we're guaranteed to hold the DB lock until
        #   this call exits. Should be a callable (like a thread Event's wait()
        #   method)
        # * _notify_in_critical_section should be called when we acquire the DB
        #   lock, so other processes which are waiting for us to have it can
        #   proceed
        # These options should ONLY be used when testing.
        _notify_in_critical_section = options.get(
            '_notify_critical_section', None)
        _wait_for_release = options.get('_wait_for_release', None)
        _nowait = "NOWAIT" if options.get('_nowait', False) else ""
        _nowait_query = "NOWAIT" if _nowait else ""
        nonce_expire_stamp = int(time.time()) - openid.store.nonce.SKEW
        test_email_pattern = settings.EMAIL_ADDRESS_PATTERN.replace(
            '+', '\+').replace('.', '\.') % "[^@]+"
        # The prefix is used to leverage a DB index to narrow down the
        # set of matching records.
        prefix_index = settings.EMAIL_ADDRESS_PATTERN.index("%s")
        test_email_prefix = settings.EMAIL_ADDRESS_PATTERN[:prefix_index]
        if options['date_created'] is None:
            date_created = date.today() + timedelta(days=1)
        else:
            parsed = datetime.strptime(options['date_created'], '%Y-%m-%d')
            date_created = parsed.date()
        query_subst = {
            'limit': limit,
            '_nowait': _nowait_query,
            'nonce_expire_stamp': nonce_expire_stamp
        }
        queries = {
            'sessions': SESSION_SQL % query_subst,
            'nonces': NONCES_SQL % query_subst,
        }
        verbosity = int(options['verbosity'])

        testdata = options.get('testdata')
        if testdata:
            self.clean_testdata(
                test_email_pattern, test_email_prefix,
                date_created, limit, verbosity,
                _nowait=_nowait,
                _notify_in_critical_section=_notify_in_critical_section,
                _wait_for_release=_wait_for_release)

        selected_queries = [query for query in queries
                            if options.get(query)]
        if not selected_queries and not testdata:
            self.stdout.write(NO_ITEMS)

        # These flags are used to call the testing-related methods only once.
        crit_notify = False
        already_waited = False
        for item in selected_queries:
            if verbosity >= 1:
                self.stdout.write("\nCleaning %s..." % item)
            with transaction.atomic():
                cursor = connection.cursor()
                cursor.execute(queries[item])
                if (callable(_notify_in_critical_section) and
                        not crit_notify):
                    _notify_in_critical_section()
                    crit_notify = True
                if callable(_wait_for_release) and not already_waited:
                    # Actually call the _wait_for_release callable, once it
                    # returns we're clear to continue execution
                    _wait_for_release()
                    already_waited = True
                while cursor.rowcount > 0:
                    if verbosity >= 2:
                        self.stdout.write(".")
                    cursor.execute(queries[item])

    def clean_testdata(self, email_pattern, email_prefix, date_created, limit,
                       verbosity=0, _nowait=False,
                       _notify_in_critical_section=None,
                       _wait_for_release=None):
        kwargs = {'email__iregex': email_pattern,
                  'email__istartswith': email_prefix,
                  'date_created__lt': date_created}

        if verbosity >= 1:
            self.stdout.write("\nCleaning accounts...\n")

        # These flags are used to call the testing-related methods only once.
        crit_notify = False
        already_waited = False
        while True:
            with transaction.atomic():
                # Lock emails for update since they'll be deleted
                email_ids = EmailAddress.objects.select_for_update(
                    nowait=_nowait).filter(**kwargs).values_list('pk')[:limit]
                # No explicit select_for_update as we'll simply delete the
                # queryset
                accounts = Account.objects.filter(
                    emailaddress__in=email_ids)
                if not accounts:
                    break

                if verbosity >= 2:
                    self.stdout.write("\tDeleting %d accounts..." % (
                        accounts.count(),))
                accounts.delete()
                if verbosity >= 2:
                    self.stdout.write('\t [OK]\n')
                if (callable(_notify_in_critical_section) and not crit_notify):
                    _notify_in_critical_section()
                    crit_notify = True
                if callable(_wait_for_release) and not already_waited:
                    # Actually call the _wait_for_release callable, once it
                    # returns we're clear to continue execution
                    _wait_for_release()
                    already_waited = True
