# Copyright 2013 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

import time
from collections import OrderedDict
from multiprocessing import Process
from optparse import make_option
from uuid import uuid4

import psutil
from django.contrib.auth.hashers import (
    BCryptPasswordHasher,
    PBKDF2PasswordHasher,
)
from django.core.management.base import BaseCommand

from identityprovider.auth import BCrypt13PasswordHasher
from identityprovider.utils import encrypt_launchpad_password


PASSWORD = str(uuid4())


class BCrypt9PasswordHasher(BCryptPasswordHasher):
    rounds = 9


class BCrypt8PasswordHasher(BCryptPasswordHasher):
    rounds = 8


bcrypt13_hasher = BCrypt13PasswordHasher()
bcrypt12_hasher = BCryptPasswordHasher()
bcrypt9_hasher = BCrypt9PasswordHasher()
bcrypt8_hasher = BCrypt8PasswordHasher()
pbkdf2_hasher = PBKDF2PasswordHasher()


def launchpad():
    encrypt_launchpad_password(PASSWORD)


def encrypt_with_django(hasher):
    salt = hasher.salt()
    hasher.encode(PASSWORD, salt)


def bcrypt13():
    encrypt_with_django(bcrypt13_hasher)


def bcrypt12():
    encrypt_with_django(bcrypt12_hasher)


def bcrypt9():
    encrypt_with_django(bcrypt9_hasher)


def bcrypt8():
    encrypt_with_django(bcrypt8_hasher)


def pbkdf2():
    encrypt_with_django(pbkdf2_hasher)


CYPHERS = OrderedDict({
    'bcrypt13': bcrypt13,
    'bcrypt12': bcrypt12,
    'bcrypt9': bcrypt9,
    'bcrypt8': bcrypt8,
    'launchpad': launchpad,
    'pbkdf2': pbkdf2,
})


def execute(cypher, iterations):
    cypher_function = CYPHERS[cypher]
    for i in range(iterations):
        cypher_function()


class Command(BaseCommand):

    option_list = BaseCommand.option_list + (
        make_option('-i', '--iterations', dest='iterations', default=1000,
                    help='Iterations to check and create hashed passwords.'),
        make_option('-t', '--time', dest='collection_time', default=0.1,
                    type='float',
                    help='Time between each result collection.'),
        make_option('-c', '--cypher', dest='cypher',
                    help='Cypher type (choose between: {})'.format(
                        CYPHERS.keys()),
                    choices=CYPHERS.keys()),
    )
    help = 'Launches a password cypher performance test.'

    results = OrderedDict()

    def handle(self, *args, **options):
        iterations = int(options['iterations'])
        cypher_name = options['cypher']
        collection_time = float(options['collection_time'])

        self.stdout.write('Checking for {} iterations, interval of {}'
                          .format(iterations, collection_time))

        if cypher_name is None:
            cyphers = sorted(CYPHERS.keys())
        else:
            cyphers = [cypher_name]

        for cypher in cyphers:
            self.stdout.write('Testing cypher {}'.format(cypher))
            self.test_cypher(cypher, iterations, collection_time)

        for cypher in cyphers:
            results = self.results[cypher]
            data = results['data']
            length = len(data)
            cpu = sum(datum['cpu'] for datum in data) / length
            rss_memory = sum(datum['rss_memory'] for datum in data) / length
            vms_memory = sum(datum['vms_memory'] for datum in data) / length
            elapsed_time = results['elapsed_time']

            self.stdout.write('Results for {}:'.format(cypher))
            self.stdout.write('\tAverage CPU: {}'.format(cpu))
            self.stdout.write('\tAverage RSS memory: {}'.format(rss_memory))
            self.stdout.write('\tAverage VMS memory: {}'.format(vms_memory))
            self.stdout.write('\tElapsed time: {}'.format(elapsed_time))

    def test_cypher(self, cypher, iterations, collection_time):
        data = []
        self.results[cypher] = {
            'data': data,
        }
        p = Process(target=execute, args=(cypher, iterations))
        p.start()
        process = psutil.Process(p.pid)
        start_time = time.time()
        while p.is_alive():
            mem_info = process.get_memory_info()
            data.append({
                'cpu': process.get_cpu_percent(),
                'rss_memory': mem_info[0],
                'vms_memory': mem_info[1],
            })
            time.sleep(collection_time)
        elapsed_time = time.time() - start_time
        self.results[cypher]['elapsed_time'] = elapsed_time
