import logging

from datetime import timedelta

from django.conf import settings
from django.core.management.base import BaseCommand
from django.db import transaction
from django.utils.timezone import now

from identityprovider import emailutils
from identityprovider.models import Account
from identityprovider.models.const import AccountStatus, EmailStatus


class Command(BaseCommand):

    account_action = 'suspend'
    account_status = AccountStatus.ACTIVE
    action = 'suspension'
    action_after_days = 'SUSPEND_UNVERIFIED_ACCOUNT_AFTER_DAYS'
    warn_before_days = 'WARN_SUSPEND_UNVERIFIED_ACCOUNT_BEFORE_DAYS'
    help = ('Notify owners of unverified accounts about upcoming suspension, '
            'and suspend unverified accounts already notified.')

    def _filter_by_date(self, queryset, threshold):
        # do not care about time of creation, just the day has to match
        result = queryset.filter(
            date_created__year=threshold.year,
            date_created__month=threshold.month,
            date_created__day=threshold.day)
        return result

    def _notify(self, accounts):
        for account in accounts:
            if account.preferredemail is None:
                logging.warning('Can not notify %r about %s since no '
                                'preferredemail is set.', account, self.action)
                continue

            days_of_warning = getattr(settings, self.warn_before_days)
            try:
                emailutils.send_action_required_warning(
                    account, days_of_warning, self.account_action)
            except:
                logging.exception('Error while notifying %s for %r:',
                                  self.action, account)
            else:
                logging.info('Notified %r about future account %s.',
                             account, self.action)

    def _do_action(self, accounts):
        for account in accounts:
            # copy the account before performing the action for future logging
            copied = Account.objects.get(id=account.id)
            days_of_warning = getattr(settings, self.warn_before_days)

            # need to cache the email before changing the account
            email = None
            if account.preferredemail is not None:
                email = account.preferredemail.email
            else:
                logging.warning('Can not notify %r about %s since no '
                                'preferredemail is set.', copied, self.action)
            try:
                getattr(account, self.account_action)()
            except:
                logging.exception('Error while applying %s to %r:',
                                  self.action, copied)
            else:
                logging.info('Account %s for %s succeeded.',
                             self.action, copied)
                if email is not None:
                    try:
                        emailutils.send_action_applied_notice(
                            email, copied.displayname,
                            days_of_warning, self.account_action)
                    except:
                        logging.exception(
                            'Error while notifying %s to %r:',
                            self.action, copied)

    def handle(self, *args, **options):
        # Used for tests.
        # * _nowait specifies whether the select_for_update query should have
        #   NOWAIT added, so it immediately exceptions out if it can't get
        #   a lock (where default behavior is to block until the lock can be
        #   obtained).
        # * _wait_for_release should block (possibly until some mechanism
        #   releases its block), so we're guaranteed to hold the DB lock until
        #   this call exits. Should be a callable (like a thread Event's wait()
        #   method)
        # * _notify_in_critical_section signals we  have the DB lock,
        #   so other processes which are waiting for us to have it can proceed
        #   (like a thread Event's set() method)
        # These options should ONLY be used when testing.
        _nowait = options.get('_nowait', False)
        _notify_in_critical_section = options.get('_notify_critical_section',
                                                  None)
        _wait_for_release = options.get('_wait_for_release', None)
        with transaction.atomic():
            # grab all active accounts with zero verified emails
            accounts = Account.objects.select_for_update(
                nowait=_nowait).filter(status=self.account_status)
            accounts = accounts.exclude(
                emailaddress__status__in=[EmailStatus.PREFERRED,
                                          EmailStatus.VALIDATED])
            right_now = now()
            do_action_days = getattr(settings, self.action_after_days)
            notify_days = do_action_days - getattr(settings,
                                                   self.warn_before_days)
            assert notify_days > 0

            # filter those that were created 'notify_days' ago for notification
            notification_threshold = right_now - timedelta(days=notify_days)
            need_notification = self._filter_by_date(accounts,
                                                     notification_threshold)
            self._notify(need_notification)

            # filter those that were created 'do_action_days' ago for action
            action_threshold = right_now - timedelta(days=do_action_days)
            need_action = self._filter_by_date(accounts, action_threshold)
            self._do_action(need_action)
            if callable(_notify_in_critical_section):
                _notify_in_critical_section()
            if callable(_wait_for_release):
                # Actually call the _wait_for_release callable, once it returns
                # we're clear to continue execution
                _wait_for_release()
