# Copyright 2010-2016 Canonical Ltd.  This software is licensed under
# the GNU Affero General Public License version 3 (see the file
# LICENSE).

from __future__ import absolute_import
from __future__ import unicode_literals

import logging

from django.conf import settings
from django.contrib.auth import get_backends
from django.contrib.auth.models import User, update_last_login, make_password
from django.contrib.auth.signals import user_logged_in
from django.core.exceptions import ValidationError
from django.core.urlresolvers import reverse
from django.db import models, IntegrityError
from django.utils.functional import cached_property
from django.utils.translation import ugettext_lazy as _
from django.utils.timezone import datetime, now, timedelta
from model_utils.managers import PassThroughManager
from oauth_backend.models import Consumer as V1Consumer, Token as V1Token

from identityprovider.models.const import (
    AccountCreationRationale,
    AccountStatus,
    EmailStatus,
    TokenScope,
)
from identityprovider.utils import (
    generate_openid_identifier,
    generate_random_string_len_50,
    generate_random_string_len_60,
    generate_random_string_len_70,
    get_object_or_none,
)

from identityprovider.validators import Errors, get_password_validator
from webservices.launchpad import get_lp_username

__all__ = [
    'Account',
    'AccountPassword',
    'DisplaynameField',
    'PasswordField',
    'LPOpenIdIdentifier',
    'Token',
]


logger = logging.getLogger(__name__)
# Disconnect default signal handler, which tries to update last_login
# field on the db level, but on the Account object this is a virtual
# property
user_logged_in.disconnect(update_last_login)


def update_account_last_login(**kwargs):
    account = kwargs['user']
    account.last_login = now()
    # there's no ned to save, as property handler does that

user_logged_in.connect(update_account_last_login)


class AccountQuerySet(models.query.QuerySet):

    def verified(self):
        return self.filter(
            emailaddress__status__in=(EmailStatus.VALIDATED,
                                      EmailStatus.PREFERRED)).distinct()


class AccountManager(PassThroughManager):
    """Extend the default manager.

    Add a method that takes care of creating an Account, and the necessary
    AccountPassword, EmailAddress objects.

    """

    def create_account(self, displayname, email_address, password,
                       creation_rationale=None, salt=None,
                       password_encrypted=False, email_validated=True,
                       openid_identifier=None, **kwargs):
        from identityprovider.models import EmailAddress

        errors = Errors()

        email_status = (
            EmailStatus.PREFERRED if email_validated else EmailStatus.NEW)
        email = EmailAddress(email=email_address, status=email_status)

        with errors.collect('email'):
            email.full_clean()

        # do validation before account (and related) creation/save
        if password is None:
            # https://docs.djangoproject.com/en/1.6/topics/auth/passwords/
            # django.contrib.auth.hashers.make_password
            # If password is None then a concatenation of
            # UNUSABLE_PASSWORD_PREFIX and a random string will be returned
            # which disallows logins. Additional random string reduces chances
            # of gaining access to staff or superuser accounts.
            password = make_password(None, salt=salt)
        elif not password_encrypted:
            # an encrypted password is assumed to be valid
            # this will raise a ValidationError if password does not
            # comply with policy
            with errors.collect('password'):
                # Use already validated email for password policy check
                validate_policy = get_password_validator(email=email_address)
                validate_policy(password)
                password = make_password(password, salt=salt)

        if not displayname:
            errors['displayname'] = ['Field required']

        if errors:
            raise ValidationError(errors)

        if creation_rationale is None:
            creation_rationale = (
                AccountCreationRationale.OWNER_CREATED_LAUNCHPAD)

        kwargs.update({
            'displayname': displayname,
            'creation_rationale': creation_rationale,
            'status': AccountStatus.ACTIVE,
        })

        if openid_identifier is not None:
            kwargs['openid_identifier'] = openid_identifier

        account = Account.objects.create(**kwargs)
        email.account = account
        email.save()
        password = AccountPassword.objects.create(
            account=account,
            password=password)
        return account

    def get_by_email(self, email):
        return get_object_or_none(self.select_related(),
                                  emailaddress__email__iexact=email)

    def active_by_openid(self, openid_identifier):
        return get_object_or_none(self, openid_identifier=openid_identifier,
                                  status=AccountStatus.ACTIVE)


class DisplaynameField(models.TextField):
    def __init__(self, null=False, **kwargs):
        super(DisplaynameField, self).__init__(null=null, **kwargs)


class PasswordField(models.TextField):
    def __init__(self, null=False, **kwargs):
        super(PasswordField, self).__init__(null=null, **kwargs)


class Account(models.Model):
    date_created = models.DateTimeField(default=now, editable=False)
    creation_rationale = models.IntegerField(
        choices=AccountCreationRationale._get_choices())
    # creation source could be seen as redundant, but since creation_rationale
    # only allows for an integer from a fixed list, we can't let API users to
    # set a custom string, and we want that so we can identify creation source.
    creation_source = models.TextField(blank=True, null=True)
    status = models.IntegerField(choices=AccountStatus._get_choices())
    date_status_set = models.DateTimeField(default=now)
    displayname = DisplaynameField()
    openid_identifier = models.TextField(default=generate_openid_identifier,
                                         unique=True)
    oauth_secret = models.TextField(
        default=generate_random_string_len_50, unique=True, null=True)
    status_comment = models.TextField(blank=True, null=True)
    old_openid_identifier = models.TextField(blank=True, null=True,
                                             db_index=True)
    preferredlanguage = models.TextField(blank=True, null=True)
    twofactor_required = models.BooleanField(default=False)
    twofactor_attempts = models.SmallIntegerField(default=0)
    warn_about_backup_device = models.BooleanField(default=True)

    # FIXME: once all accounts have been migrated to a True/False value,
    # replace with BooleanField
    need_password_reset = models.NullBooleanField()

    # The username data in the DB is replicated from LP with some latency
    # (usually seconds, sometimes more), so we set need_username_refresh
    # whenever we inform LP of a username change, so we know we have to check
    # with the LP API until the change replicates.
    need_username_refresh = models.NullBooleanField()

    objects = AccountManager.for_queryset_class(AccountQuerySet)()

    OWNS_OR_IN_ANY_TEAM = """SELECT team.id
    FROM lp_person AS person
    JOIN lp_openididentifier ON person.account = lp_openididentifier.account
    JOIN lp_person AS team ON team.name = ANY(%s)
    LEFT OUTER JOIN lp_teamparticipation ON lp_teamparticipation.team = team.id
    WHERE lp_openididentifier.identifier = %s
    AND (lp_teamparticipation.person = person.id
         OR team.teamowner = person.id);
    """

    class Meta:
        app_label = 'identityprovider'
        db_table = 'account'
        verbose_name = _('account')
        verbose_name_plural = _('accounts')
        permissions = (
            ('api_view_account_details',
             'Can view account details through API calls'),
            ('api_edit_account_details',
             'Can edit account details through API calls'),
        )

    def __unicode__(self):
        return self.displayname

    def get_absolute_url(self):
        return reverse('server-identity', args=[self.openid_identifier])

    @property
    def need_backup_device_warning(self):
        return self.warn_about_backup_device and self.devices.count() == 1

    @property
    def paper_devices_needing_renewal(self):
        devices = self.devices.filter(device_type='paper')
        size = settings.TWOFACTOR_PAPER_CODES
        for device in devices:
            used = device.counter % size
            if used > (size - settings.TWOFACTOR_PAPER_CODES_WARN_RENEWAL):
                yield device

    @property
    def is_verified(self):
        return self.verified_emails().count() > 0

    # Act as an oauth custom consumer, to be used with v2 tokens
    @property
    def key(self):
        return self.openid_identifier

    @property
    def secret(self):
        if not self.oauth_secret:
            # Migrate data as we go: need to reuse secrets for old Consumer
            # so OAuth signatures keep being valid.
            try:
                consumer = V1Consumer.objects.get(
                    user__username=self.openid_identifier)
            except V1Consumer.DoesNotExist:
                self.oauth_secret = generate_random_string_len_50()
            else:
                self.oauth_secret = consumer.secret
            self.save()

        return self.oauth_secret

    def emails(self, limit=None):
        result = self.emailaddress_set.order_by('-date_created')
        if limit is not None:
            result = result[:limit]
        else:
            result = result.all()
        return result

    def verified_emails(self):
        """Returns all verified e-mail addresses.

        Include preferred email, with preferred sort first.

        """
        email = self.emailaddress_set.verified()
        return email.order_by('-status', 'email')

    def unverified_emails(self):
        email = self.emailaddress_set.filter(status=EmailStatus.NEW)
        return email.order_by('email')

    def last_authenticated_sites(self, limit=10):
        sites = self.openidrpsummary_set.order_by('-date_last_used')
        return sites if limit is None else sites[:limit]

    def sites_with_active_sessions(self, age_hours=24):
        """
        Return list of sites on which user *may* have active sessions

        Current rule of thumb is to return all sites accessed within previous
        24 hours.
        """
        max_date_last_used = now() - timedelta(hours=age_hours)
        sites = self.openidrpsummary_set.filter(
            date_last_used__gte=max_date_last_used)
        return sites.order_by('-date_last_used')

    def set_deleted(self):
        # setting this status triggers _do_deleted
        self.status = AccountStatus.DELETED
        self.save()

    def _do_deleted(self):
        self.user.is_active = False
        self.user.save()

        self.authtoken_set.all().delete()
        self.emailaddress_set.delete()
        self.accountpassword.password = 'invalid'
        self.accountpassword.save()
        self.invalidate_oauth_tokens()

    def _get_preferredemail(self):
        email = getattr(self, '_preferredemail', None)

        if email is None or not email.is_verified:
            try:
                email = None
                account_emails = self.emailaddress_set.filter(
                    status=EmailStatus.PREFERRED)
                if account_emails.count() > 0:
                    email = account_emails[0]

                if not email:
                    # Try to determine a suitable address, and mark it
                    # as preferred.
                    emails = self.emailaddress_set.filter(
                        status=EmailStatus.VALIDATED)
                    if emails.count() > 0:
                        email = emails.order_by('date_created')[0]
                        email.status = EmailStatus.PREFERRED
                        email.save()
                        logger.info(
                            "Updating preferred email to %r for account %s",
                            email, self
                        )

                if not email:
                    # we have no validated email, so use the first NEW email
                    # but don't save it
                    emails = self.emailaddress_set.filter(
                        status=EmailStatus.NEW)
                    if emails.count() > 0:
                        email = emails.order_by('date_created')[0]

                self._preferredemail = email
            except:
                # no error in preferred email should raise, as this breaks
                # login completely
                logger.exception(
                    'Error in getting preferredemail for account %s:', self)
                self._preferredemail = None
        return self._preferredemail

    def _set_preferredemail(self, email):
        if not email.is_verified:
            raise ValidationError('Email must be verified')
        current = self.preferredemail
        if current is not None:
            current.status = EmailStatus.VALIDATED
            current.save()
            if current != email:
                # avoid circular import
                from identityprovider import emailutils
                emailutils.send_preferred_changed_notification(
                    current.email, email.email)
        email.status = EmailStatus.PREFERRED
        email.save()
        self._preferredemail = email

    preferredemail = property(_get_preferredemail, _set_preferredemail)

    @cached_property
    def person(self):
        # Importing it here to prevent cyclic import issue
        from identityprovider.models.person import Person
        try:
            return Person.objects.extra(
                tables=('lp_openididentifier',),
                where=('lp_openididentifier.account = lp_person.account',
                       'lp_openididentifier.identifier = %s'),
                params=(self.openid_identifier,),
            ).get()
        except Person.DoesNotExist:
            pass

    def person_in_any_team(self, teams):
        from identityprovider.models.person import Person
        team_names = [
            team.name if isinstance(team, Person) else team
            for team in teams]
        teams = Person.objects.raw(
            self.OWNS_OR_IN_ANY_TEAM,
            params=(team_names, self.openid_identifier))
        result = len(list(teams)) > 0
        return result

    def person_in_team(self, team):
        return self.person_in_any_team([team])

    @cached_property
    def person_name(self):
        db_username = self.person.name if self.person else None
        if not self.need_username_refresh:
            return db_username
        # The user has changed their username recently, so get the latest data
        # from LP.
        lp_username = get_lp_username(self.openid_identifier)
        if lp_username == db_username and not settings.READ_ONLY_MODE:
            # The new name has replicated to the SSO DB, so we can stop asking
            # LP every time.
            self.need_username_refresh = False
            self.save(update_fields=['need_username_refresh'])
        return lp_username

    @property
    def user(self):
        if not hasattr(self, '_user'):
            # When in read only mode we don't want to create users so we'll
            # return what exist or None if not there
            if settings.READ_ONLY_MODE:
                self._user = get_object_or_none(
                    User, username=self.openid_identifier)
            else:
                if self.preferredemail:
                    self._user, _ = User.objects.get_or_create(
                        username=self.openid_identifier,
                        defaults={'email': self.preferredemail.email})
                else:
                    return None

        return self._user

    @property
    def username(self):
        return self.user.username

    def _get_last_login(self):
        if self.user is not None:
            return self.user.last_login

    def _set_last_login(self, login_datetime):
        # we can live without last_login
        if settings.READ_ONLY_MODE:
            return

        if self.user is not None:
            self.user.last_login = login_datetime
            self.user.save()

    last_login = property(_get_last_login, _set_last_login)

    @property
    def can_reactivate(self):
        return (self.status == AccountStatus.DEACTIVATED or
                self.status == AccountStatus.NOACCOUNT)

    @property
    def can_reset_password(self):
        return (self.is_active or self.can_reactivate)

    @property
    def is_active(self):
        return (self.status == AccountStatus.ACTIVE)

    def is_authenticated(self):
        return True

    @property
    def is_staff(self):
        # Logging into the admin system with this breaks things
        return False

    @property
    def is_superuser(self):
        return False

    @property
    def encrypted_password(self):
        try:
            accountpassword = self.accountpassword
            password = accountpassword.password
        except AccountPassword.DoesNotExist:
            password = None
        return password

    def set_password(self, password, salt=None):
        validate_policy = get_password_validator(account=self)
        validate_policy(password)
        try:
            accountpassword = self.accountpassword
        except AccountPassword.DoesNotExist:
            accountpassword = AccountPassword(account=self, password='invalid')
        accountpassword.password = make_password(password, salt=salt)
        accountpassword.save()
        # clear the password reset flag
        self.need_password_reset = False
        self.save()

    def has_module_perms(self, app_label):
        if not self.is_active:
            return False
        for backend in get_backends():
            if hasattr(backend, "has_module_perms"):
                if backend.has_module_perms(self, app_label):
                    return True
        return False

    def has_perm(self, perm):
        if not self.is_active:
            return False
        for backend in get_backends():
            if hasattr(backend, "has_perm"):
                if backend.has_perm(self, perm):
                    return True
        return False

    def create_oauth_token(self, token_name, scope=TokenScope.VERSION_2):
        token, _ = self.get_or_create_oauth_token(token_name, scope=scope)
        return token

    def get_or_create_oauth_token(
            self, token_name, scope=TokenScope.VERSION_2):
        token = None
        if scope is None:
            consumer, _ = V1Consumer.objects.get_or_create(
                user=self.user, defaults={'secret': self.secret})
            token = consumer.token_set.filter(name=token_name).order_by(
                '-created_at').first()
            if token is not None:
                # if multiple tokens are present, keep the newest one
                token.updated_at = now()
                token.save()
                created = False
            else:
                token = consumer.token_set.create(name=token_name)
                created = True
        else:
            i = 1
            name = token_name
            while token is None:
                try:
                    token, created = Token.objects.get_or_create(
                        name=name, consumer=self, scope=scope)
                except IntegrityError:
                    name = token_name + str(i)
                i += 1

            if not created:
                token.date_updated = now()
                token.save()

        return token, created

    def oauth_tokens(self, limit=None, scope=TokenScope.VERSION_2):
        if not scope:
            result = V1Token.objects.filter(
                consumer__user__username=self.openid_identifier).order_by(
                '-updated_at')
        else:
            result = self.token_set.filter(scope=scope).order_by(
                '-date_updated')

        if limit is not None:
            result = result[:limit]
        else:
            result = result.all()

        return result

    def create_v1_oauth_token(self, token_name):
        return self.create_oauth_token(token_name, scope=None)

    def get_or_create_v1_oauth_token(self, token_name):
        return self.get_or_create_oauth_token(token_name, scope=None)

    def v1_oauth_tokens(self, limit=None):
        return self.oauth_tokens(limit=limit, scope=None)

    def oauth_token_by_key(self, token_key):
        token = self.token_set.filter(key=token_key).first()
        if not token:
            token = V1Token.objects.filter(
                consumer__user__username=self.openid_identifier,
                pk=token_key).first()
        return token

    def invalidate_oauth_tokens(self):
        tokens = V1Token.objects.filter(
            consumer__user__username=self.openid_identifier).all()
        tokens.delete()
        V1Consumer.objects.filter(
            user__username=self.openid_identifier).all().delete()
        self.token_set.all().delete()

    def must_require_twofactor(self):
        """Whether this account must mandatorily do twofactor.

        This is calculated from the twofactor_required flag and team
        memberships.

        """
        # Force twofactor_required for those accounts in specific teams
        result = self.person_in_any_team(settings.TWOFACTOR_MANDATORY_TEAMS)
        if result and not (
                self.twofactor_required and self.warn_about_backup_device):
            # Since twofactor will be mandatory, also make sure the warning
            # for the lack of backup device is shown
            self.twofactor_required = True
            self.warn_about_backup_device = True
            self.save(update_fields=['twofactor_required',
                                     'warn_about_backup_device'])

        return result

    def has_twofactor_devices(self):
        """Return if this Account has any two-factor devices."""
        return self.devices.exists()

    def increment_twofactor_attempts(self):
        """Increment the 2FA attempt counter."""
        self.twofactor_attempts = self.twofactor_attempts + 1
        self.save()

    def reset_twofactor_attempts(self):
        """Reset the 2FA attempt counter."""
        self.twofactor_attempts = 0
        self.save()

    def deactivate(self):
        self.status = AccountStatus.DEACTIVATED
        self.save()

    def suspend(self):
        self.status = AccountStatus.SUSPENDED
        self.save()

    def save(self, force_insert=False, force_update=False, **kwargs):
        if settings.READ_ONLY_MODE:
            return

        # if we suspend the account, reset the password too, to force the user
        # to reset it after the account is re-enabled
        # also, saving a new password will invalidate all oauth tokens
        if self.status == AccountStatus.SUSPENDED:
            try:
                # by setting a plain text value we make sure its not going to
                # match against anything
                self.accountpassword.password = 'invalid'
                self.accountpassword.save()
            except AccountPassword.DoesNotExist:
                # no password, so nothing to reset
                pass
        elif self.status == AccountStatus.DELETED:
            self._do_deleted()

        # update date_status_set only when the status value changes
        try:
            old_status = Account.objects.get(id=self.id).status
            if self.status != old_status:
                # update date_status_set
                self.date_status_set = now()
        except Account.DoesNotExist:
            pass

        # now go and effectively save it
        super(Account, self).save(force_insert, force_update, **kwargs)

    def reset_password(self, email=None, reason='password_leak'):
        if email is not None:
            account = Account.objects.get_by_email(email)
            if account != self:
                return False
        else:
            email = self.preferredemail.email

        self.need_password_reset = True
        self.save()
        from identityprovider import emailutils
        emailutils.send_password_reset_email(self, email, reason=reason)
        return True


class AccountPassword(models.Model):
    account = models.OneToOneField(Account, db_column=b'account')
    password = PasswordField()
    last_leak_check = models.DateTimeField(null=True)
    date_changed = models.DateTimeField(null=True, editable=False)

    LEAK_REFERENCE_DATE = datetime(2000, 01, 01, 0, 0, 0)

    class Meta:
        app_label = 'identityprovider'
        db_table = 'accountpassword'
        verbose_name = _('account password')
        verbose_name_plural = _('account passwords')

    def __unicode__(self):
        return _("Password for %s") % unicode(self.account)

    def save(self, *args, **kwargs):
        """Save the instance.

        On save, last_leak_check will be reset to far in the past.  This will
        cause a leak check the next time this password is used for login.  We
        want this whether the "account" or "password" fields are updated, since
        any of those changes make the password potentially leak-vulnerable.

        Usual "update_fields" behavior applies (i.e. last_leak_check will NOT
        be auto_updated if any fields are specified with update_fields).

        Also, we keep the last time the password was changed.
        """
        if 'last_leak_check' not in kwargs.get('update_fields', []):
            self.last_leak_check = AccountPassword.LEAK_REFERENCE_DATE
        self.date_changed = now()
        super(AccountPassword, self).save(*args, **kwargs)


class Token(models.Model):

    consumer = models.ForeignKey(Account)
    name = models.TextField()
    scope = models.TextField(
        choices=TokenScope._get_choices(), default=TokenScope.VERSION_2)
    key = models.TextField(
        primary_key=True, default=generate_random_string_len_60)
    secret = models.TextField(default=generate_random_string_len_70)
    date_created = models.DateTimeField(default=now)
    date_updated = models.DateTimeField(default=now)

    class Meta:
        app_label = 'identityprovider'
        unique_together = ('consumer', 'name')

    def __unicode__(self):
        return self.key

    @property
    def user(self):
        return self.consumer


class LPOpenIdIdentifier(models.Model):
    """Mapping between Launchpad accounts and OpenID identifiers."""

    identifier = models.TextField(unique=True, primary_key=True)
    lp_account = models.IntegerField(null=False, db_column=b'account',
                                     db_index=True)
    date_created = models.DateTimeField(null=False, default=now)

    class Meta:
        app_label = 'identityprovider'
        db_table = 'lp_openididentifier'
        verbose_name = _("LP OpenID Identifier")
        verbose_name_plural = _("LP OpenID Identifiers")

    def __unicode__(self):
        return _("LP OpenID Identifier for %s") % unicode(self.lp_account)
