# Copyright 2010 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

from __future__ import unicode_literals

import logging
import string

from hashlib import sha256

from django.core.urlresolvers import reverse
from django.core.validators import validate_email
from django.db import models
from django.http import Http404
from django.utils.timezone import now

from identityprovider.models.account import (
    Account,
    DisplaynameField,
    PasswordField,
)
from identityprovider.models.const import AuthTokenType
from identityprovider.utils import generate_random_string

__all__ = (
    'AUTHTOKEN_LENGTH',
    'AUTHTOKEN_PATTERN',
    'AuthToken',
    'verify_token_string'
)


# Vowels and "ambiguous" characters are not used for authtokens to avoid
# inadvertent curse words and confusion.
AUTHTOKEN_CHARS = "bcdfghjkmnpqrtvwxyzBCDFGHJKLMNPQRTVWXYZ2346789"
AUTHTOKEN_LENGTH = 20
HASH_LENGTH = 64
HASH_NAME = "sha256"

# For now, we have some older tokens (20 chars) still active, while we
# introduce the shorter tokens.  After a while, we could probably
# bound this pattern's repetition more tightly.
# AUTHTOKEN_PATTERN = '[A-Za-z0-9]{%s}' % AUTHTOKEN_LENGTH
AUTHTOKEN_PATTERN = '[A-Za-z0-9]+'

logger = logging.getLogger(__name__)

urls_mapping = {
    AuthTokenType.INVALIDATEEMAIL: 'invalidate_email',
    AuthTokenType.VALIDATEEMAIL: 'confirm_email',
    AuthTokenType.PASSWORDRECOVERY: 'reset_password',
}


class AuthTokenManager(models.Manager):

    valid_types = [
        AuthTokenType.INVALIDATEEMAIL,
        AuthTokenType.VALIDATEEMAIL,
        AuthTokenType.PASSWORDRECOVERY,
        AuthTokenType.VALIDATEGPG,
        AuthTokenType.VALIDATESIGNONLYGPG,
    ]

    def create(self, **kwargs):
        token_type = kwargs.get('token_type', -1)
        if token_type not in self.valid_types:
            raise ValueError(
                "token_type is not a valid AuthToken type: %s" % token_type
            )

        redirection_url = kwargs.get('redirection_url', None)
        if isinstance(redirection_url, str):
            kwargs['redirection_url'] = unicode(redirection_url)

        # Obtain a RAW token, this will be sent to the user.  If one is
        # given as part of the kwargs, we honor that, otherwise a new
        # random and unique token is generated.
        if 'raw_token' in kwargs:
            raw_token = kwargs['raw_token']
            del kwargs['raw_token']
        else:
            raw_token = create_unique_token_for_table()
        # we'll store the hash only, so a DB compromise doesn't yield
        # ready-to-use tokens (similar to storing hashed passwords)
        kwargs['hashed_token'] = AuthToken.hash_raw(raw_token)

        token = super(AuthTokenManager, self).create(**kwargs)
        token.full_clean()  # validate fields, including email
        # Add a non-db attribute, this stores the raw token while this
        # instance of AuthToken exists (so be sure to perform any
        # actions that need the raw token, such as sending out e-mail,
        # with the same instance).
        token.raw_token = raw_token
        return token


class AuthToken(models.Model):
    date_created = models.DateTimeField(
        default=now, blank=True, editable=False, db_index=True)
    date_consumed = models.DateTimeField(blank=True, null=True, db_index=True)
    token_type = models.IntegerField(choices=AuthTokenType._get_choices())
    hashed_token = models.TextField(unique=True, db_column=b'token')
    requester = models.ForeignKey(
        Account, db_column=b'requester', null=True, blank=True)
    requester_email = models.TextField(null=True, blank=True)
    email = models.TextField(db_index=True, validators=[validate_email])
    redirection_url = models.TextField(null=True, blank=True)

    displayname = DisplaynameField(null=True, blank=True)
    password = PasswordField(null=True, blank=True)
    fingerprint = models.CharField(max_length=60, blank=True)

    objects = AuthTokenManager()

    class Meta:
        app_label = 'identityprovider'
        db_table = u'authtoken'

    def __unicode__(self):
        try:
            return self.raw_token
        except AttributeError:
            return self.hashed_token

    def get_absolute_url(self):
        """Return absolute URL for this AuthToken.

        The absolute URL references the raw_token.

        :raises: AttributeError if there is no raw_token attribute.
        This can happen if the method gets called "out of scope", that is,
        against an AuthToken that was loaded from the database and whose
        raw_token has already been destroyed.
        """
        url_name = urls_mapping.get(self.token_type, None)
        # The next line will "implicitly" raise AttributeError if there's
        # no raw_token.
        kwargs = dict(authtoken=self.raw_token, email_address=self.email)
        if url_name is None:
            url_name = 'claim_token'
            kwargs = dict(authtoken=self.raw_token)

        return reverse(url_name, kwargs=kwargs)

    @classmethod
    def get_token(cls, **kwargs):
        """objects.get-like method to transparently handle hashing and old tokens.

        It's a wrapper which does two things:
           * Hashes incoming raw_token parameters to relieve callers from
             the responsibility and knowledge of doing this.
           * Augments the search to also look for "old-style" tokens where
             the "hashed_token" field is not hashed.
        """
        raw_token = hashed_token = None
        if 'raw_token' in kwargs:
            raw_token = kwargs['raw_token']
            hashed_token = cls.hash_raw(raw_token)
            del kwargs['raw_token']
        # kwargs now contains nothing about the tokens.
        query = models.Q(**kwargs)
        if raw_token:
            # Augment the query with the token conditions
            query = query & (models.Q(hashed_token=hashed_token) |
                             models.Q(hashed_token=raw_token))

        return cls.objects.get(query)

    @staticmethod
    def hash_raw(raw_token):
        """Hash a raw token returning the hash to store in the DB."""
        return sha256(raw_token).hexdigest()

    @property
    def active(self):
        return self.date_consumed is None

    @property
    def is_hashed(self):
        return (len(self.hashed_token) == HASH_LENGTH and
                all(c in string.hexdigits for c in self.hashed_token))

    @property
    def short_token(self):
        """Return a shortened representation for e.g. admin display.

        Hashed tokens are shortened to ~21 characters (depending on hash name
        length), while unhashed tokens are returned verbatim.
        """

        if self.is_hashed:
            return "(%s: %s...%s)" % (
                HASH_NAME, self.hashed_token[:4], self.hashed_token[-4:])
        else:
            return self.hashed_token

    def consume(self):
        self.date_consumed = now()
        self.save()

        result = AuthToken.objects.filter(
            email=self.email, token_type=self.token_type,
            requester=self.requester, date_consumed=None)
        right_now = now()
        for token in result:
            token.date_consumed = right_now
            token.save()


def create_unique_token_for_table(token_length=AUTHTOKEN_LENGTH,
                                  obj=AuthToken, column='hashed_token'):
    """Create a new unique token in a table.

    Generates a token and makes sure a hash for it does not already exist in
    the table and column specified.

    In order to handle "old-style" tokens, it's also verified that the newly
    created token does not directly exist in the same column.

    :param token_length: The length for the token string
    :param column: Database column where the token will be stored.

    :return: A new token string
    """
    raw_token = generate_random_string(token_length, charset=AUTHTOKEN_CHARS)
    hashed_token = AuthToken.hash_raw(raw_token)
    params_hashed = {column: hashed_token}
    params_raw = {column: raw_token}
    q = models.Q(**params_hashed) | models.Q(**params_raw)
    try:
        while obj.objects.filter(q).count() > 0:
            raw_token = generate_random_string(token_length,
                                               charset=AUTHTOKEN_CHARS)
            hashed_token = AuthToken.hash_raw(raw_token)
            params_hashed = {column: hashed_token}
            params_raw = {column: raw_token}
            q = models.Q(**params_hashed) | models.Q(**params_raw)
    except Exception:
        pass
    return raw_token


def verify_token_string(raw_token_string, email):
    """Verify and return a token by raw token string.

    Returns an unconsumed AuthToken object for the given email
    and raw_token (uses get_token so will correctly find both
    old- and new-style matching tokens).

    :return: An AuthToken, or None if no matching token was found.
    """
    try:
        return AuthToken.get_token(raw_token=raw_token_string, email=email,
                                   date_consumed=None)
    except AuthToken.DoesNotExist:
        return None


def get_authtoken_or_404(**kwargs):
    """Get tokens with get_token semantics but 404 if not found."""
    try:
        return AuthToken.get_token(**kwargs)
    except AuthToken.DoesNotExist:
        raise Http404("No AuthToken matches the given query")
