# Copyright 2010, 2012 Canonical Ltd.  This software is licensed under
# the GNU Affero General Public License version 3 (see the file
# LICENSE).

from __future__ import unicode_literals

import logging

from django.core.validators import validate_email
from django.db import models
from django.utils.timezone import now
from django.utils.translation import ugettext_lazy as _
from model_utils.managers import PassThroughManager

from identityprovider.models.account import Account
from identityprovider.models.person import Person
from identityprovider.models.const import EmailStatus

__all__ = (
    'EmailAddress', 'InvalidatedEmailAddress',
)


logger = logging.getLogger(__name__)


class EmailAddressQuerySet(models.query.QuerySet):
    def verified(self):
        return self.filter(
            status__in=(EmailStatus.VALIDATED, EmailStatus.PREFERRED))


class EmailAddressManager(PassThroughManager):
    """A custom manager for email address."""


class EmailAddress(models.Model):
    email = models.TextField(validators=[validate_email])
    lp_person = models.IntegerField(
        db_column=b'person', blank=True, null=True, editable=False)
    status = models.IntegerField(choices=EmailStatus._get_choices())
    date_created = models.DateTimeField(
        default=now, blank=True, editable=False)
    account = models.ForeignKey(
        Account, db_column=b'account', blank=True, null=True)

    objects = EmailAddressManager.for_queryset_class(EmailAddressQuerySet)()

    class Meta:
        app_label = 'identityprovider'
        db_table = u'emailaddress'
        verbose_name_plural = _('Email addresses')

    def __unicode__(self):
        return self.email

    @property
    def is_preferred(self):
        return self.status == EmailStatus.PREFERRED

    @property
    def is_verified(self):
        return self.status in (EmailStatus.VALIDATED, EmailStatus.PREFERRED)

    def validate(self):
        if self.status == EmailStatus.NEW:
            self.status = EmailStatus.VALIDATED
            self.save()

    def invalidate(self):
        account = self.account
        if account is None and self.lp_person:
            try:
                person = Person.objects.get(id=self.lp_person)
                account = person.account
            except Person.DoesNotExist:
                pass

        invalidated_email = None
        if account:
            # create invalidated entry
            invalidated_email = InvalidatedEmailAddress.objects.create(
                email=self.email, account=account,
                date_created=self.date_created)
        else:
            logger.warning(
                "Could not create invalidated entry for %s, no associated "
                "account found", self.email)
        # and delete from emails table
        self.delete()
        return invalidated_email


class InvalidatedEmailAddress(models.Model):
    email = models.TextField(validators=[validate_email])
    date_created = models.DateTimeField(blank=True, editable=False)
    date_invalidated = models.DateTimeField(default=now, null=True, blank=True)
    account = models.ForeignKey(
        Account, db_column=b'account', blank=True, null=True)
    account_notified = models.BooleanField(default=False)

    class Meta:
        app_label = 'identityprovider'
        db_table = u'invalidated_emailaddress'
        verbose_name_plural = _('Invalidated email addresses')

    def __unicode__(self):
        return self.email
