# Copyright 2010 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

from __future__ import unicode_literals

from django.db import models

from identityprovider.const import PERSON_VISIBILITY_PUBLIC
from identityprovider.models.account import Account, LPOpenIdIdentifier
from identityprovider.models.const import AccountCreationRationale

__all__ = (
    'Person',
    'TeamParticipation',
    'get_team_memberships_for_user',
)


class Person(models.Model):
    displayname = models.TextField(null=True, blank=True)
    teamowner = models.IntegerField(db_column=b'teamowner',
                                    null=True, blank=True)
    teamdescription = models.TextField(null=True, blank=True)
    name = models.TextField(unique=True, null=True)
    language = models.IntegerField(
        db_column=b'language', null=True, blank=True)
    fti = models.TextField(null=True)
    defaultmembershipperiod = models.IntegerField(null=True, blank=True)
    defaultrenewalperiod = models.IntegerField(null=True, blank=True)
    subscriptionpolicy = models.IntegerField(default=1, null=True)
    merged = models.IntegerField(db_column=b'merged',
                                 null=True, blank=True)
    datecreated = models.DateTimeField(auto_now_add=True, null=True)
    addressline1 = models.TextField(null=True, blank=True)
    addressline2 = models.TextField(null=True, blank=True)
    organization = models.TextField(null=True, blank=True)
    city = models.TextField(null=True, blank=True)
    province = models.TextField(null=True, blank=True)
    country = models.IntegerField(
        db_column=b'country', null=True, blank=True)
    postcode = models.TextField(null=True, blank=True)
    phone = models.TextField(null=True, blank=True)
    homepage_content = models.TextField(null=True, blank=True)
    icon = models.IntegerField(db_column=b'icon',
                               null=True, blank=True)
    mugshot = models.IntegerField(db_column=b'mugshot',
                                  null=True, blank=True)
    hide_email_addresses = models.NullBooleanField(default=False)
    creation_rationale = models.IntegerField(
        null=True, blank=True, choices=AccountCreationRationale._get_choices())
    creation_comment = models.TextField(null=True, blank=True)
    registrant = models.IntegerField(db_column=b'registrant',
                                     null=True, blank=True)
    logo = models.IntegerField(db_column=b'logo',
                               null=True, blank=True)
    renewal_policy = models.IntegerField(default=10, null=True)
    personal_standing = models.IntegerField(default=0, null=True)
    personal_standing_reason = models.TextField(null=True, blank=True)
    mail_resumption_date = models.DateField(null=True, blank=True)
    mailing_list_auto_subscribe_policy = models.IntegerField(default=1,
                                                             null=True)
    mailing_list_receive_duplicates = models.NullBooleanField(default=True)
    visibility = models.IntegerField(default=1, null=True)
    verbose_bugnotifications = models.NullBooleanField(default=False)
    lp_account = models.IntegerField(null=True, db_column=b'account',
                                     unique=True)

    class Meta:
        app_label = 'identityprovider'
        db_table = u'lp_person'

    def __unicode__(self):
        return self.displayname if self.displayname else self.name

    def in_team(self, team):
        # Just to be fully compatible with
        # lp:lib/registry/model/person.py:Person.inTeam()
        if isinstance(team, (str, unicode)):
            team_name = team
            team = None
        else:
            team_name = team.name

        # avoid a whole select and do a COUNT() only query
        result = TeamParticipation.objects.filter(
            team__name=team_name, person=self).exists()
        if not result:
            if team is not None:
                result = self.id == team.teamowner
            else:  # delay queries when only the team_name was provided
                result = Person.objects.filter(
                    name=team_name, teamowner=self.id).exists()

        return result

    @property
    def account(self):
        if self.lp_account is not None:
            open_ids = LPOpenIdIdentifier.objects.filter(
                lp_account=self.lp_account)
            if len(open_ids) > 0:
                openid_identifier = open_ids[0].identifier
                accounts = Account.objects.filter(
                    openid_identifier=openid_identifier)
                if len(accounts) > 0:
                    return accounts[0]

    def is_team(self):
        return self.teamowner is not None

    @property
    def time_zone(self):
        try:
            return self.personlocation.time_zone
        except:
            return None


class PersonLocation(models.Model):
    date_created = models.DateTimeField(null=True)
    person = models.OneToOneField(Person, db_column=b'person', null=True)
    latitude = models.FloatField(null=True, blank=True)
    longitude = models.FloatField(null=True, blank=True)
    time_zone = models.TextField(null=True, blank=True)
    last_modified_by = models.IntegerField(
        db_column=b'last_modified_by', null=True)
    date_last_modified = models.DateTimeField(null=True)
    visible = models.NullBooleanField(default=True)
    locked = models.NullBooleanField(default=False)

    class Meta:
        app_label = 'identityprovider'
        db_table = u'lp_personlocation'


class TeamParticipation(models.Model):
    team = models.ForeignKey(Person, db_column=b'team', null=True,
                             related_name='team_participations')
    person = models.ForeignKey(Person, db_column=b'person', null=True)

    class Meta:
        app_label = 'identityprovider'
        db_table = u'lp_teamparticipation'
        unique_together = ('team', 'person')

    def __unicode__(self):
        return "%s in %s" % (unicode(self.person), unicode(self.team))


def get_team_memberships_for_user(team_names, user, include_private=False):
    if user.person is None:
        # Someone that has no person associated with the account can't
        # be a member of any teams.
        return []
    teams = TeamParticipation.objects.filter(
        person=user.person,
        team__name__in=team_names,
    )
    if not include_private:
        teams = teams.filter(
            team__visibility=PERSON_VISIBILITY_PUBLIC)
    # By default it's a query set result, not a true list
    return list(teams.values_list('team__name', flat=True))
