# Copyright 2010-2012 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

from __future__ import unicode_literals

import logging
from datetime import timedelta

from django.db import models
from django.conf import settings
from django.utils.translation import ugettext_lazy as _
from django.utils.timezone import now
from gargoyle import gargoyle
from oath import accept_hotp, hotp

from identityprovider.models.account import Account


__all__ = ['AuthenticationDevice']
TWOFACTOR_LOGIN = 'two_factor_login'
TWOFACTOR_SYNC_ERROR = _("None of the devices could be synchronized "
                         "to the provided one-time passwords.")
logger = logging.getLogger(__name__)


def get_otp_type(otp):
    """Returns the format of otp, return None if not supported"""
    l = len(otp)
    if l == 6:
        return 'dec6'
    elif l == 8:
        return 'dec8'
    return None


def authenticate(account, otp):
    """Authenticates oath_token against the devices for an account."""
    for device in account.devices.order_by('-counter'):
        if device.authenticate(otp):
            return True
    # avoid circular import (identityprovider.login import
    # identityprovider.models which import from .twofactor import *
    from identityprovider.login import AuthenticationError
    raise AuthenticationError(_("The password is invalid"))


def sync(account, otps):
    """Tries to sync one of the devices with the provided otps."""
    for device in account.devices.order_by('-counter'):
        if device.sync(otps):
            return True
    # avoid circular import (identityprovider.login import
    # identityprovider.models which import from .twofactor import *
    from identityprovider.login import AuthenticationError
    raise AuthenticationError(TWOFACTOR_SYNC_ERROR)


def is_twofactor_enabled(request):
    if settings.READ_ONLY_MODE:
        return False
    else:
        return gargoyle.is_active('TWOFACTOR', request)


def login(request):
    request.session[TWOFACTOR_LOGIN] = now()


def logout(request):
    if TWOFACTOR_LOGIN in request.session:
        del request.session[TWOFACTOR_LOGIN]


def is_upgraded(request, must_be_fresh=False):
    """Return whether the user's session has been upgraded by a 2factor.

    If must_be_fresh is set, then the TWOFACTOR_LOGIN value has to be newer
    than settings.TWOFACTOR_FRESHNESS seconds.

    """
    login = request.session.get(TWOFACTOR_LOGIN)
    if login is None:
        result = False
    else:
        threshold = settings.TWOFACTOR_TTL
        if must_be_fresh:
            threshold = min(threshold, settings.TWOFACTOR_FRESHNESS)

        result = login + timedelta(seconds=threshold) >= now()
    return result


def is_authenticated(request):
    """Return whether the user is fully authenticated.

    According to his own 2nd-factor preference. This should be used as a
    replacement for request.user.is_authenticated() almost everywhere.
    """

    account = request.user

    # User is at least 1st-factor authenticated
    if not account.is_authenticated():
        return False

    # twofactor is enabled and account requires (or enforces) 2nd factor
    if user_requires_twofactor_auth(request, account):
        return is_upgraded(request)

    return True


def site_requires_twofactor_auth(request, token, rpconfig):
    if not is_twofactor_enabled(request):
        return False
    if rpconfig is not None and rpconfig.twofactor_required:
        return True
    # TODO: Check if site requested (through PAPE) 2f.
    return False


def user_requires_twofactor_auth(request, account):
    result = is_twofactor_enabled(request) and (
        # the account either explicitely set to always require 2FA or it's
        # part of a team that enforces it.
        account.twofactor_required or account.must_require_twofactor())
    return result


class AuthenticationDevice(models.Model):
    account = models.ForeignKey(Account, related_name='devices')
    key = models.TextField()
    name = models.TextField()
    counter = models.IntegerField(default=0)
    device_type = models.TextField(null=True)

    class Meta:
        app_label = 'identityprovider'
        # order devices by id by default
        ordering = ('id',)

    def _get_otp_type(self, otp):
        # avoid circular import (identityprovider.login import
        # identityprovider.models which import from .twofactor import *
        from identityprovider.login import AuthenticationError

        otp_type = get_otp_type(otp)
        if not otp_type:
            raise AuthenticationError(_("Invalid otp length, should be 6 or "
                                        "8 characters"))
        return otp_type

    def _sync_to_last_otp(self, otps, counters_to_test, otp_type):
        otps_length = len(otps)
        window_otps = [hotp(self.key, next_counter, otp_type)
                       for next_counter in counters_to_test]
        for counter in counters_to_test:
            slice_start = counter - self.counter
            slice_end = slice_start + otps_length
            next_counter = counter + otps_length
            next_otps = window_otps[slice_start:slice_end]
            if next_otps == otps:
                self.counter = next_counter
                self.save()
                return True

        return False

    def authenticate(self, otp):
        """Authenticate the code against the device"""
        otp_type = self._get_otp_type(otp)
        valid, new_counter = accept_hotp(
            self.key, otp, self.counter, otp_type, drift=settings.HOTP_DRIFT,
            backward_drift=settings.HOTP_BACKWARDS_DRIFT,
        )
        if valid:
            self.counter = new_counter
            self.save()
        return valid

    def sync(self, otps):
        """Synchronize the counter according to provided otps.

        Return False if it couldn't be synch'd."""
        for otp in otps:
            # Only the last otp_type is reused, but we get all of them to check
            # if the format is correct.
            otp_type = self._get_otp_type(otp)
        max_counter = self.counter + settings.HOTP_SYNC_WINDOW + len(otps) - 1
        counters_to_test = range(self.counter, max_counter + 1)
        return self._sync_to_last_otp(otps, counters_to_test, otp_type)
