# Copyright 2013 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

import itertools

from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.db.utils import DEFAULT_DB_ALIAS

__all__ = [
    'NOP_DB_ALIAS',
    'ReadOnlyModeRouter',
]

NOP_DB_ALIAS = 'nop'
NOP_DB_ENGINE = 'identityprovider.nop_backend'


class ReadOnlyModeRouter(object):
    """Database router which handles read-only mode"""

    def __init__(self):
        # Make sure that 'nop' database is present in the
        # configuration.
        if NOP_DB_ALIAS not in settings.DATABASES:
            raise ImproperlyConfigured(
                "Missing '{0}' database".format(NOP_DB_ALIAS))

        # And has right database engine.
        if settings.DATABASES[NOP_DB_ALIAS]['ENGINE'] != NOP_DB_ENGINE:
            msg = "'{0}' should have '{1}' engine".format(
                NOP_DB_ALIAS, NOP_DB_ENGINE)
            raise ImproperlyConfigured(msg)

        # If no slaves are configured, just read from default database
        # as normal.
        if settings.DATABASE_SLAVES:
            self.slaves = itertools.cycle(settings.DATABASE_SLAVES)
        else:
            self.slaves = itertools.cycle([DEFAULT_DB_ALIAS])

    def _model_label(self, model):
        return "{0}.{1}".format(
            model._meta.app_label, model._meta.object_name)

    def db_for_read(self, model, **hints):
        # Special case for session data, always read from master
        if self._model_label(model) == 'sessions.Session':
            return DEFAULT_DB_ALIAS

        return self.slaves.next()

    def db_for_write(self, model, **hints):
        if not settings.READ_ONLY_MODE:
            return DEFAULT_DB_ALIAS

        if self._model_label(model) in settings.WRITABLE_RO_MODE_MODELS:
            return DEFAULT_DB_ALIAS

        return NOP_DB_ALIAS

    def allow_relation(self, obj1, obj2, **hints):
        return True

    def allow_migrate(self, db, model):
        return db == DEFAULT_DB_ALIAS
