# Copyright 2010 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).
import logging
import traceback
import urllib2

from django.dispatch import Signal
from django.conf import settings
from django.contrib.auth.signals import user_logged_in
from django.core.cache import cache
from django.db.models.signals import post_save, pre_delete
from django.utils.timezone import now

from oauth.oauth import OAuthRequest

from identityprovider.const import SESSION_TOKEN_KEY, SESSION_TOKEN_NAME
from identityprovider.models import Account, AccountPassword, AuthLog, Token
from identityprovider.models.const import AuthLogType, TokenScope
from webservices.utils import http_request_with_timeout


logger = logging.getLogger(__name__)


account_created = Signal(providing_args=["openid_identifier"])
account_details_changed = Signal(providing_args=["openid_identifier"])
account_email_added = Signal(providing_args=["openid_identifier"])
account_email_validated = Signal(providing_args=["openid_identifier"])

application_token_created = Signal(providing_args=["openid_identifier"])
application_token_invalidated = Signal(providing_args=["openid_identifier"])

login_failed = Signal(providing_args=['credentials', 'request'])
# Django's user_logged_in signal only fires when creds are authenticated
# against Accounts, login_succeeded fires whenever a user is verified as
# being authorized, for instance, when SSO checks that it has an existing login
# cookie.
login_succeeded = Signal(providing_args=['user', 'request'])


def account_change_notify(sender, openid_identifier, **kwargs):
    update_url = getattr(settings, "SSO_ACCOUNT_UPDATE_URL")
    if update_url:
        try:
            http_request_with_timeout(update_url,
                                      {"openid_identifier": openid_identifier})
        except urllib2.URLError:
            pass


# Wire up all signals with notification function
account_created.connect(account_change_notify)
account_details_changed.connect(account_change_notify)
account_email_added.connect(account_change_notify)
account_email_validated.connect(account_change_notify)
application_token_created.connect(account_change_notify)
application_token_invalidated.connect(account_change_notify)


def invalidate_account_oauth_tokens(sender, instance, created, **kwargs):
    if not created:
        # invalidate oauth tokens on password change
        update_fields = kwargs.get('update_fields', None)
        # Cancel invalidation if only last_leak_check was changed
        # (it was the only field specified in update_fields).
        # If no update_fields was given or it contains anything in addition
        # to last_leak_check, then we still invalidate.
        force_invalidate = (
            not update_fields or
            'last_leak_check' not in update_fields or
            len(update_fields) != 1)

        if force_invalidate:
            instance.account.invalidate_oauth_tokens()


post_save.connect(
    invalidate_account_oauth_tokens, sender=AccountPassword,
    dispatch_uid='identityprovider.AccountPassword.post_save')


def log_login_succeeded(sender, user, request, **kwargs):
    """Log when an a user is authenticated.

    This can be either when login/password are requested/verified
    or when the user has an existing auth cookie (and thus bypasses
    login/password/2fa).
    """
    logtype = kwargs.get('authlogtype', AuthLogType.OPENID_NEW)
    email = kwargs.get('email', '')
    # user should always be an Account instance.
    log_entry = {"log_type": logtype,
                 "when_created": now(),
                 "login_email": email,
                 "account": user,
                 "referer": request.META.get("HTTP_REFERER", ""),
                 "remote_ip": request.META.get("REMOTE_ADDR", ""),
                 "user_agent": request.META.get("HTTP_USER_AGENT", "")}
    AuthLog.objects.create(**log_entry)

# connect to user login
login_succeeded.connect(
    log_login_succeeded,
    dispatch_uid='identityprovider.login_success_logging')


def log_login_failed(sender, credentials, request, **kwargs):
    logtype = kwargs.get('authlogtype', AuthLogType.OPENID_FAIL)
    log_entry = {"log_type": logtype,
                 "when_created": now(),
                 "login_email": credentials.get('email', ""),
                 "referer": request.META.get("HTTP_REFERER", ""),
                 "remote_ip": request.META.get("REMOTE_ADDR", ""),
                 "user_agent": request.META.get("HTTP_USER_AGENT", "")}
    AuthLog.objects.create(**log_entry)

# connect to login failure.
login_failed.connect(
    log_login_failed, dispatch_uid='identityprovider.login_failed_logging')


def set_session_oauth_token(sender, user, request, **kwargs):
    # user is an Account instance here

    headers = {'Authorization': request.META.get('HTTP_AUTHORIZATION', '')}
    orequest = OAuthRequest.from_request(request.method, '', headers=headers)

    if orequest and 'oauth_token' in orequest.parameters:
        # check for token in headers (handle login_by_token case)
        token_key = orequest.get_parameter('oauth_token')
    else:
        try:
            # check for already existent token
            token_key = user.token_set.get(scope=TokenScope.WEB_LOGIN).key
        except Token.DoesNotExist:
            # no token, this session will be invalidated when RO mode is off
            token_key = ''

        if not token_key and not settings.READ_ONLY_MODE:
            token_key = user.create_oauth_token(
                token_name=SESSION_TOKEN_NAME, scope=TokenScope.WEB_LOGIN).key

    request.session[SESSION_TOKEN_KEY] = token_key


# connect to user login when sender=Account
user_logged_in.connect(
    set_session_oauth_token, sender=Account,
    dispatch_uid='identityprovider.user_logged_in')


def track_failed_login(sender, credentials, request, **kwargs):
    ip = request.META.get('HTTP_X_FORWARDED_FOR', None)
    if ip is not None:
        ip = ip.split(', ')[0]
    else:
        ip = request.META.get('REMOTE_ADDR', '')

    key = "failed-login-%s" % ip
    attempts = cache.get(key, 0) + 1
    if attempts >= settings.MAX_FAILED_LOGIN_ATTEMPTS:
        # log and/or notify
        logger.warning('Too many failed login attempts. count=%s ip=%s',
                       attempts, ip)

    cache.set(key, attempts, settings.FAILED_LOGIN_WINDOW)


# connect to login failure.
login_failed.connect(
    track_failed_login,
    dispatch_uid='identityprovider.login_failed')


def track_isdtest_account_deletion(sender, instance, using, **kwargs):
    if instance.displayname == "ISD Test":
        logger.warning(
            "Deleted isd test account, %s",
            "".join(traceback.format_list(traceback.extract_stack()))
        )

pre_delete.connect(track_isdtest_account_deletion, sender=Account)
