import hashlib
from django.contrib.staticfiles.storage import (
    StaticFilesStorage,
)
from whitenoise.django import (
    HelpfulExceptionMixin,
    GzipStaticFilesMixin,
    # import from whitenoise to get backport if needed
    ManifestStaticFilesStorage,
)


# SCA customisation to ManifestStaticFilesStorage
class SSOManifestStaticFilesStorage(ManifestStaticFilesStorage):
    """Customise behaviour to ignore yui, and use long hashes"""

    def post_process(self, paths, *args, **kwargs):
        # don't post process yui files, as they have broken url() fragments in
        # the css, which breaks this parser. YUI will never change anyway, so
        # cache busting urls are not needed
        paths = dict(p for p in paths.items() if 'yui/' not in p[0])

        all_post_processed = super(SSOManifestStaticFilesStorage,
                                   self).post_process(paths, *args, **kwargs)
        for original, path, processed in all_post_processed:
            yield original, path, processed

    def url(self, url, force=False):
        if 'yui/' in url:
            # skip ManifestStaticFilesStorage handling all together, just want
            # default behaviour
            # This is not nice, but don't know better way
            return StaticFilesStorage.url(self, url)
        else:
            return super(SSOManifestStaticFilesStorage, self).url(url, force)

    def file_hash(self, clean_name, content=None):
        # override odd upstream decision to only use first 12 chars of md5
        if content is None:
            return None
        md5 = hashlib.md5()
        for chunk in content.chunks():
            md5.update(chunk)
        return md5.hexdigest()


class SSOGzipManifestStaticFilesStorage(
        HelpfulExceptionMixin, GzipStaticFilesMixin,
        SSOManifestStaticFilesStorage):
    pass
