from django.conf import settings
from django.core.cache import cache
from oauth_backend.models import Token as V1Token
from oauth.oauth import OAuthDataStore

from identityprovider.models import Account, Token


class SSODataStore(OAuthDataStore):
    """Custom data store for SSO.

    Ensure tokens are retrieved either for version 1 or version 2.

    """

    DEFAULT_NONCE_TIMEOUT = 3600  # one hour

    def __init__(self, oauth_request=None):
        """To serve as a Piston datastore we'll need provide this signature.

        We later won't use the oauth_request, so we can ignore it here.
        """
        super(SSODataStore, self).__init__()

    def lookup_token(self, token_type, token_field):
        """Return the Token object for the given token_field value.

        The value for token_type should always be 'access' as only such tokens
        are stored in database.

        """
        assert token_type == 'access'

        try:
            token = Token.objects.get(pk=token_field)
        except Token.DoesNotExist:
            token = V1Token.objects.filter(pk=token_field).first()

        return token

    def lookup_consumer(self, consumer_key):
        """Return the Consumer object for the given consumer_key value."""
        return Account.objects.active_by_openid(consumer_key)

    @property
    def _nonce_timeout_seconds(self):
        return getattr(
            settings, 'OAUTH_NONCE_TIMEOUT', self.DEFAULT_NONCE_TIMEOUT)

    def _nonce_key(self, consumer_key, token_key, nonce):
        return 'oauth-%s-%s-%s' % (consumer_key, token_key, nonce)

    def lookup_nonce(self, consumer, token, nonce):
        """Look up for the given nonce in the internal storage.

        If the nonce is found for the given consumer and key, return it.
        If not, record its usage (to prevent re-use) and return None.

        """
        key = self._nonce_key(consumer.key, token.pk, nonce)
        result = cache.get(key)
        if result is None:
            cache.set(key, nonce, timeout=self._nonce_timeout_seconds)

        return result
