from django.core.urlresolvers import reverse

from identityprovider.models.account import Account
from identityprovider.tests.helpers import FunctionalTestCase
from identityprovider.tests import DEFAULT_USER_PASSWORD


class StandaloneLoginTestCase(FunctionalTestCase):

    def test_login(self):
        """Logging into the base url without an OpenID request.

        We allow users accessing the base url directly (instead of
        being sent there from an OpenID RP) to login, so that they can
        see/change their details.

        """
        response = self.client.get(self.base_url)
        self.assertContains(
            response,
            '<span class="trusted-rp-name">Ubuntu One</span> '
            '<span class="action-title">log in</span>')
        self.assertIn("user_intention_create", response.content)

        response = self.login()
        # Once successfully logged in, they will see the details of their
        # account.
        self.assertContains(response, "Test Account")
        self.assertContains(response, self.default_email)

    def _do_register_account(self):
        """Creating a new account."""
        url = reverse('new_account')
        self.client.get(url)  # set cookie
        data = dict(
            displayname='New User',
            email=self.new_email,
            password=DEFAULT_USER_PASSWORD,
            passwordconfirm=DEFAULT_USER_PASSWORD,
            accept_tos=True,
        )
        return self.client.post(url, data=data, follow=True)

    def test_register(self):
        """Creating a new account."""
        response = self._do_register_account()
        self.assertContains(response, "Your account was created successfully")

        self.assertRedirects(response, reverse('account-index'))

        account = Account.objects.get_by_email(email=self.new_email)
        self.assertEqual(account.displayname, 'New User')
        self.assertTrue(account.person is None)

    def test_password_reset(self, reset_client=False):
        """Resetting the password."""
        # user should be registered with a confirmed email address
        self._do_register_account()
        self.client.logout()

        response = self.client.get('/+forgot_password', follow=True)
        self.assertContains(response, 'Step 1 of 3: Type your email address')

        response = self.client.post(
            '/+forgot_password', data=dict(email=self.new_email), follow=True)
        self.assertContains(response, 'Step 2 of 3: Check your email')

        link = self.get_reset_password_link_from_mail()
        response = self.client.get(link)
        token = self._get_raw_token_from_mail("resetpassword")
        link = reverse('reset_password',
                       kwargs=dict(authtoken=token,
                                   email_address=self.new_email))
        self.assertRedirects(response, link)

        if reset_client:
            # try to reset password out of session
            self.reset_client()

        response = self.client.get(link)
        self.assertContains(response, 'Step 3 of 3: Choose a new password')

        new_pwd = 'new Passw0rd'
        data = dict(password=new_pwd, passwordconfirm=new_pwd)
        response = self.client.post(link, data=data)
        self.assertRedirects(response, reverse('account-index'))

        # Log in with the new password.
        self.client.logout()
        success = self.client.login(username=self.new_email, password=new_pwd)
        self.assertTrue(success)
        response = self.client.get(reverse('account-index'))
        self.assertContains(response, "New User")

    def test_password_reset_out_of_session(self):
        self.test_password_reset(reset_client=True)

    def test_add_new_email(self):
        """Adding an e-mail address."""
        response = self.login()

        response = self.client.get(reverse('account-edit'))
        link = self.get_attribute_from_response(
            response,
            'a[data-qa-id="manage_email_addresses_link"]',
            'href')
        response = self.client.get(link)
        self.assertContains(response, 'Your email addresses')
        self.assertNotContains(response, self.new_email)
        self.assertContains(response, 'Add email address')

        data = dict(newemail=self.new_email)
        response = self.client.post(reverse('new_email'),
                                    data=data, follow=True)
        self.assertContains(response, 'Validate your email address')
        msg = ('just emailed %s (from noreply@ubuntu.com) with '
               'instructions on validating your email address.')
        self.assertContains(response, msg % self.new_email)

        link = self.get_newemail_link_from_mail()
        response = self.client.get(link)
        token = self._get_raw_token_from_mail("newemail")
        link = reverse('confirm_email',
                       kwargs=dict(authtoken=token,
                                   email_address=self.new_email))
        self.assertRedirects(response, link)

        response = self.client.get(link)
        self.assertContains(response, 'Validate %s?' % self.new_email)
        msg = 'Are you sure you want to confirm and validate this email'
        self.assertContains(response, msg)
        button = self.submit_from_response(response)
        self.assertIn('Yes, I\'m sure', button.text_content())

        # confirm new email
        response = self.client.post(link)

        self.client.logout()
        response = self.login(email=self.new_email)

        self.assertContains(response, "Test Account")
        self.assertContains(response, self.new_email)
