# Copyright 2010, 2012 Canonical Ltd.  This software is licensed under
# the GNU Affero General Public License version 3 (see the file
# LICENSE).

from django.contrib import admin
from django.contrib.admin.sites import (
    AdminSite,
    AlreadyRegistered,
    NotRegistered,
)
from django.conf import settings
from django.contrib.auth.models import User, check_password
from django.core.urlresolvers import reverse
from django.db import connections
from django.utils.dateformat import format
from django.utils.timezone import now
from mock import MagicMock, Mock, patch
from pyquery import PyQuery

from identityprovider.admin import (
    AccountPasswordInline,
    ApproxCountQuerySet,
    AuthenticationDeviceInline,
    AuthenticationDeviceInlineForm,
    EmailAddressAdmin,
    EmailAddressInline,
    InvalidatedEmailAddress,
    InvalidatedEmailAddressAdmin,
    LargeTableUserAdmin,
    OpenIDRPConfigForm,
    build_auth_logs_link,
    build_tokens_link,
    build_auth_tokens_link,
)
from identityprovider.models import (
    Account,
    AccountPassword,
    AuthLog,
    APIUser,
    AuthenticationDevice,
    EmailAddress,
    LPOpenIdIdentifier,
    OpenIDRPConfig,
    Person,
)
from identityprovider.models import authtoken
from identityprovider.models.const import (
    AccountStatus,
    AuthLogType,
    EmailStatus,
)
from identityprovider.tests.utils import SSOBaseTestCase


class AdminTestCase(SSOBaseTestCase):

    def setUp(self):
        super(AdminTestCase, self).setUp()
        self.admin = self.factory.make_account(
            is_superuser=True, password="admin007")
        assert self.client.login(
            username=self.admin.user.email, password="admin007")
        self.account = self.factory.make_account()

    def test_registered_models(self):
        for model in (Account, OpenIDRPConfig, APIUser, EmailAddress, Person):
            self.assertRaises(AlreadyRegistered, admin.site.register, model)

        for model in (AccountPassword, LPOpenIdIdentifier):
            self.assertRaises(NotRegistered, admin.site.unregister, model)

    def test_openidrpconfig_allowed_user_attribs_checkboxes_postable(self):
        trust_root = 'http://localhost/bla/'
        displayname = 'My Test RP'
        description = 'Bla'
        allowed_user_attribs = ['fullname', 'email']
        creation_rationale = 13

        data = {'trust_root': trust_root,
                'displayname': displayname,
                'description': description,
                'allowed_user_attribs': allowed_user_attribs,
                'creation_rationale': creation_rationale,
                }
        add_view = reverse('admin:identityprovider_openidrpconfig_add')
        response = self.client.get(add_view)
        response = self.client.post(add_view, data)
        self.assertEqual(302, response.status_code)
        # We don't get the ID back, so ensure we only have one entity and
        # assume it's the correct one.  This is racy, but the alternative is
        # another request to the list screen to scrape the ID from there.
        self.assertEqual(OpenIDRPConfig.objects.count(), 1)
        rpconfig = OpenIDRPConfig.objects.get()
        self.assertEqual(rpconfig.trust_root, trust_root)
        self.assertEqual(rpconfig.displayname, displayname)
        self.assertEqual(rpconfig.description, description)
        self.assertEqual(sorted(rpconfig.allowed_user_attribs.split(',')),
                         sorted(allowed_user_attribs))
        self.assertEqual(rpconfig.creation_rationale, creation_rationale)

    def test_openidrpconfig_allowed_user_attribs_checkboxes_getable(self):
        data = {'trust_root': 'http://localhost/bla/',
                'displayname': 'My Test RP',
                'description': 'Bla',
                'allowed_user_attribs': 'fullname',
                'creation_rationale': '13',
                }
        rpconfig = OpenIDRPConfig.objects.create(**data)
        change_view = reverse(
            'admin:identityprovider_openidrpconfig_change',
            args=(rpconfig.id,))
        response = self.client.get(change_view)
        dom = PyQuery(response.content)
        checked = dom.find('input[checked=checked]')
        self.assertEqual(len(checked), 1)
        self.assertEqual(checked[0].value, 'fullname')

    def test_inline_models(self):
        expected_inlines = [AccountPasswordInline, EmailAddressInline,
                            AuthenticationDeviceInline]
        registered_inlines = admin.site._registry[Account].inlines
        self.assertEqual(registered_inlines, expected_inlines)

    def test_account_overview(self):
        changelist_view = reverse('admin:identityprovider_account_changelist')
        r = self.client.get(changelist_view)
        self.assertNotContains(r, self.account.preferredemail)
        self.assertContains(r, build_tokens_link(self.account))
        self.assertContains(r, build_auth_tokens_link(self.account))
        self.assertContains(r, build_auth_logs_link("Account", self.account))

    def post_account_change(self, success=True, **kwargs):
        account_id = self.account.id
        account_password_id = self.account.accountpassword.id
        email = self.account.preferredemail
        change_view = reverse(
            'admin:identityprovider_account_change', args=(account_id,))
        parameters = {
            'creation_rationale': str(self.account.creation_rationale),
            'status': str(self.account.status),
            'displayname': self.account.displayname,
            'openid_identifier': self.account.openid_identifier,
            'accountpassword-TOTAL_FORMS': '1',
            'accountpassword-INITIAL_FORMS': '1',
            'accountpassword-0-id': str(account_password_id),
            'accountpassword-0-account': str(account_id),
            'accountpassword-0-last_leak_check_0': "01/01/2015",
            'accountpassword-0-last_leak_check_1': "00:00:00",
            'emailaddress_set-TOTAL_FORMS': '1',
            'emailaddress_set-INITIAL_FORMS': '1',
            'emailaddress_set-0-id': str(email.id),
            'emailaddress_set-0-account': str(account_id),
            'emailaddress_set-0-email': email,
            'emailaddress_set-0-status': str(email.status),
            'devices-TOTAL_FORMS': '0',
            'devices-INITIAL_FORMS': '0',
        }
        parameters.update(kwargs)
        response = self.client.post(change_view, parameters)

        if success:
            status_code = response.status_code
            # If view returns with 200 on POST, it means the form has errors.
            if status_code == 200:
                self.assertEqual(response.context['errors'], [])

            # Any non-error status code would be fine here, but right now
            # it redirects with a 302.
            self.assertEqual(
                status_code, 302,
                'Response did not redirect as expected, response is: %s' %
                response)

        return response

    def test_account_change(self):
        new_email = 'mark2@example.com'
        new_password = 'blah'
        parameters = {
            'emailaddress_set-0-email': new_email,
            'accountpassword-0-password': new_password,
        }
        self.post_account_change(**parameters)

        account = Account.objects.get(id=self.account.id)
        self.assertEqual(account.preferredemail.email, new_email)
        self.assertTrue(check_password(
            new_password, account.accountpassword.password))

    def test_accountpassword_change_resets_last_leak_check(self):
        # Force the last leak check to now
        last_leak_check = now()
        self.account.accountpassword.last_leak_check = last_leak_check
        self.account.accountpassword.save(update_fields=['last_leak_check'])
        assert self.account.accountpassword.last_leak_check == last_leak_check
        new_password = 'blah'
        parameters = {
            'accountpassword-0-password': new_password,
        }
        self.post_account_change(**parameters)
        account = Account.objects.get(id=self.account.id)
        # Now the last_leak_check should be reset to LEAK_REFERENCE_DATE
        self.assertEqual(account.accountpassword.last_leak_check,
                         AccountPassword.LEAK_REFERENCE_DATE)

    def test_account_delete(self):
        response = self.post_account_change(status=str(AccountStatus.DELETED))
        self.assertEqual(response.status_code, 302)
        account = Account.objects.get(id=self.account.id)
        self.assertEqual(account.status, AccountStatus.DELETED)
        self.assertEqual(account.emailaddress_set.count(), 0)

    def test_device_change(self):
        device = AuthenticationDevice.objects.create(
            account=self.account,
            key='some key',
            name='Some device',
            counter=124,
            device_type=None,
        )
        parameters = {
            'devices-TOTAL_FORMS': '1',
            'devices-INITIAL_FORMS': '1',
            'devices-0-id': str(device.id),
            'devices-0-key': 'some key',
            'devices-0-name': 'Some device',
            'devices-0-counter': '123',
            'devices-0-device_type': 'paper',
        }
        self.post_account_change(**parameters)

        device = AuthenticationDevice.objects.get(id=device.id)
        self.assertEqual(device.counter, 123)
        self.assertEqual(device.device_type, 'paper')

    def get_device_inlineform(self, device_type):
        device = AuthenticationDevice.objects.create(
            account=self.account,
            key='some key',
            name='Some device',
            counter=124,
            device_type=None,
        )
        post_data = {
            'key': 'other key',
            'name': 'A device',
            'counter': '123',
            'device_type': device_type,
        }
        form = AuthenticationDeviceInlineForm(post_data, instance=device)
        return form

    def test_device_inlineform_invalid_type(self):
        form = self.get_device_inlineform(device_type='foobar')
        device = AuthenticationDevice.objects.all()[0]
        expected_defaults = {
            'id': device.id,
            'key': 'some key',
            'name': 'Some device',
            'counter': 124,
            'device_type': None,
        }
        self.assertEqual(form.initial, expected_defaults)
        self.assertEqual(form.is_valid(), False)

        msg = ('Select a valid choice. foobar is not one of the '
               'available choices.')
        expected_errors = {'device_type': [msg]}
        self.assertEqual(form.errors, expected_errors)

    def test_device_inlineform_valid_type(self):
        form = self.get_device_inlineform(device_type='None')
        self.assertEqual(form.is_valid(), True)

    def test_multiple_preferred_emails(self):
        email = self.account.preferredemail
        parameters = {
            'emailaddress_set-TOTAL_FORMS': '2',
            'emailaddress_set-1-account': str(self.account.id),
            'emailaddress_set-1-email': 'failure@example.com',
            'emailaddress_set-1-status': str(EmailStatus.PREFERRED),
        }
        r = self.post_account_change(success=False, **parameters)

        account = Account.objects.get(pk=self.account.pk)
        self.assertContains(r, 'Only one email address can be preferred.')
        self.assertEqual(len(account.emailaddress_set.filter(
                         status=EmailStatus.PREFERRED)), 1)
        self.assertEqual(account.preferredemail, email)

    def test_inline_forms(self):
        change_view = reverse('admin:identityprovider_account_change',
                              args=(self.account.id,))
        r = self.client.get(change_view)
        self.assertEqual(r.status_code, 200)
        self.assertContains(r,
                            '<input type="password" '
                            'name="accountpassword-0-password" '
                            'id="id_accountpassword-0-password">',
                            html=True)

    def test_add_api_user(self):
        add_view = reverse('admin:identityprovider_apiuser_add')
        r = self.client.post(add_view, {
            'username': 'api_user',
            'password': 'backend'})
        self.assertEqual(r.status_code, 302)

    def test_email_overview(self):
        changelist_view = reverse(
            'admin:identityprovider_emailaddress_changelist')
        r = self.client.get(changelist_view)
        self.assertContains(r, build_tokens_link(self.account))
        self.assertContains(r, build_auth_tokens_link(self.account))
        self.assertContains(
            r, build_auth_logs_link("Email", self.account.preferredemail))


class RPAdminTestCase(SSOBaseTestCase):

    def setUp(self):
        super(RPAdminTestCase, self).setUp()

        OpenIDRPConfig.objects.create(trust_root="test", displayname="test",
                                      description="test")
        admin = self.factory.make_account(
            is_superuser=True, password="admin007")
        assert self.client.login(
            username=admin.user.email, password="admin007")

    def _get_form(self):
        path = reverse("admin:identityprovider_openidrpconfig_add")
        return self.client.get(path)

    def _get_table(self):
        path = reverse("admin:identityprovider_openidrpconfig_changelist")
        return self.client.get(path)

    def test_form_has_2f_checkbox(self):
        r = self._get_form()
        dom = PyQuery(r.content)
        el = dom.find('input[name="twofactor_required"]')[0]
        self.assertEqual(el.get('type'), 'checkbox')

    def test_table_has_2f_column(self):
        r = self._get_table()
        self.assertContains(r, "Twofactor required</a></div>")

    def test_table_2f_is_editable(self):
        r = self._get_table()
        dom = PyQuery(r.content)
        el = dom.find('input[name="form-0-twofactor_required"]')[0]
        self.assertEqual(el.get('type'), 'checkbox')

    def test_table_is_filterable_by_2f(self):
        r = self._get_table()
        self.assertContains(r, "<h3> By twofactor required </h3>")

    def test_trust_root_url_is_preserved(self):
        rpconfig = OpenIDRPConfig.objects.create(
            trust_root='http://example.org',
            displayname="test", description="test")

        data = {
            'displayname': rpconfig.displayname,
            'description': rpconfig.description,
            'trust_root': 'http://example.org?param=value',
            'creation_rationale': 13
        }
        form = OpenIDRPConfigForm(data=data, instance=rpconfig)
        self.assertTrue(form.is_valid())
        form.save()

        rpconfig = OpenIDRPConfig.objects.get(pk=rpconfig.pk)
        self.assertEqual(rpconfig.trust_root, 'http://example.org?param=value')


class EmailAddressAdminTestCase(SSOBaseTestCase):
    fixtures = ["test"]
    model = EmailAddress
    modeladmin = EmailAddressAdmin

    def setUp(self):
        super(EmailAddressAdminTestCase, self).setUp()
        self.admin = self.modeladmin(self.model, AdminSite())

    @patch('identityprovider.admin.reverse')
    def test_account_link(self, mock_reverse):
        mock_reverse.return_value = '/foo'
        xss_displayname = "Sample<script>alert('busted');</script>Person & Co."
        email = EmailAddress.objects.get(email='test@canonical.com')
        email.account.displayname = xss_displayname

        expected = (
            '<a href="/foo">Sample&lt;script&gt;alert(&#39;busted&#39;);&lt;'
            '/script&gt;Person &amp; Co.</a> (OpenID: %s)')
        openid = email.account.openid_identifier
        self.assertEqual(self.admin.account_link(email), expected % openid)


class InvalidatedEmailAddressAdminTestCase(EmailAddressAdminTestCase):

    model = InvalidatedEmailAddress
    modeladmin = InvalidatedEmailAddressAdmin


class ApproxCountQuerySetTestCase(SSOBaseTestCase):
    def setUp(self):
        super(ApproxCountQuerySetTestCase, self).setUp()
        # use User model here just to have a real model and avoid
        # lots of mocking
        self.qs = ApproxCountQuerySet(model=User)
        self.qs.root_query_set = Mock()

        mock_connections = self.patch('identityprovider.admin.connections',
                                      new=MagicMock(spec=connections))

        self.mock_cursor = mock_connections['default'].cursor()

    def test_count_is_cached(self):
        self.qs._result_cache = [Mock()]
        self.assertEqual(self.qs.count(), 1)
        self.assertFalse(self.mock_cursor.execute.called)

    def test_count_estimate(self):
        self.mock_cursor.fetchone.return_value = ['42']
        self.assertEqual(self.qs.count(), 42)

    def test_count_estimate_with_distinct(self):
        self.qs.query.distinct = True
        self.mock_cursor.fetchone.return_value = ['42']
        self.assertEqual(self.qs.count(), 42)

    def assert_count_real(self):
        with patch.object(self.qs.query, 'get_count') as mock_get_count:
            self.assertEqual(self.qs.count(), mock_get_count.return_value)

    def test_count_real_with_where(self):
        self.qs.query.where = Mock()
        self.assert_count_real()

    def test_count_real_wth_high_mark(self):
        self.qs.query.high_mark = 100
        self.assert_count_real()

    def test_count_real_with_low_mark(self):
        self.qs.query.low_mark = 23
        self.assert_count_real()

    def test_count_real_with_select(self):
        self.qs.query.select = Mock()
        self.assert_count_real()

    def test_count_real_with_grou_by(self):
        self.qs.query.group_by = Mock()
        self.assert_count_real()

    def test_count_real_with_having(self):
        self.qs.query.having = Mock()
        self.assert_count_real()


class LargeTableUserAdminTestCase(SSOBaseTestCase):

    def test_get_queryset(self):
        user_admin = LargeTableUserAdmin(User, AdminSite())
        name = 'django.contrib.admin.options.ModelAdmin.get_queryset'
        with patch(name) as mock_qs:
            user_admin.get_queryset(None)
        mock_qs.return_value._clone.assert_called_once_with(
            klass=ApproxCountQuerySet)

    def test_filtered_queryset(self):
        account = self.factory.make_account(
            is_superuser=True, password='12345678')
        admin_user = account.user

        assert admin_user.is_superuser
        assert self.client.login(
            username=admin_user.email, password='12345678')

        response = self.client.get(reverse('admin:auth_user_changelist'),
                                   {'q': admin_user.username[:2]})
        self.assertContains(response, admin_user.email)


class AuthTokenAdminTestCase(SSOBaseTestCase):

    def setUp(self):
        super(AuthTokenAdminTestCase, self).setUp()
        self.admin = self.factory.make_account(
            is_superuser=True, password="admin007")
        assert self.client.login(
            username=self.admin.user.email, password="admin007")
        self.account = self.factory.make_account()

    def test_search_hashes_tokens(self):
        """Searching for a raw token string should hash prior to querying."""
        test_token = self.factory.make_authtoken(
            email=self.account.preferredemail.email)
        response = self.client.get(
            reverse('admin:identityprovider_authtoken_changelist'),
            {'q': test_token.raw_token})
        self.assertContains(response, '1 auth token')
        self.assertContains(response, test_token.short_token)

    def test_search_finds_old_style_tokens(self):
        test_token = self.factory.make_authtoken(
            email=self.account.preferredemail.email)
        # Turn authtoken into an old-style token
        test_token.hashed_token = test_token.raw_token
        test_token.save()

        response = self.client.get(
            reverse('admin:identityprovider_authtoken_changelist'),
            {'q': test_token.raw_token})
        self.assertContains(response, test_token.email)
        self.assertContains(response, '1 auth token')
        self.assertNotContains(response, authtoken.HASH_NAME)


class AuthLogAdminTestCase(SSOBaseTestCase):

    def setUp(self):
        super(AuthLogAdminTestCase, self).setUp()
        self.admin = self.factory.make_account(
            is_superuser=True, password="admin007")
        assert self.client.login(
            username=self.admin.user.email, password="admin007")
        self.account = self.factory.make_account()

    def create_and_get_response_changelist(self, fields):
        """Creation of record and basic non-test-specific checks."""
        AuthLog.objects.create(**fields)
        view_url = reverse('admin:identityprovider_authlog_changelist')
        response = self.client.get(view_url)
        self.assertContains(response, "Authentication logs")
        self.assertContains(response, ">Account<")
        self.assertContains(response, ">Login email<")
        self.assertContains(response, '1.2.3.4')
        self.assertContains(
            response, format(
                fields['when_created'], settings.DATETIME_FORMAT))
        return response

    def test_successful_log_display(self):
        creation_time = now()
        fields = {
            'log_type': AuthLogType.OPENID_NEW,
            'when_created': creation_time,
            'account': self.account,
            'login_email': self.account.preferredemail.email,
            'referer': "http://www.example.com/referred_from",
            'remote_ip': '1.2.3.4',
            'user_agent': 'Godzilla 1.0',
        }
        response = self.create_and_get_response_changelist(fields)
        # A new OpenID login will contain both the account (displayname)
        # and the e-mail address used for logging in.
        self.assertContains(response, self.account.displayname)
        self.assertContains(response, self.account.preferredemail.email)
        self.assertContains(
            response, AuthLogType._label_for(AuthLogType.OPENID_NEW))
        # Should have a link to the account
        self.assertContains(
            response, reverse('admin:identityprovider_account_change',
                              args=(self.account.id,)))

    def test_successful_existing_log_display(self):
        creation_time = now()
        fields = {
            'log_type': AuthLogType.OPENID_EXISTING,
            'when_created': creation_time,
            'account': self.account,
            'referer': "http://www.example.com/referred_from",
            'remote_ip': '1.2.3.4',
            'user_agent': 'Godzilla 1.0',
        }
        response = self.create_and_get_response_changelist(fields)
        # An existing OpenID login only has the account link as we don't get
        # the raw credentials again (and fetching preferredemail is expensive
        # to do in the admin).
        self.assertContains(response, self.account.displayname)
        self.assertNotContains(response, self.account.preferredemail.email)
        self.assertContains(
            response, AuthLogType._label_for(AuthLogType.OPENID_EXISTING))
        # Should have a link to the account
        self.assertContains(
            response, reverse('admin:identityprovider_account_change',
                              args=(self.account.id,)))

    def test_failed_log_display(self):
        creation_time = now()
        fields = {
            'log_type': AuthLogType.OPENID_FAIL,
            'when_created': creation_time,
            'login_email': "failed@example.com",
            'referer': "http://www.example.com/referred_from",
            'remote_ip': '1.2.3.4',
            'user_agent': 'Godzilla 1.0',
        }
        response = self.create_and_get_response_changelist(fields)
        # Email instead of displayname, as we have no account
        self.assertContains(response, "failed@example.com")
        self.assertContains(
            response, AuthLogType._label_for(AuthLogType.OPENID_FAIL))
        # Should NOT have a link to the account
        self.assertNotContains(
            response, reverse('admin:identityprovider_account_change',
                              args=(self.account.id,)))

    def create_and_get_response_detail(self, fields):
        """Creation of record and basic non-test-specific checks - detail.

        This tests the detailed ("change") view of a record.
        """
        log_item = AuthLog.objects.create(**fields)
        view_url = reverse(
            'admin:identityprovider_authlog_change', args=(log_item.id,))
        response = self.client.get(view_url)
        self.assertContains(response, ">Account:<")
        self.assertContains(response, ">Login email:<")
        self.assertContains(response, '1.2.3.4')
        self.assertContains(
            response, format(
                fields['when_created'], settings.DATETIME_FORMAT))
        return response

    def test_detailed_succeeded_log_display(self):
        creation_time = now()
        fields = {
            'account': self.account,
            'log_type': AuthLogType.OPENID_NEW,
            'when_created': creation_time,
            'login_email': "ok@example.com",
            'referer': "http://www.example.com/referred_from",
            'remote_ip': '1.2.3.4',
            'user_agent': 'Godzilla 1.0',
        }
        response = self.create_and_get_response_detail(fields)
        self.assertContains(response, "ok@example.com")
        self.assertContains(
            response, AuthLogType._label_for(AuthLogType.OPENID_NEW))
        # Should NOT have a link to the account
        self.assertNotContains(
            response, reverse('admin:identityprovider_account_change',
                              args=(self.account.id,)))
        self.assertContains(response, self.account.displayname)

    def test_detailed_failed_log_display(self):
        creation_time = now()
        fields = {
            'log_type': AuthLogType.OPENID_FAIL,
            'when_created': creation_time,
            'login_email': "fail@example.com",
            'referer': "http://www.example.com/referred_from",
            'remote_ip': '1.2.3.4',
            'user_agent': 'Godzilla 1.0',
        }
        response = self.create_and_get_response_detail(fields)
        self.assertContains(
            response, format(
                fields['when_created'], settings.DATETIME_FORMAT))
        self.assertContains(response, "fail@example.com")
        self.assertContains(
            response, AuthLogType._label_for(AuthLogType.OPENID_FAIL))
        # Should NOT have a link to the account
        self.assertNotContains(
            response, reverse('admin:identityprovider_account_change',
                              args=(self.account.id,)))
        self.assertNotContains(response, self.account.displayname)
