import json
from contextlib import contextmanager

import requests
from django.conf import settings
from django.conf.urls import patterns, url
from django.http import HttpResponse
from django.views.decorators.csrf import csrf_exempt
from timeline import Timeline

from identityprovider.apiutils import (
    get_api_client,
    DjangoWSGIAdapter,
    TimelineRecordingApiClient,
    TimelineRecordingApiSession,
)
from identityprovider.tests.utils import SSOBaseTestCase


class GetApiClientTestCase(SSOBaseTestCase):

    def setUp(self):
        super(GetApiClientTestCase, self).setUp()
        self.mock_request = self.patch('requests.Session.request')

    def assert_django_adapters(self, api, request):
        http = api.session.adapters['http://']
        self.assertIsInstance(http, DjangoWSGIAdapter)
        self.assertEqual(http.parent_request, request)
        https = api.session.adapters['https://']
        self.assertIsInstance(https, DjangoWSGIAdapter)
        self.assertEqual(https.parent_request, request)

    def test_api_client_uses_internal_with_test_client(self):
        request = self.factory.make_request('/', method='POST')
        api = get_api_client(request)
        self.assert_django_adapters(api, request)

    def test_returns_timeline_recording_client(self):
        request = self.factory.make_request('/', method='POST')
        api = get_api_client(request)
        self.assertIsInstance(api, TimelineRecordingApiClient)

    def test_makes_timer_function_from_request(self):
        timeline = Timeline()
        request = self.factory.make_request(
            '/', method='POST', **{'timeline.timeline': timeline})
        api = get_api_client(request)
        api.session.post('/accounts')
        # 2 because we use allow_nested, as the request may be routed
        # internally if API_USE_INTERNAL is set.
        self.assertEqual(2, len(timeline.actions))
        self.assertEqual('ssoclient-POST-start', timeline.actions[0].category)
        self.assertEqual('/accounts', timeline.actions[0].detail)

    def assert_http_host_set(self, headers, client_host):
        request = self.factory.make_request('/foo/bar', **headers)

        assert settings.API_HOST != client_host

        path = '/zaraza'
        local_api_url = settings.API_HOST + settings.API_URL
        for method in ('GET', 'POST', 'PATCH', 'DELETE'):
            api = get_api_client(request)
            api.session.request(method, path)

            self.mock_request.assert_called_once_with(
                method, local_api_url + path, headers={'Host': client_host})
            self.mock_request.reset_mock()

    def test_passes_correct_http_host_using_header_value(self):
        server_name = 'super-odd-server-name'
        port = 6666
        client_host = '%s:%s' % (server_name, port)

        headers = {
            'HTTP_HOST': client_host,
            'SERVER_NAME': 'other-name',
            'SERVER_PORT': 7777}
        self.assert_http_host_set(headers, client_host)

    def test_http_host_fallback(self):
        server_name = 'super-odd-server-name'
        port = 6666
        client_host = '%s:%s' % (server_name, port)
        headers = {'SERVER_NAME': server_name, 'SERVER_PORT': port}
        self.assert_http_host_set(headers, client_host)


class TimelineRecordingApiClientTestCase(SSOBaseTestCase):

    def test_session_is_timeline_recording(self):
        api = TimelineRecordingApiClient('localhost')
        self.assertIsInstance(api.session, TimelineRecordingApiSession)

    def test_passes_timer_to_session(self):
        timer = object()
        api = TimelineRecordingApiClient('localhost', timer=timer)
        self.assertEqual(timer, api.session.timer)

    def test_passes_http_host_to_session(self):
        http_host = object()
        api = TimelineRecordingApiClient('localhost', http_host=http_host)
        self.assertEqual(http_host, api.session.http_host)


class TimelineRecordingApiSessionTestCase(SSOBaseTestCase):

    def setUp(self):
        super(TimelineRecordingApiSessionTestCase, self).setUp()
        self.mock_request = self.patch('requests.Session.request')

    def test_uses_timer(self):
        expected_response = object()
        self.mock_request.return_value = expected_response
        called_args = []

        @contextmanager
        def timer_fn(category, detail, allow_nested=False):
            called_args.append((category, detail, allow_nested))
            yield

        endpoint = 'http://localhost/'
        session = TimelineRecordingApiSession(endpoint, timer=timer_fn)
        path = 'test'
        response = session.request('GET', path)
        self.assertEqual(expected_response, response)
        self.assertEqual([('ssoclient-GET', path, True)], called_args)

    def test_http_host(self):
        host = 'http://example.com'
        endpoint = 'http://localhost'
        session = TimelineRecordingApiSession(endpoint, http_host=host)
        path = '/foo'

        for method in ('GET', 'POST', 'PATCH', 'DELETE'):
            session.request(method, path)
            self.mock_request.assert_called_once_with(
                method, endpoint + path, headers={'Host': host})
            self.mock_request.reset_mock()


@csrf_exempt
def testview(request):
    """Serializes the request data back as a json response, for inspection"""
    data = {
        'GET': request.GET,
        'COOKIES': request.COOKIES,
        'META': dict((k, v) for k, v in request.META.items()
                     if k not in ('wsgi.input', 'wsgi.errors'))
    }
    if request.META.get('HTTP_CONTENT_TYPE') == 'application/json':
        data['POST'] = json.loads(request.read())
    else:
        data['POST'] = request.POST
    return HttpResponse(
        content=json.dumps(data),
        content_type='application/json'
    )

urlpatterns = patterns(
    '',
    url(r'^test', testview),
    url(r'^set_language$',
        lambda r: r,
        name='set_language')
)


class DjangoWSGIAdapterTestCase(SSOBaseTestCase):

    urls = 'identityprovider.tests.test_apiutils'
    factory_headers = {
        'SERVER_NAME': "apitest",
        'SERVER_PORT': "667",
        'wsgi.url_scheme': 'https',
    }

    def assert_meta(self, meta):
        self.assertEqual(meta['REMOTE_ADDR'], '127.0.0.1')
        self.assertEqual(meta['SERVER_NAME'], 'apitest')
        self.assertEqual(meta['SERVER_PORT'], '667')
        self.assertEqual(meta['wsgi.url_scheme'], 'https')

    def setUp(self):
        super(DjangoWSGIAdapterTestCase, self).setUp()
        self.session = requests.session()
        self.original_request = self.factory.make_request(
            '/', method='POST', **self.factory_headers)
        self.session.mount('http://', DjangoWSGIAdapter(self.original_request))

    def test_django_wsgi_adaptor_get(self):
        response = self.session.get(
            'http://abc/test?a=1',
            headers={
                'Accept': 'application/json',
                'X-Custom-Header': 'Hi',
            },
            cookies={'test': 'cookie'}
        )
        data = response.json()
        get = data['GET']
        self.assertEqual(get['a'], '1')
        self.assertEqual(data['POST'], {})
        self.assert_meta(data['META'])
        self.assertEqual(data['META']['HTTP_X_CUSTOM_HEADER'], 'Hi')
        self.assertEqual(data['COOKIES']['test'], 'cookie')

    def test_django_wsgi_adaptor_post_json(self):
        response = self.session.post(
            'http://abc/test',
            data=json.dumps({'a': 1}),
            headers={
                'Accept': 'application/json',
                'X-Custom-Header': 'Hi',
                'Content-Type': 'application/json'
            },
            cookies={'test': 'cookie'}
        )
        data = response.json()
        self.assertEqual(data['GET'], {})
        self.assertEqual(data['POST']['a'], 1)
        self.assert_meta(data['META'])
        self.assertEqual(data['META']['HTTP_CONTENT_TYPE'], 'application/json')
        self.assertEqual(data['META']['CONTENT_TYPE'], 'application/json')
        self.assertEqual(data['META']['HTTP_X_CUSTOM_HEADER'], 'Hi')
        self.assertEqual(data['COOKIES']['test'], 'cookie')
