# encoding: utf-8
# Copyright 2010-2016 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

import base64
import binascii
import json
import os

from contextlib import contextmanager
from datetime import datetime, timedelta
from uuid import uuid4

from django.conf import settings
from django.contrib.auth.hashers import get_hasher
from django.contrib.auth.models import (
    check_password,
    is_password_usable,
    make_password,
)
from django.core.exceptions import ValidationError
from django.db import connection
from django.test.utils import override_settings
from django.utils.timezone import now
from mock import call, Mock, patch
from pymacaroons import Verifier

from identityprovider.auth import (
    LaunchpadBackend,
    SSOOAuthAuthentication,
    SSORequestValidator,
    _decrypt_caveat,
    basic_authenticate,
    build_discharge_macaroon,
    build_discharge_macaroon_from_root,
    refresh_discharge,
    refresh_macaroons,
    validate_oauth_signature,
)
from identityprovider.login import AuthenticationError, AccountDeactivated
from identityprovider.models import (
    Account,
    AccountPassword,
    AuthLog,
    EmailAddress,
)
from identityprovider.models.const import (
    AccountStatus,
    EmailStatus,
    TokenScope,
)
from identityprovider.readonly import ReadOnlyManager
from identityprovider.tests import DEFAULT_API_PASSWORD, DEFAULT_USER_PASSWORD
from identityprovider.tests.utils import SSOBaseTestCase
from identityprovider.utils import (
    encrypt_launchpad_password,
    get_cypher_migration_status,
    validate_launchpad_password,
)


class AuthLogTestCaseMixin(object):

    @contextmanager
    def assert_authlog_matches(self, expected_values):
        """Verify that AuthLog records match expected values.

        Use this around a block of code that will create an AuthLog record. The
        record will be checked as follows:

        * Verify that a single record exists in AuthLog.
        * Verify that the record matches expected_values. Keys present in the
        expected_values dict will be checked against fields with the same
        name for the expected values.
        * Verify that the when_created attribute of the AuthLog record
        is in the expected date range for when the record was created.
        """
        before = now()
        yield
        after = now()
        self.assertEqual(AuthLog.objects.all().count(), 1)
        log_item = AuthLog.objects.all()[0]
        self.assertGreaterEqual(log_item.when_created, before)
        self.assertLessEqual(log_item.when_created, after)
        self.assertIn(expected_values, AuthLog.objects.all().values(
            *expected_values.keys()))


class LaunchpadBackendTestCase(SSOBaseTestCase):

    email = 'mark@example.com'

    def setUp(self):
        super(LaunchpadBackendTestCase, self).setUp()
        self.account = self.factory.make_account(
            email=self.email, password=DEFAULT_USER_PASSWORD)
        self.backend = LaunchpadBackend()

    def assert_datetime_in_range(self, to_check, before, after):
        """Helper to check that a datetime is within range (before, after)."""
        self.assertGreaterEqual(to_check, before)
        self.assertGreaterEqual(after, to_check)

    def test_authenticate_with_email_status_not_in_expected_one(self):
        email_address = EmailAddress.objects.get(
            email__iexact=self.email)
        email_address.status = 9999
        email_address.save()

        result = self.backend.authenticate(self.email,
                                           DEFAULT_USER_PASSWORD)

        self.assertIsNone(result)

    def test_get_user_does_not_exist(self):
        user = self.backend.get_user(9999)
        self.assertIsNone(user)

    def test_authenticate_with_email_case_insensitive(self):
        # Make sure authentication works as expected
        account1 = self.backend.authenticate(self.email,
                                             DEFAULT_USER_PASSWORD)
        self.assertIsNotNone(account1)

        # Try using different case for email
        account2 = self.backend.authenticate('Mark@Example.com',
                                             DEFAULT_USER_PASSWORD)
        self.assertIsNotNone(account2)

        # Make sure both accounts are the same
        self.assertEqual(account1, account2)

    def test_authenticate_with_launchpad_encrypted_password(self):
        """The backend is able to authenticate with Launchpad cypher."""
        ap = self.account.accountpassword
        ap.password = encrypt_launchpad_password(DEFAULT_USER_PASSWORD)
        ap.save()

        account = self.backend.authenticate(self.email, DEFAULT_USER_PASSWORD)
        self.assertEqual(account, self.account)

    def test_changes_cypher_after_verifying_with_launchpad(self):
        """Change to the Django cypher when needed."""
        ap = self.account.accountpassword
        ap.password = encrypt_launchpad_password(DEFAULT_USER_PASSWORD)
        ap.save()
        self.assertFalse(check_password(DEFAULT_USER_PASSWORD, ap.password))
        self.assertTrue(validate_launchpad_password(DEFAULT_USER_PASSWORD,
                                                    ap.password))

        account = self.backend.authenticate(self.email, DEFAULT_USER_PASSWORD)
        password = account.accountpassword.password
        self.assertTrue(check_password(DEFAULT_USER_PASSWORD, password))
        self.assertFalse(validate_launchpad_password(DEFAULT_USER_PASSWORD,
                                                     password))

    def test_changes_cypher_do_not_removes_oauth_tokens(self):
        """Change to the Django cypher when needed."""
        ap = self.account.accountpassword
        ap.password = encrypt_launchpad_password(DEFAULT_USER_PASSWORD)
        ap.save()
        assert not check_password(DEFAULT_USER_PASSWORD, ap.password)
        assert validate_launchpad_password(DEFAULT_USER_PASSWORD, ap.password)

        # create a bunch of tokens for the account
        self.account.create_oauth_token('test-web', scope=TokenScope.WEB_LOGIN)
        tokens = []
        for i in xrange(3):
            t1 = self.account.create_oauth_token('test-v1-%s' % i, scope=None)
            t2 = self.account.create_oauth_token(
                'test-v2-%s' % i, scope=TokenScope.VERSION_2)
            tokens.append(t1.key)
            tokens.append(t2.key)
        assert self.account.oauth_tokens(scope=None).count() == 3
        assert self.account.oauth_tokens(
            scope=TokenScope.VERSION_2).count() == 3
        assert self.account.oauth_tokens(
            scope=TokenScope.WEB_LOGIN).count() == 1

        account = self.backend.authenticate(self.email, DEFAULT_USER_PASSWORD)
        assert account.id == self.account.id

        actual = sorted(
            [t.key for t in account.oauth_tokens(scope=None)] +
            [t.key for t in account.oauth_tokens(scope=TokenScope.VERSION_2)])
        self.assertEqual(sorted(tokens), actual)

    def test_authenticate_with_django_encrypted_password(self):
        """The backend is able to authenticate with Django cypher."""
        ap = self.account.accountpassword
        ap.password = make_password(DEFAULT_USER_PASSWORD)
        ap.save()

        account = self.backend.authenticate(self.email, DEFAULT_USER_PASSWORD)
        self.assertEqual(account, self.account)

    def test_authenticate_updates_last_leak_check_with_date(self):
        """Time of last leak check is updated when authenticating."""
        ap = self.account.accountpassword
        ap.password = make_password(DEFAULT_USER_PASSWORD)
        ap.save()

        before = now()
        account = self.backend.authenticate(self.email, DEFAULT_USER_PASSWORD)
        after = now()
        last_leak_check = account.accountpassword.last_leak_check
        self.assert_datetime_in_range(last_leak_check, before, after)

    def test_authenticate_updates_last_leak_check_with_null(self):
        """Time of last leak check is updated when authenticating - null case.

        Newly-created accounts have last_leak_check set to the past as part of
        the save process. AccountPasswords which already existed in the
        database will have last_leak_check = None (null value).
        """
        ap = self.account.accountpassword
        ap.password = make_password(DEFAULT_USER_PASSWORD)
        ap.save()
        ap.last_leak_check = None
        ap.save(update_fields=['last_leak_check'])
        assert Account.objects.all()[0].accountpassword.last_leak_check is None

        before = now()
        account = self.backend.authenticate(self.email, DEFAULT_USER_PASSWORD)
        after = now()
        last_leak_check = account.accountpassword.last_leak_check
        self.assert_datetime_in_range(last_leak_check, before, after)

    def test_doesnt_save_password_again_if_not_needed(self):
        """The backend doesn't update the password if not needed."""
        ap = self.account.accountpassword
        ap.password = make_password(DEFAULT_USER_PASSWORD)
        ap.save()

        with patch.object(settings, 'DEBUG', True):
            self.backend.authenticate(self.email, DEFAULT_USER_PASSWORD)

        for query in connection.queries:
            sql = query['sql']
            if sql.startswith('UPDATE "accountpassword" SET "password"'):
                self.fail("Shouldn't have updated the password.")

    def test_authenticate_account_active(self):
        # make sure account is active
        self.assertEqual(self.account.status, AccountStatus.ACTIVE)
        # make sure authentication succeeds
        response = self.backend.authenticate(self.email,
                                             DEFAULT_USER_PASSWORD)
        self.assertEqual(response, self.account)

    def test_authenticate_with_request(self):
        token = self.account.create_oauth_token('test')
        request = self.factory.make_authenticated_request(token=token)

        self.assertEqual(
            self.backend.authenticate(request=request), self.account)

    def test_authenticate_with_request_v1_tokens_not_valid(self):
        token = self.account.create_v1_oauth_token('test')
        request = self.factory.make_authenticated_request(token=token)

        self.assertIsNone(self.backend.authenticate(request=request))

    def test_authenticate_with_request_account_inactive(self):
        token = self.account.create_oauth_token('test')

        for status, _ in AccountStatus._get_choices():
            self.account.status = status
            self.account.save()
            # need a new request each time, for nonce freshness
            request = self.factory.make_authenticated_request(token=token)
            # make sure authentication fails if not active
            if status == AccountStatus.ACTIVE:
                self.assertEqual(
                    self.backend.authenticate(request=request), self.account)
            else:
                self.assertIsNone(self.backend.authenticate(request=request))

    def test_authenticate_account_inactive(self):
        for status, _ in AccountStatus._get_choices():
            if status == AccountStatus.ACTIVE:
                # skip as this is tested elsewhere
                continue

            self.account.status = status
            self.account.save()

            # make sure authentication fails
            response = self.backend.authenticate(self.email,
                                                 DEFAULT_USER_PASSWORD)
            self.assertEqual(response, None)

    def test_authenticate_account_no_password(self):
        self.account.accountpassword.delete()

        response = self.backend.authenticate(self.email,
                                             DEFAULT_USER_PASSWORD)

        self.assertIsNone(response)
        account = Account.objects.get_by_email(self.email)
        self.assertIsNotNone(account)

    def test_authenticate_account_with_leaked_password(self):
        leak = self.factory.make_leaked_credential(
            email=self.email, password=DEFAULT_USER_PASSWORD)
        # make sure account is not flagged for password reset
        self.assertFalse(self.account.need_password_reset)

        assert leak.password != self.account.encrypted_password

        before = now()
        account = self.backend.authenticate(
            self.email, DEFAULT_USER_PASSWORD)
        after = now()
        # verify account was flagged for password reset
        self.assertTrue(account.need_password_reset)
        # Verify the last_leak_check was updated
        self.assert_datetime_in_range(
            account.accountpassword.last_leak_check, before, after)

    def test_authenticate_account_read_only_skips_leak_checks(self):
        # Since password can't be changed anyway, should skip leak checks
        # and updates of last_leak_check in AP records.
        leak = self.factory.make_leaked_credential(
            email=self.email, password=DEFAULT_USER_PASSWORD)
        # make sure account is not flagged for password reset
        self.assertFalse(self.account.need_password_reset)

        assert leak.password != self.account.encrypted_password
        last_leak_check = self.account.accountpassword.last_leak_check

        rm = ReadOnlyManager()
        rm.set_readonly()
        self.addCleanup(rm.clear_readonly)

        account = self.backend.authenticate(
            self.email, DEFAULT_USER_PASSWORD)
        # verify account was NOT flagged for password reset
        self.assertFalse(account.need_password_reset)
        # Verify the last_leak_check was NOT updated
        self.assertEqual(account.accountpassword.last_leak_check,
                         last_leak_check)

    def test_authenticate_account_with_leaked_password_old_leak(self):
        # Test that leak is NOT checked (since the accountpassword's
        # last_leak_check is newer than the leak's when_created). This is
        # checked somewhat indirectly: the password IS leaked so it would be
        # flagged, so we know if it wasn't flagged it's because it was not
        # checked.
        leak = self.factory.make_leaked_credential(
            email=self.email, password=DEFAULT_USER_PASSWORD)
        # make sure account is not flagged for password reset
        self.assertFalse(self.account.need_password_reset)
        self.account.accountpassword.last_leak_check = now()
        self.account.accountpassword.save(update_fields=['last_leak_check'])

        assert leak.password != self.account.encrypted_password

        before = now()
        account = self.backend.authenticate(
            self.email, DEFAULT_USER_PASSWORD)
        after = now()
        # verify account was not flagged for password reset
        self.assertFalse(account.need_password_reset)
        # Verify the last_leak_check was updated anyway
        self.assert_datetime_in_range(
            account.accountpassword.last_leak_check, before, after)

    def test_authenticate_account_with_leaked_password_no_call_validate(self):
        # Ensure that an older leak does not even call _validate_raw_password,
        # since that's the whole point of verifying leak dates.
        leak = self.factory.make_leaked_credential(
            email=self.email, password=DEFAULT_USER_PASSWORD)
        # make sure account is not flagged for password reset
        self.assertFalse(self.account.need_password_reset)
        self.account.accountpassword.last_leak_check = now()
        self.account.accountpassword.save(update_fields=['last_leak_check'])

        assert leak.password != self.account.encrypted_password

        # Mock the backend's _validate_raw_password method. The mock returns
        # what _validate_raw_password would if the password matched.
        self.backend._validate_raw_password = Mock(
            return_value=self.account.accountpassword)
        self.backend.authenticate(self.email, DEFAULT_USER_PASSWORD)
        # Check that validate_raw_password was not called with a
        # hashed_password and update=False, which means it was called from
        # password_leaked, which we don't want.
        self.assertNotIn(
            call(self.account.accountpassword, DEFAULT_USER_PASSWORD,
                 hashed_password=leak.password, update=False),
            self.backend._validate_raw_password.mock_calls)

    def test_password_leaked_no_account(self):
        self.factory.make_leaked_credential(email=self.email,
                                            password=DEFAULT_USER_PASSWORD)
        self.account.delete()

        leaked = LaunchpadBackend().password_leaked(
            self.email, DEFAULT_USER_PASSWORD, None)
        self.assertTrue(leaked)

    def test_has_perm(self):
        account = self.factory.make_account(permissions=['foo', 'bar'])

        backend = LaunchpadBackend()
        self.assertTrue(backend.has_perm(account, 'identityprovider.foo'))
        self.assertTrue(backend.has_perm(account, 'identityprovider.bar'))
        self.assertFalse(backend.has_perm(account, 'identityprovider.baz'))

        self.assertTrue(backend.has_perm(account.user, 'identityprovider.foo'))
        self.assertTrue(backend.has_perm(account.user, 'identityprovider.bar'))
        self.assertFalse(
            backend.has_perm(account.user, 'identityprovider.baz'))


@override_settings(PASSWORD_HASHERS=[
    'identityprovider.auth.BCrypt13PasswordHasher',
    'django.contrib.auth.hashers.PBKDF2PasswordHasher'])
class LaunchpadBackendHashMigrationTestCase(SSOBaseTestCase):

    def test_changes_hasher_upon_authentication(self):
        """The backend changes from PBKDF2 to bcrypt upon authentication."""
        email = 'mark@example.com'
        account = self.factory.make_account(
            email=email, password=DEFAULT_USER_PASSWORD)
        ap = account.accountpassword
        ap.password = make_password(DEFAULT_USER_PASSWORD,
                                    hasher='pbkdf2_sha256')
        ap.save()
        backend = LaunchpadBackend()

        account = backend.authenticate(email, DEFAULT_USER_PASSWORD)

        encrypted = account.accountpassword.password
        hasher = get_hasher('bcrypt13')
        self.assertTrue(hasher.verify(DEFAULT_USER_PASSWORD, encrypted))

    def make_password(self, encrypted_password):
        self.factory.make_account(password=encrypted_password,
                                  password_encrypted=True)

    def test_gets_cypher_migration_status(self):
        self.make_password(make_password(str(uuid4())))
        self.make_password(encrypt_launchpad_password(str(uuid4())))
        self.make_password(make_password(str(uuid4())))

        result = get_cypher_migration_status()

        self.assertEqual(result['new_hashes'], 2)
        self.assertEqual(result['old_hashes'], 1)

    def test_uses_bcrypt_by_default(self):
        password = 'some pass'
        salt = get_hasher('bcrypt13').salt()

        default_password = make_password(password, salt=salt)
        expected_password = make_password(password, salt=salt,
                                          hasher='bcrypt13')

        self.assertEqual(default_password, expected_password)


class SSORequestValidatorTestCase(SSOBaseTestCase):

    def setUp(self):
        super(SSORequestValidatorTestCase, self).setUp()
        self.validator = SSORequestValidator()
        self.account = self.factory.make_account()
        self.client_key = self.account.openid_identifier
        self.token = self.account.create_oauth_token(token_name='v2 test')
        self.v1_token = self.account.create_v1_oauth_token(token_name='test')
        # This mock request is really a custom object from the oauthlib library
        # that for now we don't use at all (is not even a Django request so
        # it would be misleading to use the request factory here).
        self.mock_request = object()

    def test_enforce_ssl(self):
        with self.settings(OAUTH_ENFORCE_SSL=False):
            self.assertEqual(self.validator.enforce_ssl, False)

        with self.settings(OAUTH_ENFORCE_SSL=True):
            self.assertEqual(self.validator.enforce_ssl, True)

    def test_timestamp_lifetime(self):
        self.assertEqual(self.validator.timestamp_lifetime,
                         settings.OAUTH_REQUEST_MAX_AGE_SECONDS)

    def test_nonce_length(self):
        self.assertEqual(self.validator.nonce_length, (5, 100))

    def test_client_key_length(self):
        self.assertEqual(self.validator.client_key_length, (7, 7))

    def test_access_token_length(self):
        self.assertEqual(self.validator.access_token_length, (50, 60))

    def test_dummy_client(self):
        self.assertEqual(self.validator.dummy_client, 'dummy_client')

    def test_dummy_access_token(self):
        self.assertEqual(self.validator.dummy_access_token,
                         'dummy_access_token')

    def test_nonce_with_underscore(self):
        self.assertTrue(self.validator.check_nonce('NkYb2bjEbPd82_cXfH'))

    def test_nonce_with_other_allowed_chars(self):
        for i in ('-', '.', '~'):
            nonce = 'NkYb2bjEbPd82%scXfH' % i
            self.assertTrue(
                self.validator.check_nonce(nonce),
                'Nonce %r should be valid' % nonce)

    def test_validate_client_key(self):
        valid_key = self.validator.validate_client_key(
            self.client_key, request=self.mock_request)
        self.assertTrue(valid_key)

    def test_validate_access_token(self):
        valid_access_token = self.validator.validate_access_token(
            self.client_key, self.token.key, request=self.mock_request)
        self.assertTrue(valid_access_token)

    def test_validate_realms(self):
        valid_realms = self.validator.validate_realms(
            self.client_key, self.token.key, request=self.mock_request)
        self.assertTrue(valid_realms)

    def test_validate_timestamp_and_nonce(self):
        result = self.validator.validate_timestamp_and_nonce(
            self.client_key, timestamp=123456789, nonce=12345,
            request=self.mock_request, access_token=self.token.key)
        self.assertTrue(result)

        # same nonce again, not valid
        result = self.validator.validate_timestamp_and_nonce(
            self.client_key, timestamp=123456789, nonce=12345,
            request=self.mock_request, access_token=self.token.key)
        self.assertFalse(result)

    def test_validate_timestamp_and_nonce_expiration_setting(self):
        with self.settings(OAUTH_NONCE_TIMEOUT_SECONDS=0):
            result = self.validator.validate_timestamp_and_nonce(
                self.client_key, timestamp=123456789, nonce=12345,
                request=self.mock_request, access_token=self.token.key)
            self.assertTrue(result)

        # same nonce again, valid since timeout is zero
        result = self.validator.validate_timestamp_and_nonce(
            self.client_key, timestamp=123456789, nonce=12345,
            request=self.mock_request, access_token=self.token.key)
        self.assertTrue(result)

    def test_get_client_secret(self):
        result = self.validator.get_client_secret(
            self.client_key, request=self.mock_request)
        self.assertEqual(result, self.account.secret)

    def test_get_client_secret_invalid_key(self):
        result = self.validator.get_client_secret(
            '1234567890', request=self.mock_request)
        self.assertEqual(result, self.validator.dummy_client)

    def test_get_access_token_secret_matches_v2(self):
        result = self.validator.get_access_token_secret(
            self.client_key, self.token.key, request=self.mock_request)
        self.assertEqual(result, self.token.secret)

    def test_get_access_token_secret_matches_v1(self):
        result = self.validator.get_access_token_secret(
            self.client_key, self.v1_token.key, request=self.mock_request)
        self.assertEqual(result, self.v1_token.secret)

    def test_get_access_token_secret_no_match(self):
        result = self.validator.get_access_token_secret(
            self.client_key, '12345678901234567890', request=self.mock_request)
        self.assertEqual(result, self.validator.dummy_access_token)

    def test_get_access_token_secret_client_key_must_match_for_v2(self):
        other_account = self.factory.make_account()
        other_token = other_account.create_oauth_token(token_name='other')

        result = self.validator.get_access_token_secret(
            self.client_key, other_token.key, request=self.mock_request)
        self.assertEqual(result, self.validator.dummy_access_token)

        result = self.validator.get_access_token_secret(
            other_account.openid_identifier, self.token.key,
            request=self.mock_request)
        self.assertEqual(result, self.validator.dummy_access_token)

        result = self.validator.get_access_token_secret(
            other_account.openid_identifier, other_token.key,
            request=self.mock_request)
        self.assertEqual(result, other_token.secret)

        result = self.validator.get_access_token_secret(
            self.client_key, self.token.key, request=self.mock_request)
        self.assertEqual(result, self.token.secret)

    def test_get_access_token_secret_with_filter_matches_v2_token(self):
        token_filter = lambda token: 'test' not in token.name  # matches both
        validator = SSORequestValidator(token_filter)

        result = validator.get_access_token_secret(
            self.client_key, 'mismatch', request=self.mock_request)
        self.assertEqual(result, self.validator.dummy_access_token)

        result = validator.get_access_token_secret(
            self.client_key, self.token.key, request=self.mock_request)
        self.assertEqual(result, self.validator.dummy_access_token)

        result = validator.get_access_token_secret(
            self.client_key, self.v1_token.key, request=self.mock_request)
        self.assertEqual(result, self.validator.dummy_access_token)

    def test_get_access_token_secret_with_filter_matches_v1_token(self):
        token_filter = lambda token: token.name != 'test'
        validator = SSORequestValidator(token_filter)

        result = validator.get_access_token_secret(
            self.client_key, 'mismatch', request=self.mock_request)
        self.assertEqual(result, self.validator.dummy_access_token)

        result = validator.get_access_token_secret(
            self.client_key, self.token.key, request=self.mock_request)
        self.assertEqual(result, self.token.secret)

        result = validator.get_access_token_secret(
            self.client_key, self.v1_token.key, request=self.mock_request)
        self.assertEqual(result, self.validator.dummy_access_token)


class ValidateOAuthSignatureTestCase(SSOBaseTestCase):

    http_url = 'http://example.com/'
    http_method = 'GET'

    def setUp(self):
        super(ValidateOAuthSignatureTestCase, self).setUp()
        self.account = self.factory.make_account()
        self.token = self.account.create_oauth_token(token_name='test')

    def make_headers(self, **kwargs):
        kwargs.setdefault('token', self.token)
        kwargs.setdefault('url', self.http_url)
        kwargs.setdefault('http_method', self.http_method)

        _, headers, _ = self.factory.make_oauth_authorization_from_token(
            **kwargs)
        return headers

    def assert_valid(self, valid, oauth_request, token=None):
        if token is None:
            token = self.token
        self.assertTrue(valid)
        self.assertEqual(
            oauth_request.client_key, self.account.openid_identifier)
        self.assertEqual(oauth_request.resource_owner_key, token.key)
        self.assertEqual(oauth_request.uri, self.http_url)
        self.assertEqual(oauth_request.http_method, self.http_method)
        self.assertEqual(oauth_request.body, '')

    def assert_invalid(self, valid, oauth_request):
        self.assertFalse(valid)
        self.assertIsNone(oauth_request)

    def test_invalid_signature_http_url_mismatch(self):
        headers = self.make_headers(url='http://foo.com/')
        result = validate_oauth_signature(
            'http://foo.com/bar', self.http_method, headers)

        self.assert_invalid(*result)

    def test_invalid_signature_http_method_mismatch(self):
        headers = self.make_headers(http_method='GET')
        result = validate_oauth_signature(
            self.http_url, 'POST', headers)

        self.assert_invalid(*result)

    def test_invalid_signature_token_mismatch(self):
        other_token = self.factory.make_account().create_oauth_token('foo')
        headers = self.make_headers(token=other_token, consumer=self.account)
        result = validate_oauth_signature(
            self.http_url, self.http_method, headers)

        self.assert_invalid(*result)

    def test_invalid_signature_header_reuse(self):
        headers = self.make_headers()

        result = validate_oauth_signature(
            self.http_url, self.http_method, headers)
        self.assert_valid(*result)

        result = validate_oauth_signature(
            self.http_url, self.http_method, headers)
        self.assert_invalid(*result)

    def test_valid_signature(self):
        headers = self.make_headers()
        result = validate_oauth_signature(
            self.http_url, self.http_method, headers)

        self.assert_valid(*result)

    def test_valid_signature_with_filter(self):
        headers = self.make_headers()
        result = validate_oauth_signature(
            self.http_url, self.http_method, headers, lambda t: True)

        self.assert_valid(*result)

    def test_invalid_signature_with_filter(self):
        headers = self.make_headers()
        result = validate_oauth_signature(
            self.http_url, self.http_method, headers, lambda t: False)

        self.assert_invalid(*result)

    def test_invalid_signature_with_partial_filter(self):
        token = self.token
        headers = self.make_headers(token=token)
        result = validate_oauth_signature(
            self.http_url, self.http_method, headers,
            lambda t: t.name != token.name)

        self.assert_invalid(*result)

    def test_valid_signature_v1_token(self):
        token = self.account.create_v1_oauth_token('v1')
        headers = self.make_headers(token=token)
        result = validate_oauth_signature(
            self.http_url, self.http_method, headers)

        self.assert_valid(*result, token=token)

    def test_valid_signature_with_filter_v1_token(self):
        token = self.account.create_v1_oauth_token('v1')
        headers = self.make_headers(token=token)
        result = validate_oauth_signature(
            self.http_url, self.http_method, headers, lambda t: True)

        self.assert_valid(*result, token=token)

    def test_invalid_signature_with_filter_v1_token(self):
        token = self.account.create_v1_oauth_token('v1')
        headers = self.make_headers(token=token)
        result = validate_oauth_signature(
            self.http_url, self.http_method, headers, lambda t: False)

        self.assert_invalid(*result)

    def test_invalid_signature_with_partial_filter_v1_token(self):
        token = self.account.create_v1_oauth_token('v1')
        headers = self.make_headers(token=token)
        result = validate_oauth_signature(
            self.http_url, self.http_method, headers,
            lambda t: 'v1' not in t.name)

        self.assert_invalid(*result)

    def test_valid_signature_custom_nonce(self):
        headers = self.make_headers(nonce='NkYb2bjEbPd82_cXfH')
        result = validate_oauth_signature(
            self.http_url, self.http_method, headers)

        self.assert_valid(*result)

    def test_valid_signature_other_chars_in_nonce(self):
        for i in ('-', '.', '_', '~'):
            headers = self.make_headers(nonce='NkYb2bjEbPd82%scXfH' % i)
            result = validate_oauth_signature(
                self.http_url, self.http_method, headers)
            self.assert_valid(*result)


class SSOOAuthAuthenticationTestCase(SSOBaseTestCase):

    def setUp(self):
        super(SSOOAuthAuthenticationTestCase, self).setUp()
        self.auth = SSOOAuthAuthentication()
        self.account = self.factory.make_account()
        self.token = self.account.create_oauth_token('test')

    def assert_authenticated(self, request):
        response = self.auth.is_authenticated(request)

        self.assertEqual(response, True)
        self.assertEqual(getattr(request, 'user', None), self.account)
        self.assertEqual(getattr(request, 'consumer', None), self.account)
        self.assertEqual(
            getattr(request, 'throttle_extra', None), self.account.id)

    def assert_not_authenticated(self, request):
        response = self.auth.is_authenticated(request)

        self.assertEqual(response, False)
        self.assertIsNone(getattr(request, 'user', None))
        sentinel = object()
        self.assertEqual(getattr(request, 'consumer', sentinel), sentinel)
        self.assertEqual(
            getattr(request, 'throttle_extra', sentinel), sentinel)

    def test_is_authenticated_invalid_request(self):
        request = self.factory.make_authenticated_request(token=self.token)
        request.META['HTTP_AUTHORIZATION'] = 'foobar'
        self.assert_not_authenticated(request)

    def test_is_authenticated_some_error(self):
        request = self.factory.make_authenticated_request(token=self.token)
        mock_validate_token = Mock(side_effect=ValueError)
        with patch('identityprovider.auth.validate_oauth_signature',
                   mock_validate_token):
            self.assert_not_authenticated(request)

    def test_is_authenticated_no_consumer_token(self):
        request = self.factory.make_authenticated_request(token=self.token)

        new_sig = request.META['HTTP_AUTHORIZATION'].replace(
            self.account.key, 'foobar')
        request.META['HTTP_AUTHORIZATION'] = new_sig

        self.assert_not_authenticated(request)

    def test_is_authenticated_inactive_account(self):
        for status, _ in AccountStatus._get_choices():
            self.account.status = status
            self.account.save()

            request = self.factory.make_authenticated_request(token=self.token)
            # make sure authentication fails if not active
            if status == AccountStatus.ACTIVE:
                self.assert_authenticated(request)
            else:
                self.assert_not_authenticated(request)

    def test_is_authenticated_v1_token(self):
        request = self.factory.make_authenticated_request(
            token=self.account.create_v1_oauth_token('v1token'))

        self.assert_not_authenticated(request)

    def test_is_authenticated(self):
        request = self.factory.make_authenticated_request(token=self.token)

        self.assert_authenticated(request)

    def test_challenge(self):
        auth = SSOOAuthAuthentication(realm='foo')
        response = auth.challenge()
        self.assertEqual(response.status_code, 401)
        self.assertEqual(response.content, 'Authorization Required')
        self.assertEqual(response['WWW-Authenticate'], 'OAuth realm="foo"')


class BasicAuthenticateTestCase(SSOBaseTestCase):
    def test_apiuser_authenticate(self):
        apiuser = self.factory.make_apiuser()
        user = basic_authenticate(apiuser.username, DEFAULT_API_PASSWORD)
        self.assertEqual(user, apiuser)

    def test_apiuser_authenticate_with_latin1_username(self):
        username = u'ço'
        apiuser = self.factory.make_apiuser(username=username)
        user = basic_authenticate(username.encode('latin1'),
                                  DEFAULT_API_PASSWORD)
        self.assertEqual(user, apiuser)

    def test_apiuser_authenticate_with_utf8_username(self):
        username = u'ço'
        apiuser = self.factory.make_apiuser(username=username)
        user = basic_authenticate(username.encode('utf-8'),
                                  DEFAULT_API_PASSWORD)
        self.assertEqual(user, apiuser)

    def test_launchpadbackend_authenticate(self):
        account = self.factory.make_account()
        user = basic_authenticate(account.preferredemail.email,
                                  DEFAULT_USER_PASSWORD)
        self.assertEqual(user, account)

    @patch('identityprovider.auth.LaunchpadBackend.authenticate')
    @patch('identityprovider.auth.APIUser.authenticate')
    def test_password_is_decoded(self, mock_apiuser_authenticate,
                                 mock_lp_backend_authenticate):
        # make sure apiuser check fails so lp backend is tried too
        mock_apiuser_authenticate.return_value = None

        username = u'Franço'
        password = u'Franço'

        basic_authenticate(username.encode('latin1'),
                           password.encode('latin1'))

        mock_apiuser_authenticate.assert_called_once_with(
            username, password)
        mock_lp_backend_authenticate.assert_called_once_with(
            username, password)


class CypherLearningTestCase(SSOBaseTestCase):

    def test_launchpad_pass_is_not_usable_by_django(self):
        self.assertFalse(is_password_usable(encrypt_launchpad_password('foo')))

    def test_django_pass_is_usable_by_django(self):
        self.assertTrue(is_password_usable(make_password('foo')))


class ValidateRawPasswordTestCase(SSOBaseTestCase):

    def setUp(self):
        super(ValidateRawPasswordTestCase, self).setUp()
        self.email = 'mark@example.com'
        self.account = self.factory.make_account(
            email=self.email, password=DEFAULT_USER_PASSWORD)
        self.backend = LaunchpadBackend()
        self.category = None
        self.detail = None

    @contextmanager
    def dummy_timer(self, category, detail):
        self.category = category
        self.detail = detail
        yield

    def validate_and_verify_expectations(
            self, ap, password, expected_category, expected_detail,
            hashed_password=None, timer=None):
        """Encapsulate password validation and checking of expectations."""
        result_ap = self.backend._validate_raw_password(
            ap, password, hashed_password=hashed_password, timer=timer)
        if ap:
            self.assertIsInstance(result_ap, AccountPassword)
        self.assertEqual(self.category, expected_category)
        self.assertEqual(self.detail, expected_detail)

    def test_validate_raw_password_no_timer(self):
        self.validate_and_verify_expectations(
            self.account.accountpassword, DEFAULT_USER_PASSWORD,
            expected_category=None, expected_detail=None)

    def test_validate_raw_password_uses_context_manager(self):
        self.validate_and_verify_expectations(
            self.account.accountpassword, DEFAULT_USER_PASSWORD,
            expected_category="django-password-checking",
            expected_detail=str(self.account.id), timer=self.dummy_timer)

    def test_validate_raw_password_identifier_no_account_password(self):
        self.validate_and_verify_expectations(
            None, DEFAULT_USER_PASSWORD,
            expected_category="django-password-checking",
            expected_detail="(direct password comparison)",
            hashed_password=make_password(DEFAULT_USER_PASSWORD),
            timer=self.dummy_timer)


class BuildMacaroonFromRootDischargeTestCase(SSOBaseTestCase):
    """Test the deprecated build_discharge_macaroon_from_root.

    These tests are kept separated as that function won't evolve no more.
    """

    def setUp(self):
        super(BuildMacaroonFromRootDischargeTestCase, self).setUp()
        self.root_macaroon, self.macaroon_random_key, _ = self.build_macaroon(
            version=0)

    def test_root_macaroon_corrupt(self):
        self.assertRaises(ValidationError, build_discharge_macaroon_from_root,
                          "fake account", "I'm a seriously corrupted macaroon")

    def test_root_macaroon_not_for_sso(self):
        macaroon, _, _ = self.build_macaroon(
            service_location="other service", version=0)
        self.assertRaises(AuthenticationError,
                          build_discharge_macaroon_from_root,
                          "fake account", macaroon.serialize())

    def test_proper_discharging(self):
        # build the input and call
        real_account = self.factory.make_account()
        before = now()
        discharge_macaroon = build_discharge_macaroon_from_root(
            real_account, self.root_macaroon.serialize())
        after = now()

        # test
        def checker(caveat):
            """Assure all caveats inside the discharged macaroon are ok."""
            source, key, value = caveat.split("|", 2)
            if source != settings.MACAROON_SERVICE_LOCATION:
                # just checking the discharge one, not the root
                return True

            if key == 'valid_since':
                valid_since = datetime.strptime(value, '%Y-%m-%dT%H:%M:%S.%f')
                self.assertGreater(valid_since, before)
                self.assertGreater(after, valid_since)
                return True

            if key == 'last_auth':
                last_auth = datetime.strptime(value, '%Y-%m-%dT%H:%M:%S.%f')
                self.assertGreater(last_auth, before)
                self.assertGreater(after, last_auth)
                return True

            if key == 'account':
                acc = json.loads(base64.b64decode(value).decode("utf8"))
                self.assertEqual(acc['openid'], real_account.openid_identifier)
                self.assertEqual(acc['email'],
                                 real_account.preferredemail.email)
                self.assertEqual(acc['displayname'], real_account.displayname)
                self.assertEqual(acc['is_verified'], real_account.is_verified)
                return True

            if key == 'expires':
                expires = datetime.strptime(value, '%Y-%m-%dT%H:%M:%S.%f')
                before_plus_ttl = before + timedelta(
                    seconds=settings.MACAROON_TTL)
                after_plus_ttl = after + timedelta(
                    seconds=settings.MACAROON_TTL)
                self.assertGreater(expires, before_plus_ttl)
                self.assertGreater(after_plus_ttl, expires)
                return True

            # we're not validating an SSO from the discharged macaroon, fail!
            return False

        v = Verifier()
        v.satisfy_general(checker)
        v.verify(self.root_macaroon, self.macaroon_random_key,
                 [discharge_macaroon])


class BuildMacaroonDischargeTestCase(SSOBaseTestCase):

    def test_caveat_id_corrupt(self):
        self.assertRaises(ValidationError, build_discharge_macaroon,
                          "fake account", "I'm a seriously corrupted caveatid")

    def test_caveat_id_not_for_sso(self):
        macaroon, _, _ = self.build_macaroon(
            service_location="other service", version=1)
        sso_caveat = self.get_sso_caveat(
            macaroon, service_location="other service")

        self.assertRaises(ValidationError, build_discharge_macaroon,
                          "fake account", sso_caveat)

    def test_proper_discharging(self):
        # build the input and call
        root_macaroon, _, random_key = self.build_macaroon(version=1)
        sso_caveat = self.get_sso_caveat(root_macaroon)

        real_account = self.factory.make_account()
        before = now()
        discharge_macaroon = build_discharge_macaroon(
            real_account, sso_caveat.caveat_id)
        after = now()

        # test
        def checker(caveat):
            """Assure all caveats inside the discharged macaroon are ok."""
            source, key, value = caveat.split("|", 2)

            if key == 'valid_since':
                valid_since = datetime.strptime(value, '%Y-%m-%dT%H:%M:%S.%f')
                self.assertGreater(valid_since, before)
                self.assertGreater(after, valid_since)
                return True

            if key == 'last_auth':
                last_auth = datetime.strptime(value, '%Y-%m-%dT%H:%M:%S.%f')
                self.assertGreater(last_auth, before)
                self.assertGreater(after, last_auth)
                return True

            if key == 'account':
                acc = json.loads(base64.b64decode(value).decode("utf8"))
                self.assertEqual(acc['openid'], real_account.openid_identifier)
                self.assertEqual(acc['email'],
                                 real_account.preferredemail.email)
                self.assertEqual(acc['displayname'], real_account.displayname)
                self.assertEqual(acc['is_verified'], real_account.is_verified)
                return True

            if key == 'expires':
                expires = datetime.strptime(value, '%Y-%m-%dT%H:%M:%S.%f')
                before_plus_ttl = before + timedelta(
                    seconds=settings.MACAROON_TTL)
                after_plus_ttl = after + timedelta(
                    seconds=settings.MACAROON_TTL)
                self.assertGreater(expires, before_plus_ttl)
                self.assertGreater(after_plus_ttl, expires)
                return True

            # we're not validating an SSO from the discharged macaroon, fail!
            return False

        v = Verifier()
        v.satisfy_general(checker)
        v.verify(discharge_macaroon, random_key, [])


class MacaroonHelpersTestCase(SSOBaseTestCase):

    def setUp(self):
        super(MacaroonHelpersTestCase, self).setUp()
        self.stats_increment = self.patch(
            'identityprovider.auth.stats.increment')

    def test_get_caveat_ok_v0(self):
        info = {
            'some stuff': 'foo',
            'more stuff': 'bar',
        }
        _, test_rsa_pub_key = self.setup_key_pair()
        info_encrypted = base64.b64encode(
            test_rsa_pub_key.encrypt(json.dumps(info)))

        # check
        caveat_info = _decrypt_caveat(info_encrypted)
        self.assertEqual(caveat_info, info)
        self.stats_increment.assert_any_call(
            'macaroon.auth.caveat_version', key='0')

    def test_get_caveat_ok_v1(self):
        random_key = binascii.hexlify(os.urandom(32))
        _, test_rsa_pub_key = self.setup_key_pair()
        caveat_id = json.dumps({
            'version': 1,
            'secret': base64.b64encode(
                test_rsa_pub_key.encrypt(random_key)),
        })

        # check
        caveat_info = _decrypt_caveat(caveat_id)
        self.assertEqual(caveat_info, {
            'version': 1,
            '3rdparty': random_key,
        })
        self.stats_increment.assert_any_call(
            'macaroon.auth.caveat_version', key='1')

    def test_get_caveat_v0_badly_encrypted(self):
        self.assertRaises(ValidationError,
                          _decrypt_caveat, b"not really well encrypted stuff")

    def test_get_caveat_v1_bad_version(self):
        random_key = binascii.hexlify(os.urandom(32))
        _, test_rsa_pub_key = self.setup_key_pair()
        caveat_id = json.dumps({
            'version': 7,
            'secret': base64.b64encode(
                test_rsa_pub_key.encrypt(random_key)),
        })
        self.assertRaises(ValidationError, _decrypt_caveat, caveat_id)

    def test_get_caveat_v1_badly_encrypted(self):
        _, test_rsa_pub_key = self.setup_key_pair()
        caveat_id = json.dumps({
            'version': 1,
            'secret': "badly encrypted",
        })
        self.assertRaises(ValidationError, _decrypt_caveat, caveat_id)


class MacaroonRefreshFromRootTestCase(SSOBaseTestCase):
    """Test the deprecated refresh_macaroons.

    These tests are kept separated as that function won't evolve no more.
    """

    def setUp(self):
        super(MacaroonRefreshFromRootTestCase, self).setUp()
        self.root_macaroon, self.macaroon_random_key, _ = self.build_macaroon(
            version=0)

        # discharge the test macaroon
        self.account = self.factory.make_account()
        self.discharge_macaroon = build_discharge_macaroon_from_root(
            self.account, self.root_macaroon.serialize())

    def test_root_macaroon_corrupt(self):
        self.assertRaises(
            ValidationError, refresh_macaroons,
            "Corrupted macaroon", self.discharge_macaroon.serialize())

    def test_discharge_macaroon_corrupt(self):
        self.assertRaises(
            ValidationError, refresh_macaroons,
            self.root_macaroon.serialize(), "Seriously corrupted macaroon")

    def test_macaroons_dont_verify_ok(self):
        # just get *another* discharge so it's not for the same root macaroon
        other_root, _, _ = self.build_macaroon(version=0)
        other_discharge = build_discharge_macaroon_from_root(
            self.account, other_root.serialize())
        self.assertRaises(AuthenticationError, refresh_macaroons,
                          self.root_macaroon.serialize(),
                          other_discharge.serialize())

    def test_deactivated_account(self):
        self.account.deactivate()
        self.assertRaises(AccountDeactivated, refresh_macaroons,
                          self.root_macaroon.serialize(),
                          self.discharge_macaroon.serialize())

    def test_password_changed(self):
        self.account.set_password("a new password")
        self.assertRaises(AuthenticationError, refresh_macaroons,
                          self.root_macaroon.serialize(),
                          self.discharge_macaroon.serialize())

    def test_password_with_no_datetime(self):
        # simulate an "old" account password (before we started to keep the
        # changed date); note that I'm bypassing the AccountPassword save(),
        # which will update the attribute that I want in Null! (and we confirm
        # that in the assert)
        self.account.accountpassword.date_changed = None
        super(AccountPassword, self.account.accountpassword).save()
        assert self.account.accountpassword.date_changed is None

        # check that auths ok
        refresh_macaroons(self.root_macaroon.serialize(),
                          self.discharge_macaroon.serialize())

    def test_proper_refreshing(self):
        old_discharge = self.discharge_macaroon  # just rename for readability
        service_location = settings.MACAROON_SERVICE_LOCATION

        def get_value(search_key):
            for caveat in old_discharge.first_party_caveats():
                source, key, value = caveat.caveat_id.split("|", 2)
                if source == service_location and key == search_key:
                    return value

        # get old values from the macaroon and also change the account to see
        # that reflected
        old_last_auth = get_value('last_auth')
        old_expires = get_value('expires')
        new_mail = self.factory.make_email_for_account(
            self.account, status=EmailStatus.PREFERRED)
        self.account.displayname = "New test display name"
        self.account.save()

        # call!
        before = now()
        new_discharge = refresh_macaroons(self.root_macaroon.serialize(),
                                          old_discharge.serialize())
        after = now()

        # test
        def checker(caveat):
            """Assure all caveats inside the discharged macaroon are ok."""
            source, key, value = caveat.split("|", 2)
            if source != service_location:
                # just checking the NEW discharge one, not the root
                return True

            if key == 'valid_since':
                valid_since = datetime.strptime(value, '%Y-%m-%dT%H:%M:%S.%f')
                self.assertGreater(valid_since, before)
                self.assertGreater(after, valid_since)
                return True

            if key == 'last_auth':
                self.assertEqual(value, old_last_auth)
                return True

            if key == 'account':
                acc = json.loads(base64.b64decode(value).decode("utf8"))
                self.assertEqual(acc['openid'], self.account.openid_identifier)
                self.assertEqual(acc['email'], new_mail.email)
                self.assertEqual(acc['displayname'], "New test display name")
                self.assertEqual(acc['is_verified'], self.account.is_verified)
                return True

            if key == 'expires':
                self.assertEqual(value, old_expires)
                return True

            # we're not validating an SSO from the discharged macaroon, fail!
            return False

        # verify using the NEW discharge macaroon
        v = Verifier()
        v.satisfy_general(checker)
        v.verify(self.root_macaroon, self.macaroon_random_key,
                 [new_discharge])


class MacaroonRefreshTestCase(SSOBaseTestCase):

    def setUp(self):
        super(MacaroonRefreshTestCase, self).setUp()
        root_macaroon, _, self.random_key = self.build_macaroon(version=1)
        (caveat,) = [c for c in root_macaroon.third_party_caveats()
                     if c.location == settings.MACAROON_SERVICE_LOCATION]

        # get a discharge for the test macaroon
        self.account = self.factory.make_account()
        self.discharge_macaroon = build_discharge_macaroon(
            self.account, caveat.caveat_id)

    def test_discharge_macaroon_corrupt(self):
        self.assertRaises(
            ValidationError, refresh_discharge, "Seriously corrupted macaroon")

    def test_deactivated_account(self):
        self.account.deactivate()
        self.assertRaises(AccountDeactivated, refresh_discharge,
                          self.discharge_macaroon.serialize())

    def test_password_changed(self):
        self.account.set_password("a new password")
        self.assertRaises(AuthenticationError, refresh_discharge,
                          self.discharge_macaroon.serialize())

    def test_password_with_no_datetime(self):
        # simulate an "old" account password (before we started to keep the
        # changed date); note that I'm bypassing the AccountPassword save(),
        # which will update the attribute that I want in Null! (and we confirm
        # that in the assert)
        self.account.accountpassword.date_changed = None
        super(AccountPassword, self.account.accountpassword).save()
        assert self.account.accountpassword.date_changed is None

        # check that auths ok
        refresh_discharge(self.discharge_macaroon.serialize())

    def test_proper_refreshing(self):
        old_discharge = self.discharge_macaroon  # just rename for readability
        service_location = settings.MACAROON_SERVICE_LOCATION

        def get_value(search_key):
            for caveat in old_discharge.first_party_caveats():
                source, key, value = caveat.caveat_id.split("|", 2)
                if source == service_location and key == search_key:
                    return value

        # get old values from the macaroon and also change the account to see
        # that reflected
        old_last_auth = get_value('last_auth')
        old_expires = get_value('expires')
        new_mail = self.factory.make_email_for_account(
            self.account, status=EmailStatus.PREFERRED)
        self.account.displayname = "New test display name"
        self.account.save()

        # call!
        before = now()
        new_discharge = refresh_discharge(old_discharge.serialize())
        after = now()

        # test
        def checker(caveat):
            """Assure all caveats inside the discharged macaroon are ok."""
            source, key, value = caveat.split("|", 2)

            if key == 'valid_since':
                valid_since = datetime.strptime(value, '%Y-%m-%dT%H:%M:%S.%f')
                self.assertGreater(valid_since, before)
                self.assertGreater(after, valid_since)
                return True

            if key == 'last_auth':
                self.assertEqual(value, old_last_auth)
                return True

            if key == 'account':
                acc = json.loads(base64.b64decode(value).decode("utf8"))
                self.assertEqual(acc['openid'], self.account.openid_identifier)
                self.assertEqual(acc['email'], new_mail.email)
                self.assertEqual(acc['displayname'], "New test display name")
                self.assertEqual(acc['is_verified'], self.account.is_verified)
                return True

            if key == 'expires':
                self.assertEqual(value, old_expires)
                return True

            # we're not validating an SSO from the discharged macaroon, fail!
            return False

        # verify using the NEW discharge macaroon
        v = Verifier()
        v.satisfy_general(checker)
        v.verify(new_discharge, self.random_key, [])
