from django.core.management import call_command, CommandError
from django.test import TestCase

from identityprovider.models import (
    Account,
    EmailAddress,
    LPOpenIdIdentifier,
    Person,
)
from identityprovider.models.const import EmailStatus
from identityprovider.tests import DEFAULT_USER_PASSWORD


class AddEmailToAccountCommandTestCase(TestCase):

    def setUp(self):
        super(AddEmailToAccountCommandTestCase, self).setUp()
        # create an account
        self.account = Account.objects.create_account(
            'test', 'test@canonical.com', DEFAULT_USER_PASSWORD,
            openid_identifier='openid')
        # create a LP accont
        person = Person.objects.create(name='test', lp_account=42)
        # link LP account with SSO account
        LPOpenIdIdentifier.objects.create(
            identifier='openid', lp_account=person.lp_account)

    def test_add_by_email(self):
        call_command('add_canonical_email', 'new@canonical.com',
                     email='test@canonical.com')
        new_email = EmailAddress.objects.get(email='new@canonical.com')
        self.assertEqual(new_email.status, EmailStatus.VALIDATED)
        self.assertEqual(new_email.account, self.account)

    def test_add_by_email_with_unknown_email(self):
        with self.assertRaises(CommandError) as cm:
            call_command('add_canonical_email', 'new@canonical.com',
                         email='foo@canonical.com')
        self.assertEqual(
            "No account found matching email 'foo@canonical.com'.",
            str(cm.exception))

    def test_add_by_openid(self):
        call_command('add_canonical_email', 'new@canonical.com',
                     openid='openid')
        new_email = EmailAddress.objects.get(email='new@canonical.com')
        self.assertEqual(new_email.status, EmailStatus.VALIDATED)
        self.assertEqual(new_email.account, self.account)

    def test_add_by_openid_with_unknown_openid(self):
        with self.assertRaises(CommandError) as cm:
            call_command('add_canonical_email', 'new@canonical.com',
                         openid='foo')
        self.assertEqual("No account found matching openid 'foo'.",
                         str(cm.exception))

    def test_add_by_lp_username(self):
        call_command('add_canonical_email', 'new@canonical.com',
                     lp_username='test')
        new_email = EmailAddress.objects.get(email='new@canonical.com')
        self.assertEqual(new_email.status, EmailStatus.VALIDATED)
        self.assertEqual(new_email.account, self.account)

    def test_add_by_lp_username_with_unknown_lp_username(self):
        with self.assertRaises(CommandError) as cm:
            call_command('add_canonical_email', 'new@canonical.com',
                         lp_username='foo')
        self.assertEqual("No account found matching LP username 'foo'.",
                         str(cm.exception))

    def test_add_by_lp_username_without_account(self):
        # make sure the LP account exists
        person = Person.objects.create(name='foo')
        # but is not linked to any sso account
        assert person.account is None

        with self.assertRaises(CommandError) as cm:
            call_command('add_canonical_email', 'new@canonical.com',
                         lp_username='foo')
        self.assertEqual("No account found matching LP username 'foo'.",
                         str(cm.exception))

    def test_add_no_options(self):
        with self.assertRaises(CommandError) as cm:
            call_command('add_canonical_email')
        self.assertEqual(
            'Need to specify one of --email, --openid or --lp-username',
            str(cm.exception))

    def test_add_no_email(self):
        with self.assertRaises(CommandError) as cm:
            call_command('add_canonical_email', email='test@canonical.com')
        self.assertEqual('You need to specify an email to add.',
                         str(cm.exception))

    def test_add_multiple_emails(self):
        with self.assertRaises(CommandError) as cm:
            call_command('add_canonical_email', 'new1@canonical.com',
                         'new2@canonical.com', email='test@canonical.com')
        self.assertEqual('You can only specify a single email to add.',
                         str(cm.exception))

    def test_add_existing_email(self):
        call_command('add_canonical_email', 'new@canonical.com',
                     email='test@canonical.com')
        with self.assertRaises(CommandError) as cm:
            call_command('add_canonical_email', 'new@canonical.com',
                         email='test@canonical.com')
        self.assertEqual('Cannot add given email for it is already claimed.',
                         str(cm.exception))

    def test_add_non_canonical_email(self):
        with self.assertRaises(CommandError) as cm:
            call_command('add_canonical_email', 'new@example.com',
                         email='test@canonical.com')
        self.assertEqual('You can only add @canonical.com email addresses.',
                         str(cm.exception))
