from collections import defaultdict
from StringIO import StringIO

from django.core.management import call_command
from django.utils.timezone import now, timedelta

from identityprovider.models import AuthToken, Account
from identityprovider.tests.utils import SSOBaseTestCase


class CleanAuthTokensCommandTestCase(SSOBaseTestCase):

    email = 'test@canonical.com'

    def setUp(self):
        super(CleanAuthTokensCommandTestCase, self).setUp()

        self.account = self.factory.make_account(email=self.email)

    def call_the_command(self, **kwargs):
        output_file = StringIO()
        call_command('clean_old_authtokens', stdout=output_file, **kwargs)
        output_file.seek(0)
        return output_file.read()

    def test_clean_default_cutoff(self):
        deletable = self.factory.make_authtoken(
            email=self.account.preferredemail.email, requester=self.account)
        deletable.date_created = now() - timedelta(days=32)
        deletable.save()
        safe = self.factory.make_authtoken(
            email=self.account.preferredemail.email, requester=self.account,
            date_created=now() - timedelta(days=31))
        accounts_before = Account.objects.all().count()

        output = self.call_the_command()

        accounts_after = Account.objects.all().count()
        tokens_after = AuthToken.objects.all()
        self.assertIn(safe, tokens_after)
        self.assertNotIn(deletable, tokens_after)
        # Paranoid, ensure the account wasn't deleted
        self.assertEqual(accounts_before, accounts_after)
        self.assertIn("Total Auth Tokens: 2", output)
        self.assertIn("Deleting 1 tokens older than 32 days.", output)

    def test_dry_run(self):
        deletable = self.factory.make_authtoken(
            email=self.account.preferredemail.email, requester=self.account,
            date_created=now() - timedelta(days=32))
        safe = self.factory.make_authtoken(
            email=self.account.preferredemail.email, requester=self.account,
            date_created=now() - timedelta(days=31))

        output = self.call_the_command(dry_run=True)

        tokens_after = AuthToken.objects.all()
        self.assertIn(safe, tokens_after)
        self.assertIn(deletable, tokens_after)
        self.assertIn("Total Auth Tokens: 2", output)
        self.assertIn("1 tokens older than 32 days would be deleted.", output)

    def test_clean_specific_cutoff(self):
        deletable = self.factory.make_authtoken(
            email=self.account.preferredemail.email, requester=self.account,
            date_created=now() - timedelta(days=12))
        safe = self.factory.make_authtoken(
            email=self.account.preferredemail.email, requester=self.account,
            date_created=now() - timedelta(days=11))
        accounts_before = Account.objects.all().count()

        output = self.call_the_command(max_age=12)

        accounts_after = Account.objects.all().count()
        tokens_after = AuthToken.objects.all()
        self.assertIn(safe, tokens_after)
        self.assertNotIn(deletable, tokens_after)
        # Paranoid, ensure the account wasn't deleted
        self.assertEqual(accounts_before, accounts_after)
        self.assertIn("Total Auth Tokens: 2", output)
        self.assertIn("Deleting 1 tokens older than 12 days.", output)

    def test_clean_string_param(self):
        # Parameters from command-line are passed as strings to the command.
        # Ensure this doesn't blow up.
        output = self.call_the_command(max_age="12")

        self.assertIn("Deleting 0 tokens older than 12 days.", output)

    def test_clean_non_int_string_param(self):
        # Ensure an non-convertible parameter exits with a reasonable error.
        with self.assertRaises(ValueError):
            self.call_the_command(max_age="a")

    def test_multiple_accounts_tokens_ages(self):
        auth_tokens = defaultdict(list)
        accounts = []
        for i in xrange(10):
            account_mail = "{}-me@example.com".format(i)
            account = self.factory.make_account(email=account_mail)
            accounts.append(account)
            # for each account, create tokens with age from
            # 0 (created today) to $account_number
            # (so account 0 will have no tokens, account9 will have 9 tokens)
            # Ages range from 0 days (starting in account1) to 8 days (one for
            # account9)
            for age in xrange(i):
                days_ago = now() - timedelta(days=age)
                auth_tokens[age].append(self.factory.make_authtoken(
                    email=account_mail, requester=account,
                    date_created=days_ago)
                )
        # Starting with 9 days, check if we have the expected number of tokens
        # with that age, try to delete tokens older than that age, and ensure
        # the expected tokens are deleted and the others are still there.  We
        # need to start backwards because starting with shorter cutoff dates
        # will delete all older tokens in bulk.
            # we have:
            # 9 tokens with age 0
            # 8 tokens with age 1
            # ...
            # 2 tokens with age 7
            # 1  token with age 8
            # 0 tokens with age 9
        for cutoff in xrange(9, 0, -1):
            expected_tokens = 9 - cutoff
            output = self.call_the_command(max_age=cutoff)
            self.assertIn(
                "Deleting {} tokens older than {} days".format(
                    expected_tokens, cutoff), output)
            existing_tokens = AuthToken.objects.all()
            for age in xrange(9):
                for token in auth_tokens[age]:
                    if age >= cutoff:
                        self.assertNotIn(token, existing_tokens)
                    else:
                        self.assertIn(token, existing_tokens)
