import sys
import openid.store.nonce

from datetime import timedelta
from time import time
from StringIO import StringIO
from mock import Mock, patch

from django.contrib.sessions.models import Session
from django.core.management import call_command
from django.utils.timezone import now

from identityprovider.models import (
    Account,
    AccountPassword,
    EmailAddress,
    OpenIDNonce,
    OpenIDAssociation,
)
from identityprovider.tests.utils import (
    ConcurrentCommandMixin,
    SSOBaseTestCase,
)


class CleanupCommandTestCase(SSOBaseTestCase, ConcurrentCommandMixin):

    def populate(self, num=50):
        for i in xrange(num):
            # Creation and running of the tests takes less than a second,
            # and because the cutoff for reaping uses int(time()), we need
            # to ensure that nonces are at least one second older than the
            # reaping time.
            nonce_expired_timestamp = time() - openid.store.nonce.SKEW - 1
            unique = str(i) * 10
            server_url = 'http://localhost/' + str(i)
            Session.objects.create(session_key=unique, expire_date=now())
            OpenIDNonce.objects.create(server_url=server_url,
                                       timestamp=nonce_expired_timestamp)
            OpenIDAssociation.objects.create(
                server_url=server_url, handle=unique, secret=unique,
                issued=time(), lifetime=100, assoc_type='foo')

        self.assertEqual(num, Session.objects.count())
        self.assertEqual(num, OpenIDNonce.objects.count())
        self.assertEqual(num, OpenIDAssociation.objects.count())

    def make_test_accounts(self, count=0, date_created=None):
        for i in xrange(count):
            email = self.factory.make_email_address(prefix='isdtest+',
                                                    domain='canonical.com')

            account = self.factory.make_account(email=email)
            if date_created is not None:
                account.date_created = date_created
                account.save()
                EmailAddress.objects.filter(email=email).update(
                    date_created=date_created)

    def test_cleanup(self):
        self.populate()

        mock_stdout = StringIO()
        with patch.object(sys, 'stdout', mock_stdout):
            call_command('cleanup', verbosity=0)
        mock_stdout.seek(0)
        self.assertTrue('No items selected to clean up.' in mock_stdout.read())

    def test_cleanup_all(self):
        self.populate()
        call_command('cleanup', sessions=True, nonces=True, verbosity=0)

        self.assertEqual(
            0, Session.objects.filter(expire_date__lt=now()).count())
        self.assertEqual(0, OpenIDNonce.objects.filter(timestamp__lt=(
            int(time()) - openid.store.nonce.SKEW)).count())

    def test_cleanup_sessions(self):
        self.populate()
        # Before cleaning, ensure we do have sessions that will be reaped
        self.assertNotEqual(
            0, Session.objects.filter(expire_date__lt=now()).count(),
            "No sessions match the reaping criteria")
        call_command('cleanup', sessions=True, verbosity=0)

        self.assertEqual(
            0, Session.objects.filter(expire_date__lt=now()).count())
        self.assertEqual(50, OpenIDNonce.objects.count())

    def test_cleanup_nonces(self):
        self.populate()
        # Before cleaning, ensure we do have nonces that will be cleaned
        self.assertNotEqual(0, OpenIDNonce.objects.filter(timestamp__lt=(
            int(time()) - openid.store.nonce.SKEW)).count(),
            "No nonces match the reaping criteria")
        call_command('cleanup', nonces=True, verbosity=0)

        self.assertEqual(50, Session.objects.count())
        self.assertEqual(0, OpenIDNonce.objects.filter(timestamp__lt=(
            int(time()) - openid.store.nonce.SKEW)).count())

    def test_cleanup_sessions_concurrent(self):
        self.populate()
        expire_before = now()
        # Before cleaning, ensure we do have nonces that will be cleaned
        self.assertNotEqual(
            0,
            Session.objects.filter(expire_date__lt=expire_before).count(),
            "No sessions match the reaping criteria")

        # Run command in 2 threads and ensure one of them exceptions out due
        # to db locking.
        self.concurrent_command_locking_check('cleanup', sessions=True,
                                              _nowait=True, verbosity=0)
        self.assertEqual(
            0, Session.objects.filter(expire_date__lt=expire_before).count())
        self.assertEqual(50, OpenIDNonce.objects.count())

    def test_cleanup_nonces_concurrent(self):
        self.populate()
        # Before cleaning, ensure we do have nonces that will be cleaned
        self.assertNotEqual(0, OpenIDNonce.objects.filter(timestamp__lt=(
            int(time()) - openid.store.nonce.SKEW)).count(),
            "No nonces match the reaping criteria")

        # Run command in 2 threads and ensure one of them exceptions out due
        # to db locking.
        self.concurrent_command_locking_check('cleanup', nonces=True,
                                              _nowait=True, verbosity=0)
        self.assertEqual(50, Session.objects.count())
        self.assertEqual(0, OpenIDNonce.objects.filter(timestamp__lt=(
            int(time()) - openid.store.nonce.SKEW)).count())

    def assert_testdata(self, accounts=0, emails=0, passwords=0):
        tomorrow = now().date() + timedelta(days=1)
        test_emails = EmailAddress.objects.filter(
            email__iregex=r'^isdtest\+[^@]+@canonical\.com$',
            date_created__lt=tomorrow)
        test_accounts = Account.objects.filter(
            displayname__startswith='Test Account')
        test_accountpasswords = AccountPassword.objects.filter(
            account__in=test_accounts)
        self.assertEqual(test_emails.count(), emails)
        self.assertEqual(test_accounts.count(), accounts)
        self.assertEqual(test_accountpasswords.count(), passwords)

    def test_cleanup_testdata_by_date(self):
        today = now().date()
        yesterday = today - timedelta(days=1)
        self.make_test_accounts(count=10, date_created=yesterday)
        self.make_test_accounts(count=5)
        self.assert_testdata(emails=15, accounts=15, passwords=15)

        call_command('cleanup', testdata=True,
                     date_created=today.strftime('%Y-%m-%d'),
                     verbosity=0)

        self.assert_testdata(emails=5, accounts=5, passwords=5)

    def test_cleanup_testdata_limit(self):
        self.make_test_accounts(count=10)

        mock_stdout = StringIO()
        call_command('cleanup', testdata=True, limit=5,
                     verbosity=2, stdout=mock_stdout)
        expected_string = "Deleting 5 accounts..."
        # Ensure it reported deleting 5 accounts twice (for a total of 10)
        self.assertEqual(mock_stdout.getvalue().count(expected_string), 2)

    def test_cleanup_testdata(self):
        self.make_test_accounts(count=10)
        self.assert_testdata(emails=10, accounts=10, passwords=10)

        call_command('cleanup', testdata=True, verbosity=0)

        self.assert_testdata(emails=0, accounts=0, passwords=0)

    def test_cleanup_testdata_keep_special_account(self):
        self.make_test_accounts(count=10)
        self.assert_testdata(emails=10, accounts=10, passwords=10)

        # Make an isdtest@canonical.com account, this is special
        # and we need to test it survives cleanup.
        isd_email = "isdtest@canonical.com"
        isd_date = now().date() - timedelta(days=365)
        account = self.factory.make_account(email=isd_email)
        account.date_created = isd_date
        account.save()
        EmailAddress.objects.filter(email=isd_email).update(
            date_created=isd_date)

        call_command('cleanup', testdata=True, verbosity=0)

        # Sanity check: No emails matching usual test data pattern: 1 account:
        # 1 password.
        self.assert_testdata(emails=0, accounts=1, passwords=1)
        # Now verify that the special email is still there (it's linked to the
        # existing test account and password but it doesn't match the usual
        # test e-mail pattern so is not found by assert_testdata.
        self.assertEqual(
            EmailAddress.objects.filter(email='isdtest@canonical.com').count(),
            1)

    def test_cleanup_testdata_concurrent(self):
        self.make_test_accounts(count=20)
        self.assert_testdata(emails=20, accounts=20, passwords=20)

        # Run command in 2 threads and ensure one of them exceptions out due
        # to db locking.
        self.concurrent_command_locking_check('cleanup', testdata=True,
                                              _nowait=True, verbosity=0)

        self.assert_testdata(emails=0, accounts=0, passwords=0)

    def test_clean_testdata_verbose_progress(self):
        self.make_test_accounts(count=20)
        mock_stdout = Mock()

        call_command('cleanup', testdata=True, limit=5, verbosity=2,
                     stdout=mock_stdout)

        # 2 stdout writes per batch + 1 write for all accounts
        self.assertEqual(mock_stdout.write.call_count, 9)

    def test_clean_testdata_silent_progress(self):
        self.make_test_accounts(count=10)
        mock_stdout = Mock()

        call_command('cleanup', testdata=True, limit=5, verbosity=1,
                     stdout=mock_stdout)

        # 1 stdout write for all accounts
        self.assertEqual(mock_stdout.write.call_count, 1)

    def test_cleanup_orphaned_accounts(self):
        self.make_test_accounts(count=10)
        emails = EmailAddress.objects.filter(
            email__iregex=r'^isdtest\+[^@]+@canonical\.com$')
        email_ids = emails.values_list('pk')[:5]
        EmailAddress.objects.filter(pk__in=email_ids).delete()

        call_command('cleanup', testdata=True, verbosity=0)

        self.assert_testdata(emails=0, accounts=5, passwords=5)
