# Copyright 2016 Canonical Ltd.
# All Rights Reserved

import os
import tempfile
import unittest

from identityprovider import crypto


class CryptoTestCase(unittest.TestCase):

    def assert_keys_pair(self, pub_key, priv_key):
        """Assure that the keys pair is useful."""
        data = os.urandom(64)
        encrypted = pub_key.encrypt(data)
        decrypted = priv_key.decrypt(encrypted)
        self.assertEqual(decrypted, data)

    def test_parse_public_key(self):
        filepath = os.path.join(os.path.dirname(__file__),
                                'fixtures', 'test_rsa_pub_key')
        with open(filepath, "rb") as fh:
            raw_data = fh.read()
        pub_key = crypto.parse_public_key(raw_data)
        filepath = os.path.join(os.path.dirname(__file__),
                                'fixtures', 'test_rsa_priv_key')
        priv_key = crypto.load_private_key(filepath)
        self.assert_keys_pair(pub_key, priv_key)

    def test_parse_priv_key(self):
        filepath = os.path.join(os.path.dirname(__file__),
                                'fixtures', 'test_rsa_pub_key')
        pub_key = crypto.load_public_key(filepath)
        filepath = os.path.join(os.path.dirname(__file__),
                                'fixtures', 'test_rsa_priv_key')
        with open(filepath, "rb") as fh:
            raw_data = fh.read()
        priv_key = crypto.parse_private_key(raw_data)
        self.assert_keys_pair(pub_key, priv_key)

    def test_loaded_keys_ok(self):
        filepath = os.path.join(os.path.dirname(__file__),
                                'fixtures', 'test_rsa_pub_key')
        pub_key = crypto.load_public_key(filepath)
        filepath = os.path.join(os.path.dirname(__file__),
                                'fixtures', 'test_rsa_priv_key')
        priv_key = crypto.load_private_key(filepath)
        self.assert_keys_pair(pub_key, priv_key)

    def test_load_public_key_nofile(self):
        key = crypto.load_public_key("/badpath")
        self.assertIsNone(key)

    def test_load_public_key_corrupted(self):
        with tempfile.NamedTemporaryFile() as tmp_key_file:
            with open(tmp_key_file.name, "wb") as fh:
                fh.write("garbage, not real key bytes")
            key = crypto.load_public_key(tmp_key_file.name)
        self.assertIsNone(key)

    def test_load_priv_key_nofile(self):
        key = crypto.load_private_key("/badpath")
        self.assertIsNone(key)

    def test_load_priv_key_corrupted(self):
        with tempfile.NamedTemporaryFile() as tmp_key_file:
            with open(tmp_key_file.name, "wb") as fh:
                fh.write("garbage, not real key bytes")
            key = crypto.load_private_key(tmp_key_file.name)
        self.assertIsNone(key)
