# -*- coding: utf-8 -*-

# Copyright 2010 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

import re
import unittest

from django.core import mail
from django.core.urlresolvers import reverse
from django.template.loader import render_to_string
from django.test.utils import override_settings
from django.utils.timezone import now
from gpgservice_client.testing.keyserver.tests.keys import SampleDataKeys
from mock import patch

from identityprovider.emailutils import (
    _should_add_invalidation_link,
    format_address,
    send_account_suspended_warning,
    send_action_applied_notice,
    send_action_required_warning,
    send_fancy_email,
    send_gpg_validation_message,
    send_impersonation_email,
    send_invalidation_email_notice,
    send_invitation_after_password_reset,
    send_new_user_email,
    send_notification_to_invalidated_email_address,
    send_password_reset_email,
    send_preferred_changed_notification,
    send_templated_email,
    send_twofactor_failure_warning,
    send_validation_email_request,
)
from identityprovider.models import AuthToken, EmailAddress
from identityprovider.models.const import AuthTokenType, EmailStatus
from identityprovider.tests.utils import SSOBaseTestCase
from webui.gpg import SSOGPGClient
from webui.tests.gpg_fixture import SSOGPGServiceFixture


NOREPLY_FROM_ADDRESS = 'noreply@example.com'
NOREPLY_FROM_NAME = 'Cool Name'
RECIPIENTS = ['foo@zaraza.com', 'bar@zaraza.com']


class FormatAddressTestCase(unittest.TestCase):

    def assert_valid_address(self, actual, expected):
        self.assertEqual(actual, expected)
        # ensure address will not oops when going thru django email system
        result = mail.send_mail(
            'subject', 'msg', 'from@foo.com', [actual])
        self.assertEqual(result, 1)

    def test_address_and_name(self):
        result = format_address('foo@bar.com', 'Name')
        self.assert_valid_address(result, 'Name <foo@bar.com>')

    def test_address_empty_name(self):
        result = format_address('foo@bar.com', '')
        self.assert_valid_address(result, 'foo@bar.com')

    def test_address_no_name(self):
        result = format_address('foo@bar.com')
        self.assert_valid_address(result, 'foo@bar.com')

    def test_address_no_ascii(self):
        result = format_address('moño@bar.com')
        self.assert_valid_address(result, u'moño@bar.com')

    def test_unicode_name(self):
        result = format_address('foo.bar@canonical.com', u'F\xc3\xb4 Bar')
        expected = '=?utf-8?b?RsODwrQgQmFy?= <foo.bar@canonical.com>'
        self.assert_valid_address(result, expected)

    def test_unicode_name_bytestring(self):
        result = format_address('foo.bar@canonical.com', 'F\xc3\xb4 Bar')
        expected = '=?utf-8?q?F=C3=B4_Bar?= <foo.bar@canonical.com>'
        self.assert_valid_address(result, expected)

    def test_unicode_addr(self):
        email = u'f\xc3\xb4.bar@canonical.com'
        name = u'Foo Bar'
        result = format_address(email, name)
        self.assert_valid_address(result, u'%s <%s>' % (name, email))

    def test_unicode_addr_no_name(self):
        email = u'f\xc3\xb4.bar@canonical.com'
        result = format_address(email)
        self.assert_valid_address(result, email)

    def test_quoting(self):
        result = format_address('foo.bar@canonical.com', 'Foo [Baz] Bar')
        self.assert_valid_address(
            result, '"Foo \\[Baz\\] Bar" <foo.bar@canonical.com>')

    def test_long_name(self):
        result = format_address('long.name@example.com', 'a ' * 100)
        self.assertFalse('\n' in result)


@override_settings(NOREPLY_FROM_ADDRESS=NOREPLY_FROM_ADDRESS)
class SendTemplatedEmailTestCase(SSOBaseTestCase):

    def setUp(self):
        super(SendTemplatedEmailTestCase, self).setUp()

        self.mock_send = self.patch(
            'identityprovider.emailutils.send_mail'
        )
        self.mock_render = self.patch(
            'identityprovider.emailutils.render_to_string'
        )

    def test_no_from(self):
        send_templated_email("", "", {}, '')
        self.assertEqual(
            self.mock_send.call_args[0][-2],
            NOREPLY_FROM_ADDRESS,
        )

    def test_single_email(self):
        send_templated_email("", "", {}, 'test@abc.com')
        self.assertEqual(
            self.mock_send.call_args[0][-1],
            ['test@abc.com']
        )

    def test_multiple_emails(self):
        send_templated_email("", "", {}, ['test@abc.com', 'test2@abc.com'])
        self.assertEqual(
            self.mock_send.call_args[0][-1],
            ['test@abc.com', 'test2@abc.com']
        )


@override_settings(
    NOREPLY_FROM_ADDRESS=NOREPLY_FROM_ADDRESS,
    NOREPLY_FROM_NAME=NOREPLY_FROM_NAME)
class SendFancyEmailTestCase(SSOBaseTestCase):

    template_name = "email/preferred-changed.txt"

    def test_fancy_email(self, context=None):
        if context is None:
            context = {'new_preferred': 'foo@example.com'}

        send_fancy_email(
            "subject", self.template_name, context, 'test@abc.com',
            from_address='zaraza@ubuntu.com')

        self.assertEqual(len(mail.outbox), 1)
        email = mail.outbox[0]
        self.assertEqual(email.to, ['test@abc.com'])
        self.assertEqual(email.from_email, 'zaraza@ubuntu.com')
        self.assertEqual(email.subject, '%s: subject' % NOREPLY_FROM_NAME)
        self.assertEqual(
            email.body, render_to_string(self.template_name, context))

    def test_fancy_email_with_context(self):
        self.test_fancy_email(context={'new_preferred': 'bar'})

    def test_fancy_email_no_from(self):
        send_fancy_email(
            "subject", self.template_name, {}, 'test@abc.com')
        self.assertEqual(len(mail.outbox), 1)
        email = mail.outbox[0]
        self.assertEqual(
            email.from_email,
            '%s <%s>' % (NOREPLY_FROM_NAME, NOREPLY_FROM_ADDRESS))


@override_settings(
    NOREPLY_FROM_ADDRESS=NOREPLY_FROM_ADDRESS,
    NOREPLY_FROM_NAME=NOREPLY_FROM_NAME)
class SendEmailTestCase(SSOBaseTestCase):

    email = 'testing@canonical.com'
    status = EmailStatus.VALIDATED  # the EmailStatus of the target email
    redirection_url = 'http://foo.example.com'
    template_string_if_invalid = 'TEMPLATE_STRING_IF_INVALID'
    tokens_count = 1

    def setUp(self):
        super(SendEmailTestCase, self).setUp()
        self.account = self.factory.make_account(email='test@canonical.com')
        self.factory.make_email_for_account(
            email='testing@canonical.com', status=EmailStatus.VALIDATED,
            account=self.account)
        self.factory.make_email_for_account(
            email='testtest@canonical.com', status=EmailStatus.NEW,
            account=self.account)
        self.factory.make_email_for_account(
            email='testtesttest@canonical.com',
            account=self.account).invalidate()
        self.request = self.factory.make_request()

        p = self.settings(
            TEMPLATE_STRING_IF_INVALID=self.template_string_if_invalid,
        )
        p.enable()
        self.addCleanup(p.disable)

        # Do not mock send_fancy_email since we need to confirm that no
        # variable gets unrendered in the email.

        if self.status is not None:
            emailaddress = EmailAddress.objects.get(email=self.email)
            assert emailaddress.status == self.status
        else:
            assert EmailAddress.objects.filter(email=self.email).count() == 0
        assert AuthToken.objects.all().count() == 0

    def assert_tokens(self, target_email=None, requester_email=None,
                      displayname=None, password=None, tokens_count=None,
                      **kwargs):
        tokens = AuthToken.objects.all()
        if tokens_count is None:
            tokens_count = self.tokens_count
        self.assertEqual(tokens.count(), tokens_count)

        if target_email is None:
            target_email = self.email

        redirection_url = kwargs.get('redirection_url', self.redirection_url)

        for token in tokens:
            self.assertEqual(token.requester, self.account)
            self.assertEqual(token.requester_email, requester_email)
            self.assertEqual(token.email, target_email)
            self.assertEqual(token.redirection_url, redirection_url)
            if token.token_type != AuthTokenType.INVALIDATEEMAIL:
                self.assertEqual(token.displayname, displayname)
                self.assertEqual(token.password, password)
            else:
                self.assertIsNone(token.displayname)
                self.assertIsNone(token.password)

    def get_context(self, token_type, **kwargs):
        """
        Build a context to be used when rendering a test e-mail.

        :param token_type: The type of token to load from the database.
        :param kwargs: Optional arguments, one or both of token or
            invalidate_token. If given, instead of trying to load
            tokens of token_type, the specified tokens will be used directly.
            This is mainly useful for access to transient attributes which
            are not stored in the database.
        """
        db_token = AuthToken.objects.get(token_type=token_type)
        token = kwargs.get('token', db_token)
        if kwargs.get('token'):
            self.assertEqual(db_token, token)
        requester = getattr(self.account, 'displayname', token.displayname)
        token_url = self.request.build_absolute_uri(token.get_absolute_url())
        context = {
            'requester': requester,
            'requester_email': token.requester_email,
            'toaddress': self.email,
            'token': token.raw_token,
            'token_url': token_url,
        }

        # If the account does not have any validated/preferred email
        # address, ensure an invalidation token was also created
        if self.account and self.account.verified_emails().count() == 0:
            db_invalidate_token = AuthToken.objects.get(
                token_type=AuthTokenType.INVALIDATEEMAIL)
            invalidate_token = kwargs.get(
                'invalidate_token', db_invalidate_token)
            if kwargs.get('invalidate_token'):
                self.assertEqual(db_invalidate_token, invalidate_token)
            token_url = self.request.build_absolute_uri(
                invalidate_token.get_absolute_url())
            context['invalidate_email_url'] = token_url

        return context

    def assert_email_properly_formatted(self, subject, template, context=None,
                                        token_type=None, **kwargs):
        if context is None:
                # Call get_context with kwargs. get_context may care about
                # token or invalidate_token. This is used to pass around
                # tokens without retrieving them from the database, since
                # for testing purposes we may need to have their transient
                # (non-db-stored) attributes still available.
                context = self.get_context(token_type, **kwargs)
        # Remove token from kwargs otherwise it will overwrite the
        # token (raw_token) that get_context set.
        kwargs.pop('token', None)
        context.update(kwargs)

        self.assertEqual(len(mail.outbox), 1)
        email = mail.outbox[0]
        self.assertEqual(email.subject,
                         NOREPLY_FROM_NAME + ': ' + subject)
        self.assertEqual(render_to_string(template, context), email.body)
        self.assertNotIn(self.template_string_if_invalid, email.body)

        # XXX: due to https://code.djangoproject.com/ticket/19915
        # the setting TEMPLATE_STRING_IF_INVALID will not be used even if the
        # context does not provide an expected variable to be defined, so we
        # need to do this ugly yet effective test to ensure all variables are
        # defined:
        self.assertNotIn('  ', email.body)

        invalidate_msg = ('If you wish to report this email being incorrectly '
                          'used, please click the following link:')
        invalidate_url = context.get('invalidate_email_url', '')
        if invalidate_url:
            self.assertIn(invalidate_msg, email.body)
            self.assertIn(invalidate_url, email.body)
        else:
            self.assertNotIn(invalidate_msg, email.body)

    def test_send_impersonation_email(self):
        send_impersonation_email(self.sso_root_url, self.email)

        url = self.request.build_absolute_uri(
            reverse('forgot_password'))
        context = {
            'forgotten_password_url': url,
            'to_email': self.email,
        }
        self.assert_email_properly_formatted(
            'Warning', 'email/impersonate-warning.txt', context)
        email = mail.outbox[0]
        self.assertIn("If it wasn't you, no need to worry. Your account is "
                      "safe and there is nothing you need to do.",
                      email.body)

    def test_send_new_user_email_no_platform(self):
        token, invalidate_token = send_new_user_email(
            self.sso_root_url, self.account, self.email, self.redirection_url)

        self.assert_tokens()
        self.assert_email_properly_formatted(
            'Finish your registration', 'email/welcome.txt',
            token_type=AuthTokenType.VALIDATEEMAIL, token=token,
            invalidate_token=invalidate_token)

    @patch('identityprovider.emailutils.logger')
    def test_send_new_user_email_invalid_platform(self, mock_logger):
        token, invalidate_token = send_new_user_email(
            self.sso_root_url, self.account, self.email,
            self.redirection_url, platform='foo')

        self.assert_tokens()
        self.assert_email_properly_formatted(
            'Finish your registration', 'email/welcome.txt',
            token_type=AuthTokenType.VALIDATEEMAIL, token=token,
            invalidate_token=invalidate_token)

    def test_send_new_user_email_platform_desktop(self):
        token, invalidate_token = send_new_user_email(
            self.sso_root_url, self.account, self.email, self.redirection_url,
            platform='desktop')

        self.assert_tokens()
        self.assert_email_properly_formatted(
            'Finish your registration', 'email/desktop-newuser.txt',
            token_type=AuthTokenType.VALIDATEEMAIL, token=token,
            invalidate_token=invalidate_token)

    def test_send_new_user_email_platform_mobile(self):
        self.account = None
        token, invalidate_token = send_new_user_email(
            self.sso_root_url, self.account, self.email, self.redirection_url,
            platform='mobile')

        self.assert_tokens()
        self.assert_email_properly_formatted(
            'Finish your registration', 'email/welcome.txt',
            token_type=AuthTokenType.VALIDATEEMAIL, token=token,
            invalidate_token=invalidate_token)

    def test_send_password_reset_email_with_redirection_url(self):
        token, invalidate_token = send_password_reset_email(
            self.sso_root_url, self.account, self.email, self.redirection_url)

        self.assert_tokens(requester_email=self.email)
        self.assert_email_properly_formatted(
            'Forgotten Password', 'email/forgottenpassword.txt',
            token_type=AuthTokenType.PASSWORDRECOVERY, token=token,
            invalidate_token=invalidate_token)

    def test_send_password_reset_email_no_redirection_url(self):
        token, invalidate_token = send_password_reset_email(
            self.sso_root_url, self.account, self.email)

        self.assert_tokens(requester_email=self.email, redirection_url=None)
        self.assert_email_properly_formatted(
            'Forgotten Password', 'email/forgottenpassword.txt',
            token_type=AuthTokenType.PASSWORDRECOVERY, token=token,
            invalidate_token=invalidate_token)

    def test_send_password_reset_email_with_custom_reason(self):
        token, invalidate_token = send_password_reset_email(
            self.sso_root_url, self.account, self.email,
            reason='password_leak|adobe')

        self.assert_tokens(requester_email=self.email, redirection_url=None)
        token_type = AuthTokenType.PASSWORDRECOVERY
        # Tokens are passed here instead of in assert_email_properly_formatted
        context = self.get_context(token_type, token=token,
                                   invalidate_token=invalidate_token)
        # make sure the invalidation link is not present in the context
        context.pop('invalidate_email_url', None)
        self.assert_email_properly_formatted(
            'You need to change your password',
            'email/password_leak/adobe.txt',
            context=context, token_type=token_type)

    def test_send_password_reset_email_with_generic_leak_template(self):
        token, invalidate_token = send_password_reset_email(
            self.sso_root_url, self.account, self.email,
            reason='password_leak|foo')

        self.assert_tokens(requester_email=self.email, redirection_url=None)
        token_type = AuthTokenType.PASSWORDRECOVERY
        # Tokens are passed here instead of in assert_email_properly_formatted
        context = self.get_context(token_type, token=token,
                                   invalidate_token=invalidate_token)
        # make sure the invalidation link is not present in the context
        context.pop('invalidate_email_url', None)
        self.assert_email_properly_formatted(
            'You need to change your password',
            'email/password_leak/default.txt',
            context=context, token_type=token_type)

    def test_send_password_reset_email_with_invalid_reason(self):
        send_password_reset_email(self.sso_root_url, self.account, self.email,
                                  reason='invalid')

        self.assert_tokens(requester_email=self.email, redirection_url=None,
                           tokens_count=0)
        self.assertEqual(len(mail.outbox), 0)

    def test_leaked_password_reset_email_without_invalidation_link(self):
        send_password_reset_email(self.sso_root_url, self.account, self.email,
                                  reason='password_leak|adobe')

        self.assertEqual(len(mail.outbox), 1)
        mail_body = mail.outbox[0].body
        self.assertNotIn('If you wish to report this email',
                         mail_body)

    def test_send_validation_email_request_with_preferred_email(self):
        assert self.account.preferredemail is not None

        token, invalidate_token = send_validation_email_request(
            self.sso_root_url, self.account, self.email, self.redirection_url)

        email = self.account.preferredemail.email
        self.assert_tokens(requester_email=email)
        self.assert_email_properly_formatted(
            'Validate your email address', 'email/validate-email.txt',
            token_type=AuthTokenType.VALIDATEEMAIL,
            token=token, invalidate_token=invalidate_token)

    def test_send_validation_email_request_no_preferredemail(self):
        self.account.emailaddress_set.all().delete()
        self.account._preferredemail = None
        assert self.account.preferredemail is None

        token, invalidate_token = send_validation_email_request(
            self.sso_root_url, self.account, self.email, self.redirection_url)

        # two tokens are created, one for validation, and one for invalidation,
        # since the email being validated is not in the system, so the real
        # owner would want to invalidate it
        self.assert_tokens(tokens_count=2)
        self.assert_email_properly_formatted(
            'Validate your email address', 'email/validate-email.txt',
            token_type=AuthTokenType.VALIDATEEMAIL, token=token,
            invalidate_token=invalidate_token)

    def test_send_validation_email_request_no_redirection_url(self):
        token, invalidate_token = send_validation_email_request(
            self.sso_root_url, self.account, self.email)

        email = self.account.preferredemail.email
        self.assert_tokens(requester_email=email, redirection_url=None)
        self.assert_email_properly_formatted(
            'Validate your email address', 'email/validate-email.txt',
            token_type=AuthTokenType.VALIDATEEMAIL, token=token,
            invalidate_token=invalidate_token)

    def _assert_send_invalidation_email_notice(self):
        invalidated_email = 'invalid@example.com'
        send_invalidation_email_notice(self.sso_root_url, self.account,
                                       invalidated_email)

        verify_emails_link = self.request.build_absolute_uri(
            reverse('account-emails'))
        context = dict(
            display_name=self.account.displayname,
            invalidated_email=invalidated_email,
            to_email=self.account.preferredemail,
        )
        if self.account.unverified_emails().count() > 0:
            context['verify_emails_link'] = verify_emails_link
        subject = 'The email address {email} was removed from your account'
        self.assert_email_properly_formatted(
            subject.format(email=invalidated_email),
            'email/email-invalidated.txt',
            context=context)

        content = mail.outbox[0].body
        optionals = ('We strongly recommend you take the time to verify any '
                     'unverified email address by visiting this link',
                     verify_emails_link)
        for optional in optionals:
            if self.account.unverified_emails().count() == 0:
                self.assertNotIn(optional, content)
            else:
                self.assertIn(optional, content)

    def test_send_invalidation_email_notice_with_verify_email_link(self):
        assert self.account.unverified_emails().count() > 0
        self._assert_send_invalidation_email_notice()

    def test_send_invalidation_email_notice_no_verify_email_link(self):
        self.account.emailaddress_set.update(status=EmailStatus.VALIDATED)
        assert self.account.unverified_emails().count() == 0
        self._assert_send_invalidation_email_notice()

    def test_send_notification_to_invalidated_email_address(self):
        invalidated_email = 'invalid@example.com'
        send_notification_to_invalidated_email_address(self.sso_root_url,
                                                       self.account,
                                                       invalidated_email)
        context = dict(
            display_name=self.account.displayname,
            invalidated_email=invalidated_email,
            to_email=self.account.preferredemail,
        )
        subject = 'The email address {email} has been invalidated'
        self.assert_email_properly_formatted(
            subject.format(email=invalidated_email),
            'email/invalidate_email_confirmation.txt',
            context=context)

    def test_send_preferred_changed_notification(self):
        new_preferred = 'a@foo.com'
        send_preferred_changed_notification(self.email, new_preferred)

        context = {
            'new_preferred': new_preferred,
        }
        self.assert_email_properly_formatted(
            'E-mail change notification', 'email/preferred-changed.txt',
            context=context)

    def test_send_invitation_after_password_reset(self):
        url = self.request.build_absolute_uri(reverse('new_account'))
        send_invitation_after_password_reset(self.sso_root_url, self.email)

        context = {
            'email': self.email, 'signup': url,
        }
        self.assert_email_properly_formatted(
            'Password reset request', 'email/invitation.txt', context=context)

    def test_send_action_required_warning(self, action='suspend',
                                          result='suspended'):
        # since this applies to account with no validated nor preferred email,
        # remove those:
        EmailAddress.objects.filter(
            account=self.account, status__in=[EmailStatus.PREFERRED,
                                              EmailStatus.VALIDATED]).delete()
        assert self.account.verified_emails().count() == 0
        url = self.request.build_absolute_uri(reverse('account-emails'))

        token, invalidate_token = send_action_required_warning(
            self.sso_root_url, self.account, days_of_warning=8, action=action)

        # tokens will always be 2, since we need the validate email token *and*
        # the invalidate email
        self.assert_tokens(
            target_email=self.account.preferredemail.email,
            redirection_url=None, tokens_count=2)
        self.assert_email_properly_formatted(
            'Account to be %s - action required' % result,
            'email/account-action-required.txt',
            token_type=AuthTokenType.VALIDATEEMAIL,
            created=self.account.date_created,
            emails_url=url, days_of_warning=8, action=action, token=token,
            invalidate_token=invalidate_token)

    def test_send_action_applied_notice(self, action='suspend',
                                        result='suspended'):
        # since this applies to account with no validated nor preferred email,
        # remove those:
        EmailAddress.objects.filter(
            account=self.account, status__in=[EmailStatus.PREFERRED,
                                              EmailStatus.VALIDATED]).delete()

        send_action_applied_notice(
            self.account.preferredemail.email, self.account.displayname,
            days_of_warning=8, action=action)

        context = dict(
            display_name=self.account.displayname,
            days_of_warning=8, action=action,
        )
        self.assert_email_properly_formatted(
            'Account %s' % result, 'email/account-%s-applied.txt' % action,
            context=context)

    def test_send_action_required_warning_delete(self):
        self.test_send_action_required_warning(action='delete',
                                               result='deleted')

    def test_send_action_applied_notice_delete(self):
        self.test_send_action_applied_notice(action='delete',
                                             result='deleted')


class SendEmailToPreferredAddressTestCase(SendEmailTestCase):

    email = 'test@canonical.com'
    status = EmailStatus.PREFERRED

    # all the same asserts as parent, PREFERRED behaves just like VALIDATED


class SendEmailToNewAddressTestCase(SendEmailTestCase):

    email = 'testtest@canonical.com'
    status = EmailStatus.NEW
    tokens_count = 2

    # assert changes since the invalidation link has to added to emails

    def get_context(self, token_type=None, **kwargs):
        context = super(SendEmailToNewAddressTestCase, self).get_context(
            token_type, **kwargs)
        invalidate_token = kwargs.get(
            'invalidate_token',
            AuthToken.objects.get(
                token_type=AuthTokenType.INVALIDATEEMAIL))
        token_url = self.request.build_absolute_uri(
            invalidate_token.get_absolute_url())
        context['invalidate_email_url'] = token_url
        return context


class SendEmailToUnknownAddressTestCase(SendEmailToNewAddressTestCase):

    email = 'newemail@foo.com'
    status = None
    tokens_count = 2

    # all the same asserts as parent, unknown email behaves just like NEW


@override_settings(
    NOREPLY_FROM_ADDRESS=NOREPLY_FROM_ADDRESS,
    TWOFACTOR_FAILURE_NOTIFICATION_EMAILS=RECIPIENTS)
class SendTwoFactorFailureWarningTestCase(SSOBaseTestCase):

    csrf_cookie = '5SlVp2IVk9qIqRrKtqjXwwsO7XM1YDoM'
    http_cookie = (
        'csrftoken=%s;'
        '__utma=188566824.1278775677.1372447381.1372800383.1372873493.3;'
        '__utmz=188566824.1372447381.1.1.utmcsr=(direct)|'
        'utmccn=(direct)|utmcmd=(none);'
        'ssosessionid=m5gd76ivjj793dk0t333qftzncnwr5iq; C=1;'
        '__utmb=188566824.51.10.1372873493; __utmc=188566824' % csrf_cookie)
    headers = {
        'CSRF_COOKIE': '5SlVp2IVk9qIqRrKtqjXwwsO7XM1YDoM',
        'CONTENT_LENGTH': '69',
        'CONTENT_TYPE': 'application/x-www-form-urlencoded',
        'HTTP_ACCEPT': 'text/html,application/xhtml+xml,application/xml;q=0.9',
        'HTTP_ACCEPT_ENCODING': 'gzip, deflate',
        'HTTP_ACCEPT_LANGUAGE': 'en-US,en;q=0.5',
        'HTTP_CONNECTION': 'keep-alive',
        'HTTP_COOKIE': http_cookie,
        'HTTP_HOST': '10.0.3.48:8008',
        'HTTP_DNT': '1',
        'HTTP_REFERER': 'http://10.0.3.48:8008/two_factor_auth',
        'HTTP_USER_AGENT': ('Mozilla/5.0 (X11; Ubuntu; Linux x86_64; rv:22.0) '
                            'Gecko/20100101 Firefox/22.0'),
        'PATH_INFO': '/two_factor_auth',
        'REMOTE_ADDR': '10.0.3.1',
        'REMOTE_HOST': '',
        'REQUEST_METHOD': 'POST',
        'SERVER_NAME': 'sso-precise',
        'SERVER_PORT': '8008',
        'SERVER_PROTOCOL': 'HTTP/1.1',
        'SERVER_SOFTWARE': 'WSGIServer/0.1 Python/2.7.3',
        'TZ': 'UTC',
    }

    def setUp(self):
        super(SendTwoFactorFailureWarningTestCase, self).setUp()
        mock_now = self.patch('identityprovider.emailutils.now')
        mock_now.return_value = self.now = now()

        self.account = self.factory.make_account()

    def send_email(self, headers=None):
        if headers is None:
            headers = self.headers.copy()
        send_twofactor_failure_warning(self.account, headers)

        self.assertEqual(len(mail.outbox), 1)
        email = mail.outbox[0]
        return email

    def test_failure_attempts(self):
        self.account.twofactor_attempts = 18

        email = self.send_email()
        msg = 'Multiple 2FA failures (%s)' % self.account.twofactor_attempts
        self.assertIn(msg, email.body)

    def test_account_info(self):
        email = self.send_email()
        self.assertIn(self.account.openid_identifier, email.body)
        self.assertIn(self.account.displayname, email.body)

    def test_account_person_none(self):
        assert self.account.person is None

        email = self.send_email()
        self.assertIn('lp_person=None', email.body)

    def test_account_person_not_none(self):
        self.factory.make_person(account=self.account)
        assert self.account.person is not None

        email = self.send_email()
        self.assertIn('lp_person=%s' % self.account.person.name, email.body)

    def test_empty_headers(self):
        email = self.send_email(headers={})
        self.assertIn('Headers from last request are:\n<no headers>',
                      email.body)

    def test_http_cookie_removed(self):
        email = self.send_email()
        self.assertNotIn('HTTP_COOKIE', email.body)
        self.assertNotIn(self.http_cookie, email.body)

    def test_csrf_cookie_removed(self):
        email = self.send_email()
        self.assertNotIn('CSRF_COOKIE', email.body)
        self.assertNotIn(self.csrf_cookie, email.body)

    def test_headers(self):
        email = self.send_email()

        expected = '\n'.join(
            sorted('%r: %r' % (k, v) for k, v in self.headers.iteritems()
                   if k not in ('HTTP_COOKIE', 'CSRF_COOKIE'))
        )
        self.assertIn('Headers from last request are:\n%s' % expected,
                      email.body)

    def test_current_date_sent(self):
        email = self.send_email()
        self.assertIn(self.now.isoformat(), email.body)

    def test_subject(self):
        email = self.send_email()
        msg = 'Multiple 2FA failures (%s) for %s' % (
            self.account.twofactor_attempts, self.account.displayname)
        self.assertEqual(msg, email.subject)

    def test_from(self):
        email = self.send_email()
        self.assertEqual(NOREPLY_FROM_ADDRESS, email.from_email)

    def test_recipients(self):
        email = self.send_email()
        self.assertEqual(RECIPIENTS, email.recipients())


@override_settings(
    NOREPLY_FROM_ADDRESS=NOREPLY_FROM_ADDRESS,
    TWOFACTOR_FAILURE_NOTIFICATION_EMAILS=RECIPIENTS)
class SendAccountSuspendedTestCase(SSOBaseTestCase):

    def setUp(self):
        super(SendAccountSuspendedTestCase, self).setUp()
        mock_now = self.patch('identityprovider.emailutils.now')
        mock_now.return_value = self.now = now()

        self.account = self.factory.make_account()

    def send_email(self, reason='foo bar baz'):
        send_account_suspended_warning(self.account, reason)

        self.assertEqual(len(mail.outbox), 1)
        email = mail.outbox[0]
        return email

    def test_account_info(self):
        email = self.send_email()
        self.assertIn(self.account.openid_identifier, email.body)
        self.assertIn(self.account.displayname, email.body)

    def test_account_person_none(self):
        assert self.account.person is None

        email = self.send_email()
        self.assertIn('lp_person=None', email.body)

    def test_account_person_not_none(self):
        self.factory.make_person(account=self.account)
        assert self.account.person is not None

        email = self.send_email()
        self.assertIn('lp_person=%s' % self.account.person.name, email.body)

    def test_current_date_sent(self):
        email = self.send_email()
        self.assertIn(self.now.isoformat(), email.body)

    def test_reason(self):
        reason = 'Very serious reason'
        email = self.send_email(reason)
        self.assertIn('Reason: ' + reason, email.body)

    def test_subject(self):
        email = self.send_email()
        msg = 'Account suspended for %s' % self.account.displayname
        self.assertEqual(msg, email.subject)

    def test_from(self):
        email = self.send_email()
        self.assertEqual(NOREPLY_FROM_ADDRESS, email.from_email)

    def test_recipients(self):
        email = self.send_email()
        self.assertEqual(RECIPIENTS, email.recipients())


class ShouldAddInvalidationLinkTests(SSOBaseTestCase):

    def test_an_email_means_false(self):
        email = self.factory.make_email_address()
        account = self.factory.make_account()
        self.factory.make_email_for_account(account, email=email,
                                            status=EmailStatus.VALIDATED)
        self.assertEqual(
            False, _should_add_invalidation_link(email))

    def test_an_email_means_false_different_case(self):
        email = self.factory.make_email_address()
        account = self.factory.make_account()
        self.factory.make_email_for_account(account, email=email,
                                            status=EmailStatus.VALIDATED)
        self.assertEqual(
            False, _should_add_invalidation_link(email.upper()))

    def test_no_emails_means_true(self):
        email = self.factory.make_email_address()
        self.assertEqual(True, _should_add_invalidation_link(email))

    def test_new_emails_means_true(self):
        email = self.factory.make_email_address()
        account = self.factory.make_account()
        self.factory.make_email_for_account(account, email=email,
                                            status=EmailStatus.NEW)
        self.assertEqual(True, _should_add_invalidation_link(email))


class SendGPGValidationMessageTestCase(SSOBaseTestCase):

    def setUp(self):
        super(SendGPGValidationMessageTestCase, self).setUp()
        self.use_fixture(SSOGPGServiceFixture())
        self.account = self.factory.make_account()
        self.root_url = 'http://testserver'

    def extract_encrypted_block(self, text):
        """Extract a PGP encrypted block from a larger body of text."""
        regex = (
            "-----BEGIN PGP MESSAGE-----\n"
            "Version: GnuPG v1\n\n"
            ".*?"
            "-----END PGP MESSAGE-----"
            )
        match = re.search(regex, text, re.DOTALL)
        return text[match.start():match.end()] if match else None

    def assert_email_contains_pgp_block(self, email):
        self.assertIsNotNone(self.extract_encrypted_block(email.body))

    def assert_email_does_not_contain_pgp_block(self, email):
        self.assertIsNone(self.extract_encrypted_block(email.body))

    def test_raises_assertion_error_on_bad_token_type(self):
        bad_token = self.factory.make_authtoken(
            token_type=AuthTokenType.VALIDATEEMAIL,
            email=self.account.preferredemail)
        key_details = dict()
        gpg_client = SSOGPGClient()

        self.assertRaises(
            AssertionError,
            send_gpg_validation_message,
            self.root_url,
            key_details,
            bad_token,
            self.account,
            gpg_client,
        )

    def test_can_send_email_for_encryption_key(self):
        token = self.factory.make_authtoken(
            token_type=AuthTokenType.VALIDATEGPG,
            email=self.account.preferredemail,
            fingerprint=SampleDataKeys.regular_key,
        )
        gpg_client = SSOGPGClient()
        key_details = gpg_client.getKeyDetailsFromKeyServer(
            SampleDataKeys.regular_key)

        send_gpg_validation_message(
            self.root_url,
            key_details,
            token,
            self.account,
            gpg_client,
        )

        self.assertEqual(1, len(mail.outbox))
        email = mail.outbox[0]
        self.assert_email_contains_pgp_block(email)

    def test_can_send_email_for_sign_only_key(self):
        token = self.factory.make_authtoken(
            token_type=AuthTokenType.VALIDATESIGNONLYGPG,
            email=self.account.preferredemail,
            fingerprint=SampleDataKeys.sign_only_key,
        )
        gpg_client = SSOGPGClient()
        key_details = gpg_client.getKeyDetailsFromKeyServer(
            SampleDataKeys.sign_only_key)

        send_gpg_validation_message(
            self.root_url,
            key_details,
            token,
            self.account,
            gpg_client,
        )

        self.assertEqual(1, len(mail.outbox))
        email = mail.outbox[0]
        self.assert_email_does_not_contain_pgp_block(email)
