from unittest import TestCase
from urllib import urlencode

from gargoyle import gargoyle

from identityprovider.gargoyle_conditions import (
    LPTeamConditionSet,
    Regex,
    Team,
)
from identityprovider.models.account import Account
from identityprovider.models.const import EmailStatus
from identityprovider.tests.utils import SSOBaseTestCase


class TestTeam(TestCase):

    def test_is_active_team_not_found(self):
        condition = 'my-team'
        value = ['team1', 'team2']
        self.assertFalse(Team(condition).is_active(condition, value))

    def test_is_active_no_teams(self):
        condition = 'my-team'
        value = []
        self.assertFalse(Team(condition).is_active(condition, value))

    def test_is_active_team_found(self):
        condition = 'my-team'
        value = ['team1', 'my-team']
        self.assertTrue(Team(condition).is_active(condition, value))


class LPTeamConditionSetTestCase(SSOBaseTestCase):

    def setUp(self):
        super(LPTeamConditionSetTestCase, self).setUp()
        self.conditions = LPTeamConditionSet(Account)

    def test_get_namespace(self):
        self.assertEqual(self.conditions.get_namespace(), 'lp_team')

    def test_get_group_label(self):
        self.assertEqual(self.conditions.get_group_label(), 'LP Team')

    def test_can_execute(self):
        account = self.factory.make_account()
        self.assertTrue(self.conditions.can_execute(account))

    def test_can_execute_no_account(self):
        self.assertFalse(self.conditions.can_execute(None))

    def test_get_field_value_wrong_field_name(self):
        expected = []
        account = self.factory.make_account()
        team = self.factory.make_team('team1')
        self.factory.add_account_to_team(account, team)

        self.assertEqual(
            self.conditions.get_field_value(account, 'foo'), expected)

    def test_get_field_value_no_person(self):
        expected = []
        account = self.factory.make_account()

        self.assertEqual(
            self.conditions.get_field_value(account, 'team'), expected)

    def test_get_field_value_person(self):
        expected = ['team1']
        account = self.factory.make_account()
        team = self.factory.make_team('team1')
        self.factory.add_account_to_team(account, team)

        self.assertEqual(
            self.conditions.get_field_value(account, 'team'), expected)
        self.assertTrue(account.person.in_team('team1'))

    def test_get_field_value_query_count(self):
        expected_count = 1
        account = self.factory.make_account()
        team = self.factory.make_team('team1')
        self.factory.add_account_to_team(account, team)
        func = [self.conditions.get_field_value, account, 'team']
        self.assertNumQueries(expected_count, *func)
        # Add the user to a second team and check that the number of
        # queries is the same.
        team2 = self.factory.make_team('team2')
        self.factory.add_account_to_team(account, team2)
        self.assertNumQueries(expected_count, *func)


class AccountConditionSetTestCase(SSOBaseTestCase):

    key_name = 'test'

    def setUp(self):
        super(AccountConditionSetTestCase, self).setUp()
        condition_set = ('identityprovider.gargoyle_conditions.'
                         'AccountConditionSet(identityprovider.account)')
        self.conditionally_enable_flag(
            self.key_name, 'email', 'isdtest\+.*@canonical\.com',
            condition_set)
        self.account = self.factory.make_account()

    def test_is_active_no_preferred_email(self):
        self.account.preferredemail.status = EmailStatus.VALIDATED
        self.assertFalse(gargoyle.is_active(self.key_name, self.account))

    def test_is_active_preferred_email_not_matching(self):
        self.assertNotEqual(self.account.preferredemail, None)
        self.assertFalse(gargoyle.is_active(self.key_name))

    def test_is_active_preferred_email_matching(self):
        email = self.factory.make_email_for_account(
            self.account, 'isdtest+test@canonical.com', EmailStatus.PREFERRED)
        self.account.preferredemail = email
        self.assertTrue(gargoyle.is_active(self.key_name, self.account))

    def test_is_active_preferred_email_is_none(self):
        self.account.emailaddress_set.all().delete()

        account = Account.objects.get(id=self.account.id)
        self.assertEqual(account.preferredemail, None)
        self.assertFalse(gargoyle.is_active(self.key_name, account))

    def test_is_active_non_validated_email(self):
        self.factory.make_email_for_account(
            self.account, 'isdtest+test@canonical.com', EmailStatus.NEW)
        self.assertTrue(gargoyle.is_active(self.key_name, self.account))


class RequestDataConditionSetTestCase(SSOBaseTestCase):

    key_name = 'test'

    def setUp(self):
        super(RequestDataConditionSetTestCase, self).setUp()
        condition_set = (
            'identityprovider.gargoyle_conditions.RequestDataConditionSet')
        self.conditionally_enable_flag(
            self.key_name, 'email', 'isdtest(\+[^@]*)?@canonical\.com',
            condition_set)

    def test_is_active_no_email_in_post_data(self):
        request = self.factory.make_request(path='', method='POST', data={})
        self.assertFalse(gargoyle.is_active(self.key_name, request))

    def test_is_active_no_email_in_get_data(self):
        request = self.factory.make_request()
        self.assertFalse(gargoyle.is_active(self.key_name, request))

    def test_is_active_email_in_post_data_not_match(self):
        request = self.factory.make_request(
            path='', method='POST', data={'email': 'foo@foo.com'})
        self.assertFalse(gargoyle.is_active(self.key_name, request))

    def test_is_active_email_in_post_data_match(self):
        request = self.factory.make_request(
            path='', method='POST', data={'email': 'isdtest@canonical.com'})
        self.assertTrue(gargoyle.is_active(self.key_name, request))
        request = self.factory.make_request(
            path='', method='POST', data={'email': 'isdtest+foo@canonical.com'}
        )
        self.assertTrue(gargoyle.is_active(self.key_name, request))

    def test_is_active_email_in_get_data_not_match(self):
        request = self.factory.make_request(path='?email=foo@foo.com')
        self.assertFalse(gargoyle.is_active(self.key_name, request))

    def test_is_active_email_in_get_data_match(self):
        data = {'email': 'isdtest@canonical.com'}
        qs = '?' + urlencode(data)
        request = self.factory.make_request(path=qs)
        self.assertTrue(gargoyle.is_active(self.key_name, request))

        data['email'] = 'isdtest+foo@canonical.com'
        qs = '?' + urlencode(data)
        request = self.factory.make_request(path=qs)
        self.assertTrue(gargoyle.is_active(self.key_name, request))


class RegexTestCase(TestCase):

    def setUp(self):
        super(RegexTestCase, self).setUp()
        self.field = Regex()
        self.condition = u'isdtest(\\+[^@]*)?@canonical\\.com'

    def test_is_active_match(self):
        result = self.field.is_active(self.condition,
                                      'isdtest+foo@canonical.com')
        self.assertTrue(result)

    def test_is_active_no_match(self):
        result = self.field.is_active(self.condition, 'foo@foo.com')
        self.assertFalse(result)

    def test_is_active_null_value(self):
        result = self.field.is_active(self.condition, None)
        self.assertFalse(result)

    def test_is_active_multiple_values(self):
        emails = ['isdtest+foo@canonical.com', 'someother@example.com']
        result = self.field.is_active(self.condition, emails)
        self.assertTrue(result)


class GeoIPConditionSetTestCase(SSOBaseTestCase):

    key_name = 'test'

    def setUp(self):
        super(GeoIPConditionSetTestCase, self).setUp()
        condition_set = (
            'identityprovider.gargoyle_conditions.GeoIPConditionSet')
        self.conditionally_enable_flag(
            self.key_name, 'country_code', 'CN',
            condition_set, exclude=True)

    def test_is_active_from_china(self):
        IP_ADDRESSES_CHINA = ('1.1.0.1', '111.0.0.1', '222.222.222.222', )
        request = self.factory.make_request()
        for ip_address in IP_ADDRESSES_CHINA:
            request.META['REMOTE_ADDR'] = ip_address
            self.assertFalse(gargoyle.is_active(self.key_name, request))

    def test_is_active_not_from_china(self):
        IP_ADDRESSES = ('127.0.0.1', '128.128.128.128', '190.224.32.64')
        request = self.factory.make_request()
        for ip_address in IP_ADDRESSES:
            request.META['REMOTE_ADDR'] = ip_address
            self.assertTrue(gargoyle.is_active(self.key_name, request))
