# Copyright 2011 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

import re

from django.conf import settings
from django.conf.urls import patterns, url
from django.contrib.auth.models import User, AnonymousUser
from django.http import HttpResponse, HttpResponseRedirect
from django.test import TestCase, modify_settings
from mock import patch
from openid.message import IDENTIFIER_SELECT, OPENID1_URL_LIMIT, OPENID2_NS
from piston.resource import Resource

from identityprovider.const import SESSION_TOKEN_KEY
from identityprovider.middleware.useraccount import (
    UserAccountConversionMiddleware)
from identityprovider.middleware.oopsplease import OopsPleaseMiddleware
from identityprovider.middleware.readonly import ReadOnlyMiddleware
from identityprovider.middleware.statsd import PistonRequestTimingMiddleware
from identityprovider.models import Account, OpenIDRPConfig
from identityprovider.models.const import TokenScope
from identityprovider.oops_config import ErrorReporter
from identityprovider.tests import DEFAULT_USER_PASSWORD
from identityprovider.tests.utils import SSOBaseTestCase
from urls import urlpatterns as original_patterns


def _extract_csrf_token(response):
    csrf_token = response.cookies.get(settings.CSRF_COOKIE_NAME)
    if csrf_token:
        return csrf_token.value

    csrf_field = re.search('<input [^>]*name=[\'"]csrfmiddlewaretoken[\'"]'
                           '[^>]*/>', response.content)
    return csrf_field and re.search(' value=[\'"]([^\'"]+)[\'"]',
                                    csrf_field.group()).group(1)


class UserAccountConversionMiddlewareTestCase(SSOBaseTestCase):

    def setUp(self):
        super(UserAccountConversionMiddlewareTestCase, self).setUp()

        self.account = self.factory.make_account()
        self.middleware = UserAccountConversionMiddleware()
        self.user = self.account.user

    def test_when_not_admin_and_user_object_in_request(self):
        request = self.factory.make_request(user=self.user, session=True)

        # also needs a token
        token = self.account.create_oauth_token(
            'Test Web Login', scope=TokenScope.WEB_LOGIN)
        request.session[SESSION_TOKEN_KEY] = token.key

        self.middleware.process_request(request)
        self.assertIsInstance(request.user, Account)

    def test_when_not_admin_and_user_object_in_request_have_no_account(self):
        request = self.factory.make_request(
            user=User(username='fail'), session=True)

        self.middleware.process_request(request)
        self.assertIsInstance(request.user, AnonymousUser)

    def test_when_admin_and_account_object_in_request(self):
        request = self.factory.make_request(
            path='/admin/test', user=self.account)

        self.middleware.process_request(request)
        self.assertIsInstance(request.user, User)

    def test_when_admin_and_account_object_in_request_have_no_user(self):
        self.account.openid_identifier = "non-existing"
        request = self.factory.make_request(
            path='/admin/test', user=self.account, session=True)

        self.middleware.process_request(request)
        self.assertIsInstance(request.user, AnonymousUser)

    def test_session_without_token_logs_out(self):
        request = self.factory.make_request(user=self.account, session=True)

        self.middleware.process_request(request)
        self.assertIsInstance(request.user, AnonymousUser)

    def test_invalid_session_token_logs_out(self):
        request = self.factory.make_request(user=self.account, session=True)
        request.session[SESSION_TOKEN_KEY] = '123'

        self.middleware.process_request(request)
        self.assertIsInstance(request.user, AnonymousUser)

    def test_read_only_mode_session(self):
        request = self.factory.make_request(user=self.account, session=True)
        request.session[SESSION_TOKEN_KEY] = ''

        with self.settings(READ_ONLY_MODE=True):
            self.middleware.process_request(request)

        self.assertIsInstance(request.user, Account)

    def test_valid_session_token(self):
        token = self.account.create_oauth_token(
            'Test Web Login', scope=TokenScope.WEB_LOGIN)
        request = self.factory.make_request(user=self.account, session=True)
        request.session[SESSION_TOKEN_KEY] = token.key

        self.middleware.process_request(request)
        self.assertIsInstance(request.user, Account)


# The CSRF middleware requires a session cookie in order to activate.
# The tests perform a login in order to acquire this session cookie.
class CSRFMiddlewareTestCase(SSOBaseTestCase):

    email = 'mark@example.com'

    def setUp(self):
        super(CSRFMiddlewareTestCase, self).setUp()
        self.factory.make_account(email=self.email)

    def _land(self, client=None):
        client = client or self.client
        client.handler.enforce_csrf_checks = True
        r = client.get('/')
        return r, _extract_csrf_token(r)

    def _login(self, client=None, csrf_token=None):
        client = client or self.client
        client.handler.enforce_csrf_checks = True
        form = {'email': self.email, 'password': DEFAULT_USER_PASSWORD}
        if not csrf_token:
            # get the token from the login page
            r, csrf_token = self._land()
        form['csrfmiddlewaretoken'] = csrf_token
        r = client.post('/+login', form)
        self.assertIn(settings.SESSION_COOKIE_NAME, client.cookies)
        return r

    def _logout(self, client=None):
        client = client or self.client
        client.handler.enforce_csrf_checks = True
        return client.get('/+logout')

    def test_allow_with_token(self):
        self._login()
        r = self.client.get('/')
        csrf_token = _extract_csrf_token(r)
        self.assertTrue(csrf_token)
        data = {
            'displayname': 'Mark Shuttleworthy',
            'csrfmiddlewaretoken': csrf_token
        }
        r = self.client.post('/+edit', data)
        self.assertNotEquals(r.status_code, 403)

    def test_forbid_without_token(self):
        self._login()
        r = self.client.post('/+edit', {'displayname': 'Mark Shuttleworthy'})
        self.assertEqual(r.status_code, 403)

    def test_forbid_with_forged_token(self):
        self._login()
        r = self.client.post('/+edit', {
            'displayname': 'Mark Shuttleworthy',
            'csrfmiddlewaretoken': '0'})
        self.assertEqual(r.status_code, 403)

    def test_forbid_with_stolen_token(self):
        self._login()
        token = 'a'
        r, token = self._land()
        self.assertTrue(token)
        # Get a new session, but don't invalidate the old session
        self.client.cookies.clear()
        self._login()
        r = self.client.post('/+edit', {
            'displayname': 'Mark Shuttleworthy',
            'csrfmiddlewaretoken': token})
        self.assertEqual(r.status_code, 403)

    def test_ajax(self):
        self._login()
        _, token = self._land()
        headers = {'HTTP_X_CSRFTOKEN': token}
        r = self.client.post('/+edit', {'displayname': 'Mark Shuttleworthy'},
                             **headers)
        self.assertNotEquals(r.status_code, 403)

    def _login_multiple_windows(self, start_with_old_cookie,
                                logout_before_stale_login):

        # Multiple windows actually share a single browser state.
        window1 = self.client
        window2 = self.client

        if start_with_old_cookie:
            r1, csrf_token1 = self._land(window1)
            r1 = self._login(window1, csrf_token1)
            r1 = self._logout(window1)

        r1, csrf_token1 = self._land(window1)

        r2, csrf_token2 = self._land(window2)
        self._login(window2, csrf_token2)

        if logout_before_stale_login:
            self._logout(window2)

        # due to fix in django 1.5.5 csrf token is rotated on login
        # see: https://docs.djangoproject.com/en/1.5/releases/1.5.5/#properly-rotate-csrf-token-on-login  # noqa
        r1 = self._login(window1, csrf_token2)
        self.assertEqual(r1.status_code, 403)

        # but after requesting a new one, login works as expected
        r3, csrf_token3 = self._land(window1)
        r3 = self._login(window1, csrf_token3)
        self.assertNotEqual(r3.status_code, 403)

    def test_multiple_windows_no_cookie(self):
        self._login_multiple_windows(start_with_old_cookie=False,
                                     logout_before_stale_login=False)

    def test_multiple_windows_old_cookie(self):
        self._login_multiple_windows(start_with_old_cookie=True,
                                     logout_before_stale_login=False)

    def test_multiple_windows_no_cookie_logged_out(self):
        self._login_multiple_windows(start_with_old_cookie=False,
                                     logout_before_stale_login=True)

    def test_multiple_windows_old_cookie_logged_out(self):
        self._login_multiple_windows(start_with_old_cookie=True,
                                     logout_before_stale_login=True)

    def test_csrf_in_openid_forms(self):
        trust_root = 'http://localhost/'
        return_to = trust_root + (OPENID1_URL_LIMIT * 'a')

        # Setup the trust_root to be trusted for OpenID operations.
        self.rpconfig = OpenIDRPConfig(trust_root=trust_root)
        self.rpconfig.save()

        # 1. Establish a shared secret with the Provider.
        data = {
            'openid.mode': 'associate',
            'openid.assoc_type': 'HMAC-SHA1',
        }
        r = self.client.get('/+openid', data)
        [assoc_handle] = re.findall('assoc_handle:(.*)', r.content)

        # Reset cookies, just in case, because after this we pretend
        # to be the user.
        self.client.cookies.clear()

        # ... but preset the test cookie.
        self.preseed_cookie_check()

        # 2. RP directs user to Provider for checkid_setup and login.
        data = {
            'openid.mode': 'checkid_setup',
            'openid.realm': trust_root,
            'openid.return_to': return_to,
            'openid.ns': OPENID2_NS,
            'openid.identity': IDENTIFIER_SELECT,
            'openid.claimed_id': 'http://openid.launchpad.dev/+id/mark_oid',
            'openid.assoc_handle': assoc_handle,
        }
        r = self.client.get('/+openid', data, follow=True)
        self.assertTemplateUsed(r, 'registration/login.html')

        [oid_token] = re.findall(r'http://[^/]+/([^/]+)/\+decide',
                                 r.redirect_chain[-1][0])

        # 3. User logs in to SSO, but is allowed to decide whether to
        # continue back to RP.
        data = {
            'email': self.email,
            'password': DEFAULT_USER_PASSWORD
        }
        r = self.client.post('/%s/+login' % oid_token, data, follow=True)
        self.assertTemplateUsed(r, 'server/decide.html')

        csrf_token = _extract_csrf_token(r)
        # We *must* have a CSRF token on the form that submits to SSO.
        self.assertNotEquals(csrf_token, None)

        # 4. User decides to continue on to RP.
        data = {
            'csrfmiddlewaretoken': csrf_token,
            'ok': 'ok'
        }
        r = self.client.post('/%s/+decide' % oid_token, data, follow=True)
        self.assertTemplateUsed(r, 'server/post-assertion.html')
        self.assertContains(r, '<form ')
        self.assertContains(r, 'method="post"')

        # We *must not* have a CSRF token on the form that submits
        # to the RP.
        self.assertEqual(_extract_csrf_token(r), None)

    def test_csrf_cookie_is_secure_and_httponly(self):
        with self.settings(CSRF_COOKIE_SECURE=True, CSRF_COOKIE_HTTPONLY=True):
            response, _ = self._land()
        csrf_cookie = response.cookies.get(settings.CSRF_COOKIE_NAME)
        self.assertEqual(csrf_cookie['secure'], True)
        self.assertEqual(csrf_cookie['httponly'], True)


class ReadOnlyMiddlewareTestCase(SSOBaseTestCase):

    @patch('identityprovider.readonly.os')
    def test_check_for_readonly_when_ro_false_and_file(self, mock_os):
        mock_os.path.exists.return_value = True
        middleware = ReadOnlyMiddleware()
        with self.settings(READ_ONLY_MODE=False):
            middleware.process_request(None)

            self.assertTrue(settings.READ_ONLY_MODE)

    @patch('identityprovider.readonly.os')
    def test_check_for_readonly_when_ro_true_and_file(self, mock_os):
        mock_os.path.exists.return_value = True
        middleware = ReadOnlyMiddleware()
        with self.settings(READ_ONLY_MODE=True):
            middleware.process_request(None)

            self.assertTrue(settings.READ_ONLY_MODE)

    @patch('identityprovider.readonly.os')
    def test_check_for_readonly_when_ro_false_and_no_file(self, mock_os):
        mock_os.path.exists.return_value = False
        middleware = ReadOnlyMiddleware()
        with self.settings(READ_ONLY_MODE=False):
            middleware.process_request(None)

            self.assertFalse(settings.READ_ONLY_MODE)

    @patch('identityprovider.readonly.os')
    def test_check_for_readonly_when_ro_true_and_no_file(self, mock_os):
        mock_os.path.exists.return_value = False
        middleware = ReadOnlyMiddleware()
        with self.settings(READ_ONLY_MODE=True):
            middleware.process_request(None)

            self.assertFalse(settings.READ_ONLY_MODE)


class OopsPleaseMiddlewareTests(SSOBaseTestCase):

    def setUp(self):
        super(OopsPleaseMiddlewareTests, self).setUp()
        self.reports = []
        self.reported_ids = None

    def collect_oopses(self, report):
        self.reports.append(report)
        self.reported_ids = [0]
        return self.reported_ids

    def get_reporter(self):
        reporter = ErrorReporter()
        reporter.oops_config.publisher = self.collect_oopses
        return reporter

    def do_request(self, middleware, request, response=None, reporter=None):
        if response is None:
            response = {}
        with patch(
            'identityprovider.middleware.oopsplease.global_error_reporter',
                reporter):
            return middleware.process_response(request, response)

    def full_request(self, query_string):
        middleware = OopsPleaseMiddleware()
        reporter = self.get_reporter()
        request = self.factory.make_request('/', data=query_string)
        self.do_request(middleware, request, reporter=reporter)
        return request

    def test_no_oops_for_no_query_string(self):
        self.full_request({})
        self.assertEqual([], self.reports)

    def test_oops_for_query_string(self):
        self.full_request({OopsPleaseMiddleware.QUERY_STRING_KEY: 1})
        self.assertEqual(1, len(self.reports))

    def test_uses_request(self):
        request = self.full_request(
            {OopsPleaseMiddleware.QUERY_STRING_KEY: 1})
        self.assertEqual(request.build_absolute_uri(), self.reports[0]['url'])

    def test_makes_exception(self):
        self.full_request({OopsPleaseMiddleware.QUERY_STRING_KEY: 1})
        self.assertEqual('ManuallyGeneratedOops', self.reports[0]['type'])

    def test_returns_response(self):
        middleware = OopsPleaseMiddleware()
        response = object()
        request = self.factory.make_request()
        self.assertEqual(
            response, self.do_request(middleware, request, response=response))

    def test_adds_header_to_response(self):
        middleware = OopsPleaseMiddleware()
        reporter = self.get_reporter()
        request = self.factory.make_request(
            '/', data={OopsPleaseMiddleware.QUERY_STRING_KEY: 1})
        response = self.do_request(middleware, request, reporter=reporter)
        self.assertEqual(repr(self.reported_ids), response['X-Oops-Id'])


class TestHandler(object):
    pass


class PistonRequestTimingMiddlewareTests(SSOBaseTestCase):

    def setUp(self):
        super(PistonRequestTimingMiddlewareTests, self).setUp()
        self.mock_statsd_process_view = self.patch(
            'django_statsd.middleware.'
            'GraphiteRequestTimingMiddleware.process_view')

    def test_uses_handler_for_piston_resource(self):
        resource = Resource(handler=TestHandler)
        middleware = PistonRequestTimingMiddleware()
        request = self.factory.make_request()

        middleware.process_view(request, resource, None, None)

        self.mock_statsd_process_view.assert_called_once_with(
            request, resource.handler, None, None)

    def test_uses_name_for_function(self):
        middleware = PistonRequestTimingMiddleware()
        request = self.factory.make_request()

        def handler():
            pass

        middleware.process_view(request, handler, None, None)

        self.mock_statsd_process_view.assert_called_once_with(
            request, handler, None, None)


def dummy(request):
    dummy_html = """<!DOCTYPE html><body>
        <form action="." method="POST"></form>
        </body></html>"""

    if request.method == 'POST':
        return HttpResponseRedirect('/')

    return HttpResponse(dummy_html)


urlpatterns = original_patterns + patterns(
    '',
    url(r'^honeypot-test/$', 'identityprovider.tests.test_middleware.dummy',
        name='honeypot-test'),
)


@modify_settings(MIDDLEWARE_CLASSES={
    'append': 'identityprovider.middleware.honeypot.HoneypotMiddleware'
})
class HoneypotEnabledTestCase(TestCase):

    urls = 'identityprovider.tests.test_middleware'
    url = '/honeypot-test/'

    def test_honeypot_field_name_compatible_with_openid_dance(self):
        self.assertTrue(settings.HONEYPOT_FIELD_NAME.startswith('openid.'))

    def test_get_has_honeypot_field(self):
        response = self.client.get(self.url)

        self.assertContains(response, '<form action="." method="POST">')
        honeypot = '<input type="text" name="%s" value="" />'
        self.assertContains(response, honeypot % settings.HONEYPOT_FIELD_NAME)

    def test_post_fails_without_honeypot_field(self):
        response = self.client.post(self.url)
        self.assertContains(
            response, 'Bad bot, go away! Request aborted.', status_code=400)

    def test_post_fails_with_non_empty_honeypot_field(self):
        response = self.client.post(
            self.url, data={settings.HONEYPOT_FIELD_NAME: 'x'})
        self.assertContains(
            response, 'Bad bot, go away! Request aborted.', status_code=400)

    def test_post_success_with_empty_honeypot_field(self):
        response = self.client.post(
            self.url, data={settings.HONEYPOT_FIELD_NAME: ''})
        self.assertRedirects(response, '/')


class HoneypotWhitelistedTestCase(TestCase):

    def assert_honeypot_not_required(self, url, **kwargs):
        response = self.client.post(url, **kwargs)
        # ensure the test makes sense, ie that the answer is not 404
        self.assertNotEqual(
            response.status_code, 404,
            'POST to URL %s should not result in 404.' % url)
        # we could have a 400 from a bad request to the API, for example
        if response.status_code == 400:
            self.assertNotContains(response, 'Bad bot', status_code=400)

    def test_post_to_api_v1_not_required_to_have_honeypot(self):
        for i in ('1.0/', '1.1/'):
            url = '/api/' + i
            self.assert_honeypot_not_required(
                url, **{'HTTP_ACCEPT': 'application/vd.sun.wadl+xml'})

    def test_post_to_api_v2_not_required_to_have_honeypot(self):
        self.assert_honeypot_not_required('/api/v2/accounts')

    def test_post_to_openid_urls_not_required_to_have_honeypot(self):
        self.assert_honeypot_not_required('/+openid')

    def test_post_to_openid_urls_with_lang_not_required_to_have_honeypot(self):
        for i in ('en', 'cn', 'es'):
            url = '/%s/+openid' % i
            self.assert_honeypot_not_required(url)

    def test_post_to_saml_urls_not_required_to_have_honeypot(self):
        for saml_base in ('/+saml', '/saml'):
            for i in ('', '/process'):
                url = saml_base + i
                self.assert_honeypot_not_required(
                    url, data={'SAMLRequest': 'foo', 'RelayState': 'bar'})
