# Copyright 2010-2016 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).


from django.conf import settings
from django.contrib.auth.hashers import get_hasher, is_password_usable
from django.contrib.auth.models import User, check_password, make_password
from django.core.exceptions import ValidationError
from django.db import IntegrityError, connection
from django.test.utils import override_settings
from django.utils.timezone import datetime, now, timedelta
from mock import Mock, patch
from oauth_backend.models import Consumer as OAuthConsumer, Token as OAuthToken

from identityprovider.login import (
    AuthenticationError,
    AccountDeactivated,
    AccountSuspended,
    PasswordPolicyError,
    authenticate_user,
)
from identityprovider.models.account import (
    Account,
    AccountPassword,
    AccountQuerySet,
)
from identityprovider.models.emailaddress import EmailAddress
from identityprovider.models.const import (
    AccountStatus,
    AccountCreationRationale,
    EmailStatus,
    TokenScope,
)
from identityprovider.readonly import ReadOnlyManager
from identityprovider.tests import DEFAULT_USER_PASSWORD
from identityprovider.tests.utils import SSOBaseTestCase
from identityprovider.views.testing import MockLaunchpad
from u1testutils.logging import LogHandlerTestCase


class MockBackend(object):

    def __init__(self, value):
        self.value = value

    def has_module_perms(self, account, app_label):
        return self.value

    def has_perm(self, account, perm):
        return self.value


class AccountQuerySetTestCase(SSOBaseTestCase):

    def setUp(self):
        super(AccountQuerySetTestCase, self).setUp()
        self.qs = AccountQuerySet(model=Account)

    def test_verified_list_only_verified_accounts(self):
        accounts = self.qs.verified()
        for account in accounts:
            emails = account.emailaddress_set.all()
            self.assertTrue(any([email.is_verified for email in emails]))


class AccountManagerTestCase(SSOBaseTestCase):

    def test_get_by_email_no_accounts(self):
        with self.assertNumQueries(1):
            account = Account.objects.get_by_email(
                'somethin@which.not.exists.com')

        self.assertTrue(account is None)

    def test_get_by_email_fails(self):
        account = self.factory.make_account()
        email = 'somethin@which.not.exists.com'
        assert account.preferredemail != email
        with self.assertNumQueries(1):
            account = Account.objects.get_by_email(email)

        self.assertTrue(account is None)

    def test_get_by_email_exists(self):
        email = 'a@a.com'
        account = self.factory.make_account(email=email)
        with self.assertNumQueries(1):
            result = Account.objects.get_by_email(email)

        self.assertEqual(account, result)

    def test_create_account_with_invalid_password(self):
        salt = get_hasher('default').salt()
        kwargs = dict(
            displayname='displayname', email_address='email@host.local',
            password='test', salt=salt,
            creation_rationale=AccountCreationRationale.USER_CREATED,
        )
        self.assertRaises(ValidationError, Account.objects.create_account,
                          **kwargs)

    def test_create_account_with_invalid_email_address(self):
        salt = get_hasher('default').salt()
        kwargs = dict(
            displayname='displayname', email_address='not_an_email',
            password=DEFAULT_USER_PASSWORD, salt=salt,
            creation_rationale=AccountCreationRationale.USER_CREATED,
        )
        self.assertRaises(ValidationError, Account.objects.create_account,
                          **kwargs)

    def test_create_account_with_rationale(self):
        salt = get_hasher('default').salt()
        account = Account.objects.create_account(
            displayname='displayname', email_address='email@host.local',
            password=DEFAULT_USER_PASSWORD,
            creation_rationale=AccountCreationRationale.USER_CREATED,
            salt=salt,
        )

        self.assertEqual(account.displayname, 'displayname')
        self.assertEqual(account.creation_rationale,
                         AccountCreationRationale.USER_CREATED)
        self.assertEqual(account.status, AccountStatus.ACTIVE)

        emails = EmailAddress.objects.filter(account=account)
        self.assertEqual(emails.count(), 1)
        email = emails[0]
        self.assertEqual(email.email, 'email@host.local')
        self.assertEqual(email.status, EmailStatus.PREFERRED)

        passwords = AccountPassword.objects.filter(account=account)
        self.assertEqual(passwords.count(), 1)
        password = passwords[0]
        self.assertEqual(password.password,
                         make_password(DEFAULT_USER_PASSWORD, salt=salt))

    def test_create_account_with_no_rationale(self):
        salt = get_hasher('default').salt()
        account = Account.objects.create_account(
            displayname='displayname', email_address='email@host.local',
            password=DEFAULT_USER_PASSWORD, salt=salt)

        self.assertEqual(account.displayname, 'displayname')
        self.assertEqual(account.creation_rationale,
                         AccountCreationRationale.OWNER_CREATED_LAUNCHPAD)
        self.assertEqual(account.status, AccountStatus.ACTIVE)

        emails = EmailAddress.objects.filter(account=account)
        self.assertEqual(emails.count(), 1)
        email = emails[0]
        self.assertEqual(email.email, 'email@host.local')
        self.assertEqual(email.status, EmailStatus.PREFERRED)

        passwords = AccountPassword.objects.filter(account=account)
        self.assertEqual(passwords.count(), 1)
        password = passwords[0]
        self.assertEqual(password.password,
                         make_password(DEFAULT_USER_PASSWORD, salt=salt))

    def test_create_account_with_email_validated(self):
        salt = get_hasher('default').salt()
        account = Account.objects.create_account(
            displayname='displayname', email_address='email@host.local',
            password=DEFAULT_USER_PASSWORD, salt=salt, email_validated=False)

        self.assertEqual(account.displayname, 'displayname')
        self.assertEqual(account.creation_rationale,
                         AccountCreationRationale.OWNER_CREATED_LAUNCHPAD)
        self.assertEqual(account.status, AccountStatus.ACTIVE)

        emails = EmailAddress.objects.filter(account=account)
        self.assertEqual(emails.count(), 1)
        email = emails[0]
        self.assertEqual(email.email, 'email@host.local')
        self.assertEqual(email.status, EmailStatus.NEW)

        passwords = AccountPassword.objects.filter(account=account)
        self.assertEqual(passwords.count(), 1)
        password = passwords[0]
        self.assertEqual(password.password,
                         make_password(DEFAULT_USER_PASSWORD, salt=salt))

    def test_create_account_no_password(self):
        account = Account.objects.create_account(
            displayname='John Doe', email_address='email@host.local',
            password=None)

        self.assertFalse(is_password_usable(account.accountpassword.password))

    def test_create_account_no_email(self):
        with self.assertRaises(ValidationError) as ctx:
            Account.objects.create_account(
                displayname='John Doe', email_address=None,
                password=DEFAULT_USER_PASSWORD)

        e = ctx.exception
        self.assertEqual(
            e.message_dict, {'email': ['This field cannot be null.']})

    def test_create_account_no_displayname(self):
        with self.assertRaises(ValidationError) as ctx:
            Account.objects.create_account(
                displayname=None, email_address='email@host.local',
                password=DEFAULT_USER_PASSWORD)

        e = ctx.exception
        self.assertEqual(e.message_dict, {'displayname': ['Field required']})

    def test_create_account_displayname_empty(self):
        with self.assertRaises(ValidationError) as ctx:
            Account.objects.create_account(
                displayname='', email_address='email@host.local',
                password=DEFAULT_USER_PASSWORD)

        e = ctx.exception
        self.assertEqual(e.message_dict, {'displayname': ['Field required']})

    def test_create_account_duplicated_email(self):
        email = 'email@foo.com'
        account = Account.objects.create_account(
            displayname='Foo', email_address=email, password='12345678')
        with self.assertRaises(IntegrityError) as ctx:
            Account.objects.create_account(
                displayname='Bar', email_address=email, password='87654321')

        stored = Account.objects.get_by_email(email)
        self.assertEqual(account, stored)
        msg = 'duplicate key value violates unique constraint'
        self.assertIn(msg, unicode(ctx.exception))

    def test_active_by_openid_no_accounts(self):
        result = Account.objects.active_by_openid('123456')
        self.assertIsNone(result)

    def test_active_by_openid_no_match(self):
        account = self.factory.make_account()
        assert account.openid_identifier != '123456'
        result = Account.objects.active_by_openid('123456')
        self.assertIsNone(result)

    def test_active_by_openid(self):
        account = self.factory.make_account()
        openid = account.openid_identifier
        for status, name in AccountStatus._get_choices():
            account.status = status
            account.save()
            result = Account.objects.active_by_openid(openid)
            if status == AccountStatus.ACTIVE:
                assert account.is_active
                self.assertEqual(result, account)
            else:
                self.assertIsNone(result)

    def test_verified_list_only_verified_accounts(self):
        accounts = Account.objects.verified()
        for account in accounts:
            emails = account.emailaddress_set.all()
            self.assertTrue(any([email.is_verified for email in emails]))

    def test_verified_no_duplicates(self):
        accounts = Account.objects.verified()
        self.assertEqual(accounts.count(), accounts.distinct().count())


@override_settings(READ_ONLY_MODE=False)
class AccountTestCase(SSOBaseTestCase):

    def prepare_perms(self, has_perm):
        account = self.factory.make_account()

        def mock_get_backends():
            return [MockBackend(has_perm)]

        self.patch('identityprovider.models.account.get_backends',
                   new=mock_get_backends)
        return account

    def test_null_last_login_when_user_doesnt_exist(self):
        account = self.factory.make_account()
        account.user.delete()
        self.assertTrue(account.last_login is None)

    def test_null_last_login_when_user_exists_but_never_logged_in(self):
        account = self.factory.make_account()
        assert account.user.username == account.openid_identifier

        self.assertTrue(account.last_login is None)

    def test_set_last_login_when_user_does_not_exists(self):
        account = self.factory.make_account()
        account.last_login = now()

        self.assertTrue(account.last_login is not None)

    def test_set_last_login_when_readonly(self):
        readonly_manager = ReadOnlyManager()
        account = self.factory.make_account()
        account.last_login = last_login = datetime(2010, 01, 01)
        self.assertEqual(account.last_login, last_login)

        assert not settings.READ_ONLY_MODE
        readonly_manager.set_readonly()
        self.addCleanup(readonly_manager.clear_readonly)

        account.last_login = the_now = now()
        self.assertEqual(account.last_login, last_login)
        self.assertNotEqual(account.last_login, the_now)

    def test_set_last_login_when_no_preferredemail(self):
        account = self.factory.make_account()
        account.emailaddress_set.all().update(status=EmailStatus.NEW)
        account.last_login = now()
        # Test it has actually created a User object
        self.assertEqual(
            1, User.objects.filter(username=account.openid_identifier).count())

    def test_is_active(self):
        account = self.factory.make_account()

        self.assertTrue(account.is_active)

    def test_is_staff(self):
        account = self.factory.make_account()

        self.assertFalse(account.is_staff)

    def test_is_authenticated(self):
        account = self.factory.make_account()

        self.assertTrue(account.is_authenticated)

    def test_is_superuser(self):
        account = self.factory.make_account()

        self.assertFalse(account.is_superuser)

    def test_is_not_verified(self):
        account = self.factory.make_account(email_validated=False)
        self.assertFalse(account.is_verified)

    def test_is_verified(self):
        account = self.factory.make_account()
        self.assertTrue(account.is_verified)

    def test_twofactor_attempts_default(self):
        account = self.factory.make_account()
        self.assertEqual(account.twofactor_attempts, 0)

    def test_increment_twofactor_attempts(self):
        account = self.factory.make_account()
        assert account.twofactor_attempts == 0

        account.increment_twofactor_attempts()

        # reload account
        account = Account.objects.get(id=account.id)
        self.assertEqual(account.twofactor_attempts, 1)

    def test_reset_twofactor_attempts(self):
        account = self.factory.make_account()
        assert account.twofactor_attempts == 0

        account.increment_twofactor_attempts()
        account.increment_twofactor_attempts()
        account.increment_twofactor_attempts()
        account.increment_twofactor_attempts()
        assert account.twofactor_attempts == 4

        account.reset_twofactor_attempts()

        # reload account
        account = Account.objects.get(id=account.id)
        self.assertEqual(account.twofactor_attempts, 0)

    def test_has_module_perms_when_not_active(self):
        account = self.factory.make_account()
        account.status = AccountStatus.NOACCOUNT

        self.assertFalse(account.has_module_perms('app'))

    def test_has_module_perms_returning_true(self):
        account = self.prepare_perms(True)
        self.assertTrue(account.has_module_perms('app'))

    def test_has_module_perms_returning_false(self):
        account = self.prepare_perms(False)
        self.assertFalse(account.has_module_perms('app'))

    def test_has_perm_returning_true(self):
        account = self.prepare_perms(True)
        self.assertTrue(account.has_perm('can_change_account'))

    def test_has_perm_returning_false(self):
        account = self.prepare_perms(False)
        self.assertFalse(account.has_perm('can_change_account'))

    def test_has_perm_when_account_is_not_active(self):
        account = self.factory.make_account()
        account.status = AccountStatus.NOACCOUNT

        self.assertFalse(account.has_perm('can_change_account'))

    def test_new_account_has_preferred_email_with_flag(self):
        account = self.factory.make_account(email_validated=False)
        self.assertIsNotNone(account.preferredemail)

    def test_set_preferred_with_unverified_email(self):
        account = self.factory.make_account(email_validated=False)
        with self.assertRaises(ValidationError):
            account.preferredemail = account.emailaddress_set.get()

    def test_can_reset_password_without_verified_email(self):
        account = self.factory.make_account(email_validated=False)
        self.assertTrue(account.can_reset_password)

    def test_set_password_validates(self):
        account = self.factory.make_account()
        new_password = 'some NEW password 123'
        original_password = account.accountpassword.password

        validator = Mock()
        with patch('identityprovider.models.account.get_password_validator',
                   return_value=validator) as mock_validate:
            account.set_password(new_password)

        mock_validate.assert_called_once_with(account=account)
        validator.assert_called_once_with(new_password)
        self.assertNotEqual(account.accountpassword.password,
                            original_password)

    def test_sets_password_with_django_cypher(self):
        password = 'Some password from 2013'
        account = self.factory.make_account()

        account.set_password(password)

        self.assertTrue(check_password(password, account.encrypted_password))

    def test_set_password_validation_failure(self):
        account = self.factory.make_account()
        original_password = account.accountpassword.password
        with self.assertRaises(ValidationError):
            account.set_password('short')

        account = Account.objects.get(id=account.id)
        self.assertEqual(account.accountpassword.password, original_password)

    def test_set_password_resets_password_reset_flag(self):
        account = self.factory.make_account()
        account.need_password_reset = True
        account.save()

        account.set_password(DEFAULT_USER_PASSWORD)

        self.assertFalse(account.need_password_reset)

    def test_validated_email_address_is_preferred_email_by_default(self):
        account = self.factory.make_account(email_validated=False)
        assert account.verified_emails().count() == 0
        email_address = account.emailaddress_set.create(
            email='foobar@canonical.com',
            status=EmailStatus.VALIDATED)
        self.assertEqual(account.preferredemail.email, 'foobar@canonical.com')

        email_address = EmailAddress.objects.get(pk=email_address.id)
        self.assertEqual(email_address.status, EmailStatus.PREFERRED)

    def test_account_suspend_reset_password(self):
        account = self.factory.make_account(password=DEFAULT_USER_PASSWORD,
                                            status=AccountStatus.SUSPENDED)

        account_password = AccountPassword.objects.get(account=account)
        self.assertEqual(account_password.password, 'invalid')

    def test_save_when_readonly(self):
        readonly_manager = ReadOnlyManager()
        account = self.factory.make_account()
        assert account.status == AccountStatus.ACTIVE

        assert not settings.READ_ONLY_MODE
        readonly_manager.set_readonly()
        self.addCleanup(readonly_manager.clear_readonly)

        account.suspend()
        # refresh account from db
        account = Account.objects.get(id=account.id)
        self.assertEqual(account.status, AccountStatus.ACTIVE)

    def test_save_when_no_password(self):
        account = self.factory.make_account()
        # remove all passwords
        AccountPassword.objects.filter(account=account).delete()

        # reload account, to bypass django orm cache
        account = Account.objects.get(pk=account.pk)
        # trigger test
        account.status = AccountStatus.SUSPENDED
        account.save()

        self.assertEqual(
            AccountPassword.objects.filter(account=account).count(), 0)

    def test_sites_with_active_sessions_when_returns_empty_result(self):
        account = self.factory.make_account()
        sites_count = account.sites_with_active_sessions().count()
        self.assertEqual(sites_count, 0)

    def test_sites_with_active_sessions_when_one_site_is_active(self):
        account = self.factory.make_account()

        account.openidrpsummary_set.create(
            date_last_used=now(),
            trust_root='http://trust-root.example.com',
            openid_identifier='openid-identifier')

        account.openidrpsummary_set.create(
            date_last_used=now() - timedelta(days=3),
            trust_root='http://trust-root-2.example.com')

        sites_count = account.sites_with_active_sessions().count()
        self.assertEqual(sites_count, 1)

    @patch('identityprovider.models.account.now')
    def test_sites_with_active_sessions_use_utc(self, mock_now):
        mock_now.return_value = now()

        account = self.factory.make_account()
        sites_count = account.sites_with_active_sessions().count()

        self.assertEqual(sites_count, 0)
        self.assertEqual(mock_now.called, True)

    def test_warn_about_backup_device(self):
        account = self.factory.make_account()
        # use assertEqual instead to assertTrue to ensure attr is a boolean
        self.assertEqual(account.warn_about_backup_device, True)

    def test_need_backup_device_warning_no_devices(self):
        account = self.factory.make_account()
        assert account.warn_about_backup_device
        assert account.devices.count() == 0
        self.assertFalse(account.need_backup_device_warning)

    def test_need_backup_device_warning_one_device(self):
        account = self.factory.make_account()
        self.factory.make_device(account=account)

        assert account.warn_about_backup_device
        assert account.devices.count() == 1
        self.assertTrue(account.need_backup_device_warning)

    def test_need_backup_device_warning_one_device_warn_setting_false(self):
        account = self.factory.make_account()
        account.warn_about_backup_device = False
        account.save()
        self.factory.make_device(account=account)

        assert not account.warn_about_backup_device
        assert account.devices.count() == 1
        self.assertFalse(account.need_backup_device_warning)

    def test_need_backup_device_warning_two_devices(self):
        account = self.factory.make_account()
        self.factory.make_device(account=account)
        self.factory.make_device(account=account)

        assert account.warn_about_backup_device
        assert account.devices.count() == 2
        self.assertFalse(account.need_backup_device_warning)

    def test_deactivate(self):
        account = self.factory.make_account()
        assert account.status != AccountStatus.DEACTIVATED

        account.deactivate()
        self.assertEqual(account.status, AccountStatus.DEACTIVATED)

        # the status change was also saved in the DB
        account = Account.objects.get(id=account.id)
        self.assertEqual(account.status, AccountStatus.DEACTIVATED)

    def test_suspend(self):
        account = self.factory.make_account()
        assert account.status != AccountStatus.SUSPENDED

        account.suspend()
        self.assertEqual(account.status, AccountStatus.SUSPENDED)

        # the status change was also saved in the DB
        account = Account.objects.get(id=account.id)
        self.assertEqual(account.status, AccountStatus.SUSPENDED)

    def test_deleted_status_deletes_account(self):
        account = self.factory.make_account()
        # create some AuthTokens
        self.factory.make_account_token(account)
        self.factory.make_account_token(account)

        account.status = AccountStatus.DELETED
        account.save()

        account = Account.objects.get(id=account.id)
        self.assertEqual(account.accountpassword.password, 'invalid')
        self.assertFalse(account.is_active)
        self.assertEqual(account.emails().count(), 0)
        self.assertEqual(account.authtoken_set.count(), 0)

        user = User.objects.get(username=account.openid_identifier)
        self.assertFalse(user.is_active)

    def test_set_deleted_no_user(self):
        account = self.factory.make_account()
        account.user.delete()
        assert User.objects.filter(
            username=account.openid_identifier).count() == 0

        account.set_deleted()
        self.assertEqual(account.status, AccountStatus.DELETED)

        account = Account.objects.get(id=account.id)
        self.assertEqual(account.accountpassword.password, 'invalid')
        self.assertFalse(account.is_active)
        self.assertEqual(account.emails().count(), 0)

        user = User.objects.get(username=account.openid_identifier)
        self.assertFalse(user.is_active)

    def test_set_deleted_clears_tokens(self):
        account = self.factory.make_account()

        # creating oauth token also creates the User for the account
        account.create_oauth_token('foo', scope=TokenScope.VERSION_2)
        account.create_oauth_token('bar', scope=TokenScope.WEB_LOGIN)
        account.create_oauth_token('foo old', scope=None)

        account.set_deleted()

        user = User.objects.get(username=account.openid_identifier)
        self.assertFalse(user.is_active)

        consumers = OAuthConsumer.objects.filter(
            user__username=account.openid_identifier)
        self.assertEqual(consumers.count(), 0)

        tokens = OAuthToken.objects.filter(
            consumer__user__username=account.openid_identifier)
        self.assertEqual(tokens.count(), 0)

    def test_reset_password_default_email(self):
        account = self.factory.make_account()

        name = 'identityprovider.emailutils.send_password_reset_email'
        with patch(name) as mock_send_email:
            requested = account.reset_password()

        self.assertTrue(requested)
        mock_send_email.assert_called_once_with(
            account, account.preferredemail.email, reason='password_leak')
        account = Account.objects.get(id=account.id)
        self.assertTrue(account.need_password_reset)

    def test_reset_password_default_reason(self):
        account = self.factory.make_account()
        email = account.preferredemail.email

        name = 'identityprovider.emailutils.send_password_reset_email'
        with patch(name) as mock_send_email:
            requested = account.reset_password(email)

        self.assertTrue(requested)
        mock_send_email.assert_called_once_with(
            account, email, reason='password_leak')
        account = Account.objects.get(id=account.id)
        self.assertTrue(account.need_password_reset)

    def test_reset_password_custom_reason(self):
        account = self.factory.make_account()
        email = account.preferredemail.email

        name = 'identityprovider.emailutils.send_password_reset_email'
        with patch(name) as mock_send_email:
            requested = account.reset_password(email, reason='forgot_password')

        self.assertTrue(requested)
        mock_send_email.assert_called_once_with(
            account, email, reason='forgot_password')
        account = Account.objects.get(id=account.id)
        self.assertTrue(account.need_password_reset)

    def test_reset_password_with_invalid_email(self):
        account = self.factory.make_account()

        requested = account.reset_password('invalid@foo.com')
        self.assertFalse(requested)
        account = Account.objects.get(id=account.id)
        self.assertFalse(account.need_password_reset)

    def test_person_name(self):
        account = self.factory.make_account()
        person = self.factory.make_person(account=account)
        self.assertEqual(person.name, account.person_name)

    def test_person_name_no_person(self):
        account = self.factory.make_account()
        self.assertIsNone(account.person_name)

    def test_person_name_need_refresh(self):
        lp = MockLaunchpad()
        self.patch(
            'identityprovider.models.account.get_lp_username',
            new=lp.get_lp_username)

        account = self.factory.make_account()
        self.assertIsNone(account.person_name)

        # SSO doesn't pick up LP username changes immediately due to
        # replication lag.
        lp.set_lp_username(account.openid_identifier, 'foo')
        del account.person
        del account.person_name
        self.assertIsNone(account.person_name)

        # But if the need_username_refresh flag is set, person_name knows to
        # check with the LP API directly.
        account.need_username_refresh = True
        del account.person
        del account.person_name
        self.assertEqual(account.person_name, 'foo')
        self.assertTrue(account.need_username_refresh)

        # Once the change replicates, need_username_refresh is unset when
        # person_name is read.
        lp.flush_to_db()
        del account.person
        del account.person_name
        self.assertTrue(account.need_username_refresh)
        self.assertEqual(account.person_name, 'foo')
        self.assertFalse(account.need_username_refresh)

        del account.person
        del account.person_name
        self.assertEqual(account.person_name, 'foo')


class LogISDAccountDeletionTestCase(LogHandlerTestCase):

    """Test for isd deletion logging, this is a temporary class."""

    del_message = "Deleted isd test account"

    def test_no_log_on_delete_normal_account(self):
        account = Account.objects.create_account(
            displayname='John Doe', email_address='email@host.local',
            password=None)
        account.delete()
        # Ensure there's no log of this
        log_messages = [self.del_message in r.getMessage()
                        for r in self.memento_handler.records]
        self.assertNotIn(True, log_messages)

    def test_log_on_delete_isdtest(self):
        account = Account.objects.create_account(
            displayname='ISD Test', email_address='isdtest@canonical.com',
            password=None)
        account.delete()
        log_messages = [r.getMessage() for r in self.memento_handler.records]

        # Ensure we logged the deletion
        self.assertIn(True, [self.del_message in msg for msg in log_messages])
        # Ensure the traceback at least contains the test module's name
        self.assertIn(
            True, ["test_models_account.py" in msg for msg in log_messages])


class AccountMustRequireTwofactorTestCase(SSOBaseTestCase):

    def assert_twofactor_settings_unchanged(self, account):
        reloaded = Account.objects.get(pk=account.pk)
        for i in ('twofactor_required', 'warn_about_backup_device'):
            self.assertEqual(getattr(account, i), getattr(reloaded, i))

    def assert_twofactor_settings_all_true(self, account):
        reloaded = Account.objects.get(pk=account.pk)
        for i in ('twofactor_required', 'warn_about_backup_device'):
            self.assertTrue(getattr(account, i))
            self.assertTrue(getattr(reloaded, i))

    def test_must_not_require_twofactor_no_team_in_settings(self):
        account = self.factory.make_account()

        with self.settings(TWOFACTOR_MANDATORY_TEAMS=[]):
            self.assertFalse(account.must_require_twofactor())

        self.assert_twofactor_settings_unchanged(account)

    def test_must_not_require_twofactor_not_in_team(self):
        account = self.factory.make_account()
        team = self.factory.make_team()
        assert not account.person_in_team(team)

        with self.settings(TWOFACTOR_MANDATORY_TEAMS=[team.name]):
            self.assertFalse(account.must_require_twofactor())

        self.assert_twofactor_settings_unchanged(account)

    def test_must_require_twofactor_when_in_team(self):
        account = self.factory.make_account(warn_about_backup_device=False)
        assert not account.twofactor_required
        assert not account.warn_about_backup_device

        team = self.factory.make_team()
        self.factory.add_account_to_team(account, team)
        assert account.person_in_team(team)

        with self.settings(TWOFACTOR_MANDATORY_TEAMS=[team.name]):
            self.assertTrue(account.must_require_twofactor())

        self.assert_twofactor_settings_all_true(account)


class AccountTeamMembershipTestCase(SSOBaseTestCase):

    def _assert_in_team(self, account, expected, method, **kwargs):
        # reload account in case person is cached
        account = Account.objects.get(pk=account.pk)

        with self.assertNumQueries(1):
            result = method(**kwargs)

        self.assertEqual(result, expected)

    def assert_in_team(self, account, team):
        self._assert_in_team(
            account, expected=True, method=account.person_in_team, team=team)

    def assert_not_in_team(self, account, team):
        self._assert_in_team(
            account, expected=False, method=account.person_in_team, team=team)

    def assert_in_teams(self, account, teams):
        self._assert_in_team(
            account, expected=True, method=account.person_in_any_team,
            teams=teams)

    def assert_not_in_teams(self, account, teams):
        self._assert_in_team(
            account, expected=False, method=account.person_in_any_team,
            teams=teams)

    def test_no_person_queries(self):
        account = self.factory.make_account()

        with self.assertNumQueries(1):
            self.assertIsNone(account.person)

    def test_with_person_queries(self):
        account = self.factory.make_account()
        person = self.factory.make_person(account=account)
        # ensure the person is not cached in the instance
        account = Account.objects.get(pk=person.account.pk)

        with self.assertNumQueries(1):
            self.assertEqual(account.person, person)

    def test_no_person_with_team_name(self):
        account = self.factory.make_account()
        self.assert_not_in_team(account, 'foo')

    def test_no_person_with_team_object(self):
        account = self.factory.make_account()
        self.assert_not_in_team(account, self.factory.make_team(name='foo'))

    def test_person_with_team_name(self):
        account = self.factory.make_account(teams=['foo'])
        self.assert_in_team(account, 'foo')

    def test_person_with_team_object(self):
        team = self.factory.make_team(name='foo')
        account = self.factory.make_account(teams=[team.name])
        self.assert_in_team(account, team)

    def test_person_with_many_teams(self):
        account = self.factory.make_account(teams=['foo', 'bar', 'baz'])
        self.assert_in_team(account, 'foo')
        self.assert_in_team(account, 'bar')
        self.assert_in_team(account, 'baz')

    def test_person_with_team_prefix(self):
        account = self.factory.make_account(teams=['foo'])
        self.assert_not_in_team(account, 'fooo')
        self.assert_not_in_team(account, 'fo')

    def test_other_team_name(self):
        self.factory.make_team(name='bar')
        account = self.factory.make_account(teams=['foo'])
        self.assert_not_in_team(account, 'bar')

    def test_other_team_object(self):
        team = self.factory.make_team(name='bar')
        account = self.factory.make_account(teams=['foo'])
        self.assert_not_in_team(account, team)

    def test_other_unexisting_team(self):
        account = self.factory.make_account(teams=['foo'])
        self.assert_not_in_team(account, 'bar')

    def test_other_user_in_desired_team(self):
        account = self.factory.make_account(teams=['foo'])
        other = self.factory.make_account(teams=['bar'])

        self.assert_not_in_team(account, 'bar')
        self.assert_in_team(account, 'foo')
        self.assert_in_team(other, 'bar')
        self.assert_not_in_team(other, 'foo')

    def test_no_membership_but_team_ownership(self):
        account = self.factory.make_account(teams=['foo'])
        self.factory.make_team(name='bar', owner=account.person)

        self.assert_in_team(account, 'foo')
        self.assert_in_team(account, 'bar')

    def test_many_teams(self):
        account = self.factory.make_account(teams=['foo', 'bar', 'baz'])

        self.assert_in_teams(account, ['foo', 'bar', 'baz'])
        self.assert_in_teams(account, ['foo'])
        self.assert_in_teams(account, ['zaraza', 'foo', 'other'])
        self.assert_not_in_teams(account, ['zaraza', 'fooo', 'other'])

    def test_many_teams_including_ownership(self):
        account = self.factory.make_account(teams=['foo', 'baz'])
        assert account.person is not None
        self.factory.make_team(name='bar', owner=account.person)

        self.assert_in_teams(account, ['foo'])
        self.assert_in_teams(account, ['foo', 'bar'])
        self.assert_in_teams(account, ['bar'])

    def test_many_teams_many_ownerships(self):
        account = self.factory.make_account()
        person = self.factory.make_person(account=account)
        assert account.person == person
        for i in xrange(3):
            self.factory.make_team(name='team-%s' % i, owner=account.person)
            self.factory.make_team(name='other-team-%s' % i)

        # need a second query to check team ownership
        self.assert_in_teams(account, ['team-1'])
        # but no extra query for more than one team
        self.assert_in_teams(account, ['team-1', 'team-2'])

    def test_many_teams_performance(self):
        accounts = []
        teams = []
        for i in xrange(3):
            teams.append('team-%i' % i)
            for j in xrange(10):
                accounts.append(self.factory.make_account(teams=teams))

        openid = accounts[0].openid_identifier
        cursor = connection.cursor()
        cursor.execute(
            "EXPLAIN ANALYZE " + Account.OWNS_OR_IN_ANY_TEAM, [teams, openid])
        row = cursor.fetchone()
        # This is not orthodox but is the only way to ensure
        # Account.OWNS_OR_IN_ANY_TEAM does not perform a query that will
        # include a gazillion rows in the nested join loop.
        self.assertIn(' rows=1 width=4', unicode(row))


class AccountEmailsTestCase(SSOBaseTestCase):

    def setUp(self):
        super(AccountEmailsTestCase, self).setUp()
        self.account = self.factory.make_account()

    def test_limit_emails(self):
        assert self.account.emails().count() == 1

        expected = [self.account.preferredemail]
        for i in xrange(10):
            date_created = now() + timedelta(seconds=i)
            email = EmailAddress.objects.create(
                account=self.account, email='emails-test-%s@example.com' % i,
                status=EmailStatus.NEW if i % 2 else EmailStatus.VALIDATED,
                date_created=date_created)
            expected.append(email)

        expected.reverse()

        # no limit should return all
        emails = self.account.emails()
        self.assertEqual(emails.count(), 11)
        self.assertEqual(list(emails), expected)

        # limit should return the latest 5
        emails = self.account.emails(limit=5)
        self.assertEqual(emails.count(), 5)
        self.assertEqual(list(emails), expected[:5])

    def test_verified_emails_has_preferred_first(self):
        account = self.factory.make_account()
        self.factory.make_email_for_account(account,
                                            status=EmailStatus.VALIDATED)
        emails = account.verified_emails()
        self.assertTrue(len(emails) > 1)
        self.assertEqual(emails[0].status, EmailStatus.PREFERRED)
        self.assertEqual(emails[1].status, EmailStatus.VALIDATED)


class AccountPasswordTestCase(SSOBaseTestCase):

    def test_unicode(self):
        account = self.factory.make_account(
            openid_identifier='oid', displayname='displayname')
        self.assertEqual(
            unicode(account.accountpassword), u'Password for displayname')

    def test_last_leak_check_correct_on_new_ap(self):
        account = Account(openid_identifier='oid', displayname='displayname')
        password = AccountPassword(account=account, password='password')
        self.assertIsNone(password.last_leak_check)

    def test_last_leak_check_correct_on_account_creation(self):
        self.factory.make_account()
        # the AP last_leak_check is set to this epoch as part of
        # the save/update process.
        self.assertEqual(AccountPassword.objects.all()[0].last_leak_check,
                         AccountPassword.LEAK_REFERENCE_DATE)

    def test_can_set_last_leak_check(self):
        account = self.factory.make_account()
        the_time = now()
        account.accountpassword.last_leak_check = the_time
        account.accountpassword.save(update_fields=['last_leak_check'])
        self.assertEqual(
            the_time, AccountPassword.objects.all()[0].last_leak_check)

    def test_last_leak_check_reset_on_save(self):
        account = self.factory.make_account()
        # Last leak checked now
        account.accountpassword.last_leak_check = now()
        account.accountpassword.save(update_fields=['last_leak_check'])
        # Now the password is changed
        account.accountpassword.password = "Somethingelse"
        account.accountpassword.save()
        self.assertEqual(
            AccountPassword.objects.all()[0].last_leak_check,
            AccountPassword.LEAK_REFERENCE_DATE)

    def test_changed_password(self):
        account = self.factory.make_account()

        # get timestamp from creation
        created_at = account.accountpassword.date_changed

        # change the password
        account.accountpassword.password = "Somethingelse"
        account.accountpassword.save()

        # verify!
        self.assertGreater(account.accountpassword.date_changed, created_at)


class CreateOAuthTokenForAccountTestCase(SSOBaseTestCase):

    def setUp(self):
        super(CreateOAuthTokenForAccountTestCase, self).setUp()
        self.account = self.factory.make_account(
            displayname='test', email='test-oauth@example.com',
            password=DEFAULT_USER_PASSWORD)
        self.user = self.account.user

    def test_when_account_has_associated_consumer(self):
        consumer, _ = OAuthConsumer.objects.get_or_create(user=self.user)

        token = self.account.create_oauth_token('new-token', scope=None)

        self.assertEqual(token.consumer.id, consumer.id)

    def test_consumer_with_scope(self):
        token = self.account.create_oauth_token(
            'new-token', scope=TokenScope.VERSION_2)

        self.assertEqual(token.consumer.id, self.account.id)

    def test_get_or_create_no_scope(self):
        consumer, _ = OAuthConsumer.objects.get_or_create(user=self.user)

        token, created = self.account.get_or_create_oauth_token(
            'new-token', scope=None)

        self.assertTrue(created)
        self.assertEqual(token.consumer.id, consumer.id)

        token2, created = self.account.get_or_create_oauth_token(
            'new-token', scope=None)
        self.assertFalse(created)
        self.assertEqual(token2, token)
        self.assertNotEqual(token2.updated_at, token.updated_at)

        token3, created = self.account.get_or_create_oauth_token(
            'new-token2', scope=None)
        self.assertTrue(created)
        self.assertNotEqual(token, token3)
        self.assertEqual(token3.consumer.id, consumer.id)

    def test_get_or_create(self):
        token, created = self.account.get_or_create_oauth_token(
            'new-token', scope=TokenScope.VERSION_2)

        self.assertTrue(created)
        self.assertEqual(token.consumer, self.account)

        token2, created = self.account.get_or_create_oauth_token(
            'new-token', scope=TokenScope.VERSION_2)
        self.assertFalse(created)
        self.assertEqual(token2, token)
        self.assertGreaterEqual(token2.date_updated, token.date_updated)

        token3, created = self.account.get_or_create_oauth_token(
            'new-token-new', scope=TokenScope.VERSION_2)
        self.assertTrue(created)
        self.assertNotEqual(token, token3)
        self.assertEqual(token3.consumer, self.account)

    def test_get_or_create_when_multiple_results(self):
        consumer, _ = OAuthConsumer.objects.get_or_create(user=self.user)

        # create multiple tokens
        token1 = consumer.token_set.create(name='new-token')
        token2 = consumer.token_set.create(name='new-token')
        token2.created_at = token1.created_at + timedelta(seconds=10)
        token2.save()

        token, created = self.account.get_or_create_oauth_token(
            'new-token', scope=None)

        self.assertEqual(token, token2)
        self.assertFalse(created)

    def test_get_or_create_existing_token_name_with_scope(self):
        # try to create multiple tokens
        token, created = self.account.get_or_create_oauth_token(
            'new-token', scope=TokenScope.VERSION_2)
        self.assertEqual(token.consumer, self.account)
        self.assertTrue(created)

        token2, created = self.account.get_or_create_oauth_token(
            'new-token', scope=TokenScope.VERSION_2)
        self.assertEqual(token, token2)
        self.assertFalse(created)

    def test_get_or_create_existing_token_name_with_different_scope(self):
        token, created = self.account.get_or_create_oauth_token(
            'new-token', scope=TokenScope.VERSION_2)
        self.assertEqual(token.consumer, self.account)
        self.assertTrue(created)

        token2, created = self.account.get_or_create_oauth_token(
            'new-token', scope=TokenScope.WEB_LOGIN)
        self.assertNotEqual(token, token2)
        self.assertEqual(token.consumer, self.account)
        self.assertEqual(token2.name, token.name + '1')
        self.assertTrue(created)

    def test_get_when_account_has_no_associated_consumer(self):
        OAuthConsumer.objects.filter(user=self.user).delete()

        token = self.account.create_oauth_token(
            'new-token', scope=None)

        self.assertEqual(token.consumer.user.id, self.user.id)

    def test_get_or_create_when_account_has_no_associated_consumer(self):
        OAuthConsumer.objects.filter(user=self.user).delete()

        token, created = self.account.get_or_create_oauth_token(
            'new-token', scope=None)

        self.assertTrue(created)
        self.assertEqual(token.consumer.user.id, self.user.id)

    def test_invalidate_oauth_tokens(self):
        self.account.create_oauth_token('foo', scope=TokenScope.VERSION_2)
        self.account.create_oauth_token('bar', scope=TokenScope.WEB_LOGIN)
        self.account.create_oauth_token('foo old', scope=None)

        assert self.account.oauth_tokens(scope=None).count() > 0
        assert self.account.oauth_tokens(
            scope=TokenScope.VERSION_2).count() > 0
        assert self.account.oauth_tokens(
            scope=TokenScope.WEB_LOGIN).count() > 0

        self.account.invalidate_oauth_tokens()

        self.assertEqual(self.account.oauth_tokens(scope=None).count(), 0)
        self.assertEqual(
            self.account.oauth_tokens(scope=TokenScope.VERSION_2).count(), 0)
        self.assertEqual(
            self.account.oauth_tokens(scope=TokenScope.WEB_LOGIN).count(), 0)

    def test_limit_oauth_tokens_no_scope(self):
        assert self.account.oauth_tokens(scope=None).count() == 0

        expected = []
        for i in xrange(10):
            token, _ = self.account.get_or_create_oauth_token(
                'token-%s' % i, scope=None)
            token.updated_at = token.created_at + timedelta(seconds=i)
            token.save()
            expected.append(token)

        expected.reverse()

        assert self.account.oauth_tokens(
            scope=TokenScope.VERSION_2).count() == 0, 'No v2 tokens'

        # no limit should return all
        tokens = self.account.oauth_tokens(scope=None)
        self.assertEqual(tokens.count(), 10)
        self.assertEqual(list(tokens), expected)

        # limit should return the latest 5
        tokens = self.account.oauth_tokens(limit=5, scope=None)
        self.assertEqual(tokens.count(), 5)
        self.assertEqual(list(tokens), expected[:5])

    def test_limit_oauth_tokens(self):
        assert self.account.oauth_tokens(
            scope=TokenScope.VERSION_2).count() == 0

        expected = []
        for i in xrange(10):
            token = self.account.create_oauth_token(
                'token-%s' % i, scope=TokenScope.VERSION_2)
            token.date_updated = token.date_created + timedelta(seconds=i)
            token.save()
            expected.append(token)

        expected.reverse()

        # no limit should return all
        tokens = self.account.oauth_tokens(scope=TokenScope.VERSION_2)
        self.assertEqual(tokens.count(), 10)
        self.assertEqual(list(tokens), expected)

        # limit should return the latest 5
        tokens = self.account.oauth_tokens(limit=5, scope=TokenScope.VERSION_2)
        self.assertEqual(tokens.count(), 5)
        self.assertEqual(list(tokens), expected[:5])

    def test_oauth_token_by_key(self):
        tokens = {}
        for i in xrange(5):
            token = self.account.create_oauth_token(
                'token-%s' % i, scope=TokenScope.VERSION_2)
            tokens[token.key] = token

        for key, token in tokens.iteritems():
            self.assertEqual(self.account.oauth_token_by_key(key), token)

    def test_oauth_token_by_key_old_token(self):
        tokens = {}
        for i in xrange(5):
            token = self.account.create_oauth_token('token-%s' % i, scope=None)
            tokens[token.key] = token

        for key, token in tokens.iteritems():
            self.assertEqual(self.account.oauth_token_by_key(key), token)

    def test_oauth_token_by_key_no_token(self):
        token = self.account.create_oauth_token(
            'token', scope=TokenScope.VERSION_2)
        token_key = token.key
        token.delete()

        self.assertIsNone(self.account.oauth_token_by_key(token_key))

    def test_oauth_token_by_key_not_owner_no_scope(self):
        other = self.factory.make_account()
        other_token = other.create_oauth_token('other', scope=None)

        self.assertEqual(
            other.oauth_token_by_key(other_token.key), other_token)
        self.assertIsNone(self.account.oauth_token_by_key(other_token.key))

    def test_oauth_token_by_key_not_owner_with_scope(self):
        other = self.factory.make_account()
        other_token = other.create_oauth_token(
            'other', scope=TokenScope.VERSION_2)

        self.assertEqual(
            other.oauth_token_by_key(other_token.key), other_token)
        self.assertIsNone(self.account.oauth_token_by_key(other_token.key))


class AuthenticateUserTestCase(SSOBaseTestCase):

    def setUp(self):
        super(AuthenticateUserTestCase, self).setUp()
        self.email = 'test@example.com'
        self.password = DEFAULT_USER_PASSWORD
        self.account = self.factory.make_account(
            email=self.email, password=self.password)
        self.user = self.account.user

    def assert_failed_login(self, exception):
        with self.assertRaises(exception):
            authenticate_user(self.email, self.password)

    def test_account_suspended(self):
        self.account.status = AccountStatus.SUSPENDED
        self.account.save()
        self.assert_failed_login(AccountSuspended)

    def test_account_deactivated(self):
        self.account.status = AccountStatus.DEACTIVATED
        self.account.save()
        self.assert_failed_login(AccountDeactivated)

    def test_failed_login(self):
        self.password = 'wrong password'
        self.assert_failed_login(AuthenticationError)

    def test_new_email_allowed(self):
        email = self.account.emailaddress_set.get(email=self.email)
        email.status = EmailStatus.NEW
        email.save()

        result = authenticate_user(self.email, self.password)

        self.assertEqual(result, self.account)

    def test_authenticate_user(self):
        result = authenticate_user(self.email, self.password)
        self.assertEqual(result, self.account)

    def test_invalidated_email(self):
        email_obj = self.account.emailaddress_set.get(email=self.email)
        email_obj.invalidate()

        # this is not distinguishable if another user registered with
        # the same email from a password mismatch
        with self.assertRaises(AuthenticationError):
            authenticate_user(self.email, self.password)

    def test_password_policy_error(self):
        unicode_error_message = u'FAIL\xac'
        validator = Mock(side_effect=ValidationError(unicode_error_message))

        with patch('identityprovider.login.get_password_validator',
                   return_value=validator) as mock_get_validator:
            with self.assertRaises(PasswordPolicyError) as ctx:
                authenticate_user(self.email, self.password)

        exception = ctx.exception
        self.assertEqual(unicode(exception), unicode_error_message)
        self.assertEqual(exception.account, self.account)

        mock_get_validator.assert_called_once_with(account=self.account,
                                                   request=None)
        validator.assert_called_once_with(self.password)

    def test_account_need_password_reset(self):
        self.account.need_password_reset = True
        self.account.save()

        with self.assertRaises(PasswordPolicyError) as ctx:
            authenticate_user(self.email, self.password)

        self.assertIn('there was a security breach', str(ctx.exception))
        self.assertEqual(ctx.exception.account, self.account)

    def test_non_account_instance(self):
        user = User.objects.create_user(username='test', password='12345678')

        with self.assertRaises(AuthenticationError):
            authenticate_user(user.username, '12345678')
