# Copyright 2010 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

from django.contrib.auth.hashers import get_hasher
from django.contrib.auth.models import make_password, check_password

from identityprovider.models.api import APIUser
from identityprovider.utils import (
    encrypt_launchpad_password,
    generate_salt,
    validate_launchpad_password,
)
from identityprovider.tests.utils import SSOBaseTestCase


class APIUserTestCase(SSOBaseTestCase):

    def setUp(self):
        super(APIUserTestCase, self).setUp()

        hasher = get_hasher()

        self.salt = hasher.salt()
        self.user = APIUser(username='username')
        self.user.set_password('password', salt=self.salt)
        self.user.save()

    def test_set_password(self):
        self.user.set_password('otherpassword', salt=self.salt)
        expected = make_password('otherpassword', salt=self.salt)
        self.assertEqual(self.user.password, expected)

    def test_verify_password(self):
        self.user.set_password('password', salt=self.salt)
        self.assertFalse(self.user._verify_password('otherpassword'))
        self.assertTrue(self.user._verify_password('password'))

    def test_verify_password_with_launchpad(self):
        """Ensure backwards-compatibility with Launchpad cypher."""
        self.user.password = encrypt_launchpad_password(
            'password', salt=generate_salt())
        self.assertFalse(self.user._verify_password('otherpassword'))
        self.assertTrue(self.user._verify_password('password'))

    def test_changes_cypher_after_verifying_with_launchpad(self):
        """Change to the Django cypher when needed."""
        self.user.password = encrypt_launchpad_password(
            'password', salt=generate_salt())
        self.user.save()
        self.assertFalse(check_password('password', self.user.password))
        self.assertTrue(validate_launchpad_password('password',
                                                    self.user.password))

        # This verification should trigger a change in the password in
        # the model instance itself.
        self.user._verify_password('password')

        user = APIUser.objects.get(username=self.user.username)
        self.assertTrue(check_password('password', user.password))
        self.assertFalse(validate_launchpad_password('password',
                                                     user.password))

    def test_verify_password_with_django(self):
        self.user.password = make_password('password')
        self.assertFalse(self.user._verify_password('otherpassword'))
        self.assertTrue(self.user._verify_password('password'))

    def test_authenticate_user_exists(self):
        user = APIUser.authenticate('username', 'password')
        self.assertEqual(user, self.user)

    def test_authenticate_wrong_password(self):
        user = APIUser.authenticate('username', 'otherpassword')
        self.assertEqual(user, None)

    def test_authenticate_user_does_not_exist(self):
        user = APIUser.authenticate('bad_user', 'password')
        self.assertEqual(user, None)

    def test_unicode(self):
        self.assertEqual(unicode(self.user), u'username')
