from django.db import DataError
from django.utils.timezone import now

from identityprovider.models.authlog import AuthLog
from identityprovider.models.const import AuthLogType
from identityprovider.tests import DEFAULT_USER_PASSWORD
from identityprovider.tests.utils import SSOBaseTestCase


class AuthLogTestCase(SSOBaseTestCase):

    login_email = "test@canonical.com"
    login_password = DEFAULT_USER_PASSWORD

    def setUp(self):
        super(AuthLogTestCase, self).setUp()
        self.account = self.factory.make_account(email=self.login_email,
                                                 password=self.login_password)
        self.log_types = AuthLogType._get_choices()

    def test_create_success(self):
        """Successful login attempt should create record with account link.

        It also includes an e-mail address.
        """
        fields = {
            'log_type': AuthLogType.OPENID_NEW,
            'when_created': now(),
            'account': self.account,
            'login_email': self.account.preferredemail,
            'referer': "http://www.example.com/referred_from",
            'remote_ip': '1.2.3.4',
            'user_agent': 'Godzilla 1.0',
        }
        log_item = AuthLog.objects.create(**fields)
        for field, value in fields.items():
            self.assertEqual(getattr(log_item, field), value)

    def test_create_existing_success_with_no_email(self):
        """Successful login attempt should create record with account link.

        This checks the case where the attempt does NOT have a login_email
        value (which can happen if it's a repeat OpenID request where we don't
        get the email and password again).
        """
        fields = {
            'log_type': AuthLogType.OPENID_NEW,
            'when_created': now(),
            'account': self.account,
            'referer': "http://www.example.com/referred_from",
            'remote_ip': '1.2.3.4',
            'user_agent': 'Godzilla 1.0',
        }
        log_item = AuthLog.objects.create(**fields)
        for field, value in fields.items():
            self.assertEqual(getattr(log_item, field), value)
        # Ensure login_email has the correct default, since we didn't send it
        # explicitly.
        self.assertEqual(log_item.login_email, '')

    def test_create_failure(self):
        """Failed login attempt should record requested username."""
        fields = {
            'log_type': AuthLogType.OPENID_FAIL,
            'when_created': now(),
            'login_email': self.login_email,
            'referer': "http://www.example.com/referred_from",
            'remote_ip': '1.2.3.4',
            'user_agent': 'Godzilla 1.0',
        }
        log_item = AuthLog.objects.create(**fields)
        for field, value in fields.items():
            self.assertEqual(getattr(log_item, field), value)

    def test_create_ipv6(self):
        """IPv6 remote_ip should be saved correctly."""
        fields = {
            'log_type': AuthLogType.OPENID_FAIL,
            'when_created': now(),
            'login_email': self.login_email,
            'referer': "http://www.example.com/referred_from",
            'remote_ip': '2001:0db8:85a3:0000:0000:8a2e:0370:7334',
            'user_agent': 'Godzilla 1.0',
        }
        log_item = AuthLog.objects.create(**fields)
        for field, value in fields.items():
            self.assertEqual(getattr(log_item, field), value)

    def test_create_invalid_ip_fails(self):
        """Invalid remote_ip should give DataError."""
        fields = {
            'log_type': AuthLogType.OPENID_FAIL,
            'when_created': now(),
            'login_email': self.login_email,
            'referer': "http://www.example.com/referred_from",
            'remote_ip': 'this is not a valid IP Address',
            'user_agent': 'Godzilla 1.0',
        }
        with self.assertRaises(DataError):
            AuthLog.objects.create(**fields)

    def check_log_string(self, log_type, string_component):
        fields = {
            'log_type': log_type,
            'when_created': now(),
            'account': self.account,
            'referer': "http://www.example.com/referred_from",
            'remote_ip': '1.2.3.4',
            'user_agent': 'Godzilla 1.0',
        }
        log_item = AuthLog(**fields)
        self.assertEqual(
            unicode(log_item), u'%s: %s (1.2.3.4)' % (
                string_component, self.account))

    def test_expected_log_unicodes(self):
        for log_type, string_component in (
                (AuthLogType.OPENID_NEW, "OPENID_NEW"),
                (AuthLogType.OPENID_EXISTING, "OPENID_EXISTING"),
                (AuthLogType.OPENID_FAIL, "OPENID_FAIL"),
                (AuthLogType.API_NEW, "API_NEW"),
                (AuthLogType.API_EXISTING, "API_EXISTING"),
                (AuthLogType.API_FAIL, "API_FAIL"),
        ):
            self.check_log_string(log_type, string_component)

    def test_log_labels(self):
        expected_labels = {
            AuthLogType.OPENID_FAIL: "Failed web login",
            AuthLogType.OPENID_NEW: "New web login",
            AuthLogType.OPENID_EXISTING: "Existing web login",
            AuthLogType.API_FAIL: "Failed API login",
            AuthLogType.API_NEW: "New API login",
            AuthLogType.API_EXISTING: "Existing API login"
        }
        for log_type, label in expected_labels.items():
            self.assertEqual(AuthLogType._label_for(log_type), label)

    def test_log_item_labels(self):
        fields = {
            'when_created': now(),
            'account': self.account,
            'referer': "http://www.example.com/referred_from",
            'remote_ip': '1.2.3.4',
            'user_agent': 'Godzilla 1.0',
        }
        for log_type_id, log_type_str in AuthLogType._get_choices():
            fields['log_type'] = log_type_id
            log_item = AuthLog(**fields)
            self.assertEqual(
                log_item.log_type_label(), AuthLogType._label_for(log_type_id))
