# coding: utf-8
# Copyright 2010 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

import random

from django.core.exceptions import MultipleObjectsReturned, ValidationError
from django.http import Http404
from django.conf import settings
from django.core.urlresolvers import reverse
from django.test.utils import override_settings
from django.utils.timezone import now
from mock import patch

from identityprovider.models import (
    AuthToken,
    EmailAddress,
    authtoken,
    verify_token_string
)
from identityprovider.models.const import AuthTokenType
from identityprovider.tests.utils import SSOBaseTestCase
from identityprovider.utils import generate_random_string


class AuthTokenManagerTestCase(SSOBaseTestCase):

    def test_valid_email(self):
        emails = {'valid': ['kiko.async@hotmail.com', 'kiko+async@hotmail.com',
                            'kiko-async@hotmail.com', 'kiko_async@hotmail.com',
                            'kiko@async.com.br', 'kiko@canonical.com',
                            'kiko@UBUNTU.COM', 'kiko@gnu.info',
                            'user@z.de', 'bob=dobbs@example.com',
                            'keith@risby-family.co.uk'],
                  'invalid': ['user@z..de', 'user@.z.de', 'i@tv',
                              'keith@risby-family-.co.uk',
                              'keith@-risby-family.co.uk']}
        for email in emails['valid']:
            t = AuthToken.objects.create(
                email=email, token_type=AuthTokenType.VALIDATEEMAIL)
            self.assertEqual(t.email, email)
        for email in emails['invalid']:
            self.assertRaises(
                ValidationError, AuthToken.objects.create, email=email,
                token_type=AuthTokenType.VALIDATEEMAIL)

    def test_token_invalid_type(self):
        good_types = [AuthTokenType.PASSWORDRECOVERY,
                      AuthTokenType.VALIDATEEMAIL,
                      AuthTokenType.INVALIDATEEMAIL,
                      AuthTokenType.VALIDATEGPG,
                      AuthTokenType.VALIDATESIGNONLYGPG]
        for token_type in good_types:
            token = AuthToken.objects.create(email='mark@example.com',
                                             token_type=token_type)
            self.assertEqual(token.token_type, token_type)

        bad_types = [t[0] for t in AuthTokenType._get_choices()
                     if t[0] not in good_types]
        for token_type in bad_types:
            self.assertRaises(ValueError, AuthToken.objects.create,
                              email='mark@example.com', token_type=token_type)

    def test_token_url_contains_raw_token(self):
        token = AuthToken.objects.create(
            email='mark@example.com', token_type=AuthTokenType.VALIDATEEMAIL)
        self.assertIn(token.raw_token, token.get_absolute_url())

    def test_token_is_set(self):
        token = AuthToken.objects.create(
            email='mark@example.com', token_type=AuthTokenType.VALIDATEEMAIL)
        self.assertIsNotNone(token.hashed_token)
        # The actual stored value for "token" should be the SHA256 checksum
        self.assertEqual(len(token.hashed_token), 64)
        # Hashing the raw token attribute should yield the checksum
        self.assertEqual(token.hashed_token,
                         AuthToken.hash_raw(token.raw_token))

    def test_token_is_not_overwritten(self):
        token_string = '123LMN456ZKW'
        token = AuthToken.objects.create(
            email='mark@example.com', raw_token=token_string,
            token_type=AuthTokenType.VALIDATEEMAIL)
        self.assertEqual(token.hashed_token, AuthToken.hash_raw(token_string))

    def test_raw_token_not_stored(self):
        token = AuthToken.objects.create(
            email='mark@example.com', token_type=AuthTokenType.VALIDATEEMAIL)
        self.assertIsNotNone(token.raw_token)
        token = AuthToken.objects.get(email='mark@example.com')
        # A token retrieved from the database should NOT have a raw_token
        # attribute
        with self.assertRaises(AttributeError):
            token.raw_token

    def test_can_create_validate_gpg_token_with_fingerprint(self):
        AuthToken.objects.create(
            email='foo@bar.com',
            token_type=AuthTokenType.VALIDATEGPG,
            fingerprint='A' * 40,
        )
        AuthToken.objects.create(
            email='foo@bar.com',
            token_type=AuthTokenType.VALIDATESIGNONLYGPG,
            fingerprint='0' * 40,
        )


class AuthTokenTestCase(SSOBaseTestCase):

    email = 'test@canonical.com'
    new_email = 'email@domain.com'

    def setUp(self):
        super(AuthTokenTestCase, self).setUp()
        self.account = self.factory.make_account(email=self.email)

        assert EmailAddress.objects.filter(email=self.new_email).count() == 0

    def test_get_absolute_url(self):
        for t, name in AuthTokenType._get_choices():
            token_string = str(t).ljust(6, 'x')
            # First request a token with a valid type to trick the manager into
            # creating it...
            token = AuthToken.objects.create(
                token_type=AuthTokenType.VALIDATEEMAIL,
                email=self.new_email, raw_token=token_string,
            )
            # Now assign the type we want and save it, bypassing the manager.
            token.token_type = t
            token.save()
            name = 'claim_token'
            kwargs = dict(authtoken=token_string)
            if t in authtoken.urls_mapping:
                name = authtoken.urls_mapping[t]
                kwargs['email_address'] = self.new_email

            url = reverse(name, kwargs=kwargs)
            self.assertEqual(token.get_absolute_url(), url)

    def test_consume(self):
        email = 'mark@example.com'
        token_type = AuthTokenType.VALIDATEEMAIL
        # need more than one token to fully test the method
        token = AuthToken.objects.create(
            email=email, token_type=token_type, raw_token='a')
        token2 = AuthToken.objects.create(
            email=email, token_type=token_type, raw_token='b')
        self.assertIsNone(token.date_consumed)
        self.assertIsNone(token2.date_consumed)

        token.consume()

        tokens = AuthToken.objects.filter(
            email=token.email, token_type=token.token_type,
            requester=token.requester)
        self.assertEqual(tokens.count(), 2)
        self.assertIsNotNone(tokens[0].date_consumed)
        self.assertIsNotNone(tokens[1].date_consumed)

    @patch('identityprovider.models.authtoken.now')
    def test_consume_use_utc(self, mock_now):
        mock_now.return_value = now()

        email = 'mark@example.com'
        token_type = AuthTokenType.VALIDATEEMAIL
        # need more than one token to fully test the method
        token = AuthToken.objects.create(
            email=email, token_type=token_type, raw_token='a')
        token2 = AuthToken.objects.create(
            email=email, token_type=token_type, raw_token='b')
        self.assertIsNone(token.date_consumed)
        self.assertIsNone(token2.date_consumed)

        token.consume()

        tokens = AuthToken.objects.filter(
            email=token.email, token_type=token.token_type,
            requester=token.requester)
        self.assertEqual(tokens.count(), 2)
        self.assertEqual(tokens[0].date_consumed, mock_now.return_value)
        self.assertEqual(tokens[1].date_consumed, mock_now.return_value)
        self.assertEqual(mock_now.called, True)

    def assert_rng_is_deterministic(self):
        """Verify the RNG is predictable to ensure tests are valid."""
        rng = settings.RANDOM_GENERATOR
        rng.seed(42)
        string1 = generate_random_string(10)
        rng.seed(42)
        string2 = generate_random_string(10)
        self.assertEqual(string1, string2)

    @override_settings(RANDOM_GENERATOR=random)
    def test_create_unique_token_uses_correct_charset(self):
        # We override RNG to avoid eating up entropy from SystemRandom since
        # we'll generate a very long string.
        # A length-100 string is virtually guaranteed to contain a character
        # outside AUTHTOKEN_CHARS, if the code is somehow wrong. (See
        # cumulative binomial distribution).

        token1 = authtoken.create_unique_token_for_table(
            100, AuthToken, 'hashed_token')
        self.assertTrue(all(ch in authtoken.AUTHTOKEN_CHARS for ch in token1),
                        "Generated token contains unexpected characters")

    @override_settings(RANDOM_GENERATOR=random)
    def test_create_unique_token_for_table(self):
        rng = settings.RANDOM_GENERATOR
        # Ensure we're using RNG which honors seed() and is deterministic.
        self.assert_rng_is_deterministic()
        # initialize randomizer
        rng.seed(42)
        # create a token
        token1 = authtoken.create_unique_token_for_table(
            1, AuthToken, 'hashed_token')
        # use that token within an AuthToken
        AuthToken.objects.create(
            email='mark@example.com',
            token_type=AuthTokenType.VALIDATEEMAIL, raw_token=token1)

        # reinitialize randomizer to force same token
        rng.seed(42)
        token2 = authtoken.create_unique_token_for_table(
            1, AuthToken, 'hashed_token')
        self.assertNotEqual(token1, token2)

    @override_settings(RANDOM_GENERATOR=random)
    def test_create_unique_token_for_table_old_style(self):
        rng = settings.RANDOM_GENERATOR
        # Ensure we're using RNG which honors seed() and is deterministic.
        self.assert_rng_is_deterministic()

        # initialize randomizer
        rng.seed(42)

        # create a token
        token1 = authtoken.create_unique_token_for_table(
            1, AuthToken, 'hashed_token')
        # Create an AuthToken
        clash_token = AuthToken.objects.create(
            email='mark@example.com',
            token_type=AuthTokenType.VALIDATEEMAIL)
        # Turn it into an old-style AuthToken (i.e. it was here
        # before and is not hashed) with the token1 value.
        clash_token.hashed_token = token1
        clash_token.save()

        # reinitialize randomizer to force same token
        rng.seed(42)
        token2 = authtoken.create_unique_token_for_table(
            1, AuthToken, 'hashed_token')
        self.assertNotEqual(token1, token2)

    @override_settings(RANDOM_GENERATOR=random)
    def test_create_unique_token_for_table_double_match(self):
        """Test the remote far-fetched case of two tokens matching."""
        rng = settings.RANDOM_GENERATOR
        # Ensure we're using RNG which honors seed() and is deterministic.
        self.assert_rng_is_deterministic()
        # initialize randomizer
        rng.seed(42)

        # create a token
        token1 = authtoken.create_unique_token_for_table(
            1, AuthToken, 'hashed_token')
        # Create an AuthToken - this is hashed and will match token1
        AuthToken.objects.create(
            email='mark@example.com',
            token_type=AuthTokenType.VALIDATEEMAIL, raw_token=token1)
        # Create an old-style token - this will also match token1 but is
        # not hashed.
        clash_token2 = AuthToken.objects.create(
            email='mark@example.com',
            token_type=AuthTokenType.VALIDATEEMAIL)
        clash_token2.hashed_token = token1
        clash_token2.save()

        # reinitialize randomizer to force same token
        rng.seed(42)
        token2 = authtoken.create_unique_token_for_table(
            1, AuthToken, 'hashed_token')
        self.assertNotEqual(token1, token2)

    def test_active(self):
        t = AuthToken.objects.create(
            token_type=AuthTokenType.VALIDATEEMAIL, email=self.new_email)
        self.assertTrue(t.active)

        t.consume()
        self.assertFalse(t.active)

    def test_get_tokens_finds_hashed_token(self):
        token = AuthToken.objects.create(
            token_type=AuthTokenType.VALIDATEEMAIL, email=self.new_email)
        raw_token = token.raw_token
        found_token = AuthToken.get_token(
            token_type=AuthTokenType.VALIDATEEMAIL,
            email=self.new_email, raw_token=raw_token)
        self.assertEqual(found_token, token)

    def test_get_tokens_non_ascii_chars_fails(self):
        # This should only happen if non-ascii chars make it past
        # user input validation, so it should exception out.
        AuthToken.objects.create(
            token_type=AuthTokenType.VALIDATEEMAIL, email=self.new_email)
        weird_token = u'é'  # Accented e

        with self.assertRaises(UnicodeEncodeError):
            AuthToken.get_token(
                token_type=AuthTokenType.VALIDATEEMAIL,
                email=self.new_email, raw_token=weird_token)

    def test_get_tokens_finds_old_style_non_hashed_token(self):
        token = AuthToken.objects.create(
            token_type=AuthTokenType.VALIDATEEMAIL, email=self.new_email)
        # Hack the token to make it an old-style token with raw token stored
        # in the "hashed_token" field.
        token.hashed_token = "foobar"
        token.save()
        # get_token should magically do the right thing
        found_token = AuthToken.get_token(
            token_type=AuthTokenType.VALIDATEEMAIL,
            email=self.new_email, raw_token="foobar")
        self.assertEqual(found_token, token)

    def test_get_tokens_with_duplicates_raises_exception(self):
        # Since a duplicated token should never appear, the designed/tested
        # behavior is to propagate MultipleObjectsReturned exception.
        token_hashed = AuthToken.objects.create(
            token_type=AuthTokenType.VALIDATEEMAIL, email=self.new_email)
        token_raw = AuthToken.objects.create(
            token_type=AuthTokenType.VALIDATEEMAIL, email=self.new_email)
        # Modify the raw token so the value it stores is the same as the
        # hashed token's raw token (confusing, I know).
        token_raw.hashed_token = token_hashed.raw_token
        token_raw.save()
        with self.assertRaises(MultipleObjectsReturned):
            AuthToken.get_token(
                token_type=AuthTokenType.VALIDATEEMAIL,
                email=self.new_email, raw_token=token_hashed.raw_token)

    def test_verify_token_string_new_style(self):
        token_hashed = AuthToken.objects.create(
            token_type=AuthTokenType.VALIDATEEMAIL, email=self.new_email)
        token_verified = verify_token_string(
            raw_token_string=token_hashed.raw_token,
            email=self.new_email)
        self.assertEqual(token_hashed, token_verified)

    def test_verify_token_string_old_style(self):
        token_old = AuthToken.objects.create(
            token_type=AuthTokenType.VALIDATEEMAIL, email=self.new_email)
        token_old.hashed_token = token_old.raw_token
        token_old.save()
        token_verified = verify_token_string(
            raw_token_string=token_old.raw_token,
            email=self.new_email)
        self.assertEqual(token_old, token_verified)

    def test_get_authtoken_or_404_finds(self):
        token = AuthToken.objects.create(
            token_type=AuthTokenType.VALIDATEEMAIL, email=self.new_email)
        found_token = authtoken.get_authtoken_or_404(
            token_type=AuthTokenType.VALIDATEEMAIL,
            email=self.new_email, raw_token=token.raw_token)
        self.assertEqual(token, found_token)

    def test_get_authtoken_or_404_does_not_find(self):
        AuthToken.objects.create(
            token_type=AuthTokenType.VALIDATEEMAIL, email=self.new_email)
        with self.assertRaises(Http404):
            authtoken.get_authtoken_or_404(
                token_type=AuthTokenType.VALIDATEEMAIL,
                email=self.new_email, raw_token="bogustoken")

    def test_is_hashed_hashed_token(self):
        token = AuthToken.objects.create(
            token_type=AuthTokenType.VALIDATEEMAIL, email=self.new_email)
        self.assertTrue(token.is_hashed)

    def test_is_hashed_non_hashed_token(self):
        token = AuthToken.objects.create(
            token_type=AuthTokenType.VALIDATEEMAIL, email=self.new_email)
        token.hashed_token = "unhashedtoken"
        token.save()
        self.assertFalse(token.is_hashed)

    def test_is_hashed_non_hashed_token_correct_length(self):
        token = AuthToken.objects.create(
            token_type=AuthTokenType.VALIDATEEMAIL, email=self.new_email)
        token.hashed_token = token.hashed_token[:-1] + "x"
        token.save()
        self.assertEqual(len(token.hashed_token), authtoken.HASH_LENGTH)
        self.assertFalse(token.is_hashed)

    def test_is_hashed_weird_hashed_token(self):
        token = AuthToken.objects.create(
            token_type=AuthTokenType.VALIDATEEMAIL, email=self.new_email)
        token.hashed_token = token.hashed_token * 2
        token.save()
        self.assertFalse(token.is_hashed)

    def test_short_token_hashed_token(self):
        token = AuthToken.objects.create(
            token_type=AuthTokenType.VALIDATEEMAIL, email=self.new_email)
        self.assertNotIn(token.hashed_token, token.short_token)

    def test_short_token_non_hashed_token(self):
        token = AuthToken.objects.create(
            token_type=AuthTokenType.VALIDATEEMAIL, email=self.new_email)
        token.hashed_token = "unhashedtoken"
        token.save()
        self.assertEqual("unhashedtoken", token.short_token)
