# Copyright 2010 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

from contextlib import contextmanager
from cStringIO import StringIO
from unittest import skip
from urllib2 import HTTPError, URLError

from django.conf import settings
from django.test import TestCase
from django.test.utils import override_settings
from mock import Mock, patch

from identityprovider.models.captcha import (
    Captcha,
    CaptchaResponse,
    CaptchaV2,
    NewCaptchaError,
    VerifyCaptchaError,
)
from identityprovider.tests.utils import SSOBaseTestCase


class CaptchaResponseTestCase(TestCase):

    def test_no_data(self):
        r = CaptchaResponse(200, None)
        self.assertEqual(r.data(), None)

    def test_data_from_response(self):
        resp = StringIO('this is the response')
        r = CaptchaResponse(200, resp)
        self.assertEqual(r._data, None)
        self.assertEqual(r.data(), 'this is the response')

    def test_data_exists(self):
        r = CaptchaResponse(200, None)
        r._data = 'this is the data'
        self.assertEqual(r.data(), 'this is the data')


class CaptchaOpenTestCase(TestCase):

    def test_open_properly_handles_all_expected_error_codes(self):
        handled_codes = [111, 113, 408, 500, 502, 503, 504]
        Captcha._setup_opener()
        self.addCleanup(setattr, Captcha, 'opener', None)

        def build_mock_open(code):
            def mock_open(request):
                if code < 200:
                    raise URLError([code, "reason"])
                else:
                    raise HTTPError(
                        None, code, "reason", None, StringIO("test"))
            return mock_open

        for code in handled_codes:
            Captcha.opener.open = build_mock_open(code)
            response = Captcha._open(None)
            self.assertTrue(response.is_error)
            self.assertEqual(response.reason, "reason")
            self.assertEqual(response.code, code)
            if code >= 200:
                self.assertEqual(response.body, "test")


@override_settings(CAPTCHA_PUBLIC_KEY='public', CAPTCHA_PRIVATE_KEY='private')
class CaptchaBaseTestCase(SSOBaseTestCase):

    def setUp(self):
        super(CaptchaBaseTestCase, self).setUp()
        # patch Captcha so it never hits the real network
        self.captcha_response = CaptchaResponse(
            code=666, response=StringIO('\n'))
        self.mock_captcha_open = self.patch(
            'identityprovider.models.captcha.CaptchaBase._open',
            return_value=self.captcha_response)


class CaptchaTestCase(CaptchaBaseTestCase):

    def test_opener_is_created_during_init(self):
        Captcha.opener = None
        Captcha('test')
        self.assertTrue(Captcha.opener is not None)

    @patch('urllib.urlencode')
    def test_non_ascii_captcha_solution(self, mock_urlencode):
        captcha_solution = u'\xa3'
        encoded_solution = captcha_solution.encode('utf-8')
        mock_urlencode.return_value = 'some text'

        captcha = Captcha.new()
        self.mock_captcha_open.return_value.is_error = True
        captcha.verify(captcha_solution, 'foobar', '')

        captcha_response = mock_urlencode.call_args[0][0]['response']
        self.assertEqual(captcha_response, encoded_solution)

    def test_opener_is_created_when_proxy_is_required(self):
        Captcha.opener = None

        with self.settings(CAPTCHA_PROXIES={'http': 'https://foo.com'}):
            Captcha('test')

        self.assertTrue(Captcha.opener is not None)

    def test_serialize(self):
        captcha = Captcha('id', 'image')

        self.assertEqual(captcha.serialize(),
                         {'captcha_id': 'id', 'image_url': 'image'})

    def test_new_captcha_error(self):
        self.captcha_response.is_error = True
        with self.assertRaises(NewCaptchaError) as ctx:
            Captcha.new()

        e = ctx.exception
        self.assertEqual(e.dummy.captcha_id, None)
        self.assertEqual(e.dummy.image_url, None)

    def test_captcha_open(self):
        self.mock_captcha_open.side_effect = URLError((-1, 'error'))
        self.assertRaises(URLError, Captcha.new)

    def test_new_captcha_uses_timer(self):
        passed_info = []

        @contextmanager
        def timer_fn(category, detail):
            passed_info.append((category, detail))
            yield

        self.captcha_response.code = 200

        Captcha.new(timer=timer_fn)

        self.assertEqual(1, len(passed_info))
        self.assertEqual('captcha-new', passed_info[0][0])
        self.assertEqual(
            self.mock_captcha_open.call_args[0][0], passed_info[0][1])


class CaptchaV2TestCase(CaptchaBaseTestCase):

    def test_opener_is_created_during_init(self):
        CaptchaV2.opener = None
        CaptchaV2()
        self.assertTrue(CaptchaV2.opener is not None)

    def test_new_captcha_uses_timer(self):
        passed_info = []

        @contextmanager
        def timer_fn(category, detail):
            passed_info.append((category, detail))
            yield

        self.captcha_response.code = 200

        CaptchaV2(timer=timer_fn)

        self.assertEqual(1, len(passed_info))
        self.assertEqual('captcha-new', passed_info[0][0])

    def test_verify_uses_timer(self):
        passed_info = []

        @contextmanager
        def timer_fn(category, detail):
            passed_info.append((category, detail))
            yield
        c = CaptchaV2()
        c.verify(None, 'localhost', '', timer=timer_fn)
        self.assertEqual('captcha-verify', passed_info[0][0])
        self.assertEqual(settings.CAPTCHA_V2_VERIFY_URL, passed_info[0][1])


class CaptchaVerifyTestCase(CaptchaBaseTestCase):

    CAPTCHA = Captcha
    VERIFICATION_URL = settings.CAPTCHA_VERIFY_URL
    RESPONSE_OK = 'true\nok'

    def test_verify_calls_check_whitelist(self):
        captcha = self.CAPTCHA(None)
        captcha.check_whitelist = Mock(return_value=False)

        result = captcha.verify(None, 'localhost', 'foo')
        self.assertFalse(result)
        self.assertTrue(self.mock_captcha_open.called)

        captcha.check_whitelist.assert_called_with('foo')

    def test_check_whitelist_match(self):
        captcha = self.CAPTCHA(None)

        regexps = ['not a match', '^foo$']
        with self.settings(EMAIL_WHITELIST_REGEXP_LIST=regexps):
            result = captcha.verify(None, 'localhost', 'foo')
            self.assertTrue(result)
            self.assertFalse(self.mock_captcha_open.called)

    def test_check_whitelist_no_match(self):
        captcha = self.CAPTCHA(None)

        regexps = ['not a match', '^bar$']
        with self.settings(EMAIL_WHITELIST_REGEXP_LIST=regexps):
            result = captcha.verify(None, 'localhost', 'foo')
            self.assertFalse(result)
            self.assertTrue(self.mock_captcha_open.called)

    def test_verify_already_verified(self):
        c = self.CAPTCHA(None)
        c._verified = True
        r = c.verify(None, None, '')
        self.assertTrue(r)

    def test_verify_response_ok(self):
        self.captcha_response.code = 200
        self.captcha_response.response = StringIO(self.RESPONSE_OK)

        c = self.CAPTCHA.new()
        self.assertTrue(c.verify(None, "localhost", ''))
        self.assertEqual(c.message, 'ok')

    def test_verify_no_captcha_id(self):
        c = self.CAPTCHA(None)

        self.captcha_response.is_error = True
        self.assertFalse(c.verify(None, 'localhost', ''))
        self.assertEqual(c.message, 'no-challenge')

    def test_verify_error(self):
        c = self.CAPTCHA(None)
        c.captcha_id = 'challenge-id'

        self.captcha_response.is_error = True
        self.assertRaises(VerifyCaptchaError, c.verify, None, 'localhost', '')

    def test_verify_uses_timer(self):
        passed_info = []

        @contextmanager
        def timer_fn(category, detail):
            passed_info.append((category, detail))
            yield
        c = self.CAPTCHA(None)
        c.verify(None, 'localhost', '', timer=timer_fn)
        self.assertEqual('captcha-verify', passed_info[0][0])
        self.assertEqual(self.VERIFICATION_URL, passed_info[0][1])


class CaptchaV2VerifyTestCase(CaptchaVerifyTestCase):

    CAPTCHA = CaptchaV2
    VERIFICATION_URL = settings.CAPTCHA_V2_VERIFY_URL
    RESPONSE_OK = '{"success": true}'

    @skip("This (inherited) test makes no sense for CaptchaV2")
    def test_verify_no_captcha_id(self):
        pass
