# -*- coding: utf-8 -*-

from datetime import timedelta

from django.core.exceptions import ValidationError
from django.db import IntegrityError
from django.utils.timezone import now
from mock import Mock, patch

from identityprovider.models import (
    EmailAddress,
    InvalidatedEmailAddress,
)
from identityprovider.models.const import EmailStatus
from identityprovider.models.emailaddress import EmailAddressQuerySet
from identityprovider.tests.utils import SSOBaseTestCase


class EmailAddressQuerySetTestCase(SSOBaseTestCase):

    def setUp(self):
        super(EmailAddressQuerySetTestCase, self).setUp()
        self.qs = EmailAddressQuerySet(model=EmailAddress)

    def test_verified_list_only_verified_emails(self):
        emails = self.qs.verified()
        for email in emails:
            self.assertTrue(email.is_verified)


class EmailAddressManagerTestCase(SSOBaseTestCase):

    def test_verified(self):
        account = self.factory.make_account()  # this gives a validated address
        for i in xrange(5):
            EmailAddress.objects.create(email='%s@foo.com' % i,
                                        status=EmailStatus.NEW,
                                        account=account)
        for i in xrange(3):
            EmailAddress.objects.create(email='%s@example.com' % i,
                                        status=EmailStatus.VALIDATED,
                                        account=account)
        for i in xrange(3):
            EmailAddress.objects.create(email='%s@bar.com' % i,
                                        status=EmailStatus.PREFERRED,
                                        account=account)

        assert EmailAddress.objects.all().count() == 12

        emails = EmailAddress.objects.verified()
        self.assertEqual(len(emails), 7)
        for email in emails:
            self.assertTrue(email.is_verified)


class EmailAddressTestCase(SSOBaseTestCase):

    email = 'test@canonical.com'

    def setUp(self):
        super(EmailAddressTestCase, self).setUp()
        self.account = self.factory.make_account(email=self.email)

    def test_emailaddress_with_account(self):
        email = EmailAddress.objects.get(email=self.email)
        self.assertEqual(self.account, email.account)

    def test_emailaddress_is_verified(self):
        email = EmailAddress.objects.get(email=self.email)
        assert email.status == EmailStatus.PREFERRED
        self.assertTrue(email.is_verified)
        email.status = EmailStatus.VALIDATED
        self.assertTrue(email.is_verified)
        email.status = EmailStatus.NEW
        self.assertFalse(email.is_verified)

    def test_emailaddress_validate(self):
        email = EmailAddress.objects.get(email=self.email)
        email.status = EmailStatus.NEW
        email.validate()
        self.assertTrue(email.status, EmailStatus.VALIDATED)
        email.status = EmailStatus.PREFERRED
        email.validate()
        self.assertTrue(email.status, EmailStatus.PREFERRED)

    def test_emailaddress_invalidate(self):
        email = EmailAddress.objects.get(email=self.email)
        invalidated = email.invalidate()
        self.assertTrue(isinstance(invalidated, InvalidatedEmailAddress))
        self.assertEqual(email.email, invalidated.email)
        self.assertEqual(email.account, invalidated.account)
        self.assertEqual(email.date_created, invalidated.date_created)
        # is not available anymore
        emails = EmailAddress.objects.filter(email=self.email)
        self.assertFalse(emails.exists())

    def test_emailaddress_invalidate_with_person(self):
        email_address = self.factory.make_email_address()
        account = self.factory.make_account(email=email_address)
        person = self.factory.make_person(account=account)
        # set up email with lp_person instead of account
        email = EmailAddress.objects.get(email=email_address)
        email.account = None
        email.lp_person = person.id
        email.save()

        invalidated = email.invalidate()
        self.assertTrue(isinstance(invalidated, InvalidatedEmailAddress))
        self.assertEqual(email.email, invalidated.email)
        self.assertEqual(person.account, invalidated.account)
        self.assertEqual(email.date_created, invalidated.date_created)
        # is not available anymore
        emails = EmailAddress.objects.filter(email=email_address)
        self.assertFalse(emails.exists())

    @patch('identityprovider.models.emailaddress.logger')
    def test_emailaddress_invalidate_without_account(self, mock_logger):
        email_address = self.factory.make_email_address()
        self.factory.make_account(email=email_address)
        # set up email with lp_person instead of account
        email = EmailAddress.objects.get(email=email_address)
        email.account = None
        # invalid person id
        email.lp_person = 12345
        email.save()

        invalidated = email.invalidate()
        self.assertEqual(invalidated, None)
        mock_logger.warning.assert_called_once_with(
            "Could not create invalidated entry for %s, "
            "no associated account found", email_address
        )
        # original is not available anymore
        emails = EmailAddress.objects.filter(email=email_address)
        self.assertFalse(emails.exists())


class InvalidatedEmailAddressTestCase(SSOBaseTestCase):

    def test_date_created_not_default(self):
        self.assertRaises(
            IntegrityError,
            InvalidatedEmailAddress.objects.create, email='foo@foo.com')

    def test_email_valid(self):
        for valid in ('a@a.com', 'aaaa.bbbb@foo.com', 'a@foo.net'):
            email = InvalidatedEmailAddress.objects.create(
                email=valid, date_created=now())
            email.full_clean()  # no failure

    def test_email_invalid(self):
        for invalid in ('', '.', '@', '@foo', '@foo.x', 'a@foo.x', 'foo@bar',
                        'a@@a.com'):
            email = InvalidatedEmailAddress.objects.create(
                email=invalid, date_created=now())
            self.assertRaises(
                ValidationError, email.full_clean)

    def test_unicode(self):
        email = u'zaraza♥foo@example.com'
        invalid = InvalidatedEmailAddress.objects.create(
            email=email, date_created=now())
        self.assertEqual(email, unicode(invalid))

    def test_date_invalidated(self):
        some_date = now() + timedelta(days=30)
        invalid = InvalidatedEmailAddress.objects.create(
            email='foo@example.com', date_created=now(),
            date_invalidated=some_date)

        self.assertEqual(invalid.date_invalidated, some_date)

    def test_date_invalidated_default(self):
        mock_date = Mock()
        mock_date.return_value = now()
        InvalidatedEmailAddress._meta.get_field_by_name(
            'date_invalidated')[0].default = mock_date
        invalid = InvalidatedEmailAddress.objects.create(
            email='foo@example.com', date_created=now())

        self.assertEqual(invalid.date_invalidated, mock_date.return_value)

    def test_account_notified(self):
        invalid = InvalidatedEmailAddress.objects.create(
            email='foo@example.com', date_created=now())
        self.assertFalse(invalid.account_notified)
