# Copyright 2010 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

import base64
import json
from datetime import timedelta
from time import time

from django.utils.timezone import now
from mock import patch
from openid.association import Association
from openid.store.nonce import SKEW

from identityprovider.models.openidmodels import (
    DjangoOpenIDStore,
    OpenIDAssociation,
    OpenIDAuthorization,
    OpenIDNonce,
    OpenIDRPConfig,
    OpenIDRPSummary,
)
from identityprovider.tests.utils import SSOBaseTestCase
from identityprovider.readonly import ReadOnlyManager


class DjangoOpenIDStoreTestCase(SSOBaseTestCase):

    def setUp(self):
        super(DjangoOpenIDStoreTestCase, self).setUp()
        self.server_url = 'http://server.example.com'
        self.store = DjangoOpenIDStore()
        self.assoc, _ = OpenIDAssociation.objects.get_or_create(
            server_url=self.server_url,
            handle="handle", secret="secret".encode("base64"),
            issued=time(), lifetime=1000, assoc_type='HMAC-SHA1')

    def test_store_association_when_association_already_exists(self):
        association = Association("handle", "secret", 2, 2, 'HMAC-SHA1')

        self.store.storeAssociation(self.server_url, association)
        openid_association = OpenIDAssociation.objects.get(
            server_url=self.server_url, handle="handle")

        self.assertEqual(openid_association.issued, 2)

    def test_get_association_with_handle_none(self):
        association = self.store.getAssociation(self.server_url)

        self.assertEqual(association.handle, "handle")

    def test_get_association_when_lifetime_is_zero(self):
        openid_association = OpenIDAssociation.objects.get(
            server_url=self.server_url, handle="handle")
        openid_association.lifetime = 0
        openid_association.save()

        association = self.store.getAssociation(self.server_url)

        self.assertTrue(association is None)

    def test_use_nonce_when_expired(self):
        self.assertFalse(
            self.store.useNonce("url", time() + SKEW + 100, "salt"))

    def test_use_nonce_when_nonce_exists(self):
        OpenIDNonce.objects.create(
            server_url="http://example.com", timestamp=time(), salt="salt")

        self.assertFalse(
            self.store.useNonce("http://example.com", time(), "salt"))

    def test_use_nonce_when_nonce_does_not_exist(self):
        self.assertTrue(
            self.store.useNonce("http://example.com", time(), "salt"))
        OpenIDNonce.objects.get(
            server_url="http://example.com", salt="salt").delete()

    def test_getAssociation_not_existing(self):
        OpenIDAssociation.objects.filter(server_url=self.server_url).delete()
        assoc = self.store.getAssociation(self.server_url)
        self.assertEqual(assoc, None)

    def test_getAssociation_existing_no_handle(self):
        expected = Association(
            self.assoc.handle, base64.b64decode(str(self.assoc.secret)),
            int(self.assoc.issued), self.assoc.lifetime, self.assoc.assoc_type)
        assoc = self.store.getAssociation(self.server_url)
        self.assertEqual(assoc, expected)

    def test_getAssociation_expired(self):
        # make sure the association has expired
        self.assoc.lifetime = 0
        self.assoc.save()

        assoc = self.store.getAssociation(self.server_url)
        self.assertEqual(assoc, None)
        assocs = OpenIDAssociation.objects.filter(server_url=self.server_url)
        self.assertEqual(assocs.count(), 0)

    def test_getAssociation_existing_same_handle(self):
        expected = Association(
            self.assoc.handle, base64.b64decode(str(self.assoc.secret)),
            int(self.assoc.issued), self.assoc.lifetime, self.assoc.assoc_type)
        assoc = self.store.getAssociation(self.server_url, 'handle')
        self.assertEqual(assoc, expected)

    def test_getAssociation_existing_different_handle(self):
        assoc = self.store.getAssociation(self.server_url, 'otherhandle')
        self.assertEqual(assoc, None)

    def test_cleanupNonce(self):
        OpenIDNonce.objects.create(server_url=self.server_url, timestamp=0,
                                   salt='')

        self.assertEqual(OpenIDNonce.objects.all().count(), 1)
        self.store.cleanupNonce()
        self.assertEqual(OpenIDNonce.objects.all().count(), 0)

    def test_cleanupAssociations(self):
        self.assoc.lifetime = 0
        self.assoc.save()

        self.assertEqual(OpenIDAssociation.objects.all().count(), 1)
        self.store.cleanupAssociations()
        self.assertEqual(OpenIDAssociation.objects.all().count(), 0)


class OpenIDRPConfigTestCase(SSOBaseTestCase):

    def setUp(self):
        super(OpenIDRPConfigTestCase, self).setUp()
        self.rp_config = OpenIDRPConfig.objects.create()

    def test_logo_url_property_when_logo_is_none(self):
        self.rp_config.logo = None

        self.assertTrue(self.rp_config.logo_url is None)

    def test_logo_url_property_when_logo_is_relative(self):
        with self.settings(MEDIA_URL='http://our_sso_server/media/'):
            self.rp_config.logo = 'test.png'
            self.assertEqual(self.rp_config.logo_url,
                             'http://our_sso_server/media/test.png')

    def test_logo_url_property_when_logo_is_absolute(self):
        with self.settings(MEDIA_URL='http://our_sso_server/media/'):
            external_logo_url = 'http://www.consumer/logo.png'
            self.rp_config.logo = external_logo_url
            self.assertEqual(self.rp_config.logo_url, external_logo_url)

    def test_unicode(self):
        self.rp_config.displayname = 'MyOpenIDRPConfig'
        self.rp_config.save()
        self.assertEqual(unicode(self.rp_config), u'MyOpenIDRPConfig')

    def test_twofactor_required_default(self):
        self.assertFalse(self.rp_config.twofactor_required)


class OpenIDAssociationTestCase(SSOBaseTestCase):

    def setUp(self):
        super(OpenIDAssociationTestCase, self).setUp()
        for i in xrange(5):
            OpenIDAssociation.objects.create(
                server_url='http://test.com',
                handle='handle%s' % i,
                secret=('secret%s' % i).encode('base64'),
                issued=time(), lifetime=1000, assoc_type='HMAC-SHA1')

    def test_create_association(self):
        # If this creation doesn't fail then our primary key is right,
        # as we've already got several associations with the same
        # trust root.
        OpenIDAssociation.objects.create(
            server_url="http://test.com",
            handle="handle", secret="secret".encode("base64"),
            issued=time(), lifetime=1000, assoc_type='HMAC-SHA1')


class OpenIDAuthorizationTestCase(SSOBaseTestCase):

    def setUp(self):
        super(OpenIDAuthorizationTestCase, self).setUp()
        self.account = self.factory.make_account()
        self.trust_root = 'http://foo.bar.baz'
        self.expires = now() + timedelta(1)

    def create_auth(self):
        return OpenIDAuthorization.objects.create(
            account=self.account, client_id=None, date_expires=self.expires,
            trust_root=self.trust_root)

    def test_authorize_when_readonly(self):
        rm = ReadOnlyManager()
        rm.set_readonly()
        self.addCleanup(rm.clear_readonly)

        expires = now()
        OpenIDAuthorization.objects.authorize(self.account, self.trust_root,
                                              expires)
        self.assertRaises(
            OpenIDAuthorization.DoesNotExist,
            OpenIDAuthorization.objects.get, account=self.account,
            client_id=None, trust_root=self.trust_root)

    @patch('identityprovider.models.openidmodels.now')
    def test_authorize_expires(self, mock_now):
        mock_now.return_value = now()

        OpenIDAuthorization.objects.authorize(self.account, self.trust_root)

        auth = OpenIDAuthorization.objects.get(
            account=self.account, client_id=None, trust_root=self.trust_root)
        self.assertEqual(auth.date_expires, mock_now.return_value)

    @patch('identityprovider.models.openidmodels.now')
    def test_authorize_existing(self, mock_now):
        mock_now.return_value = now()
        auth1 = self.create_auth()
        OpenIDAuthorization.objects.authorize(self.account, self.trust_root)
        auth2 = OpenIDAuthorization.objects.get(pk=auth1.id)
        self.assertEqual(auth2.date_expires, mock_now.return_value)

    @patch('identityprovider.models.openidmodels.now')
    def test_authorize_not_existing(self, mock_now):
        mock_now.return_value = now()
        OpenIDAuthorization.objects.authorize(self.account, self.trust_root,
                                              None)
        auth = OpenIDAuthorization.objects.get(
            account=self.account, client_id=None, trust_root=self.trust_root)
        self.assertEqual(auth.date_expires, mock_now.return_value)

    def test_is_authorized_generic(self):
        self.create_auth()
        self.assertTrue(OpenIDAuthorization.objects.is_authorized(
            self.account, self.trust_root, 'client'))

    def test_is_authorized_client_id(self):
        self.assertFalse(OpenIDAuthorization.objects.is_authorized(
            self.account, self.trust_root, 'client'))
        OpenIDAuthorization.objects.authorize(
            self.account, self.trust_root,
            expires=self.expires, client_id='client')
        self.assertTrue(OpenIDAuthorization.objects.is_authorized(
            self.account, self.trust_root, 'client'))

    def test_is_not_authorized(self):
        self.assertFalse(OpenIDAuthorization.objects.is_authorized(
            self.account, self.trust_root, 'client'))


class OpenIDRPSummaryTestCase(SSOBaseTestCase):

    approved_data = {
        'test1': {'one': [1, 2], 'two': [2, 3]},
        'test2': {'three': [3, 4], 'four': [4, 5]}}

    def setUp(self):
        super(OpenIDRPSummaryTestCase, self).setUp()
        self.account = self.factory.make_account()
        identity = self.factory.make_request().build_absolute_uri(
            self.account.get_absolute_url())
        self.openid_identity_url = identity
        self.trust_root = 'http://foo.bar.baz'

    def create_summary(self):
        summary = OpenIDRPSummary.objects.create(
            account=self.account,
            trust_root=self.trust_root, openid_identifier='oid')
        summary.set_approved_data(self.approved_data)
        summary.save()
        return summary

    def test_record_when_readonly(self):
        rm = ReadOnlyManager()
        rm.set_readonly()

        try:
            summary = OpenIDRPSummary.objects.record(
                self.account, self.trust_root, self.openid_identity_url)
            self.assertEqual(summary, None)
            self.assertRaises(
                OpenIDRPSummary.DoesNotExist,
                OpenIDRPSummary.objects.get, account=self.account,
                trust_root=self.trust_root, openid_identifier=None)
        finally:
            rm.clear_readonly()

    def test_record_with_existing_and_no_openid_identifier(self):
        summary1 = self.create_summary()
        summary2 = OpenIDRPSummary.objects.record(
            self.account, self.trust_root, self.openid_identity_url)
        summary2.set_approved_data(self.approved_data)
        self.assertEqual(summary1.total_logins, 1)
        self.assertEqual(summary2.total_logins, 1)
        self.assertNotEqual(summary1, summary2)

    def test_record_existing_same_oid(self):
        summary1 = self.create_summary()
        summary2 = OpenIDRPSummary.objects.record(
            self.account, self.trust_root, 'oid')
        summary2 = OpenIDRPSummary.objects.get(pk=summary2.pk)
        self.assertEqual(summary2.total_logins, 2)
        self.assertEqual(summary1, summary2)

    def test_record_existing_different_oid(self):
        summary1 = self.create_summary()
        summary2 = OpenIDRPSummary.objects.record(
            self.account, self.trust_root, 'oid1',
            approved_data=self.approved_data)
        self.assertEqual(summary1.total_logins, 1)
        self.assertEqual(summary2.total_logins, 1)
        self.assertNotEqual(summary1, summary2)

    def test_record_not_existing(self):
        summary1 = OpenIDRPSummary.objects.record(
            self.account, self.trust_root, 'oid')
        self.assertEqual(summary1.total_logins, 1)
        summary2 = OpenIDRPSummary.objects.get(
            account=self.account,
            trust_root=self.trust_root, openid_identifier='oid')
        self.assertEqual(summary1, summary2)

    def test_approved_data(self):
        summary = self.create_summary()
        summary.set_approved_data(self.approved_data)
        summary.save()
        summary2 = OpenIDRPSummary.objects.get(
            account=self.account,
            trust_root=self.trust_root, openid_identifier='oid')
        expected = self.approved_data.copy()
        expected.update({
            'user_attribs': {
                'requested': [],
                'approved': [],
            }
        })
        self.assertEqual(summary2.get_approved_data(), expected)

    def test_get_approved_data_with_sreg(self):
        summary = self.create_summary()
        summary.set_approved_data({'sreg': {'requested': ['email', 'fullname'],
                                            'approved': ['email']}})
        approved_data = summary.get_approved_data()
        self.assertEqual(approved_data, {
            'sreg': {
                'requested': ['email', 'fullname'],
                'approved': ['email'],
            },
            'user_attribs': {
                'requested': list(set(['email', 'fullname'])),
                'approved': ['email'],
            }
        })

    def test_get_approved_data_with_ax(self):
        summary = self.create_summary()
        summary.set_approved_data({
            'ax': {
                'requested': ['email', 'account_verified'],
                'approved': ['account_verified'],
            }
        })
        approved_data = summary.get_approved_data()
        self.assertEqual(approved_data, {
            'ax': {
                'requested': ['email', 'account_verified'],
                'approved': ['account_verified'],
            },
            'user_attribs': {
                'requested': list(set(['email', 'account_verified'])),
                'approved': ['account_verified'],
            }
        })

    def test_get_approved_data_mixed_sreg_ax(self):
        summary = self.create_summary()
        summary.set_approved_data({
            'sreg': {
                'requested': ['email', 'fullname'],
                'approved': ['email'],
            },
            'ax': {
                'requested': ['email', 'account_verified'],
                'approved': ['account_verified'],
            }
        })
        approved_data = summary.get_approved_data()
        self.assertEqual(approved_data, {
            'sreg': {
                'requested': ['email', 'fullname'],
                'approved': ['email'],
            },
            'ax': {
                'requested': ['email', 'account_verified'],
                'approved': ['account_verified'],
            },
            'user_attribs': {
                'requested': list(set(['email', 'fullname',
                                       'account_verified'])),
                'approved': list(set(['email', 'account_verified'])),
            }
        })

    def test_get_approved_data_persistent(self):
        summary = self.create_summary()
        summary.set_approved_data({
            'ax': {
                'requested': ['email', 'account_verified'],
                'approved': ['account_verified'],
            }
        })
        approved_data = summary.get_approved_data()
        # refresh object from db
        summary = OpenIDRPSummary.objects.get(id=summary.id)
        self.assertEqual(json.loads(summary.approved_data), approved_data)

    def test_get_approved_data_with_user_attribs(self):
        summary = self.create_summary()
        approved_data = {
            'sreg': {
                'requested': ['email', 'fullname'],
                'approved': [],
            },
            'user_attribs': {
                'requested': ['email'],
                'approved': ['email'],
            }
        }
        summary.set_approved_data(approved_data)
        self.assertEqual(summary.get_approved_data(), approved_data)
