# Copyright 2010 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).
from random import randint

from identityprovider.const import (
    PERSON_VISIBILITY_PUBLIC,
    PERSON_VISIBILITY_PRIVATE_MEMBERSHIP,
)
from identityprovider.models import (
    Account,
    LPOpenIdIdentifier,
    Person,
    TeamParticipation,
    get_team_memberships_for_user,
)
from identityprovider.models.const import (
    AccountCreationRationale,
    AccountStatus,
)
from identityprovider.tests.utils import SSOBaseTestCase


class PersonTestCase(SSOBaseTestCase):

    def setUp(self):
        super(PersonTestCase, self).setUp()
        lp_account = randint(100, 10000)
        LPOpenIdIdentifier.objects.create(identifier='oid',
                                          lp_account=lp_account)
        self.person1 = Person.objects.create(
            displayname='Person', name='person', lp_account=lp_account)
        self.person2 = Person.objects.create(displayname='Other', name='other')
        self.team1 = Person.objects.create(name='team',
                                           teamowner=self.person2.id)
        self.team2 = Person.objects.create(name='other-team',
                                           teamowner=self.person1.id)

    def test_unicode(self):
        self.assertEqual(unicode(self.person1), u'Person')

    def test_in_team_no_team(self):
        self.assertFalse(self.person1.in_team('no-team'))

    def test_in_team_by_team_name(self):
        TeamParticipation.objects.create(team=self.team1, person=self.person1)
        with self.assertNumQueries(1):
            self.assertTrue(self.person1.in_team('team'))

    def test_in_team_by_team_object(self):
        TeamParticipation.objects.create(team=self.team1, person=self.person1)
        with self.assertNumQueries(1):
            self.assertTrue(self.person1.in_team(self.team1))

    def test_not_in_team(self):
        self.assertFalse(self.person1.in_team('team'))

    def test_in_team_no_teamparticipation_same_owner(self):
        Person.objects.create(name='otherteam', teamowner=self.person1.id)
        # we can not workaround doing a second query
        with self.assertNumQueries(2):
            self.assertTrue(self.person1.in_team('otherteam'))

    def test_account_when_no_account(self):
        self.assertEqual(self.person1.account, None)

    def test_account_when_account(self):
        account = Account.objects.create(
            creation_rationale=AccountCreationRationale.USER_CREATED,
            status=AccountStatus.ACTIVE, displayname='Person',
            openid_identifier='oid')
        self.assertEqual(self.person1.account, account)


class TeamParticipationTestCase(SSOBaseTestCase):

    def test_unicode(self):
        team = Person.objects.create(displayname='Team', name='team')
        person = Person.objects.create(displayname='Person', name='person')
        tp = TeamParticipation.objects.create(team=team, person=person)
        self.assertEqual(unicode(tp), u'Person in Team')


class TeamMembershipTest(SSOBaseTestCase):

    def _set_team_to_private_membership(self, team_name):
        self._set_team_privacy(team_name, PERSON_VISIBILITY_PRIVATE_MEMBERSHIP)

    def _set_team_to_public_membership(self, team_name):
        self._set_team_privacy(team_name, PERSON_VISIBILITY_PUBLIC)

    def _set_team_privacy(self, team_name, privacy):
        team = Person.objects.get(name=team_name)
        team.visibility = privacy
        team.save()

    def setUp(self):
        super(TeamMembershipTest, self).setUp()
        self.account = self.factory.make_account(teams=['ubuntu-team'])
        assert self.account.person is not None

    def test_get_team_memberships_on_personless_account(self):
        account = self.factory.make_account()
        assert account.person is None
        memberships = get_team_memberships_for_user(
            ['ubuntu-team'], account, False)
        self.assertEqual(memberships, [])

    def test_get_team_memberships_when_team_is_owned(self):
        # Someone that owns a team, but isn't a member isn't considered
        # a member
        self.factory.make_team(name='owned-team', owner=self.account.person)
        memberships = get_team_memberships_for_user(
            ['owned-team'], self.account, False)
        self.assertEqual(memberships, [])

    def test_get_team_memberships_when_team_is_visible(self):
        memberships = get_team_memberships_for_user(
            ['ubuntu-team'], self.account, False)
        self.assertEqual(memberships, ['ubuntu-team'])

    def test_get_team_membership_when_team_is_not_visible(self):
        self._set_team_to_private_membership('ubuntu-team')
        memberships = get_team_memberships_for_user(
            ['ubuntu-team', 'myteam'], self.account, False)
        self.assertEqual(memberships, [])

    def test_team_is_private_but_you_can_see_them(self):
        self._set_team_to_private_membership('ubuntu-team')
        memberships = get_team_memberships_for_user(
            ['ubuntu-team', 'myteam'], self.account, True)
        self.assertEqual(memberships, ['ubuntu-team'])
