import time
from datetime import date

from bzrlib.errors import NotBranchError
from django.conf import settings
from gargoyle.testutils import switches
from mock import Mock, patch

from identityprovider.models.account import Account
from identityprovider.preflight_checks import get_version_info
from identityprovider.tests import DEFAULT_USER_PASSWORD
from identityprovider.tests.utils import (
    SSOBaseTestCase,
    TeamConditionSelectiveMixin,
)


class PreflightTestCase(SSOBaseTestCase, TeamConditionSelectiveMixin):

    URL = '/preflight/'
    key = 'PREFLIGHT'

    def setUp(self):
        self.email = 'email@domain.local'
        self.password = 'testpassword'
        super(PreflightTestCase, self).setUp()

        self.account = self.make_account()
        self.enable_flag()

        self.mock_open = self.patch(
            'identityprovider.preflight_checks.Branch.open_containing')
        self.mock_branch = Mock()
        self.mock_branch.nick = 'trunk'
        self.mock_branch.revno.return_value = 42
        self.mock_tip = Mock(timestamp=123456789)
        self.mock_branch.repository.get_revision.return_value = self.mock_tip
        # make sure open_containing returns a tuple
        self.mock_open.return_value = (self.mock_branch, '')

    def test_anonymous(self):
        response = self.client.get(self.URL)

        self.assertEqual(404, response.status_code)

    def test_account_without_person_receives_404(self):
        account = Account.objects.create_account(
            "test", "abcd@example.com", DEFAULT_USER_PASSWORD)
        assert account.person is None
        assert self.client.login(
            username=account.preferredemail.email,
            password=DEFAULT_USER_PASSWORD)

        response = self.client.get(self.URL)
        self.assertEqual(response.status_code, 404)

    def test_person_not_in_the_required_team_will_get_404(self):
        account = self.factory.make_account(
            email="test@canonical.com", password=DEFAULT_USER_PASSWORD)
        assert self.client.login(
            username=account.preferredemail.email,
            password=DEFAULT_USER_PASSWORD)
        response = self.client.get(self.URL)
        self.assertEqual(response.status_code, 404)

    @switches(PREFLIGHT=False)
    def test_flag_not_active(self):
        login = self.client.login(username=self.email, password=self.password)
        self.assertTrue(login)

        response = self.client.get(self.URL)

        self.assertEqual(404, response.status_code)

    @switches(PREFLIGHT=True)
    def test_success(self):
        login = self.client.login(username=self.email, password=self.password)
        self.assertTrue(login)

        response = self.client.get(self.URL)

        self.assertEqual(200, response.status_code)
        libs = set([x['name'] for x in response.context[0]['versions']])
        expected = set(
            ['Django', 'Platform', 'Python', 'config'] +
            list(settings.INSTALLED_APPS))

        self.assertEqual(expected, libs)

    @patch('identityprovider.preflight_checks.import_module')
    def test_version_from_branch(self, mock_import):
        mod = Mock()
        mod.__name__ = 'foo'
        mod.__file__ = '/some/path'
        mod.__version__ = '1.2.3'
        mock_import.return_value = mod

        tip_date = date(2013, 6, 24)
        mock_revision = Mock()
        mock_revision.timestamp = time.mktime(tip_date.timetuple())
        self.mock_branch.repository.get_revision.return_value = mock_revision

        version_info = get_version_info('foo')

        self.assertEqual(version_info,
                         {'name': 'foo', 'version': 'trunk@42 (13.06.24)'})

    @patch('identityprovider.preflight_checks.import_module')
    def test_version_from_version_attribute(self, mock_import):
        mod = Mock()
        mod.__name__ = 'foo'
        mod.__file__ = '/some/path'
        mod.__version__ = '1.2.3'
        mock_import.return_value = mod

        self.mock_open.side_effect = NotBranchError(mod.__file__)

        version_info = get_version_info(mod)

        self.assertEqual(version_info,
                         {'name': 'foo', 'version': '1.2.3'})

    @patch('identityprovider.preflight_checks.import_module')
    def test_version_without_version_attribute(self, mock_import):
        mod = Mock()
        mod.__name__ = 'foo'
        mod.__file__ = '/some/path'
        mock_import.return_value = mod

        self.mock_open.side_effect = NotBranchError(mod.__file__)

        version_info = get_version_info(mod)

        self.assertEqual(version_info,
                         {'name': 'foo', 'version': 'version not available'})

    def test_config_branch_version(self):
        tip_date = date(2013, 6, 24)
        mock_revision = Mock()
        mock_revision.timestamp = time.mktime(tip_date.timetuple())
        self.mock_branch.nick = 'config'
        self.mock_branch.repository.get_revision.return_value = mock_revision

        version_info = get_version_info('config')

        expected = {'name': 'config', 'version': 'config@42 (13.06.24)'}
        self.assertEqual(version_info, expected)

    def test_preflight_basic_auth(self):
        auth = self.email, self.password
        encoded = ('%s:%s' % auth).encode('base64').replace('\n', '')
        r = self.client.get(self.URL, HTTP_AUTHORIZATION='Basic %s' % encoded)

        self.assertTemplateUsed(r, "preflight/overview.html")
