# Copyright 2013 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

from django.db.utils import DEFAULT_DB_ALIAS
from django.conf import settings
from django.contrib.sessions.models import Session
from django.core.exceptions import ImproperlyConfigured
from django.test import SimpleTestCase

from identityprovider.models.emailaddress import EmailAddress
from identityprovider.models.person import Person
from identityprovider.routers import ReadOnlyModeRouter, NOP_DB_ALIAS


class ReadOnlyModeRouterConfigTestCase(SimpleTestCase):

    def test_raises_exception_if_nop_db_is_missing(self):
        with self.settings(DATABASES={}):
            self.assertRaises(ImproperlyConfigured, ReadOnlyModeRouter)

    def test_raises_exception_if_wrong_db_engine(self):
        with self.settings(DATABASES={'nop': {'ENGINE': 'wrong'}}):
            self.assertRaises(ImproperlyConfigured, ReadOnlyModeRouter)

    def test_without_slaves_configured_default_is_returned(self):
        with self.settings(DATABASE_SLAVES=[]):
            router = ReadOnlyModeRouter()
            self.assertEqual(router.slaves.next(), DEFAULT_DB_ALIAS)


class ReadOnlyModeRouterRWTestCase(SimpleTestCase):

    def setUp(self):
        super(ReadOnlyModeRouterRWTestCase, self).setUp()
        overrides = self.settings(DATABASE_SLAVES=['s1', 's2'])
        overrides.enable()
        self.addCleanup(overrides.disable)

        self.router = ReadOnlyModeRouter()

    def test_allow_migrate(self):
        self.assertTrue(self.router.allow_migrate(DEFAULT_DB_ALIAS, None))
        for slave in settings.DATABASE_SLAVES:
            self.assertFalse(self.router.allow_migrate(slave, None))

    def test_allow_relation(self):
        self.assertTrue(self.router.allow_relation(None, None))

    def test_db_for_write(self):
        self.assertEqual(self.router.db_for_write(Person),
                         DEFAULT_DB_ALIAS)

    def test_db_for_read(self):
        for slave in settings.DATABASE_SLAVES:
            self.assertEqual(self.router.db_for_read(Person), slave)

    def test_db_for_read_when_no_slaves(self):
        with self.settings(DATABASE_SLAVES=[]):
            router = ReadOnlyModeRouter()

        self.assertEqual(router.db_for_read(Person), DEFAULT_DB_ALIAS)


class ReadOnlyModeRouterROTestCase(SimpleTestCase):

    def setUp(self):
        super(ReadOnlyModeRouterROTestCase, self).setUp()
        writable_models = [
            'sessions.Session',
            'identityprovider.Account',
        ]
        overrides = self.settings(
            READ_ONLY_MODE=True,
            WRITABLE_RO_MODE_MODELS=writable_models
        )
        overrides.enable()
        self.addCleanup(overrides.disable)

        self.router = ReadOnlyModeRouter()

    def test_db_for_write_for_django_session(self):
        self.assertEqual(self.router.db_for_write(Session), DEFAULT_DB_ALIAS)

    def test_db_for_write_for_email_address(self):
        self.assertEqual(self.router.db_for_write(EmailAddress), NOP_DB_ALIAS)
