# -*- coding: utf-8 -*-

# Copyright 2013 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

from django.conf import settings
from django.contrib.auth.signals import user_logged_in
from django.core.urlresolvers import reverse
from django.db.models.signals import post_save
from django.utils.timezone import now

from mock import Mock, patch

from identityprovider.const import SESSION_TOKEN_KEY, SESSION_TOKEN_NAME
from identityprovider.models.const import AuthLogType, TokenScope
from identityprovider.signals import (
    invalidate_account_oauth_tokens,
    login_failed,
    log_login_failed,
    log_login_succeeded,
    set_session_oauth_token,
    track_failed_login,
    login_succeeded
)
from identityprovider.readonly import ReadOnlyManager
from identityprovider.tests import DEFAULT_USER_PASSWORD
from identityprovider.tests.test_auth import AuthLogTestCaseMixin
from identityprovider.tests.utils import SSOBaseTestCase


class SessionTokenOnLoginTestCase(SSOBaseTestCase):

    def setUp(self):
        super(SessionTokenOnLoginTestCase, self).setUp()
        self.email = 'test@canonical.com'
        self.account = self.factory.make_account(email=self.email)

    def test_signal_connected(self):
        # r[1] is a weak ref
        registered_functions = [r[1]() for r in user_logged_in.receivers]
        self.assertIn(set_session_oauth_token, registered_functions)

    def test_listener_default(self):
        assert self.account.oauth_tokens().count() == 0

        self.client.login(username=self.email, password=DEFAULT_USER_PASSWORD)
        oauth_tokens = self.account.token_set.filter(
            scope=TokenScope.WEB_LOGIN).all()
        self.assertEqual(oauth_tokens.count(), 1)
        self.assertEqual(
            self.client.session.get(SESSION_TOKEN_KEY), oauth_tokens[0].key)

    def test_listener_token_name_exists_wrong_scope(self):
        self.account.create_oauth_token(
            token_name=SESSION_TOKEN_NAME, scope=TokenScope.VERSION_2)

        assert self.account.token_set.count() == 1
        assert self.account.token_set.filter(
            scope=TokenScope.WEB_LOGIN).count() == 0

        self.client.login(username=self.email, password=DEFAULT_USER_PASSWORD)
        oauth_tokens = self.account.token_set.filter(
            scope=TokenScope.WEB_LOGIN).all()
        self.assertEqual(oauth_tokens.count(), 1)
        self.assertEqual(SESSION_TOKEN_NAME + '1', oauth_tokens[0].name)
        self.assertEqual(
            self.client.session.get(SESSION_TOKEN_KEY), oauth_tokens[0].key)

    def test_listener_wrong_token_name_but_correct_scope(self):
        token = self.account.create_oauth_token(
            token_name='foo bar baz', scope=TokenScope.WEB_LOGIN)

        assert self.account.token_set.count() == 1
        assert self.account.token_set.filter(
            scope=TokenScope.WEB_LOGIN).count() == 1

        self.client.login(username=self.email, password=DEFAULT_USER_PASSWORD)
        oauth_tokens = self.account.token_set.filter(
            scope=TokenScope.WEB_LOGIN).all()
        self.assertEqual(oauth_tokens.count(), 1)
        self.assertEqual(token, oauth_tokens[0])
        self.assertEqual(
            self.client.session.get(SESSION_TOKEN_KEY), oauth_tokens[0].key)

    def test_listener_checks_for_token_in_header(self):
        assert self.client.session.get(SESSION_TOKEN_KEY) is None

        url = reverse('login_by_token')
        token = self.factory.make_oauth_token(account=self.account)
        oauth_header = self.factory.make_authorization_header_from_token(
            url, token)

        self.client.get(url, **oauth_header)

        oauth_tokens = self.account.oauth_tokens()
        self.assertIn(token, oauth_tokens)
        self.assertEqual(
            self.client.session.get(SESSION_TOKEN_KEY), token.key)

    def test_listener_read_only_mode_no_token(self):
        rm = ReadOnlyManager()
        rm.set_readonly()
        self.addCleanup(rm.clear_readonly)

        self.client.login(username=self.email, password=DEFAULT_USER_PASSWORD)
        self.assertEqual(self.account.token_set.all().count(), 0)
        self.assertEqual(self.client.session.get(SESSION_TOKEN_KEY), '')

    def test_listener_read_only_mode_previous_token_no_web_login(self):
        self.account.create_oauth_token(token_name=SESSION_TOKEN_NAME)

        rm = ReadOnlyManager()
        rm.set_readonly()
        self.addCleanup(rm.clear_readonly)

        self.client.login(username=self.email, password=DEFAULT_USER_PASSWORD)
        self.assertEqual(
            self.client.session.get(SESSION_TOKEN_KEY), '')

    def test_listener_read_only_mode_previous_token(self):
        token = self.account.create_oauth_token(
            token_name=SESSION_TOKEN_NAME, scope=TokenScope.WEB_LOGIN)

        rm = ReadOnlyManager()
        rm.set_readonly()
        self.addCleanup(rm.clear_readonly)

        self.client.login(username=self.email, password=DEFAULT_USER_PASSWORD)
        self.assertEqual(
            self.client.session.get(SESSION_TOKEN_KEY), token.key)


class InvalidateOauthTokenOnPasswordChangeTestCase(SSOBaseTestCase):

    def test_signal_connected(self):
        # r[1] is a weak ref
        registered_functions = [r[1]() for r in post_save.receivers]
        self.assertIn(invalidate_account_oauth_tokens, registered_functions)

    def test_listener_on_password_change(self):
        account = self.factory.make_account()
        account.create_oauth_token('foo')
        account.create_oauth_token('bar', scope=TokenScope.WEB_LOGIN)
        assert account.token_set.all().count() == 2

        account.set_password('fooBar123')
        self.assertEqual(account.oauth_tokens().count(), 0)

    def test_listener_on_last_leak_check_change(self):
        # In this case we don't want the account's tokens to change at all.
        account = self.factory.make_account()
        account.create_oauth_token('foo')
        account.create_oauth_token('bar', scope=TokenScope.WEB_LOGIN)
        assert account.token_set.all().count() == 2

        old_oauth_tokens = account.oauth_tokens()
        account.accountpassword.last_leak_check = now()
        account.accountpassword.save(update_fields=['last_leak_check'])
        self.assertItemsEqual(account.oauth_tokens(), old_oauth_tokens)


class TrackFailedLoginTestCase(SSOBaseTestCase):
    def test_signal_connected(self):
        registered_functions = [r[1]() for r in login_failed.receivers]
        self.assertIn(track_failed_login, registered_functions)

    def test_increment_count_on_failed_login_with_forwarded_header(self):
        sender = Mock()
        credentials = dict(email='foo@foo.com', password='foo')
        request = Mock()
        request.META = {'HTTP_X_FORWARDED_FOR': '10.1.2.3, 10.4.5.6'}

        with patch('identityprovider.signals.cache') as mock_cache:
            mock_cache.get.return_value = 0
            track_failed_login(sender, credentials, request)

        mock_cache.get.assert_called_once_with('failed-login-10.1.2.3', 0)
        mock_cache.set.assert_called_once_with(
            'failed-login-10.1.2.3', 1, settings.FAILED_LOGIN_WINDOW)

    def test_increment_count_on_failed_login_without_forwarded_header(self):
        sender = Mock()
        credentials = dict(email='foo@foo.com', password='foo')
        request = Mock()
        request.META = {'REMOTE_ADDR': '10.1.2.3'}

        with patch('identityprovider.signals.cache') as mock_cache:
            mock_cache.get.return_value = 0
            track_failed_login(sender, credentials, request)

        mock_cache.get.assert_called_once_with('failed-login-10.1.2.3', 0)
        mock_cache.set.assert_called_once_with(
            'failed-login-10.1.2.3', 1, settings.FAILED_LOGIN_WINDOW)

    def test_log_on_max_failed_logins(self):
        sender = Mock()
        credentials = dict(email='foo@foo.com', password='foo')
        request = Mock()
        request.META = {'HTTP_X_FORWARDED_FOR': '1.2.3.4'}

        with patch('identityprovider.signals.logger') as mock_logger, \
                patch('identityprovider.signals.cache') as mock_cache:
            mock_cache.get.return_value = (
                settings.MAX_FAILED_LOGIN_ATTEMPTS - 1)
            track_failed_login(sender, credentials, request)

        mock_logger.warning.assert_called_once_with(
            'Too many failed login attempts. count=%s ip=%s',
            settings.MAX_FAILED_LOGIN_ATTEMPTS, '1.2.3.4')


class AuthLoggingTestCase(SSOBaseTestCase, AuthLogTestCaseMixin):

    login_email = "test@canonical.com"
    login_password = DEFAULT_USER_PASSWORD

    def setUp(self):
        super(AuthLoggingTestCase, self).setUp()
        self.account = self.factory.make_account(email=self.login_email,
                                                 password=self.login_password)

    def test_signal_connected(self):
        registered_functions_login_failed = [
            r[1]() for r in login_failed.receivers]
        registered_functions_login_succeeded = [
            r[1]() for r in login_succeeded.receivers]
        self.assertIn(log_login_succeeded,
                      registered_functions_login_succeeded)
        self.assertIn(log_login_failed,
                      registered_functions_login_failed)

    def test_log_on_failed_login(self):
        sender = Mock()
        credentials = dict(email='foo@foo.com', password='foo')
        request = self.factory.make_request('login', method='POST')
        request.META = {'REMOTE_ADDR': '10.1.2.3',
                        'HTTP_REFERER': 'http://referred.example.com',
                        'HTTP_USER_AGENT': 'Mothra 1.0'}
        expected_values = {
            'login_email': 'foo@foo.com',
            'log_type': AuthLogType.OPENID_FAIL,
            'referer': request.META['HTTP_REFERER'],
            'user_agent': request.META['HTTP_USER_AGENT'],
            'remote_ip': request.META['REMOTE_ADDR'],
            'account': None
        }
        with self.assert_authlog_matches(expected_values):
            log_login_failed(sender, credentials, request)

    def test_log_on_failed_login_api(self):
        """Ensure the login_failed signal honors the authlogtype parameter."""
        sender = Mock()
        credentials = dict(email='foo@foo.com', password='foo')
        request = self.factory.make_request('login', method='POST')
        request.META = {'REMOTE_ADDR': '10.1.2.3',
                        'HTTP_REFERER': 'http://referred.example.com',
                        'HTTP_USER_AGENT': 'Mothra 1.0'}
        expected_values = {
            'login_email': 'foo@foo.com',
            'log_type': AuthLogType.API_FAIL,
            'referer': request.META['HTTP_REFERER'],
            'user_agent': request.META['HTTP_USER_AGENT'],
            'remote_ip': request.META['REMOTE_ADDR'],
            'account': None
        }
        with self.assert_authlog_matches(expected_values):
            log_login_failed(
                sender, credentials, request, authlogtype=AuthLogType.API_FAIL)

    def test_log_on_successful_new_login(self):
        sender = Mock()
        request = self.factory.make_request('login', method='POST')
        request.META = {'REMOTE_ADDR': '10.1.2.3',
                        'HTTP_REFERER': 'http://referred.example.com',
                        'HTTP_USER_AGENT': 'Mothra 1.0'}
        expected_values = {
            'login_email': 'foo@bar.com',
            'log_type': AuthLogType.OPENID_NEW,
            'referer': request.META['HTTP_REFERER'],
            'user_agent': request.META['HTTP_USER_AGENT'],
            'remote_ip': request.META['REMOTE_ADDR'],
            'account': self.account.id
        }
        with self.assert_authlog_matches(expected_values):
            log_login_succeeded(
                sender, self.account, request, email='foo@bar.com')

    def test_log_on_successful_existing_login(self):
        sender = Mock()
        request = self.factory.make_request('login', method='POST')
        request.META = {'REMOTE_ADDR': '10.1.2.3',
                        'HTTP_REFERER': 'http://referred.example.com',
                        'HTTP_USER_AGENT': 'Mothra 1.0'}
        expected_values = {
            'login_email': 'foo@bar.com',
            'log_type': AuthLogType.OPENID_EXISTING,
            'referer': request.META['HTTP_REFERER'],
            'user_agent': request.META['HTTP_USER_AGENT'],
            'remote_ip': request.META['REMOTE_ADDR'],
            'account': self.account.id
        }

        with self.assert_authlog_matches(expected_values):
            log_login_succeeded(
                sender, self.account, request, email='foo@bar.com',
                authlogtype=AuthLogType.OPENID_EXISTING)
