# Copyright 2010 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

from oauth.oauth import OAuthDataStore

from identityprovider.models.const import AccountStatus
from identityprovider.store import SSODataStore
from identityprovider.tests.utils import SSOBaseTestCase


class SSODataStoreTestCase(SSOBaseTestCase):

    def setUp(self):
        super(SSODataStoreTestCase, self).setUp()
        self.ds = SSODataStore()
        self.consumer = self.factory.make_account()
        self.token = self.consumer.create_oauth_token('test')

    def test_data_store_creation_piston_compatible(self):
        store = SSODataStore(oauth_request=object())
        self.assertIsInstance(store, OAuthDataStore)

    def test_lookup_token_wrong_type(self):
        self.assertRaises(
            AssertionError, self.ds.lookup_token, 'value', self.token.key)

    def test_lookup_token_exists(self):
        token = self.ds.lookup_token('access', self.token.key)
        self.assertEqual(token, self.token)

    def test_lookup_token_has_expected_attributes(self):
        token = self.ds.lookup_token('access', self.token.key)
        self.assertEqual(token.key, self.token.key)
        self.assertEqual(token.secret, self.token.secret)
        self.assertEqual(token.user, self.consumer)

    def test_lookup_token_not_exists(self):
        token_key = self.token.key
        self.token.delete()
        token = self.ds.lookup_token('access', token_key)
        self.assertIsNone(token)

    def test_lookup_consumer_exists(self):
        result = self.ds.lookup_consumer(self.consumer.key)
        self.assertEqual(result, self.consumer)

    def test_lookup_consumer_not_exists(self):
        consumer = self.ds.lookup_consumer('foo')
        self.assertIsNone(consumer)

    def test_lookup_consumer_inactive(self):
        for status, _ in AccountStatus._get_choices():
            self.consumer.status = status
            self.consumer.save()

            result = self.ds.lookup_consumer(self.consumer.key)

            if status == AccountStatus.ACTIVE:
                self.assertEqual(result, self.consumer)
            else:
                self.assertIsNone(result)

    def test_lookup_nonce_does_not_exist(self):
        nonce = self.ds.lookup_nonce(self.consumer, self.token, 'nonce')
        self.assertIsNone(nonce)

    def test_lookup_nonce_does_exist(self):
        nonce = self.ds.lookup_nonce(self.consumer, self.token, 'nonce')
        assert nonce is None

        nonce = self.ds.lookup_nonce(self.consumer, self.token, 'nonce')
        self.assertEqual(nonce, 'nonce')

    def test_lookup_nonce_expiration_default(self):
        self.ds.DEFAULT_NONCE_TIMEOUT = 0

        nonce = self.ds.lookup_nonce(self.consumer, self.token, 'nonce')
        self.assertIsNone(nonce)

        nonce = self.ds.lookup_nonce(self.consumer, self.token, 'nonce')
        self.assertIsNone(nonce)

    def test_lookup_nonce_expiration_setting(self):
        self.ds.DEFAULT_NONCE_TIMEOUT = 10

        with self.settings(OAUTH_NONCE_TIMEOUT=0):
            nonce = self.ds.lookup_nonce(self.consumer, self.token, 'nonce')
        self.assertIsNone(nonce)

        nonce = self.ds.lookup_nonce(self.consumer, self.token, 'nonce')
        self.assertIsNone(nonce)

    def test_lookup_nonce_expiration_setting_wins_over_default(self):
        self.ds.DEFAULT_NONCE_TIMEOUT = 0

        with self.settings(OAUTH_NONCE_TIMEOUT=100):
            nonce = self.ds.lookup_nonce(self.consumer, self.token, 'nonce')
        self.assertIsNone(nonce)

        nonce = self.ds.lookup_nonce(self.consumer, self.token, 'nonce')
        self.assertEqual(nonce, 'nonce')

    def test_cache_key_unique_per_consumer(self):
        nonce = self.ds.lookup_nonce(self.consumer, self.token, 'nonce')
        self.assertIsNone(nonce)

        nonce = self.ds.lookup_nonce(
            self.factory.make_account(), self.token, 'nonce')
        self.assertIsNone(nonce)

    def test_cache_key_unique_per_token(self):
        nonce = self.ds.lookup_nonce(self.consumer, self.token, 'nonce')
        self.assertIsNone(nonce)

        token = self.consumer.create_oauth_token('other token')
        nonce = self.ds.lookup_nonce(self.consumer, token, 'nonce')
        self.assertIsNone(nonce)

    def test_cache_key_unique_per_nonce(self):
        nonce = self.ds.lookup_nonce(self.consumer, self.token, 'nonce')
        self.assertIsNone(nonce)

        nonce = self.ds.lookup_nonce(self.consumer, self.token, 'nonci')
        self.assertIsNone(nonce)
