# Copyright 2010 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).
import os
import random
from datetime import date, datetime, time, timedelta
from uuid import uuid4

from django.conf import settings
from django.contrib.sessions.models import Session
from django.core.exceptions import SuspiciousOperation
from django.test import TestCase
from django.test.utils import override_settings

from identityprovider import utils
from identityprovider.models.person import Person
from identityprovider.utils import (
    canonical_url,
    clean_sessions_until,
    generate_random_string,
    polite_form_errors,
    redirection_url_for_token,
    validate_launchpad_password,
)
from identityprovider.tests.utils import SSOBaseTestCase
from identityprovider.utils import get_provider_url


class CanonicalUrlTestCase(SSOBaseTestCase):

    def test_when_object_is_not_person(self):
        self.assertTrue(canonical_url(None) is None)

    def test_when_view_name_is_none(self):
        url = canonical_url(Person(name="test"), view_name=None)
        self.assertEqual(url, "https://launchpad.net/~test")

    def test_when_view_name_is_not_none(self):
        url = canonical_url(Person(name="test"), view_name="TEST")
        self.assertEqual(url, "https://launchpad.net/~test/TEST")


class PoliteFormErrorsTestCase(SSOBaseTestCase):

    def test_polite_errors(self):
        text = "Enter a valid e-mail address."
        errors = {'email': [text]}

        polite_form_errors(errors)

        self.assertTrue(errors['email'][0] != text)


class ValidateLaunchpadPasswordTestCase(SSOBaseTestCase):

    def test_when_there_is_binascii_error_is_raised(self):
        self.assertFalse(validate_launchpad_password('a', 'b'))


class RedirectionURLForTokenTestCase(SSOBaseTestCase):

    def test_if_token_is_none(self):
        url = redirection_url_for_token(None)
        self.assertEqual(url, "/")

    def test_if_token_is_not_none(self):
        token = "ABCDEFGH" * 2
        url = redirection_url_for_token(token)
        self.assertEqual(url, "/%s/+decide" % token)


class SessionTestCase(SSOBaseTestCase):

    def create_session(self, expire_date):
        return Session.objects.create(
            session_key=str(uuid4()),
            session_data='some data',
            expire_date=expire_date,
        )

    def test_deletes_sessions_created_until_provided_date(self):
        creation_date = date.today()
        expire_date = creation_date + timedelta(
            seconds=settings.SESSION_COOKIE_AGE)

        self.create_session(expire_date)
        self.create_session(expire_date - timedelta(days=1))
        self.create_session(expire_date - timedelta(days=2))
        session4 = self.create_session(expire_date + timedelta(days=1))

        with self.assertNumQueries(1):
            clean_sessions_until(creation_date)

        sessions = list(Session.objects.all())

        self.assertEqual(sessions, [
            session4,
        ])

    def test_deletes_sessions_with_limit(self):
        creation_date = date.today()
        expire_date = creation_date + timedelta(
            seconds=settings.SESSION_COOKIE_AGE)

        self.create_session(expire_date)
        self.create_session(expire_date - timedelta(days=1))
        session3 = self.create_session(expire_date - timedelta(days=2))
        session4 = self.create_session(expire_date + timedelta(days=1))
        session5 = self.create_session(expire_date + timedelta(days=2))

        with self.assertNumQueries(1):
            clean_sessions_until(creation_date, limit=2)

        sessions = list(Session.objects.all())

        self.assertEqual(sessions, [
            session3,
            session4,
            session5,
        ])

    def test_deletes_sessions_until_exact_date_and_not_a_second_more(self):
        creation_date = date.today()
        expire_date = creation_date + timedelta(
            seconds=settings.SESSION_COOKIE_AGE)

        self.create_session(expire_date)
        self.create_session(expire_date - timedelta(days=1))
        self.create_session(expire_date - timedelta(days=2))
        session4 = self.create_session(datetime.combine(expire_date,
                                                        time(second=1)))

        with self.assertNumQueries(1):
            clean_sessions_until(creation_date)

        sessions = list(Session.objects.all())

        self.assertEqual(sessions, [
            session4,
        ])

    def test_doesnt_delete_sessions_if_dry_running(self):
        creation_date = date.today()
        expire_date = creation_date + timedelta(
            seconds=settings.SESSION_COOKIE_AGE)

        session1 = self.create_session(expire_date)
        session2 = self.create_session(expire_date - timedelta(days=1))
        session3 = self.create_session(expire_date - timedelta(days=2))
        session4 = self.create_session(expire_date + timedelta(days=1))

        with self.assertNumQueries(1):
            clean_sessions_until(creation_date, dry_run=True)

        sessions = list(Session.objects.all())

        self.assertEqual(sessions, [
            session1,
            session2,
            session3,
            session4,
        ])


@override_settings(ALLOWED_HOSTS=['domain1.org', 'testserver'])
class GetProviderURLTestCase(SSOBaseTestCase):

    def test_get_provider_url_from_request(self):
        kwargs = {
            'HTTP_HOST': 'domain1.org',
            'wsgi.url_scheme': 'https',
        }
        request = self.factory.make_request(**kwargs)
        url = get_provider_url(request)
        self.assertEqual(url, 'https://domain1.org/+openid')

    def test_get_provider_url_invalid_domain(self):
        kwargs = {
            'HTTP_HOST': 'example.org',
            'wsgi.url_scheme': 'http',
        }
        request = self.factory.make_request(**kwargs)
        with self.assertRaises(SuspiciousOperation):
            get_provider_url(request)


class BlacklistedPasswordTestCase(SSOBaseTestCase):

    def setUp(self):
        super(BlacklistedPasswordTestCase, self).setUp()
        # Tests will override with various sets, make sure we cleanup
        self.patch(utils, 'blacklisted_password_set', new=set())

    @override_settings(PASSWORD_BLACKLIST_FILE="bl_test_filename")
    def test_populate_password_blacklist(self):
        raw_pass_data = ("cthulhu50\nr'lyeh25\ndunwich600\nDunWich600"
                         "\n\r        \r\n   e   \nCthulhu50")
        self.addCleanup(os.unlink, settings.PASSWORD_BLACKLIST_FILE)
        with open(settings.PASSWORD_BLACKLIST_FILE, "w") as bl_file:
            bl_file.write(raw_pass_data)
        expected_passwords = {
            'cthulhu50', "r'lyeh25", 'dunwich600', '        ', '   e   '}
        utils.populate_password_blacklist()
        self.assertEqual(utils.get_password_blacklist(),
                         expected_passwords)

    def test_password_is_not_blacklisted(self):
        utils.blacklisted_password_set = set(['bogus', 'blacklist'])
        self.assertFalse(utils.password_is_blacklisted('goodpass'))

    def test_password_is_blacklisted(self):
        utils.blacklisted_password_set = set(['bogus', 'blacklist'])
        self.assertTrue(utils.password_is_blacklisted('blacklist'))

    def test_password_is_blacklisted_variable_case(self):
        utils.blacklisted_password_set = set(['bogus', 'blacklist'])
        self.assertTrue(utils.password_is_blacklisted('BlackList'))


class GenerateRandomStringTestCase(TestCase):

    def test_random_seed_not_used(self):
        # SystemRandom doesn't honor the seed() method, so this should
        # test we're not using the weak RNG by default.
        random.seed(7337)
        string1 = generate_random_string(10)
        random.seed(7337)
        string2 = generate_random_string(10)
        self.assertNotEqual(string1, string2)

    def test_specify_charset(self):
        # Charset deliberately contains only characters not in utils.ALPHA_NUM
        charset = '!@#$%^&*()'
        string1 = generate_random_string(200, charset=charset)
        self.assertTrue(all([ch in charset for ch in string1]))
