from unittest import TestCase

from django.core.exceptions import ValidationError
from mock import Mock, sentinel

from identityprovider.tests import DEFAULT_USER_PASSWORD
from identityprovider import utils
from identityprovider.tests.utils import SSOBaseTestCase, TimelineActionMixin
from identityprovider.validators import (
    BasicValidator,
    CanonicalValidator,
    Errors,
    get_password_validator,
)


ACCOUNT, PASSWORD = sentinel.account, sentinel.password


class PasswordValidatorTestsMixin(object):

    def test_password_too_short(self):
        with self.assertRaises(ValidationError):
            self.validator('abcd3Fg')

    def test_password_invalid_chars(self):
        with self.assertRaises(ValidationError):
            self.validator(u'abcd\xe1Fgh')

    def test_password_none(self):
        with self.assertRaises(ValidationError):
            self.validator(None)

    def test_default_password(self):
        self.assertEqual(self.validator(DEFAULT_USER_PASSWORD), None)

    def test_valid_password(self):
        self.assertEqual(self.validator('abcD3fgh'), None)


class PasswordPolicyValidatorTestCase(PasswordValidatorTestsMixin, TestCase):

    validator = get_password_validator()


class BasicValidatorPolicyTestCase(PasswordValidatorTestsMixin, TestCase):

    validator = BasicValidator()

    def test_match(self):
        validator = BasicValidator()
        result = validator.match()
        self.assertEqual(result, True)


class ValidatorWithRequestTimingTestCase(SSOBaseTestCase, TimelineActionMixin):

    validator = BasicValidator()

    def test_validate_leaked_password_uses_timeline(self):
        a_password = "foobar1274"
        request = self.factory.make_request(timeline=True)
        account = self.factory.make_account(password=a_password)
        # Add a leak so the validator goes into the code path that
        # checks it using the timer
        self.factory.make_leaked_credential(
            account.preferredemail.email, a_password, source='me')
        validator = BasicValidator(account=account, request=request)
        with self.assertRaises(ValidationError):
            validator(a_password)
        self.assert_timeline_contains(request=request,
                                      category='django-password-checking',
                                      detail=str(account.id))


class CanonicalValidatorTestCase(PasswordValidatorTestsMixin, SSOBaseTestCase):

    validator = CanonicalValidator()

    def test_password_too_short_8_chars(self):
        with self.assertRaises(ValidationError):
            self.validator('abcd3Fg0')

    def test_password_too_short_9_chars(self):
        with self.assertRaises(ValidationError):
            self.validator('abcd3Fg0_')

    def test_invalid_missing_uppercase_and_digits(self):
        with self.assertRaises(ValidationError):
            self.validator('abcdefghijklmn')

    def test_invalid_missing_digits(self):
        with self.assertRaises(ValidationError):
            self.validator('abcdefGHijklmn')

    def test_invalid_missing_uppercase(self):
        with self.assertRaises(ValidationError):
            self.validator('abcdef12ijklmn')

    def test_default_password(self):
        with self.assertRaises(ValidationError):
            self.validator(DEFAULT_USER_PASSWORD)

    def test_valid_password(self):
        self.assertEqual(self.validator('abcdefA1xx'), None)

    def test_match_none(self):
        validator = CanonicalValidator()
        self.assertEqual(validator.match(), False)

    def test_match_account_not_in_team(self):
        account = Mock()
        account.person_in_team.return_value = False

        validator = CanonicalValidator(account=account)
        self.assertEqual(validator.match(), False)
        account.person_in_team.assert_called_once_with('canonical')

    def test_match_account_in_team(self):
        account = Mock()
        account.person_in_team.return_value = True

        validator = CanonicalValidator(account=account)
        self.assertEqual(validator.match(), True)
        account.person_in_team.assert_called_once_with('canonical')

    def test_blacklisted_password_case_insensitive(self):
        """Passwords in the blacklist should be rejected (case-insensitive)."""
        self.patch(utils, 'blacklisted_password_set',
                   new=set(['bogus', 'blacklist2']))
        with self.assertRaises(ValidationError) as ctx:
            self.validator('blacklIst2')
        self.assertIn("it is very common and susceptible to attack",
                      str(ctx.exception))

    def test_blacklisted_password_not_short_length(self):
        """Blacklist should not be used if Validator rejects it first."""
        self.patch(utils, 'blacklisted_password_set',
                   new=set(['bogus', 'blacklist2']))
        with self.assertRaises(ValidationError) as ctx:
            self.validator('bogus')
        self.assertNotIn("it is very common and susceptible to attack",
                         str(ctx.exception))


class PasswordValidatorTestCase(SSOBaseTestCase):

    def test_canonical_validator_matches_first(self):
        mock_canonical_validator_match = self.patch(
            'identityprovider.validators.CanonicalValidator.match',
            return_value=True)
        mock_basic_validator_match = self.patch(
            'identityprovider.validators.BasicValidator.match',
            return_value=True)

        validator = get_password_validator(account=ACCOUNT)

        self.assertIsInstance(validator, CanonicalValidator)
        mock_canonical_validator_match.assert_called_once_with()
        self.assertFalse(mock_basic_validator_match.called)

    def test_basic_validator_used_if_canonical_does_not_match(self):
        mock_canonical_validator_match = self.patch(
            'identityprovider.validators.CanonicalValidator.match',
            return_value=False)
        mock_basic_validator_match = self.patch(
            'identityprovider.validators.BasicValidator.match',
            return_value=True)

        validator = get_password_validator(account=ACCOUNT)

        self.assertIsInstance(validator, BasicValidator)
        mock_canonical_validator_match.assert_called_once_with()
        mock_basic_validator_match.assert_called_once_with()

    def test_validate_password_policy_fails(self):
        self.patch(
            'identityprovider.validators.CanonicalValidator.match',
            return_value=False)
        self.patch(
            'identityprovider.validators.BasicValidator.match',
            return_value=False)

        with self.assertRaises(ValidationError):
            get_password_validator(ACCOUNT)


class ErrorsTestCase(SSOBaseTestCase):

    def test_collect_no_key(self):
        errors = Errors()
        with errors.collect():
            raise ValidationError('Some Error')
        self.assertEqual(errors['__all__'][0].message, 'Some Error')

    def test_collect_no_key_merges_dict(self):
        errors = Errors()
        with errors.collect():
            raise ValidationError({'k': 'v', 'x': 'y'})
        self.assertEqual(errors['k'], ['v'])
        self.assertEqual(errors['x'], ['y'])

    def test_collect_with_key(self):
        errors = Errors()
        with errors.collect('key'):
            raise ValidationError('Some Error')
        self.assertEqual(errors['key'], ['Some Error'])

    def test_collect_with_key_flattens_dict(self):
        errors = Errors()
        with errors.collect('key'):
            raise ValidationError({'k': 'v', 'x': 'y'})
        self.assertEqual(sorted(errors['key']), ['v', 'y'])
