# Copyright 2010 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

from datetime import datetime
from django.conf import settings
from django.test import TestCase
from django.test.utils import override_settings

from identityprovider.tests.utils import SSOBaseTestCase
from identityprovider.widgets import (
    CommaSeparatedWidget,
    LPUsernameWidget,
    ReadOnlyDateTimeWidget,
    ROAwareSelect,
    ROAwareTextInput,
    StatusWidget,
)


class WidgetRenderingTestCaseBase(SSOBaseTestCase):

    def assert_renders_correctly(self, expected, *args, **kwargs):
        """Assert that 'self.widget''s rendered output equals 'expected'.

        The *args and **kwargs are passed to the widgets render method.

        This assertion works for lazy string proxies, which makes it more
        useful than assertHTMLEqual.
        """
        observed = self.widget.render(*args, **kwargs)
        self.assertHTMLEqual(unicode(observed), expected)


@override_settings(READ_ONLY_MODE=False)
class ROAwareTextInputTestCase(WidgetRenderingTestCaseBase):

    def setUp(self):
        super(ROAwareTextInputTestCase, self).setUp()
        self.widget = ROAwareTextInput()

    def test_render_when_readonly(self):
        settings.READ_ONLY_MODE = True

        self.assert_renders_correctly(
            '<span class="rofield"></span>',
            'test',
            None,
        )
        self.assert_renders_correctly(
            '<span class="rofield">value</span>',
            'test',
            'value',
        )

    def test_render_when_not_readonly(self):
        self.assert_renders_correctly(
            '<input type="text" name="test" />',
            'test',
            None,
        )
        self.assert_renders_correctly(
            '<input type="text" name="test" value="value" />',
            'test',
            'value',
        )


@override_settings(READ_ONLY_MODE=False)
class ROAwareSelectTestCase(WidgetRenderingTestCaseBase):

    def setUp(self):
        super(ROAwareSelectTestCase, self).setUp()
        self.choices = (('1', 'One'), ('2', 'Two'))
        self.widget = ROAwareSelect()

    def test_render_when_readonly(self):
        settings.READ_ONLY_MODE = True

        self.assert_renders_correctly(
            '<span class="rofield"></span>',
            'test',
            'value',
            choices=self.choices
        )

    def test_render_when_readonly_selected(self):
        settings.READ_ONLY_MODE = True

        choices = self.choices + (('value', 'The Value'),)
        self.assert_renders_correctly(
            '<span class="rofield">The Value</span>',
            'test',
            'value',
            choices=choices,
        )

    def test_render_when_not_readonly(self):
        expected = """<select name="test">
<option value="1">One</option>
<option value="2">Two</option>
</select>"""
        self.assert_renders_correctly(
            expected,
            'test',
            'value',
            choices=self.choices,
        )


class CommaSeparatedWidgetTestCase(TestCase):

    def setUp(self):
        super(CommaSeparatedWidgetTestCase, self).setUp()
        self.choices = (('1', 'One'), ('2', 'Two'), ('3', 'Three'))
        self.widget = CommaSeparatedWidget()

    def test_render_when_value_is_none(self):
        r = self.widget.render('test', None, choices=self.choices)
        self.assertTrue('checked' not in r)

    def test_render_when_value_is_list(self):
        r = self.widget.render('test', ['1', '2'], choices=self.choices)
        self.assertEqual(r.count('checked="checked"'), 2)

    def test_render_when_value_is_comma_separated_list(self):
        r = self.widget.render('test', "1,2", choices=self.choices)
        self.assertEqual(r.count('checked="checked"'), 2)


class StatusWidgetTestCase(WidgetRenderingTestCaseBase):

    def setUp(self):
        super(StatusWidgetTestCase, self).setUp()
        self.choices = (('1', 'One'), ('2', 'Two'))
        self.widget = StatusWidget()

    def test_render_not_date_status_set(self):
        expected = """<select name="test">
<option value="1">One</option>
<option value="2">Two</option>
</select> """
        self.assert_renders_correctly(
            expected,
            'test',
            'value',
            choices=self.choices,
        )

    def test_render_date_status_set(self):
        self.widget.date_status_set = datetime(2010, 01, 01)
        expected = """<select name="test">
<option value="1">One</option>
<option value="2">Two</option>
</select> Set on 2010-01-01 00:00:00"""
        self.assert_renders_correctly(
            expected,
            'test',
            'value',
            choices=self.choices,
        )


class ReadOnlyDateTimeWidgetTestCase(TestCase):

    def setUp(self):
        super(ReadOnlyDateTimeWidgetTestCase, self).setUp()
        self.widget = ReadOnlyDateTimeWidget()

    def test_render_string(self):
        self.widget.value = 'value'
        r = self.widget.render('test', None)
        self.assertEqual(r, 'value')

    def test_render_int(self):
        self.widget.value = 4
        r = self.widget.render('test', None)
        self.assertEqual(r, '4')

    def test_render_bool(self):
        self.widget.value = False
        r = self.widget.render('test', None)
        self.assertEqual(r, 'False')

    def test_render_datetime(self):
        self.widget.value = datetime(2010, 01, 01)
        r = self.widget.render('test', None)
        self.assertEqual(r, '2010-01-01 00:00:00')


class LPUsernameWidgetTestCase(WidgetRenderingTestCaseBase):

    def setUp(self):
        super(LPUsernameWidgetTestCase, self).setUp()
        self.widget = LPUsernameWidget()
        self.account = self.factory.make_account()
        assert self.account.person is None
        self.widget.account = self.account

    def test_render_account_with_person(self):
        self.factory.make_person(account=self.account)
        assert self.account.person is not None

        expected = '<a href="https://launchpad.net/~{name}">{name}</a>'.format(
            name=self.widget.account.person.name)
        self.assert_renders_correctly(
            expected,
            'test',
            None,
        )

    def test_render_no_account(self):
        # unlink account
        self.widget.account = None
        r = self.widget.render('test', None)
        self.assertEqual(r, '')

    def test_render_account_no_person(self):
        r = self.widget.render('test', None)
        self.assertEqual(r, '')
