# Copyright 2010-2013 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

import base64
import binascii
import httplib
import itertools
import json
import os
import threading
import urllib
import urllib2

from cStringIO import StringIO
from datetime import timedelta

import gargoyle

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.asymmetric import rsa
from django.conf import settings
from django.core.cache import cache
from django.core.management import call_command
from django.db import connection, OperationalError
from django.test import TestCase, TransactionTestCase
from django.utils.importlib import import_module
from gargoyle.models import SELECTIVE, Switch
from mock import Mock, patch
from openid.message import IDENTIFIER_SELECT
from pymacaroons import Macaroon
from pyquery import PyQuery

# import signals to ensure all the handlers are connected and tests use the
# same logic as prod systems
from identityprovider import crypto, signed, signals  # noqa
from identityprovider.macaroon import MacaroonRequest
from identityprovider.tests import DEFAULT_USER_PASSWORD
from identityprovider.tests.factory import SSOObjectFactory
from identityprovider.utils import generate_random_string
from identityprovider.views import server


# call autodiscover to ensure that the custom condition sets are registered
gargoyle.autodiscover()


MISSING_BACKUP_DEVICE = """
<div class="message warning" id="missing_backup_device">
  <p>
    We strongly recommend having two authentication devices,
    a <strong>primary</strong> device and a <strong>secondary</strong>
    or backup device.
    </br>
    Having two authentication devices means you can continue to access your
    account with your secondary device should your primary device be lost or
    stolen.
    </br>
    </br>
    Click to <a href="{add_device_link}">add a backup device</a>.
  </p>
</div>

"""

EXHAUSTED_WARNING = """Your printed list of backup codes is nearly used up.
  Please print a new list for the following devices."""


def assert_exausted_warning(testcase, devices, response):
    """Helper to assert warning is correctly displayed"""
    dom = PyQuery(response.content)
    warning = dom.find('#exhausted_warning')
    testcase.assertEqual(len(warning), 1)
    elems = warning.find('p')
    testcase.assertEqual(elems[0].text.strip(), EXHAUSTED_WARNING)
    testcase.assertEqual(len(devices), len(elems) - 1)
    for device, elem in zip(devices, elems[1:]):
        testcase.assertIn(device.name, elem.text)
        link = elem.find('a')
        testcase.assertNotEqual(link, None)
        testcase.assertIn("Generate New Codes", link.find('span').text)


class SSOBaseTestCaseMixin(object):

    factory = SSOObjectFactory()

    @property
    def sso_root_url(self):
        return self.factory.make_request().build_absolute_uri()

    def patch(self, target, *args, **kwargs):
        """Perform monkeypatching and clean up once test is done.

        Monkeypatch the given 'target' with whatever value is given in the
        'new' named parameter. If not 'new' is given, an instance of Mock()
        will be used by default.

        If 'target' is not an object, it's expected to be the Python import
        module patch describing the namespace to be monkeypatched.

        If 'target' is an object, the second paramenter has to be the attribute
        name of 'target' to be monkeypatched.

        """
        if 'new' not in kwargs:
            kwargs.setdefault('new_callable', Mock)
        if isinstance(target, basestring):
            # Mock target
            p = patch(target, *args, **kwargs)
        else:
            # Patch a target attribute
            p = patch.object(target, *args, **kwargs)

        # result will be of the same type as param 'new'
        result = p.start()
        self.addCleanup(p.stop)
        return result

    def isolate_memcache(self):
        """Remove entries added by a given test when cleaning up.

        Known use cases were leaving entries in the cache may affect other
        tests, causing them to fail because they assume a clean cache
        (including those that are using this method): rate limitation,
        invalid credentials.

        """
        # Tests can add entries into memcache. This makes it a shared resource
        # that needs special care, especially when running concurrently.  In
        # that case, subprocesses can access memcache concurrently and they are
        # protected from each other by defining a KEY_PREFIX using
        # os.getpid(). But no tests should ever call cache.clear() because this
        # removes *all* entries, including the ones from the other
        # subprocesses. Tracking the entries added by a test and deleting them
        # when cleaning up, provides the same result: all entries created by
        # the cache are removed, leaving the cache clean.

        # Sanity checks as stastd is already monkey-patching things. Failure at
        # this point will point to a change in the test environment.
        self.assertEqual('StatsdTracker', type(cache).__name__)
        true_cache = cache.cache
        self.assertEqual('django.core.cache', type(true_cache).__module__)
        self.assertEqual('DefaultCacheProxy', type(true_cache).__name__)
        # Record all keys added to the cache
        keys = set()
        # There are only a few methods we need to spy to capture all added keys
        orig_set = true_cache.set
        orig_add = true_cache.add

        def set_capturing_key(key, *args, **kwargs):
            keys.add(key)
            return orig_set(key, *args, **kwargs)

        def add_capturing_key(key, *args, **kwargs):
            keys.add(key)
            return orig_add(key, *args, **kwargs)

        # Install the spies that will record the added keys
        self.patch(true_cache, 'set', new=set_capturing_key)
        self.patch(true_cache, 'add', new=add_capturing_key)
        # To make the memcache log more readable, we insert our own key
        cache.set(self.id(), os.getpid())
        # Remove the added keys when cleaning up
        self.addCleanup(cache.delete_many, keys)

    def setup_key_pair(self, _cache=[]):
        """Build the key pairs (cached, as it takes quite some time now!)."""
        # get or build the keys
        if _cache:
            test_rsa_priv_key, test_rsa_pub_key = _cache
        else:
            priv_key = rsa.generate_private_key(
                public_exponent=65537, key_size=2048,
                backend=default_backend())
            pub_key = priv_key.public_key()
            test_rsa_priv_key = crypto.Crypto(priv_key)
            test_rsa_pub_key = crypto.Crypto(pub_key)
            _cache.append(test_rsa_priv_key)
            _cache.append(test_rsa_pub_key)

        # set up config for priv key
        p = self.settings(CRYPTO_SSO_PRIVKEY=test_rsa_priv_key)
        p.enable()
        self.addCleanup(p.disable)
        return test_rsa_priv_key, test_rsa_pub_key

    def build_macaroon(self, service_location=None, version=1):
        if service_location is None:
            service_location = settings.MACAROON_SERVICE_LOCATION
        test_rsa_priv_key, test_rsa_pub_key = self.setup_key_pair()

        # create a Macaroon with the proper third party caveat
        macaroon_random_key = binascii.hexlify(os.urandom(32))
        root_macaroon = Macaroon(
            location='The store ;)',
            key=macaroon_random_key,
            identifier='A test macaroon',
        )
        random_key = binascii.hexlify(os.urandom(32))
        if version == 0:
            info = {
                'roothash': macaroon_random_key,
                '3rdparty': random_key,
            }
            payload = base64.b64encode(
                test_rsa_pub_key.encrypt(json.dumps(info)))
        else:
            payload = json.dumps({
                'version': 1,
                'secret': base64.b64encode(
                    test_rsa_pub_key.encrypt(random_key)),
            })
        root_macaroon.add_third_party_caveat(
            service_location, random_key, payload)
        return root_macaroon, macaroon_random_key, random_key

    def get_sso_caveat(self, macaroon, service_location=None):
        """Extract the SSO caveat from a macaroon."""
        if service_location is None:
            service_location = settings.MACAROON_SERVICE_LOCATION
        (sso_caveat,) = [c for c in macaroon.third_party_caveats()
                         if c.location == service_location]
        return sso_caveat

    def build_macaroon_request(self):
        """Build a macaroon and a MacaroonRequest to discharge it."""
        root_macaroon, _, _ = self.build_macaroon(version=1)
        sso_caveat = self.get_sso_caveat(root_macaroon)
        return MacaroonRequest(sso_caveat.caveat_id)

    def use_fixture(self, fixture):
        """Set up 'fixture' to be used with the current test.

        This is a low fidelity backport of testtools useFixture for
        environments where testtools cannot be used.

        Note that this will discard any details the fixture provides, so debug
        logs won't be preserved.
        """
        try:
            fixture.setUp()
            self.addCleanup(fixture.cleanUp)
        except:
            fixture.cleanUp()
            raise
        return fixture

    def preseed_cookie_check(self):
        # Sets up the cookie usually set by SSO to validate the client has
        # cookies support. This avoids an extra redirect to /+cookie, then back
        # to the invoking URL, which can cause some tests to misbehave and/or
        # cause a RedirectCycleError from Django 1.8's overzealous redirect
        # loop detection.
        self.client.cookies['C'] = '1'


class SSOBaseTestCase(SSOBaseTestCaseMixin, TestCase):

    def setUp(self):
        super(SSOBaseTestCase, self).setUp()
        self.isolate_memcache()

    def get_query(self, response):
        query_start = response['Location'].find('?')
        query_str = response['Location'][query_start + 1:]
        query_items = [tuple(item.split('=')) for item in query_str.split('&')]
        query = dict(query_items)
        return query

    def _assign_token_to_rpconfig(self, rpconfig, token=None):
        if token is None:
            token = generate_random_string(16)
        request = {'openid.mode': 'checkid_setup',
                   'openid.trust_root': rpconfig.trust_root,
                   'openid.return_to': rpconfig.trust_root,
                   'openid.identity': IDENTIFIER_SELECT}
        openid_server = server._get_openid_server(request)
        orequest = openid_server.decodeRequest(request)
        session = self.client.session
        if isinstance(session, dict):
            # Workaround for https://code.djangoproject.com/ticket/11475
            engine = import_module(settings.SESSION_ENGINE)
            session = engine.SessionStore()
            session.save()
            session_cookie = settings.SESSION_COOKIE_NAME
            self.client.cookies[session_cookie] = session.session_key
            cookie_data = {
                'max-age': None,
                'path': '/',
                'domain': settings.SESSION_COOKIE_DOMAIN,
                'secure': settings.SESSION_COOKIE_SECURE or None,
                'expires': None,
            }
            self.client.cookies[session_cookie].update(cookie_data)
        session[token] = signed.dumps(orequest, settings.SECRET_KEY)
        session.save()
        return token

    def conditionally_enable_flag(self, key, field_name, value, condition_set,
                                  exclude=False):
        """Setup a gargoyle flag for 'key' in SELECTIVE status."""
        Switch.objects.create(key=key, status=SELECTIVE)
        switch = gargoyle.gargoyle[key]
        switch.add_condition(
            condition_set=condition_set,
            field_name=field_name,
            condition=value,
            exclude=exclude,
        )
        self.addCleanup(
            switch.remove_condition,
            condition_set=condition_set,
            field_name=field_name,
            condition=value,
        )


class SSOBaseTransactionTestCase(SSOBaseTestCaseMixin, TransactionTestCase):

    def setUp(self):
        super(SSOBaseTransactionTestCase, self).setUp()
        self.isolate_memcache()


class AuthenticatedTestCase(SSOBaseTestCase):

    login_email = "test@canonical.com"
    login_password = DEFAULT_USER_PASSWORD

    def setUp(self):
        super(AuthenticatedTestCase, self).setUp()
        self.account = self.factory.make_account(email=self.login_email,
                                                 password=self.login_password)
        assert self.client.login(
            username=self.login_email, password=self.login_password)


class MockHandler(urllib2.HTTPHandler):

    def __init__(self):
        self.requests = []

    def set_next_response(self, code, body):
        self.next_response = (code, body)

    def http_open(self, req):
        self.requests.append(req)
        code, body = self.next_response
        response = urllib.addinfourl(StringIO(body), {}, req.get_full_url())
        response.code = code
        response.msg = httplib.responses[code]
        return response


def test_concurrently(times):
    """
    Add this decorator to small pieces of code that you want to test
    concurrently to make sure they don't raise exceptions when run at the
    same time.  E.g., some Django views that do a SELECT and then a subsequent
    INSERT might fail when the INSERT assumes that the data has not changed
    since the SELECT.
    """

    def test_concurrently_decorator(test_func):

        def wrapper(*args, **kwargs):
            exceptions = []
            import threading

            def call_test_func():
                try:
                    test_func(*args, **kwargs)
                except Exception as e:
                    exceptions.append(e)
                    raise
                finally:
                    connection.close()

            threads = []
            for i in range(times):
                threads.append(threading.Thread(target=call_test_func))
            for t in threads:
                t.start()
            for t in threads:
                t.join()
            if exceptions:
                raise Exception('test_concurrently intercepted %s exceptions: '
                                '%s' % (len(exceptions), exceptions))
        return wrapper

    return test_concurrently_decorator


class TeamConditionSelectiveMixin(object):

    team = 'awesomener'
    key = ''

    def enable_flag(self):
        field_name = 'team'
        value = self.team
        condition_set = (
            'identityprovider.gargoyle_conditions.LPTeamConditionSet(lp_team)')
        self.conditionally_enable_flag(self.key, field_name, value,
                                       condition_set)

    def make_account(self):
        account = self.factory.make_account(email=self.email,
                                            password=self.password)
        # bind account with the team
        team = self.factory.make_team(name=self.team)
        self.factory.add_account_to_team(account, team)
        return account


class TimelineActionMixin(object):

    def assert_timeline_contains(self, request, category, detail):
        timeline = request.META.get('timeline.timeline')
        self.assertIsNotNone(timeline)
        password_actions = [action for action in timeline.actions
                            if action.category == category]
        self.assertEqual(len(password_actions), 1)
        self.assertEqual(password_actions[0].detail, detail)
        self.assertGreaterEqual(password_actions[0].duration,
                                timedelta(seconds=0))


class ConcurrentCommandMixin(object):

    def concurrent_command_locking_check(self, command_to_test, **kwargs):
        """Helper to test concurrent, db-locking command methods.

        Will call the command_to_test (given as a string, as the argument to
        call_command) with **kwargs, in two threads (identical calls).
        Exceptions raised during execution will be stored and checked
        once threads exit. Expects to find a specific exception message.
        Optional methods are passed and expected to be called by the commands
        under test at specific times to control thread interaction and make
        the tests non-timing-dependent.

        This helper is very specific to testing concurrent execution of
        management commands which try to acquire an update lock on the
        database.
        """
        # This is done somewhat indirectly since the point is for the lock
        # to be mostly transparent. To test, the _nowait=True parameter is
        # passed to the command, which will cause it to exception out if it's
        # unable to get the lock (production behavior is to block until it can
        # get a lock). So we can test for the locking exception.
        # The thread will call this function which just calls the command,
        # stores the expected exception in a parent-accessible container,
        # and closes the connection afterwards.
        def command_and_close(excp_container, **kwargs):
            # Get the event we need to fire if the thread crashes
            _notify_critical_section = kwargs['_notify_critical_section']
            try:
                call_command(command_to_test, **kwargs)
            except OperationalError as excp:
                excp_container.append(excp)
                _notify_critical_section()
            finally:
                # DB connections in threads are not auto-closed by django:
                # check out #database-connection-s-thread-locality in
                # https://docs.djangoproject.com/en/dev/releases/1.4/
                # The thread will call this function which just calls the
                # command, stores the expected exception in a parent-accessible
                # container,
                connection.close()
        # THREADING IS SCARY
        # To make tests independent of timings:
        # 1- Threads start running and will eventually call the method passed
        # as _notify_critical_section. This increases a counter and when the
        # counter is equal to the number of threads (meaning either both
        # threads entered the critical section or crashed, as the method is
        # also called when the thread raises an exception) the
        # _notify_critical_section event is set.  Before exiting the critical
        # section, threads should call the passed _wait_for_release method
        # which will block until the master sets the sync_event event.
        # 2- The master starts the threads, waits on the
        # _notify_critical_section event.  When it gets that, it means both
        # threads are either in the critical section, blocked on sync_event, or
        # have crashed and exited with an exception.
        # 3- Once the master gets the _notify_critical_section event, it can
        # ask any waiting threads to terminate (by setting the wait_forever
        # event), join them and perform the necessary test checks.
        # CAVEAT: Callers should ensure they only call _notify_critical_section
        # ONCE when entering the critical section, otherwise maybe one thread
        # could call it num_thread times while the others haven't yet, which
        # would incorrectly signal that all threads are in the critical
        # section.
        sync_event = threading.Event()
        notify_critical_section_event = threading.Event()
        counter = itertools.count(1)  # Supposed to be thread-safe!
        num_threads = 2

        def thread_entered_crit_sec():
            cval = counter.next()  # cval = cval++
            if (cval == num_threads):
                notify_critical_section_event.set()

        # Augment kwargs with the events
        kwargs.update({'_wait_for_release': sync_event.wait,
                       '_notify_critical_section': thread_entered_crit_sec})
        thrown_exceptions = []
        threads = [
            threading.Thread(target=command_and_close,
                             args=(thrown_exceptions,), kwargs=kwargs)
            for t in range(num_threads)
        ]
        [thread.start() for thread in threads]
        # Both threads are running. Now wait for the wakeup event:
        # The timeout is a safeguard with an ample-enough time.
        notify_critical_section_event.wait(10)
        sync_event.set()
        [thread.join() for thread in threads]
        # Finally, assert that one of the threads got OperationalError: could
        # not obtain lock on row in relation "account"
        self.assertEqual(len(thrown_exceptions), 1)
        self.assertIsInstance(thrown_exceptions[0], OperationalError)
        self.assertIn("could not obtain lock on row",
                      thrown_exceptions[0].message)
