# Copyright 2010 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

from __future__ import absolute_import
from __future__ import unicode_literals

import binascii
import hashlib
import logging
import string
from datetime import timedelta


from django import forms
from django.conf import settings
from django.contrib.sessions.models import Session
from django.core.exceptions import ObjectDoesNotExist
from django.core.urlresolvers import reverse
from django.db import connection, transaction
from django.db.utils import ProgrammingError
from django.utils.translation import ugettext_lazy as _

SALT_LENGTH = 20
ALPHA_NUM = string.ascii_letters + string.digits

# call populate_password_blacklist at startup to preload the blacklist
blacklisted_password_set = None

logger = logging.getLogger(__name__)


def canonical_url(obj, request=None, rootsite=None,
                  path_only_if_possible=False, view_name=None):
    """Very dumb implementation of canonical_url()."""
    from identityprovider.models import Person
    if isinstance(obj, Person):
        if view_name is None:
            return 'https://launchpad.net/~%s' % obj.name
        else:
            return 'https://launchpad.net/~%s/%s' % (obj.name, view_name)
    else:
        return None


# TODO: Remove after migrating all users' passwords to the Django cypher
def encrypt_launchpad_password(plaintext, salt=None):
    plaintext = str(plaintext)
    if salt is None:
        salt = generate_salt()
    result = binascii.b2a_base64(
        hashlib.sha1(plaintext + salt).digest() + salt.encode('utf-8'))
    return result[:-1]


def generate_openid_identifier():
    # On devel mode, when starting the DB, the initial migration is run,
    # and django will execute the account.openid_identifier's default method,
    # which is this one. The SQL for this store procedure will be ready
    # only AFTER the Account model was created, so there is a chicken-egg
    # situation that we workaround with this try-except.
    # During normal operation, the stored procedure will be used.
    try:
        with transaction.atomic():
            cursor = connection.cursor()
            cursor.execute("SELECT generate_openid_identifier();")
            row = cursor.fetchone()
            result = row[0]
    except ProgrammingError:
        result = generate_random_string(7)
    return result


def generate_random_string(length, prefix='', charset=ALPHA_NUM):
    rng = settings.RANDOM_GENERATOR
    result = prefix + ''.join(rng.choice(charset) for x in xrange(length))
    return unicode(result)


def generate_random_string_len_50():
    return generate_random_string(50)


def generate_random_string_len_60():
    return generate_random_string(60)


def generate_random_string_len_70():
    return generate_random_string(70)


def generate_salt():
    """ From lib/canonical/launchpad/webapp/authentication.py """
    # Salt can be any length, but not more than about 37 characters
    # because of limitations of the binascii module.
    return generate_random_string(SALT_LENGTH)


def polite_form_errors(errors):
    if 'email' in errors:
        if errors['email'][0] == _("Enter a valid e-mail address."):
            errors['email'][0] = _("Please enter a valid email address.")


def add_non_field_error(form, exc):
    """Adds a generic error message to a form from an exception"""
    # apparently _errors is part of the public api
    # (https://docs.djangoproject.com/en/dev/ref/forms/validation/
    #   #described-later)
    errors = form._errors.setdefault(
        forms.forms.NON_FIELD_ERRORS,
        form.error_class())
    if hasattr(exc, 'messages'):  # django.forms.ValidationError
        errors.extend(exc.messages)
    else:  # plain old Exception
        errors.append(unicode(exc))


# TODO: Remove after migrating all users' passwords to the Django cypher
def validate_launchpad_password(plaintext, encrypted):
    encrypted = str(encrypted)
    plaintext = str(plaintext)
    try:
        ref = binascii.a2b_base64(encrypted)
    except binascii.Error:
        # Not valid base64.
        return False
    salt = ref[20:]
    v = binascii.b2a_base64(
        hashlib.sha1(plaintext + salt).digest() + salt)[:-1]
    pw1 = (v or '').strip()
    pw2 = (encrypted or '').strip()
    return pw1 == pw2


def get_object_or_none(qs_or_model, **kwargs):
    if hasattr(qs_or_model, 'objects'):
        qs_or_model = qs_or_model.objects
    try:
        return qs_or_model.get(**kwargs)
    except ObjectDoesNotExist:
        return None


def redirection_url_for_token(token):
    if token is not None:
        # Redirected from an OpenID request, must get back after confirmation
        return reverse('server-decide', kwargs=dict(token=token))
    else:
        return '/'


def clean_sessions_until(creation_date, limit=None, dry_run=False):
    expire_date = creation_date + timedelta(
        seconds=settings.SESSION_COOKIE_AGE)
    with transaction.atomic():
        sessions = Session.objects.filter(
            expire_date__lte=expire_date)

        if limit is not None:
            sessions = Session.objects.filter(session_key__in=sessions[:limit])

        if dry_run:
            query = str(sessions.query)
            clean_lines = [
                'Would have deleted sessions from query: {}'.format(query),
                'Sessions to be deleted (session_key expire_date)',
            ]
            for session in sessions:
                clean_lines.append('{} - {}'.format(session.session_key,
                                                    session.expire_date))
            clean_lines.append(
                'Total sessions that would have been deleted: {}'.format(
                    len(sessions)))
            return '\n'.join(clean_lines)
        else:
            # Doing a delete on the entire sessions query will get a lock on
            # the table implicitly (due to transaction.atomic) so this command
            # should be safe to run concurrently.
            sessions.delete()


def get_cypher_migration_status():
    """Get the password cypher migration status.

    Returns a dict with two keys:
    - 'new_hashes': the amount of passwords already migrated;
    - 'old_hashes': the amount of passwords still to be migrated.
    """

    cursor = connection.cursor()

    new_hash_pattern = 'bcrypt13'
    hash_counts = {
        'new_hashes': 0,
        'old_hashes': 0,
    }

    cursor.execute(r"""
    select hash_type, count(hash_type) from (
        select
            case
                when password like '{0}%%' then 'new_hashes'
                else 'old_hashes'
            end
        from accountpassword
    ) as hash_type
    group by hash_type
    """.format(new_hash_pattern))
    for result in cursor.fetchall():
        if result[0] == '(new_hashes)':
            hash_counts['new_hashes'] = result[1]
        else:
            hash_counts['old_hashes'] = result[1]

    return hash_counts


def get_provider_url(request):
    uri = reverse('server-openid')
    return request.build_absolute_uri(uri)


def password_is_blacklisted(password):
    """Looks up a password in the blacklist set.

    It's implemented as a method to allow for some preprocessing of the
    incoming password (e.g. lowercasing).
    """
    normalized_password = password.lower()
    return normalized_password in get_password_blacklist()


def get_password_blacklist():
    """Get a set representing all the blacklisted passwords."""
    return blacklisted_password_set or set()


def populate_password_blacklist():
    """Load the password blacklist from a file into a module variable."""
    global blacklisted_password_set
    blacklisted_password_set = set()
    try:
        with open(settings.PASSWORD_BLACKLIST_FILE, "r") as blacklist_file:
            # The blacklist should be case-insensitive so entries are stored
            # always in lowercase
            blacklisted_password_set = set(
                entry.strip("\r\n").lower() for entry in blacklist_file)
        # This avoids doing entry.rstrip('\r\n') in the if condition
        # of the generator (would have to be done twice for every entry)
        if "" in blacklisted_password_set:
            blacklisted_password_set.remove("")
        success = True
        logger.debug("Loaded %s blacklisted passwords from %s",
                     len(blacklisted_password_set),
                     settings.PASSWORD_BLACKLIST_FILE)
    # If PASSWORD_BLACKLIST_FILE can't be read, just log and exit. The
    # blacklisted_password_set will still be a set, albeit an empty one.
    except (OSError, IOError):
        logger.exception("Couldn't load password blacklist")
        success = False
    return success
