from contextlib import contextmanager
import re

from django.core.exceptions import ValidationError
from django.utils.translation import ugettext_lazy as _

from identityprovider.timeline_helpers import get_request_timing_function
from identityprovider.utils import password_is_blacklisted

PASSWORD_POLICY_ERROR = _(
    'Password must be at least {min_length} characters long')
PASSWORD_POLICY_HELP_TEXT = _(
    'Password with at least {min_length} characters')
CANONICAL_PASSWORD_POLICY_ERROR = _(
    "Password must consist of lower- and upper-case characters, and at "
    "least one digit."
)
PASSWORD_BLACKLISTED_ERROR = _(
    "Password cannot be used because it is very common and susceptible to "
    "attack. Please choose a different one.")
PASSWORD_LEAKED = _('Password is unsafe because it has been exposed '
                    'by a security breach on another website.')


class BasicValidator(object):
    """Generic password validator.

    Accounts are valid if:
        - password contains at least 8 characters
        - password contains only ascii characters
    """

    MIN_LENGTH = 8
    PASSWORD_POLICY_HELP_TEXT = PASSWORD_POLICY_HELP_TEXT.format(
        min_length=MIN_LENGTH)

    def __init__(self, account=None, email=None, request=None):
        super(BasicValidator, self).__init__()
        self.account = account
        self.email = email
        self.timer = get_request_timing_function(request) if request else None

    def match(self):
        """Return True for all accounts."""
        # Note that for new users account will be None
        # The basic validator always matches
        return True

    def __call__(self, password):
        """Check the password complies with the policy.

        Validation is successful if:
            - password contains at least 8 characters
            - password contains only ascii characters
            - password doesn't match leaked credentials for account
            - password doesn't match leaked credentials for email

        Raises ValidationError on failure.
        """
        # Implement the basic password policy
        try:
            # ascii only!
            str(password)
        except UnicodeEncodeError:
            raise ValidationError(_("Invalid characters in password"))
        if password is None or len(password) < self.MIN_LENGTH:
            raise ValidationError(
                PASSWORD_POLICY_ERROR.format(min_length=self.MIN_LENGTH))
        if self.password_leaked(password):
            raise ValidationError(PASSWORD_LEAKED)

    def password_leaked(self, password):
        # import here to avoid a circular import issue
        from identityprovider.auth import LaunchpadBackend
        from identityprovider.models import AccountPassword

        backend = LaunchpadBackend()
        try:
            ap = self.account.accountpassword
            emails = [em.email for em in self.account.emails()]
        except (AttributeError, AccountPassword.DoesNotExist):
            ap = None
            emails = [self.email] if self.email else []
        for email in emails:
            if backend.password_leaked(email, password, ap, timer=self.timer):
                return True
        return False


class CanonicalValidator(BasicValidator):
    """Password validator for Canonical accounts.

    Accounts are valid if:
        - account is member of the 'canonical' team
        - password contains at least 10 characters
        - password contains only ascii characters
        - password contains at least one lower case letter
        - password contains at least one upper case letter
        - password contains at least one number
    """

    MIN_LENGTH = 10
    PASSWORD_POLICY_HELP_TEXT = PASSWORD_POLICY_HELP_TEXT.format(
        min_length=MIN_LENGTH)

    def match(self):
        """Return True if the account is part of the 'canonical' team."""
        return self.account is not None and self.account.person_in_team(
            'canonical')

    def __call__(self, password):
        """Check the password complies with the policy.

        Validation is successful if:
            - password contains at least one lower case letter
            - password contains at least one upper case letter
            - password contains at least one number
            - password is not blacklisted

        Raises ValidationError on failure.
        """
        super(CanonicalValidator, self).__call__(password)

        has_lower = False
        has_upper = False
        has_number = False
        if re.search(r'[a-z]', password):
            has_lower = True
        if re.search(r'[A-Z]', password):
            has_upper = True
        if re.search(r'[0-9]', password):
            has_number = True

        if not all([has_lower, has_upper, has_number]):
            raise ValidationError(CANONICAL_PASSWORD_POLICY_ERROR)

        if password_is_blacklisted(password):
            raise ValidationError(_(PASSWORD_BLACKLISTED_ERROR))


def get_password_validator(account=None, email=None, request=None):
    """Return an appropriate password validation function for the account."""

    for validator_class in [CanonicalValidator, BasicValidator]:
        validator = validator_class(account, email, request=request)
        if validator.match():
            # return to exit as we have found a match
            return validator

    # this should never happen as the BasicValidator should always match!
    raise ValidationError(PASSWORD_POLICY_ERROR)


class Errors(dict):
    GENERAL = "__all__"

    def _new_error_dict(self, e):
        result = {}
        for k, v in getattr(e, 'error_dict', {}).items():
            result[k] = []
            for i in v:
                result[k].extend(i.messages)
        return result

    @contextmanager
    def collect(self, key=None):
        try:
            yield
        except ValidationError as e:
            if key is not None:
                self[key] = e.messages
            else:
                e.update_error_dict(self)
                # since django 1.7, when creating a ValidationError with a dict
                # as parameter, the internal dict will have lists of
                # ValidationErros instances
                self.update(self._new_error_dict(e))
