# Copyright 2010, 2012 Canonical Ltd.  This software is licensed under
# the GNU Affero General Public License version 3 (see the file
# LICENSE).

from datetime import timedelta
import logging
import re
import urllib
import urlparse

from django.conf import settings
from django.contrib import auth, messages
from django.core.urlresolvers import reverse
from django.http import (
    Http404,
    HttpResponse,
    HttpResponseBadRequest,
    HttpResponseForbidden,
    HttpResponseRedirect,
)
from django.template import loader
from django.shortcuts import get_object_or_404, render
from django.utils.decorators import decorator_from_middleware
from django.utils import translation
from django.utils.translation import ugettext_lazy as _
from django.utils.timezone import now
from django.views.decorators.csrf import csrf_exempt
from gargoyle.decorators import switch_is_active
from openid import oidutil
from openid.extensions import ax, pape
from openid.extensions.sreg import SRegRequest, SRegResponse
from openid.message import IDENTIFIER_SELECT, registerNamespaceAlias
from openid.server.server import (
    BROWSER_REQUEST_MODES,
    ENCODE_URL,
    CheckIDRequest,
    ProtocolError,
    Server,
)
from openid.server.trustroot import TrustRoot
from openid.urinorm import urinorm
from openid.yadis.constants import YADIS_HEADER_NAME

from identityprovider import signed
from identityprovider.auth import build_discharge_macaroon
from identityprovider.const import (
    AX_DATA_FIELDS,
    MACAROON_NS,
    LAUNCHPAD_TEAMS_NS,
)
from identityprovider.forms import (
    MacaroonRequestForm,
    PreAuthorizeForm,
    TeamsRequestForm,
    UserAttribsRequestForm,
)
from identityprovider.macaroon import (
    MacaroonRequest,
    MacaroonResponse,
)
from identityprovider.middleware.xrds import XRDSMiddleware
from identityprovider.models import (
    Account,
    DjangoOpenIDStore,
    OpenIDAuthorization,
    OpenIDRPConfig,
    OpenIDRPSummary,
    twofactor,
    get_team_memberships_for_user,
)
from identityprovider.models.const import AuthLogType
from identityprovider.models.twofactor import is_twofactor_enabled
from identityprovider.signals import login_succeeded
from identityprovider.teams import TeamsRequest
from identityprovider.utils import generate_random_string, get_provider_url
from identityprovider.views import utils

SITE_REQUIRES_VERIFIED = _(
    'The site {rp_name} requires that you verify your email address before '
    'accessing its contents.')

accept_xrds = decorator_from_middleware(XRDSMiddleware)
registerNamespaceAlias(LAUNCHPAD_TEAMS_NS, 'lp')
logger = logging.getLogger(__name__)

# So we can map OpenID HTTP responses to real Django response classes
RESPONSE_CLASSES = {
    200: HttpResponse,
    302: HttpResponseRedirect,
    400: HttpResponseBadRequest,
    403: HttpResponseForbidden,
    404: Http404,
}


def make_response(code, content, content_type):
    ResponseClass = RESPONSE_CLASSES.get(code)
    if ResponseClass is not None:
        response = ResponseClass(content, content_type=content_type)
    else:
        response = HttpResponse(content, content_type=content_type,
                                status=code)
    return response


# moved from views.i18n to be able to split django apps and
# avoid circular deps
def set_language_info(request, response, lang):
    if lang not in settings.SUPPORTED_LANGUAGES:
        lang = 'en'
    if hasattr(request, 'session'):
        request.session['django_language'] = lang
    else:
        response.set_cookie(settings.LANGUAGE_COOKIE_NAME, lang)


@csrf_exempt
@accept_xrds
def openid_provider(request, lang=None):
    if lang not in settings.SUPPORTED_LANGUAGES:
        # Next we check for a primary language.
        if lang is not None:
            lang = lang.split('_')[0]  # e.g. de_CH becomes de.
        if lang not in settings.SUPPORTED_LANGUAGES:
            lang = translation.get_language_from_request(request)
    translation.activate(lang)
    provider_url = get_provider_url(request)
    openid_server = _get_openid_server(provider_url)
    if request.method == 'GET':
        querydict = dict(request.GET.items())
    else:
        querydict = dict(request.POST.items())
    logger.debug("querydict = %s", str(querydict))
    try:
        orequest = openid_server.decodeRequest(querydict)
        response = _process_openid_request(request, orequest, openid_server)
    except ProtocolError as e:
        response = _handle_openid_error(e)
    set_language_info(request, response, lang)
    return response


def _process_openid_request(request, orequest, openid_server):
    if not orequest:
        return render(request, 'server/server_info.html')

    if orequest.mode in ("checkid_immediate", "checkid_setup"):
        if (utils.get_rpconfig(orequest.trust_root) is None and
                getattr(settings, 'SSO_RESTRICT_RP', True)):
            # This is an untrusted RP.  We don't play with these for now.
            logger.debug("Untrusted RP: %s", orequest.trust_root)
            token = generate_random_string(16)
            request.session[token] = signed.dumps(orequest,
                                                  settings.SECRET_KEY)
            response = HttpResponseRedirect('/%s/+untrusted' % token)
        else:
            response = _handle_user_response(request, orequest)
    else:
        oresponse = openid_server.handleRequest(orequest)
        response = _django_response(request, oresponse)
    return response


def _handle_openid_error(error):
    if error.whichEncoding() == ENCODE_URL:
        url = error.encodeToURL()
        return HttpResponseRedirect(url)
    else:
        response = HttpResponse(error.encodeToKVForm())
        response['Content-Type'] = 'text/plain; charset=utf-8'
        return response


def _handle_user_response(request, orequest):
    response = None
    rp_config = utils.get_rpconfig(orequest.trust_root)
    user_verified_for_rp = _user_is_verified_for_rp(request.user, rp_config)
    if orequest.immediate:
        auto_authorized = _is_auto_authorized_rp(rp_config)
        if (auto_authorized and user_verified_for_rp and
                twofactor.is_authenticated(request) and
                _is_identity_owner(request, orequest)):
            if orequest.idSelect():
                openid_identity_url = request.build_absolute_uri(
                    request.user.get_absolute_url())
                oresponse = orequest.answer(
                    True, identity=openid_identity_url)
            else:
                oresponse = orequest.answer(True)
            _add_user_attribs(request, orequest, oresponse)
            _check_team_membership(request, orequest, oresponse,
                                   immediate=True)
            _add_macaroon(request, orequest, oresponse)
            response = _django_response(request, oresponse, True)
        else:
            oresponse = orequest.answer(False)
            response = _django_response(request, oresponse)
    elif not _is_valid_openid_for_this_site(request, orequest.identity):
        context = {
            'trust_root': orequest.trust_root,
            'identifier': orequest.identity,
            'continue_url': orequest.answer(False).encodeToURL(),
        }
        response = render(request, 'server/invalid_identifier.html', context)
    elif (user_verified_for_rp and _openid_is_authorized(request, orequest)):
        if orequest.idSelect():
            openid_identity_url = request.build_absolute_uri(
                request.user.get_absolute_url())
            oresponse = orequest.answer(
                True, identity=openid_identity_url)
        else:
            oresponse = orequest.answer(True)
        _add_user_attribs(request, orequest, oresponse)
        _check_team_membership(request, orequest, oresponse, immediate=True)
        _add_macaroon(request, orequest, oresponse)
        response = _django_response(request, oresponse, True)
    elif (twofactor.is_authenticated(request) and not
          _is_identity_owner(request, orequest)):
        oresponse = orequest.answer(False)
        response = _django_response(request, oresponse, True)
    else:
        token = generate_random_string(16)
        request.session[token] = signed.dumps(orequest,
                                              settings.SECRET_KEY)
        response = HttpResponseRedirect(
            reverse('server-decide', kwargs=dict(token=token)))
    referer = request.META.get('HTTP_REFERER')
    if referer:
        response.set_cookie('openid_referer', referer)
    return response


def _is_valid_openid_for_this_site(request, identity):
    try:
        identity = urinorm(identity)
        idparts = urlparse.urlparse(identity)
        root_url = request.build_absolute_uri('/')
        srvparts = urlparse.urlparse(root_url)
        if identity == IDENTIFIER_SELECT:
            return True

        if (idparts.port != srvparts.port or
                idparts.scheme != srvparts.scheme):
            return False

        if (idparts.hostname != srvparts.hostname and
                not srvparts.hostname.endswith(".%s" % idparts.hostname)):
            return False

        accept_path_patterns = [
            '^/$',
            '^/\+id/[a-zA-Z0-9\-_\.]+$',
            '^/~[a-zA-Z0-9\-_\.]+$',
        ]
        for pattern in accept_path_patterns:
            if re.match(pattern, idparts.path) is not None:
                return True
        return False
    except:
        return False


def _get_orequest(request, token):
    """Returns the OpenID request for the given request from the
    user's browser.  May throw an exception if there is no OpenID
    request, or if it is invalid."""
    raw_orequest = request.session.get(token, None)
    return signed.loads(raw_orequest, settings.SECRET_KEY)


def decide(request, token):
    try:
        orequest = _get_orequest(request, token)
        rpconfig = utils.get_rpconfig(orequest.trust_root)
    except:
        return HttpResponse("Invalid OpenID transaction")

    if not request.user.is_authenticated():
        # XXX: need to remove this circular dep to the webui app
        from webui.views import ui
        return ui.LoginView.as_view()(request, token, rpconfig=rpconfig)

    if not _user_is_verified_for_rp(request.user, rpconfig):
        name = rpconfig.displayname if rpconfig else orequest.trust_root
        messages.warning(
            request,
            SITE_REQUIRES_VERIFIED.format(rp_name=name)
        )
        return HttpResponseRedirect(reverse('account-emails'))

    site_requires_twofactor = twofactor.site_requires_twofactor_auth(
        request, token, rpconfig)
    if (not twofactor.is_authenticated(request) or
            (site_requires_twofactor and not twofactor.is_upgraded(request))):
        if is_twofactor_enabled(request):
            return HttpResponseRedirect(reverse('twofactor', args=[token]))
        else:
            return _process_decide(request, orequest, decision=False)

    if ('ok' in request.POST or
            (rpconfig is not None and rpconfig.auto_authorize)):
        return _process_decide(request, orequest, decision=True)

    sreg_request = SRegRequest.fromOpenIDRequest(orequest)
    ax_request = ax.FetchRequest.fromOpenIDRequest(orequest)
    teams_request = TeamsRequest.fromOpenIDRequest(orequest)
    macaroon_request = MacaroonRequest.fromOpenIDRequest(orequest)
    try:
        openid_identity_url = request.build_absolute_uri(
            request.user.get_absolute_url())
        summary = OpenIDRPSummary.objects.get(
            account=request.user, trust_root=orequest.trust_root,
            openid_identifier=openid_identity_url)
        approved_data = summary.get_approved_data()
    except OpenIDRPSummary.DoesNotExist:
        approved_data = {}

    user_attribs_form = UserAttribsRequestForm(
        request, sreg_request, ax_request, rpconfig,
        approved_data=approved_data.get('user_attribs'))
    teams_form = TeamsRequestForm(request, teams_request, rpconfig,
                                  approved_data=approved_data.get('teams'))
    macaroon_form = MacaroonRequestForm(
        request, macaroon_request, rpconfig, orequest.trust_root,
        approved_data=approved_data.get('macaroon'))
    context = {
        'account': request.user,
        'trust_root': orequest.trust_root,
        'rpconfig': rpconfig,
        'user_attribs_form': user_attribs_form,
        'teams_form': teams_form,
        'macaroon_form': macaroon_form,
        'token': token,
        'sane_trust_root': _request_has_sane_trust_root(orequest)
    }
    # Log existing authentication request (i.e. checked via pre-verified
    # cookie).
    login_succeeded.send(
        sender=request.user, user=request.user, request=request,
        authlogtype=AuthLogType.OPENID_EXISTING)
    return render(request, 'server/decide.html', context)


def _request_has_sane_trust_root(openid_request):
    """Return True if the RP's trust root looks sane."""
    assert openid_request is not None, (
        'Could not find the OpenID request')
    trust_root = TrustRoot.parse(openid_request.trust_root)
    return trust_root.isSane()


def pre_authorize(request):
    logger.debug("HTTP_REFERER for this request: %s\n >>> \n",
                 request.META.get('HTTP_REFERER', 'None'))
    if request.method == 'GET':
        form = PreAuthorizeForm(request.GET)
    else:
        form = PreAuthorizeForm(request.POST)
    if form.is_valid():
        try:
            trust_root, callback, referer = _get_valid_pre_auth_data(request,
                                                                     form)
        except:
            # Unauthorized trust root or referrer.
            _clear_pre_auth_session_data(request)
            return HttpResponseBadRequest()

        if twofactor.is_authenticated(request):
            logger.debug("Approved for %s, %s\n >>> \n", referer, request.user)
            client_id = request.session.session_key
            hours = getattr(settings, 'PRE_AUTHORIZATION_VALIDITY', 2)
            expires = now() + timedelta(hours=hours)
            OpenIDAuthorization.objects.authorize(
                request.user, trust_root, expires, client_id)
            return HttpResponseRedirect(callback)
        else:
            request.session['pre_auth_referer'] = referer
            request.session['pre_auth_referer_for'] = trust_root
            next = "%s?%s" % (request.META.get('PATH_INFO'),
                              request.META.get('QUERY_STRING', ''))
            return HttpResponseRedirect('/+login?next=%s' %
                                        urllib.quote(next))
    else:
        _clear_pre_auth_session_data(request)
        return HttpResponseBadRequest()


def _get_valid_pre_auth_data(request, form):
    trust_root = form.cleaned_data['trust_root']
    callback = form.cleaned_data['callback']
    http_referer = _get_pre_auth_referer(request, trust_root)

    if http_referer is None:
        raise Exception("Pre-auth not approved")

    for line in getattr(settings, 'OPENID_PREAUTHORIZATION_ACL', []):
        referer, acl_trust_root = line
        if http_referer.startswith(referer) and trust_root == acl_trust_root:
            return (trust_root, callback, http_referer)
    raise Exception("Pre-auth not approved")


def _get_pre_auth_referer(request, trust_root):
    sess_referer = request.session.get('pre_auth_referer', None)
    sess_referer_for = request.session.get('pre_auth_referer_for', None)
    if sess_referer is not None and sess_referer_for is not None:
        _clear_pre_auth_session_data(request)
        if sess_referer_for == trust_root:
            logger.debug("getting referer %s from session", sess_referer)
            logger.debug("http referer was %s",
                         request.META.get('HTTP_REFERER', ''))
            return sess_referer
        else:
            return None
    else:
        return request.META.get('HTTP_REFERER', None)


def _clear_pre_auth_session_data(request):
    try:
        del request.session['pre_auth_referer']
        del request.session['pre_auth_referer_for']
    except:
        pass


def cancel(request, token):
    try:
        raw_orequest = request.session.get(token, None)
        orequest = signed.loads(raw_orequest, settings.SECRET_KEY)
    except:
        return HttpResponse("Invalid OpenID transaction")
    if twofactor.is_authenticated(request):
        return _process_decide(request, orequest, decision=False)
    else:
        provider_url = get_provider_url(request)
        oresponse = orequest.answer(False, provider_url)
        response = _django_response(request, oresponse)
        return response


def xrds(request):
    logger.debug("xrds()")
    context = {
        'endpoint_url': get_provider_url(request),
    }
    resp = render(request, 'server/openidapplication-xrds.xml', context)
    resp['Content-type'] = 'application/xrds+xml'
    return resp


@accept_xrds
def identity_page(request, identifier):
    account = get_object_or_404(Account, openid_identifier=identifier)
    if not account.is_active:
        raise Http404()
    openid_identity_url = request.build_absolute_uri(
        account.get_absolute_url())
    context = {
        'provider_url': get_provider_url(request),
        'identity_url': openid_identity_url,
        'display_name': account.displayname,
    }
    resp = render(request, 'server/person.html', context)
    resp[YADIS_HEADER_NAME] = request.build_absolute_uri(
        reverse('server-xrds', args=[account.openid_identifier]))
    return resp


def xrds_identity_page(request, identifier):
    account = get_object_or_404(Account, openid_identifier=identifier)
    if not account.is_active:
        raise Http404()
    openid_identity_url = request.build_absolute_uri(
        account.get_absolute_url())
    context = {
        'provider_url': get_provider_url(request),
        'identity_url': openid_identity_url,
    }
    resp = render(request, 'server/person-xrds.xml', context)
    resp['Content-type'] = 'application/xrds+xml'
    return resp


def _openid_is_authorized(request, openid_request):
    logger.debug("openid_is_authorized(%s, %s)",
                 openid_request.identity, openid_request.trust_root)

    rpconfig = utils.get_rpconfig(openid_request.trust_root)

    if (not twofactor.is_authenticated(request) or
            not _is_identity_owner(request, openid_request)):
        logger.debug("openid_is_authorized() -> False (id_owner)")
        return False
    elif (twofactor.site_requires_twofactor_auth(request, None, rpconfig) and
            not twofactor.is_upgraded(request)):
        logger.debug("openid_is_authorized() -> False (rpconfig)")
        return False

    elif _should_reauthenticate(openid_request, request.user):
        logger.debug("openid_is_authorized() -> True (should_reauthenticate)")
        auth.logout(request)
        return False

    elif _is_auto_authorized_rp(rpconfig):
        logger.debug("openid_is_authorized() -> True (rpconfig)")
        return True

    else:
        ret = OpenIDAuthorization.objects.is_authorized(
            request.user, openid_request.trust_root,
            request.session.session_key)
        logger.debug("openid_is_authorized() -> %s", ret)
        return ret


def _is_auto_authorized_rp(rp):
    return rp is not None and rp.auto_authorize


def _user_is_verified_for_rp(user, rp):
    if user.is_authenticated():
        return user.is_verified or (rp is not None and rp.allow_unverified)
    return False


def _should_reauthenticate(openid_request, user):
    """Should the user re-enter their password?

    Return True if the user entered their password more than
    max_auth_age seconds ago. Return False otherwise.

    The max_auth_age parameter is defined in the OpenID Provider
    Authentication Policy Extension.
    http://openid.net/
        specs/openid-provider-authentication-policy-extension-1_0-07.html

    This parameter contains the maximum number of seconds before which
    an authenticated user must enter their password again. By default,
    there is no such maximum and if the user is logged in Launchpad, they
    can simply click-through to Sign In the relying party.

    But if the relaying party provides a value for that parameter, the
    user most have logged in not more than that number of seconds ago,
    Otherwise, they'll have to enter their password again.
    """
    pape_request = pape.Request.fromOpenIDRequest(openid_request)

    # If there is no parameter, the login is valid.
    if pape_request is None or pape_request.max_auth_age is None:
        logger.debug("No pape request")
        return False

    try:
        max_auth_age = int(pape_request.max_auth_age)
    except ValueError:
        logger.debug("pape:max_auth_age parameter should be an integer: %s",
                     pape_request.max_auth_age)
        return False

    cutoff = now() - timedelta(seconds=max_auth_age)
    logger.debug("%s", user.last_login)
    return user.last_login <= cutoff


def _django_response(request, oresponse, auth_success=False, orequest=None):
    """ Convert an OpenID response into a Django HttpResponse """
    provider_url = get_provider_url(request)
    webresponse = _get_openid_server(provider_url).encodeResponse(oresponse)

    # This is a workaround for the fact the the openid library returns bare
    # HTML form markup instead of a complete HTML document. See
    # https://github.com/openid/python-openid/pull/31/files which has been
    # merged, but not released.
    if webresponse.body and oresponse.request.mode in BROWSER_REQUEST_MODES:
        content_type = "text/html; charset=utf-8"
        body = oidutil.autoSubmitHTML(webresponse.body)
    else:
        content_type = "text/plain; charset=utf-8"
        body = webresponse.body

    response = make_response(
        webresponse.code,
        body,
        content_type
    )

    response.status_code = webresponse.code
    for key, value in webresponse.headers.items():
        response[key] = value
        logger.debug("response[%s] = %s", key, value)
    logger.debug("response_body = %s", webresponse.body)
    if auth_success and isinstance(oresponse.request, CheckIDRequest):
        logger.debug("oresponse.fields = %s", str(oresponse.fields))
        approved_data = _get_approved_data(request, orequest)
        openid_identity_url = request.build_absolute_uri(
            request.user.get_absolute_url())
        OpenIDRPSummary.objects.record(
            request.user,
            oresponse.request.trust_root,
            openid_identity_url,
            approved_data)
    return response


def _get_approved_data(request, orequest):
    """Given an HTTP request and an OpenID request, return a nested dict of
    values requested in the request and approved by the user.
    """
    if not orequest:
        return None

    approved_data = {}
    rpconfig = utils.get_rpconfig(orequest.trust_root)

    sreg_request = SRegRequest.fromOpenIDRequest(orequest)
    ax_request = ax.FetchRequest.fromOpenIDRequest(orequest)

    user_attribs_form = UserAttribsRequestForm(
        request, sreg_request, ax_request, rpconfig)
    if user_attribs_form.has_data:
        approved_data['user_attribs'] = {
            'requested': user_attribs_form.data.keys(),
            'approved': user_attribs_form.data_approved_for_request.keys()}

    args = orequest.message.getArgs(LAUNCHPAD_TEAMS_NS)
    team_names = args.get('query_membership')
    if team_names:
        team_names = team_names.split(',')
        teams_form = TeamsRequestForm(
            request,
            TeamsRequest.fromOpenIDRequest(orequest),
            rpconfig,
        )
        approved_data['teams'] = {
            'requested': team_names,
            'approved': teams_form.teams_approved_by_user}

    macaroon_args = orequest.message.getArgs(MACAROON_NS)
    if macaroon_args.get('caveat_id'):
        macaroon_form = MacaroonRequestForm(
            request, MacaroonRequest.fromOpenIDRequest(orequest), rpconfig,
            orequest.trust_root)
        approved_data['macaroon'] = {
            'requested': macaroon_form.data.keys(),
            'approved': macaroon_form.data_approved_for_request.keys()}

    return approved_data


def _is_identity_owner(request, openid_request):
    assert request.user is not None, (
        "user should be logged in by now.")
    openid_identity_url = request.build_absolute_uri(
        request.user.get_absolute_url())
    ret = (openid_request.idSelect() or
           openid_request.identity == openid_identity_url)
    logger.debug("_is_identity_owner() -> %s", ret)
    return ret


def _add_user_attribs(request, openid_request, openid_response):
    # Add ax and sreg result data
    sreg_request = SRegRequest.fromOpenIDRequest(openid_request)
    ax_request = ax.FetchRequest.fromOpenIDRequest(openid_request)
    rpconfig = utils.get_rpconfig(openid_request.trust_root)
    form = UserAttribsRequestForm(
        request, sreg_request, ax_request, rpconfig)
    if form.data_approved_for_request:
        sreg_response = SRegResponse.extractResponse(
            sreg_request, form.data_approved_for_request)
        openid_response.addExtension(sreg_response)
        if ax_request is not None:
            ax_response = ax.FetchResponse(ax_request)
            for k, v in form.data_approved_for_request.iteritems():
                if AX_DATA_FIELDS.getNamespaceURI(k) in ax_request:
                    ax_response.addValue(AX_DATA_FIELDS.getNamespaceURI(k), v)
            openid_response.addExtension(ax_response)


def _process_decide(request, orequest, decision):
    openid_identity_url = request.build_absolute_uri(
        request.user.get_absolute_url())
    oresponse = orequest.answer(decision, identity=openid_identity_url)
    if decision:
        # If they use PAPE, let them know of the last logged in time.
        pape_request = pape.Request.fromOpenIDRequest(orequest)
        if pape_request:
            last_login = request.user.last_login
            pape_response = pape.Response(
                auth_time=last_login.strftime('%Y-%m-%dT%H:%M:%SZ'))
            oresponse.addExtension(pape_response)

        OpenIDAuthorization.objects.authorize(
            request.user,
            orequest.trust_root,
            client_id=request.session.session_key)
        _add_user_attribs(request, orequest, oresponse)
        # if there's no submitted POST data, this is an auto-authorized
        # (immediate) request
        immediate = not request.POST
        _check_team_membership(request, orequest, oresponse,
                               immediate=immediate)
        _add_macaroon(request, orequest, oresponse)
    r = _django_response(request, oresponse, decision, orequest)
    if r.content:
        # Only user-visible content is generated from this view.  Wrap
        # it up and set the Content-Type.
        r.content = loader.render_to_string("server/post-assertion.html", {
            'form': r.content})
        r['Content-Type'] = 'text/html; charset=utf-8'
        # Also, as this is POSTing to another server, disabled CSRF
        # protection.
        r.csrf_exempt = True
    return r


def _get_openid_server(provider_url):
    logger.debug("_get_server()")
    store = DjangoOpenIDStore()
    openid_server = Server(store, op_endpoint=provider_url)
    return openid_server


def _check_team_membership(request, orequest, oresponse, immediate=True):
    """Perform team membership checks.

    If any team membership checks have been requested as part of
    the OpenID request, annotate the response with the list of
    teams the user is actually a member of.
    """
    assert request.user is not None, (
        'Must be logged in to calculate team membership')
    if request.user.person is None:
        return
    args = orequest.message.getArgs(LAUNCHPAD_TEAMS_NS)
    team_names = args.get('query_membership')
    if not team_names:
        return
    team_names = team_names.split(',')
    if immediate:
        try:
            summary = OpenIDRPSummary.objects.get(
                account=request.user, trust_root=orequest.trust_root,
                openid_identifier=request.user.openid_identifier)
            approved_data = summary.get_approved_data()
        except OpenIDRPSummary.DoesNotExist:
            approved_data = {}
        if 'teams' in approved_data:
            teams = ','.join(approved_data['teams'].get('approved', []))
        else:
            rpconfig = utils.get_rpconfig(orequest.trust_root)
            teams = ','.join(get_team_memberships_for_user(
                team_names, request.user,
                rpconfig and rpconfig.can_query_any_team))
    else:
        form = TeamsRequestForm(
            request,
            TeamsRequest.fromOpenIDRequest(orequest),
            utils.get_rpconfig(orequest.trust_root),
        )
        teams = ','.join(form.teams_approved_by_user)
    oresponse.fields.namespaces.addAlias(LAUNCHPAD_TEAMS_NS, 'lp')
    oresponse.fields.setArg(LAUNCHPAD_TEAMS_NS, 'is_member', teams)


def _add_macaroon(request, openid_request, openid_response):
    """Add discharge macaroon if requested and approved."""
    macaroon_request = MacaroonRequest.fromOpenIDRequest(openid_request)
    rpconfig = utils.get_rpconfig(openid_request.trust_root)
    form = MacaroonRequestForm(
        request, macaroon_request, rpconfig, openid_request.trust_root)
    if form.data_approved_for_request:
        discharge_macaroon = build_discharge_macaroon(
            request.user, macaroon_request.caveat_id)
        macaroon_response = MacaroonResponse.extractResponse(
            macaroon_request, discharge_macaroon.serialize())
        openid_response.addExtension(macaroon_response)


def untrusted(request, token):
    raw_orequest = request.session.get(token, None)
    orequest = signed.loads(raw_orequest, settings.SECRET_KEY)
    context = {'trust_root': orequest.trust_root}
    return render(request, 'server/untrusted.html', context)


@switch_is_active('LOGIN_BY_TOKEN')
def login_by_token(request):
    user = auth.authenticate(request=request)
    if user is None:
        return HttpResponseForbidden()
    auth.login(request, user)

    next_step = request.GET.get('next')
    rpconfig = None
    if next_step:
        rpconfig = OpenIDRPConfig.objects.for_url(next_step)
    if next_step:
        if rpconfig or utils.is_safe_redirect_url(next_step):
            return HttpResponseRedirect(next_step)

        msg = _("Unknown redirect URL '{url}'")
        messages.warning(request, msg.format(url=next_step))

    return HttpResponseRedirect('/')
