# Copyright 2010 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

from random import randint

from django.http import HttpResponse
from django.shortcuts import get_object_or_404

from identityprovider.models import (
    Account,
    LPOpenIdIdentifier,
    Person,
)
from identityprovider.oops_config import ManuallyGeneratedOops
from identityprovider.views.utils import require_testing_enabled
from identityprovider.utils import get_provider_url
from webservices.launchpad import (
    NameAlreadyTaken,
    NotPlaceholderAccount,
)


@require_testing_enabled
def openid_consumer(request):
    msg = ['Consumer received %s' % request.method]
    args = getattr(request, request.method)
    for arg in sorted(args):
        msg.append("%s:%s" % (arg, args[arg]))
        if arg == 'openid.user_setup_url':
            msg.append('\n')
    return HttpResponse('\n'.join(msg))


@require_testing_enabled
def delegate_profile(request, username, version=0):
    person = get_object_or_404(Person, name=username)
    try:
        version = int(version)
    except:
        version = 0
    if person.account is not None and person.account.is_active:
        headers = []
        openid_identity_url = request.build_absolute_uri(
            person.account.get_absolute_url())
        header_data = {
            'server': get_provider_url(request),
            'openid': openid_identity_url,
        }
        if version in [0, 1]:
            headers += """
            <link rel="openid.server"
                href="%(server)s" />
            <link rel="openid.delegate"
                href="%(openid)s" />"""
        if version in [0, 2]:
            headers += """
            <link rel="openid2.provider"
                href="%(server)s" />
            <link rel="openid2.local_id"
                href="%(openid)s" />
            <meta http-equiv="X-XRDS-Location"
                content="%(openid)s/+xrds" />"""
        headers = "".join(headers) % header_data
    else:
        headers = ""
    return HttpResponse("""
<html>
    <head>
        <title>Test delegation profile for: ~%(name)s</title>
        %(headers)s
    </head>
    <body>
        <h1>Test profile for ~%(name)s</h1>
    </body>
</html>""" % {'name': person.name, 'headers': headers})


# Doesn't require settings.TESTING, as while it is for testing, we want
# to test on prod at times to ensure that oopses are working as
# expected. It also doesn't disclose any information or impact the
# performance of the system.
def error(request):
    """ Raise an internal server error """
    raise ManuallyGeneratedOops("This error is a test.  Please ignore.")


class MockLaunchpad(object):
    """Mock Launchpad, currently just for set_lp_username."""

    def __init__(self):
        self.placeholders = {}
        self.non_placeholders = {}

    def add_user(self, openid_identifier, username):
        self.non_placeholders[openid_identifier] = username

    def set_lp_username(self, openid_identifier, username,
                        dry_run=False):
        if (username in self.placeholders.values() or
                username in self.non_placeholders.values()):
            raise NameAlreadyTaken(400, "Username taken")
        if openid_identifier in self.non_placeholders:
            raise NotPlaceholderAccount(400, "Non-placeholder account")
        if not dry_run:
            self.placeholders[openid_identifier] = username

    def flush_to_db(self):
        items = self.placeholders.items() + self.non_placeholders.items()
        for openid_identifier, username in items:
            account = Account.objects.active_by_openid(openid_identifier)
            if account is None or account.person is None:
                lp_account = randint(100, 1000)
                LPOpenIdIdentifier.objects.create(
                    identifier=openid_identifier, lp_account=lp_account)
                person = Person.objects.create(
                    displayname=username, name=username,
                    lp_account=lp_account)
                if account is not None:
                    del account.person
            else:
                person = account.person
            person.name = username
            person.save()

    def get_lp_username(self, openid_identifier):
        return (
            self.placeholders.get(openid_identifier)
            or self.non_placeholders.get(openid_identifier))
