import uuid

import oops_wsgi

from canonical_raven.middleware import Sentry

from django.conf import settings
from oops_wsgi import django as oops_wsgi_django
from timeline_django import setup as timeline_django_setup
from timeline_django.wsgi import make_app as timeline_django_make_app
from timeline.wsgi import make_app as timeline_wsgi_make_app
from whitenoise.django import DjangoWhiteNoise

from versioninfo import version_info

from identityprovider.oops_config import global_error_reporter


class OOPSWSGIHandler(oops_wsgi_django.OOPSWSGIHandler):
    """Pre-generates an oops id on exception."""

    def handle_uncaught_exception(self, request, resolver, exc_info):
        # Generate OOPS id early so that we can render the error page right
        # away and not depend on passing thread locals around.
        if 'oops.report' in request.environ:
            unique_id = uuid.uuid4().hex
            request.environ['oops.report']['id'] = "OOPS-%s" % unique_id
        return super(OOPSWSGIHandler, self).handle_uncaught_exception(
            request, resolver, exc_info)

    def __call__(self, environ, start_response):
        def start_response_with_exc_info(status, headers, exc_info=None):
            """Custom start_response callback for wsgi."""
            # This will pass the exception information back to oops_wsgi.  It
            # should not be necessary once
            # https://code.djangoproject.com/ticket/16674 has been merged.
            if exc_info is None:
                exc_info = environ['oops.context'].get('exc_info', None)
            return start_response(status, headers, exc_info)
        return super(OOPSWSGIHandler, self).__call__(
            environ, start_response_with_exc_info)


class BzrRevnoMiddleware(object):
    REVNO_HEADER = ('X-Bzr-Revision-Number', version_info['revno'])

    def __init__(self, app):
        self.app = app

    def __call__(self, environ, start_response):

        def custom_start_response(status, headers, exc_info=None):
            headers.append(self.REVNO_HEADER)
            return start_response(status, headers, exc_info)

        return self.app(environ, custom_start_response)


def make_app():
    """Encapsulate our SSO handler in middleware."""

    app = OOPSWSGIHandler()

    # First we create the config that defines what to do with the oopses.
    config = global_error_reporter.setup()

    # Install the hooks to record the sql queries
    timeline_django_setup.setup_for_requests()

    # Set up the timeline in the django and wsgi envs
    app = timeline_django_make_app(app)
    app = timeline_wsgi_make_app(app)

    app = BzrRevnoMiddleware(app)
    app = DjangoWhiteNoise(app)

    # Then we wrap the django app in the oops one
    app = oops_wsgi.make_app(
        app, config, oops_on_status=['500'],
        soft_start_timeout=settings.HANDLER_TIMEOUT_MILLIS)
    if settings.USE_SENTRY:
        app = Sentry(app, soft_start_timeout=settings.HANDLER_TIMEOUT_MILLIS)
    return app
