# Copyright 2010 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

import os
import random
import string
import struct

from django.conf import settings
from django.contrib.auth.models import User
from django.db import models
from django.utils.timezone import now


__all__ = ['Token', 'Consumer']


TOKEN_LENGTH = getattr(settings, 'OAUTH_TOKEN_LENGTH', 50)
TOKEN_SECRET_LENGTH = getattr(settings, 'OAUTH_TOKEN_SECRET_LENGTH', 50)
CONSUMER_SECRET_LENGTH = getattr(settings, 'OAUTH_CONSUMER_SECRET_LENGTH', 30)


def _set_seed():
    if (not hasattr(_set_seed, 'seed') and os.path.exists("/dev/random")):
        data = open("/dev/random").read(struct.calcsize('Q'))
        random.seed(struct.unpack('Q', data))
        _set_seed.seed = True


def generate_random_string(length):
    _set_seed()
    return ''.join(random.choice(string.ascii_letters)
                   for x in range(length))


def generate_random_string_secret():
    return generate_random_string(CONSUMER_SECRET_LENGTH)


def generate_random_string_token():
    return generate_random_string(TOKEN_LENGTH)


class Consumer(models.Model):
    """An OAuth consumer (client)."""

    user = models.OneToOneField(User, related_name='oauth_consumer')

    secret = models.CharField(
        max_length=255, blank=True, default=generate_random_string_secret,
    )

    created_at = models.DateTimeField(auto_now_add=True)
    updated_at = models.DateTimeField(auto_now=True)

    class Meta:
        db_table = 'oauth_consumer'

    def __unicode__(self):
        return self.key

    @property
    def key(self):
        return self.user.username


class Token(models.Model):
    """An OAuth access token."""

    consumer = models.ForeignKey(Consumer)

    token = models.CharField(
        max_length=TOKEN_LENGTH, default=generate_random_string_token,
        primary_key=True)

    token_secret = models.CharField(
        max_length=TOKEN_SECRET_LENGTH, default=generate_random_string_secret)

    name = models.CharField(max_length=255, blank=True)

    created_at = models.DateTimeField(default=now)
    updated_at = models.DateTimeField(default=now)

    class Meta:
        db_table = 'oauth_token'

    def __unicode__(self):
        return self.token

    # This class will be used from Piston oauth handlers, which assume the
    # Token instance have key, secret and user attributes.

    @property
    def key(self):
        return self.token

    @property
    def secret(self):
        return self.token_secret

    @property
    def user(self):
        return self.consumer.user

    def serialize(self):
        return {
            'consumer_key': self.consumer.key,
            'consumer_secret': self.consumer.secret,
            'token': self.token,
            'token_secret': self.token_secret,
            'name': self.name,
            'created': str(self.created_at),
            'updated': str(self.updated_at),
        }
