import unittest
import os
import shutil
import subprocess
import sys
import tempfile

from django.conf import settings
from django.test import runner as dj_runner
from gargoyle import gargoyle
from olstests import suites
from olstestsdjango import runner

try:
    import cProfile as profile
except ImportError:
    import profile


class SSOTestSuite(suites.TestSuite):
    """Test suite for the canonical identity provider project.

    This centralizes the setup for the test environment and applies to all
    tests. Use only for setups that can't be done at a lower level.

    """

    def setUp(self):
        super(SSOTestSuite, self).setUp()
        # gargoyle caches settings. Setting timeout to 0 prevents this,
        # increasing test isolation.
        gargoyle.timeout = 0
        # We need unique identifiers for several pieces, the process pid is a
        # good source of uniqueness which works for concurrent runs on the same
        # host as well as subprocesses for a single run.
        pid = os.getpid()
        # We want the db to be unique for each worker (or the master if no
        # concurrency is involved).
        for k, v in settings.DATABASES.items():
            v['TEST']['NAME'] = 'test_{}_{}'.format(v['NAME'], pid)
        # Caches need to be unique, we get there by using a unique key prefix
        settings.CACHES['default']['KEY_PREFIX'] = '{}-'.format(pid)
        # Make oops dir unique
        settings.OOPSES['publishers'][0]['error_dir'] = os.path.join(
            settings.LOGS_DIR, 'www-oops-{}'.format(pid))
        # Read-only mode relies on the 'db.readonly' file in
        # DBFAILOVER_FLAG_DIR, this needs to be unique
        settings.DBFAILOVER_FLAG_DIR = tempfile.mkdtemp()
        self.addCleanup(shutil.rmtree, settings.DBFAILOVER_FLAG_DIR, True)
        # 'interactive' in the call below is used internally to autoclobber dbs
        # (using 'not interactive'). 'verbosity' is set to 1 to only display
        # creating/deletion of the databases.
        old_config = dj_runner.setup_databases(verbosity=1,
                                               interactive=False)
        self.addCleanup(self.teardown_databases, old_config)

    def teardown_databases(self, old_config):
        old_names, mirrors = old_config
        for connection, old_name, destroy in old_names:
            if destroy:
                connection.creation.destroy_test_db(old_name, verbosity=1)


def subprocess_run(args):
    try:
        return subprocess.check_output(args,
                                       stderr=subprocess.STDOUT,
                                       stdin=subprocess.PIPE)
    except subprocess.CalledProcessError as e:
        # By default, the bare exception won't reveal the output from the
        # failed command, we do before re-raising.
        sys.stderr.write('Command {} failed with {}'.format(
            ' '.join(args), e.output))
        raise


class SSOContainerSuite(suites.SubprocessedSuite):

    def setUp(self):
        super(SSOContainerSuite, self).setUp()
        self.worker_dir = os.getcwd()
        # Spawn the ephemeral container
        self.worker_name = 'sso-worker-{:>02d}'.format(self.unique)
        subprocess_run(['uci-vms', 'start', self.worker_name])
        self.addCleanup(subprocess_run, ['uci-vms', 'stop', self.worker_name])
        # Finish container setup (food for a make target?)
        subprocess_run(['uci-vms', 'shell', self.worker_name,
                        '(', 'cd', self.worker_dir, '&&',
                        'make', 'start-db',
                        ')'])
        # Build the command running the tests, connecting the pipes
        self.cmd = ['uci-vms', 'shell', self.worker_name,
                    # And now we run the tests inside the container
                    '(', 'cd', self.worker_dir, '&&',
                    'make', 'test',
                    'ARGS="-f subunit --pick {}/{}"'.format(
                        self.unique, self.factor),
                    ')']


class NovaSuite(suites.SubprocessedSuite):

    def setUp(self):
        super(NovaSuite, self).setUp()
        self.worker_dir = '~/sso/work'
        # Spawn the ephemeral container
        self.worker_name = 'sso-nova-worker-{:>02d}'.format(self.unique)
        # Devs may want to leave the db running
        self.start_db()
        # Build the command running the tests, connecting the pipes
        self.cmd = ['uci-vms2', 'shell', self.worker_name,
                    # And now we run the tests inside the container
                    '(', 'cd', self.worker_dir, '&&',
                    'make', 'test',
                    # -c4 below is ok for m1.large
                    'ARGS="-f subunit --pick {}/{} -c4"'.format(
                        self.unique, self.factor),
                    ')']

    def start_db(self):
        """Start the databse (and stop it when the suite has run).

        Depending on dev preference, the tests can start/stop the db for each
        run or assume that it is started.
        """
        subprocess_run(['uci-vms2', 'shell', self.worker_name,
                        '(', 'cd', self.worker_dir,
                        # FIXME: We *can* check that the remote branch is up to
                        # date or not, and push --overwrite from the host then
                        # bzr update in the container and 'make bootstrap' only
                        # when needed -- vila 2016-01-08

                        # '&&', 'bzr', 'pull', '-v',
                        # '&&', 'make', 'bootstrap'
                        '&&', 'make', 'start-db'
                        ')'])
        self.addCleanup(subprocess_run, ['uci-vms2', 'shell', self.worker_name,
                                         '(', 'cd', self.worker_dir, '&&',
                                         'make', 'stop-db',
                                         ')'])


class TestRunner(runner.TestRunner):

    split_suites = dict(fork=suites.ForkedSuite,
                        lxc=SSOContainerSuite,
                        nova=NovaSuite)
    wrapping_test_suite = SSOTestSuite


class ProfilingTestSuiteRunner(TestRunner):

    def run_suite(self, suite, **kwargs):
        runner = unittest.TextTestRunner(
            verbosity=self.verbosity, failfast=self.failfast).run

        profile.runctx(
            'result = run_tests(suite)',
            {
                'run_tests': runner,
                'suite': suite,
            },
            locals(),
            getattr(settings, 'TEST_PROFILE', None),
        )
        return locals()['result']
