# Copyright 2012-2013 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).
from identityprovider.models import OpenIDRPConfig
from saml2idp import google_apps, salesforce
from saml2idp.base import Processor
from saml2idp.exceptions import CannotHandleAssertion


__all__ = [
    'ExpensifyProcessor',
    'GoogleAppsProcessor',
    'SalesForceProcessor',
    'SalesForceAttributeProcessor',
]


class CanonicalOverrides(object):

    def _get_canonical_email(self, account):
        preferred = None
        # user has @canonical email address?
        emails = account.emailaddress_set.filter(
            email__iendswith='@canonical.com')
        for email in emails:
            # only consider emails formatted like first.last@canonical.com
            name, _ = email.email.split('@')
            if '.' in name and '+' not in name:
                preferred = email.email
                break
        if preferred is None and account.preferredemail in emails:
            # support preferred email if it's a @canonical.com email
            preferred = account.preferredemail.email
        return preferred

    def _determine_subject(self):
        account = self._django_request.user
        preferred = account.preferredemail.email

        # check if RP prefers @canonical emails
        trust_root = self._request_params.get('ACS_URL')
        if trust_root is not None:
            rpconfig = OpenIDRPConfig.objects.for_url(trust_root)
            if rpconfig is not None and rpconfig.prefer_canonical_email:
                canonical_email = self._get_canonical_email(account)
                if canonical_email is not None:
                    preferred = canonical_email

        self._subject = preferred


class GoogleAppsProcessor(CanonicalOverrides, google_apps.Processor):
    """
    Google Apps SAML 2.0 Processor, specific to Canonical Identity Provider.
    """


class SalesForceProcessor(CanonicalOverrides, salesforce.Processor):
    """
    SalesForce SAML 2.0 Processor, specific to Canonical Identity Provider.
    """


class SalesForceAttributeProcessor(CanonicalOverrides, salesforce.Processor):
    """
    SalesForce SAML 2.0 Processor, specific to Canonical Identity Provider.
    Adds attributes for SalesForce.
    """

    def _validate_request(self):
        """Validate the _saml_request.

        Throw an Exception if the validation does not succeed.
        """
        # skip salesforce.Processor as that requires the acs_url to be on the
        # salesforce.com domain
        Processor._validate_request(self)

        acs_domain = self._sp_config.get('acs_domain', 'salesforce.com')
        acs_name = self._sp_config.get('acs_name', 'SalesForce')

        if not ".%s" % acs_domain in self._request_params['ACS_URL']:
            raise CannotHandleAssertion(
                "AssertionConsumerService is not a %s URL." % acs_name)

    def _determine_audience(self):
        self._audience = self._sp_config.get(
            'saml_url',
            'https://saml.salesforce.com')

    def _get_canonical_email(self, account):
        """
        For the SalesForce portal, use email+portal@canonical.com.
        """
        email = super(SalesForceAttributeProcessor,
                      self)._get_canonical_email(account)
        if email is not None:
            username, domain = email.split('@')
            pattern = self._sp_config.get('email_pattern',
                                          '{username}+portal@{domain}')
            email = pattern.format(username=username, domain=domain)
            return email

    def _format_assertion(self):
        self._assertion_params['ATTRIBUTES'] = self._sp_config.get(
            'attributes', {})
        super(SalesForceAttributeProcessor, self)._format_assertion()


class ExpensifyProcessor(CanonicalOverrides, google_apps.Processor):
    """
    Expensify SAML 2.0 Processor, specific to Canonical Identity Provider.
    """

    def _validate_request(self):
        # override validation as the google_apps.Processor constrains requests
        # to the google.com domain only
        Processor._validate_request(self)
        if '.expensify.com/' not in self._request_params['ACS_URL']:
            raise CannotHandleAssertion(
                'AssertionConsumerService is not an Expensify URL.')


class CornerstoneProcessor(CanonicalOverrides, google_apps.Processor):
    """
    Cornerstone SAML 2.0 Processor, specific to Canonical Identity Provider.
    """
    def _validate_request(self):
        # override validation as the salesforce.Processor constrains requests
        # to the salesforce.com domain only
        Processor._validate_request(self)
        if '.csod.com/' not in self._request_params['ACS_URL']:
            raise CannotHandleAssertion(
                'AssertionConsumerService is not a Cornerstone URL.')

    def _decode_request(self):
        """
        Decodes non compressed request data.
        """
        Processor._decode_request(self)
