"""Tests for SAML2 Processors.

A lot of this is borrowed directly from the saml2idp tests.
"""
import base64
import copy

from django.core.urlresolvers import reverse
from django.test.client import Client
from gargoyle.models import GLOBAL, Switch
from gargoyle.testutils import switches
from mock import patch
from pyquery import PyQuery
from saml2idp import codex, saml2idp_metadata
from saml2idp.exceptions import CannotHandleAssertion
from saml2idp.tests import salesforce, google_apps
from saml2idp.urls import deeplink_url_patterns

import identityprovider.urls

from identityprovider.models import (
    Account,
    EmailAddress,
    OpenIDRPConfig,
    twofactor,
)
from identityprovider.models.const import EmailStatus
from identityprovider.tests import DEFAULT_USER_PASSWORD
from identityprovider.tests.utils import AuthenticatedTestCase


SAML_TEAM = 'saml2team'  # Name of team enabled for SAML 2.0


def find_cert_path():
    """
    Find the right path to certificates
    """
    from identityprovider import tests
    return tests.__path__[0] + '/cert/'


def get_saml_response(response):
    tree = PyQuery(response.content)
    inputtag = tree.find('input[name="SAMLResponse"]')
    assert len(inputtag) == 1
    encoded_response = inputtag[0].get('value')
    samlresponse = base64.b64decode(encoded_response)
    return samlresponse


class GoogleAppsAssertionTestCase(AuthenticatedTestCase):
    """
    Test SAML Assertions for a Google Apps Service Point.

    Sub-classes with identical behavior, but different ACS_URL or REQUEST_DATA
    simply specify their own attributes on the class.
    """
    ACS_URL = google_apps.GOOGLE_APPS_ACS
    REQUEST_DATA = google_apps.REQUEST_DATA
    PROCESSOR = 'ubuntu_sso_saml.processors.GoogleAppsProcessor'

    def setUp(self, *args, **kwargs):
        super(GoogleAppsAssertionTestCase, self).setUp(*args, **kwargs)
        # Save settings.
        new_saml2idp_config = copy.deepcopy(saml2idp_metadata.SAML2IDP_CONFIG)
        self.patch(saml2idp_metadata, 'SAML2IDP_CONFIG',
                   new=new_saml2idp_config)
        path = find_cert_path()
        cert_file = path + 'certificate.pem'
        priv_file = path + 'private-key.pem'
        test_config = {
            'certificate_file': cert_file,
            'private_key_file': priv_file,
            'signing': True,
        }
        saml2idp_metadata.SAML2IDP_CONFIG.update(test_config)

        saml2idp_metadata.SAML2IDP_REMOTES['foobar'] = {
            'acs_url': self.ACS_URL,
            'processor': self.PROCESSOR,
        }
        self.addCleanup(saml2idp_metadata.SAML2IDP_REMOTES.pop, 'foobar')

        # enable TWOFACTOR flag
        switch = Switch.objects.create(key='TWOFACTOR', status=GLOBAL)
        self.addCleanup(switch.delete)

        # mock twofactor auth
        self.patch(twofactor, 'authenticate')
        self.saml2team = self.factory.make_team(SAML_TEAM)

    def test_authnrequest_handled(self):
        """ AuthnRequest isn't handled by a Processor. """
        # Arrange/Act:
        response = self.client.get('/+saml', data=self.REQUEST_DATA,
                                   follow=False)

        # Assert:
        self.assertEqual(response.status_code, 302)

    def test_authnrequest_csrf_exempt(self):
        client = Client(enforce_csrf_checks=True)
        response = client.post('/+saml', data=self.REQUEST_DATA,
                               follow=False)
        self.assertEqual(response.status_code, 302)

    def test_user_logged_in(self):
        """ Unable to produce a valid Assertion for SAML 2-enabled user. """
        # Act:
        self.factory.add_account_to_team(self.account, self.saml2team)
        # upgrade to 2f session
        response = self.client.post(reverse('twofactor'),
                                    {'oath_token': '123456'})
        response = self.client.get('/+saml', data=self.REQUEST_DATA,
                                   follow=True)
        samlresponse = get_saml_response(response)

        # Assert:
        self.assertIn(self.login_email, samlresponse)

    def test_user_has_different_preferred_email(self):
        """ User's preferred email is not being passed as SAML identifier. """
        # make sure no @canonical.com email address is used so the preferred
        # email is returned
        self.login_email = 'mark@example.com'
        self.account = self.factory.make_account(
            email=self.login_email, password=self.login_password)
        assert self.client.login(
            username=self.login_email, password=self.login_password)
        # Act:
        self.factory.add_account_to_team(self.account, self.saml2team)
        # upgrade to 2f session
        response = self.client.post(reverse('twofactor'),
                                    {'oath_token': '123456'})
        # create alternate email
        # and make sure the email is linked to the account
        alternate_email = 'alternate@example.com'
        alternate = EmailAddress.objects.create(
            email=alternate_email, account=self.account,
            status=EmailStatus.VALIDATED)
        # and make the alternate email the preferred
        self.account.preferredemail = alternate
        self.account.save()

        account = Account.objects.get(pk=self.account.pk)
        self.assertEqual(account.preferredemail, alternate)
        response = self.client.get('/+saml', data=self.REQUEST_DATA,
                                   follow=True)
        samlresponse = get_saml_response(response)

        # Assert:
        self.assertIn(alternate_email, samlresponse)

    def test_user_not_2f_but_no_rpconfig(self):
        response = self.client.get(
            '/+saml', data=self.REQUEST_DATA, follow=True)

        self.assertRedirects(response, reverse('login_process'))

    def test_user_not_2f_but_rpconfig_without_flag(self):
        OpenIDRPConfig.objects.create(trust_root=self.ACS_URL)

        response = self.client.get(
            '/+saml', data=self.REQUEST_DATA, follow=True)

        self.assertRedirects(response, reverse('login_process'))

    def test_user_not_2f_but_not_required(self):
        OpenIDRPConfig.objects.create(
            trust_root=self.ACS_URL, twofactor_required=True)

        with switches(TWOFACTOR=False):
            response = self.client.get(
                '/+saml', data=self.REQUEST_DATA, follow=True)

        self.assertRedirects(response, reverse('login_process'))

    def test_user_not_2f_but_required(self):
        # make sure the rpconfig entry requires 2f
        OpenIDRPConfig.objects.create(
            trust_root=self.ACS_URL, twofactor_required=True)

        with switches(TWOFACTOR=True):
            response = self.client.get('/+saml', data=self.REQUEST_DATA,
                                       follow=True)

        # make sure the last redirect was for twofactor
        self.assertIn(reverse('twofactor'), response.redirect_chain[-1][0])

    def setup_saml_emails(self, prefer_canonical_email=False,
                          email='first.last@canonical.com',
                          status=EmailStatus.VALIDATED):
        self.factory.add_account_to_team(self.account, self.saml2team)
        # add @canonical.com email address to the account
        account = EmailAddress.objects.get(email=self.login_email).account
        EmailAddress.objects.create(
            account=account, email=email, status=status)
        # create an RP that prefers @canonical emails
        OpenIDRPConfig.objects.create(
            trust_root=self.ACS_URL,
            prefer_canonical_email=prefer_canonical_email)

        return account.preferredemail.email

    def do_saml_request(self):
        # do twofactor dance
        self.client.post(reverse('twofactor'), {'oath_token': '123456'})
        # do saml dance
        response = self.client.get('/+saml', data=self.REQUEST_DATA,
                                   follow=True)
        samlresponse = get_saml_response(response)
        return samlresponse

    def test_canonical_email_is_preferred(self):
        preferred = self.setup_saml_emails(prefer_canonical_email=True)
        samlresponse = self.do_saml_request()
        # verify @canonical.com email address is returned despite it's not the
        # preferredemail
        self.assertIn('first.last@canonical.com', samlresponse)
        self.assertNotEqual('first.last@canonical.com', preferred)

    def test_canonical_email_is_preferred_despite_case(self):
        email = 'FIRST.LAST@CANONICAL.COM'
        preferred = self.setup_saml_emails(prefer_canonical_email=True,
                                           email=email)
        samlresponse = self.do_saml_request()
        # verify @canonical.com email address is returned despite it's not the
        # preferredemail
        self.assertIn(email, samlresponse)
        self.assertNotEqual(email, preferred)

    def test_only_long_form_canonical_email_is_preferred(self):
        email = self.login_email
        # add non-preferred long form email address
        EmailAddress.objects.create(
            account=self.account, email='first.last+alias@canonical.com',
            status=EmailStatus.VALIDATED)
        preferred = self.setup_saml_emails(prefer_canonical_email=True)
        self.assertEqual(preferred, email)
        self.assertNotEqual(preferred, 'first.last@canonical.com')

        samlresponse = self.do_saml_request()
        # verify @canonical.com email address is returned despite it's not the
        # preferredemail
        self.assertIn('first.last@canonical.com', samlresponse)
        self.assertNotIn(email, samlresponse)

    def test_canonical_email_is_not_preferred(self):
        preferred = self.setup_saml_emails(prefer_canonical_email=False)
        samlresponse = self.do_saml_request()
        # verify @canonical.com email address is *not* returned despite because
        # the flag was not set
        self.assertIn(preferred, samlresponse)
        self.assertNotEqual('first.last@canonical.com', preferred)

    def test_canonical_email_without_rpconfig(self):
        preferred = self.setup_saml_emails()
        # make sure there is no RP for this url
        OpenIDRPConfig.objects.filter(trust_root=self.ACS_URL).delete()

        samlresponse = self.do_saml_request()
        # verify @canonical.com email address is *not* returned despite because
        # the flag was not set
        self.assertIn(preferred, samlresponse)
        self.assertNotEqual('first.last@canonical.com', preferred)

    def test_short_form_canonical_email_ignored(self):
        # start off a clean slate
        short_email = EmailAddress.objects.get(
            account=self.account, email='test@canonical.com')
        short_email.status = EmailStatus.VALIDATED
        short_email.save()
        EmailAddress.objects.create(account=self.account, email='foo@foo.com',
                                    status=EmailStatus.PREFERRED)
        self.login_email = 'foo@foo.com'

        # make sure the short email it's not the preferred address
        preferred = self.setup_saml_emails(prefer_canonical_email=True)
        assert preferred == 'foo@foo.com'
        # make sure there's no long form @canonical email
        EmailAddress.objects.filter(
            account=self.account, email='first.last@canonical.com').delete()

        samlresponse = self.do_saml_request()
        # verify @canonical.com email address is *not* returned
        self.assertIn(preferred, samlresponse)
        self.assertNotEqual('first.last@canonical.com', preferred)
        self.assertNotEqual('test@canonical.com', preferred)

    def test_non_canonical_email(self):
        # prepare account for SAML
        account = self.factory.make_account()
        preferred = account.preferredemail.email
        self.factory.add_account_to_team(account, self.saml2team)
        OpenIDRPConfig.objects.create(trust_root=self.ACS_URL,
                                      prefer_canonical_email=True)

        # make sure account has only one email
        assert account.emailaddress_set.count() == 1
        # and it's not a @canonical.com one
        assert '@canonical.com' not in preferred

        assert self.client.login(
            username=preferred, password=DEFAULT_USER_PASSWORD)

        samlresponse = self.do_saml_request()
        self.assertIn(preferred, samlresponse)


class SalesForceAssertionTestCase(GoogleAppsAssertionTestCase):
    """
    Test SAML Assertions for a SalesForce Service Point.
    """
    ACS_URL = salesforce.SALESFORCE_ACS
    REQUEST_DATA = salesforce.REQUEST_DATA
    PROCESSOR = 'ubuntu_sso_saml.processors.SalesForceProcessor'


class SalesForcePortalAssertionTestCase(SalesForceAssertionTestCase):
    ACS_URL = salesforce.SALESFORCE_ACS
    REQUEST_DATA = salesforce.REQUEST_DATA
    PROCESSOR = 'ubuntu_sso_saml.processors.SalesForceAttributeProcessor'

    def setUp(self, *args, **kwargs):
        super(SalesForcePortalAssertionTestCase, self).setUp(*args, **kwargs)
        target = 'https://somesite.salesforce.com/%(target)s'
        saml2idp_metadata.SAML2IDP_REMOTES['foobar'].update({
            'links': {
                # NOTE: This doesn't match the SALESFORCE_ACS,
                # but it will in production.
                r'portal/(?P<target>\w+)': target,
            },
            'attributes': {
                'organization_id': 'SAMPLE_ORG_ID',
                'portal_id': 'SAMPLE_PORTAL_ID',
            },
            'email_pattern': '{username}+portal@{domain}',
        })

        # Update the site's URLs to include the new pattern.
        self.patch(identityprovider.urls, 'urlpatterns',
                   new=identityprovider.urls.urlpatterns[:])
        identityprovider.urls.urlpatterns += deeplink_url_patterns(
            'ubuntu_sso_saml.views',
            r'^\+saml/init/%s$', 'saml_init'
        )

    def test_canonical_email_is_preferred(self):
        preferred = self.setup_saml_emails(prefer_canonical_email=True)
        samlresponse = self.do_saml_request()
        self.assertIn('first.last+portal@canonical.com', samlresponse)
        self.assertNotEqual('first.last+portal@canonical.com', preferred)

    def test_canonical_email_is_preferred_despite_case(self):
        email = 'FIRST.LAST@CANONICAL.COM'
        portal_email = 'FIRST.LAST+portal@CANONICAL.COM'
        preferred = self.setup_saml_emails(prefer_canonical_email=True,
                                           email=email)
        samlresponse = self.do_saml_request()
        # verify @canonical.com email address is returned despite it's not the
        # preferredemail
        self.assertIn(portal_email, samlresponse)
        self.assertNotEqual(portal_email, preferred)

    def test_only_long_form_canonical_email_is_preferred(self):
        email = self.login_email
        portal_email = 'first.last+portal@canonical.com'
        # add non-preferred long form email address
        EmailAddress.objects.create(
            account=self.account, email='first.last+alias@canonical.com',
            status=EmailStatus.VALIDATED)
        preferred = self.setup_saml_emails(prefer_canonical_email=True)
        self.assertEqual(preferred, email)
        self.assertNotEqual(preferred, 'first.last@canonical.com')

        samlresponse = self.do_saml_request()
        # verify @canonical.com email address is returned despite it's not the
        # preferredemail
        self.assertIn(portal_email, samlresponse)
        self.assertNotEqual(portal_email, preferred)

    def test_non_canonical_email(self):
        # prepare account for SAML
        account = self.factory.make_account()
        preferred = account.preferredemail.email
        self.factory.add_account_to_team(account, self.saml2team)
        OpenIDRPConfig.objects.create(trust_root=self.ACS_URL,
                                      prefer_canonical_email=True)

        # make sure account has only one email
        assert account.emailaddress_set.count() == 1
        # and it's not a @canonical.com one
        assert '@canonical.com' not in preferred

        assert self.client.login(
            username=preferred, password=DEFAULT_USER_PASSWORD)

        samlresponse = self.do_saml_request()
        self.assertIn(preferred, samlresponse)
        self.assertNotIn('+portal@', samlresponse)

    def test_portal_deeplink(self):
        """Unable to produce a valid Portal Assertion.

        (for SAML 2-enabled user)
        """
        # Act:
        self.factory.add_account_to_team(self.account, self.saml2team)
        # upgrade to 2f session
        response = self.client.post(reverse('twofactor'),
                                    {'oath_token': '123456'})
        response = self.client.get(
            '/+saml/init/portal/SOMETARGETID', data=self.REQUEST_DATA,
            follow=True)
        samlresponse = get_saml_response(response)

        tree = PyQuery(response.content)
        relsttag = tree.find('input[name="RelayState"]')
        assert len(relsttag) == 1
        relaystate = relsttag[0].get('value')

        samlsoup = PyQuery(samlresponse.replace('xmlns:', 'xmlnamespace:'))

        org_attr = samlsoup.find('attribute[name="organization_id"]')
        assert len(org_attr) == 1
        org_val = org_attr[0].find('attributevalue')

        port_attr = samlsoup.find('attribute[name="portal_id"]')
        assert len(port_attr) == 1
        port_val = port_attr[0].find('attributevalue')

        # Assert:
        self.assertIn(self.login_email, samlresponse)
        self.assertEqual(relaystate,
                         'https://somesite.salesforce.com/SOMETARGETID')
        self.assertEqual(org_val.text, 'SAMPLE_ORG_ID')
        self.assertEqual(port_val.text, 'SAMPLE_PORTAL_ID')

    def test_portal_deeplink_with_custom_attributes(self):
        # patch remote config
        saml2idp_metadata.SAML2IDP_REMOTES['foobar'].update({
            'attributes': {
                'organization_id': 'SAMPLE_ORG_ID',
            },
        })

        # Act:
        self.factory.add_account_to_team(self.account, self.saml2team)
        # upgrade to 2f session
        response = self.client.post(reverse('twofactor'),
                                    {'oath_token': '123456'})
        response = self.client.get(
            '/+saml/init/portal/SOMETARGETID', data=self.REQUEST_DATA,
            follow=True)
        samlresponse = get_saml_response(response)

        tree = PyQuery(response.content)
        relsttag = tree.find('input[name="RelayState"]')
        assert len(relsttag) == 1
        relaystate = relsttag[0].get('value')

        samlsoup = PyQuery(samlresponse.replace('xmlns:', 'xmlnamespace:'))

        org_attr = samlsoup.find('attribute[name="organization_id"]')
        assert len(org_attr) == 1
        org_val = org_attr[0].find('attributevalue')

        port_attr = samlsoup.find('attribute[name="portal_id"]')
        # this remote doesn't have a portal_id attribute
        assert len(port_attr) == 0

        # Assert:
        self.assertIn(self.login_email, samlresponse)
        self.assertEqual(relaystate,
                         'https://somesite.salesforce.com/SOMETARGETID')
        self.assertEqual(org_val.text, 'SAMPLE_ORG_ID')

    def test_custom_email_pattern(self):
        # patch remote config
        saml2idp_metadata.SAML2IDP_REMOTES['foobar'].update({
            'email_pattern': '{username}@hr.{domain}'
        })

        preferred = self.setup_saml_emails(prefer_canonical_email=True)
        # make sure transformed email is not a valid email for account
        account = Account.objects.get_by_email(self.login_email)
        account_emails = [email.email
                          for email in account.emailaddress_set.all()]
        self.assertNotIn('first.last@hr.canonical.com', account_emails)
        # and make sure the preferred canonical email address is in there
        self.assertIn('first.last@canonical.com', account_emails)
        # even if it's not the preferred email
        self.assertNotEqual(preferred, 'first.last@canonical.com')

        samlresponse = self.do_saml_request()
        # make sure the login email was transformed
        self.assertIn('first.last@hr.canonical.com', samlresponse)

    def test_custom_email_pattern_with_short_form_canonical_email(self):
        # patch remote config
        saml2idp_metadata.SAML2IDP_REMOTES['foobar'].update({
            'email_pattern': '{username}@hr.{domain}'
        })

        preferred = self.setup_saml_emails(prefer_canonical_email=True)
        account = Account.objects.get_by_email(self.login_email)
        # make sure long-form email is not registered with account
        EmailAddress.objects.filter(
            account=account, email='first.last@canonical.com').delete()
        # make sure transformed email is not a valid email for account
        account_emails = [email.email
                          for email in account.emailaddress_set.all()]
        self.assertNotIn('test@hr.canonical.com', account_emails)
        self.assertNotIn('first.last@hr.canonical.com', account_emails)
        # and make sure the preferred canonical email address is in there
        self.assertIn('test@canonical.com', account_emails)
        self.assertNotIn('first.last@canonical.com', account_emails)
        # even if it's not the preferred email
        self.assertNotEqual(preferred, 'first.last@canonical.com')

        samlresponse = self.do_saml_request()
        # make sure the login email was transformed
        self.assertIn('test@hr.canonical.com', samlresponse)

    def make_saml_request(self, acs_url=None, saml_url=None):
        saml_request = self.REQUEST_DATA['SAMLRequest']
        if acs_url or saml_url:
            saml_request = base64.b64decode(saml_request)
            if acs_url:
                saml_request = saml_request.replace(
                    'https://login.salesforce.com', acs_url)
            if saml_url:
                saml_request = saml_request.replace(
                    'https://saml.salesforce.com', saml_url)
            saml_request = base64.b64encode(saml_request)
        return saml_request

    def test_custom_acs_domain(self):
        # patch remote config
        saml2idp_metadata.SAML2IDP_REMOTES['foobar'].update({
            'acs_domain': 'cloudforce.com',
            'acs_url': 'https://login.cloudforce.com'
        })
        saml_request = self.make_saml_request(
            acs_url='https://login.cloudforce.com')

        with patch.dict(self.REQUEST_DATA, SAMLRequest=saml_request):
            samlresponse = self.do_saml_request()
        self.assertIn('https://login.cloudforce.com', samlresponse)

    @patch('saml2idp.registry.logger.debug')
    def test_custom_acs_name(self, mock_debug):
        saml2idp_metadata.SAML2IDP_REMOTES['foobar'].update({
            'acs_domain': 'cloudforce.com',
            'acs_name': 'CloudForce',
        })

        with self.assertRaises(CannotHandleAssertion):
            self.do_saml_request()
        self.assertIn('CloudForce', str(mock_debug.call_args))

    def test_custom_saml_url(self):
        # patch remote config
        saml2idp_metadata.SAML2IDP_REMOTES['foobar'].update({
            'saml_url': 'https://saml.cloudforce.com',
        })
        saml_request = self.make_saml_request(
            saml_url='https://saml.cloudforce.com')

        with patch.dict(self.REQUEST_DATA, SAMLRequest=saml_request):
            samlresponse = self.do_saml_request()
        self.assertIn('https://saml.cloudforce.com', samlresponse)


class ExpensifyAssertionTestCase(GoogleAppsAssertionTestCase):
    """Test SAML Assertions for an Expensify Service Point."""

    ACS_URL = (
        'https://www.expensify.com/authentication/saml/loginCallback?'
        'domain=example.com')
    SAML_REQUEST = codex.deflate_and_base64_encode("""
<?xml version="1.0" encoding="UTF-8"?>
<samlp:AuthnRequest xmlns:samlp="urn:oasis:names:tc:SAML:2.0:protocol"
                    xmlns:saml="urn:oasis:names:tc:SAML:2.0:assertion"
                    ID="fsdajhfdskjhfdajfadfhksfhdklfhkfjshlfsahkdhsa"
                    Version="2.0"
                    IssueInstant="2015-09-24T15:40:06Z"
                    Destination="https://www.expensify.com/authentication/saml/loginCallback?domain=example.com"
                    ProtocolBinding="urn:oasis:names:tc:SAML:2.0:bindings:HTTP-POST"
                    AssertionConsumerServiceURL="https://www.expensify.com/authentication/saml/loginCallback?domain=example.com"
                    >
    <saml:Issuer>https://www.expensify.com</saml:Issuer>
    <samlp:NameIDPolicy Format="urn:oasis:names:tc:SAML:1.1:nameid-format:emailAddress"
                        AllowCreate="true"
                        />
    <samlp:RequestedAuthnContext Comparison="exact">
        <saml:AuthnContextClassRef>urn:oasis:names:tc:SAML:2.0:ac:classes:PasswordProtectedTransport</saml:AuthnContextClassRef>
    </samlp:RequestedAuthnContext>
</samlp:AuthnRequest>
""")  # noqa
    RELAY_STATE = (
        'https://www.expensify.com/authentication/saml/login.php')
    REQUEST_DATA = {
        'SAMLRequest': SAML_REQUEST,
        'RelayState': RELAY_STATE,
    }
    PROCESSOR = 'ubuntu_sso_saml.processors.ExpensifyProcessor'


class CornerstoneAssertionTestCase(GoogleAppsAssertionTestCase):
    """Test SAML Assertions for an Expensify Service Point."""

    ACS_URL = 'https://site.csod.com/samldefault.aspx'
    SAML_REQUEST = base64.b64encode("""
<?xml version="1.0" encoding="UTF-8"?>
<AuthnRequest xmlns:xsd="http://www.w3.org/2001/XMLSchema"
              xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
              ID="THE_ID"
              Version="2.0"
              IssueInstant="2016-01-25T15:18:25.9931866Z"
              Destination="https://login.staging.ubuntu.com/saml"
              ProtocolBinding="urn:oasis:names:tc:SAML:2.0:bindings:HTTP-POST"
              AssertionConsumerServiceURL="https://site.csod.com/samldefault.aspx"
              xmlns="urn:oasis:names:tc:SAML:2.0:protocol"
              >
    <Issuer xmlns="urn:oasis:names:tc:SAML:2.0:assertion">https://site.csod.com</Issuer>
    <Signature xmlns="http://www.w3.org/2000/09/xmldsig#">
        <SignedInfo>
            <CanonicalizationMethod Algorithm="http://www.w3.org/2001/10/xml-exc-c14n#" />
            <SignatureMethod Algorithm="http://www.w3.org/2001/04/xmldsig-more#rsa-sha256" />
            <Reference URI="#_THE_ID">
                <Transforms>
                    <Transform Algorithm="http://www.w3.org/2000/09/xmldsig#enveloped-signature" />
                    <Transform Algorithm="http://www.w3.org/2001/10/xml-exc-c14n#" />
                </Transforms>
                <DigestMethod Algorithm="http://www.w3.org/2001/04/xmlenc#sha256" />
                <DigestValue>DIGEST_VALUE</DigestValue>
            </Reference>
        </SignedInfo>
        <SignatureValue>SIGNATURE_VALUE</SignatureValue>
    </Signature>
    <NameIDPolicy Format="urn:oasis:names:tc:SAML:1.1:nameid-format:unspecified" />
</AuthnRequest>
""")  # noqa
    RELAY_STATE = 'https://site.csod.com/'
    REQUEST_DATA = {
        'SAMLRequest': SAML_REQUEST,
        'RelayState': RELAY_STATE,
    }
    PROCESSOR = 'ubuntu_sso_saml.processors.CornerstoneProcessor'
