# Copyright 2016 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).
import json
import re
import urllib2
import urlparse

from django.test.utils import override_settings

from identityprovider.tests.utils import (
    MockHandler,
    SSOBaseTestCase,
)
from webservices.launchpad import (
    NameAlreadyTaken,
    NoSuchAccount,
    NotPlaceholderAccount,
    SSHKeyAdditionDataError,
    SSHKeyAdditionTypeError,
    SSHKeyCompromisedError,
    add_lp_ssh_key,
    delete_lp_ssh_key,
    get_lp_ssh_keys,
    get_lp_username,
    set_lp_username,
)


@override_settings(
    LP_API_URL='http://api.launchpad.net/test',
    LP_API_CONSUMER_KEY='SSO', LP_API_TOKEN='sso-token',
    LP_API_TOKEN_SECRET='sso-secret')
class LPUsernameAPITestCase(SSOBaseTestCase):

    def setUp(self):
        super(LPUsernameAPITestCase, self).setUp()
        self.lp_api_handler = MockHandler()
        self.mock_lp_api_opener = self.patch(
            'webservices.launchpad._lp_api_opener',
            return_value=urllib2.build_opener(self.lp_api_handler))

    def test_set_lp_username(self):
        self.lp_api_handler.set_next_response(200, '')
        self.assertIsNone(set_lp_username('foo', 'bar', dry_run=True))

        [request] = self.lp_api_handler.requests
        self.assertEqual('POST', request.get_method())
        self.assertEqual(['Authorization'], request.headers.keys())
        self.assertEqual(
            'OAuth oauth_nonce="...", oauth_timestamp="...", '
            'oauth_version="1.0", oauth_signature_method="PLAINTEXT", '
            'oauth_consumer_key="SSO", oauth_token="sso-token", '
            'oauth_signature="%26sso-secret"',
            re.sub(r'"\d+"', '"..."', request.headers['Authorization']))
        self.assertEqual(
            'http://api.launchpad.net/test/people', request.get_full_url())
        self.assertEqual(
            {'ws.op': ['setUsernameFromSSO'], 'openid_identifier': ['foo'],
             'name': ['bar'], 'dry_run': ['true']},
            urlparse.parse_qs(request.data))

    def test_set_lp_username_taken(self):
        self.lp_api_handler.set_next_response(
            400,
            'name: taken-name is already in use by another person or team.')
        self.assertRaises(
            NameAlreadyTaken, set_lp_username, 'foo', 'taken-name')

    def test_set_lp_username_blacklisted(self):
        self.lp_api_handler.set_next_response(
            400,
            'name: The name &#x27;private-name&#x27; has been blocked by the '
            'Launchpad administrators. Contact Launchpad Support if you want '
            'to use this name.')
        self.assertRaises(
            NameAlreadyTaken, set_lp_username, 'foo', 'private-name')

    def test_set_lp_username_existing(self):
        self.lp_api_handler.set_next_response(
            400,
            'An account for that OpenID identifier already exists.')
        self.assertRaises(
            NotPlaceholderAccount, set_lp_username, 'taken-openid',
            'bar')

    def test_get_lp_username(self):
        self.lp_api_handler.set_next_response(200, json.dumps("bar"))
        self.assertEqual("bar", get_lp_username('foo'))

        [request] = self.lp_api_handler.requests
        self.assertEqual('GET', request.get_method())
        self.assertEqual(['Authorization'], request.headers.keys())
        self.assertEqual(
            'OAuth oauth_nonce="...", oauth_timestamp="...", '
            'oauth_version="1.0", oauth_signature_method="PLAINTEXT", '
            'oauth_consumer_key="SSO", oauth_token="sso-token", '
            'oauth_signature="%26sso-secret"',
            re.sub(r'"\d+"', '"..."', request.headers['Authorization']))
        self.assertEqual(
            'http://api.launchpad.net/test/people?'
            'ws.op=getUsernameForSSO&openid_identifier=foo',
            request.get_full_url())


@override_settings(
    LP_API_URL='http://api.launchpad.net/test',
    LP_API_CONSUMER_KEY='SSO', LP_API_TOKEN='sso-token',
    LP_API_TOKEN_SECRET='sso-secret')
class LPSSHKeyAPITestCase(SSOBaseTestCase):

    def setUp(self):
        super(LPSSHKeyAPITestCase, self).setUp()
        self.lp_api_handler = MockHandler()
        self.mock_lp_api_opener = self.patch(
            'webservices.launchpad._lp_api_opener',
            return_value=urllib2.build_opener(self.lp_api_handler))

    def assert_authorization_headers(self, request):
        self.assertEqual(['Authorization'], request.headers.keys())
        self.assertEqual(
            'OAuth oauth_nonce="...", oauth_timestamp="...", '
            'oauth_version="1.0", oauth_signature_method="PLAINTEXT", '
            'oauth_consumer_key="SSO", oauth_token="sso-token", '
            'oauth_signature="%26sso-secret"',
            re.sub(r'"\d+"', '"..."', request.headers['Authorization']))

    def test_get_ssh_keys_nonexistant(self):
        self.lp_api_handler.set_next_response(200, 'null')
        self.assertIsNone(get_lp_ssh_keys('openid'))

        self.assertEqual(1, len(self.lp_api_handler.requests))
        request = self.lp_api_handler.requests[0]
        self.assertEqual('GET', request.get_method())
        self.assert_authorization_headers(request)

        self.assertEqual(
            'http://api.launchpad.net/test/people?'
            'ws.op=getSSHKeysForSSO&openid_identifier=openid',
            request.get_full_url())

    def test_get_ssh_keys_no_keys(self):
        self.lp_api_handler.set_next_response(200, '[]')
        self.assertEqual([], get_lp_ssh_keys('openid'))

    def test_add_ssh_key_with_nonexistant_user(self):
        self.lp_api_handler.set_next_response(
            400, "No account found for openid identifier 'openid'")
        self.assertRaises(
            NoSuchAccount,
            add_lp_ssh_key,
            'openid', 'ssh-rsa keytext comment', False
        )
        [request] = self.lp_api_handler.requests
        self.assertEqual('POST', request.get_method())
        self.assert_authorization_headers(request)

    def test_add_ssh_key_with_bad_key_type(self):
        self.lp_api_handler.set_next_response(
            400, "Invalid SSH key type: 'flub'")
        self.assertRaises(
            SSHKeyAdditionTypeError,
            add_lp_ssh_key,
            'openid', 'flub keytext comment', False
        )
        [request] = self.lp_api_handler.requests
        self.assertEqual('POST', request.get_method())
        self.assert_authorization_headers(request)

    def test_add_ssh_key_with_bad_key_data(self):
        self.lp_api_handler.set_next_response(
            400, "Invalid SSH key data: 'keywithnospaces'")
        self.assertRaises(
            SSHKeyAdditionDataError,
            add_lp_ssh_key,
            'openid', 'keywithnospaces', False
        )
        [request] = self.lp_api_handler.requests
        self.assertEqual('POST', request.get_method())
        self.assert_authorization_headers(request)

    def test_add_ssh_key_with_compromised_key(self):
        self.lp_api_handler.set_next_response(
            400,
            "This key cannot be added as it is known to be compromised.")
        self.assertRaises(
            SSHKeyCompromisedError,
            add_lp_ssh_key,
            'openid', 'compromised_key', False
        )
        [request] = self.lp_api_handler.requests
        self.assertEqual('POST', request.get_method())
        self.assert_authorization_headers(request)

    def test_add_ssh_key_works(self):
        self.lp_api_handler.set_next_response(200, "")
        self.assertIsNone(add_lp_ssh_key('openid', 'key_data', False))
        [request] = self.lp_api_handler.requests
        self.assertEqual('POST', request.get_method())
        self.assert_authorization_headers(request)

    def test_delete_ssh_key_with_nonexistant_user(self):
        self.lp_api_handler.set_next_response(
            400, "No account found for openid identifier 'openid'")
        self.assertRaises(
            NoSuchAccount,
            delete_lp_ssh_key,
            'openid', 'ssh-rsa keytext comment', False
        )
        [request] = self.lp_api_handler.requests
        self.assertEqual('POST', request.get_method())
        self.assert_authorization_headers(request)

    def test_delete_ssh_key_with_bad_key_type(self):
        self.lp_api_handler.set_next_response(
            400, "Invalid SSH key type: 'flub'")
        self.assertRaises(
            SSHKeyAdditionTypeError,
            delete_lp_ssh_key,
            'openid', 'flub keytext comment', False
        )
        [request] = self.lp_api_handler.requests
        self.assertEqual('POST', request.get_method())
        self.assert_authorization_headers(request)

    def test_delete_ssh_key_with_bad_key_data(self):
        self.lp_api_handler.set_next_response(
            400, "Invalid SSH key data: 'keywithnospaces'")
        self.assertRaises(
            SSHKeyAdditionDataError,
            delete_lp_ssh_key,
            'openid', 'keywithnospaces', False
        )
        [request] = self.lp_api_handler.requests
        self.assertEqual('POST', request.get_method())
        self.assert_authorization_headers(request)

    def test_delete_ssh_key_works(self):
        self.lp_api_handler.set_next_response(200, "")
        self.assertIsNone(delete_lp_ssh_key('openid', 'key_data', False))
        [request] = self.lp_api_handler.requests
        self.assertEqual('POST', request.get_method())
        self.assert_authorization_headers(request)
