# Copyright 2016 Canonical Ltd.  This software is licensed under
# the GNU Affero General Public License version 3 (see the file LICENSE).

from django import forms
from django.utils.html import escape, format_html
from django.utils.translation import ugettext_lazy as _
from gpgservice_client import (
    GPGServiceException,
    sanitize_fingerprint,
)

from identityprovider.models import AuthToken
from identityprovider.models.const import AuthTokenType
from webservices.launchpad import (
    LaunchpadAPIError,
    add_lp_ssh_key,
    delete_lp_ssh_key,
    get_lp_ssh_keys,
)
from webui.gpg import (
    SSOGPGClient,
    compare_gpg_fingerprints,
    get_user_id_for_account,
    validation_phrase_for_authtoken,
)
from webui.views import gpg_messages


class ClaimOpenPGPKeyForm(forms.Form):

    """A form to validate the claiming of an OpenPGP key."""

    placeholder_text = _('Example: {fingerprint}').format(
        fingerprint='27E0 7815 B47C 0397 90D5  8589 27D9 A27B F3F9 6058')
    fingerprint = forms.CharField(
        widget=forms.TextInput(
            attrs={'placeholder': placeholder_text, 'class': 'input-row'}),
        error_messages={'required': gpg_messages.bad_fingerprint_format()},
    )

    def clean_fingerprint(self):
        client = SSOGPGClient()
        fingerprint = self.cleaned_data['fingerprint']

        sane_fingerprint = sanitize_fingerprint(fingerprint)
        if sane_fingerprint is None:
            raise forms.ValidationError(gpg_messages.bad_fingerprint_format())
        if client.getKeyByFingerprint(sane_fingerprint) is not None:
            raise forms.ValidationError(
                gpg_messages.key_already_imported(fingerprint))

        key_details = client.getKeyDetailsFromKeyServer(
            sane_fingerprint)
        if key_details is None:
            raise forms.ValidationError(gpg_messages.unknown_key())
        if key_details['expired']:
            raise forms.ValidationError(gpg_messages.key_expired(fingerprint))
        if key_details['revoked']:
            raise forms.ValidationError(gpg_messages.key_revoked(fingerprint))
        # The view needs a copy of key_details. To save ourselves an expensive
        # round-trip, we expose it here:
        self.cleaned_data['key_details'] = key_details
        return sane_fingerprint


class VerifySignedTextForm(forms.Form):

    clearsigned = forms.CharField(
        label=_('Clearsigned Text'),
        widget=forms.Textarea(),
        error_messages={'required': gpg_messages.missing_signed_text()}
    )

    def __init__(self, data=None, token=None):
        """Validate form based on 'data'.

        'token' should be the authtoken we're claiming.
        """
        if data and not token:
            raise ValueError("'token' must be supplied with 'data'")
        super(VerifySignedTextForm, self).__init__(data)
        self._token = token

    def clean_clearsigned(self):
        signed_text = self.cleaned_data['clearsigned'].strip()
        if not signed_text:
            raise forms.ValidationError(gpg_messages.missing_signed_text())

        gpg_client = SSOGPGClient()
        try:
            fingerprint, content = gpg_client.verifySignedContent(signed_text)
            token = self._token
            if not compare_gpg_fingerprints(fingerprint, token.fingerprint):
                raise forms.ValidationError(
                    gpg_messages.signing_key_mismatch(fingerprint))

            validation_phrase = validation_phrase_for_authtoken(token)
            if content.split() != validation_phrase.split():
                raise forms.ValidationError(
                    gpg_messages.validation_phrase_mismatch())
        except GPGServiceException as e:
            raise forms.ValidationError(str(e))
        return signed_text


class PendingGPGTokenField(forms.ModelMultipleChoiceField):
    """A ModelMultipleChoiceField for displaying AuthTokens for GPG keys."""

    def label_from_instance(self, token):
        """Decide how to display 'token'.

        Tokens for GPG keys should display the keyid for the key in its
        display format.
        """
        return gpg_messages.display_keyid_long(
            token.displayname, token.fingerprint)


class PendingGPGValidationTokenForm(forms.Form):
    tokens = PendingGPGTokenField(
        required=True,
        widget=forms.CheckboxSelectMultiple,
        label=(''),
        queryset=AuthToken.objects.none()
    )

    def __init__(self, account, *args, **kwargs):
        """Form for pending GPG key validation tokens.

        :param account: The account to display tokens for. Should be the
            currently logged in account.
        """
        super(PendingGPGValidationTokenForm, self).__init__(*args, **kwargs)
        pending_tokens = AuthToken.objects.filter(
            requester=account,
            token_type__in=[
                AuthTokenType.VALIDATEGPG,
                AuthTokenType.VALIDATESIGNONLYGPG,
            ],
            date_consumed=None
        ).order_by('fingerprint')
        self.fields['tokens'].queryset = pending_tokens


class _GPGKeysForm(forms.Form):

    """A form that lists GPG keys for a particular user.

    This form talks to the gpgservice to display keys that belong to the user.
    It must be instantiated with two parameters: the account to display keys
    for, and whether to display enabled or disabled keys. An example
    instantiation from a view might look like this for an unbound form::

        enabled_keys_form = _GPGKeysForm(request.account, True)

    ...and like this for a bound form::

        enabled_keys_form = _GPGKeysForm(request.account, True, request.POST)

    Keys are identified by their fingerprint, which is also what the gpgservice
    client uses. The form does not alter keys in any way.

    Instead of using this class directly, use EnabledGPGKeysForm and
    DisabledGPGKeysForm instead.
    """

    keys = forms.MultipleChoiceField(
        required=True,
        widget=forms.CheckboxSelectMultiple,
        label=('')
    )

    def __init__(self, account, enabled=True, *args, **kwargs):
        id_string = "id_%s_%%s" % ("enabled" if enabled else "disabled")
        super(_GPGKeysForm, self).__init__(*args, auto_id=id_string, **kwargs)
        user_id = get_user_id_for_account(account)
        client = SSOGPGClient()
        self.keys = client.getKeysForOwner(user_id)['keys']
        self.enabled = enabled
        self.fields['keys'].choices = self._get_user_active_keys

    def _get_user_active_keys(self):
        for key in self.keys:
            if key['enabled'] == self.enabled:
                display = "{size}{algorithm}/{id}".format(**key)
                if not key['can_encrypt']:
                    display = _('{keyid} (sign only)').format(keyid=display)
                yield (key['fingerprint'], display)

    def num_choices(self):
        # Since the 'choices' attribute is an iterator, templates can't use it
        # to determine whether there are keys to render or not, since the
        # iterator object is always truthy. This function is used in the
        # template instead.
        return len(list(self.fields['keys'].choices))


class EnabledGPGKeysForm(_GPGKeysForm):

    def __init__(self, account, *args, **kwargs):
        super(EnabledGPGKeysForm, self).__init__(
            account, enabled=True, *args, **kwargs)


class DisabledGPGKeysForm(_GPGKeysForm):

    def __init__(self, account, *args, **kwargs):
        super(DisabledGPGKeysForm, self).__init__(
            account, enabled=False, *args, **kwargs)


class SSHKeyField(forms.CharField):

    def __init__(self, *args, **kwargs):
        placeholder = kwargs.pop('placeholder')
        attrs = {
            'placeholder': placeholder,
        }
        if kwargs.pop('disabled', False):
            attrs['disabled'] = 'disabled'

        kwargs['widget'] = forms.Textarea(attrs=attrs)
        super(SSHKeyField, self).__init__(*args, **kwargs)


class ClaimSSHKeyForm(forms.Form):

    def __init__(self, account, *args, **kwargs):
        super(ClaimSSHKeyForm, self).__init__(*args, **kwargs)
        self.account = account
        if not self.account.is_verified:
            disabled = True
            placeholder = _(
                "You must verify your account and set a username first.")
        elif self.account.person_name is None:
            disabled = True
            placeholder = _("You must set a username first.")
        else:
            disabled = False
            placeholder = _('SSH Key Text.')
        self.fields['ssh_key'] = SSHKeyField(
            required=True,
            disabled=disabled,
            placeholder=placeholder,
        )

    def clean_ssh_key(self):
        keytext = self.cleaned_data['ssh_key'].strip()
        try:
            # Ask launchpad for a dry_run so nothing is actually changed:
            add_lp_ssh_key(self.account.openid_identifier, keytext,
                           dry_run=True)
        except LaunchpadAPIError as e:
            raise forms.ValidationError(e.message)
        return keytext

    def add_key_to_launchpad(self):
        keytext = self.cleaned_data['ssh_key']
        add_lp_ssh_key(self.account.openid_identifier, keytext, dry_run=False)


def format_ssh_key(ssh_key):
    """Decide how to display 'ssh_key'.

    SSH keys are freeform text, so users absolutely *will* try and add keys
    that contain javascript, so we need to make sure they're escaped when
    being displayed to the user.
    """
    try:
        key_type, key_text, comment = ssh_key.split(' ', 2)
    except ValueError:
        # This should never happen, as launchpad is supposed to guarantee a
        # valid ssh key format. However, we need to exercise this code path
        # in tests to ensure that users malicously POST'ing bad key formats
        # will get a ValidationError.
        # Just return the escaped data - this will never be run in production:
        return format_html(
            '<label>Bad Key Data: <span>{key}</span></label>',
            key=escape(ssh_key))
    else:
        return format_html(
            '<div>'
            '  <label>{key_type_label}<span>{key_type}</span></label> '
            '  <label>{key_text_label}<span>{key_text}</span></label> '
            '  <label>{comment_label}<span>{comment}</span></label> '
            '</div>',
            key_type_label=_("Key Type: "),
            key_type=key_type,
            key_text_label=_("Key Text: "),
            key_text=key_text,
            comment_label=_("Comment: "),
            comment=comment,
        )


class DeleteSSHKeyForm(forms.Form):

    ssh_keys = forms.MultipleChoiceField(
        required=True,
        widget=forms.CheckboxSelectMultiple,
    )

    def __init__(self, account, *args, **kwargs):
        super(DeleteSSHKeyForm, self).__init__(*args, **kwargs)
        self.account = account
        keys = get_lp_ssh_keys(self.account.openid_identifier)
        choices = [(key, format_ssh_key(key)) for key in keys]
        self.fields['ssh_keys'].choices = choices

    def delete_keys_from_launchpad(self):
        for key in self.cleaned_data['ssh_keys']:
            delete_lp_ssh_key(self.account.openid_identifier, key,
                              dry_run=False)
