# Copyright 2016 Canonical Ltd.  This software is licensed under
# the GNU Affero General Public License version 3 (see the file LICENSE).

"""SSO GPGService Client Implementation."""

from urlparse import urljoin

from django.conf import settings
from gpgservice_client import GPGClient

from identityprovider.models.const import AuthTokenType


class SSOGPGClient(GPGClient):

    def __init__(self):
        try:
            super(SSOGPGClient, self).__init__(
                settings.GPGSERVICE_ENDPOINT, settings.GPGSERVICE_TIMEOUT)
        except TypeError:
            raise RuntimeError(
                "Error: GPG functionality is not available unless the "
                "'GPGSERVICE_ENDPOINT' setting has been set to the URL of "
                "a running gpgservice instance. See README for more details."
            )


def get_user_id_for_account(account):
    """Build a user-id key for use with gpgservice for 'account'.

    OpenPGP keys stored in gpgservice are stored per 'user-id'. This is treated
    as a free-text field by gpgservice, but launchpad and SSO use the user's
    openid url. This is in the form:

    https://login.ubuntu.com/+id/XXXXXX

    Since this string is used as the key for storing and retrieving GPG keys,
    it is critical that the string is consistent for the same Person/Account.
    This function ensures that the string returned is consistent regardless
    of whether the user is accessing login.ubuntu.com or login.launchpad.net.
    """
    return urljoin(settings.SSO_ROOT_URL, account.get_absolute_url())


def compare_gpg_fingerprints(lhs, rhs):
    """Compare two GPG tokens for equality.

    Returns True if they are equal, false otherwise.
    """
    return lhs.replace(' ', '').lower() == rhs.replace(' ', '').lower()


def validation_phrase_for_authtoken(token):
    """Generate a validation phrase for a sign-only authtoken.

    The generated string MUST NOT BE translated, to prevent malicous users
    submitting a translation that happens to match a publicly available signed
    text from the target key.

    :raises ValueError: if 'token' is not a token for a sign-only GPG key.
    """
    if token.token_type != AuthTokenType.VALIDATESIGNONLYGPG:
        raise ValueError(
            "Token must be of type VALIDATESIGNONLYGPG, not %r"
            % token.token_type)
    return (
        'Please register %s to the\n'
        'Ubuntu One user with the email address %s.\n'
        '%s') % (
        token.fingerprint,
        token.requester.preferredemail,
        token.date_created.strftime('%Y-%m-%d %H:%M:%S'))
