# Copyright 2016 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

import os
import subprocess
import sys

from django.test import override_settings
from fixtures import (
    Fixture,
    EnvironmentVariable,
)
from gpgservice_client.testing import (
    GPGKeyServiceFixture,
    TestKeyServerFixture,
)


def _find_twistd_path():
    """Get the path to the 'twistd' executable.

    This is necessary since SSO uses --system-site-packages when creating the
    virtualenv, and the system tests are run on as twisted installed site-wide
    so the twisted package is never installed in the virtualenv.

    This function looks in the current active virtualenv first, then in the
    site-wide PATH.
    """
    twistd_path = os.path.join(
        os.path.dirname(sys.executable), 'twistd')
    if not os.path.exists(twistd_path):
        twistd_path = subprocess.check_output(
            ['which', 'twistd']).decode().strip()
        if not os.path.exists(twistd_path):
            raise RuntimeError(
                "Could not find twistd executable in virtualenv or system "
                "$PATH. Twisted must be installed for this fixture to work.")
    return twistd_path


class SSOGPGServiceFixture(Fixture):

    # In the case where twistd is installed globally this somewhat expensive
    # (see _find_twistd_path definition above), so we call it once and store
    # it as a cvar. The error case where twistd isn't installed anywhere should
    # never happen, since it's a dependency of SSO.
    twistd_path = _find_twistd_path()

    """A fixture to set up and run a test key server and the gpgservice.

    The test keyserver is shipped as part of the gpgservice_client package. It
    ships with several sample keys that can be used in tests. Details of these
    keys can be found in:

        gpgservice_client.testing.keyserver.tests.keys.SampleDataKeys

    Using this fixture is as simple as calling::

        self.use_fixture(SSOGPGServiceFixture())

    ...in your test. For the duration of that test the 'GPGSERVICE_ENDPOINT'
    setting will be set appropriately, and webui.gpg.SSOGPGClient instances
    will talk to the running gpgservice instance.

    The services are cleaned up at the end of the test case, thereby
    maintaining test isolation.
    """

    def setUp(self):
        super(SSOGPGServiceFixture, self).setUp()

        self.useFixture(EnvironmentVariable('TWISTD_PATH', self.twistd_path))
        self.keyserver = self.useFixture(TestKeyServerFixture())
        self.gpgservice = self.useFixture(
            GPGKeyServiceFixture(
                self.keyserver.bind_host, self.keyserver.daemon_port))
        new_settings = override_settings(
            GPGSERVICE_ENDPOINT=self.gpgservice.root_url)
        new_settings.enable()
        self.addCleanup(new_settings.disable)
