from __future__ import unicode_literals

from datetime import timedelta

from django.conf import settings
from django.contrib.auth.models import AnonymousUser
from django.core.urlresolvers import reverse
from django.http import (
    Http404,
    HttpResponseRedirect,
)
from django.utils.timezone import now
from gargoyle.testutils import switches

from mock import ANY, Mock, patch

from identityprovider.models import InvalidatedEmailAddress, twofactor
from identityprovider.models.const import EmailStatus
from identityprovider.tests.utils import SSOBaseTestCase

from webui.decorators import (
    EMAIL_INVALIDATED,
    ratelimit,
    redirect_home_if_logged_in,
    require_twofactor_authenticated,
    require_twofactor_enabled,
    sso_login_required,
)


@sso_login_required(require_twofactor=True,
                    require_twofactor_freshness=True)
def view(request, *args, **kwargs):
    return 'SUCCESS'


class RequireTwofactorEnabledTestCase(SSOBaseTestCase):

    def setUp(self):
        super(RequireTwofactorEnabledTestCase, self).setUp()

        @require_twofactor_enabled
        def view(request):
            return 'SUCCESS'
        self.view = view

    def test_decorator_with_twofactor_enabled(self):
        request = self.factory.make_request(session=True)
        name = 'webui.decorators.twofactor.is_twofactor_enabled'
        with patch(name) as mock_enabled:
            mock_enabled.return_value = True
            response = self.view(request)
        self.assertEqual(response, 'SUCCESS')

    def test_decorator_with_twofactor_disabled(self):
        request = self.factory.make_request(session=True)
        name = 'webui.decorators.twofactor.is_twofactor_enabled'
        with patch(name) as mock_enabled:
            mock_enabled.return_value = False
            with self.assertRaises(Http404):
                self.view(request)


class RatelimitTestCase(SSOBaseTestCase):

    @patch('webui.decorators.now')
    def test_keys_to_check_use_utc(self, mock_now):
        mock_now.return_value = now()

        request = self.factory.make_request(session=True)
        limiter = ratelimit()
        with patch.object(limiter, 'key_extra') as mock_key_extra:
            mock_key_extra.return_value = None
            limiter.keys_to_check(request)

        self.assertEqual(mock_now.called, True)

    @patch('webui.decorators.now')
    def test_current_key_use_utc(self, mock_now):
        mock_now.return_value = now()

        request = self.factory.make_request(session=True)
        limiter = ratelimit()
        with patch.object(limiter, 'key_extra') as mock_key_extra:
            mock_key_extra.return_value = None
            limiter.current_key(request)

        self.assertEqual(mock_now.called, True)

    def test_disallowed_logging(self):
        limiter = ratelimit()
        limiter.remote_ip = lambda req: 33

        user1 = Mock(spec=['openid_identifier'])
        user2 = Mock(spec=['username'])
        request = self.factory.make_request(session=True)

        user1.openid_identifier = user2.username = 'Username'
        user1.id = user2.id = 99

        with patch('webui.decorators.logger') as mock_logger:
            for user in user1, user2:
                mock_logger.warning.reset_mock()
                request.user = user
                limiter.disallowed(request)
                mock_logger.warning.assert_called_once_with(
                    ANY, 'Username', 33, 99)


class SSOLoginRequiredBaseTestCase(SSOBaseTestCase):

    def setUp(self):
        super(SSOLoginRequiredBaseTestCase, self).setUp()
        self.calls = []

        @sso_login_required
        def view(request, *args, **kwargs):
            self.calls.append((request, args, kwargs))
            return "SUCCESS"

        self.view = view

    def assert_redirects_to_twofactor(self, response, next_url='/'):
        self.assertEqual(response.status_code, 302)
        self.assertEqual(
            response['Location'], reverse('twofactor') + '?next=' + next_url)

    def assert_view_called_once(self, response, request=None):
        self.assertEqual(response, 'SUCCESS')
        if request is None:
            request = self.request
        # the decorated view was called once
        self.assertEqual(self.calls, [(request, (), {})])
        self.calls = []  # reset


class SSOLoginRequiredTestCase(SSOLoginRequiredBaseTestCase):

    def mock_request(self, authed=True, twofactor_login=None):
        if authed:
            user = self.factory.make_account()
        else:
            user = AnonymousUser()

        request = self.factory.make_request(
            path='/target', user=user, session=True, messages=True)
        if twofactor_login is not None:
            request.session[twofactor.TWOFACTOR_LOGIN] = twofactor_login

        return request

    def assert_redirects_to_twofactor(self, response, next_url='/target'):
        super(SSOLoginRequiredTestCase, self).assert_redirects_to_twofactor(
            response, next_url)

    def test_django_login_check_still_works(self):
        request = self.mock_request(False)
        response = self.view(request)

        self.assertEqual(response.status_code, 302)
        self.assertEqual(response['Location'], '/+login?next=/target')

    @switches(TWOFACTOR=True)
    def test_allowed_2f_not_required(self):
        request = self.mock_request()
        assert not twofactor.user_requires_twofactor_auth(
            request, request.user)

        response = self.view(request)
        self.assert_view_called_once(response, request)

    @switches(TWOFACTOR=True)
    def test_allowed_when_2f_required_by_user_and_2f_authed(self):
        request = self.mock_request(twofactor_login=now())
        request.user.twofactor_required = True

        assert twofactor.user_requires_twofactor_auth(request, request.user)
        assert twofactor.is_upgraded(request)

        response = self.view(request)
        self.assert_view_called_once(response, request)

    @switches(TWOFACTOR=True)
    def test_allowed_when_2f_required_by_keyword_and_2f_authed(self):
        request = self.mock_request(twofactor_login=now())

        assert not twofactor.user_requires_twofactor_auth(
            request, request.user)
        assert twofactor.is_upgraded(request)

        @sso_login_required(require_twofactor=True)
        def view(request, *args, **kwargs):
            return "SUCCESS"

        response = view(request)
        self.assertEqual(response, 'SUCCESS')

    @switches(TWOFACTOR=True)
    def test_redirected_when_2f_required_by_user_and_not_2f_authed(self):
        request = self.mock_request(
            twofactor_login=now() - timedelta(seconds=settings.TWOFACTOR_TTL))
        request.user.twofactor_required = True

        assert twofactor.user_requires_twofactor_auth(request, request.user)
        assert not twofactor.is_upgraded(request)

        response = self.view(request)
        self.assert_redirects_to_twofactor(response)

    @switches(TWOFACTOR=True)
    def test_redirected_when_2f_required_by_keyword_and_not_2f_authed(self):
        request = self.mock_request(
            twofactor_login=now() - timedelta(seconds=settings.TWOFACTOR_TTL))

        assert not twofactor.user_requires_twofactor_auth(
            request, request.user)
        assert not twofactor.is_upgraded(request)

        @sso_login_required(require_twofactor=True)
        def view(request, *args, **kwargs):
            return "SUCCESS"

        response = view(request)
        self.assert_redirects_to_twofactor(response)

    @switches(TWOFACTOR=False)
    def test_decoration_no_params(self):
        # Needs two-factor authentication to be disabled so we get 'SUCCESS'
        # rather than a redirect.  Because of
        # https://github.com/disqus/gargoyle/issues/57, we cannot rely on the
        # switch being in any particular state, so we use a decorator.
        request = self.mock_request()
        response = self.view(request)
        # if this works, the decorator handles this invocation properly
        self.assert_view_called_once(response, request)

    @switches(TWOFACTOR=False)
    def test_decoration_with_params(self):
        # Needs two-factor authentication to be disabled so we get 'SUCCESS'
        # rather than a redirect.  Because of
        # https://github.com/disqus/gargoyle/issues/57, we cannot rely on the
        # switch being in any particular state, so we use a decorator.

        @sso_login_required(login_url='/alt_login')
        def view(request, *args, **kwargs):
            return "SUCCESS"

        response = view(self.mock_request())
        # if this works, the decorator handles this invocation properly
        self.assertEqual(response, 'SUCCESS')

    @switches(TWOFACTOR=True)
    def test_require_twofactor_freshness_when_no_twofactor(self):
        request = self.mock_request(twofactor_login=now())
        request.user.twofactor_required = True

        assert twofactor.user_requires_twofactor_auth(request, request.user)
        assert twofactor.is_upgraded(request, must_be_fresh=True)

        @sso_login_required(require_twofactor_freshness=True)
        def view(request, *args, **kwargs):
            return "SUCCESS"

        response = view(self.mock_request())
        self.assertEqual(response, 'SUCCESS')

    @switches(TWOFACTOR=True)
    def test_require_twofactor_freshness_when_twofactor_old(self):
        # fake an old 2f session
        request = self.mock_request(
            twofactor_login=now() - timedelta(seconds=settings.TWOFACTOR_TTL))
        request.user.twofactor_required = True

        assert twofactor.user_requires_twofactor_auth(request, request.user)
        assert not twofactor.is_upgraded(request, must_be_fresh=True)

        @sso_login_required(require_twofactor_freshness=True)
        def view(request, *args, **kwargs):
            return "SUCCESS"

        response = view(request)
        self.assert_redirects_to_twofactor(response)

    @switches(TWOFACTOR=True)
    def test_require_twofactor_freshness_when_twofactor_non_fresh(self):
        delta = timedelta(seconds=settings.TWOFACTOR_FRESHNESS)
        request = self.mock_request(twofactor_login=now() - delta)
        request.user.twofactor_required = True

        assert twofactor.user_requires_twofactor_auth(request, request.user)
        assert not twofactor.is_upgraded(request, must_be_fresh=True)

        @sso_login_required(require_twofactor_freshness=True)
        def view(request, *args, **kwargs):
            return "SUCCESS"

        response = view(request)
        self.assert_redirects_to_twofactor(response)

    def test_require_twofactor_freshness_when_twofactor_fresh(self):
        request = self.mock_request(twofactor_login=now())
        assert twofactor.is_upgraded(request, must_be_fresh=True)

        @sso_login_required(require_twofactor_freshness=True)
        def view(request, *args, **kwargs):
            return "SUCCESS"

        response = view(request)
        self.assertEqual(response, 'SUCCESS')

    def test_account_must_require_two_factor(self):
        request = self.mock_request()
        team = self.factory.make_team()
        self.factory.add_account_to_team(request.user, team)

        with self.settings(TWOFACTOR_MANDATORY_TEAMS=[team.name]):
            response = self.view(request)

        self.assert_redirects_to_twofactor(response)


class SSOLoginRequiredInvalidatedEmailsTestCase(SSOLoginRequiredBaseTestCase):

    def setUp(self):
        super(SSOLoginRequiredInvalidatedEmailsTestCase, self).setUp()
        self.account = self.factory.make_account()
        for email in self.account.emailaddress_set.all():
            email.invalidate()

        assert self.account.emailaddress_set.count() == 0
        assert self.account.invalidatedemailaddress_set.count() > 0

        self.request = self.factory.make_request(
            session=True, user=self.account)

    def test_correct_template(self):
        response = self.view(self.request)
        self.assertEqual(
            response.template_name,
            'account/user_logged_out_no_valid_emails.html')

    def test_explanation_shown(self):
        response = self.view(self.request)
        response.render()
        msgs = (
            'You have no valid email addresses linked to your account',
            'To reactivate the account please contact our support team',
            settings.SUPPORT_FORM_URL,
        )
        for msg in msgs:
            self.assertIn(msg, response.content)
        self.assertNotIn(self.account.displayname, response.content)

    def test_logged_out(self):
        self.view(self.request)

        # the decorated view was not called
        self.assertEqual(self.calls, [])
        # user was logged out
        self.assertIsInstance(self.request.user, AnonymousUser)
        self.assertFalse(self.request.user.is_authenticated())


class SSOLoginRequiredNoEmailsTestCase(
        SSOLoginRequiredInvalidatedEmailsTestCase):

    def setUp(self):
        super(SSOLoginRequiredNoEmailsTestCase, self).setUp()
        self.account.emailaddress_set.all().delete()
        self.account.invalidatedemailaddress_set.all().delete()


class SSOLoginRequiredInvalidatedEmailsWarningTestCase(
        SSOLoginRequiredBaseTestCase):

    def setUp(self):
        super(SSOLoginRequiredInvalidatedEmailsWarningTestCase, self).setUp()
        self.account = self.factory.make_account()
        self.invalid = self.factory.make_email_for_account(
            self.account, status=EmailStatus.NEW).invalidate()
        assert not self.invalid.account_notified

        assert self.account.emailaddress_set.count() == 1
        assert self.account.invalidatedemailaddress_set.count() == 1

        self.request = self.factory.make_request(
            session=True, user=self.account)

        # patch messages framework to track messages
        self.mock_messages = self.patch('webui.decorators.messages')

    def assert_warning_shown_for_email(self, response, email):
        self.assert_view_called_once(response)
        # email is marked as notified
        invalid = InvalidatedEmailAddress.objects.get(id=email.id)
        self.assertTrue(invalid.account_notified)
        # message was shown
        self.mock_messages.warning.assert_called_with(
            self.request, EMAIL_INVALIDATED.format(email=invalid))

    def test_warning_shown(self):
        response = self.view(self.request)
        self.assert_warning_shown_for_email(response, email=self.invalid)

    def test_two_invalidated_emails(self):
        other = self.factory.make_email_for_account(
            self.account, status=EmailStatus.NEW).invalidate()
        assert self.account.invalidatedemailaddress_set.count() == 2

        # call the view for the first time
        response = self.view(self.request)

        # view was called and warning was shown for the older email
        self.assert_warning_shown_for_email(response, email=self.invalid)
        # the other email was unchanged
        unchanged = InvalidatedEmailAddress.objects.get(id=other.id)
        self.assertFalse(unchanged.account_notified)
        self.assertEqual(other, unchanged)

        # call the view again
        response = self.view(self.request)

        # the 'other' email has to be processed as a warning
        self.assert_warning_shown_for_email(response, email=other)

    def test_no_warning_shown_when_no_email(self):
        self.account.invalidatedemailaddress_set.all().delete()

        response = self.view(self.request)

        self.assert_view_called_once(response)
        # message was not shown
        self.assertFalse(self.mock_messages.warning.called)

    def test_warning_shown_only_if_account_not_notified_before(self):
        self.account.invalidatedemailaddress_set.update(account_notified=True)

        response = self.view(self.request)

        self.assert_view_called_once(response)
        # email is unchanged
        invalid = InvalidatedEmailAddress.objects.get(id=self.invalid.id)
        self.assertEqual(invalid, self.invalid)
        # message was not shown
        self.assertFalse(self.mock_messages.warning.called)


class SSOLoginRequiredDeviceWarningTestCase(SSOLoginRequiredBaseTestCase):

    def setUp(self):
        super(SSOLoginRequiredDeviceWarningTestCase, self).setUp()
        self.account = self.factory.make_account(twofactor_required=True)
        self.request = self.factory.make_request(
            session=True, messages=True, user=self.account)

        p = switches(TWOFACTOR=True)
        p.patch()
        self.addCleanup(p.unpatch)

    def assert_messages(self, expected):
        actual = [unicode(m.message) for m in self.request._messages]
        self.assertEqual(actual, expected)

    def test_no_warning_messages_no_twofactor_required(self):
        self.account.twofactor_required = False
        response = self.view(self.request)

        self.assert_view_called_once(response)
        self.assert_messages([])

    def test_no_warning_messages_twofactor_valid(self):
        assert self.account.twofactor_required
        twofactor.login(self.request)

        response = self.view(self.request)

        self.assert_view_called_once(response)
        self.assert_messages([])

    def test_no_warning_messages_twofactor_not_valid_many_devices(self):
        assert self.account.twofactor_required
        self.factory.make_device(account=self.account)

        response = self.view(self.request)

        self.assert_redirects_to_twofactor(response)
        self.assert_messages([])

    def test_no_devices_and_adding_a_device(self):
        assert self.account.twofactor_required
        assert not self.account.has_twofactor_devices()

        self.request.path = reverse('device-addition')

        response = self.view(self.request)

        # twofactor is skipped just this once to allow for a device to be added
        self.assert_view_called_once(response)
        self.assert_messages([])


class RedirectHomeIfLoggedInTestCase(SSOBaseTestCase):

    def setUp(self):
        super(RedirectHomeIfLoggedInTestCase, self).setUp()
        self.request = self.factory.make_request(session=True)

        @redirect_home_if_logged_in
        def view(request):
            return 'SUCCESS'
        self.view = view

    def test_decorator_with_logged_in_user(self):
        account = self.factory.make_account()
        self.request.user = account
        response = self.view(self.request)
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response['Location'], reverse('account-index'))

    def test_decorator_with_anonymous_user(self):
        self.request.user = AnonymousUser()
        response = self.view(self.request)
        self.assertEqual(response, 'SUCCESS')


class TwoFactorAuthenticationDecoratorTestCase(SSOBaseTestCase):

    def setUp(self):
        super(TwoFactorAuthenticationDecoratorTestCase, self).setUp()

        @require_twofactor_authenticated()
        def view(request):
            return 'SUCCESS'
        self.view = view

    def set_twofactor_authenticated(self, is_authenticated):
        mock = self.patch('webui.decorators.twofactor.is_authenticated')
        mock.return_value = is_authenticated

    def patch_django_messages_framework(self):
        """Patch the django messages framework.

        This enables the assert_warning_message_string call below.

        This is required since we're calling the view directly, rather than
        doing a full-stack test, so the messages middleware isn't involved.
        """
        self._message_mock = self.patch('webui.decorators.messages')

    def assert_warning_message_string(self, expected_string):
        """Assert that a single warning was created with 'expected_string'.

        This is required since the messages framework is typically called with
        a ugettext proxy and a request. This method removes some of the
        complexities involved in making the assertion.

        To use this assertion method, 'patch_django_messages_framework' must
        be called first. An AssertionError will be raised otherwise.
        """
        mock = getattr(self, '_message_mock', None)
        assert mock is not None, (
            "Error: You must call patch_django_messages_framework before "
            "using assert_warning_message_string."
        )
        self.assertEqual(1, mock.warning.call_count)
        _, observed = mock.warning.call_args[0]
        self.assertEqual(expected_string, observed)

    def test_decorator_with_twofactor_enabled(self):
        request = self.factory.make_request(session=True)
        self.set_twofactor_authenticated(True)

        response = self.view(request)
        self.assertEqual('SUCCESS', response)

    def test_decorator_with_twofactor_disabled_and_default_message(self):
        request = self.factory.make_request(session=True)
        self.set_twofactor_authenticated(False)
        self.patch_django_messages_framework()

        resp = self.view(request)

        self.assertIsInstance(resp, HttpResponseRedirect)
        self.assert_warning_message_string(
            "You must log in to access this page.")

    def test_decorator_with_twofactor_disabled_and_nondefault_message(self):
        @require_twofactor_authenticated("Log in! Use your Yubikey!")
        def view(request):
            return 'SUCCESS'
        self.view = view

        request = self.factory.make_request(session=True)
        self.set_twofactor_authenticated(False)
        self.patch_django_messages_framework()

        resp = self.view(request)

        self.assertIsInstance(resp, HttpResponseRedirect)
        self.assert_warning_message_string("Log in! Use your Yubikey!")
