# Copyright 2016 Canonical Ltd.  This software is licensed under
# the GNU Affero General Public License version 3 (see the file LICENSE).

from __future__ import unicode_literals

from datetime import datetime
from textwrap import dedent

from gpgservice_client import sanitize_fingerprint
from gpgservice_client.testing.keyserver.tests.keys import SampleDataKeys

from identityprovider.models.const import AuthTokenType
from identityprovider.tests.utils import SSOBaseTestCase
from webservices.launchpad import (
    NoSuchAccount,
    SSHKeyAdditionDataError,
    SSHKeyAdditionTypeError,
    SSHKeyCompromisedError,
)
from webui.forms import (
    ClaimOpenPGPKeyForm,
    ClaimSSHKeyForm,
    DeleteSSHKeyForm,
    DisabledGPGKeysForm,
    EnabledGPGKeysForm,
    PendingGPGValidationTokenForm,
    VerifySignedTextForm,
    format_ssh_key,
)
from webui.tests.gpg_fixture import SSOGPGServiceFixture
from webui.views import gpg_messages


class ClaimOpenPGPKeyFormTestCase(SSOBaseTestCase):

    def setUp(self):
        super(ClaimOpenPGPKeyFormTestCase, self).setUp()
        self.use_fixture(SSOGPGServiceFixture())
        self.account = self.factory.make_account()

    def assert_fingerprint_field_has_error(self, form, error):
        self.assertIn('fingerprint', form.errors)
        self.assertIn(error, form.errors['fingerprint'])

    def test_unbound_form_has_fingerprint_field(self):
        form = ClaimOpenPGPKeyForm()

        self.assertIn('fingerprint', form.fields)

    def test_errors_on_empty_fingerprint(self):
        form = ClaimOpenPGPKeyForm({'fingerprint': ''})

        self.assert_fingerprint_field_has_error(
            form, gpg_messages.bad_fingerprint_format())

    def test_errors_on_invalid_fingerprint(self):
        form = ClaimOpenPGPKeyForm({'fingerprint': 'Invalid fingerprint'})

        self.assert_fingerprint_field_has_error(
            form, gpg_messages.bad_fingerprint_format())

    def test_errors_on_already_imported_key(self):
        fingerprint = SampleDataKeys.regular_key
        self.factory.add_gpgkey_to_account(self.account, fingerprint)
        form = ClaimOpenPGPKeyForm({'fingerprint': fingerprint})

        self.assert_fingerprint_field_has_error(
            form, gpg_messages.key_already_imported(fingerprint))

    def test_errors_on_unknown_key(self):
        # Thomi's fingerprint is unknown to the test key server:
        fingerprint = '9C25 0773 25C2 161B 7CE3  C35B D078 66BB 82EF E30D'
        form = ClaimOpenPGPKeyForm({'fingerprint': fingerprint})

        self.assert_fingerprint_field_has_error(
            form, gpg_messages.unknown_key())

    def test_errors_on_expired_key(self):
        fingerprint = SampleDataKeys.expired_key
        form = ClaimOpenPGPKeyForm({'fingerprint': fingerprint})

        self.assert_fingerprint_field_has_error(
            form, gpg_messages.key_expired(fingerprint))

    def test_errors_on_revoked_key(self):
        fingerprint = SampleDataKeys.revoked_key
        form = ClaimOpenPGPKeyForm({'fingerprint': fingerprint})

        self.assert_fingerprint_field_has_error(
            form, gpg_messages.key_revoked(fingerprint))

    def test_sets_key_details_attribute_on_success(self):
        fingerprint = SampleDataKeys.regular_key
        form = ClaimOpenPGPKeyForm({'fingerprint': fingerprint})

        self.assertTrue(form.is_valid())
        self.assertIn('key_details', form.cleaned_data)
        self.assertIn('can_encrypt', form.cleaned_data['key_details'])
        self.assertTrue(form.cleaned_data['key_details']['can_encrypt'])


class VerifySignedTextFormTestCase(SSOBaseTestCase):

    def setUp(self):
        super(VerifySignedTextFormTestCase, self).setUp()
        self.use_fixture(SSOGPGServiceFixture())
        self.account = self.factory.make_account()

    def assert_clearsigned_field_has_error(self, form, error):
        self.assertIn('clearsigned', form.errors)
        self.assertIn(error, form.errors['clearsigned'])

    def make_sign_only_token(self, fingerprint):
        # The tests below use pre-signed text blocks to work around the
        # difficulty of signing text at test run time. Because of that, we
        # need to carefully control how the auth tokens are generated: We need
        # to make sure that the email address, date created and fingerprint
        # values set on the authtoken never change.
        fingerprint = sanitize_fingerprint(fingerprint)
        self.account = self.factory.make_account(
            email='someone@somewhere.com',
        )
        token = self.factory.make_authtoken(
            token_type=AuthTokenType.VALIDATESIGNONLYGPG,
            email=self.account.preferredemail.email,
            fingerprint=fingerprint,
            requester=self.account,
            date_created=datetime(2016, 4, 12, 16, 0, 0),
        )
        return token

    def test_missing_signed_text(self):
        token = self.make_sign_only_token(SampleDataKeys.sign_only_key)
        form = VerifySignedTextForm({'clearsigned': ''}, token)

        self.assert_clearsigned_field_has_error(
            form, gpg_messages.missing_signed_text()
        )

    def test_text_signed_with_wrong_key(self):
        token = self.make_sign_only_token(SampleDataKeys.sign_only_key)
        form = VerifySignedTextForm(
            {'clearsigned': CORRECT_TEXT_SIGNED_WITH_DIFFERENT_KEY}, token)

        self.assert_clearsigned_field_has_error(
            form, gpg_messages.signing_key_mismatch(SampleDataKeys.regular_key)
        )

    def test_modified_text_signed_with_correct_key(self):
        token = self.make_sign_only_token(SampleDataKeys.sign_only_key)
        form = VerifySignedTextForm(
            {'clearsigned': MODIFIED_TEXT_SIGNED_WITH_CORRECT_KEY}, token)

        self.assert_clearsigned_field_has_error(
            form, gpg_messages.validation_phrase_mismatch()
        )

    def test_correct_text_signed_with_unknown_key(self):
        token = self.make_sign_only_token(SampleDataKeys.sign_only_key)
        form = VerifySignedTextForm(
            {'clearsigned': CORRECT_TEXT_SIGNED_WITH_UNKNOWN_KEY}, token)

        self.assert_clearsigned_field_has_error(
            form, 'Error: Could not find GPG public key.'
        )

    def test_correct_text_signed_with_correct_key(self):
        token = self.make_sign_only_token(SampleDataKeys.sign_only_key)
        form = VerifySignedTextForm(
            {'clearsigned': CORRECT_TEXT_SIGNED_WITH_CORRECT_KEY}, token)

        self.assertTrue(form.is_valid())
        self.assertEqual(
            CORRECT_TEXT_SIGNED_WITH_CORRECT_KEY,
            form.cleaned_data['clearsigned'])


# The correct text, signed with the correct key:
CORRECT_TEXT_SIGNED_WITH_CORRECT_KEY = dedent("""\
    -----BEGIN PGP SIGNED MESSAGE-----
    Hash: SHA1

    Please register 447DBF38C4F9C4ED752246B77D88913717B05A8F to the
    Ubuntu One user with the email address someone@somewhere.com.
    2016-04-12 16:00:00
    -----BEGIN PGP SIGNATURE-----
    Version: GnuPG v1

    iEYEARECAAYFAlcsCwIACgkQfYiRNxewWo/oVgCfcCAoLug0dsVclV0xMKsLlnOY
    bgYAn0faP2lW74vKDkQSmr6gGZvn9tEm
    =KAby
    -----END PGP SIGNATURE-----""")

# Correct text signed with a different key (that we know about)
CORRECT_TEXT_SIGNED_WITH_DIFFERENT_KEY = dedent("""\
    -----BEGIN PGP SIGNED MESSAGE-----
    Hash: SHA1

    Please register 447DBF38C4F9C4ED752246B77D88913717B05A8F to the
    Ubuntu One user with the email address someone@somewhere.com.
    2016-04-12 16:00:00
    -----BEGIN PGP SIGNATURE-----
    Version: GnuPG v1

    iEYEARECAAYFAlcsC20ACgkQ2yWXVgK6Xvb/AQCdFycuh4emAeP3b5KXOVsxn6RD
    JfQAni9VWo6/bik8Kun+7N38LLLwk8m8
    =LXhA
    -----END PGP SIGNATURE-----""")

# signed with the correct key, but the text was changed:
MODIFIED_TEXT_SIGNED_WITH_CORRECT_KEY = dedent("""\
    -----BEGIN PGP SIGNED MESSAGE-----
    Hash: SHA1

    Please DO NOT register 447DBF38C4F9C4ED752246B77D88913717B05A8F to the
    Ubuntu One user with the email address someone@somewhere.com.
    2016-04-12 16:00:00
    -----BEGIN PGP SIGNATURE-----
    Version: GnuPG v1

    iEYEARECAAYFAlcsC64ACgkQfYiRNxewWo+LaACdHdPGy9OD7ClTJsMkBkAMW53p
    W6QAn2WpaL/W7Bbg5OalYRVV8KoctJ9u
    =ciux
    -----END PGP SIGNATURE-----""")

# The text is correct, but it's signed with an unknown key:
CORRECT_TEXT_SIGNED_WITH_UNKNOWN_KEY = dedent("""\
    -----BEGIN PGP SIGNED MESSAGE-----
    Hash: SHA1

    Please register 447DBF38C4F9C4ED752246B77D88913717B05A8F to the
    Ubuntu One user with the email address someone@somewhere.com.
    2016-04-12 16:00:00
    -----BEGIN PGP SIGNATURE-----
    Version: GnuPG v1

    iQIcBAEBAgAGBQJXLAwMAAoJELCMjm0FjU8pyYkP/3chVIGFCM9qwY+gxh8EIDQ2
    VgGQb3D+VBz302sSb7e6/yHXqWct5KuGceV57oEKh7ktMghiJbu0O9tELazR31LH
    MpjqL+FenMTvrx26LV9mp3HOmHYhIR3xbtPHtALeNP/FvfzJKpS3Doe/fX0FxPH7
    G+1VJnVSHD02FkNg0sLGLTGjC4VUG0tR/vL4DNY5DNcM/SpTOGvmTVAj+ODyNlbW
    3hAz2QweZJRT6o4LxZU03B4Oh54WivvaZFWvvpRwOusFunp55om0xH7IEKY/qDcn
    QeDb17SBn64vxh9nohFcrDv+nFCweW4vFPgyZiG/ZQMmBbEM6f0fW/swKFHh/NXP
    iSZ96diqROfMXuoLRTV3LLk+OfIFdpmCjm4A9X5MECzWIovpnAQB3C8AzjYOuil/
    ByGby9XVUdavkc9fdggnxyE/yZqp3neSu1qhOIN2T1M1QJlRsI9VdYMFUrZ/sSKl
    3eDuEFa2E4xwdozbJPB4piGSPt3lra5sGosl6KMYh8Q8slaL73DPjYNKn3fy2ftP
    C9BIhRDyCg7U4QLYKoUK52GaNSwn34sc1WjvCzo3BMoegUVsT/sta+zZKDbXalIE
    zq2YCMixKTKfWUhJXuZPhWuDCVT1xaZgOlQyvL8kcG2rk9jv7FuUqrg6j8tT2jTw
    3I0l1CMSQHW94vugh0ZH
    =XOLk
    -----END PGP SIGNATURE-----""")


class PendingGPGValidationTokenFormTestCase(SSOBaseTestCase):

    def setUp(self):
        super(PendingGPGValidationTokenFormTestCase, self).setUp()
        self.account = self.factory.make_account()

    def add_encryption_key_pending_token(self, fingerprint, **kwargs):
        default_args = {
            'requester': self.account,
            'token_type': AuthTokenType.VALIDATEGPG,
            'fingerprint': fingerprint,
            'email': self.account.preferredemail.email
        }
        default_args.update(kwargs)
        return self.factory.make_authtoken(**default_args)

    def add_sign_only_key_pending_token(self, fingerprint, **kwargs):
        default_args = {
            'requester': self.account,
            'token_type': AuthTokenType.VALIDATESIGNONLYGPG,
            'fingerprint': fingerprint,
            'email': self.account.preferredemail.email
        }
        default_args.update(kwargs)
        return self.factory.make_authtoken(**default_args)

    def test_form_with_no_tokens(self):
        form = PendingGPGValidationTokenForm(self.account)

        self.assertFalse(form.is_bound)
        self.assertEqual(0, len(form.fields['tokens'].choices))

    def test_form_with_multiple_tokens(self):
        t1 = self.add_encryption_key_pending_token(SampleDataKeys.regular_key)
        t2 = self.add_sign_only_key_pending_token(SampleDataKeys.sign_only_key)

        form = PendingGPGValidationTokenForm(self.account)

        self.assertFalse(form.is_bound)
        self.assertEqual(2, len(form.fields['tokens'].choices))
        choice_labels = [v for k, v in form.fields['tokens'].choices]
        self.assertIn(
            str(gpg_messages.display_fingerprint(
                t1.fingerprint, markup=False)),
            choice_labels[1]
        )
        self.assertIn(
            str(gpg_messages.display_fingerprint(
                t2.fingerprint, markup=False)),
            choice_labels[0]
        )

    def test_can_select_multiple_fingerprint(self):
        t1 = self.add_encryption_key_pending_token(SampleDataKeys.regular_key)
        t2 = self.add_sign_only_key_pending_token(SampleDataKeys.sign_only_key)

        form = PendingGPGValidationTokenForm(
            self.account,
            data={'tokens': [t1.id, t2.id]}
        )

        self.assertTrue(form.is_bound)
        self.assertTrue(form.is_valid())

        selected_tokens = form.cleaned_data['tokens']
        self.assertEqual(2, len(selected_tokens))
        self.assertIn(t1, selected_tokens)
        self.assertIn(t2, selected_tokens)

    def test_can_select_single_fingerprint(self):
        self.add_encryption_key_pending_token(SampleDataKeys.regular_key)
        t1 = self.add_sign_only_key_pending_token(SampleDataKeys.sign_only_key)

        form = PendingGPGValidationTokenForm(
            self.account,
            data={'tokens': [t1.id]}
        )

        self.assertTrue(form.is_bound)
        self.assertTrue(form.is_valid())

        selected_tokens = form.cleaned_data['tokens']
        self.assertEqual(1, len(selected_tokens))
        self.assertIn(t1, selected_tokens)

    def test_errors_with_invalid_choice(self):
        t1 = self.add_encryption_key_pending_token(SampleDataKeys.regular_key)

        form = PendingGPGValidationTokenForm(
            self.account,
            data={'tokens': [t1.id + 1]}
        )

        self.assertFalse(form.is_valid())
        self.assertIn('tokens', form.errors)
        self.assertEqual(1, len(form.errors['tokens']))
        token_error = form.errors['tokens'][0]
        self.assertTrue(token_error.startswith("Select a valid choice."))

    def test_errors_with_no_choice(self):
        self.add_encryption_key_pending_token(SampleDataKeys.regular_key)

        form = PendingGPGValidationTokenForm(
            self.account,
            data={'tokens': ''}
        )

        self.assertFalse(form.is_valid())
        self.assertIn('tokens', form.errors)
        self.assertEqual(["This field is required."], form.errors['tokens'])

    def test_does_not_include_consumed_tokens(self):
        token = self.add_encryption_key_pending_token(
            SampleDataKeys.regular_key)
        token.consume()
        form = PendingGPGValidationTokenForm(self.account)

        self.assertFalse(form.is_bound)
        self.assertEqual(0, len(form.fields['tokens'].choices))


class GPGKeysFormTestCase(SSOBaseTestCase):

    def setUp(self):
        super(GPGKeysFormTestCase, self).setUp()
        self.use_fixture(SSOGPGServiceFixture())
        self.account = self.factory.make_account()

    def test_form_with_no_keys(self):
        form = EnabledGPGKeysForm(self.account)
        self.assertEqual(0, len(list(form.fields['keys'].choices)))

    def test_form_with_one_key(self):
        self.factory.add_gpgkey_to_account(
            self.account, SampleDataKeys.regular_key)
        form = EnabledGPGKeysForm(self.account)
        self.assertEqual(1, len(list(form.fields['keys'].choices)))

    def test_form_can_filter_only_disabled_keys(self):
        disabled_key = SampleDataKeys.regular_key.replace(' ', '')
        enabled_key = SampleDataKeys.sign_only_key.replace(' ', '')
        self.factory.add_gpgkey_to_account(
            self.account, disabled_key, keysize=4096, algorithm='R',
            enabled=False)
        self.factory.add_gpgkey_to_account(
            self.account, enabled_key, enabled=True)

        form = DisabledGPGKeysForm(self.account)
        choices = list(form.fields['keys'].choices)
        expected_display_format = "4096R/%s" % disabled_key[-8:]

        self.assertEqual([(disabled_key, expected_display_format)], choices)

    def test_form_can_filter_only_enabled_keys(self):
        disabled_key = SampleDataKeys.regular_key.replace(' ', '')
        enabled_key = SampleDataKeys.sign_only_key.replace(' ', '')
        self.factory.add_gpgkey_to_account(
            self.account, disabled_key, enabled=False)
        self.factory.add_gpgkey_to_account(
            self.account, enabled_key, keysize=4096, algorithm='R',
            can_encrypt=False, enabled=True
        )

        form = EnabledGPGKeysForm(self.account)
        choices = list(form.fields['keys'].choices)
        expected_display_format = "4096R/%s (sign only)" % enabled_key[-8:]

        self.assertEqual([(enabled_key, expected_display_format)], choices)

    def test_form_errors_on_invalid_selection(self):
        actual_fingerprint = SampleDataKeys.regular_key
        wrong_fingerprint = 'A' * 40
        self.factory.add_gpgkey_to_account(self.account, actual_fingerprint)

        form = EnabledGPGKeysForm(
            self.account, data={'keys': [wrong_fingerprint]})

        self.assertFalse(form.is_valid())
        self.assertIn('keys', form.errors)
        self.assertEqual(
            form.errors['keys'],
            ["Select a valid choice. %s is not one of the available choices."
             % wrong_fingerprint]
        )

    def test_form_errors_on_no_selection(self):
        form = EnabledGPGKeysForm(self.account, data={'keys': []})

        self.assertFalse(form.is_valid())
        self.assertIn('keys', form.errors)
        self.assertEqual(['This field is required.'], form.errors['keys'])

    def test_can_select_single_fingerprint(self):
        fingerprint = SampleDataKeys.regular_key.replace(' ', '')
        self.factory.add_gpgkey_to_account(self.account, fingerprint)

        form = EnabledGPGKeysForm(self.account, data={'keys': [fingerprint]})
        self.assertTrue(form.is_valid())
        self.assertEqual([fingerprint], form.cleaned_data['keys'])

    def test_can_select_multiple_fingerprints(self):
        fingerprint1 = SampleDataKeys.regular_key.replace(' ', '')
        fingerprint2 = SampleDataKeys.sign_only_key.replace(' ', '')
        self.factory.add_gpgkey_to_account(self.account, fingerprint1)
        self.factory.add_gpgkey_to_account(self.account, fingerprint2)

        form = EnabledGPGKeysForm(
            self.account, data={'keys': [fingerprint1, fingerprint2]})
        self.assertTrue(form.is_valid())
        self.assertEqual(
            sorted([fingerprint1, fingerprint2]),
            sorted(form.cleaned_data['keys'])
        )

    def test_form_key_display_for_encryption_keys(self):
        fingerprint = SampleDataKeys.regular_key.replace(' ', '')
        keyid = fingerprint[-8:]
        self.factory.add_gpgkey_to_account(
            self.account,
            fingerprint,
            keysize=2048,
            algorithm='R',
            can_encrypt=True
        )

        form = EnabledGPGKeysForm(self.account)

        _, form_display = list(form.fields['keys'].choices)[0]
        self.assertEqual('2048R/%s' % keyid, form_display)

    def test_form_key_display_for_sign_only_keys(self):
        fingerprint = SampleDataKeys.sign_only_key.replace(' ', '')
        keyid = fingerprint[-8:]
        self.factory.add_gpgkey_to_account(
            self.account,
            fingerprint,
            keysize=4096,
            algorithm='D',
            can_encrypt=False
        )

        form = EnabledGPGKeysForm(self.account)

        _, form_display = list(form.fields['keys'].choices)[0]
        self.assertEqual('4096D/%s (sign only)' % keyid, form_display)

    def test_choices_does_not_consume_generator(self):
        self.factory.add_gpgkey_to_account(self.account)
        form = EnabledGPGKeysForm(self.account)

        self.assertEqual(1, form.num_choices())
        self.assertEqual(1, form.num_choices())


class ClaimSSHKeyFormTestCase(SSOBaseTestCase):

    def assert_field_is_disabled(self, field):
        self.assertIn('disabled', field.widget.attrs)

    def assert_field_placeholder_equals(self, field, expected):
        self.assertEqual(expected, str(field.widget.attrs['placeholder']))

    def make_full_account(self):
        """A convenience function to make a 'full' account.

        A 'full' account in this context means that the email has been verified
        and a launchpad username has been set.
        """
        account = self.factory.make_account(email_validated=True)
        self.factory.make_person(account=account)
        return account

    def patch_lp_add_key(self, raise_exception=None):
        """Patch the 'add_lp_ssh_key', creating realistic side effects.

        :param raise_exception: If specified, must be one of NoSuchAccount,
            SSHKeyAdditionTypeError, SSHKeyAdditionDataError or
            SSHKeyCompromisedError. Exceptions will be raised with the same
            message as if they'd been returned from launchpad.
        """
        exception_messages = {
            NoSuchAccount: "No account found for openid identifier '{openid}'",
            SSHKeyAdditionTypeError: "Invalid SSH key type: '{sshkeytype}'",
            SSHKeyAdditionDataError: "Invalid SSH key data: '{sshkey}'",
            SSHKeyCompromisedError: "This key cannot be added as it is "
                                    "known to be compromised.",
        }

        def _side_effect(openid_id, key_text, dry_run):
            key_type = key_text.split(' ')[0]
            message = exception_messages[raise_exception].format(
                openid=openid_id,
                sshkeytype=key_type,
                sshkey=key_text,
            )
            raise raise_exception(400, message)

        mock = self.patch('webui.forms.add_lp_ssh_key')
        if raise_exception is not None:
            mock.side_effect = _side_effect
        return mock

    def assert_form_field_has_error(self, form, field, expected_error):
        self.assertIn(field, form.errors)
        self.assertEqual([expected_error], form.errors[field])

    def test_is_disabled_with_unverified_account(self):
        account = self.factory.make_account(email_validated=False)
        form = ClaimSSHKeyForm(account)

        self.assert_field_is_disabled(form.fields['ssh_key'])
        self.assert_field_placeholder_equals(
            form.fields['ssh_key'],
            "You must verify your account and set a username first.",
        )

    def test_is_disabled_with_no_username_set(self):
        account = self.factory.make_account(email_validated=True)
        form = ClaimSSHKeyForm(account)

        self.assert_field_is_disabled(form.fields['ssh_key'])
        self.assert_field_placeholder_equals(
            form.fields['ssh_key'],
            "You must set a username first.",
        )

    def test_form_not_disabled(self):
        account = self.make_full_account()
        form = ClaimSSHKeyForm(account)

        self.assertNotIn('disabled', form.fields['ssh_key'].widget.attrs)
        self.assert_field_placeholder_equals(
            form.fields['ssh_key'],
            "SSH Key Text.",
        )

    def test_form_errors_with_missing_Data(self):
        account = self.make_full_account()
        form = ClaimSSHKeyForm(account, {'ssh_key': ''})

        self.assertFalse(form.is_valid())
        self.assert_form_field_has_error(
            form, 'ssh_key', 'This field is required.')

    def test_form_uses_dry_run_when_cleaning_data(self):
        account = self.make_full_account()
        form = ClaimSSHKeyForm(
            account, {'ssh_key': 'ssh-rsa keytext comment'})

        mock_lp_add = self.patch_lp_add_key()

        self.assertTrue(form.is_valid())
        mock_lp_add.assert_called_once_with(
            account.openid_identifier, 'ssh-rsa keytext comment', dry_run=True)

    def test_form_converts_lp_NoSuchAccount_to_form_error(self):
        account = self.make_full_account()
        form = ClaimSSHKeyForm(
            account, {'ssh_key': 'ssh-rsa keytext comment'})

        self.patch_lp_add_key(NoSuchAccount)
        self.assertFalse(form.is_valid())
        self.assert_form_field_has_error(
            form,
            'ssh_key',
            "No account found for openid identifier '%s'"
            % account.openid_identifier,
        )

    def test_form_converts_lp_SSHKeyAdditionTypeError_to_form_error(self):
        account = self.make_full_account()
        form = ClaimSSHKeyForm(
            account, {'ssh_key': 'ssh-flub keytext comment'})

        self.patch_lp_add_key(SSHKeyAdditionTypeError)
        self.assertFalse(form.is_valid())
        self.assert_form_field_has_error(
            form,
            'ssh_key',
            "Invalid SSH key type: 'ssh-flub'",
        )

    def test_form_converts_lp_SSHKeyAdditionDataError_to_form_error(self):
        account = self.make_full_account()
        form = ClaimSSHKeyForm(
            account, {'ssh_key': 'badkey'})

        self.patch_lp_add_key(SSHKeyAdditionDataError)
        self.assertFalse(form.is_valid())
        self.assert_form_field_has_error(
            form,
            'ssh_key',
            "Invalid SSH key data: 'badkey'",
        )

    def test_form_converts_lp_SSHKeyCompromisedError_to_form_error(self):
        account = self.make_full_account()
        form = ClaimSSHKeyForm(
            account, {'ssh_key': 'ssh-dss foo bar'})

        self.patch_lp_add_key(SSHKeyCompromisedError)
        self.assertFalse(form.is_valid())
        self.assert_form_field_has_error(
            form,
            'ssh_key',
            "This key cannot be added as it is known to be compromised.",
        )

    def test_form_adds_ssh_key(self):
        account = self.make_full_account()
        form = ClaimSSHKeyForm(
            account, {'ssh_key': 'ssh-rsa foo bar'})

        mock_lp_add = self.patch_lp_add_key()
        self.assertTrue(form.is_valid())
        mock_lp_add.reset_mock()
        form.add_key_to_launchpad()

        mock_lp_add.assert_called_once_with(
            account.openid_identifier,
            'ssh-rsa foo bar',
            dry_run=False)


class FormatSSHKeyTestCase(SSOBaseTestCase):

    def test_escapes_xss(self):
        bad_content = '<script>window.alert("boo!");</script>'
        bad_key = ' '.join((bad_content,) * 3)
        formatted_key = format_ssh_key(bad_key)
        self.assertNotIn(bad_content, formatted_key)
        self.assertIn(
            '&lt;script&gt;window.alert(&quot;boo!&quot;);&lt;/script&gt;',
            formatted_key
        )


class DeleteSSHKeyFormTestCase(SSOBaseTestCase):

    SAMPLE_RSA_KEY = 'ssh-rsa keytext_goes_here comment goes here'
    SAMPLE_DSA_KEY = 'ssh-dss more_keytext_here this is a comment'

    def setUp(self):
        super(DeleteSSHKeyFormTestCase, self).setUp()
        self.account = self.factory.make_account()

    def patch_lp_delete_key(self, raise_exception=None):
        """Patch the 'delete_lp_ssh_key', creating realistic side effects.

        :param raise_exception: If specified, must be one of NoSuchAccount,
            SSHKeyAdditionTypeError or SSHKeyAdditionDataError Exceptions will
            be raised with the same message as if they'd been returned from
            launchpad.
        """
        exception_messages = {
            NoSuchAccount: "No account found for openid identifier '{openid}'",
            SSHKeyAdditionTypeError: "Invalid SSH key type: '{sshkeytype}'",
            SSHKeyAdditionDataError: "Invalid SSH key data: '{sshkey}'",
        }

        def _side_effect(openid_id, key_text, dry_run):
            key_type = key_text.split(' ')[0]
            message = exception_messages[raise_exception].format(
                openid=openid_id,
                sshkeytype=key_type,
                sshkey=key_text,
            )
            raise raise_exception(400, message)

        mock = self.patch('webui.forms.delete_lp_ssh_key')
        if raise_exception is not None:
            mock.side_effect = _side_effect
        return mock

    def patch_get_keys(self, returned_keys=[]):
        mock = self.patch('webui.forms.get_lp_ssh_keys')
        mock.return_value = returned_keys
        return mock

    def assert_form_field_has_error(self, form, field, expected_error):
        self.assertIn(field, form.errors)
        self.assertEqual([expected_error], form.errors[field])

    def test_form_passes_account_id_to_get_lp_keys(self):
        mock_get_keys = self.patch_get_keys()
        DeleteSSHKeyForm(self.account)

        mock_get_keys.assert_called_once_with(self.account.openid_identifier)

    def test_form_with_no_keys(self):
        self.patch_get_keys()
        form = DeleteSSHKeyForm(self.account)

        self.assertEqual(0, len(form.fields['ssh_keys'].choices))

    def test_form_multiple_keys(self):
        self.patch_get_keys([self.SAMPLE_RSA_KEY, self.SAMPLE_DSA_KEY])
        form = DeleteSSHKeyForm(self.account)

        self.assertEqual(2, len(form.fields['ssh_keys'].choices))
        rsa_choice, dsa_choice = form.fields['ssh_keys'].choices
        self.assertEqual(
            (self.SAMPLE_RSA_KEY, format_ssh_key(self.SAMPLE_RSA_KEY)),
            rsa_choice,
        )
        self.assertEqual(
            (self.SAMPLE_DSA_KEY, format_ssh_key(self.SAMPLE_DSA_KEY)),
            dsa_choice,
        )

    def test_form_deletes_ssh_key(self):
        self.patch_get_keys([self.SAMPLE_RSA_KEY])
        form = DeleteSSHKeyForm(
            self.account, {'ssh_keys': [self.SAMPLE_RSA_KEY]})

        mock_lp_delete = self.patch_lp_delete_key()
        self.assertTrue(form.is_valid())
        mock_lp_delete.reset_mock()
        form.delete_keys_from_launchpad()

        mock_lp_delete.assert_called_once_with(
            self.account.openid_identifier,
            self.SAMPLE_RSA_KEY,
            dry_run=False)
