# Copyright 2016 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

from datetime import datetime

from identityprovider.tests.utils import SSOBaseTestCase
from identityprovider.models.const import AuthTokenType
from identityprovider.models.authtoken import AuthToken
from webui.gpg import (
    compare_gpg_fingerprints,
    SSOGPGClient,
    validation_phrase_for_authtoken,
)
from webui.tests.gpg_fixture import SSOGPGServiceFixture


class GPGServiceFixtureTestCase(SSOBaseTestCase):

    def test_can_talk_to_service_with_fixture(self):
        self.use_fixture(SSOGPGServiceFixture())
        client = SSOGPGClient()

        keys = client.getKeysForOwner('name16_oid')['keys']
        self.assertEqual(1, len(keys))


class GPGFingerprintEqualityTestCase(SSOBaseTestCase):

    def test_ignores_case(self):
        self.assertTrue(compare_gpg_fingerprints('a' * 40, 'A' * 40))

    def test_ignores_whitespace(self):
        self.assertTrue(compare_gpg_fingerprints('A ' * 40, 'A' * 40))

    def test_different_lengths(self):
        self.assertFalse(compare_gpg_fingerprints('0' * 32, '0' * 40))

    def test_different_contents(self):
        self.assertFalse(
            compare_gpg_fingerprints(
                'AAAA DEAD BEEF FEED FACE  0000 DEAD BEEF FEED FACE',
                '1234 FACE 3726 F00D 2836  1723 B34F F32D F34D F433',
            ))


class AuthTokenValidationPhraseTestCase(SSOBaseTestCase):

    def setUp(self):
        super(AuthTokenValidationPhraseTestCase, self).setUp()
        self.fingerprint = 'A000B111C222D333E444F55566667777888899990000'
        self.requester = self.factory.make_account()
        self.date_created = datetime(2016, 5, 6, 14, 29, 13)

    def test_validation_phrase(self):
        token = self.factory.make_authtoken(
            token_type=AuthTokenType.VALIDATESIGNONLYGPG,
            fingerprint=self.fingerprint,
            date_created=self.date_created,
            email=self.requester.preferredemail,
            requester=self.requester,
        )
        expected = (
            'Please register %s to the\n'
            'Ubuntu One user with the email address %s.\n'
            '%s') % (
            self.fingerprint,
            self.requester.preferredemail,
            self.date_created.strftime('%Y-%m-%d %H:%M:%S'))
        self.assertEqual(expected, validation_phrase_for_authtoken(token))

    def test_raises_ValueError_for_bad_token_types(self):
        bad_token_types = [t for t in AuthToken.objects.valid_types
                           if t is not AuthTokenType.VALIDATESIGNONLYGPG]
        for bad_token_type in bad_token_types:
            bad_token = self.factory.make_authtoken(
                token_type=bad_token_type,
                fingerprint=self.fingerprint,
                date_created=self.date_created,
                email=self.requester.preferredemail,
                requester=self.requester,
            )
            self.assertRaises(
                ValueError, validation_phrase_for_authtoken, bad_token)


class GPGClientTestCase(SSOBaseTestCase):

    def test_unconfigured_client_raises_helpful_expection(self):
        with self.settings(GPGSERVICE_ENDPOINT=None):
            with self.assertRaises(RuntimeError) as cm:
                SSOGPGClient()
            self.assertEqual(
                "Error: GPG functionality is not available unless the "
                "'GPGSERVICE_ENDPOINT' setting has been set to the URL of "
                "a running gpgservice instance. See README for more details.",
                str(cm.exception)
            )
