# Copyright 2010 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

from __future__ import unicode_literals

from django.conf import settings
from django.core import mail

from identityprovider.fields import FIELD_REQUIRED, INVALID_EMAIL
from identityprovider.login import LOGIN_INCORRECT
from identityprovider.models.authtoken import AuthToken
from identityprovider.tests.utils import SSOBaseTestCase


class LoginTest(SSOBaseTestCase):

    def setUp(self):
        super(LoginTest, self).setUp()
        self.preseed_cookie_check()

    def test_loginform_password_did_not_match_error(self):
        r = self.client.post('/+login', {'email': "mark@example.com",
                                         'password': "foo"})
        self.assertFormError(r, 'form', None, unicode(LOGIN_INCORRECT))

    def test_loginform_wrong_email_error(self):
        r = self.client.post('/+login', {'email': "foo@"})
        self.assertFormError(r, 'form', 'email', unicode(INVALID_EMAIL))

    def test_loginform_password_required_error(self):
        r = self.client.post('/+login', {'email': "mark@example.com",
                                         'passwrd': ""})
        self.assertFormError(r, 'form', 'password', unicode(FIELD_REQUIRED))


class HybridLoginNewAccountTest(SSOBaseTestCase):

    def test_create_account_form_required_field(self):
        self.preseed_cookie_check()
        response = self.client.get('/+login')

        self.assertContains(response, "login_create_account_radio")
        self.assertContains(response, "login_form")
        self.assertContains(response, "create_account_form")


class NewAccountTest(SSOBaseTestCase):

    def test_newaccount_existing(self):
        query = {
            'email': 'nobody@debian.org',
        }
        self.preseed_cookie_check()
        response = self.client.post('/+forgot_password', query)
        self.assertEqual(response.status_code, 200)
        self.assertIn("Reset password", response.content)
        self.assertIn("nobody@debian.org", response.content)
        self.assertEqual(len(mail.outbox), 1)
        self.assertEqual(
            mail.outbox[0].subject,
            u"%s: Password reset request" % settings.NOREPLY_FROM_NAME)
        self.assertIn('nobody@debian.org', mail.outbox[0].body)
        self.assertIn('+new_account', mail.outbox[0].body)
        self.assertFormError(response, 'form', 'email',
                             'No account associated with nobody@debian.org')

    def test_newaccount_no_tos_accept(self):
        self.preseed_cookie_check()
        r = self.client.post('/+new_account', {
            'email': "mark@example.com",
            'password': "foofoofoo",
            'passwordconfirm': "foofoofoo",
            'displayname': 'bar'
        })

        s = ("Check the box to indicate that "
             "you agree with our terms of use:")

        self.assertFormError(r, 'form', 'accept_tos', s)


class ForgottenPasswordTest(SSOBaseTestCase):

    fixtures = ["test"]

    def test_forgottenpass_nonexisting(self):
        query = {
            'displayname': 'Tester',
            'email': 'test@canonical.com',
            'password': 'Testing123',
            'passwordconfirm': 'Testing123',
            'accept_tos': True,
        }
        self.preseed_cookie_check()
        response = self.client.post('/+new_account', query)
        self.assertContains(response, "test@canonical.com")
        self.assertEqual(len(mail.outbox), 1)
        self.assertEqual(mail.outbox[0].subject,
                         u"%s: Warning" % settings.NOREPLY_FROM_NAME)
        mail.outbox = []

    def test_resetform_success(self):
        query = {
            'email': 'test@canonical.com',
        }
        self.preseed_cookie_check()
        response = self.client.post('/+forgot_password', query)
        self.assertContains(response, "Reset password")
        self.assertContains(response, "test@canonical.com")
        self.assertEqual(len(mail.outbox), 1)
        self.assertEqual(
            mail.outbox[0].subject,
            u"%s: Forgotten Password" % settings.NOREPLY_FROM_NAME)
        mail.outbox = []

        AuthToken.objects.all().delete()
