# Copyright 2010 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

from __future__ import unicode_literals

import BeautifulSoup
from django.core import mail
from django.core.urlresolvers import reverse
from gargoyle.testutils import switches
from gpgservice_client.testing.keyserver.tests.keys import SampleDataKeys

from identityprovider.models import AuthToken
from identityprovider.models.const import AuthTokenType
from identityprovider.tests import DEFAULT_USER_PASSWORD
from identityprovider.tests.utils import SSOBaseTestCase
from webui.gpg import (
    get_user_id_for_account,
    SSOGPGClient,
)
from webui.tests.gpg_fixture import SSOGPGServiceFixture
from webui.views import gpg_messages
from webui.views.account import gpg_keys


class GPGTestCase(SSOBaseTestCase):

    def setUp(self):
        super(GPGTestCase, self).setUp()
        self.use_fixture(SSOGPGServiceFixture())

    def assert_response_includes_message(self, response, expected_message):
        self.assertIn('messages', response.context)
        messages = [m.message for m in response.context['messages']]
        self.assertIn(expected_message, messages)

    def assert_gpg_key_notification_sent(self):
        self.assertEqual(1, len(mail.outbox))
        email = mail.outbox[0]
        self.assertEqual('Ubuntu One: Confirm your OpenPGP Key', email.subject)
        self.assertEqual(
            'Ubuntu One OpenPGP Key Confirmation <noreply@ubuntu.com>',
            email.from_email
        )
        self.assertEqual([self.account.preferredemail.email], email.to)

    def assert_form_labels_not_seen_before(self, form, seen_ids):
        """Assert that all the labels in 'form' use an id not seen before.

        :param form: Must be a BeautifulSoup node that represents a <form> tag.
        :param seen_ids: Must be a set that will be used to record the ids seen
            so far.
        """
        for label in form.findAll('label'):
            for_id = [v for k, v in label.attrs if k == 'for']
            if for_id:
                self.assertEqual(1, len(for_id))
                for_id = for_id[0]
                self.assertNotIn(for_id, seen_ids)
                seen_ids.add(for_id)

    def _get_keys_for_account(self):
        user_id = get_user_id_for_account(self.account)
        return SSOGPGClient().getKeysForOwner(user_id)['keys']

    def assert_account_has_disabled_key(self, fingerprint):
        keys = self._get_keys_for_account()
        found_keys = [k for k in keys if k['fingerprint'] == fingerprint]
        self.assertEqual(
            1, len(found_keys),
            "Could not find key with fingerprint %r for account %r"
            % (fingerprint, self.account)
        )
        key = found_keys[0]
        self.assertFalse(
            key['enabled'],
            "Key with fingerprint %r on account %r is enabled "
            "(expected disabled)" % (fingerprint, self.account)
        )

    def assert_account_has_pending_gpg_token(self, fingerprint, token_type):
        try:
            AuthToken.get_token(
                fingerprint=fingerprint,
                token_type=token_type,
                requester=self.account,
                date_consumed=None,
            )
        except AuthToken.DoesNotExist:
            self.fail(
                "Account %r does not have a pending auth token of type %r "
                "for fingerprint %r" % (self.account, token_type, fingerprint))

    def create_user_and_login(self, name='person-2'):
        self.account = self.factory.make_account(
            password=DEFAULT_USER_PASSWORD,
            email='example_user@example.com')
        assert self.client.login(
            username=self.account.preferredemail.email,
            password=DEFAULT_USER_PASSWORD)

    @switches(GPG_SERVICE=False)
    def test_gpg_keys_menu_item_not_rendered_without_feature_flag(self):
        self.create_user_and_login()
        resp = self.client.get(reverse('account-index'))

        self.assertNotContains(resp, '<span>GPG Keys</span>', html=True)

    @switches(GPG_SERVICE=True)
    def test_gpg_keys_menu_item_is_rendered_with_feature_flag(self):
        self.create_user_and_login()

        resp = self.client.get(reverse('account-index'))

        self.assertContains(resp, '<span>GPG Keys</span>', html=True)

    @switches(GPG_SERVICE=False)
    def test_gpg_keys_view_returns_404_without_feature_flag(self):
        self.create_user_and_login()

        resp = self.client.get(reverse('gpg_keys'))
        self.assertEqual(404, resp.status_code)

    @switches(GPG_SERVICE=True)
    def test_gpg_keys_view_returns_200_with_feature_flag(self):
        self.create_user_and_login()

        resp = self.client.get(reverse('gpg_keys'))
        self.assertContains(
            resp, '<h1 class="u1-h-main">GPG Keys</h1>', html=True)

    def test_gpg_keys_view_returns_200_with_feature_flag_conditional(self):
        self.create_user_and_login()
        condition_set = ('identityprovider.gargoyle_conditions.'
                         'AccountConditionSet(identityprovider.account)')
        self.conditionally_enable_flag(
            'GPG_SERVICE',
            'email',
            self.account.preferredemail.email,
            condition_set
        )
        resp = self.client.get(reverse('gpg_keys'))
        self.assertContains(
            resp, '<h1 class="u1-h-main">GPG Keys</h1>', html=True)

    def test_gpg_keys_view_returns_404_without_feature_flag_conditional(self):
        self.create_user_and_login()
        # Set a conditional flag for a second account to verify that it really
        # is conditional:
        special_account = self.factory.make_account()
        condition_set = ('identityprovider.gargoyle_conditions.'
                         'AccountConditionSet(identityprovider.account)')
        self.conditionally_enable_flag(
            'GPG_SERVICE',
            'email',
            special_account.preferredemail.email,
            condition_set
        )
        resp = self.client.get(reverse('gpg_keys'))
        self.assertEqual(404, resp.status_code)

    @switches(GPG_SERVICE=True)
    def test_gpg_keys_view_requires_authentication(self):
        self.preseed_cookie_check()
        resp = self.client.get(reverse('gpg_keys'))
        self.assertRedirects(
            resp,
            reverse('login') + '?next=' + reverse('gpg_keys'),
        )

    @switches(GPG_SERVICE=True)
    def test_gpg_keys_view_disabled_in_readonly_mode(self):
        account = self.factory.make_account()
        with self.settings(READ_ONLY_MODE=True):
            request = self.factory.make_request(user=account)
            resp = gpg_keys(request)
        self.assertContains(
            resp,
            "<h1 class='main'>We're currently in read-only mode.</h1>",
            status_code=403,
            html=True)

    @switches(GPG_SERVICE=True)
    def test_can_create_pending_token_for_encryption_key(self):
        self.create_user_and_login()
        resp = self.client.post(
            reverse('gpg_keys'),
            data={
                'claim_gpg': '',
                'fingerprint': SampleDataKeys.regular_key
            },
            follow=True,
        )

        self.assertRedirects(resp, reverse('gpg_keys'))
        internal_fingerprint_format = SampleDataKeys.regular_key.replace(
            ' ', '')
        token = AuthToken.get_token(fingerprint=internal_fingerprint_format)
        self.assertEqual(self.account, token.requester)
        self.assertEqual(internal_fingerprint_format, token.fingerprint)
        self.assertIsNotNone(token.date_created)
        self.assertEqual(self.account.preferredemail.email, token.email)
        self.assertEqual(AuthTokenType.VALIDATEGPG, token.token_type)
        self.assertContains(
            resp,
            "An email with further instructions has been sent to %s"
            % self.account.preferredemail,
        )
        self.assert_gpg_key_notification_sent()

    @switches(GPG_SERVICE=True)
    def test_can_create_pending_token_for_sign_only_key(self):
        self.create_user_and_login()
        resp = self.client.post(
            reverse('gpg_keys'),
            data={
                'claim_gpg': '',
                'fingerprint': SampleDataKeys.sign_only_key
            },
            follow=True,
        )

        self.assertRedirects(resp, reverse('gpg_keys'))
        internal_fingerprint_format = SampleDataKeys.sign_only_key.replace(
            ' ', '')
        token = AuthToken.get_token(fingerprint=internal_fingerprint_format)
        self.assertEqual(self.account, token.requester)
        self.assertEqual(internal_fingerprint_format, token.fingerprint)
        self.assertEqual(self.account.preferredemail.email, token.email)
        self.assertIsNotNone(token.date_created)
        self.assertEqual(AuthTokenType.VALIDATESIGNONLYGPG, token.token_type)
        self.assertContains(
            resp,
            "An email with further instructions has been sent to %s"
            % self.account.preferredemail,
        )
        self.assert_gpg_key_notification_sent()

    @switches(GPG_SERVICE=True)
    def test_claim_token_form_errors_appear_on_view(self):
        self.create_user_and_login()
        resp = self.client.post(
            reverse('gpg_keys'),
            data={
                'claim_gpg': '',
                'fingerprint': 'not a valid fingerprint!'
            },
        )
        self.assertFormError(
            resp,
            'claim_form',
            'fingerprint',
            gpg_messages.bad_fingerprint_format(),
        )

    @switches(GPG_SERVICE=True)
    def test_pending_key_form_not_shown_with_no_pending_tokens(self):
        self.create_user_and_login()
        resp = self.client.get(reverse('gpg_keys'))
        self.assertNotContains(resp, "Keys pending validation")

    @switches(GPG_SERVICE=True)
    def test_pending_key_form_shown_with_pending_tokens(self):
        self.create_user_and_login()
        self.factory.make_authtoken(
            requester=self.account,
            token_type=AuthTokenType.VALIDATEGPG,
            fingerprint=SampleDataKeys.regular_key,
            email=self.account.preferredemail
        )
        resp = self.client.get(reverse('gpg_keys'))

        self.assertContains(resp, "Keys pending validation")

    @switches(GPG_SERVICE=True)
    def test_can_cancel_pending_token(self):
        self.create_user_and_login()
        fingerprint = SampleDataKeys.regular_key.replace(' ', '')
        token = self.factory.make_authtoken(
            requester=self.account,
            token_type=AuthTokenType.VALIDATEGPG,
            fingerprint=fingerprint,
            email=self.account.preferredemail
        )

        resp = self.client.post(
            reverse('gpg_keys'),
            data={'cancel_pending': '', 'tokens': [token.id]},
            follow=True
        )

        self.assertRedirects(resp, reverse('gpg_keys'))
        self.assert_response_includes_message(
            resp, gpg_messages.pending_tokens_cancelled([fingerprint]))
        self.assertRaises(
            AuthToken.DoesNotExist,
            AuthToken.get_token,
            fingerprint=fingerprint
        )

    @switches(GPG_SERVICE=True)
    def test_pending_form_errors_are_rendered(self):
        self.create_user_and_login()
        fingerprint = SampleDataKeys.regular_key.replace(' ', '')
        token = self.factory.make_authtoken(
            requester=self.account,
            token_type=AuthTokenType.VALIDATEGPG,
            fingerprint=fingerprint,
            email=self.account.preferredemail
        )
        bad_choice = token.id + 1

        resp = self.client.post(
            reverse('gpg_keys'),
            data={'cancel_pending': '', 'tokens': [bad_choice]},
        )

        expected_error = (
            'Select a valid choice. %d is not one of the available choices.'
            % bad_choice
        )

        self.assertFormError(resp, 'pending_form', 'tokens', expected_error)

    @switches(GPG_SERVICE=True)
    def test_can_cancel_multiple_tokens_at_once(self):
        self.create_user_and_login()
        token1 = self.factory.make_authtoken(
            requester=self.account,
            token_type=AuthTokenType.VALIDATEGPG,
            fingerprint=SampleDataKeys.regular_key.replace(' ', ''),
            email=self.account.preferredemail
        )
        token2 = self.factory.make_authtoken(
            requester=self.account,
            token_type=AuthTokenType.VALIDATESIGNONLYGPG,
            fingerprint=SampleDataKeys.sign_only_key.replace(' ', ''),
            email=self.account.preferredemail
        )

        resp = self.client.post(
            reverse('gpg_keys'),
            data={'cancel_pending': '', 'tokens': [token1.id, token2.id]},
            follow=True
        )

        self.assertRedirects(resp, reverse('gpg_keys'))
        self.assert_response_includes_message(
            resp,
            gpg_messages.pending_tokens_cancelled(
                [SampleDataKeys.regular_key, SampleDataKeys.sign_only_key]))
        self.assertNotContains(resp, "Keys pending validation")
        self.assertRaises(
            AuthToken.DoesNotExist,
            AuthToken.get_token,
            fingerprint=SampleDataKeys.regular_key.replace(' ', '')
        )
        self.assertRaises(
            AuthToken.DoesNotExist,
            AuthToken.get_token,
            fingerprint=SampleDataKeys.sign_only_key.replace(' ', '')
        )

    @switches(GPG_SERVICE=True)
    def test_active_keys_not_shown_with_no_active_keys(self):
        self.create_user_and_login()
        resp = self.client.get(reverse('gpg_keys'))
        self.assertNotContains(resp, "Your Enabled Keys")

    @switches(GPG_SERVICE=True)
    def test_active_keys_shown_with_pending_tokens(self):
        self.create_user_and_login()
        self.factory.add_gpgkey_to_account(self.account)
        resp = self.client.get(reverse('gpg_keys'))
        self.assertContains(resp, "Your Enabled Keys")

    @switches(GPG_SERVICE=True)
    def test_can_deactivate_active_key(self):
        self.create_user_and_login()
        fingerprint = SampleDataKeys.regular_key.replace(' ', '')
        self.factory.add_gpgkey_to_account(
            self.account, fingerprint=fingerprint)

        resp = self.client.post(
            reverse('gpg_keys'),
            data={'disable_keys': '', 'keys': [fingerprint]},
            follow=True
        )

        self.assertRedirects(resp, reverse('gpg_keys'))
        self.assert_response_includes_message(
            resp, gpg_messages.keys_disabled([fingerprint]))
        self.assert_account_has_disabled_key(fingerprint)

    @switches(GPG_SERVICE=True)
    def test_deactivate_form_errors_are_rendered(self):
        self.create_user_and_login()
        fingerprint = self.factory.add_gpgkey_to_account(self.account)
        # increment our fingerprint value by one:
        bad_fingerprint = ":040x".format(int(fingerprint, 16) + 1)
        resp = self.client.post(
            reverse('gpg_keys'),
            data={'disable_keys': '', 'keys': [bad_fingerprint]},
        )

        expected_error = (
            'Select a valid choice. %s is not one of the available choices.'
            % bad_fingerprint
        )

        self.assertFormError(resp, 'disable_keys_form', 'keys', expected_error)

    @switches(GPG_SERVICE=True)
    def test_can_disable_multiple_keys_at_once(self):
        self.create_user_and_login()
        fingerprint1 = self.factory.add_gpgkey_to_account(self.account)
        fingerprint2 = self.factory.add_gpgkey_to_account(self.account)
        resp = self.client.post(
            reverse('gpg_keys'),
            data={'disable_keys': '', 'keys': [fingerprint1, fingerprint2]},
            follow=True
        )

        self.assertRedirects(resp, reverse('gpg_keys'))
        self.assert_response_includes_message(
            resp,
            gpg_messages.keys_disabled([fingerprint1, fingerprint2]))
        self.assertNotContains(resp, "Your Enabled Keys")
        self.assert_account_has_disabled_key(fingerprint1)
        self.assert_account_has_disabled_key(fingerprint2)

    @switches(GPG_SERVICE=True)
    def test_disabled_keys_not_shown_with_no_disabled_keys(self):
        self.create_user_and_login()
        resp = self.client.get(reverse('gpg_keys'))
        self.assertNotContains(resp, "Disabled Keys")

    @switches(GPG_SERVICE=True)
    def test_disabled_keys_shown_with_disabled_keys(self):
        self.create_user_and_login()
        self.factory.add_gpgkey_to_account(self.account, enabled=False)
        resp = self.client.get(reverse('gpg_keys'))
        self.assertContains(resp, "Disabled Keys")

    @switches(GPG_SERVICE=True)
    def test_can_reactivate_disabled_encryption_key(self):
        self.create_user_and_login()
        fingerprint = SampleDataKeys.regular_key.replace(' ', '')
        self.factory.add_gpgkey_to_account(
            self.account, fingerprint=fingerprint, enabled=False)

        resp = self.client.post(
            reverse('gpg_keys'),
            data={'reenable_keys': '', 'keys': [fingerprint]},
            follow=True
        )
        self.assertRedirects(resp, reverse('gpg_keys'))
        self.assert_response_includes_message(
            resp,
            gpg_messages.keys_reenabled(
                [fingerprint],
                self.account.preferredemail.email))
        self.assert_account_has_pending_gpg_token(
            fingerprint, AuthTokenType.VALIDATEGPG)

    @switches(GPG_SERVICE=True)
    def test_can_reactivate_disabled_sign_only_key(self):
        self.create_user_and_login()
        fingerprint = SampleDataKeys.sign_only_key.replace(' ', '')
        self.factory.add_gpgkey_to_account(
            self.account, fingerprint=fingerprint, enabled=False)

        resp = self.client.post(
            reverse('gpg_keys'),
            data={'reenable_keys': '', 'keys': [fingerprint]},
            follow=True
        )
        self.assertRedirects(resp, reverse('gpg_keys'))
        self.assert_response_includes_message(
            resp,
            gpg_messages.keys_reenabled(
                [fingerprint],
                self.account.preferredemail.email))
        self.assert_account_has_pending_gpg_token(
            fingerprint,
            AuthTokenType.VALIDATESIGNONLYGPG,
        )

    @switches(GPG_SERVICE=True)
    def test_reactivate_form_errors_are_rendered(self):
        self.create_user_and_login()
        fingerprint = self.factory.add_gpgkey_to_account(
            self.account, enabled=False)
        # increment our fingerprint value by one:
        bad_fingerprint = ":040x".format(int(fingerprint, 16) + 1)
        resp = self.client.post(
            reverse('gpg_keys'),
            data={'reenable_keys': '', 'keys': [bad_fingerprint]},
        )

        expected_error = (
            'Select a valid choice. %s is not one of the available choices.'
            % bad_fingerprint
        )
        self.assertFormError(
            resp, 'reenable_keys_form', 'keys', expected_error)

    @switches(GPG_SERVICE=True)
    def test_can_reactivate_multiple_keys_at_once(self):
        self.create_user_and_login()
        fingerprint1 = self.factory.add_gpgkey_to_account(
            self.account,
            fingerprint=SampleDataKeys.regular_key,
            enabled=False
        )
        fingerprint2 = self.factory.add_gpgkey_to_account(
            self.account,
            fingerprint=SampleDataKeys.sign_only_key,
            enabled=False
        )
        resp = self.client.post(
            reverse('gpg_keys'),
            data={'reenable_keys': '', 'keys': [fingerprint1, fingerprint2]},
            follow=True
        )

        self.assertRedirects(resp, reverse('gpg_keys'))
        self.assert_response_includes_message(
            resp,
            gpg_messages.keys_reenabled(
                [fingerprint1, fingerprint2],
                self.account.preferredemail.email))
        self.assert_account_has_pending_gpg_token(
            fingerprint1, AuthTokenType.VALIDATEGPG)
        self.assert_account_has_pending_gpg_token(
            fingerprint2, AuthTokenType.VALIDATESIGNONLYGPG)

    @switches(GPG_SERVICE=True)
    def test_forms_use_different_label_ids(self):
        self.create_user_and_login()
        self.factory.add_gpgkey_to_account(
            self.account,
            fingerprint=SampleDataKeys.regular_key,
            enabled=False
        )
        self.factory.add_gpgkey_to_account(
            self.account,
            fingerprint=SampleDataKeys.sign_only_key,
            enabled=True
        )

        resp = self.client.get(reverse('gpg_keys'))
        doc = BeautifulSoup.BeautifulSoup(resp.content)
        seen_ids = set()
        all_forms = doc.findAll('form')

        for form in all_forms:
            self.assert_form_labels_not_seen_before(form, seen_ids)
