# Copyright 2010 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

from __future__ import unicode_literals

from django.core.urlresolvers import reverse
from gargoyle.testutils import switches

from identityprovider.tests import DEFAULT_USER_PASSWORD
from identityprovider.tests.utils import SSOBaseTestCase
from webui.views.account import ssh_keys


class SSHViewFeatureFlagTestCase(SSOBaseTestCase):

    def create_user_and_login(self, name='person-2'):
        self.account = self.factory.make_account(
            password=DEFAULT_USER_PASSWORD,
            email='example_user@example.com')
        assert self.client.login(
            username=self.account.preferredemail.email,
            password=DEFAULT_USER_PASSWORD)

    @switches(SSH_KEY_INTEGRATION=False)
    def test_ssh_keys_menu_item_not_rendered_without_feature_flag(self):
        self.create_user_and_login()
        resp = self.client.get(reverse('account-index'))

        self.assertNotContains(resp, '<span>SSH Keys</span>', html=True)

    @switches(SSH_KEY_INTEGRATION=True)
    def test_ssh_keys_menu_item_is_rendered_with_feature_flag(self):
        self.create_user_and_login()
        resp = self.client.get(reverse('account-index'))

        self.assertContains(resp, 'SSH Keys')

    @switches(SSH_KEY_INTEGRATION=False)
    def test_ssh_keys_view_returns_404_without_feature_flag(self):
        self.create_user_and_login()
        resp = self.client.get(reverse('ssh_keys'))

        self.assertEqual(404, resp.status_code)

    @switches(SSH_KEY_INTEGRATION=True)
    def test_ssh_keys_view_returns_200_with_feature_flag(self):
        self.create_user_and_login()
        resp = self.client.get(reverse('ssh_keys'))

        self.assertEqual(200, resp.status_code)
        self.assertTemplateUsed(resp, 'account/ssh_keys.html')

    def test_ssh_keys_view_returns_200_with_feature_flag_conditional(self):
        self.create_user_and_login()
        condition_set = ('identityprovider.gargoyle_conditions.'
                         'AccountConditionSet(identityprovider.account)')
        self.conditionally_enable_flag(
            'SSH_KEY_INTEGRATION',
            'email',
            self.account.preferredemail.email,
            condition_set
        )
        resp = self.client.get(reverse('ssh_keys'))

        self.assertEqual(200, resp.status_code)
        self.assertTemplateUsed(resp, 'account/ssh_keys.html')

    def test_ssh_keys_view_returns_404_without_feature_flag_conditional(self):
        self.create_user_and_login()
        # Set a conditional flag for a second account to verify that it really
        # is conditional:
        special_account = self.factory.make_account()
        condition_set = ('identityprovider.gargoyle_conditions.'
                         'AccountConditionSet(identityprovider.account)')
        self.conditionally_enable_flag(
            'SSH_KEY_INTEGRATION',
            'email',
            special_account.preferredemail.email,
            condition_set
        )
        resp = self.client.get(reverse('ssh_keys'))

        self.assertEqual(404, resp.status_code)

    @switches(SSH_KEY_INTEGRATION=True)
    def test_ssh_keys_view_requires_authentication(self):
        self.preseed_cookie_check()
        resp = self.client.get(reverse('ssh_keys'))

        self.assertRedirects(
            resp,
            reverse('login') + '?next=' + reverse('ssh_keys'),
        )

    @switches(SSH_KEY_INTEGRATION=True)
    def test_ssh_keys_view_disabled_in_readonly_mode(self):
        account = self.factory.make_account()
        with self.settings(READ_ONLY_MODE=True):
            request = self.factory.make_request(user=account)
            resp = ssh_keys(request)

        self.assertContains(
            resp,
            "<h1 class='main'>We're currently in read-only mode.</h1>",
            status_code=403,
            html=True)
