# encoding: utf-8
# Copyright 2015 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

from __future__ import unicode_literals

from django.conf import settings
from django.core.urlresolvers import reverse
from django.utils.timezone import now, timedelta

from identityprovider.tests.utils import AuthenticatedTestCase
from identityprovider.models.const import AuthLogType


class AuthLogTestCase(AuthenticatedTestCase):

    def test_auth_log_page_empty(self):
        r = self.client.get(reverse('auth_log'))
        assert len(r.context['auth_log']) == 0
        self.assertTemplateUsed(r, 'account/auth_log.html')
        self.assertContains(r, "has no activity yet")

    def test_auth_log_page_with_auth_records(self):
        self.account.authlog_set.create(
            when_created=now(), user_agent="Godzilla 1.0",
            remote_ip='192.168.100.200', referer='http://example.com',
            login_email='', log_type=AuthLogType.OPENID_NEW)
        r = self.client.get(reverse('auth_log'))
        assert len(r.context['auth_log']) != 0
        self.assertTemplateUsed(r, 'account/auth_log.html')
        self.assertContains(r, "Time/date")
        self.assertContains(r, "Recent accesses")

    def test_auth_log_page_limit_per_page_auth_records(self):
        base_time = now() - timedelta(days=5)
        records_per_page = settings.SHOW_USER_NUM_AUTHLOGS
        ok_log_types = (
            AuthLogType.OPENID_NEW,
            AuthLogType.OPENID_EXISTING,
            AuthLogType.API_NEW,
            AuthLogType.API_EXISTING,
        )
        for count in range(records_per_page + 10):
            log_type = ok_log_types[count % len(ok_log_types)]
            access_time = base_time + timedelta(minutes=5 * count)
            self.account.authlog_set.create(
                when_created=access_time, user_agent="Godzilla 1.0",
                remote_ip='192.168.100.200', referer='http://example.com',
                login_email='', log_type=log_type)

        r = self.client.get(reverse('auth_log'))
        assert len(r.context['auth_log']) == records_per_page
        # Verify descending order
        self.assertGreater(
            r.context['auth_log'][0].when_created,
            r.context['auth_log'][1].when_created)
        self.assertContains(
            r, "The %s most recent accesses" % records_per_page)

    def test_auth_log_page_doesnt_show_failed_auth_records(self):
        # This is somewhat artificial because FAIL records are not usually
        # linked to an account.
        for log_type in (AuthLogType.OPENID_FAIL, AuthLogType.API_FAIL):
            self.account.authlog_set.create(
                when_created=now(), user_agent="Godzilla 1.0",
                remote_ip='192.168.100.200', referer='http://example.com',
                login_email='', log_type=log_type)
        r = self.client.get(reverse('auth_log'))
        assert len(r.context['auth_log']) == 0
        self.assertTemplateUsed(r, 'account/auth_log.html')
        self.assertContains(r, "has no activity yet")
