# encoding: utf-8
# Copyright 2010 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

from __future__ import unicode_literals

from urllib import quote_plus

from django.core.urlresolvers import reverse
from django.test.utils import override_settings
from openid import fetchers
from openid.consumer import consumer
from openid.consumer.discover import OpenIDServiceEndpoint, OPENID_2_0_TYPE
from openid.message import (OPENID1_URL_LIMIT, IDENTIFIER_SELECT)
from openid.yadis.constants import YADIS_CONTENT_TYPE

from identityprovider.tests.utils import SSOBaseTestCase
from identityprovider.views import server

import webui.views.consumer

from webui.views.consumer import (
    SREG_OPTIONAL,
    SREG_REQUIRED,
    to_regular_dict,
    render_index_page,
)


class MockFetcher(fetchers.Urllib2Fetcher):

    called = False

    def fetch(self, url, body=None, headers=None):
        MockFetcher.called = True
        return super(MockFetcher, self).fetch(url, body, headers)


class BaseTestCase(SSOBaseTestCase):

    def setUp(self):
        super(BaseTestCase, self).setUp()
        fetcher_module = webui.views.consumer.fetchers
        old_fetcher = fetcher_module.getDefaultFetcher().fetcher
        self.mock = MockFetcher()
        fetcher_module.setDefaultFetcher(self.mock)
        self.addCleanup(fetcher_module.setDefaultFetcher, old_fetcher)

        self.openid_data = {
            'openid_identifier': "http://localhost/+id/abcd123",
        }

    def assert_message_in_response(self, response, msg):
        for m in response.context['messages']:
            if msg in m.message:
                break  # the else block will not be executed
        else:
            self.fail('Message %r is not present in the given response.' % msg)


class FetcherTestCase(BaseTestCase):
    def test_check_custom_fetcher_called(self):
        self.client.post(reverse('start_open_id'), data=self.openid_data)
        self.assertTrue(MockFetcher.called)


class RenderIndexPageTestCase(BaseTestCase):

    def test_render_index_page_utf8(self):
        fullname = u'Dūmmẏ Ũsèṙ'.encode('utf-8')
        request = self.factory.make_request()

        template_args = {'sreg': [('fullname', fullname)],
                         'url': 'http://localhost/'}
        response = render_index_page(request, **template_args)
        headers = dict(response.items())
        self.assertEqual(headers['Content-Type'], 'text/html; charset=utf-8')
        self.assertTrue(fullname in response.content)


# make sure to enable the consumer views
@override_settings(TESTING=True)
class ConsumerTestCase(BaseTestCase):

    def setUp(self):
        super(ConsumerTestCase, self).setUp()
        self.endpoint = OpenIDServiceEndpoint()
        self.endpoint.claimed_id = 'oid'
        self.endpoint.server_url = 'http://localhost/'

    def test_start_open_id_get(self):
        # without faking the Consumer, an openid request returns error
        r = self.client.post('/consumer/', self.openid_data)
        self.assertEqual(r.status_code, 200)
        self.assertTemplateUsed(r, 'consumer/index.html')
        self.assert_message_in_response(r, 'OpenID discovery error')

    def test_consumer_view_when_not_in_debug(self):
        with self.settings(DEBUG=False):
            response = self.client.get(reverse('start_open_id'))
            self.assertEqual(response.status_code, 200)

    def test_consumer_view_when_not_in_testing(self):
        with self.settings(TESTING=False):
            response = self.client.get(reverse('start_open_id'))
            self.assertEqual(response.status_code, 404)

    def test_finish_open_id_when_not_in_testing(self):
        with self.settings(TESTING=False):
            response = self.client.get(reverse('finish_open_id'))
            self.assertEqual(response.status_code, 404)

    def test_rpXRDS_when_not_in_testing(self):
        with self.settings(TESTING=False):
            response = self.client.get(reverse('rp_xrds'))
            self.assertEqual(response.status_code, 404)


class ConsumerFakedTestCase(ConsumerTestCase):

    def setUp(self):
        super(ConsumerFakedTestCase, self).setUp()

        class MockConsumer(consumer.Consumer):
            def begin(this, url):
                auth_request = consumer.AuthRequest(self.endpoint, None)
                return auth_request

        self.patch(consumer, 'Consumer', new=MockConsumer)

    def test_to_regular_dict(self):
        self.assertEqual(to_regular_dict(self.openid_data), self.openid_data)

    def test_start_open_id_get(self):
        r = self.client.get('/consumer/')
        self.assertEqual(r.status_code, 200)
        self.assertTemplateUsed(r, 'consumer/index.html')

    def test_start_open_id_immediate(self):
        self.openid_data.update({'mode': 'immediate'})
        r = self.client.post('/consumer/', self.openid_data)
        query = self.get_query(r)
        self.assertEqual(r.status_code, 302)
        self.assertEqual(query['openid.mode'], 'checkid_immediate')
        self.assertEqual(query['openid.return_to'],
                         quote_plus('http://testserver/consumer/finish/'))

    def test_start_open_id_setup(self):
        self.openid_data.update({'mode': 'setup'})
        r = self.client.post('/consumer/', self.openid_data)
        query = self.get_query(r)
        self.assertEqual(r.status_code, 302)
        self.assertEqual(query['openid.mode'], 'checkid_setup')
        self.assertEqual(query['openid.return_to'],
                         quote_plus('http://testserver/consumer/finish/'))

    def test_start_open_id_sreg(self):
        self.openid_data.update(
            {'mode': 'setup', 'sreg': 'yes', 'sreg_fullname': SREG_REQUIRED,
             'sreg_email': SREG_REQUIRED, 'sreg_nickname': SREG_OPTIONAL})
        r = self.client.post('/consumer/', self.openid_data)
        query = self.get_query(r)
        self.assertEqual(r.status_code, 302)
        self.assertEqual(query['openid.sreg.required'],
                         quote_plus('fullname,email'))
        self.assertEqual(query['openid.sreg.optional'], 'nickname')

    def test_start_open_id_teams(self):
        request_teams = 'team1,team2,team3'
        self.openid_data.update({
            'mode': 'setup',
            'teams': 'yes',
            'request_teams': request_teams
        })
        r = self.client.post('/consumer/', self.openid_data)
        query = self.get_query(r)
        self.assertEqual(r.status_code, 302)
        self.assertEqual(query['openid.lp.query_membership'],
                         quote_plus(request_teams))

    def test_start_open_id_should_redirect(self):
        self.openid_data['mode'] = 'setup'
        r = self.client.post('/consumer/', self.openid_data)
        query = self.get_query(r)
        self.assertEqual(r.status_code, 302)
        self.assertEqual(query['openid.trust_root'],
                         quote_plus('http://testserver/consumer/'))
        self.assertEqual(query['openid.return_to'],
                         quote_plus('http://testserver/consumer/finish/'))

    def test_start_open_id_render_template(self):
        # replace consumer with OpenID 2.0 compliant one
        class MockConsumer(consumer.Consumer):
            def begin(self, url):
                endpoint = OpenIDServiceEndpoint()
                endpoint.claimed_id = 'oid'
                endpoint.server_url = 'http://localhost/'
                endpoint.type_uris = [OPENID_2_0_TYPE]
                auth_request = consumer.AuthRequest(endpoint, None)
                return auth_request
        consumer.Consumer = MockConsumer

        self.openid_data['mode'] = 'setup'
        r = self.client.post('/consumer/', self.openid_data)
        self.assertEqual(r.status_code, 200)
        self.assertTemplateUsed(r, 'consumer/request_form.html')
        self.assertTrue('openid_message'in r.context['html'])

    def test_finish_open_id_get(self):
        r = self.client.get('/consumer/finish/')
        self.assertEqual(r.status_code, 200)
        self.assertTemplateUsed(r, 'consumer/index.html')

    def test_finish_open_id_post(self):
        r = self.client.post('/consumer/finish/')
        self.assertEqual(r.status_code, 200)
        self.assertTemplateUsed(r, 'consumer/index.html')

    def test_finish_open_id_consumer_success(self):
        def mock_complete(this, request_args, return_to):
            request = {'openid.mode': 'checkid_setup',
                       'openid.trust_root': 'http://localhost/',
                       'openid.return_to': 'http://localhost/',
                       'openid.identity': IDENTIFIER_SELECT,
                       'openid.sreg.nickname': 'mynickname',
                       'openid.lp.is_member': 'myteam'}
            openid_server = server._get_openid_server(request)
            orequest = openid_server.decodeRequest(request)
            response = consumer.SuccessResponse(
                self.endpoint, orequest.message,
                signed_fields=['openid.sreg.nickname',
                               'openid.lp.is_member'])
            return response
        old_complete = consumer.Consumer.complete
        consumer.Consumer.complete = mock_complete

        r = self.client.post('/consumer/finish/', self.openid_data)
        self.assertEqual(r.context['url'], 'oid')
        self.assertEqual(r.context['sreg'], [('nickname', 'mynickname')])
        self.assertEqual(r.context['teams'], ['myteam'])

        consumer.Consumer.complete = old_complete

    def test_finish_open_id_consumer_cancel(self):

        def mock_complete(this, request_args, return_to):
            response = consumer.CancelResponse(self.endpoint)
            return response

        self.patch(consumer.Consumer, 'complete', new=mock_complete)

        r = self.client.post('/consumer/finish/', self.openid_data)
        self.assert_message_in_response(r, 'OpenID authentication cancelled.')

    def test_finish_open_id_consumer_failure(self):
        def mock_complete(this, request_args, return_to):
            response = consumer.FailureResponse(self.endpoint,
                                                message='some error')
            return response
        old_complete = consumer.Consumer.complete
        consumer.Consumer.complete = mock_complete

        r = self.client.post('/consumer/finish/', self.openid_data)
        self.assert_message_in_response(r, 'OpenID authentication failed.')
        self.assert_message_in_response(r, 'some error')

        consumer.Consumer.complete = old_complete

    def test_rpXRDS(self):
        r = self.client.get('/consumer/xrds/')
        self.assertTemplateUsed(r, 'server/openidapplication-xrds.xml')
        self.assertEqual(r['Content-Type'], YADIS_CONTENT_TYPE)

    def test_long_query_string(self):
        self.openid_data['mode'] = 'setup'
        self.openid_data['forcelongurl'] = '1'
        r = self.client.post('/consumer/', self.openid_data)
        query = self.get_query(r)
        self.assertTrue(len(query['openid.return_to']) > OPENID1_URL_LIMIT)
