# Copyright 2010-2015 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

from __future__ import unicode_literals

from urllib import quote as urlquote
from base64 import b16encode, b32encode

import mock

from django.conf import settings
from django.core.urlresolvers import reverse
from django.db import transaction
from django.test.utils import override_settings
from django.utils.timezone import now
from gargoyle.testutils import switches
from oath import hotp
from pyquery import PyQuery

from identityprovider.models import Account, AuthenticationDevice, twofactor
from identityprovider.models.const import AccountStatus
from identityprovider.tests import DEFAULT_USER_PASSWORD
from identityprovider.tests.utils import (
    MISSING_BACKUP_DEVICE,
    AuthenticatedTestCase,
    SSOBaseTransactionTestCase,
    assert_exausted_warning,
    test_concurrently,
)

from webui.decorators import ratelimit_extra
from webui.views.ui import TwoFactorView
from webui.views.devices import (
    DEVICE_DELETED,
    DEVICE_GENERATION_WARNING,
    generate_key,
    get_unique_device_name_for_user,
)

# This is the example OATH/HOTP key from RFC 4226:
# http://www.ietf.org/rfc/rfc4226.txt
KEY = '3132333435363738393031323334353637383930'


class DeviceViewsTestCaseBase(AuthenticatedTestCase):

    def setUp(self):
        super(DeviceViewsTestCaseBase, self).setUp()
        self.switch_twofactor = switches(TWOFACTOR=True)
        self.switch_twofactor.patch()
        self.addCleanup(self.switch_twofactor.unpatch)

    def do_2fa_login(self, device):
        oath_token = self.factory.make_next_otp(device)
        response = self.client.post(
            reverse('twofactor'), data=dict(oath_token=oath_token))
        self.assertRedirects(response, reverse('account-index'))
        return response

    def assert_twofactor_disabled(self, url):
        with switches(TWOFACTOR=False):
            response = self.client.get(url)
        self.assertEqual(response.status_code, 404)
        return response

    def assert_require_twofactor_upgrade(self, url, **kwargs):
        assert self.client.session.get(twofactor.TWOFACTOR_LOGIN)
        # identityprovider.models.twofactor.is_fresh == True
        # identityprovider.models.twofactor.is_upgraded == False
        with self.settings(TWOFACTOR_FRESHNESS=10000, TWOFACTOR_TTL=0):
            response = self.client.post(url, **kwargs)

        self.assertRedirects(
            response, reverse('twofactor') + '?next=' + url)
        return response

    def assert_require_twofactor_fresh(self, url, **kwargs):
        assert self.client.session.get(twofactor.TWOFACTOR_LOGIN)
        # identityprovider.models.twofactor.is_fresh == False
        # identityprovider.models.twofactor.is_upgraded == True
        with self.settings(TWOFACTOR_FRESHNESS=0, TWOFACTOR_TTL=10000):
            response = self.client.post(url, **kwargs)
        self.assertRedirects(response, reverse('twofactor') + '?next=' + url)
        return response


class DeviceListViewTestCase(DeviceViewsTestCaseBase):

    url = reverse('device-list')

    def test_twofactor_disabled(self):
        self.assert_twofactor_disabled(self.url)

    def test_list_sorted(self):
        self.factory.make_device(account=self.account, name='foo')
        # create second device
        self.factory.make_device(account=self.account, name='bar')

        response = self.client.get(self.url)
        content = response.content
        idx_device = content.find('foo')
        idx_device2 = content.find('bar')
        self.assertNotEqual(idx_device, -1)
        self.assertNotEqual(idx_device2, -1)
        self.assertTrue(idx_device < idx_device2)

    def assert_backup_device_warning(self, show_warning):
        response = self.client.get(self.url)
        backup_warning = MISSING_BACKUP_DEVICE.format(
            add_device_link=reverse('device-addition'))

        if show_warning:
            self.assertContains(response, backup_warning)
        else:
            self.assertNotContains(response, backup_warning)

    def test_backup_device_warning_no_devices(self):
        assert self.account.warn_about_backup_device
        assert self.account.devices.count() == 0

        self.assert_backup_device_warning(show_warning=False)

    def test_backup_device_warning_one_device(self):
        self.factory.make_device(account=self.account)
        assert self.account.warn_about_backup_device
        assert self.account.devices.count() == 1

        self.assert_backup_device_warning(show_warning=True)

    def test_backup_device_warning_one_device_no_warning(self):
        self.factory.make_device(account=self.account)
        self.account.warn_about_backup_device = False
        self.account.save()

        assert not self.account.warn_about_backup_device
        assert self.account.devices.count() == 1
        self.assert_backup_device_warning(show_warning=False)

    def test_backup_device_warning_two_devices(self):
        self.factory.make_device(account=self.account)
        self.factory.make_device(account=self.account)

        assert self.account.warn_about_backup_device
        assert self.account.devices.count() == 2
        self.assert_backup_device_warning(show_warning=False)

    def test_codes_exhausted_warning_not_exhausted(self):
        device = self.factory.make_device(
            account=self.account, device_type='paper')

        response = self.client.get(self.url)
        with self.assertRaises(AssertionError):
            assert_exausted_warning(self, [device], response)

    def test_codes_exhausted_warning_exhausted(self):
        counter = (settings.TWOFACTOR_PAPER_CODES -
                   settings.TWOFACTOR_PAPER_CODES_WARN_RENEWAL + 1)
        device = self.factory.make_device(
            account=self.account, device_type='paper', counter=counter)

        response = self.client.get(self.url)
        assert_exausted_warning(self, [device], response)

    def test_device_list_no_devices(self):
        assert self.account.devices.count() == 0
        assert not self.account.twofactor_required

        response = self.client.get(self.url)
        self.assertEqual(response.status_code, 200)

        tree = PyQuery(response.content)

        devices = tree.find('table#device-list')
        self.assertEqual(devices, [])

    def test_device_list_one_device(self):
        self.account.twofactor_required = True
        self.account.save()

        device = self.factory.make_device(account=self.account)
        self.do_2fa_login(device)

        response = self.client.get(self.url)
        self.assertEqual(response.status_code, 200)

        tree = PyQuery(response.content)

        rows = tree.find('table#device-list').find('tbody tr')
        self.assertEqual(len(rows), 1)
        name, command = rows.find('td')
        links = command.findall('a')
        self.assertEqual(len(links), 2)
        self.assertEqual(
            links[1].get('href'),
            reverse('device-removal', kwargs=dict(device_id=device.id)))
        self.assertEqual(name.text, device.name)


class DeviceRenameViewTestCase(DeviceViewsTestCaseBase):

    def setUp(self):
        super(DeviceRenameViewTestCase, self).setUp()
        self.device = self.factory.make_device(account=self.account)
        self.url = reverse('device-rename', args=[self.device.id])

        self.do_2fa_login(self.device)

    def test_device_rename_get(self):
        response = self.client.get(self.url)

        tree = PyQuery(response.content)
        name_input = tree.find('input[name="name"]')
        self.assertEqual(name_input[0].value, self.device.name)

    def test_device_rename_post(self):
        response = self.client.post(self.url, {'name': 'bar'})

        self.assertRedirects(response, reverse('device-list'))

        device = AuthenticationDevice.objects.get(id=self.device.id)
        self.assertEqual(device.name, 'bar')

    def test_device_rename_with_invalid_user(self):
        self.client.logout()

        new_account = self.factory.make_account()
        device = self.factory.make_device(account=new_account)
        self.client.login(username=new_account.preferredemail.email,
                          password=DEFAULT_USER_PASSWORD)
        self.do_2fa_login(device)

        # get as new user
        response = self.client.get(self.url)
        self.assertEqual(response.status_code, 404)

        # post as new user
        response = self.client.post(self.url, {'name': 'bar'})
        self.assertEqual(response.status_code, 404)

    def test_require_twofactor_upgrade(self):
        self.assert_require_twofactor_upgrade(self.url, data={'name': 'bar'})

    def test_require_twofactor_fresh(self):
        self.assert_require_twofactor_fresh(self.url)


class DeviceRemovalViewTestCase(DeviceViewsTestCaseBase):

    def setUp(self):
        super(DeviceRemovalViewTestCase, self).setUp()
        self.device = self.factory.make_device(account=self.account)
        # ensure 2fa session cookie is set
        self.do_2fa_login(self.device)

    def url(self, device=None):
        if device is None:
            device = self.device
        return reverse('device-removal', kwargs=dict(device_id=device.id))

    def test_twofactor_disabled(self):
        self.assert_twofactor_disabled(self.url())

    def test_require_twofactor_fresh(self):
        self.assert_require_twofactor_fresh(self.url())

    def test_device_not_found(self):
        url = self.url(self.device)

        self.device.delete()
        assert self.account.devices.count() == 0

        response = self.client.post(url)
        self.assertEqual(response.status_code, 404)

    def test_device_owned_by_another_user(self):
        other = self.factory.make_device()
        assert other.account != self.account

        response = self.client.post(self.url(other))
        self.assertEqual(response.status_code, 404)

    def test_get(self):
        response = self.client.get(self.url())
        self.assertContains(response, 'Are you sure')

        # check that a GET does not delete the device
        self.assertEqual(
            self.account.devices.get(id=self.device.id), self.device)

        tree = PyQuery(response.content)
        self.assertEqual(tree.find(
            '[data-qa-id="device_removal_device_name"]').text(),
            self.device.name)

    def test_post(self):
        response = self.client.post(self.url(), follow=True)

        self.assertRedirects(response, reverse('device-list'), status_code=303)
        self.assertEqual(self.account.devices.all().count(), 0)
        self.assertEqual(
            [m.message for m in response.context['messages']],
            [DEVICE_DELETED.format(name=self.device.name)])

    def test_preference_reset(self):
        self.account.twofactor_required = True
        self.account.save()

        device2 = self.factory.make_device(
            account=self.account, name='Another device')

        self.client.post(self.url())
        account = Account.objects.get(id=self.account.id)
        self.assertTrue(account.twofactor_required)

        self.client.post(self.url(device2))
        account = Account.objects.get(id=self.account.id)
        self.assertFalse(account.twofactor_required)

    def test_require_twofactor_upgrade(self):
        url = self.url()
        self.assert_require_twofactor_upgrade(url)

    def test_last_device_removal_unauthenticates(self):
        assert self.client.session[twofactor.TWOFACTOR_LOGIN] < now()
        self.client.post(self.url())
        self.assertNotIn(twofactor.TWOFACTOR_LOGIN, self.client.session)


class DeviceAdditionViewTestCase(DeviceViewsTestCaseBase):

    BASIC_TYPES = ['google', 'yubi', 'generic', 'paper']
    url = reverse('device-addition')

    def assert_successful_post(
            self, name='Some device', counter=1, device_type='generic'):

        data = {'type': device_type, 'name': name}
        if device_type != 'pape':
            data.update({'otp': '755224', 'hex_key': KEY})
        before = now()
        response = self.client.post(self.url, data=data)
        after = now()

        devices = self.account.devices.filter(
            name=name, account=self.account, counter=counter,
            device_type=device_type)
        self.assertEqual(devices.count(), 1)
        device = devices[0]

        if device_type == 'paper':
            redirect_to = reverse(
                'device-print', kwargs=dict(device_id=device.id))
        else:
            redirect_to = reverse('device-list')
        self.assertRedirects(response, redirect_to, status_code=303)

        self.assertLess(before, self.client.session[twofactor.TWOFACTOR_LOGIN])
        self.assertLess(self.client.session[twofactor.TWOFACTOR_LOGIN], after)

        return response

    def test_device_types(self):
        response = self.client.get(self.url)

        tree = PyQuery(response.content)
        type_nodes = tree.find('input[type="radio"]')

        # Note: can't just use node.value as it is None for unset radios
        self.assertEqual([dict(node.items())['value'] for node in type_nodes],
                         self.BASIC_TYPES)

    def test_initial_post(self):
        response = self.client.post(self.url, data={'type': 'generic'})

        self.assertEqual(response.status_code, 200)
        self.assertEqual(self.account.devices.count(), 0)

        tree = PyQuery(response.content)
        self.assertEqual(len(tree.find('.error')), 0)
        device_name = tree.find('input[type=text][name=name]')
        self.assertEqual(len(device_name), 1)
        self.assertEqual(device_name.val(), 'Authentication device')

    def test_initial_post_google(self):
        email = 'email@example.com'
        self.account.preferredemail.email = email
        self.account.preferredemail.save()

        with mock.patch('webui.views.devices.generate_key', return_value=KEY):
            response = self.client.post(self.url, data={'type': 'google'})

        self.assertEqual(response.status_code, 200)
        self.assertEqual(self.account.devices.count(), 0)

        tree = PyQuery(response.content)
        self.assertEqual(len(tree.find('.error')), 0)
        device_name = tree.find('input[type=text][name=name]')
        self.assertEqual(len(device_name), 1)
        self.assertEqual(device_name.val(), 'Google Authenticator')

        qrcode = tree.find('img.qrcode')
        self.assertEqual(len(qrcode), 1)
        url = qrcode[0].attrib['src']
        # correct https domain
        self.assertIn('https://chart.googleapis.com', url)
        self.assertIn(urlquote(email), url)
        self.assertIn(b32encode(KEY.decode('hex')), url)

    def test_initial_post_yubikey(self):
        with mock.patch('webui.views.devices.generate_key', return_value=KEY):
            response = self.client.post(self.url, data={'type': 'yubi'})

        self.assertEqual(response.status_code, 200)
        self.assertEqual(self.account.devices.count(), 0)

        tree = PyQuery(response.content)

        formatted_keys = tree.find('#formatted_key')
        self.assertEqual(len(formatted_keys), 1)
        formatted_key = formatted_keys[0].text.strip()
        self.assertEqual(formatted_key.replace(' ', ''), KEY)

        for key_id in ('short_press', 'long_press'):
            nodes = tree.find('#%s' % key_id)
            self.assertEqual(len(nodes), 1)
            command = nodes[0].text.strip()
            self.assertTrue(KEY.lower() in command)

    def test_successful_post_generic(self):
        self.assert_successful_post(device_type='generic')

    def test_successful_post_paper(self):
        self.assert_successful_post(
            name='Printable Backup Codes', counter=0, device_type='paper')

    def test_unrecognised_type(self):
        data = {'type': 'foobar', 'hex_key': KEY, 'name': 'Some device',
                'otp': 'some otp key'}
        response = self.client.post(self.url, data=data)

        self.assertEqual(response.status_code, 200)
        self.assertEqual(self.account.devices.count(), 0)

        tree = PyQuery(response.content)
        input_type = tree.find('input[type=radio][name=type][value=generic]')
        self.assertEqual(len(input_type), 1)

    def test_errors(self):
        data = {'type': 'generic', 'hex_key': KEY, 'name': '', 'otp': ''}
        response = self.client.post(self.url, data=data)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(self.account.devices.count(), 0)
        self.assertFormError(
            response, 'form', 'name', ['This field is required.'])
        self.assertFormError(
            response, 'form', 'otp', ['This field is required.'])

        tree = PyQuery(response.content)
        self.assertEqual(len(tree.find('.error')), 2)

    def test_add_second_device_unathenticated(self):
        self.assert_successful_post()
        assert self.account.has_twofactor_devices()

        self.assert_require_twofactor_upgrade(self.url, data={'type': 'paper'})
        self.assertEqual(self.account.devices.count(), 1)

    def test_add_second_device_authenticated(self):
        response = self.assert_successful_post()
        assert self.account.has_twofactor_devices()

        # webui.views.devices.is_upgraded == True
        with self.settings(TWOFACTOR_TTL=10000):
            response = self.assert_successful_post()

        self.assertEqual(self.account.devices.count(), 2)
        self.assertRedirects(response, reverse('device-list'), status_code=303)

    def test_require_twofactor_fresh(self):
        device = self.factory.make_device(self.account)
        assert self.account.has_twofactor_devices()
        self.do_2fa_login(device)

        self.assert_require_twofactor_fresh(self.url)


@override_settings(TWOFACTOR_PAPER_CODES=5)
class DevicePrintViewTestCase(DeviceViewsTestCaseBase):

    def setUp(self):
        super(DevicePrintViewTestCase, self).setUp()
        self.device = self.factory.make_device(
            account=self.account, key=KEY, counter=0, device_type='paper')
        self.print_url = reverse('device-print', args=[self.device.id])
        self.generate_url = reverse('device-generate', args=[self.device.id])

        # ensure 2fa session cookie is set
        self.do_2fa_login(self.device)

    def assert_generation_warning_and_button(self, counter):
        assert counter >= 0
        self.device.counter = counter
        self.device.save()

        response = self.client.get(self.print_url)

        doc = PyQuery(response.content)
        # check generate button is available
        generate_button = doc.find('a[href="%s"]' % self.generate_url)
        self.assertEqual(len(generate_button), 1)
        self.assertTrue(response.context_data['generation_enabled'])
        # check warning message is set
        self.assertEqual(
            [m.message for m in response.context['messages']],
            [DEVICE_GENERATION_WARNING])

    def print_with_counter(self, counter):
        self.device.counter = counter
        self.device.save()
        offset, position = divmod(counter, settings.TWOFACTOR_PAPER_CODES)
        offset *= settings.TWOFACTOR_PAPER_CODES

        response = self.client.get(self.print_url)

        doc = PyQuery(response.content)
        code_nodes = doc.find('#codes li')
        used = code_nodes[:position]
        unused = code_nodes[position:]

        self.assertEqual(len(used), position)
        self.assertEqual(len(unused),
                         settings.TWOFACTOR_PAPER_CODES - position)

        # check the first N are rendered with strike-through
        expected_codes = [
            hotp(self.device.key, i) for i in
            xrange(offset, offset + settings.TWOFACTOR_PAPER_CODES)]
        self.assertEqual(len(expected_codes), settings.TWOFACTOR_PAPER_CODES)
        for elem, code in zip(used, expected_codes[:position]):
            strike = elem.findall('strike')
            self.assertEqual(len(strike), 1)
            self.assertEqual(strike[0].text.strip(), code)

        # check the next N are not
        for elem, code in zip(unused, expected_codes[position:]):
            strike = elem.findall('strike')
            self.assertEqual(len(strike), 0)
            self.assertEqual(elem.text.strip(), code)

    def test_twofactor_disabled(self):
        self.assert_twofactor_disabled(self.print_url)
        self.assert_twofactor_disabled(self.generate_url)

    def test_print_codes_new_and_unused(self):
        self.print_with_counter(0)

    def test_print_codes_window_0(self):
        self.print_with_counter(settings.TWOFACTOR_PAPER_CODES - 2)

    def test_print_codes_window_1(self):
        self.print_with_counter(settings.TWOFACTOR_PAPER_CODES + 3)

    def test_print_codes_window_2(self):
        self.print_with_counter(settings.TWOFACTOR_PAPER_CODES * 2 + 3)

    def test_print_codes_window_boundry(self):
        self.print_with_counter(settings.TWOFACTOR_PAPER_CODES - 1)
        self.print_with_counter(settings.TWOFACTOR_PAPER_CODES)
        self.print_with_counter(settings.TWOFACTOR_PAPER_CODES + 1)

    def test_non_paper_404(self):
        self.device.device_type = 'generic'
        self.device.save()

        response = self.client.get(self.print_url)
        self.assertEqual(response.status_code, 404)

        response = self.client.get(self.generate_url)
        self.assertEqual(response.status_code, 404)

    def test_device_generate_get(self):
        # Be sure counter is between pages 4 and 5
        self.device.counter = settings.TWOFACTOR_PAPER_CODES * 3 + 2
        self.device.save()

        response = self.client.get(self.generate_url)

        doc = PyQuery(response.content)
        code_nodes = doc.find('#codes li')

        start = settings.TWOFACTOR_PAPER_CODES * 4
        end = settings.TWOFACTOR_PAPER_CODES * 5
        expected_codes = [
            hotp(self.device.key, i) for i in xrange(start, end)]
        codes = [node.text.strip() for node in code_nodes]
        self.assertEqual(codes, expected_codes)

    def test_device_generate_post_saves_and_redirects(self):
        # with no addition this time - testing the edge case
        self.device.counter = settings.TWOFACTOR_PAPER_CODES * 3
        self.device.save()

        response = self.client.post(self.generate_url)
        expected_redirect = reverse(
            'device-print', kwargs=dict(device_id=self.device.id))
        self.assertRedirects(response, expected_redirect)

        device = AuthenticationDevice.objects.get(id=self.device.id)
        self.assertEqual(device.counter, settings.TWOFACTOR_PAPER_CODES * 4)

    def test_device_generate_not_available_if_enough_remaining_codes(self):
        allow_gen = settings.TWOFACTOR_PAPER_CODES - 1
        self.device.counter = settings.TWOFACTOR_PAPER_CODES - allow_gen - 1
        self.device.save()

        with self.settings(TWOFACTOR_PAPER_CODES_ALLOW_GENERATION=allow_gen):
            response = self.client.get(self.print_url)
            doc = PyQuery(response.content)
            generate_button = doc.find('a.btn[href="%s"]' % self.generate_url)
            self.assertEqual(len(generate_button), 0)
            self.assertFalse(response.context_data['generation_enabled'])

    @override_settings(TWOFACTOR_PAPER_CODES_ALLOW_GENERATION=6)
    def test_device_generate_available_if_few_remaining_codes(self):
        for i in xrange(settings.TWOFACTOR_PAPER_CODES_ALLOW_GENERATION):
            self.assert_generation_warning_and_button(i)


class DeviceAuthenticationFailuretestCase(DeviceViewsTestCaseBase):

    url = reverse('twofactor')

    def setUp(self):
        super(DeviceAuthenticationFailuretestCase, self).setUp()
        self.account.twofactor_attempts = 7
        self.account.save()

    def test_twofactor_disabled(self):
        self.assert_twofactor_disabled(self.url)

    @mock.patch('webui.views.ui.logger')
    @mock.patch('webui.views.ui.auth')
    def test_too_many_attempts(self, mock_auth, mock_logger):
        with self.settings(TWOFACTOR_MAX_ATTEMPTS=0):
            response = self.client.post(self.url, {'oath_token': '123456'})
            self.assertIn('Account suspended', response.content)
            self.assertTemplateUsed(response, 'account/suspended.html')

            account = Account.objects.get(id=self.account.id)
            self.assertEqual(account.status, AccountStatus.SUSPENDED)
            mock_auth.logout.assert_called_once_with(mock.ANY)
            mock_logger.warning.assert_called_once_with(
                mock.ANY, account.openid_identifier, account.id)

    def test_failed_login_increments_attempts(self):
        self.client.post(self.url, {'oath_token': '123456'})
        account = Account.objects.get(id=self.account.id)
        self.assertEqual(account.twofactor_attempts, 8)

    def test_successful_login_resets_attempts(self):
        with mock.patch('webui.views.ui.twofactor.authenticate'):
            self.client.post(self.url, {'oath_token': '123456'})
            account = Account.objects.get(id=self.account.id)
            self.assertEqual(account.twofactor_attempts, 0)

    def test_invalid_post_increments_attempts(self):
        self.client.post(self.url, {})
        account = Account.objects.get(id=self.account.id)
        self.assertEqual(account.twofactor_attempts, 8)


class TwoFactorViewConcurrencyTestCase(SSOBaseTransactionTestCase):
    url = reverse('twofactor')

    @switches(TWOFACTOR=True)
    def test_race_condition_when_twofactor_attempt(self):
        account = self.factory.make_account(email='foo@test.com')
        self.factory.make_device(account=account, name='foo')

        def mocked(*args, **kwargs):
            import time
            # block for some time to ensure requests are concurrent
            time.sleep(0.1)

        self.patch(twofactor, 'authenticate', new=mocked)
        # We cannot easily spy memcache in the threads, so track the rate
        # limiting key for proper cleanup
        rate_limiter = ratelimit_extra(key_field='email')
        self.rl_key_used = None

        @test_concurrently(2)
        def fail_login():
            request = self.factory.make_request(
                path=self.url, data={}, method='POST', user=account,
                session=True)
            view = TwoFactorView.as_view()

            # Capture the rate limiting key
            self.rl_key_used = rate_limiter.current_key(request)

            with transaction.atomic():
                response = view(request)
            self.assertEqual(response.status_code, 200)

        fail_login()

        # Now insert the used key in the main thread cache
        rate_limiter.cache_incr(self.rl_key_used)

        account = Account.objects.get(id=account.id)
        self.assertEqual(account.twofactor_attempts, 2)


class UtilsTestCase(DeviceViewsTestCaseBase):

    def test_name_with_empty_existing(self):
        assert self.account.devices.count() == 0
        result = get_unique_device_name_for_user('test', self.account)
        self.assertEqual(result, 'test')

    def test_name_with_existing_unmatched(self):
        self.factory.make_device(account=self.account, name='some other name')
        assert self.account.devices.count() == 1
        result = get_unique_device_name_for_user('test', self.account)
        self.assertEqual(result, 'test')

    def test_name_with_existing_matched(self):
        self.factory.make_device(account=self.account, name='test')
        self.factory.make_device(account=self.account, name='test (1)')
        result = get_unique_device_name_for_user('test', self.account)
        self.assertEqual(result, 'test (2)')

    def assert_result_length(self, n):
        with mock.patch('webui.views.devices.rand_bytes') as rand:
            rand.return_value = 'a' * n
            s = generate_key(n)
            self.assertEqual(len(s), len(b16encode('a' * n)))

    def test_generate_key(self):
        self.assert_result_length(1)
        self.assert_result_length(10)
        self.assert_result_length(20)
