# Copyright 2016 Canonical Ltd.  This software is licensed under
# the GNU Affero General Public License version 3 (see the file LICENSE).

from django.template import (
    Context,
    Template,
)

from identityprovider.tests.utils import SSOBaseTestCase
from webui.views import gpg_messages


class GPGMessagesTestCase(SSOBaseTestCase):

    # These show the internal format and the correct display format for both
    # v3 (32 character) and v4 (40 character) OpenGPG fingerprints. This
    # standard is set by the gpg command line utility.
    fingerprint_internal_format_v4 = '1111222233334444555566667777888899990000'
    fingerprint_display_format_v4 = (
        '<code>'
        '1111 2222 3333 4444 5555'
        '&nbsp;&nbsp;'
        '6666 7777 8888 9999 0000'
        '</code>'
    )
    fingerprint_internal_format_v3 = '11112222333344445555666677778888'
    fingerprint_display_format_v3 = (
        '<code>'
        '11 11 22 22 33 33 44 44'
        '&nbsp;&nbsp;'
        '55 55 66 66 77 77 88 88'
        '</code>'
    )

    def assert_safe_text_contains(self, safe_text, expected):
        """Assert that 'safe text' contains 'expected' when rendered.

        Render 'safe_text' into a template and verify that the output contains
        'expected'. This is used to test that safe_text has been marked as safe
        if it contains any HTML components.
        """
        output = Template('{{ value }}').render(Context(dict(value=safe_text)))
        self.assertIn(expected, output)

    def test_display_fingerprint_works_with_v4_key(self):
        observed = gpg_messages.display_fingerprint(
            self.fingerprint_internal_format_v4)

        self.assert_safe_text_contains(
            observed,
            self.fingerprint_display_format_v4,
        )

    def test_display_fingerprint_works_with_v3_key(self):
        observed = gpg_messages.display_fingerprint(
            self.fingerprint_internal_format_v3)

        self.assert_safe_text_contains(
            observed,
            self.fingerprint_display_format_v3,
        )

    def test_display_fingerprint_raises_on_invalid_fingerprint(self):
        self.assertRaises(
            ValueError,
            gpg_messages.display_fingerprint,
            'invalid fingerprint'
        )

    def test_display_fingerprint_protects_against_xss(self):
        # construct a fingerprint that's the right length *before* escaping,
        # but will be too long after escaping. This is exactly 40 characters:
        bad_fingerprint = '<script>javascript.does_evil();</script>'
        self.assertRaises(
            ValueError,
            gpg_messages.display_fingerprint,
            bad_fingerprint
        )

    def test_key_already_imported(self):
        observed = gpg_messages.key_already_imported(
            self.fingerprint_internal_format_v4)
        self.assert_safe_text_contains(
            observed,
            self.fingerprint_display_format_v4,
        )
        observed = gpg_messages.key_already_imported(
            self.fingerprint_internal_format_v3)
        self.assert_safe_text_contains(
            observed,
            self.fingerprint_display_format_v3,
        )

    def test_key_expired(self):
        observed = gpg_messages.key_expired(
            self.fingerprint_internal_format_v4)
        self.assert_safe_text_contains(
            observed,
            self.fingerprint_display_format_v4,
        )
        observed = gpg_messages.key_expired(
            self.fingerprint_internal_format_v3)
        self.assert_safe_text_contains(
            observed,
            self.fingerprint_display_format_v3,
        )

    def test_key_revoked(self):
        observed = gpg_messages.key_revoked(
            self.fingerprint_internal_format_v4)
        self.assert_safe_text_contains(
            observed,
            self.fingerprint_display_format_v4,
        )
        observed = gpg_messages.key_revoked(
            self.fingerprint_internal_format_v3)
        self.assert_safe_text_contains(
            observed,
            self.fingerprint_display_format_v3,
        )

    def test_signing_key_mismatch(self):
        observed = gpg_messages.signing_key_mismatch(
            self.fingerprint_internal_format_v4)
        self.assert_safe_text_contains(
            observed,
            self.fingerprint_display_format_v4,
        )
        observed = gpg_messages.signing_key_mismatch(
            self.fingerprint_internal_format_v3)
        self.assert_safe_text_contains(
            observed,
            self.fingerprint_display_format_v3,
        )

    def test_key_added_to_account(self):
        observed = gpg_messages.key_added_to_account(
            self.fingerprint_internal_format_v4)
        self.assert_safe_text_contains(
            observed,
            self.fingerprint_display_format_v4,
        )
        observed = gpg_messages.key_added_to_account(
            self.fingerprint_internal_format_v3)
        self.assert_safe_text_contains(
            observed,
            self.fingerprint_display_format_v3,
        )

    def test_pending_tokens_cancelled(self):
        observed = gpg_messages.pending_tokens_cancelled(
            [self.fingerprint_internal_format_v4])
        self.assert_safe_text_contains(
            observed,
            self.fingerprint_display_format_v4,
        )
        observed = gpg_messages.pending_tokens_cancelled(
            [self.fingerprint_internal_format_v3])
        self.assert_safe_text_contains(
            observed,
            self.fingerprint_display_format_v3,
        )

    def test_keys_disabled(self):
        observed = gpg_messages.keys_disabled(
            [self.fingerprint_internal_format_v4])
        self.assert_safe_text_contains(
            observed,
            self.fingerprint_display_format_v4,
        )
        observed = gpg_messages.keys_disabled(
            [self.fingerprint_internal_format_v3])
        self.assert_safe_text_contains(
            observed,
            self.fingerprint_display_format_v3,
        )
