# coding: utf-8

# Copyright 2010 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

from __future__ import unicode_literals

from django.conf import settings
from django.contrib.auth.models import AnonymousUser
from django.core.urlresolvers import reverse
from gargoyle.testutils import switches
from mock import call, Mock
from oath import hotp

from identityprovider.fields import (
    FIELD_REQUIRED,
    INVALID_EMAIL,
    ONE_TIME_PWD_ERROR,
)
from identityprovider.forms import TwoFactorSyncForm
from identityprovider.models import Account
from identityprovider.models.twofactor import (
    TWOFACTOR_LOGIN,
    TWOFACTOR_SYNC_ERROR,
    AuthenticationDevice,
)
from identityprovider.tests.utils import SSOBaseTestCase, TimelineActionMixin
from identityprovider.tests import DEFAULT_USER_PASSWORD

from webui.constants import (
    EMAIL_EXISTS_ERROR,
    VERIFY_EMAIL_MESSAGE,
)
from webui.views.registration import (
    ACCOUNT_CREATED,
    VERIFY_EMAIL_SENT,
    WEB_CREATION_SOURCE,
    new_account,
)
from ssoclient.v2 import errors as api_errors


class RegisterTestMixin(object):

    """Constants common to other TestCases in this module."""

    URL = reverse('new_account')
    TESTDATA = {
        'email': 'test@test.com',
        'password': DEFAULT_USER_PASSWORD,
        'passwordconfirm': DEFAULT_USER_PASSWORD,
        'displayname': 'Foo bar',
        'accept_tos': True,
    }


class RegisterTestCase(SSOBaseTestCase, RegisterTestMixin):

    def setUp(self):
        """Prepare by mocking methods we're not interested in testing."""
        super(RegisterTestCase, self).setUp()
        account = self.factory.make_account()
        account.backend = 'sso'
        self.mock_authenticate = self.patch(
            'webui.views.registration.auth.authenticate', return_value=account)

        self.mock_get_api_client = self.patch(
            'webui.views.registration.get_api_client'
        )
        self.mock_api_register = self.mock_get_api_client.return_value.register
        self.mock_increment = self.patch(
            'webui.views.registration.stats.increment'
        )

    def assert_stat_calls(self, keys, name='flows.new_account', rpconfig=None):
        self.assertEqual(len(keys), len(self.mock_increment.call_args_list))
        for key, actual in zip(keys, self.mock_increment.call_args_list):
            expected = call(name, key=key, rpconfig=rpconfig)
            self.assertEqual(expected, actual)

    def assert_form_displayed(self, response, **kwargs):
        self.assertEqual(
            response.template_name,
            'registration/new_account.html'
        )
        form = response.context_data['form']
        for field, value in kwargs.items():
            self.assertIn(unicode(value), form.errors[field][0])

    def get(self, token=None, **data):
        request = self.factory.make_request(
            self.URL, session=True, data=data, user=AnonymousUser(),
            messages=True, preseed_cookie_check=True)
        response = new_account(request, token)
        response.render()
        return response

    def post(self, token=None, **data):
        request = self.factory.make_request(
            self.URL, method='POST', session=True, data=data,
            user=AnonymousUser(), messages=True, preseed_cookie_check=True)
        response = new_account(request, token)
        response.request = request
        return response

    def test_get(self):
        response = self.get()
        ctx = response.context_data
        self.assertEqual(ctx['form']['email'].value(), None)
        self.assertEqual(ctx['rpconfig'], None)
        self.assertEqual(ctx['token'], None)
        self.assert_form_displayed(response)
        self.assert_stat_calls(['requested'])

    def test_head(self):
        request = self.factory.make_request(
            self.URL, method='HEAD', user=AnonymousUser(),
            preseed_cookie_check=True)
        response = new_account(request)
        self.assertEqual(response.status_code, 405)

    def test_get_with_email(self):
        response = self.get(email='test@test.com')
        ctx = response.context_data
        self.assertEqual(ctx['form']['email'].value(), 'test@test.com')

    @switches(CAPTCHA=False)
    def test_get_optional_captcha_switch_off(self):
        response = self.get()
        self.assertEqual(response.context_data['captcha_required'], False)

    @switches(CAPTCHA=True, CAPTCHA_NEW_ACCOUNT=True)
    def test_get_optional_captcha_switch_on(self):
        response = self.get()
        self.assertEqual(response.context_data['captcha_required'], True)

    def test_post_required_fields(self):
        response = self.post()
        self.assert_form_displayed(
            response,
            email=FIELD_REQUIRED,
            password=FIELD_REQUIRED,
            passwordconfirm=FIELD_REQUIRED,
            displayname=FIELD_REQUIRED,
        )
        self.assert_stat_calls(['error.form'])

    def test_post_invalid_data(self):
        response = self.post(
            email="111",
            password="ads",
            passwordconfirm="ads",
            displayname="foo"
        )
        self.assert_form_displayed(
            response, email=INVALID_EMAIL, password='must be at least 8')
        self.assert_form_displayed(response)
        self.assert_stat_calls(['error.form'])

    def test_post_invalid_data_with_token(self):
        token = '12345678' * 2
        response = self.post(token=token, email="111")
        self.assertEqual(response.context_data['token'], token)
        self.assert_form_displayed(response)
        self.assert_stat_calls(['error.form'])

    def test_post_success(self):
        response = self.post(**self.TESTDATA)

        self.mock_api_register.assert_called_once_with(
            email=self.TESTDATA['email'],
            password=self.TESTDATA['password'],
            displayname=self.TESTDATA['displayname'],
            captcha_id=None,
            captcha_solution=None,
            create_captcha=False,
            creation_source=WEB_CREATION_SOURCE,
        )

        self.assertEqual(response.status_code, 302)
        self.assertEqual(response['Location'], '/')
        self.assert_stat_calls(['success'])
        self.assertEqual(
            response.request.session['token_email'], self.TESTDATA['email'])

    def test_post_success_with_token(self):
        token = '12345678' * 2
        response = self.post(token=token, **self.TESTDATA)

        self.assertEqual(response.status_code, 302)
        self.assertEqual(
            response['Location'],
            reverse('server-decide', kwargs=dict(token=token))
        )

    def test_post_already_registered(self):
        exc = api_errors.AlreadyRegistered(Mock())
        self.mock_api_register.side_effect = exc
        response = self.post(**self.TESTDATA)
        self.assert_form_displayed(response, email=VERIFY_EMAIL_MESSAGE)
        self.assert_stat_calls(['error.email'])

    def test_post_captcha_required(self):
        exc = api_errors.CaptchaRequired(Mock())
        self.mock_api_register.side_effect = exc
        response = self.post(**self.TESTDATA)
        self.assert_form_displayed(response)
        self.assertEqual(response.context_data['captcha_required'], True)

    def test_post_captcha_failure(self):
        mock_response = Mock()
        body = {'extra': {'captcha_message': 'XXX'}}
        exc = api_errors.CaptchaFailure(mock_response, body)
        self.mock_api_register.side_effect = exc

        response = self.post(**self.TESTDATA)
        self.assert_form_displayed(response)
        self.assertEqual(response.context_data['captcha_required'], True)
        self.assertEqual(
            response.context_data['captcha_error'],
            '&error=XXX')
        self.assert_stat_calls(['error.captcha'])

    def test_post_captcha_error(self):
        mock_response = Mock()
        body = {}
        exc = api_errors.CaptchaError(mock_response, body)
        self.mock_api_register.side_effect = exc

        response = self.post(**self.TESTDATA)
        self.assert_form_displayed(response)
        self.assertEqual(response.context_data['captcha_required'], True)
        self.assert_stat_calls(['error.captcha'])


class RegisterTimelineTestCase(
        SSOBaseTestCase, RegisterTestMixin, TimelineActionMixin):

    def setUp(self):
        """Don't mock any methods since we want to test code path behavior."""
        super(RegisterTimelineTestCase, self).setUp()
        account = self.factory.make_account()
        account.backend = 'sso'

    def make_request(self, timeline=True):
        request = self.factory.make_request(
            self.URL, method='POST', session=True, data=self.TESTDATA,
            user=AnonymousUser(), messages=True, timeline=timeline,
            preseed_cookie_check=True)
        return request

    def create_and_verify_account(self, request):
        response = new_account(request)
        self.assertEqual(response.status_code, 302)
        self.assertEqual(Account.objects.all().count(), 2)

    def test_registration_records_password_checking_timeline(self):
        request = self.make_request()
        self.create_and_verify_account(request)
        # Need the account to verify the id
        account = Account.objects.get(displayname=self.TESTDATA['displayname'])
        self.assert_timeline_contains(request=request,
                                      category='django-password-checking',
                                      detail=str(account.id))

    def test_registration_password_checking_no_timeline(self):
        request = self.make_request(timeline=False)
        self.create_and_verify_account(request)
        self.assertNotIn('timeline.timeline', request.META)

    def test_registration_password_checking_weird_timeline(self):
        request = self.make_request()
        request.META['timeline.timeline'] = []
        self.create_and_verify_account(request)
        self.assertEqual([], request.META['timeline.timeline'])


class RegisterMessagesTestCase(SSOBaseTestCase, RegisterTestMixin):

    def setUp(self):
        super(RegisterMessagesTestCase, self).setUp()
        # Before running any test, prepare the test cookie in the
        # client's cookie store.
        self.preseed_cookie_check()

    def test_success_message(self):
        # use regular django test client to assert over messages
        response = self.client.post(self.URL, data=self.TESTDATA, follow=True)

        messages = list(response.context['messages'])
        self.assertEqual(len(messages), 1)
        expected = (
            ACCOUNT_CREATED,
            VERIFY_EMAIL_SENT % dict(email_to=self.TESTDATA['email'],
                                     email_from=settings.NOREPLY_FROM_ADDRESS),
        )
        self.assertEqual(
            unicode(messages[0]), ' '.join(unicode(e) for e in expected))

    def test_error_message_when_already_registered(self):
        account = self.factory.make_account()
        self.TESTDATA['email'] = account.preferredemail.email

        response = self.client.post(self.URL, data=self.TESTDATA)

        # force cast to unicode to make sure we get a string and
        # not a lazy translation object
        self.assertContains(response, unicode(EMAIL_EXISTS_ERROR))

    def test_warning_message_when_leaked_password(self):
        self.factory.make_leaked_credential(
            email='foo@foo.com', password=DEFAULT_USER_PASSWORD)
        self.TESTDATA['email'] = 'foo@foo.com'

        response = self.client.post(self.URL, data=self.TESTDATA)

        self.assertContains(response, 'there was a security breach')
        # make sure the account was not created
        self.assertIsNone(Account.objects.get_by_email('foo@foo.com'))


class PasswordResetRequestTestCase(SSOBaseTestCase):

    url_name = 'forgot_password'

    def setUp(self):
        super(PasswordResetRequestTestCase, self).setUp()
        self.data = {'email': 'test@test.com'}

        self.mock_get_api_client = self.patch(
            'webui.views.registration.get_api_client'
        )
        self.mock_api_request_password_reset = (
            self.mock_get_api_client.return_value.request_password_reset)
        self.mock_increment = self.patch(
            'webui.views.registration.stats.increment'
        )

    def url(self, token):
        if token:
            url = reverse(self.url_name, kwargs={'token': token})
        else:
            url = reverse(self.url_name)
        return url

    def assert_stat_calls(self, keys, name='flows.forgot_password',
                          rpconfig=None):
        self.assertEqual(len(keys), len(self.mock_increment.call_args_list))
        for key, actual in zip(keys, self.mock_increment.call_args_list):
            expected = call(name, key=key, rpconfig=rpconfig)
            self.assertEqual(expected, actual)

    def assert_form_displayed(self, response, **kwargs):
        self.assertTemplateUsed(response, 'registration/forgot_password.html')
        form = response.context_data['form']
        for field, value in kwargs.items():
            self.assertIn(unicode(value), form.errors[field][0])

    def do_request(self, method, token=None, data=None, **kwargs):
        # Before running any test, prepare the test cookie.
        self.preseed_cookie_check()
        url = self.url(token)
        if data is None:
            data = {}

        response = getattr(self.client, method.lower())(
            url, data=data, **kwargs)
        return response

    def get(self, **kwargs):
        return self.do_request(method='GET', **kwargs)

    def post(self, **kwargs):
        return self.do_request(method='POST', **kwargs)

    def put(self, **kwargs):
        return self.do_request(method='PUT', **kwargs)

    def test_get(self):
        response = self.get()
        ctx = response.context_data
        self.assertEqual(ctx['form']['email'].value(), None)
        self.assertEqual(ctx['rpconfig'], None)
        self.assert_form_displayed(response)
        self.assert_stat_calls(['requested'])

    def test_get_with_email(self):
        response = self.get(data=dict(email='test@test.com'))
        ctx = response.context_data
        self.assertEqual(ctx['form']['email'].value(), 'test@test.com')

    def test_post_required_fields(self):
        response = self.post()
        self.assert_form_displayed(response, email=FIELD_REQUIRED)
        self.assert_stat_calls(['error.form'])

    def test_post_invalid_data(self):
        response = self.post(data=dict(email="111"))

        self.assert_form_displayed(response, email=INVALID_EMAIL)
        self.assert_form_displayed(response)
        self.assert_stat_calls(['error.form'])

    def test_post_invalid_data_with_token(self):
        token = '12345678' * 2
        response = self.post(token=token, data=dict(email="111"))

        self.assert_form_displayed(response)
        self.assert_stat_calls(['error.form'])

    def test_post_success(self):
        response = self.post(data=self.data)

        self.mock_api_request_password_reset.assert_called_once_with(
            email=self.data['email'], token=None)

        self.assertEqual(response.status_code, 200)
        self.assertEqual(
            self.client.session['token_email'], self.data['email'])

    def test_post_success_with_token(self):
        token = '12345678' * 2
        response = self.post(token=token, data=self.data)

        self.mock_api_request_password_reset.assert_called_once_with(
            email=self.data['email'], token=token)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(
            self.client.session['token_email'], self.data['email'])

    def test_post_email_invalidated(self):
        exc = api_errors.EmailInvalidated(Mock())
        self.mock_api_request_password_reset.side_effect = exc
        response = self.post(data=self.data)
        self.assertEqual(response.status_code, 200)
        self.assert_stat_calls(['error.email_invalidated'])

    def test_post_account_suspended(self):
        exc = api_errors.AccountSuspended(Mock())
        self.mock_api_request_password_reset.side_effect = exc
        response = self.post(data=self.data)
        self.assertEqual(response.status_code, 200)
        self.assert_stat_calls(['error.account_suspended'])

    def test_post_account_deactivated(self):
        exc = api_errors.AccountDeactivated(Mock())
        self.mock_api_request_password_reset.side_effect = exc
        response = self.post(data=self.data)
        self.assertEqual(response.status_code, 200)
        self.assert_stat_calls(['error.account_deactivated'])

    def test_post_can_not_reset_password(self):
        exc = api_errors.CanNotResetPassword(Mock())
        self.mock_api_request_password_reset.side_effect = exc
        response = self.post(data=self.data)
        self.assertEqual(response.status_code, 200)
        self.assert_stat_calls(['error.can_not_reset_password'])

    def test_post_too_many_tokens(self):
        exc = api_errors.TooManyTokens(Mock())
        self.mock_api_request_password_reset.side_effect = exc
        response = self.post(data=self.data)
        self.assertEqual(response.status_code, 200)
        self.assert_stat_calls(['error.too_many_tokens'])

    def test_reset_password(self):
        response = self.post(data=self.data)

        self.mock_api_request_password_reset.assert_called_once_with(
            email=self.data['email'], token=None)

        self.assertEqual(response.status_code, 200)
        self.assertEqual(
            self.client.session['token_email'], self.data['email'])
        self.assertEqual(response.context_data['email_heading'],
                         'Reset password')

    def test_method_not_allowed(self):
        response = self.put(data=self.data)
        self.assertEqual(response.status_code, 405)


class TwoFactorSyncTestCase(SSOBaseTestCase):
    """Tests for 2fa OTP synchronization."""

    def setUp(self):
        super(TwoFactorSyncTestCase, self).setUp()
        self.email = 'test@canonical.com'
        self.account = self.factory.make_account(
            email=self.email, password=DEFAULT_USER_PASSWORD)
        self.form = TwoFactorSyncForm()
        self.field_quantity = len(self.form.fields.keys())
        self.preseed_cookie_check()

    def fields_for(self, otps):
        return dict(zip(self.form.fields.keys(), otps))

    def counters_from(self, first_counter):
        counters_to_sync = range(first_counter,
                                 first_counter + self.field_quantity)
        return counters_to_sync

    def do_login(self):
        data = {'email': self.email, 'password': DEFAULT_USER_PASSWORD}
        url = reverse('login')
        return self.client.post(url, data)

    def test_shows_all_fields_for_synchronization(self):
        self.do_login()
        response = self.client.get(reverse('sync_2fa'))

        self.assertEqual(response.status_code, 200)
        for field_name in self.form.fields.keys():
            self.assertContains(response, field_name)

    def test_shows_all_fields_for_synchronization_in_launchpad(self):
        self.do_login()
        response = self.client.get(reverse('sync_2fa'))

        self.assertEqual(response.status_code, 200)
        for field_name in self.form.fields.keys():
            self.assertContains(response, field_name)

    def test_syncs_account_device(self):
        device = self.factory.make_device(self.account)
        self.do_login()

        counters_to_sync = self.counters_from(3)
        otps = [hotp(device.key, counter) for counter in counters_to_sync]

        response = self.client.post(reverse('sync_2fa'), self.fields_for(otps))

        self.assertRedirects(response, reverse('account-index'))
        device = AuthenticationDevice.objects.get(id=device.id)
        next_counter = counters_to_sync[-1] + 1
        self.assertEqual(device.counter, next_counter)

    def test_authenticates_user_after_synching(self):
        device = self.factory.make_device(self.account)
        self.do_login()

        counters_to_sync = self.counters_from(3)
        otps = [hotp(device.key, counter) for counter in counters_to_sync]

        self.client.post(reverse('sync_2fa'), self.fields_for(otps))

        self.assertIn(TWOFACTOR_LOGIN, self.client.session)

    def test_reloads_form_with_errors_if_invalid(self):
        device = self.factory.make_device(self.account)
        self.do_login()

        counters_to_sync = self.counters_from(3)
        otps = [hotp(device.key, counter) for counter in counters_to_sync]
        otps[1] = 'invalid-otp'

        response = self.client.post(reverse('sync_2fa'), self.fields_for(otps))

        self.assertEqual(response.status_code, 200)
        self.assertFormError(response, 'form', 'oath_token2',
                             ONE_TIME_PWD_ERROR)

    def test_shows_error_for_sync_error(self):
        device = self.factory.make_device(self.account)
        self.do_login()

        counters_to_sync = self.counters_from(device.counter +
                                              settings.HOTP_SYNC_WINDOW + 1)
        otps = [hotp(device.key, counter) for counter in counters_to_sync]

        response = self.client.post(reverse('sync_2fa'), self.fields_for(otps))

        self.assertEqual(response.status_code, 200)
        self.assertFormError(response, 'form', None, TWOFACTOR_SYNC_ERROR)

    def test_requires_login_to_sync(self):
        response = self.client.post(reverse('sync_2fa'), {})

        self.assertRedirects(response, reverse('login') + '?next=/sync_2fa')

    def test_follows_navigation_after_synching(self):
        device = self.factory.make_device(self.account)
        self.do_login()

        counters_to_sync = self.counters_from(3)
        otps = [hotp(device.key, counter) for counter in counters_to_sync]

        next_url = '/somewhere-else'
        data = self.fields_for(otps)
        data['next'] = next_url

        response = self.client.post(reverse('sync_2fa'), data)

        self.assertRedirects(response, next_url, target_status_code=404)

    def test_shows_next_url_in_form(self):
        self.do_login()
        next_url = '/somewhere-else'
        url = '{}?next={}'.format(reverse('sync_2fa'), next_url)

        response = self.client.get(url)

        self.assertContains(response, next_url)
