# coding: utf-8

# Copyright 2010 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

from __future__ import unicode_literals

import re
import urllib
import urllib2
from datetime import (
    date,
    datetime,
)
from functools import partial
from StringIO import StringIO
from textwrap import dedent
from urlparse import urlsplit
from urllib import quote

from django.conf import settings
from django.contrib.auth.models import (
    AnonymousUser,
    check_password,
    make_password
)
from django.contrib.sessions.models import Session
from django.core import mail
from django.core.cache import cache
from django.core.exceptions import ValidationError
from django.core.urlresolvers import reverse
from django.http import QueryDict
from django.utils.html import escape
from django.test.utils import override_settings
from gargoyle import gargoyle
from gargoyle.testutils import switches
from gpgservice_client.testing.keyserver.tests.keys import SampleDataKeys
from mock import Mock, patch
from pyquery import PyQuery
from timeline import Timeline
from u1testutils.logging import LogHandlerTestCase

from identityprovider import signed
from identityprovider.fields import (
    FIELD_REQUIRED,
    INVALID_EMAIL,
    ONE_TIME_PWD_ERROR,
)
from identityprovider.login import LOGIN_INCORRECT, AuthenticationError
from identityprovider.models import (
    Account,
    AuthToken,
    EmailAddress,
    OpenIDRPConfig,
    twofactor,
)
from identityprovider.models.captcha import (
    Captcha,
    CaptchaResponse,
    CaptchaV2,
)
from identityprovider.models.const import (
    AccountStatus,
    AuthLogType,
    AuthTokenType,
    EmailStatus,
)
from identityprovider.readonly import ReadOnlyManager
from identityprovider.tests import DEFAULT_USER_PASSWORD
from identityprovider.tests.test_auth import AuthLogTestCaseMixin
from identityprovider.tests.utils import (
    MockHandler,
    SSOBaseTestCase,
    TimelineActionMixin,
)
from identityprovider.utils import generate_random_string
from identityprovider.validators import PASSWORD_LEAKED
from identityprovider.views import server
from webui.tests.gpg_fixture import SSOGPGServiceFixture
from webui.views import (
    gpg_messages,
    ui,
)


DISPLAYNAME = 'Mark Shuttleworth'


def _args_list(flow_name, key=None):
    if key is None:
        key = 'success'
    return [
        ((flow_name,), {'key': 'requested', 'rpconfig': None}),
        ((flow_name,), dict(rpconfig=None, key=key)),
    ]


class BaseTestCase(SSOBaseTestCase):

    email = 'mark@example.com'
    new_email = 'person@example.com'

    def setUp(self):
        super(BaseTestCase, self).setUp()
        self.mock_increment = self.patch(
            'webui.views.ui.stats.increment')
        self.data = {'email': self.email,
                     'password': DEFAULT_USER_PASSWORD}
        self.account = self.factory.make_account(**self.data)
        self.preseed_cookie_check()

    def post_new_account(self, email=None, token=None, follow=False):
        if email is None:
            email = self.new_email
        if token is None:
            url = reverse('new_account')
        else:
            url = reverse('new_account', kwargs=dict(token=token))
        data = {
            'displayname': DISPLAYNAME,
            'email': email,
            'password': 'Testing123',
            'passwordconfirm': 'Testing123',
            'accept_tos': True
        }
        if gargoyle.is_active('CAPTCHA'):
            data['recaptcha_challenge_field'] = 'ignored'
            data['recaptcha_response_field'] = 'ignored'

        return self.client.post(url, data, follow=follow)

    def _token_url(self):
        token_string = self.mock_send_email.call_args[0][2]['token']
        token = AuthToken.get_token(raw_token=token_string)
        # Hack the token so it has the expected raw_token attribute.
        token.raw_token = token_string
        return token.get_absolute_url()

    def authenticate(self):
        assert self.client.login(
            username=self.data['email'], password=self.data['password'])

    def request_when_captcha_fails(self, url, data):
        class MockCaptcha(object):
            def __init__(self, *args):
                pass

            def verify(self, solution, ip_addr, email):
                self.message = 'no-challenge'
                return False

            @classmethod
            def new(cls, env):
                return cls()

        with patch.object(ui, 'Captcha', MockCaptcha):
            r = self.client.post(url, data)

        return r


@override_settings(LANGUAGE_CODE='es')
class SpanishUIViewsTestCase(BaseTestCase):

    def test_new_account_in_spanish(self):
        self.account.status = AccountStatus.ACTIVE
        self.account.save()

        r = self.post_new_account(email=self.email)

        self.assertEqual(r.status_code, 200)
        self.assertEqual(len(mail.outbox), 1)


class UIViewsBaseTestCase(BaseTestCase):

    def setUp(self):
        super(UIViewsBaseTestCase, self).setUp()
        self.mock_send_email = self.patch(
            'identityprovider.emailutils.send_fancy_email'
        )


class LoginTestCase(
        UIViewsBaseTestCase, AuthLogTestCaseMixin, TimelineActionMixin):

    def test_login_with_unicode_email(self):
        email = u'mark@客example.com'
        password = DEFAULT_USER_PASSWORD
        Account.objects.create_account(DISPLAYNAME, email, password)
        self.data['email'] = email.encode('utf-8')
        r = self.client.post(reverse('login'), self.data)

        self.assertRedirects(r, reverse('account-index'))

    def test_login_with_unicode_password(self):
        self.data['password'] = u'test客'.encode('utf-8')
        r = self.client.post(reverse('login'), self.data)

        self.assertFormError(r, 'form', None, LOGIN_INCORRECT)

    def test_login_with_next(self):
        self.data['next'] = reverse('account-edit')
        r = self.client.post(reverse('login'), self.data)
        self.assertRedirects(r, reverse('account-edit'))

    def test_login_failed_creates_authlog(self):
        self.data['next'] = reverse('account-edit')
        self.data['password'] = 'bad password'

        expected_values = {
            'login_email': self.data['email'],
            'log_type': AuthLogType.OPENID_FAIL,
            'referer': 'http://referer.example.com',
            'user_agent': 'Godzilla 1.0',
            'remote_ip': '127.0.0.1',
            'account': None
        }
        with self.assert_authlog_matches(expected_values):
            self.client.post(
                reverse('login'), self.data,
                HTTP_REFERER='http://referer.example.com',
                HTTP_USER_AGENT='Godzilla 1.0')

    def test_login_creates_authlog(self):
        self.data['next'] = reverse('account-edit')
        expected_values = {
            'login_email': self.data['email'],
            'log_type': AuthLogType.OPENID_NEW,
            'referer': 'http://referer.example.com',
            'user_agent': 'Godzilla 1.0',
            'remote_ip': '127.0.0.1',
            'account': self.account.id
        }
        with self.assert_authlog_matches(expected_values):
            self.client.post(
                reverse('login'), self.data,
                HTTP_REFERER='http://referer.example.com',
                HTTP_USER_AGENT='Godzilla 1.0')

    def test_login_when_account_inactive(self):
        _password = self.account.accountpassword.password

        for status, _ in AccountStatus._get_choices():
            if status == AccountStatus.ACTIVE:
                # skip as this is already tested by test_login_with_next
                continue

            self.account.status = status
            self.account.save()
            # reset password in case previous status was suspended, which
            # resets the password
            self.account.accountpassword.password = _password
            self.account.accountpassword.save()

            self.data['next'] = reverse('account-edit')
            r = self.client.post(reverse('login'), self.data)

            if status == AccountStatus.SUSPENDED:
                self.assertContains(r, 'Your account has been suspended. '
                                       'Please contact login support to '
                                       're-enable it')
            elif status == AccountStatus.DELETED:
                self.assertContains(r, 'There were some problems with the '
                                       'information you gave us.')
            else:
                self.assertContains(r, 'Your account has been deactivated. '
                                       'To reactivate it, please reset your '
                                       'password')

    def test_login_without_next_with_token(self):
        token = 'a' * 16
        r = self.client.post(reverse('login', kwargs=dict(token=token)),
                             self.data)
        self.assertEqual(r.status_code, 302)
        self.assertEqual(r['Location'], "http://testserver/%s/" % token)

    def test_login_without_next_nor_token(self):
        r = self.client.post(reverse('login'), self.data)
        self.assertRedirects(r, reverse('account-index'))

    def test_login_with_not_validated_email_permitted(self):
        self.account.emailaddress_set.create(
            email="mark-2@example.com", status=EmailStatus.NEW)
        self.data['email'] = 'mark-2@example.com'
        r = self.client.post(reverse('login'), self.data)
        self.assertRedirects(r, reverse('account-index'))

    def test_login_with_unverified_email_disabled_by_rpconfig(self):
        mock_get_rpconfig = self.patch(
            'webui.views.ui.get_rpconfig_from_request')
        rpconfig = OpenIDRPConfig(trust_root='http://localhost/foo',
                                  allow_unverified=False)
        mock_get_rpconfig.return_value = rpconfig

        # make account unverified
        self.account.emailaddress_set.all().update(status=EmailStatus.NEW)
        assert not self.account.is_verified
        self.data['email'] = 'mark@example.com'

        # use a token in the request to fake an openid request
        token = 'a' * 16
        r = self.client.post(
            reverse('login', kwargs={'token': token}), self.data)

        self.assertFormError(r, 'form', None, LOGIN_INCORRECT)

    def test_login_with_invalidated_email_error(self):
        email_obj = EmailAddress.objects.get(email="mark@example.com")
        email_obj.invalidate()

        r = self.client.post(reverse('login'), self.data)
        self.assertFormError(
            r, 'form', None,
            "This email address has been invalidated. "
            "Please contact login support.")

    def test_login_with_invalidated_email_error_when_email_readded(self):
        email_obj = EmailAddress.objects.get(email="mark@example.com")
        email_obj.invalidate()

        self.factory.make_account(email=self.email, password="password")
        r = self.client.post(reverse('login'), self.data)
        self.assertFormError(r, 'form', None, LOGIN_INCORRECT)

    def test_suspended_account_should_require_a_password_reset(self):
        """Checking regression in #620462."""
        self.account.suspend()

        # Try to log in using just suspended account
        response = self.client.post(
            reverse('login'), data=dict(email=self.email,
                                        password=DEFAULT_USER_PASSWORD))
        # This should produce an error
        suspended = ('<span class="error">Your account has been suspended. '
                     'Please contact login support to re-enable it</span>')
        self.assertContains(response, suspended)

        # re-enable the account
        self.account.status = AccountStatus.ACTIVE
        self.account.save()

        response = self.client.post(
            reverse('login'), data=dict(email=self.email,
                                        password=DEFAULT_USER_PASSWORD))

        # If the url is still the same, that means that the log in was
        # unsuccessful.
        self.assertContains(
            response,
            '<span class="error">%s</span>' % unicode(LOGIN_INCORRECT))

    def test_suspended(self):
        r = self.client.get(reverse('suspended'))
        self.assertTemplateUsed(r, 'account/suspended.html')

    def test_includes_create_account_form(self):
        r = self.client.get(reverse('login'))
        self.assertIn('create_account_form', r.context)

    def test_login_password_policy_problem_with_reset(self):
        validator = Mock(side_effect=ValidationError('THE REASON'))
        self.patch(
            'identityprovider.login.get_password_validator',
            return_value=validator)

        r = self.client.post(reverse('login'), self.data)

        # The form should redirect to reset password view, not back to login
        self.assertEqual(r.status_code, 302)
        redirect_url = r'(/token/.*/\+resetpassword/%s)' % self.email
        url_match = re.search(redirect_url, urllib2.unquote(r['location']))
        self.assertIsNotNone(url_match)
        reset_password_url = url_match.groups()[0]

        r = self.client.get(reset_password_url)
        self.assertContains(
            r, escape("Your password doesn't meet security constraints"))
        self.assertContains(r, 'THE REASON')

        validator.validate.side_effect = None
        # next request passes validation
        resp = self.client.post(reset_password_url, data={
            'passwordconfirm': 'Foo12345', 'password': 'Foo12345'})
        self.assertRedirects(resp, reverse('account-index'))

    def test_login_password_policy_problem_with_need_password_reset(self):
        self.account.need_password_reset = True
        self.account.save()

        r = self.client.post(reverse('login'), self.data)

        # The form should redirect to reset password view,
        # not back to login
        self.assertEqual(r.status_code, 302)
        redirect_url = r'(/token/.*/\+resetpassword/%s)' % self.email
        url_match = re.search(redirect_url, urllib2.unquote(r['location']))
        self.assertIsNotNone(url_match)
        reset_password_url = url_match.groups()[0]

        r = self.client.get(reset_password_url)
        self.assertContains(r, 'You need to change your password')
        self.assertContains(r, 'there was a security breach')

        # next request passes validation
        resp = self.client.post(reset_password_url, {
            'passwordconfirm': 'Foo12345', 'password': 'Foo12345'})
        self.assertRedirects(resp, reverse('account-index'))

    def test_login_password_policy_problem_with_old_token(self):
        AuthToken.objects.create(
            requester=self.account, email=self.email,
            requester_email=self.email, raw_token='foobar',
            date_consumed=date.today(),
            token_type=AuthTokenType.PASSWORDRECOVERY)

        validator = Mock(side_effect=ValidationError('THE REASON'))
        self.patch(
            'identityprovider.login.get_password_validator',
            return_value=validator)

        r = self.client.post(reverse('login'), self.data)

        # The foobar token should *not* have been used
        self.assertEqual(r.status_code, 302)
        regex = r'(/token/(.*)/\+resetpassword/%s)' % self.email
        action_match = re.search(regex, urllib2.unquote(r['location']))
        self.assertIsNotNone(action_match)
        self.assertNotEquals(action_match.groups()[1], 'foobar')

    def test_login_failed_tracked_mock_check_call(self):
        """Ensure the login_failed signal is sent with the expected params.

        This only verifies that the signal's send() method is called with
        the expected parameters; it doesn't verify the dispatcher will actually
        fire the methods registered on the signal.
        """

        request = self.factory.make_request(
            path=reverse('login'), method='POST', data=self.data, user=Mock())

        view = ui.LoginView(request=request)
        mock_send = self.patch('webui.views.ui.login_failed.send')
        self.patch(
            'webui.views.ui.authenticate_user',
            side_effect=AuthenticationError)

        view.post(request)

        mock_send.assert_called_once_with(
            sender=view, request=request,
            credentials=dict(email=self.data['email'],
                             password=self.data['password']))

    def test_login_failed_tracked_side_effects(self):
        """Ensure the login_failed signal results in expected side-effects.

        This only validates that the method registered to the signal produces
        the count increase in failed login attempts. It doesn't verify the
        method is called with the expected parameters.
        """
        request = self.factory.make_request(
            path=reverse('login'), method='POST', data=self.data, user=Mock())

        view = ui.LoginView(request=request)
        # Obtain the IP and key the same way the signal does it.
        ip = request.META.get('HTTP_X_FORWARDED_FOR', None)
        if ip is not None:
            ip = ip.split(', ')[0]
        else:
            ip = request.META.get('REMOTE_ADDR', '')
        key = "failed-login-%s" % ip

        old_attempts = cache.get(key, 0)
        self.patch(
            'webui.views.ui.authenticate_user',
            side_effect=AuthenticationError)

        view.post(request)
        new_attempts = cache.get(key, 0)
        self.assertEqual(old_attempts + 1, new_attempts)

    def test_show_warnings_when_twofactory_required_and_no_devices(self):
        self.account.twofactor_required = True
        self.account.save()

        response = self.client.post(
            reverse('login'), data=self.data, follow=True)

        self.assertRedirects(response, reverse('twofactor'))
        msgs = [m.message for m in response.context['messages']]
        self.assertEqual(msgs, [ui.TWOFACTOR_REQUIRED])

        # no 2fa form and a link to add a new device
        self.assertIsNone(response.context['form'])
        add_link = (
            'Please <a href="%s">add a new 2-factor authentication device</a>.'
        )
        self.assertContains(response, add_link % reverse('device-addition'))

    def test_logout_on_2fa_failure_but_success_regular_login(self):
        # This test makes sense when login happens via OpenID and with a
        # known OpenIDRPConfig entry that forces 2fa login.
        trust_root = 'http://example.com'
        self.factory.make_openid_rp_config(
            trust_root=trust_root, twofactor_required=True)
        params = self.factory.make_openid_request_params(
            trust_root=trust_root, return_to=trust_root)

        response = self.client.get(reverse('server-openid'), data=params)
        assert response.status_code == 302
        login_url_with_rpconfig_token = response['Location']

        self.data['oath_token'] = 123456  # invalid
        response = self.client.post(
            login_url_with_rpconfig_token, data=self.data, follow=True)

        self.assertFormError(response, 'form', '', 'The password is invalid')
        # user is not logged in
        self.assertNotIn('session_token', self.client.session)

    def test_login_with_blacklisted_password(self):
        # Update account with long enough password
        test_pass = "aPasswordToBlacklist1"
        self.account.accountpassword.password = make_password(test_pass)
        self.account.accountpassword.save()
        # Ensure the account is in the team that forces blacklisting
        self.factory.add_account_to_team(self.account,
                                         self.factory.make_team('canonical'))
        # Blacklist the test password
        self.patch('identityprovider.utils.get_password_blacklist',
                   return_value=set([test_pass.lower()]))
        # Use the now-blacklisted password when trying to login
        self.data['password'] = test_pass
        r = self.client.post(reverse('login'), self.data)
        # The form should redirect to reset password view,
        # not back to login
        self.assertEqual(r.status_code, 302)
        redirect_url = r'(/token/.*/\+resetpassword/%s)' % self.email
        url_match = re.search(redirect_url, urllib2.unquote(r['location']))
        self.assertIsNotNone(url_match)
        reset_password_url = url_match.groups()[0]
        r = self.client.get(reset_password_url)
        self.assertContains(r, escape("doesn't meet security constraints"))
        self.assertContains(r,
                            escape("it is very common and susceptible"))

        # next request passes validation
        resp = self.client.post(reset_password_url, {
            'passwordconfirm': 'FooBar12345', 'password': 'FooBar12345'})
        self.assertRedirects(resp, reverse('account-index'))

    def test_try_to_set_blacklisted_password(self):
        # Update account with long enough password so the initial login
        # doesn't force me to change it
        test_pass = "aPasswordThatMatchesThePolicy1"
        self.account.accountpassword.password = make_password(test_pass)
        self.account.accountpassword.save()
        # Ensure the account is in the team that forces blacklisting
        self.factory.add_account_to_team(self.account,
                                         self.factory.make_team('canonical'))
        # Blacklist the password we'll try to set next
        self.patch('identityprovider.utils.get_password_blacklist',
                   return_value=set(['blacklisted1']))
        # Log in with the initially-set, "good" password
        self.data['password'] = test_pass
        r = self.client.post(reverse('login'), self.data, follow=True)
        self.assertContains(r, "Choose password")
        # Now try to change it to the blacklisted one
        r = self.client.post('/', {'passwordconfirm': 'Blacklisted1',
                                   'password': 'Blacklisted1'})
        self.assertContains(r,
                            escape("it is very common and susceptible"))

    def test_django_password_checking_in_timeline(self):
        """Ensure that the request's timer_fn is propagated down."""
        empty_timeline = Timeline()
        response = self.client.post(reverse('login'), self.data, follow=True,
                                    **{'timeline.timeline': empty_timeline})
        self.assert_timeline_contains(request=response.wsgi_request,
                                      category='django-password-checking',
                                      detail=str(self.account.id))


class LogoutTestCase(UIViewsBaseTestCase):

    trust_root = 'http://example.com'

    def setUp(self):
        super(LogoutTestCase, self).setUp()
        self.authenticate()

    def logout(self, token=None, **kwargs):
        if token:
            url = reverse('logout', kwargs=dict(token=token))
        else:
            url = reverse('logout')
        response = self.client.get(url, data=kwargs)
        return response

    def set_language(self, language):
        self.addCleanup(self.client.post, reverse('set_language'),
                        data={'language': 'en'})
        # choose a language via UI
        response = self.client.post(
            reverse('set_language'), data={'language': language}, follow=True)
        assert response.status_code == 200, response.status_code

    def set_orequest_token(self, token):
        # This strange dance with session is necessary to overcome way in which
        # test.Client returns session (recreating it on every Client.session
        # access)
        session = self.client.session
        request = self.factory.make_openid_request_params()
        openid_server = server._get_openid_server(request)
        orequest = openid_server.decodeRequest(request)
        expected = signed.dumps(orequest, settings.SECRET_KEY)
        session[token] = expected
        session.save()
        return expected

    def assert_logged_out_in_sso(self, response):
        self.assertTemplateUsed(response, 'registration/logout.html')
        self.assertContains(response, 'You have been logged out')
        self.assertIsInstance(response.context['user'], AnonymousUser)

    def assert_redirect_as_per_request(self, response, redirected_to):
        self.assertRedirects(
            response, redirected_to, fetch_redirect_response=False)

    def test_logout(self):
        response = self.logout()
        self.assert_logged_out_in_sso(response)

    def test_logout_preserve_token_and_orequest(self):
        token = 'a' * 16
        expected = self.set_orequest_token(token)

        response = self.logout(token)
        self.assert_logged_out_in_sso(response)
        self.assertEqual(expected, self.client.session.get(token))

    def test_logout_displays_ga_snippet(self):
        rpconfig = self.factory.make_openid_rp_config(
            ga_snippet='[["_setAccount", "12345"]]')

        token = self._assign_token_to_rpconfig(rpconfig)
        response = self.logout(token)
        self.assertContains(response, "_gaq.push(['_setAccount', '12345']);")
        self.assert_logged_out_in_sso(response)

    def test_preferred_language_is_kept_after_logout(self):
        self.set_language('de')
        response = self.logout()

        self.assert_logged_out_in_sso(response)
        self.assertEqual(
            response.cookies[settings.LANGUAGE_COOKIE_NAME].value, 'de')

    def test_return_to_does_not_match_rpconfig(self):
        self.factory.make_openid_rp_config(trust_root=self.trust_root)

        response = self.logout(return_to='http://google.com')
        self.assert_logged_out_in_sso(response)
        self.assertNotContains(response, 'http://google.com')

    def test_return_to_does_match_rpconfig_no_site_name(self):
        self.factory.make_openid_rp_config(
            trust_root=self.trust_root, displayname='')

        return_to = self.trust_root + '/foo/bar/baz'
        response = self.logout(return_to=return_to)
        self.assert_logged_out_in_sso(response)
        self.assertContains(
            response, '<a href="{0}">Return to {0}</a>'.format(return_to))

    def test_return_to_does_match_rpconfig_with_site_name(self):
        self.factory.make_openid_rp_config(
            trust_root=self.trust_root, displayname='Example RPConfig')

        return_to = self.trust_root + '/foo/bar/baz'
        response = self.logout(return_to=return_to)
        self.assert_logged_out_in_sso(response)
        self.assertContains(
            response,
            '<a href="{0}">Return to Example RPConfig</a>'.format(return_to))

    def test_return_to_and_return_now(self):
        self.factory.make_openid_rp_config(trust_root=self.trust_root)

        return_to = self.trust_root + '/foo/bar/baz'
        response = self.logout(return_to=return_to, return_now=True)
        self.assert_redirect_as_per_request(response, return_to)

    def test_return_now_without_return_to(self):
        response = self.logout(return_now=True)
        self.assert_logged_out_in_sso(response)

    def test_return_now_preserves_language(self):
        self.factory.make_openid_rp_config(trust_root=self.trust_root)

        self.set_language('es')
        return_to = self.trust_root + '/foo/bar/baz'
        response = self.logout(return_to=return_to, return_now=True)
        self.assert_redirect_as_per_request(response, return_to)
        self.assertEqual(
            response.cookies[settings.LANGUAGE_COOKIE_NAME].value, 'es')

    def test_return_now_preserves_token(self):
        self.factory.make_openid_rp_config(trust_root=self.trust_root)

        token = 'a' * 16
        expected = self.set_orequest_token(token)
        return_to = self.trust_root + '/foo/bar/baz'
        response = self.logout(token, return_to=return_to, return_now=True)
        self.assert_redirect_as_per_request(response, return_to)
        self.assertEqual(expected, self.client.session.get(token))

    def test_return_to_no_rpconfig_and_return_now_does_not_redirect(self):
        self.factory.make_openid_rp_config(trust_root=self.trust_root)

        response = self.logout(return_to='http://google.com', return_now=True)
        self.assert_logged_out_in_sso(response)
        self.assertNotContains(response, 'http://google.com')

    def test_return_to_has_rpconfig_suffix(self):
        self.factory.make_openid_rp_config(trust_root=self.trust_root)

        return_to = self.trust_root + '.ar/foo/bar/baz'
        response = self.logout(return_to=return_to)
        self.assert_logged_out_in_sso(response)
        self.assertNotContains(response, return_to)

    def test_return_to_url_when_tricky_trust_root(self):
        self.factory.make_openid_rp_config(self.trust_root, displayname='')

        return_to = self.trust_root + '/foo/bar/baz'
        response = self.logout(return_to=return_to)
        self.assert_logged_out_in_sso(response)
        self.assertContains(
            response, '<a href="{0}">Return to {0}</a>'.format(return_to))

    def test_return_to_url_when_tilde(self):
        rpconfig = self.factory.make_openid_rp_config(self.trust_root)

        return_to = self.trust_root + '/foo/bar/baz'
        response = self.logout(return_to=return_to)
        self.assert_logged_out_in_sso(response)
        self.assertContains(
            response, '<a href="{0}">Return to {1}</a>'.format(
                return_to, rpconfig.displayname))


class EnterTokenTestCase(UIViewsBaseTestCase):

    def setUp(self):
        super(EnterTokenTestCase, self).setUp()
        self.data = {'email': self.email}

    def test_claim_unexisting_token(self):
        r = self.client.get(
            reverse('claim_token', kwargs=dict(authtoken=0)), self.data)
        self.assertEqual(r.status_code, 404)

    def test_claim_token_for_unexisting_token_type(self):
        token = self.factory.make_authtoken(email='fake@example.com')
        token.token_type = 9999
        token.save()
        r = self.client.get(
            reverse('claim_token', kwargs=dict(authtoken=token.raw_token)),
            self.data)
        self.assertEqual(r.status_code, 404)

    def test_claim_token_for_password_recovery_no_preferredemail(self):
        for email_obj in self.account.emailaddress_set.all():
            email_obj.status = EmailStatus.NEW
            email_obj.save()

        self.client.post(reverse('forgot_password'), {'email': self.email})

        # we send an email anyway despite there not being a validated email
        self.assertEqual(self.mock_send_email.call_count, 1)

    def test_claim_consumed_token(self):
        self.client.post(reverse('forgot_password'), {'email': self.email})
        url = self._token_url()

        pwd = 'Password1'
        r = self.client.post(url, {'password': pwd, 'passwordconfirm': pwd})
        self.assertRedirects(r, reverse('account-index'))

        self.client.cookies.clear()

        pwd = 'Password2'
        r = self.client.post(url, {'password': pwd, 'passwordconfirm': pwd})
        self.assertRedirects(r, reverse('bad_token'))


class NewEmailTestCase(UIViewsBaseTestCase):

    def test_config_email(self):
        self.authenticate()

        redirection_url = "http://example.com/foo/bar/baz"
        token = self.factory.make_authtoken(
            token_type=AuthTokenType.VALIDATEEMAIL, email=self.email,
            redirection_url=redirection_url)

        r = self.client.post(token.get_absolute_url(),
                             {'post': 'yes'})
        self.assertEqual(r.status_code, 302)
        self.assertEqual(r['Location'], redirection_url)

    def test_config_email_no_redirection_url(self):
        self.authenticate()

        token = self.factory.make_authtoken(
            token_type=AuthTokenType.VALIDATEEMAIL, email=self.email,
            redirection_url=None)

        r = self.client.post(token.get_absolute_url(),
                             {'post': 'yes'})
        self.assertEqual(r.status_code, 302)
        self.assertEqual(
            r['Location'],
            'http://testserver' + reverse('account-index')
        )

    def test_bad_token(self):
        r = self.client.get(reverse('bad_token'))
        self.assertEqual(r.status_code, 200)

    def test_logout_to_confirm(self):
        r = self.client.get(reverse('logout_to_confirm'))
        self.assertEqual(r.status_code, 200)

    def test_logout_to_confirm_use_request_context(self):
        response = self.client.get(reverse('logout_to_confirm'))
        msg = ('You have attempted to confirm an account while logged in as '
               'someone else.')
        self.assertContains(response, msg)


class ForgotPasswordTestCase(LogHandlerTestCase, UIViewsBaseTestCase):

    def test_forgot_password_includes_rp_analytics(self):
        rpconfig = OpenIDRPConfig.objects.create(
            trust_root='http://example.com/',
            ga_snippet='[["_setAccount", "12345"]]')
        token = self._assign_token_to_rpconfig(rpconfig)
        response = self.client.get(
            reverse('forgot_password', kwargs=dict(token=token)))
        self.assertContains(response, "_gaq.push(['_setAccount', '12345']);")

    def test_forgot_password_email_when_account_active(self):
        self.client.post(reverse('forgot_password'), {'email': self.email})

        self.assertEqual(self.mock_send_email.call_count, 1)

    def test_forgot_password_email_when_account_deactivated(self):
        """ Deactivated accounts can be reactivated using the reset password
        functionality. Bug #556878 """
        # make sure the account is deactivated
        self.account.status = AccountStatus.DEACTIVATED
        self.account.save()

        self.client.post(reverse('forgot_password'), {'email': self.email})

        self.assertEqual(self.mock_send_email.call_count, 1)
        self.assertEqual(self.account.preferredemail.email, self.email)

    def test_forgot_password_email_when_account_noaccount(self):
        # make sure the account is deactivated
        self.account.status = AccountStatus.NOACCOUNT
        self.account.save()

        self.client.post(reverse('forgot_password'), {'email': self.email})

        self.assertEqual(self.mock_send_email.call_count, 1)
        self.assertEqual(self.account.preferredemail.email, self.email)

    def test_forgot_password_email_when_account_suspended(self):
        # make sure the account is suspended
        self.account.suspend()

        self.client.post(reverse('forgot_password'), {'email': self.email})

        self.assertEqual(self.mock_send_email.call_count, 0)

    def test_forgot_password_invalid_form(self):
        r = self.client.post(reverse('forgot_password'))
        self.assertEqual(r.status_code, 200)
        self.assertTemplateUsed(r, 'registration/forgot_password.html')
        self.assertFormError(r, 'form', 'email', FIELD_REQUIRED)

    def test_forgot_password_with_token(self):
        token1 = generate_random_string(16)
        self.client.post(
            reverse('forgot_password', kwargs=dict(token=token1)),
            {'email': self.email}
        )
        # should be only one
        auth_token = AuthToken.objects.get(
            email=self.email, token_type=AuthTokenType.PASSWORDRECOVERY)
        self.assertEqual(auth_token.redirection_url,
                         reverse('server-decide', kwargs=dict(token=token1)))

    def test_reset_password_when_account_active(self):
        self.client.post(reverse('forgot_password'), {'email': self.email})

        # make sure the account is active
        self.account.status = AccountStatus.ACTIVE
        self.account.save()

        # confirm account
        data = {'password': 'Password1', 'passwordconfirm': 'Password1'}
        r = self.client.post(self._token_url(), data)
        self.assertRedirects(r, reverse('account-index'))

    def test_reset_password_default_action(self):
        self.client.post(reverse('forgot_password'), {'email': self.email})

        # make sure the account is active
        self.account.status = AccountStatus.ACTIVE
        self.account.save()

        # confirm account
        data = {'password': 'Password3', 'passwordconfirm': 'Password1'}
        r = self.client.post(self._token_url(), data)
        self.assertContains(r, ' action=""')

    def test_reset_password_email_is_case_insensitive(self):
        r = self.client.post(reverse('forgot_password'), {'email': self.email})

        # update requester email in token
        auth_token = AuthToken.objects.get(
            email=self.email, token_type=AuthTokenType.PASSWORDRECOVERY)
        auth_token.requester_email = self.email.upper()
        assert self.email.upper() != self.email
        auth_token.save()

        # make sure the account is active
        self.account.status = AccountStatus.ACTIVE
        self.account.save()

        # confirm account
        data = {'password': 'Password1', 'passwordconfirm': 'Password1'}
        r = self.client.post(self._token_url(), data)
        self.assertRedirects(r, reverse('account-index'))

    def test_reset_password_email_not_found(self):
        r = self.client.post(reverse('forgot_password'), {'email': self.email})

        # update requester email in token
        auth_token = AuthToken.objects.get(
            email=self.email, token_type=AuthTokenType.PASSWORDRECOVERY)
        auth_token.requester_email = 'another-email@example.com'
        auth_token.save()

        # make sure the account is active
        self.account.status = AccountStatus.ACTIVE
        self.account.save()

        # confirm account
        data = {'password': 'Password1', 'passwordconfirm': 'Password1'}

        mock_is_enabled = Mock(return_value=True)
        with patch.object(self.root_logger, 'isEnabledFor', mock_is_enabled):
            self.patch('webui.views.ui.logger', new=self.root_logger)
            r = self.client.post(self._token_url(), data)

        self.assertRedirects(r, reverse('bad_token'))
        self.assertLogLevelContains(
            'WARNING', "Could not reset password, invalid email address "
            "provided in the token: %s (account %s)" % (
                auth_token.requester_email, self.account.id))

    def test_reset_password_with_lowered_password_requirements(self):
        r = self.client.post(reverse('forgot_password'), {'email': self.email})

        # make sure the account is active
        self.account.status = AccountStatus.ACTIVE
        self.account.save()

        # confirm account
        data = {'password': 'password', 'passwordconfirm': 'password'}
        r = self.client.post(self._token_url(), data)
        self.assertRedirects(r, reverse('account-index'))

        account = Account.objects.get(pk=self.account.pk)
        # check password changed
        self.assertTrue(
            check_password('password', account.encrypted_password))

    def test_reset_password_when_account_active_no_password(self):
        r = self.client.post(reverse('forgot_password'), {'email': self.email})

        self.account.accountpassword.delete()
        # make sure the account is active
        self.account.status = AccountStatus.ACTIVE
        self.account.save()

        # confirm account
        data = {'password': 'Password1', 'passwordconfirm': 'Password1'}
        r = self.client.post(self._token_url(), data)
        self.assertRedirects(r, reverse('account-index'))

    def test_reset_password_when_account_deactivated(self):
        """Prevent password reset for unverified emails."""
        r = self.client.post(reverse('forgot_password'), {'email': self.email})

        # make sure the account is deactivated
        self.account.status = AccountStatus.DEACTIVATED
        self.account.save()

        # confirm account
        data = {'password': 'Password1', 'passwordconfirm': 'Password1'}
        r = self.client.post(self._token_url(), data)
        self.assertRedirects(r, reverse('account-index'))
        account = Account.objects.get(id=self.account.id)
        self.assertEqual(account.status, AccountStatus.ACTIVE)
        self.assertEqual(
            account.preferredemail.status, EmailStatus.PREFERRED)

    def test_reset_password_when_account_deactivated_but_notverified(self):
        """Prevent password reset for unverified emails."""
        r = self.client.post(reverse('forgot_password'), {'email': self.email})

        # make sure the account is deactivated
        self.account.status = AccountStatus.DEACTIVATED
        self.account.save()
        self.account.emailaddress_set.update(status=EmailStatus.NEW)

        # confirm account
        data = {'password': 'Password1', 'passwordconfirm': 'Password1'}
        r = self.client.post(self._token_url(), data)
        self.assertRedirects(r, reverse('account-index'))
        account = Account.objects.get(id=self.account.id)
        self.assertEqual(account.status, AccountStatus.ACTIVE)
        self.assertEqual(
            account.preferredemail.status, EmailStatus.PREFERRED)

    def test_reset_password_when_account_deactivated_no_preferred_email(self):
        # make sure the account is deactivated and preferred email is removed
        self.account.status = AccountStatus.DEACTIVATED
        # Blow out the cached value
        self.account.save()
        self.account.emailaddress_set.update(status=EmailStatus.NEW)

        self.account = Account.objects.get_by_email(self.email)

        self.assertEqual(self.account.preferredemail.email, self.email)

        self.client.post(reverse('forgot_password'), {'email': self.email})

        # we sent an email anyway, despite there not being a validated email
        self.assertEqual(self.mock_send_email.call_count, 1)

    def test_reset_password_when_account_no_verified_emails(self):
        self.account.emailaddress_set.update(status=EmailStatus.NEW)
        self.client.post(reverse('forgot_password'), {'email': self.email})
        # we sent an email anyway, despite there not being a validated email
        self.assertEqual(self.mock_send_email.call_count, 1)

    def test_reset_password_unverified_email_but_account_verified_email(self):
        verified = self.account.verified_emails()
        unverified = self.factory.make_email_for_account(
            self.account, status=EmailStatus.NEW)
        self.client.post(reverse('forgot_password'),
                         {'email': unverified.email})
        # email sent to verified email addresses instead
        self.assertEqual(self.mock_send_email.call_count, len(verified))

    def test_reset_password_when_account_noaccount(self):
        r = self.client.post(reverse('forgot_password'), {'email': self.email})

        # make sure the account is deactivated
        self.account.status = AccountStatus.NOACCOUNT
        self.account.save()

        # confirm account
        data = {'password': 'Password1', 'passwordconfirm': 'Password1'}
        r = self.client.post(self._token_url(), data)
        self.assertRedirects(r, reverse('account-index'))

    def test_reset_password_when_account_suspended(self):
        r = self.client.post(reverse('forgot_password'), {'email': self.email})

        self.account.suspend()

        data = {'password': 'Password1', 'passwordconfirm': 'Password1'}

        # confirm account
        r = self.client.post(self._token_url(), data)
        self.assertRedirects(r, reverse('bad_token'))

    def test_reset_password_invalid_form(self):
        # get valid session
        r = self.client.post(reverse('forgot_password'), {'email': self.email})

        # test view
        r = self.client.post(self._token_url())
        self.assertTemplateUsed(r, 'registration/reset_password.html')
        self.assertFormError(r, 'form', 'password', FIELD_REQUIRED)

    def test_reset_password_get(self):
        # get valid session
        r = self.client.post(reverse('forgot_password'), {'email': self.email})

        # test view
        r = self.client.get(self._token_url())
        self.assertTemplateUsed(r, 'registration/reset_password.html')
        self.assertEqual(r.context['form'].errors, {})

    def test_bad_method(self):
        r = self.client.put(reverse('forgot_password'), {'email': self.email})
        self.assertEqual(r.status_code, 405)

    def test_reset_password_invalidate_tokens(self):
        # make sure the account is active
        assert self.account.status == AccountStatus.ACTIVE

        self.account.create_v1_oauth_token('new-token')
        self.account.create_oauth_token('new-token')

        r = self.client.post(reverse('forgot_password'), {'email': self.email})

        # confirm account
        data = {'password': 'Password1', 'passwordconfirm': 'Password1'}
        r = self.client.post(self._token_url(), data)

        self.assertRedirects(r, reverse('account-index'))
        self.assertEqual(self.account.token_set.count(), 1)  # only web login
        self.assertEqual(
            self.account.user.oauth_consumer.token_set.count(), 0)

    def test_reset_password_no_redirection_url(self):
        self.client.post(reverse('forgot_password'), {'email': self.email})
        auth_token = AuthToken.objects.get(
            email=self.email, token_type=AuthTokenType.PASSWORDRECOVERY)
        auth_token.redirection_url = None
        auth_token.save()

        # confirm account
        data = {'password': 'Password1', 'passwordconfirm': 'Password1'}
        r = self.client.post(self._token_url(), data)
        self.assertRedirects(r, reverse('account-index'))

    def test_reset_password_leaked_password_used(self):
        self.client.post(reverse('forgot_password'), {'email': self.email})

        # make sure the account is active
        self.account.status = AccountStatus.ACTIVE
        self.account.save()
        self.factory.make_leaked_credential(
            email=self.email, password='Password1')

        # verify current password is not the one we're trying to set
        self.assertFalse(check_password('Password1',
                                        self.account.encrypted_password))
        # attempt to submit a leaked password
        data = {'password': 'Password1', 'passwordconfirm': 'Password1'}
        r = self.client.post(self._token_url(), data)
        self.assertFormError(r, 'form', 'password', [unicode(PASSWORD_LEAKED)])
        self.assertContains(r, 'there was a security breach')
        # refresh account from db
        account = Account.objects.get(id=self.account.id)
        # verify password hasn't changed
        self.assertFalse(check_password('Password1',
                                        account.encrypted_password))


class LoggedInForgotPasswordTestCase(ForgotPasswordTestCase):
    # forgot_password and reset_password views should also work
    # for logged in users, repeat tests with an authenticated user

    def setUp(self):
        super(LoggedInForgotPasswordTestCase, self).setUp()
        self.authenticate()


class ConfirmEmailTestCase(UIViewsBaseTestCase):

    def _setup_account_with_new_email(self, account=None):
        if account is None:
            account = self.account

        # create new email account
        email = self.factory.make_email_for_account(
            account, email=self.new_email, status=EmailStatus.NEW)
        # create verification token
        token = self.factory.make_authtoken(
            token_type=AuthTokenType.VALIDATEEMAIL, email=email.email)
        return email, token

    def get_email_token_validation_response(self, account=None):
        email, token = self._setup_account_with_new_email(account)
        response = self.client.get(token.get_absolute_url())
        return response, token

    def test_confirm_email_not_existing(self):
        self.authenticate()
        email, token = self._setup_account_with_new_email()
        # delete email account
        email.delete()
        # use verification token
        r = self.client.get(token.get_absolute_url())
        self.assertEqual(r.status_code, 404)

    def assertNumTokensForAddress(self, address, token_type, expected_num):
        tokens = AuthToken.objects.filter(
            date_consumed=None,
            email=address,
            token_type=token_type,
        )
        self.assertEqual(len(tokens), expected_num)

    def use_token(self, token, data=None):
        response = self.client.post(token.get_absolute_url(),
                                    data, follow=True)
        self.assertEqual(response.status_code, 200)
        return response

    def test_confirm_email_destroys_invalidation_token(self):
        self.authenticate()
        email, validate_token = self._setup_account_with_new_email()
        invalidate_token = self.factory.make_authtoken(
            token_type=AuthTokenType.INVALIDATEEMAIL,
            email=email.email,
        )
        self.use_token(validate_token)

        # now try and invalidate the same address
        response = self.use_token(invalidate_token)
        self.assertContains(response, "Unauthorized confirmation code")
        self.assertNumTokensForAddress(email.email,
                                       AuthTokenType.INVALIDATEEMAIL, 0)

    def test_confirm_email_works_with_no_invalidation_tokens(self):
        self.authenticate()
        email, validate_token = self._setup_account_with_new_email()

        # Validate the email address
        response = self.use_token(validate_token)
        self.assertEqual(response.status_code, 200)

    def test_validate_email_address_leaves_other_tokens(self):
        self.authenticate()
        email, validate_token = self._setup_account_with_new_email()
        second_account = self.factory.make_account()
        email2 = self.factory.make_email_for_account(second_account,
                                                     status=EmailStatus.NEW)
        self.factory.make_authtoken(token_type=AuthTokenType.VALIDATEEMAIL,
                                    email=email2.email)
        self.factory.make_authtoken(token_type=AuthTokenType.INVALIDATEEMAIL,
                                    email=email2.email)

        response = self.use_token(validate_token)
        self.assertEqual(response.status_code, 200)

        self.assertNumTokensForAddress(email2.email,
                                       AuthTokenType.INVALIDATEEMAIL, 1)
        self.assertNumTokensForAddress(email2.email,
                                       AuthTokenType.VALIDATEEMAIL, 1)

    def test_confirm_email_get(self):
        self.authenticate()

        email, token = self._setup_account_with_new_email()
        # use verification token
        self.client.get(token.get_absolute_url())

        # verify token has not been consumed, and email has not been
        # verified yet
        token = AuthToken.get_token(
            raw_token=token.raw_token, token_type=AuthTokenType.VALIDATEEMAIL)
        self.assertEqual(token.date_consumed, None)
        email = EmailAddress.objects.get(email=email.email)
        self.assertEqual(email.status, EmailStatus.NEW)

    def test_confirm_email_as_another_user_fails(self):
        self.authenticate()
        account = self.factory.make_account()

        r, token = self.get_email_token_validation_response(account)

        self.assertEqual(r.status_code, 404)
        self.assertEqual(0, AuthToken.objects.filter(id=token.id).count())
        email = EmailAddress.objects.get(email=self.new_email)
        self.assertEqual(EmailStatus.NEW, email.status)

    def test_confirm_email_when_not_logged_in_redirects_to_login(self):
        r, token = self.get_email_token_validation_response(self.account)

        kwargs = dict(authtoken=token, email_address=self.new_email)
        # reverse() quotes paths by default, but does not quote every char
        # sufficiently to be used as a querystring param in another URL, so
        # we still need to quote() but excluding the escaped '/' and '%' chars
        url = (reverse('login') + '?next=' +
               urllib2.quote(reverse('confirm_email', kwargs=kwargs), '/%'))
        self.assertRedirects(r, url)

    def test_confirm_email_when_not_logged_in_shows_message(self):
        email, token = self._setup_account_with_new_email()
        response = self.client.get(token.get_absolute_url(),
                                   follow=True)
        self.assertContains(
            response, 'Please log in to use this confirmation code')

    def test_invalid_email_does_not_blow_up(self):
        r = self.post_new_account(email='what<~@ever.com')
        self.assertEqual(200, r.status_code)
        self.assertContains(r, INVALID_EMAIL)

    @switches(CAPTCHA=True)
    @patch.object(CaptchaV2, 'verify', return_value=True)
    def test_confirm_email_with_valid_captcha(self, mock_verify):
        self.authenticate()
        email, validate_token = self._setup_account_with_new_email()
        response = self.use_token(validate_token,
                                  {'g-recaptcha-response': '0'})
        self.assertContains(response,
                            'The email address {} has been validated'.format(
                                email.email))
        email.refresh_from_db()
        self.assertTrue(email.is_verified)

    @switches(CAPTCHA=True)
    def test_confirm_email_with_invalid_captcha(self):
        self.authenticate()
        email, validate_token = self._setup_account_with_new_email()
        response = self.use_token(validate_token,
                                  {'g-recaptcha-response': '0'})
        self.assertContains(response, 'A captcha challenge is required')
        email.refresh_from_db()
        self.assertFalse(email.is_verified)

    @switches(CAPTCHA=True)
    def test_confirm_email_with_no_captcha(self):
        self.authenticate()
        email, validate_token = self._setup_account_with_new_email()
        response = self.use_token(validate_token)
        self.assertContains(response, 'A captcha challenge is required')
        email.refresh_from_db()
        self.assertFalse(email.is_verified)


@override_settings(CAPTCHA_PRIVATE_KEY='some-private-key')
class CaptchaVerificationTestCase(BaseTestCase):

    success_status = 302

    def setUp(self):
        super(CaptchaVerificationTestCase, self).setUp()
        mock_handler = MockHandler()
        mock_handler.set_next_response(200, 'false\nno-challenge')
        self.patch(Captcha, 'opener', new=urllib2.build_opener(mock_handler))

        p = switches(CAPTCHA=True)
        p.patch()
        self.addCleanup(p.unpatch)

    def test_new_account_when_form_validation_fails(self):
        r = self.post_new_account()
        self.assertTemplateUsed(r, 'registration/new_account.html')
        msg = 'It appears that our captcha service was unable to load'
        self.assertContains(r, msg)

    def test_new_account_captcha_whitelist(self):
        email = 'canonicaltest@gmail.com'
        pattern = '^canonicaltest(?:\+.+)?@gmail\.com$'
        with self.settings(EMAIL_WHITELIST_REGEXP_LIST=[pattern]):
            response = self.post_new_account(email=email)
            self.assertEqual(response.status_code, self.success_status)

    def test_new_account_captcha_whitelist_with_uuid(self):
        email = 'canonicaltest+something@gmail.com'
        pattern = '^canonicaltest(?:\+.+)?@gmail\.com$'
        with self.settings(EMAIL_WHITELIST_REGEXP_LIST=[pattern]):
            response = self.post_new_account(email=email)
            self.assertEqual(response.status_code, self.success_status)

    def test_new_account_captcha_whitelist_fail(self):
        email = 'notcanonicaltest@gmail.com'
        pattern = '^canonicaltest(?:\+.+)?@gmail\.com$'
        with self.settings(EMAIL_WHITELIST_REGEXP_LIST=[pattern]):
            response = self.post_new_account(email=email)
            msg = 'It appears that our captcha service was unable to load'
            self.assertContains(response, msg)

    @patch.object(Captcha, '_open')
    def test_uses_timeline_from_request(self, mock_open):
        mock_open.return_value = CaptchaResponse(200, StringIO('true\na'))
        request = Mock()
        timeline = Timeline()
        request.META = {'timeline.timeline': timeline}
        request.POST = {'recaptcha_challenge_field': 'captcha-id'}
        request.environ = {'REMOTE_ADDR': '127.0.0.1'}
        ui._verify_captcha_response(None, request, None)
        self.assertEqual(1, len(timeline.actions))
        self.assertEqual('captcha-verify', timeline.actions[0].category)


class CookiesTestCase(SSOBaseTestCase):

    def setUp(self):
        super(CookiesTestCase, self).setUp()
        self.patch('webui.decorators.disable_cookie_check', new=False)
        Session.objects.all().delete()

    def _get(self, url, *params):
        r = self.client.get(url)
        # The following line makes this cookieless:
        self.client.cookies.clear()
        if 300 <= r.status_code < 400:
            location = r['Location']
            scheme, netloc, path, query, fragment = urlsplit(location)
            return self._get(path, QueryDict(query))
        else:
            return r

    def _check_url(self, url):
        r = self._get(url)
        self.assertContains(r, '<title>Cookies required</title>')

        r = self.client.get(url, follow=True)
        self.assertNotContains(r, '<title>Cookies required</title>')

    def _count_sessions(self):
        return Session.objects.all().count()

    def test_login(self):
        self._check_url(reverse('login'))

    def test_account_index(self):
        url = reverse('account-index')

        r = self.client.get(url)

        self.assertEqual(r.status_code, 200)
        self.assertNotContains(r, '<title>Cookies required</title>')

    def test_forgot_password(self):
        self._check_url(reverse('forgot_password'))

    def test_new_account(self):
        self._check_url(reverse('new_account'))

    def test_cookie_is_sessionless(self):
        self.client.get(reverse('login'))
        self.assertEqual(self._count_sessions(), 0)

    def test_cookie_redirects_using_relative_path_only(self):
        path = '/path;part1;part2?q1=a&q2=b'
        next_url = urllib.quote_plus('http://example.com' + path)

        # testing indirectly via a decorated view, in order
        # to test the normal use case
        url = reverse('forgot_password') + '?next=' + next_url
        # trigger cookie check
        response = self.client.get(url)
        # view redirects to view for checking cookie support
        self.assertEqual(response.status_code, 302)
        self.assertIn('+cookie', response['location'])
        # follow redirect to next url
        response = self.client.get(response['location'])
        self.assertEqual(response.status_code, 302)
        # django will always build an absolte url when returning a 302 response
        # since this is being generated as part of the test run, the full
        # url will be http://testserver/...
        expected_url = 'http://testserver%s?next=%s' % (
            reverse('forgot_password'), urllib.quote_plus(path))
        self.assertEqual(response['location'], expected_url)

    def test_cookie_redirects_when_no_query_string(self):
        # testing indirectly via a decorated view, in order
        # to test the normal use case
        url = reverse('new_account')
        # trigger cookie check
        response = self.client.get(url)
        # view redirects to view for checking cookie support
        self.assertEqual(response.status_code, 302)
        self.assertIn('+cookie', response['location'])
        # follow redirect to next url
        response = self.client.get(response['location'])
        self.assertEqual(response.status_code, 302)
        # django will always build an absolte url when returning a 302 response
        # since this is being generated as part of the test run, the full
        # url will be http://testserver/...
        expected_url = 'http://testserver%s' % url
        self.assertEqual(response['location'], expected_url)


class LoginFlowStatsTestCase(BaseTestCase):

    args_list = partial(_args_list, flow_name='flows.login')

    def test_login_stats(self):
        self.client.get(reverse('login'))
        self.client.post(reverse('login'), self.data)
        self.assertEqual(self.mock_increment.call_args_list, self.args_list())

    def test_login_error_stats(self):
        self.client.get(reverse('login'))
        # make sure to pass invalid data
        self.client.post(reverse('login'), {'account': 'invalid'})
        self.assertEqual(self.mock_increment.call_args_list,
                         self.args_list(key='error'))


class ForgotPasswordFlowStatsTestCase(UIViewsBaseTestCase):

    args_list = partial(_args_list, flow_name='flows.forgot_password')

    def forgot_password(self, data=None):
        self.client.get(reverse('forgot_password'))
        if data is None:
            data = {'email': self.email}
        self.client.post(reverse('forgot_password'), data)

    def reset_password(self, valid_password=True):
        data = {'password': 'Password1', 'passwordconfirm': 'Password1'}
        if not valid_password:
            data['passwordconfirm'] = 'invalid'
        self.client.post(self._token_url(), data)

    def test_forgot_password(self):
        self.forgot_password()
        self.reset_password()
        self.assertEqual(self.mock_increment.call_args_list, self.args_list())

    def test_forgot_password_form_error(self):
        self.forgot_password(data={'foo': 'invalid'})
        self.assertEqual(self.mock_increment.call_args_list,
                         self.args_list(key='error.form'))

    def test_reset_password_token_error(self):
        self.patch(
            'webui.views.ui.verify_token_string', return_value=None)
        self.forgot_password()
        self.reset_password()
        self.assertEqual(self.mock_increment.call_args_list,
                         self.args_list(key='error.token'))

    def test_reset_password_form_error(self):
        self.forgot_password()
        self.reset_password(valid_password=False)
        self.assertEqual(self.mock_increment.call_args_list,
                         self.args_list(key='error.form'))


class BaseTwoFactorLoginTestCase(SSOBaseTestCase):
    email = 'user2f@example.com'
    password = DEFAULT_USER_PASSWORD
    logout_link = 'a[href="/+logout"]'
    url = reverse('twofactor')

    def setUp(self):
        super(BaseTwoFactorLoginTestCase, self).setUp()
        self.account = self.factory.make_account('2f User', self.email,
                                                 self.password)
        self.switch = switches(TWOFACTOR=True)
        self.switch.patch()
        self.addCleanup(self.switch.unpatch)
        self.preseed_cookie_check()

    def do_login(self, follow=False, **kwargs):
        data = {'email': self.email, 'password': self.password}
        token = kwargs.pop('token', None)
        data.update(kwargs)
        if token:
            url = reverse('login', kwargs={'token': token})
        else:
            url = reverse('login')
        return self.client.post(url, data, follow=follow)

    def do_twofactor(self, oath_token, next_url=None):
        data = {'oath_token': oath_token}
        if next_url:
            data['next'] = next_url
        return self.client.post(self.url, data)


class TwoFactorLoginTestCase(BaseTwoFactorLoginTestCase):

    def setUp(self):
        super(TwoFactorLoginTestCase, self).setUp()
        self.mock_site = self.patch(
            'webui.views.ui.twofactor.site_requires_twofactor_auth',
            return_value=True)
        self.mock_user = self.patch(
            'webui.views.ui.twofactor.user_requires_twofactor_auth',
            return_value=True)

    def test_twofactor_login_disabled_when_readonly(self):
        self.mock_site.return_value = False

        rm = ReadOnlyManager()
        rm.set_readonly()
        self.addCleanup(rm.clear_readonly)

        with switches(TWOFACTOR=True):
            self.do_login()
            response = self.client.get(self.url)
        self.assertEqual(response.status_code, 404)

    def test_when_site_does_not_require_twofactor_no_oath_field(self):
        self.mock_site.return_value = False

        response = self.client.get(reverse('login'))
        tree = PyQuery(response.content)
        self.assertEqual(len(tree.find('input[name="oath_token"]')), 0)

    def test_when_site_requires_twofactor_oath_field_is_present(self):
        response = self.client.get(reverse('login'))
        tree = PyQuery(response.content)
        inputs = tree.find('input[name="oath_token"]')
        self.assertEqual(len(inputs), 1)
        oath = inputs[0]
        self.assertEqual(oath.attrib['autocomplete'], 'off')

    def test_when_site_requires_twofactor_oath_field_is_present_lp(self):
        response = self.client.get(reverse('login'))
        tree = PyQuery(response.content)
        inputs = tree.find('input[name="oath_token"]')
        self.assertEqual(len(inputs), 1)
        oath = inputs[0]
        self.assertEqual(oath.attrib['autocomplete'], 'off')

    def test_when_user_requires_twofactor_redirected(self):
        self.mock_site.return_value = False
        r = self.do_login()
        self.assertRedirects(r, self.url)

    def test_when_user_requires_twofactor_redirected_with_token(self):
        self.mock_site.return_value = False
        token = 'a' * 16
        r = self.do_login(token=token, follow=True)
        self.assertRedirects(r, reverse('twofactor', kwargs=dict(token=token)))
        self.assertContains(r, token)

    def test_when_user_requires_twofactor_redirected_with_next(self):
        self.mock_site.return_value = False
        r = self.do_login(next=reverse('account-edit'))
        self.assertRedirects(r, '/two_factor_auth?next=/%2Bedit')

    def test_full_twofactor_login_redirected_to_next(self):
        mock_auth = self.patch(
            'webui.views.ui.twofactor.authenticate', return_value=True)
        r = self.do_login(next=reverse('account-edit'), oath_token='123456')
        mock_auth.assert_called_once_with(self.account, '123456')
        self.assertRedirects(r, reverse('account-edit'))

    def test_full_twofactor_login_illformed_token_displays_error(self):
        response = self.do_login(next=reverse('account-edit'),
                                 oath_token='XXX')
        tree = PyQuery(response.content)
        self.assertEqual(len(tree.find('input[name="oath_token"]')), 1)
        error = tree.find('#oathtoken span.error')
        self.assertEqual(len(error), 1)
        self.assertEqual(error[0].text.strip(), ONE_TIME_PWD_ERROR)

    def test_full_twofactor_login_invalid_code_displays_error(self):
        mock_auth = self.patch(
            'webui.views.ui.twofactor.authenticate',
            side_effect=ui.AuthenticationError("ERRORMSG"))
        response = self.do_login(next=reverse('account-edit'),
                                 oath_token='123456')
        tree = PyQuery(response.content)
        self.assertEqual(len(tree.find('input[name="oath_token"]')), 1)
        error = tree.find('#oathtoken span.error')
        self.assertEqual(len(error), 1)
        self.assertEqual(error[0].text.strip(), "ERRORMSG")
        mock_auth.assert_called_once_with(self.account, '123456')

    def test_full_twofactor_login_invalid_code_displays_unicode_error(self):
        self.patch(
            'webui.views.ui.twofactor.authenticate',
            side_effect=ui.AuthenticationError(u"客"))
        response = self.do_login(next=reverse('account-edit'),
                                 oath_token='123456')
        tree = PyQuery(response.content.decode('utf8'))
        self.assertEqual(len(tree.find('input[name="oath_token"]')), 1)
        error = tree.find('#oathtoken span.error')
        self.assertEqual(len(error), 1)
        self.assertTrue(error[0].text == u"客")

    def test_twofactor_step_redirected_to_next(self):
        mock_auth = self.patch(
            'webui.views.ui.twofactor.authenticate', return_value=True)
        self.client.login(username=self.email, password=self.password)
        r = self.do_twofactor(next_url=reverse('account-edit'),
                              oath_token='123456')
        self.assertRedirects(r, reverse('account-edit'))
        mock_auth.assert_called_once_with(self.account, '123456')

    def test_twofactor_step_illformed_token_displays_error(self):
        self.client.login(username=self.email, password=self.password)
        response = self.do_twofactor(next_url=reverse('account-edit'),
                                     oath_token='XXX')
        tree = PyQuery(response.content)
        self.assertEqual(len(tree.find('input[name="oath_token"]')), 1)
        error = tree.find('#oathtoken span.error')
        self.assertEqual(len(error), 1)
        self.assertEqual(error[0].text.strip(), ONE_TIME_PWD_ERROR)

    def test_twofactor_step_invalid_code_displays_error(self):
        mock_auth = self.patch(
            'webui.views.ui.twofactor.authenticate',
            side_effect=ui.AuthenticationError("ERRORMSG"))
        self.client.login(username=self.email, password=self.password)
        response = self.do_twofactor(next_url=reverse('account-edit'),
                                     oath_token='123456')
        tree = PyQuery(response.content)
        self.assertEqual(len(tree.find('input[name="oath_token"]')), 1)
        error = tree.find('#oathtoken span.error')
        self.assertEqual(len(error), 1)
        self.assertEqual(error[0].text.strip(), "ERRORMSG")
        mock_auth.assert_called_once_with(self.account, '123456')

    def test_two_factor_includes_rp_analytics(self):
        self.mock_site.return_value = False
        rpconfig = OpenIDRPConfig.objects.create(
            trust_root='http://example.com/',
            ga_snippet='[["_setAccount", "12345"]]')
        token = 'averyrandomtoken'
        response = self.do_login(token=token)
        self._assign_token_to_rpconfig(rpconfig, token=token)
        url = reverse('twofactor', kwargs=dict(token=token))
        self.assertRedirects(response, url)

        response = self.client.get(url)

        self.assertContains(response, "_gaq.push(['_setAccount', '12345']);")

    def test_twofactor_provides_logout(self):
        self.client.login(username=self.email, password=self.password)
        response = self.client.get(self.url)

        tree = PyQuery(response.content)
        logout_link = tree.find(self.logout_link)
        self.assertTrue(len(logout_link) > 0)

    def test_show_user_details(self):
        self.client.login(username=self.email, password=self.password)
        response = self.client.get(self.url)

        self.assertContains(response, self.account.displayname)

    def test_show_sync_link_on_the_side(self):
        self.client.login(username=self.email, password=self.password)
        response = self.client.get(self.url)

        self.assertContains(response, reverse('sync_2fa'))

    def test_shows_sync_link_when_otp_is_invalid_explicit(self):
        self.client.login(username=self.email, password=self.password)

        response = self.do_twofactor(oath_token='123456')

        self.assertContains(response, reverse('sync_2fa'))


class ResetPasswordLimitTestCase(SSOBaseTestCase):

    def request_password_reset(self, authtoken):
        url = reverse('reset_password', args=[authtoken, 'foo@foo.com'])
        response = self.client.get(url)
        return response

    def test_invalid_authtoken_below_threshold(self):
        self.patch(
            'webui.views.ui.ratelimit_extra.get_counters',
            return_value={None: 1})

        authtoken = 'invalid'
        response = self.request_password_reset(authtoken)
        self.assertRedirects(response, '/+bad-token')

    def test_too_many_invalid_authtokens(self):
        self.patch(
            'webui.views.ui.ratelimit_extra.get_counters',
            return_value={None: 20})

        authtoken = 'invalid'
        response = self.request_password_reset(authtoken)
        self.assertEqual(response.status_code, 403)


@override_settings(
    TWOFACTOR_FAILURE_NOTIFICATION_THRESHOLD=2,
    TWOFACTOR_FAILURE_NOTIFICATION_EMAILS=['someone@somewhere.com'],
    TWOFACTOR_FAILURE_NOTIFICATION_TEAMS=['foobarteam'])
class TwoFactorLoginWarningFailuresTestCase(BaseTwoFactorLoginTestCase):

    def setUp(self):
        super(TwoFactorLoginWarningFailuresTestCase, self).setUp()
        self.team = 'foobarteam'

        assert self.client.login(username=self.email, password=self.password)
        assert self.account.twofactor_attempts == 0
        assert not self.account.person_in_team(self.team)

    def set_threshold_delta(self, delta):
        threshold = settings.TWOFACTOR_FAILURE_NOTIFICATION_THRESHOLD
        self.account.twofactor_attempts = threshold + delta
        assert self.account.twofactor_attempts >= 0
        self.account.save()

    def set_team_to_account(self, team):
        self.factory.add_account_to_team(self.account,
                                         self.factory.make_team(team))
        assert self.account.person_in_team(team)

    def fail_twofactor_form_invalid(self):
        # the following will ensure an invalid form
        response = self.do_twofactor(oath_token='foobar')
        errors = response.context['form'].errors
        assert errors == {
            'oath_token': ['Please enter a 6-digit or 8-digit one-time '
                           'password']
        }

        return response

    def fail_twofactor_token_invalid(self):
        # form is valid but oath token is not valid
        response = self.do_twofactor(oath_token='123456')
        if 'form' not in response.context:
            self.fail('Response does not have a form: %s' % response)

        errors = response.context['form'].errors
        assert errors == {'__all__': ['The password is invalid']}

        return response

    def assert_no_emails_sent(self):
        self.assertEqual(len(mail.outbox), 0)

    def assert_too_many_failures_warning_sent(self, request):
        # reload account
        account = Account.objects.get(id=self.account.id)
        self.assertEqual(len(mail.outbox), 1)

        subject = mail.outbox[0].subject
        info = (account.twofactor_attempts, account.displayname)
        msg = 'Multiple 2FA failures (%s) for %s' % info
        self.assertEqual(msg, subject)

        body = mail.outbox[0].body
        msg += ' (lp_person=%s, openid=%s).' % (
            account.person.name, account.openid_identifier)
        self.assertIn(msg, body)
        for k, v in request.iteritems():
            self.assertIn(repr(k), body)
            self.assertIn(repr(v), body)

    def assert_account_suspended_warning_sent(self, account):
        self.assertEqual(len(mail.outbox), 1)

        subject = mail.outbox[0].subject
        msg = 'Account suspended for %s' % account.displayname
        self.assertEqual(msg, subject)

        body = mail.outbox[0].body
        msg += ' (lp_person=%s, openid=%s).' % (
            account.person.name, account.openid_identifier)
        self.assertIn(msg, body)
        reason = 'Too many 2FA failures (%s)' % account.twofactor_attempts
        self.assertIn(reason, body)

    def test_dont_notify_below_threshold_form_invalid(self):
        self.set_team_to_account(self.team)
        self.fail_twofactor_form_invalid()

        self.assert_no_emails_sent()

    def test_dont_notify_below_threshold_auth_error(self):
        self.set_team_to_account(self.team)
        self.fail_twofactor_token_invalid()

        self.assert_no_emails_sent()

    def test_team_mismatch_form_invalid(self):
        self.set_threshold_delta(-1)
        self.set_team_to_account('foobar')
        self.fail_twofactor_form_invalid()

        self.assert_no_emails_sent()

    def test_team_mismatch_auth_error(self):
        self.set_threshold_delta(-1)
        self.set_team_to_account('foobar')
        self.fail_twofactor_token_invalid()

        self.assert_no_emails_sent()

    def test_too_many_failures_form_invalid(self):
        self.set_threshold_delta(-1)
        self.set_team_to_account(self.team)
        response = self.fail_twofactor_form_invalid()

        self.assert_too_many_failures_warning_sent(request=response.request)

    def test_on_too_many_failures_auth_error(self):
        self.set_threshold_delta(-1)
        self.set_team_to_account(self.team)
        response = self.fail_twofactor_token_invalid()

        self.assert_too_many_failures_warning_sent(request=response.request)

    def test_do_not_notify_when_account_suspended_not_in_team(self):
        assert not self.account.person_in_team(self.team)

        with self.settings(TWOFACTOR_MAX_ATTEMPTS=1):
            response = self.do_twofactor(oath_token='123456')

        # reload account
        account = Account.objects.get(id=self.account.id)

        self.assertTemplateUsed(response, 'account/suspended.html')
        self.assertFalse(response.context['user'].is_authenticated())
        self.assertEqual(account.status, AccountStatus.SUSPENDED)
        # since the account is NOT on the expected team, no email is sent
        self.assert_no_emails_sent()

    def test_notify_when_account_suspended(self):
        self.set_team_to_account(self.team)

        with self.settings(TWOFACTOR_MAX_ATTEMPTS=1):
            response = self.do_twofactor(oath_token='123456')

        # reload account
        account = Account.objects.get(id=self.account.id)

        self.assertTemplateUsed(response, 'account/suspended.html')
        self.assertFalse(response.context['user'].is_authenticated())
        self.assertEqual(account.status, AccountStatus.SUSPENDED)
        # since the account is on the expected team, a notification is sent
        self.assert_account_suspended_warning_sent(account)


class TwoFactorFunctionsTestCase(SSOBaseTestCase):

    def _test_user_requires_twofactor_auth(
            self, require2f, has_devices, returns, teams=None):
        request = self.factory.make_request()

        if teams is None:
            teams = teams
        account = self.factory.make_account(
            twofactor_required=require2f, teams=teams)
        if has_devices:
            self.factory.make_device(account=account)
        assert account.has_twofactor_devices() == has_devices

        mock_flag = self.patch(
            'webui.views.ui.twofactor.is_twofactor_enabled',
            return_value=True)

        result = twofactor.user_requires_twofactor_auth(request, account)
        self.assertEqual(returns, result)
        mock_flag.assert_called_once_with(request)

        return account

    def test_not_require_twofactor_auth_with_devices(self):
        self._test_user_requires_twofactor_auth(
            require2f=False, has_devices=True, returns=False)

    def test_not_require_twofactor_auth_without_devices(self):
        self._test_user_requires_twofactor_auth(
            require2f=False, has_devices=False, returns=False)

    def test_require_twofactor_auth_without_devices(self):
        # if the user chose to force twofactor_required, ensure we make him
        # setup a device
        self._test_user_requires_twofactor_auth(
            require2f=True, has_devices=False, returns=True)

    def test_require_twofactor_auth_with_devices(self):
        self._test_user_requires_twofactor_auth(
            require2f=True, has_devices=True, returns=True)

    def test_require_twofactor_unset_mandatory_team_setting_empty(self):
        with self.settings(TWOFACTOR_MANDATORY_TEAMS=[]):
            self._test_user_requires_twofactor_auth(
                require2f=False, has_devices=False, returns=False)

    def test_require_twofactor_not_enforced_by_team_not_member(self):
        lucky_team = 'lucky-team'
        with self.settings(TWOFACTOR_MANDATORY_TEAMS=[lucky_team]):
            # account not in team, is not prompted for 2fa
            self._test_user_requires_twofactor_auth(
                require2f=False, has_devices=False, returns=False, teams=[])

    def test_require_twofactor_enforced_by_team_when_member(self):
        lucky_team = 'lucky-team'
        with self.settings(TWOFACTOR_MANDATORY_TEAMS=[lucky_team]):
            # account in team, is prompted even if twofactor_required is unset
            self._test_user_requires_twofactor_auth(
                require2f=False, has_devices=False, returns=True,
                teams=[lucky_team])

    def _make_mock_account(self, *args):
        account = self.factory.make_account()
        devices = [self.factory.make_device() for a in args]
        return account, devices

    def test_authenticate_with_no_valid_devices_throws(self):
        account = self.factory.make_account()
        self.factory.make_device(account=account)
        self.factory.make_device(account=account)

        self.assertRaises(
            ui.AuthenticationError, ui.twofactor.authenticate, account,
            '123456')

    def test_authenticate_with_first_device_succeeds(self):
        account = self.factory.make_account()
        device1 = self.factory.make_device(account=account)
        self.factory.make_device(account=account)

        otp = self.factory.make_next_otp(device1)
        self.assertTrue(ui.twofactor.authenticate(account, otp))

    def test_authenticate_with_last_device_succeeds(self):
        account = self.factory.make_account()
        self.factory.make_device(account=account)
        device2 = self.factory.make_device(account=account)

        otp = self.factory.make_next_otp(device2)
        self.assertTrue(ui.twofactor.authenticate(account, otp))


class ComboViewTestCase(SSOBaseTestCase):

    def test_combine_files(self):
        with self.settings(STATIC_ROOT='src/identityprovider/static'):
            response = self.client.get(
                settings.COMBO_URL +
                '?yui/3.15.0/widget-base/widget-base.js'
                '&identityprovider/js/ie/html5shiv.js')

        self.assertEqual(response['content-type'], 'text/javascript')
        self.assertNotContains(response, '[missing]')
        self.assertContains(
            response, '/* yui/3.15.0/widget-base/widget-base.js */')
        self.assertContains(
            response,
            '/* identityprovider/js/ie/html5shiv.js */')


class StaticPageTestCase(SSOBaseTestCase):

    def test_static_view_custom_context(self):
        response = self.client.get(reverse('ubuntuone-account'))

        # this is known to be true for this static page
        self.assertTrue(response.context['use_app_links'])

    def assertAvailable(self, url, text):
        response = self.client.get(url)
        self.assertContains(response, text)
        return response

    def assertNotAvailable(self, url):
        response = self.client.get(url)
        self.assertEqual(response.status_code, 404)
        return response

    def test_openid(self):
        response = self.client.get('/+openid')
        self.assertContains(response, 'This is Ubuntu One')

    def test_faq_page(self):
        response = self.assertAvailable(
            reverse('faq'), "Frequently asked questions")
        self.assertTemplateUsed(response, 'static/faq.html')

    def test_description_page(self):
        self.assertAvailable('/+description', "What's this?")

    def test_applications_page(self):
        account = self.factory.make_account(password=DEFAULT_USER_PASSWORD)
        assert self.client.login(
            username=account.user.email, password=DEFAULT_USER_PASSWORD)
        self.assertAvailable('/+applications', 'Applications you use')

    def test_logout(self):
        account = self.factory.make_account(password=DEFAULT_USER_PASSWORD)
        assert self.client.login(
            username=account.user.email, password=DEFAULT_USER_PASSWORD)
        self.assertAvailable('/', 'edit_account')

        # Then we log out
        response = self.client.get('/+logout')
        # and verify the logout works
        self.assertContains(response, 'You have been logged out')


class ConfirmGPGKeyClaimingBaseTestCase(SSOBaseTestCase):

    def setUp(self):
        super(ConfirmGPGKeyClaimingBaseTestCase, self).setUp()
        # Child classes provide a :cvar: but we want to shadow that with a
        # version that's in the 'internal' format to ensure that we're
        # presenting the fingerprint correctly everywhere in the UI:
        self.fingerprint = self.fingerprint.replace(' ', '')

        # To make it possible to test sign-only keys, ensure that the
        # email address of the main user is predictable:
        self.account = self.factory.make_account(
            password=DEFAULT_USER_PASSWORD,
            email='someone@somewhere.com',
        )
        # ...and likewise for the date the token was created:
        self.token = self.factory.make_authtoken(
            requester=self.account,
            token_type=self.token_type,
            email=self.account.preferredemail,
            fingerprint=self.fingerprint,
            date_created=datetime(2016, 4, 12, 16, 0, 0),
        )
        self.preseed_cookie_check()

    def login_as_account(self, account):
        assert self.client.login(
            username=account.preferredemail.email,
            password=DEFAULT_USER_PASSWORD)

    def get_claim_url_with_options(self, authtoken=None, fingerprint=None):
        """Get the URL for claiming a GPG token.

        If no options are specified, the returned url will be for the current
        token and fingerprint under test.
        """
        if authtoken is None:
            authtoken = self.token.raw_token
        if fingerprint is None:
            fingerprint = self.fingerprint
        return reverse(self.view, kwargs={
            'authtoken': authtoken,
            'fingerprint': fingerprint
        })

    def get_claim_token_view(self, **kwargs):
        return self.client.get(self.get_claim_url_with_options(**kwargs))

    def post_claim_token_view(self, **kwargs):
        follow = kwargs.pop('follow', False)
        data = kwargs.pop('data', dict())
        return self.client.post(
            self.get_claim_url_with_options(**kwargs),
            data=data,
            follow=follow,
        )


class ConfirmGPGKeyViewTestCase(ConfirmGPGKeyClaimingBaseTestCase):

    token_type = AuthTokenType.VALIDATEGPG
    fingerprint = SampleDataKeys.regular_key
    view = 'confirm_gpg_key'

    def test_must_be_authenticated_to_view_claim_token_with_get(self):
        self._test_must_be_authenticated(self.get_claim_token_view)

    def test_must_be_authenticated_to_view_claim_token_with_port(self):
        self._test_must_be_authenticated(self.post_claim_token_view)

    def _test_must_be_authenticated(self, view_func):
        resp = view_func()
        next_url = self.get_claim_url_with_options()
        self.assertRedirects(
            resp,
            reverse('login') + '?next=' + quote(next_url))

    def test_fails_with_incorrect_fingerprint_with_get(self):
        self._test_fails_with_incorrect_fingerprint(self.get_claim_token_view)

    def test_fails_with_incorrect_fingerprint_with_post(self):
        self._test_fails_with_incorrect_fingerprint(self.post_claim_token_view)

    def _test_fails_with_incorrect_fingerprint(self, view_func):
        self.login_as_account(self.account)
        wrong_fingerprint = 'A' * 40
        resp = view_func(fingerprint=wrong_fingerprint)
        self.assertRedirects(resp, '/+bad-token')

    def test_fails_with_bad_token_code_with_get(self):
        self._test_fails_with_bad_token_code(self.get_claim_token_view)

    def test_fails_with_bad_token_code_with_post(self):
        self._test_fails_with_bad_token_code(self.post_claim_token_view)

    def _test_fails_with_bad_token_code(self, view_func):
        self.login_as_account(self.account)
        wrong_token = 'ThisIsNotTheTokenYouAreLookingFor'
        resp = view_func(authtoken=wrong_token)
        self.assertRedirects(resp, '/+bad-token')

    def test_fails_with_bad_token_type_with_get(self):
        self._test_fails_with_bad_token_type(self.get_claim_token_view)

    def test_fails_with_bad_token_type_with_post(self):
        self._test_fails_with_bad_token_type(self.post_claim_token_view)

    def _test_fails_with_bad_token_type(self, view_func):
        self.login_as_account(self.account)
        bad_token = self.factory.make_authtoken(
            requester=self.account,
            token_type=AuthTokenType.VALIDATESIGNONLYGPG,
            email=self.account.preferredemail,
            fingerprint=self.fingerprint,
        )
        resp = view_func(authtoken=bad_token.raw_token)
        self.assertRedirects(resp, '/+bad-token')

    def test_token_is_destroyed_when_claimed_by_wrong_user_with_get(self):
        self._test_token_is_destroyed_when_claimed_by_wrong_user(
            self.get_claim_token_view)

    def test_token_is_destroyed_when_claimed_by_wrong_user_with_post(self):
        self._test_token_is_destroyed_when_claimed_by_wrong_user(
            self.post_claim_token_view)

    def _test_token_is_destroyed_when_claimed_by_wrong_user(self, view_func):
        wrong_user = self.factory.make_account(
            password=DEFAULT_USER_PASSWORD)
        self.login_as_account(wrong_user)
        resp = view_func()

        self.assertEqual(404, resp.status_code)
        # ensure the token was deleted:
        self.assertRaises(
            AuthToken.DoesNotExist,
            AuthToken.get_token,
            fingerprint=self.fingerprint
        )

    def test_can_render_confirmation_screen(self):
        self.login_as_account(self.account)
        resp = self.get_claim_token_view()
        expected = '<h1 class="u1-h-main">Validate OpenPGP Key</h1>'
        expected2 = '<p>{fingerprint}</p>'.format(
            fingerprint=gpg_messages.display_fingerprint(
                SampleDataKeys.regular_key))
        self.assertContains(resp, expected, html=True)
        self.assertContains(resp, expected2, html=True)

    def test_can_consume_token(self):
        self.use_fixture(SSOGPGServiceFixture())
        self.login_as_account(self.account)
        resp = self.post_claim_token_view(follow=True)
        token = AuthToken.get_token(fingerprint=self.fingerprint)

        self.assertIsNotNone(token.date_consumed)
        self.assertContains(
            resp, gpg_messages.key_added_to_account(self.fingerprint))


class ConfirmSignOnlyGPGKeyViewTestCase(ConfirmGPGKeyClaimingBaseTestCase):

    token_type = AuthTokenType.VALIDATESIGNONLYGPG
    fingerprint = SampleDataKeys.sign_only_key
    view = 'confirm_signonly_gpg_key'

    def test_must_be_authenticated_to_claim_token_with_get(self):
        self._test_must_be_authenticated_to_claim_token(
            self.get_claim_token_view)

    def test_must_be_authenticated_to_claim_token_with_post(self):
        self._test_must_be_authenticated_to_claim_token(
            self.post_claim_token_view)

    def _test_must_be_authenticated_to_claim_token(self, view_func):
        resp = view_func()
        next_url = self.get_claim_url_with_options()
        self.assertRedirects(
            resp,
            reverse('login') + '?next=' + quote(next_url))

    def test_fails_with_incorrect_fingerprint_with_get(self):
        self._test_fails_with_incorrect_fingerprint(self.get_claim_token_view)

    def test_fails_with_incorrect_fingerprint_with_post(self):
        self._test_fails_with_incorrect_fingerprint(self.post_claim_token_view)

    def _test_fails_with_incorrect_fingerprint(self, view_func):
        self.login_as_account(self.account)
        wrong_fingerprint = 'A' * 40
        resp = self.post_claim_token_view(fingerprint=wrong_fingerprint)
        self.assertRedirects(resp, '/+bad-token')

    def test_fails_with_bad_token_code_with_get(self):
        self._test_fails_with_bad_token_code(self.get_claim_token_view)

    def test_fails_with_bad_token_code_with_post(self):
        self._test_fails_with_bad_token_code(self.post_claim_token_view)

    def _test_fails_with_bad_token_code(self, view_func):
        self.login_as_account(self.account)
        wrong_token = 'ThisIsNotTheTokenYouAreLookingFor'
        resp = view_func(authtoken=wrong_token)
        self.assertRedirects(resp, '/+bad-token')

    def test_fails_with_bad_token_type_with_get(self):
        self._test_fails_with_bad_token_type(self.get_claim_token_view)

    def test_fails_with_bad_token_type_with_post(self):
        self._test_fails_with_bad_token_type(self.post_claim_token_view)

    def _test_fails_with_bad_token_type(self, view_func):
        self.login_as_account(self.account)
        bad_token = self.factory.make_authtoken(
            requester=self.account,
            token_type=AuthTokenType.VALIDATEGPG,
            email=self.account.preferredemail,
            fingerprint=self.fingerprint,
        )
        resp = view_func(authtoken=bad_token.raw_token)
        self.assertRedirects(resp, '/+bad-token')

    def test_token_is_destroyed_when_claimed_by_wrong_user_with_get(self):
        self._test_token_is_destroyed_when_claimed_by_wrong_user(
            self.get_claim_token_view)

    def test_token_is_destroyed_when_claimed_by_wrong_user_with_post(self):
        self._test_token_is_destroyed_when_claimed_by_wrong_user(
            self.post_claim_token_view)

    def _test_token_is_destroyed_when_claimed_by_wrong_user(self, view_func):
        wrong_user = self.factory.make_account(
            password=DEFAULT_USER_PASSWORD)
        self.login_as_account(wrong_user)
        resp = view_func()
        self.assertEqual(404, resp.status_code)
        # ensure the token was deleted:
        self.assertRaises(
            AuthToken.DoesNotExist,
            AuthToken.get_token,
            fingerprint=self.fingerprint
        )

    def test_can_render_confirmation_screen(self):
        self.login_as_account(self.account)
        resp = self.get_claim_token_view()
        expected = '<h1 class="u1-h-main">Validate OpenPGP Key %s?</h1>'\
            % gpg_messages.display_fingerprint(SampleDataKeys.sign_only_key)
        self.assertContains(resp, expected, html=True)

    def test_can_consume_token(self):
        self.use_fixture(SSOGPGServiceFixture())
        self.login_as_account(self.account)
        resp = self.post_claim_token_view(
            follow=True,
            data={'clearsigned': CORRECT_TEXT_SIGNED_WITH_CORRECT_KEY}
        )
        token = AuthToken.get_token(fingerprint=self.fingerprint)
        self.assertIsNotNone(token.date_consumed)
        self.assertContains(
            resp, gpg_messages.key_added_to_account(self.fingerprint))


# This text block was pre-signed with the sign-only key. The tests above use
# it to verify that the view works as expected. A more complete set of tests
# (including various GPG-related error cases) can be found in
# webui.tests.test_forms.
CORRECT_TEXT_SIGNED_WITH_CORRECT_KEY = dedent("""\
    -----BEGIN PGP SIGNED MESSAGE-----
    Hash: SHA1

    Please register 447DBF38C4F9C4ED752246B77D88913717B05A8F to the
    Ubuntu One user with the email address someone@somewhere.com.
    2016-04-12 16:00:00
    -----BEGIN PGP SIGNATURE-----
    Version: GnuPG v1

    iEYEARECAAYFAlcsCwIACgkQfYiRNxewWo/oVgCfcCAoLug0dsVclV0xMKsLlnOY
    bgYAn0faP2lW74vKDkQSmr6gGZvn9tEm
    =KAby
    -----END PGP SIGNATURE-----""")
