# Copyright 2010, 2012 Canonical Ltd.  This software is licensed under
# the GNU Affero General Public License version 3 (see the file
# LICENSE).

from __future__ import unicode_literals

import logging

from urllib import urlencode
from urlparse import parse_qs, urlparse, urlunparse

from django.conf import settings
from django.contrib import auth, messages
from django.core.urlresolvers import reverse
from django.http import HttpResponseRedirect
from django.shortcuts import get_object_or_404, render
from django.template.loader import render_to_string
from django.template.response import TemplateResponse
from django.utils.timezone import now
from django.utils.translation import ugettext as _
from django.views.decorators.vary import vary_on_headers
from gargoyle.decorators import switch_is_active
from gargoyle import gargoyle
from openid.yadis.constants import YADIS_HEADER_NAME, YADIS_CONTENT_TYPE

from identityprovider.cookies import set_test_cookie, test_cookie_worked
from identityprovider.emailutils import (
    send_invalidation_email_notice,
    send_notification_to_invalidated_email_address,
    send_validation_email_request,
    send_gpg_validation_message,
)
from identityprovider.forms import (
    DeleteAccountForm,
    EditAccountForm,
    NewEmailForm,
)
from identityprovider.models import (
    Account,
    AuthToken,
    EmailAddress,
)
from identityprovider.models.authtoken import get_authtoken_or_404
from identityprovider.models.const import (
    AccountStatus,
    AuthLogType,
    AuthTokenType,
    EmailStatus,
)
from identityprovider.models.twofactor import is_twofactor_enabled
from identityprovider.signals import (
    account_email_added,
    set_session_oauth_token,
)
from identityprovider.utils import redirection_url_for_token
from identityprovider.validators import PASSWORD_LEAKED
from identityprovider.views.server import xrds
from identityprovider.views.utils import (
    get_rpconfig_from_request,
    require_testing_enabled,
)

from webui.constants import EMAIL_EXISTS_ERROR, VERIFY_EMAIL_MESSAGE
from webui.decorators import check_readonly, sso_login_required
from webui.forms import (
    ClaimOpenPGPKeyForm,
    DisabledGPGKeysForm,
    EnabledGPGKeysForm,
    PendingGPGValidationTokenForm,
)
from webui.gpg import (
    get_user_id_for_account,
    SSOGPGClient,
)
from webui.views import gpg_messages
from webui.views.const import (
    DETAILS_UPDATED,
    EMAIL_DELETED,
    TOKEN_DELETED,
    VALIDATE_EMAIL,
    VALIDATE_EMAIL_DESC,
)
from webui.views.ui import LoginView
from webui.views.utils import (
    display_email_sent,
    set_session_email
)

logger = logging.getLogger(__name__)


@vary_on_headers('Accept')
def index(request, token=None):
    accept = request.META.get('HTTP_ACCEPT', '')
    if YADIS_CONTENT_TYPE in accept:
        return xrds(request)

    if request.user.is_authenticated():
        resp = account_edit(request, token)
    else:
        # We simply render LoginView's get() method as any posted
        # login data will be handled by the configured 'login' url.
        # (ie. /+login currently).
        resp = LoginView(request=request).get(request, token=token)
        # Perhaps the browser sent us the test cookie, or perhaps it
        # didn't.  We'll tell Django that we want one, and it will be
        # intelligent and not send it out again if the browser *did*
        # send it to us.
        set_test_cookie(resp)
    resp[YADIS_HEADER_NAME] = request.build_absolute_uri(
        reverse('server-xrds-discovery'))
    return resp


def make_relative_url(parts, **overrides):
    new_parts = {
        'path': parts.path,
        'params': parts.params,
        'query': parts.query,
        'fragment': parts.fragment,
    }
    new_parts.update(**overrides)
    relative = urlunparse(('', '', new_parts['path'], new_parts['params'],
                           new_parts['query'], new_parts['fragment']))
    return relative


def make_relative_qs(qs):
        qs_parts = parse_qs(qs)
        qs_next = qs_parts.get('next')
        if qs_next:
            qs_next = [make_relative_url(urlparse(url)) for url in qs_next]
            if len(qs_next) == 1:
                qs_next = qs_next[0]
        new_qs_parts = qs_parts.copy()
        new_qs_parts['next'] = qs_next

        new_qs = urlencode(new_qs_parts)
        return new_qs


def cookie(request):
    next_url = request.GET.get('next', '/')
    if test_cookie_worked(request):
        parts = urlparse(next_url)
        overrides = {}
        if parts.query:
            # process next qs param
            overrides['query'] = make_relative_qs(parts.query)
        # convert to relative path
        next_url = make_relative_url(parts, **overrides)
        return HttpResponseRedirect(next_url)

    # Send the cookie again, so that the user can hit reload on the
    # page we return to continue on.
    return set_test_cookie(render(request, 'cookies.html'))


@sso_login_required
def account_edit(request, token=None):
    rpconfig = get_rpconfig_from_request(request, token)
    account = request.user

    msg = 'Called account_edit with request.user being None. Impossible!'
    assert account is not None, msg

    enable_device_prefs = (is_twofactor_enabled(request) and
                           account.devices.count() > 0)
    enable_username = gargoyle.is_active('USERNAME_UI', request)
    if request.method == 'POST' and not settings.READ_ONLY_MODE:
        form = EditAccountForm(data=request.POST, instance=account,
                               enable_device_prefs=enable_device_prefs,
                               enable_username=enable_username)
        if form.is_valid():
            form.save()
            if form.password_changed:
                set_session_oauth_token(Account, account, request)
            messages.success(request, DETAILS_UPDATED)
            return HttpResponseRedirect(request.path)
    else:
        form = EditAccountForm(instance=account,
                               enable_device_prefs=enable_device_prefs,
                               enable_username=enable_username,
                               initial={'username': account.person_name})

    need_backup_device_warning = (enable_device_prefs and
                                  account.need_backup_device_warning)

    # if user has just registered, do not show warning about unverified email
    new_account_url = request.build_absolute_uri(reverse('new_account'))
    new_account = (request.META.get('HTTP_REFERER') == new_account_url)
    preferred_email = account.preferredemail  # cache it, one access to the DB
    verified = preferred_email and preferred_email.is_verified
    need_verify_email_warning = not verified and not new_account

    if enable_device_prefs:
        paper_renewals = list(request.user.paper_devices_needing_renewal)
    else:
        paper_renewals = []

    password_leaked = unicode(PASSWORD_LEAKED) in form.errors.get(
        'password', '')
    if password_leaked:
        error_msg = render_to_string('common/password_leaked_warning.txt',
                                     context={}, request=request)
        messages.warning(request, error_msg)

    context = {
        'current_section': 'account',
        'account': account,
        'account_displayname': account.displayname,
        'last_authenticated_sites': account.last_authenticated_sites(limit=10),
        'unverified_emails': account.unverified_emails(),
        'enable_device_prefs': enable_device_prefs,
        'need_backup_device_warning': need_backup_device_warning,
        'paper_devices_needing_renewal': paper_renewals,
        'need_verify_email_warning': need_verify_email_warning,
        'unverified_email': preferred_email,
        'form': form,
        'token': token,
        'rpconfig': rpconfig,
    }
    return render(request, 'account/edit.html', context)


@sso_login_required
def delete_account(request):
    account = request.user
    if request.method == 'POST' and not settings.READ_ONLY_MODE:
        account.set_deleted()
        auth.logout(request)

        return render(request, 'account/account_deleted.html')

    form = DeleteAccountForm()
    sites = account.last_authenticated_sites(limit=None)
    context = {
        'last_authenticated_sites': sites,
        'account': account,
        'form': form,
    }
    return render(request, 'account/delete.html', context)


@sso_login_required
def account_emails(request):
    account = request.user
    form = NewEmailForm()
    verified_emails = account.verified_emails()
    unverified_emails = account.unverified_emails()
    context = {
        'current_section': 'emails',
        'account': account,
        'account_displayname': account.displayname,
        'verified_emails': verified_emails,
        'unverified_emails': unverified_emails,
        'form': form,
        'num_emails': verified_emails.count() + unverified_emails.count(),
    }
    return render(request, 'account/emails.html', context)


@require_testing_enabled
@sso_login_required
def account_deactivate(request, token=None):
    account = request.user
    for email in account.emailaddress_set.all():
        email.status = EmailStatus.NEW
        email.save()
    account.status = AccountStatus.DEACTIVATED
    account.save()

    # We don't want to lose session[token] when we log the user out
    raw_orequest = request.session.get(token, None)
    auth.logout(request)
    if token is not None and raw_orequest is not None:
        request.session[token] = raw_orequest
    return HttpResponseRedirect(reverse('deactivated'))


def _send_verification_email(request, email, tokenid=None):
    # TODO: This makes use of `tokenid` if it's passed in, but the
    # call to `send_validation_email_request` always generates a new
    # token.  Is this right?

    # If there are any unverified emails that match they should be deleted.
    EmailAddress.objects.filter(email__iexact=email,
                                status=EmailStatus.NEW).delete()

    # Ensure that this account has such email address; return value is not used
    request.user.emailaddress_set.get_or_create(
        email=email, status=EmailStatus.NEW)

    redirection_url = redirection_url_for_token(tokenid)

    root_url = request.build_absolute_uri('/')
    send_validation_email_request(root_url, request.user, email,
                                  redirection_url)
    set_session_email(request.session, email)

    return display_email_sent(
        request,
        email,
        VALIDATE_EMAIL,
        VALIDATE_EMAIL_DESC.format(email_to=email,
                                   email_from=settings.NOREPLY_FROM_ADDRESS)
    )


@sso_login_required
@check_readonly
def new_email(request, token=None, emailid=None):
    form = NewEmailForm()

    if request.method == 'POST' and not settings.READ_ONLY_MODE:
        form = NewEmailForm(request.POST, account=request.user)
        if form.is_valid():
            email = form.cleaned_data['newemail']
            account_email_added.send(
                openid_identifier=request.user.openid_identifier,
                sender=None)
            return _send_verification_email(request, email, token)
        else:
            err = form.errors.get('newemail', [''])[0]
            if err == VERIFY_EMAIL_MESSAGE:
                messages.error(request, EMAIL_EXISTS_ERROR)

    context = {'form': form}
    return render(request, 'account/new_email.html', context)


@sso_login_required
def verify_email_by_id(request, token=None):
    emailid = request.GET.get('id', 0)
    email = get_object_or_404(request.user.emailaddress_set, pk=emailid,
                              status=EmailStatus.NEW)
    return _send_verification_email(request, email.email, token)


@sso_login_required
def verify_email(request, email_address, token=None):
    email = get_object_or_404(
        request.user.emailaddress_set, email=email_address,
        status=EmailStatus.NEW)
    return _send_verification_email(request, email.email, token)


def invalidate_email(request, authtoken, email_address):
    authtoken = get_authtoken_or_404(
        raw_token=authtoken, email=email_address,
        token_type=AuthTokenType.INVALIDATEEMAIL)
    if not authtoken.active:
        return HttpResponseRedirect(reverse('bad_token'))

    if request.method == 'GET':
        return TemplateResponse(
            request, 'account/confirm_email_invalidation.html',
            context=dict(email=email_address))

    authtoken.consume()

    # consume any other active token for the same email address
    for t in AuthToken.objects.filter(email=email_address, date_consumed=None):
        t.consume()

    auth.logout(request)
    response = TemplateResponse(request, 'account/invalidate_email.html',
                                context=dict(email=email_address))

    # email may not exist since the user may have tried an operation with
    # an email unknown to our system
    emails = EmailAddress.objects.filter(email__iexact=email_address)
    if emails.count() == 0:
        # This case matches, for example, registration web where the Account
        # is not created until the user validates the email address.
        # To be changed pretty soon (nessita, 2013-01-11).
        logger.info('Received a request to invalidate email %r, but a matching'
                    ' EmailAddress does not exist in the system.',
                    email_address)
        return response

    if emails.count() > 1:
        logger.warning('While trying to invalidate email %r, found more than '
                       '1 EmailAddress matching the email.', email_address)

    logger.info('Setting email %r status to invalidated.', email_address)
    email = emails[0]
    account = email.account
    email.invalidate()

    if account is None:
        logger.warning("Received a request to invalidate email %r, but the "
                       "email's account is None.", email_address)
        return response

    # notify the account of the email invalidation, prioritize validated emails
    root_url = request.build_absolute_uri('/')
    if account.preferredemail is not None:
        send_invalidation_email_notice(
            root_url,
            account,
            invalidated_email=email.email,
        )
    else:
        send_notification_to_invalidated_email_address(
            root_url,
            account,
            invalidated_email=email.email,
        )

    # invalidate all tokens by deleting them
    logger.info('Invalidating oauth tokens for %r', email_address)
    account.invalidate_oauth_tokens()

    return response


@sso_login_required
def delete_email_by_id(request, token=None):
    emailid = request.GET.get('id', 0)
    email = get_object_or_404(request.user.emailaddress_set, pk=emailid)
    return delete_email(request, email.email, token)


@sso_login_required
def delete_email(request, email_address, token=None):
    email = get_object_or_404(
        request.user.emailaddress_set, email=email_address,
        status__in=[EmailStatus.NEW, EmailStatus.VALIDATED])
    if request.method == 'POST':
        redirection_url = (
            redirection_url_for_token(token) if token
            else reverse('account-emails'))
        email.delete()
        messages.success(
            request, EMAIL_DELETED.format(email=email.email), 'temporary')
        return HttpResponseRedirect(redirection_url)

    context = {'email': email.email}
    return render(request, 'account/delete_email.html', context)


@sso_login_required
@check_readonly
def applications(request):
    if request.method == 'POST':
        token_id = request.POST.get('token_id')
        if token_id:
            token = request.user.oauth_token_by_key(token_id)
            if token is not None:
                token.delete()
                messages.success(request,
                                 TOKEN_DELETED.format(name=token.name))
        return HttpResponseRedirect('/+applications')
    else:
        tokens = [
            dict(name=t.name, date_created=t.date_created, pk=t.pk)
            for t in request.user.oauth_tokens()
        ] + [
            dict(name=t.name, date_created=t.created_at, pk=t.pk)
            for t in request.user.oauth_tokens(scope=None)
        ]
        context = {
            'current_section': 'applications',
            'tokens': tokens,
        }
        return render(request, 'account/applications.html', context)


@sso_login_required
def auth_log(request):
    full_auth_log = request.user.authlog_set.exclude(
        log_type__in=(AuthLogType.OPENID_FAIL,
                      AuthLogType.API_FAIL)).order_by('-when_created')
    full_auth_log_length = full_auth_log.count()
    auth_log_items = full_auth_log[:settings.SHOW_USER_NUM_AUTHLOGS]
    context = {
        'current_section': 'auth_log',
        'auth_log': auth_log_items,
        'full_auth_log_length': full_auth_log_length,
    }
    return render(request, 'account/auth_log.html', context)


@switch_is_active('GPG_SERVICE')
@sso_login_required
@check_readonly
def gpg_keys(request):
    context = dict(current_section='gpg_keys')
    if 'claim_gpg' in request.POST:
        claim_form = ClaimOpenPGPKeyForm(request.POST)
        if claim_form.is_valid():
            _initiate_gpg_key_claim(request, claim_form)
            return HttpResponseRedirect(reverse('gpg_keys'))
    else:
        claim_form = ClaimOpenPGPKeyForm()

    if 'cancel_pending' in request.POST:
        pending_form = PendingGPGValidationTokenForm(
            request.user, request.POST)
        if pending_form.is_valid():
            _cancel_pending_tokens(request, pending_form)
            return HttpResponseRedirect(reverse('gpg_keys'))
    else:
        pending_form = PendingGPGValidationTokenForm(request.user)

    if 'disable_keys' in request.POST:
        disable_form = EnabledGPGKeysForm(request.user, data=request.POST)
        if disable_form.is_valid():
            _deactivate_keys(request, disable_form)
            return HttpResponseRedirect(reverse('gpg_keys'))
    else:
        disable_form = EnabledGPGKeysForm(request.user)

    if 'reenable_keys' in request.POST:
        reactivate_form = DisabledGPGKeysForm(request.user, data=request.POST)
        if reactivate_form.is_valid():
            _reenable_keys(request, reactivate_form)
            return HttpResponseRedirect(reverse('gpg_keys'))
    else:
        reactivate_form = DisabledGPGKeysForm(request.user)

    context['claim_form'] = claim_form
    context['pending_form'] = pending_form
    context['disable_keys_form'] = disable_form
    context['reenable_keys_form'] = reactivate_form
    context['preferred_email'] = request.user.preferredemail

    return render(request, 'account/gpg_keys.html', context)


def _initiate_gpg_key_claim(request, claim_form):
    """Initiate the GPG claim form process.

    This function will create a GPG key verification token of the correct
    type, Send an email to the logged in users preferred email address, and
    add a message to the django message system.

    :param request: The current request being served.
    :param claim_form: A *valid* instance of ClaimOpenPGPKeyForm.
    """
    assert claim_form.is_valid(), "Claim form must be valid."
    client = SSOGPGClient()
    key_details = claim_form.cleaned_data['key_details']
    if key_details['can_encrypt']:
        token_type = AuthTokenType.VALIDATEGPG
    else:
        token_type = AuthTokenType.VALIDATESIGNONLYGPG
    token = AuthToken.objects.create(
        token_type=token_type,
        email=request.user.preferredemail.email,
        fingerprint=claim_form.cleaned_data['fingerprint'],
        displayname=key_details['displayname'],
        date_created=now(),
        requester=request.user,
    )
    send_gpg_validation_message(
        settings.SSO_ROOT_URL, key_details, token,
        request.user, client)
    messages.success(
        request,
        _("An email with further instructions has been sent to "
          "{preferred_email}").format(
            preferred_email=request.user.preferredemail))


def _cancel_pending_tokens(request, pending_form):
    """Cancel pending GPG verification tokens.

    This function will delete the pending AuthTokens listed in the
    'pending_form's 'tokens' field. It also adds a message to the django
    message system.

    :param request: The currently served request.
    :param pending_form: A *valid* instance of the
        PendingGPGValidationTokenForm.
    """
    assert pending_form.is_valid(), "Pending token form must be valid."
    tokens = pending_form.cleaned_data['tokens']
    fingerprints = []
    for token in tokens:
        fingerprints.append(token.fingerprint)
        token.delete()
    messages.success(
        request, gpg_messages.pending_tokens_cancelled(fingerprints))


def _deactivate_keys(request, disable_form):
    """Deactivate active GPG keys.

    This function will mark keys as disabled for the current logged in user.

    :param request: The currently served request.
    :param disable_form: A valid instance of the EnabledGPGKeysForm.
    """
    assert disable_form.is_valid(), "Disable keys form must be valid."
    key_fingerprints = disable_form.cleaned_data['keys']
    gpg_client = SSOGPGClient()
    user_id = get_user_id_for_account(request.user)
    for fingerprint in key_fingerprints:
        gpg_client.disableKeyForOwner(user_id, fingerprint)
    messages.success(request, gpg_messages.keys_disabled(key_fingerprints))


def _reenable_keys(request, reactivate_form):
    """Reactivate previously deactivated GPG keys.

    :param request: The currently served request.
    :param reactivate_form: A valid instance of the DisabledGPGKeysForm.
    """
    assert reactivate_form.is_valid(), "Claim form must be valid."
    client = SSOGPGClient()
    fingerprints = reactivate_form.cleaned_data['keys']
    for fingerprint in fingerprints:
        key_details = client.getKeyDetailsFromKeyServer(fingerprint)
        if key_details['can_encrypt']:
            token_type = AuthTokenType.VALIDATEGPG
        else:
            token_type = AuthTokenType.VALIDATESIGNONLYGPG
        token = AuthToken.objects.create(
            token_type=token_type,
            email=request.user.preferredemail.email,
            fingerprint=fingerprint,
            date_created=now(),
            requester=request.user,
        )
        send_gpg_validation_message(
            settings.SSO_ROOT_URL, key_details, token,
            request.user, client)
    messages.success(
        request,
        gpg_messages.keys_reenabled(
            fingerprints, request.user.preferredemail.email))


@switch_is_active('SSH_KEY_INTEGRATION')
@sso_login_required
@check_readonly
def ssh_keys(request):
    return render(request, 'account/ssh_keys.html', {})
