# Copyright 2012 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

from __future__ import unicode_literals

import re
from base64 import b16encode
from collections import namedtuple

from django.contrib import messages
from django.contrib.auth.views import redirect_to_login
from django.core.urlresolvers import reverse
from django.conf import settings
from django.http import HttpResponseRedirect
from django.shortcuts import get_object_or_404, render
from django.template.response import TemplateResponse
from django.views.generic import View
from django.utils.translation import ugettext_lazy as _
from gargoyle import gargoyle
from M2Crypto.Rand import rand_bytes
from oath import accept_hotp, hotp

from identityprovider.forms import HOTPDeviceForm, DeviceRenameForm
from identityprovider.models import AuthenticationDevice
from identityprovider.models import twofactor
from identityprovider.models.twofactor import get_otp_type

from webui.decorators import require_twofactor_enabled, sso_login_required
from webui.views.const import (
    DEVICE_ADDED,
    DEVICE_DELETED,
    DEVICE_GENERATION_WARNING,
    DEVICE_RENAMED,
    OTP_MATCH_ERROR,
)
from webui.views.utils import HttpResponseSeeOther, allow_only


DEVICE_ADDITION = 'device-addition'
DEVICE_LIST = 'device-list'
CodePageDetails = namedtuple('CodePageDetails', 'codes page start position')


def context_with_current_section(request, **kwargs):
    kwargs['current_section'] = 'devices'
    return kwargs


device_types = {
    'yubi': _('Yubikey'),
    'google': _('Google Authenticator'),
    'generic': _('Authentication device'),
    'paper': _('Printable Backup Codes'),
}


def generate_key(n):
    """Returns an OATH/HOTP key as a string of n raw bytes."""

    # An OATH/HOTP key is just bunch of random (in the "unpredictable"
    # sense) bits, of certain quantities (e.g. 160 bits or 20 bytes)
    # that are compatible with the AES algorithms.

    # From openssl's documentation:
    #
    #   RAND_bytes() puts num cryptographically strong pseudo-random
    #   bytes into buf. An error occurs if the PRNG has not been
    #   seeded with enough randomness to ensure an unpredictable byte
    #   sequence.
    #
    # openssl's RAND_bytes(num) function is available in Python as
    # M2Crypto.Rand.rand_bytes(num).

    return b16encode(rand_bytes(n))


def get_unique_device_name_for_user(name, user):
    """Returns the name with an extra number to make it unique if it exists in
       existing_names
    """
    original_name = name
    counter = 1
    existing_names = [device.name for device in user.devices.all()]
    while name in existing_names:
        name = '%s (%d)' % (original_name, counter)
        counter += 1
    return name


@sso_login_required
@require_twofactor_enabled
@allow_only('GET')
def device_list(request):
    paper_renewals = list(request.user.paper_devices_needing_renewal)
    context = context_with_current_section(
        request, device_addition_path=reverse(DEVICE_ADDITION),
        devices=request.user.devices.all(),
        need_backup_device_warning=request.user.need_backup_device_warning,
        paper_devices_needing_renewal=paper_renewals
    )
    return render(request, 'device/list.html', context)


@sso_login_required
@require_twofactor_enabled
@allow_only('GET', 'POST')
def device_addition(request):
    if (request.user.has_twofactor_devices() and
            not twofactor.is_upgraded(request, must_be_fresh=True)):
        return redirect_to_login(request.get_full_path(), reverse('twofactor'))

    if request.method == 'GET':
        context = context_with_current_section(
            request, device_list_path=reverse(DEVICE_LIST))
        return render(request, 'device/types.html', context)

    device_type = request.POST.get('type')

    if device_type not in device_types.keys():
        return render(request, 'device/types.html',
                      context_with_current_section(request))

    if device_type == 'paper':
        return _device_addition_paper(request)
    return _device_addition_standard(request, device_type)


def _device_addition_paper(request):
    hex_key = generate_key(20)
    device_name = get_unique_device_name_for_user(device_types['paper'],
                                                  request.user)
    device = _create_device(request, device_name, hex_key, 0, 'paper')
    return HttpResponseSeeOther(reverse('device-print', args=(device.id,)))


def _device_addition_standard(request, device_type):
    error = None

    if 'hex_key' in request.POST:
        hex_key = request.POST.get('hex_key')
    else:
        # TODO: 20 bytes = 160 bits; this will change based on
        # device-type.
        hex_key = generate_key(20)

    if 'name' not in request.POST:
        initial_name = get_unique_device_name_for_user(
            device_types.get(device_type), request.user)
        form = HOTPDeviceForm(initial={'name': initial_name})
    else:
        form = HOTPDeviceForm(request.POST)
        if form.is_valid():
            device_name = get_unique_device_name_for_user(
                form.cleaned_data['name'], request.user)
            otp = form.cleaned_data['otp']
            otp_type = get_otp_type(otp)
            accepted, new_counter = accept_hotp(
                hex_key, otp, 0, otp_type, drift=settings.HOTP_DRIFT,
                backward_drift=settings.HOTP_BACKWARDS_DRIFT)
            if accepted:
                _create_device(request, device_name, hex_key,
                               new_counter, device_type)
                return HttpResponseSeeOther(reverse(DEVICE_LIST))

            # Otherwise, set the error flag and fall through...
            error = OTP_MATCH_ERROR

    # Google would base32-encode, yubi would hex-encode, etc.  There
    # might even be multiple formats displayed simultaneously.
    formatted_key = re.sub('(.{4})', r'\1 ', hex_key).strip()

    context = context_with_current_section(
        request,
        device_list_path=reverse(DEVICE_LIST),
        type=device_type,
        ident="/".join([settings.TWOFACTOR_SERVICE_IDENT,
                        request.user.preferredemail.email]),
        hex_key=hex_key,
        form=form,
        formatted_key=formatted_key,
        error=error,
    )
    return render(request, 'device/addition-%s.html' % device_type, context)


def _create_device(request, device_name, hex_key, counter, device_type):
    device = AuthenticationDevice.objects.create(
        account=request.user,
        name=device_name,
        key=hex_key,
        counter=counter,
        device_type=device_type
    )
    twofactor.login(request)
    messages.success(request,
                     DEVICE_ADDED.format(name=device_name), 'temporary')
    return device


@sso_login_required(require_twofactor=True, require_twofactor_freshness=True)
@require_twofactor_enabled
@allow_only('GET')
def device_print(request, device_id):
    device = get_object_or_404(
        AuthenticationDevice, id=device_id, account=request.user,
        device_type='paper')

    details = _codes_for_position(device)
    remaining_codes = settings.TWOFACTOR_PAPER_CODES - details.position
    generation_enabled = (
        remaining_codes <= settings.TWOFACTOR_PAPER_CODES_ALLOW_GENERATION)

    if generation_enabled:
        messages.warning(request, DEVICE_GENERATION_WARNING)

    context = context_with_current_section(
        request,
        codes=details.codes,
        counter=details.position,
        device_id=device.id,
        generation_enabled=generation_enabled,
    )
    return TemplateResponse(request, 'device/print-codes.html', context)


def _codes_for_position(device, next_page=False):
    # use integer division to round the "window" boundaries
    page_size = settings.TWOFACTOR_PAPER_CODES
    page, page_position = divmod(device.counter, page_size)
    if next_page:
        page += 1
    page_start = page * page_size
    codes = [hotp(device.key, i, 'dec6')
             for i in range(page_start, page_start + page_size)]
    return CodePageDetails(codes, page, page_start, page_position)


@sso_login_required(require_twofactor=True, require_twofactor_freshness=True)
@require_twofactor_enabled
@allow_only('GET', 'POST')
def device_generate(request, device_id):
    device = get_object_or_404(
        AuthenticationDevice, id=device_id, account=request.user,
        device_type='paper')

    # find the next page of codes
    details = _codes_for_position(device, next_page=True)

    if request.method == 'GET':
        context = context_with_current_section(
            request,
            codes=details.codes,
            device_id=device.id,
        )
        return TemplateResponse(request, 'device/generate-codes.html', context)

    device.counter = details.start
    device.save()
    return HttpResponseRedirect(reverse('device-print', args=(device.id,)))


@sso_login_required(require_twofactor=True, require_twofactor_freshness=True)
@require_twofactor_enabled
@allow_only('GET', 'POST')
def device_removal(request, device_id):
    device = get_object_or_404(
        AuthenticationDevice, id=device_id, account=request.user)

    if request.method != 'POST':
        context = context_with_current_section(
            request, device_list_path=reverse(DEVICE_LIST), name=device.name)
        return render(request, 'device/removal.html', context)

    device.delete()

    # We should probably send an e-mail to the user stating which
    # device was removed. As a security measure, this would be much
    # stronger if bugs #784813, #784817, and #784818 were done.

    if not request.user.has_twofactor_devices():
        request.user.twofactor_required = False
        request.user.save()
        twofactor.logout(request)

    messages.success(request, DEVICE_DELETED.format(name=device.name))
    return HttpResponseSeeOther('/device-list')


class DeviceRenameView(View):
    def get(self, request, device_id):
        device = get_object_or_404(
            AuthenticationDevice, id=device_id, account=request.user)
        form = DeviceRenameForm({'name': device.name})
        context = context_with_current_section(
            request, device_list_path=reverse(DEVICE_LIST), form=form)
        return render(request, 'device/rename.html', context)

    def post(self, request, device_id):
        device = get_object_or_404(
            AuthenticationDevice, id=device_id, account=request.user)
        form = DeviceRenameForm(request.POST)
        if form.is_valid():
            original_name = device.name
            device.name = form.cleaned_data['name']
            device.save()
            messages.success(request,
                             DEVICE_RENAMED.format(original=original_name,
                                                   renamed=device.name))
            return HttpResponseRedirect(reverse(DEVICE_LIST))

        context = context_with_current_section(
            request, device_list_path=reverse(DEVICE_LIST), form=form)
        return render(request, 'device/rename.html', context)


device_rename = sso_login_required(
    require_twofactor=True,
    require_twofactor_freshness=True)(DeviceRenameView.as_view())


@allow_only('GET')
def device_help(request):
    if gargoyle.is_active('CAN_VIEW_SUPPORT_PHONE', request.user):
        support_phone = settings.SUPPORT_PHONE
    else:
        support_phone = ''

    context = {'support_phone': support_phone}
    return render(request, 'device/device-help.html', context)
