# Copyright 2016 Canonical Ltd.  This software is licensed under
# the GNU Affero General Public License version 3 (see the file
# LICENSE).

"""Build GPG-related error messages for user consumption."""

from django.template.loader import render_to_string
from django.utils.html import escape
from django.utils.safestring import mark_safe
from django.utils.translation import ugettext_lazy as _

from webui.utils import mark_safe_lazy


def bad_fingerprint_format():
    return render_to_string('gpg_messages/bad_fingerprint_format.html')


def key_already_imported(fingerprint):
    return render_to_string(
        'gpg_messages/key_already_imported.html',
        dict(fingerprint=display_fingerprint(fingerprint)))


def unknown_key():
    return render_to_string('gpg_messages/unknown_key.html')


def key_expired(fingerprint):
    return render_to_string(
        'gpg_messages/key_expired.html',
        dict(fingerprint=display_fingerprint(fingerprint)))


def key_revoked(fingerprint):
    return render_to_string(
        'gpg_messages/key_revoked.html',
        dict(fingerprint=display_fingerprint(fingerprint)))


def signing_key_mismatch(fingerprint):
    return render_to_string(
        'gpg_messages/signing_key_mismatch.html',
        dict(fingerprint=display_fingerprint(fingerprint)))


def validation_phrase_mismatch():
    return render_to_string('gpg_messages/validation_phrase_mismatch.html')


def key_added_to_account(fingerprint):
    return render_to_string(
        'gpg_messages/key_added_to_account.html',
        dict(fingerprint=display_fingerprint(fingerprint)))


def missing_signed_text():
    return render_to_string('gpg_messages/missing_signed_text.html')


def pending_tokens_cancelled(fingerprints):
    return render_to_string(
        'gpg_messages/pending_tokens_cancelled.html',
        dict(fingerprints=[display_fingerprint(f) for f in fingerprints]))


def keys_disabled(fingerprints):
    return render_to_string(
        'gpg_messages/keys_disabled.html',
        dict(fingerprints=[display_fingerprint(f) for f in fingerprints]))


def keys_reenabled(fingerprints, email):
    return render_to_string(
        'gpg_messages/keys_reenabled.html',
        dict(
            fingerprints=[display_fingerprint(f) for f in fingerprints],
            preferred_email=email))


def display_keyid_long(keyid, fingerprint):
    return mark_safe(
        '<a href="#" data-tooltip="{tooltip}: {fingerprint}">{id}</a>'.format(
            id=escape(keyid),
            fingerprint=display_fingerprint(fingerprint, markup=False),
            tooltip=_('GPG Fingerprint')))


def display_fingerprint(fingerprint, markup=True):
    """Correctly format 'fingerprint' for the user.

    This function returns a version of 'fingerprint' wrapped in '<code>'
    tags and with exactly the same spacing as the gpg command line utility
    produces. This ensures consistency of presentation across the UI.

    The 'fingerprint' parameter is assumed to be untrusted, and is escaped
    before any processing is done. The returned value is marked as safe.

    :raises ValueError: if 'fingerprint' is neither a v4 nor a v3 gpg
        fingerprint.
    """
    fingerprint = escape(fingerprint.replace(' ', ''))
    if len(fingerprint) == 40:
        # v4 fingerprint:
        groupsize = 4
    elif len(fingerprint) == 32:
        # v3 fingerprint:
        groupsize = 2
    else:
        raise ValueError("Invalid GPG fingerprint: %r" % fingerprint)

    groups = [
        fingerprint[s:s+groupsize]
        for s in range(0, len(fingerprint), groupsize)]
    first_half = ' '.join(groups[:len(groups)/2])
    second_half = ' '.join(groups[len(groups)/2:])
    if not markup:
        return '%s %s' % (first_half, second_half)
    else:
        return mark_safe_lazy(
            '<code>%s&nbsp;&nbsp;%s</code>' % (first_half, second_half))
